
#define NUM_ATOM_MAP           9
#define NUM_SPLIT_MAP          9
#define NUM_SPLIT_RESIDUE_LIST 50

typedef struct {
       char old_atom[5];
       char new_atom[5];
} _ATOM_MAP;

typedef struct {
        char new_residue[5];
        int  num_atom;
        _ATOM_MAP atom_map[NUM_ATOM_MAP];
} _SPLIT_MAP;

typedef struct {
       char old_residue[5];
       int  num_residue;
       _SPLIT_MAP split_map[NUM_SPLIT_MAP];
} _SPLIT_RESIDUE_LIST;

static _SPLIT_RESIDUE_LIST _split_residue_list[NUM_SPLIT_RESIDUE_LIST] = {
       { "1CU", 2,
          {
             { "CU",  1,
                {
                   { "CU",  "CU"  }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             }
          }
       },
       { "2OF", 3,
          {
             { "FE",  1,
                {
                   { "FE",  "FE",  }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             }
          }
       },
       { "3OF", 3,
          {
             { "FE",  1,
                {
                   { "FE",  "FE",  }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             }
          }
       },
       { "543", 8,
          {
             { "CA",  1,
                {
                   { "CA",   "CA"   }
                }
             },
             { "EOH", 9,
                {
                   { "C1",   "C1"   },
                   { "C2",   "C2"   },
                   { "O",    "O"    },
                   { "1H1",  "1H1"  },
                   { "2H1",  "2H1"  },
                   { "1H2",  "1H2"  },
                   { "2H2",  "2H2"  },
                   { "3H2",  "3H2"  },
                   { "HO",   "HO"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O3",   "O"   },
                   { "1HO3", "1H"  },
                   { "2HO3", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O4",   "O"   },
                   { "1HO4", "1H"  },
                   { "2HO4", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O5",   "O"   },
                   { "1HO5", "1H"  },
                   { "2HO5", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O6",   "O"   },
                   { "1HO6", "1H"  },
                   { "2HO6", "2H"  }
                }
             }
          }
       },
       { "CD1", 2,
          {
             { "CD",  1,
                {
                   { "CD",   "CD"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             }
          }
       },
       { "CD3", 4,
          {
             { "CD",  1,
                {
                   { "CD",   "CD"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O3",   "O"   },
                   { "1HO3", "1H"  },
                   { "2HO3", "2H"  }
                }
             }
          }
       },
       { "CD5", 6,
          {
             { "CD",  1,
                {
                   { "CD",   "CD"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O3",   "O"   },
                   { "1HO3", "1H"  },
                   { "2HO3", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O4",   "O"   },
                   { "1HO4", "1H"  },
                   { "2HO4", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O5",   "O"   },
                   { "1HO5", "1H"  },
                   { "2HO5", "2H"  }
                }
             }
          }
       },
       { "CO5", 6,
          {
             { "CO",  1,
                {
                   { "CO",   "CO"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O3",   "O"   },
                   { "1HO3", "1H"  },
                   { "2HO3", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O4",   "O"   },
                   { "1HO4", "1H"  },
                   { "2HO4", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O5",   "O"   },
                   { "1HO5", "1H"  },
                   { "2HO5", "2H"  }
                }
             }
          }
       },
       { "KO4", 5,
          {
             { "K",   1,
                {
                   { "K",    "K"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1H1",  "1H"  },
                   { "2H1",  "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1H2",  "1H"  },
                   { "2H2",  "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O3",   "O"   },
                   { "1H3",  "1H"  },
                   { "2H3",  "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O4",   "O"   },
                   { "1H4",  "1H"  },
                   { "2H4",  "2H"  }
                }
             }
          }
       },
       { "MH3", 2,
          {
             { "MN",  1,
                {
                   { "MN",   "MN"   }
                }
             },
             { "HOH", 2,
                {
                   { "O1",   "O"   },
                   { "H2",   "1H"  }
                }
             }
          }
       },
       { "MN5", 6,
          {
             { "MN",  1,
                {
                   { "MN",   "MN"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O3",   "O"   },
                   { "1HO3", "1H"  },
                   { "2HO3", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O4",   "O"   },
                   { "1HO4", "1H"  },
                   { "2HO4", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O5",   "O"   },
                   { "1HO5", "1H"  },
                   { "2HO5", "2H"  }
                }
             }
          }
       },
       { "MN6", 7,
          {
             { "MN",  1,
                {
                   { "MN",   "MN"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O3",   "O"   },
                   { "1HO3", "1H"  },
                   { "2HO3", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O4",   "O"   },
                   { "1HO4", "1H"  },
                   { "2HO4", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O5",   "O"   },
                   { "1HO5", "1H"  },
                   { "2HO5", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O6",   "O"   },
                   { "1HO6", "1H"  },
                   { "2HO6", "2H"  }
                }
             }
          }
       },
       { "MO1", 2,
          {
             { "MG",  1,
                {
                   { "MG",   "MG"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             }
          }
       },
       { "MO2", 3,
          {
             { "MG",  1,
                {
                   { "MG",   "MG"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             }
          }
       },
       { "MO3", 4,
          {
             { "MG",  1,
                {
                   { "MG",   "MG"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O3",   "O"   },
                   { "1HO3", "1H"  },
                   { "2HO3", "2H"  }
                }
             }
          }
       },
       { "MO4", 5,
          {
             { "MG",  1,
                {
                   { "MG",   "MG"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O3",   "O"   },
                   { "1HO3", "1H"  },
                   { "2HO3", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O4",   "O"   },
                   { "1HO4", "1H"  },
                   { "2HO4", "2H"  }
                }
             }
          }
       },
       { "MO5", 6,
          {
             { "MG",  1,
                {
                   { "MG",   "MG"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O3",   "O"   },
                   { "1HO3", "1H"  },
                   { "2HO3", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O4",   "O"   },
                   { "1HO4", "1H"  },
                   { "2HO4", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O5",   "O"   },
                   { "1HO5", "1H"  },
                   { "2HO5", "2H"  }
                }
             }
          }
       },
       { "MO6", 7,
          {
             { "MG",  1,
                {
                   { "MG",   "MG"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O3",   "O"   },
                   { "1HO3", "1H"  },
                   { "2HO3", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O4",   "O"   },
                   { "1HO4", "1H"  },
                   { "2HO4", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O5",   "O"   },
                   { "1HO5", "1H"  },
                   { "2HO5", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O6",   "O"   },
                   { "1HO6", "1H"  },
                   { "2HO6", "2H"  }
                }
             }
          }
       },
       { "MW1", 2,
          {
             { "MN",  1,
                {
                   { "MN",   "MN"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             }
          }
       },
       { "MW2", 3, 
          {
             { "MN",  1,
                {
                   { "MN",   "MN"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             }
          }
       },
       { "MW3", 4,
          {
             { "MN",  1,
                {
                   { "MN",   "MN"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O3",   "O"   },
                   { "1HO3", "1H"  },
                   { "2HO3", "2H"  }
                }
             }
          }
       },
       { "NA2", 3,
          {
             { "NA",  1,
                {
                   { "NA",   "NA"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             }
          }
       },
       { "NA5", 6,
          {
             { "NA",  1,
                {
                   { "NA",   "NA"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O3",   "O"   },
                   { "1HO3", "1H"  },
                   { "2HO3", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O4",   "O"   },
                   { "1HO4", "1H"  },
                   { "2HO4", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O5",   "O"   },
                   { "1HO5", "1H"  },
                   { "2HO5", "2H"  }
                }
             }
          }
       },
       { "NA6", 7,
          {
             { "NA",  1,
                {
                   { "NA",   "NA"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O3",   "O"   },
                   { "1HO3", "1H"  },
                   { "2HO3", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O4",   "O"   },
                   { "1HO4", "1H"  },
                   { "2HO4", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O5",   "O"   },
                   { "1HO5", "1H"  },
                   { "2HO5", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O6",   "O"   },
                   { "1HO6", "1H"  },
                   { "2HO6", "2H"  }
                }
             }
          }
       },
       { "NAO", 2,
          {
             { "NA",  1,
                {
                   { "NA",   "NA"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             }
          }
       },
       { "NAW", 4,
          {
             { "NA",  1,
                {
                   { "NA",   "NA"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O3",   "O"   },
                   { "1HO3", "1H"  },
                   { "2HO3", "2H"  }
                }
             }
          }
       },
       { "NI1", 2,
          {
             { "NI",  1,
                {
                   { "NI",   "NI"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             }
          }
       },
       { "NI2", 3,
          {
             { "NI",  1,
                {
                   { "NI",   "NI"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "3HO1", "1H"  },
                   { "4HO1", "2H"  }
                }
             }
          }
       },
       { "NI3", 4,
          {
             { "NI",  1,
                {
                   { "NI",   "NI"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O3",   "O"   },
                   { "1HO3", "1H"  },
                   { "2HO3", "2H"  }
                }
             }
          }
       },
       { "NIK", 6,
          {
             { "NI",   1,
                {
                   { "NI",   "NI"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1H1",  "1H"  },
                   { "2H1",  "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1H2",  "1H"  },
                   { "2H2",  "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O3",   "O"   },
                   { "1H3",  "1H"  },
                   { "2H3",  "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O4",   "O"   },
                   { "1H4",  "1H"  },
                   { "2H4",  "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O5",   "O"   },
                   { "1H5",  "1H"  },
                   { "2H5",  "2H"  }
                }
             }
          }
       },
       { "O4M", 5,
          {
             { "MN",  1,
                {
                   { "MN",   "MN"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O3",   "O"   },
                   { "1HO3", "1H"  },
                   { "2HO3", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O4",   "O"   },
                   { "1HO4", "1H"  },
                   { "2HO4", "2H"  }
                }
             }
          }
       },
       { "OC1", 2,
          {
             { "CA",  1,
                {
                   { "CA",   "CA"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             }
          }
       },
       { "OC2", 3,
          {
             { "CA",  1,
                {
                   { "CA",   "CA"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             }
          }
       },
       { "OC3", 4,
          {
             { "CA",  1,
                {
                   { "CA",   "CA"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O3",   "O"   },
                   { "1HO3", "1H"  },
                   { "2HO3", "2H"  }
                }
             }
          }
       },
       { "OC4", 5,
          {
             { "CA",  1,
                {
                   { "CA",   "CA"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O3",   "O"   },
                   { "1HO3", "1H"  },
                   { "2HO3", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O4",   "O"   },
                   { "1HO4", "1H"  },
                   { "2HO4", "2H"  }
                }
             }
          }
       },
       { "OC5", 6,
          {
             { "CA",  1,
                {
                   { "CA",   "CA"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O3",   "O"   },
                   { "1HO3", "1H"  },
                   { "2HO3", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O4",   "O"   },
                   { "1HO4", "1H"  },
                   { "2HO4", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O5",   "O"   },
                   { "1HO5", "1H"  },
                   { "2HO5", "2H"  }
                }
             }
          }
       },
       { "OC6", 7,
          {
             { "CA",  1,
                {
                   { "CA",   "CA"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O3",   "O"   },
                   { "1HO3", "1H"  },
                   { "2HO3", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O4",   "O"   },
                   { "1HO4", "1H"  },
                   { "2HO4", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O5",   "O"   },
                   { "1HO5", "1H"  },
                   { "2HO5", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O6",   "O"   },
                   { "1HO6", "1H"  },
                   { "2HO6", "2H"  }
                }
             }
          }
       },
       { "OC7", 8,
          {
             { "CA",  1,
                {
                   { "CA",   "CA"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O3",   "O"   },
                   { "1HO3", "1H"  },
                   { "2HO3", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O4",   "O"   },
                   { "1HO4", "1H"  },
                   { "2HO4", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O5",   "O"   },
                   { "1HO5", "1H"  },
                   { "2HO5", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O6",   "O"   },
                   { "1HO6", "1H"  },
                   { "2HO6", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O7",   "O"   },
                   { "1HO7", "1H"  },
                   { "2HO7", "2H"  }
                }
             }
          }
       },
       { "OC8", 9,
          {
             { "CA",  1,
                {
                   { "CA",   "CA"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O3",   "O"   },
                   { "1HO3", "1H"  },
                   { "2HO3", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O4",   "O"   },
                   { "1HO4", "1H"  },
                   { "2HO4", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O5",   "O"   },
                   { "1HO5", "1H"  },
                   { "2HO5", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O6",   "O"   },
                   { "1HO6", "1H"  },
                   { "2HO6", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O7",   "O"   },
                   { "1HO7", "1H"  },
                   { "2HO7", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O8",   "O"   },
                   { "1HO8", "1H"  },
                   { "2HO8", "2H"  }
                }
             }
          }
       },
       { "OCL", 2,
          {
             { "CO",  1,
                {
                   { "CO",   "CO"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             }
          }
       },
       { "OCM", 4,
          {
             { "CO",  1,
                {
                   { "CO",   "CO"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O3",   "O"   },
                   { "1HO3", "1H"  },
                   { "2HO3", "2H"  }
                }
             }
          }
       },
       { "OCN", 3,
          {
             { "CO",  1,
                {
                   { "CO",   "CO"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             }
          }
       },
       { "OCO", 7,
          {
             { "CO",  1,
                {
                   { "CO",   "CO"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O3",   "O"   },
                   { "1HO3", "1H"  },
                   { "2HO3", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O4",   "O"   },
                   { "1HO4", "1H"  },
                   { "2HO4", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O5",   "O"   },
                   { "1HO5", "1H"  },
                   { "2HO5", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O6",   "O"   },
                   { "1HO6", "1H"  },
                   { "2HO6", "2H"  }
                }
             }
          }
       },
       { "OF1", 2,
          {
             { "FE",  1,
                {
                   { "FE",  "FE",  }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             }
          }
       },
       { "OF3", 2,
          {
             { "FE",  1,
                {
                   { "FE",  "FE",  }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             }
          }
       },
       { "YH", 2,
          {
             { "Y",   1,
                {
                   { "Y",    "Y"   }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "H1",   "1H"  },
                   { "H2",   "2H"  }
                }
             }
          }
       },
       { "ZH3", 4,
          {
             { "ZN",  1,
                {
                   { "ZN",  "ZN"   }
                }
             },
             { "HOH", 2,
                {
                   { "OH1",  "O"   },
                   { "HH1",  "1H"  }
                }
             },
             { "HOH", 2,
                {
                   { "OH2",  "O"   },
                   { "HH2",  "1H"  }
                }
             },
             { "HOH", 2,
                {
                   { "OH3",  "O"   },
                   { "HH3",  "1H"  }
                }
             }
          }
       },
       { "ZN3", 2,
          {
             { "ZN",  1,
                {
                   { "ZN",  "ZN",  }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             }
          }
       },
       { "ZNO", 3,
          {
             { "ZN",  1,
                {
                   { "ZN",  "ZN",  }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "3HO1", "1H"  },
                   { "4HO1", "2H"  }
                }
             }
          }
       },
       { "ZO3", 4,
          {
             { "ZN",  1,
                {
                   { "ZN",  "ZN",  }
                }
             },
             { "HOH", 3,
                {
                   { "O1",   "O"   },
                   { "1HO1", "1H"  },
                   { "2HO1", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O2",   "O"   },
                   { "1HO2", "1H"  },
                   { "2HO2", "2H"  }
                }
             },
             { "HOH", 3,
                {
                   { "O3",   "O"   },
                   { "1HO3", "1H"  },
                   { "2HO3", "2H"  }
                }
             }
          }
       }
};
