#include <stdio.h>
#include <string.h>
#include <math.h>
#include <string>
#include <map>
#include <set>
#include <vector>

#include "_fileobj_ioutil.h"
#include "_maxit.h"
#include "assembly_parser_util.h"
#include "bond_angle_torsion.h"
#include "erf_util.h"
#include "util.h"

#define NUM_MER  20
#define NUM_UPDATE_CATEGORY 2

typedef struct {
       const char *temp_key;
       const char *label_asym_id;
       const char *label_comp_id;
       const char *label_seq_id;
       const char *auth_asym_id;
       const char *auth_comp_id;
       const char *auth_seq_id;
       const char *auth_ins_code;
} _UPDATE_ITEMS;

typedef struct {
       const char *category;
       int NumRecord;
       _UPDATE_ITEMS items[4];
} _UPDATE_CATEGORY;

const static _UPDATE_CATEGORY _update_category[NUM_UPDATE_CATEGORY] = {
      { "ndb_struct_na_base_pair", 2, {
        { "i_temp_index", "i_label_asym_id", "i_label_comp_id", "i_label_seq_id",
          "i_auth_asym_id", "i_label_comp_id", "i_auth_seq_id", "i_PDB_ins_code" },
        { "j_temp_index", "j_label_asym_id", "j_label_comp_id", "j_label_seq_id",
          "j_auth_asym_id", "j_label_comp_id", "j_auth_seq_id", "j_PDB_ins_code" } }
      },
      { "ndb_struct_na_base_pair_step", 4, {
        { "i_temp_index_1", "i_label_asym_id_1", "i_label_comp_id_1", "i_label_seq_id_1",
          "i_auth_asym_id_1", "i_label_comp_id_1", "i_auth_seq_id_1", "i_PDB_ins_code_1" },
        { "j_temp_index_1", "j_label_asym_id_1", "j_label_comp_id_1", "j_label_seq_id_1",
          "j_auth_asym_id_1", "j_label_comp_id_1", "j_auth_seq_id_1", "j_PDB_ins_code_1" },
        { "i_temp_index_2", "i_label_asym_id_2", "i_label_comp_id_2", "i_label_seq_id_2",
          "i_auth_asym_id_2", "i_label_comp_id_2", "i_auth_seq_id_2", "i_PDB_ins_code_2" },
        { "j_temp_index_2", "j_label_asym_id_2", "j_label_comp_id_2", "j_label_seq_id_2",
          "j_auth_asym_id_2", "j_label_comp_id_2", "j_auth_seq_id_2", "j_PDB_ins_code_2" } }
      }
};

const static char *_num_mers[NUM_MER+1] = {
       "UNKNOWN",
       "MONOMERIC",
       "DIMERIC",
       "TRIMERIC",
       "TETRAMERIC",
       "PENTAMERIC",
       "HEXAMERIC",
       "HEPTAMERIC",
       "OCTAMERIC",
       "NONAMERIC",
       "DECAMERIC",
       "UNDECAMERIC",
       "DODECAMERIC",
       "TRIDECAMERIC",
       "TETRADECAMERIC",
       "PENTADECAMERIC",
       "HEXADECAMERIC",
       "HEPTADECAMERIC",
       "OCTADECAMERIC",
       "NONADECAMERIC",
       "EICOSAMERIC"
};

typedef struct {
       CifString inst_ids;
       ReVarCifArray<CifString> oplist;
} _CHAIN_OPLIST;

typedef struct {
       CifString assembly_id;
       CifString type;
       CifString software;
       CifString absa;
       CifString ssa;
       CifString more;
       int total_chains;
       ReVarCifArray<_CHAIN_OPLIST> coplists;
} _ASSEMBLY;

typedef struct {
       CifString asym_id;
       CifString pdb_chnid;
       int       is_polymer;
} _ASYM_CHAINID_CORR;

typedef struct {
       CifString sym;
       CifString oper_id;
       CifString catchainids;
       ReVarCifArray<CifString> chainids;
} _SYM_CHAIN_CORR;

typedef struct {
       const char *remark;
       int index;
} _REMARK_630_TEMPLATE;

#define NUM_REMARK_630 10

const static _REMARK_630_TEMPLATE _remark_630_template[NUM_REMARK_630] = {
       { "MOLECULE TYPE: ", 1 },
       { "MOLECULE NAME: ", 2 },
       { "(M=MODEL NUMBER; RES=RESIDUE NAME; C=CHAIN IDENTIFIER;", 0 },
       { " SSSEQ=SEQUENCE NUMBER; I=INSERTION CODE.)", 0 },
       { "", 0 },
       { "  M RES C SSSEQI", 0 },
       { "SOURCE: ", 3 },
       { "TAXONOMY: ", 4 },
       { "SUBCOMP:", 0 },
       { "DETAILS: ", 5 }
};

typedef struct {
       ReVarCifArray<CifString> data;
       ReVarCifArray<CifString> subcomp;
       ReVarCifArray<ReVarCifArray<CifString> > instances;
} _REMARK_630;

static void get_coor_index(ISTable *t, _MOLECULE *molecule);
static void get_coor_index(ISTable *t, _MOLECULE *molecule, const _UPDATE_ITEMS& items);
static void get_first_and_last_residue_number(CifString &pdb_chnid, CifString &first_num,
              CifString &last_num, _MOLECULE *molecule);
static void get_coor_index1(CifString &pdb_chnid, const CifString &pdb_resnum,
               _MOLECULE *molecule, const int flag);
static int get_fingerprint(const CifString &pdb_chnid, const CifString &pdb_resnam,
      const CifString &pdb_resnum, const CifString &ins_code, _MOLECULE *molecule);
static void reverse_coor_index(ISTable *t, _MOLECULE *molecule);
static void reverse_coor_index(ISTable *t, _MOLECULE *molecule, const _UPDATE_ITEMS& items);
static void reverse_coor_index1(CifString &pdb_chnid, CifString &pdb_num,
              const CifString &Idx, _MOLECULE *molecule, const int flag);
static _RESIDUE *get_residue_from_fingerprint(const CifString &Idx, _MOLECULE *molecule);
static void get_assembly_list(CifFileObj *fobj, const char *block_id,
                 ReVarCifArray<_ASSEMBLY> &assembly_lists, const bool keep_original_count);
static void get_inhibitor_details(CifFileObj *fobj, const char *blockId,
                 const int entity_id, CifString &type, CifString &source,
                 CifString &s_details, CifString &o_details);
static COORD get_coordinate(_ATOM *atom, const NDBCELLCONST *cell, const char *SymOP);
static ISTable *pdbx_struct_assembly_Table();
static ISTable *pdbx_struct_assembly_gen_Table();
static ISTable *pdbx_struct_oper_list_Table();
static void update_remediation_atom_site_mapping_read(CifFileObj *fobj,
                       const char *block_id);
static void update_remediation_atom_site_mapping_write(CifFileObj *fobj,
                       const char *block_id);
static void get_prd_info(CifFileObj *fobj, const char *blockId, std::map<std::string,
                         std::vector<std::string> >& prd_info);

void get_wordlist_from_string_separated_by_delimits(ReVarCifArray<CifString> &wordlist,
                  char *string_buff, const char *delimit);

void _MAXIT::ndb_to_pdb_get_remark_475_and_480()
{
       if (!(_experiment_type & EXPERIMENT_TYPE_XRAY)) return;

       ReVarCifArray<ReVarCifArray<CifString> > residue_lists, atom_lists;
       ReVarCifArray<CifString> list, atom_list;
       residue_lists.Clear();
       atom_lists.Clear();
       int Mol_ID = -1;
       for (int i = 0; i < _num_mol; i++) {
            if (_num_mol > 1) Mol_ID = _molecules[i]->Mol_ID();
            _CHAIN *chain = _molecules[i]->GetFirstChain();
            while (chain) {
                 if (chain->chain_type() == ATOMN_TOKEN ||
                     chain->chain_type() == ATOMP_TOKEN) {
                      _RESIDUE *residue = chain->GetFirstResidue();
                      while (residue) {
                           int ret = residue->is_zero_occupancy_residue(Mol_ID, atom_list);
                           if (ret) {
                                list.Clear();
                                for (int j = 0; j < 5; j++) list.Add(atom_list[j]);
                                 residue_lists.Add(list);
                           } else if (atom_list.Length() > 0) {
                                atom_lists.Add(atom_list);
                           }
                           residue = chain->GetNextResidue();
                      }
                 }
                 chain = _molecules[i]->GetNextChain();
            }
       }

       One_Card_Info_Format *newCard = NULL;
       char mol_id[4], chain_id[2], insert_code[2], buffer[80];
       if (residue_lists.Length() > 0) {
            ndb_to_pdb_get_general_remark(Num_Remark_475, Remark_475, 0, 0, 0, 0);
            for (unsigned int i = 0; i < residue_lists.Length(); i++) {
                 if (residue_lists[i][0] != "-1")
                      strcpy(mol_id, residue_lists[i][0].Text());
                 else strcpy(mol_id, "   ");
                 if (residue_lists[i][1] != "")
                      strcpy(chain_id, residue_lists[i][1].Text());
                 else strcpy(chain_id, " ");
                 if (residue_lists[i][4] != "")
                      strcpy(insert_code, residue_lists[i][4].Text());
                 else strcpy(insert_code, " ");
                 sprintf(buffer, "%3s %3s %c %5s%c", mol_id, residue_lists[i][2].Text(),
                           chain_id[0], residue_lists[i][3].Text(), insert_code[0]);
                 newCard = _CardInfo[REMARK_TOKEN-1].add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(475, 1);
                 newCard->AddElementToField(buffer, 2);
            }
       }

       if (atom_lists.Length() > 0) {
            char head[80];
            ndb_to_pdb_get_general_remark(Num_Remark_480, Remark_480, 0, 0, 0, 0);
            for (unsigned int i = 0; i < atom_lists.Length(); i++) {
                 if (atom_lists[i][0] != "-1")
                      strcpy(mol_id, atom_lists[i][0].Text());
                 else strcpy(mol_id, "   ");
                 if (atom_lists[i][1] != "")
                      strcpy(chain_id, atom_lists[i][1].Text());
                 else strcpy(chain_id, " ");
                 if (atom_lists[i][4] != "")
                      strcpy(insert_code, atom_lists[i][4].Text());
                 else strcpy(insert_code, " ");
                 newCard = _CardInfo[REMARK_TOKEN-1].add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(480, 1);
                 sprintf(head, "%3s %3s %c %4s%c  ", mol_id, atom_lists[i][2].Text(),
                           chain_id[0], atom_lists[i][3].Text(), insert_code[0]);
                 newCard->AddElementToField(head, 2);
                 int num = 0;
                 for (unsigned int j = 5; j < atom_lists[i].Length(); j++) {
                      if (num == 7) {
                           newCard = _CardInfo[REMARK_TOKEN-1].add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                           newCard->AddElementToField(480, 1);
                           sprintf(head, "%3s %3s %c %4s%c  ", mol_id, atom_lists[i][2].Text(),
                                     chain_id[0], atom_lists[i][3].Text(), insert_code[0]);
                           newCard->AddElementToField(head, 2);
                           num = 0;
                      }
                      sprintf(buffer, "%-5s", atom_lists[i][j].Text());
                      newCard->AttachToField(buffer, 2);
                      num++;
                 }
            }
       }
}

void _MAXIT::ndb_to_pdb_get_remark_610()
{
       ReVarCifArray<ReVarCifArray<CifString> > residue_lists;
       ReVarCifArray<CifString> list;
       residue_lists.Clear();
       int Mol_ID = -1;
       char buffer[80];
       CifString cs;

       finding_missing_and_extra_atoms(1);

       for (int i = 0; i < _num_mol; i++) {
            if (_num_mol > 1) Mol_ID = _molecules[i]->Mol_ID();
            _CHAIN *chain = _molecules[i]->GetFirstChain();
            while (chain) {
                 if (chain->chain_type() != ATOMN_TOKEN &&
                     chain->chain_type() != ATOMP_TOKEN) {
                      _RESIDUE *residue = chain->GetFirstResidue();
                      while (residue) {
                           if (strcmp(residue->missing().Text(), "")) {
                                _ATOM *atom = residue->GetFirstAtom();
                                list.Clear();
                                sprintf(buffer, "%d", Mol_ID);
                                cs.Copy(buffer);
                                list.Add(cs);
                                cs.Copy(atom->pdb_chnid());
                                list.Add(cs);
                                cs.Copy(atom->pdb_resnam());
                                list.Add(cs);
                                sprintf(buffer, "%d", atom->pdb_resnum());
                                cs.Copy(buffer);
                                list.Add(cs);
                                cs.Copy(atom->ins_code());
                                list.Add(cs);
                                residue_lists.Add(list);
                           }
                           residue = chain->GetNextResidue();
                      }
                 }
                 chain = _molecules[i]->GetNextChain();
            }
       }

       if (residue_lists.Length() == 0) return;

       ndb_to_pdb_get_general_remark(Num_Remark_610, Remark_610, 0, 0, 0, 0);

       char mol_id[4], chain_id[2], insert_code[2];
       for (unsigned int i2 = 0; i2 < residue_lists.Length(); i2++) {
            if (residue_lists[i2][0] != "-1")
                 strcpy(mol_id, residue_lists[i2][0].Text());
            else strcpy(mol_id, "   ");
            if (residue_lists[i2][1] != "")
                 strcpy(chain_id, residue_lists[i2][1].Text());
            else strcpy(chain_id, " ");
            if (residue_lists[i2][4] != "")
                 strcpy(insert_code, residue_lists[i2][4].Text());
            else strcpy(insert_code, " ");
            sprintf(buffer, "%3s %3s %c%5s%c", mol_id, residue_lists[i2][2].Text(),
                      chain_id[0], residue_lists[i2][3].Text(), insert_code[0]);
            One_Card_Info_Format *newCard = _CardInfo[REMARK_TOKEN-1].
                      add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(610, 1);
            newCard->AddElementToField(buffer, 2);
       }
}

void _MAXIT::ndb_to_pdb_get_remark_615()
{
       if (!(_experiment_type & EXPERIMENT_TYPE_XRAY)) return;

       ReVarCifArray<ReVarCifArray<CifString> > residue_lists;
       ReVarCifArray<CifString> list, atom_list;
       residue_lists.Clear();
       int Mol_ID = -1;
       for (int i = 0; i < _num_mol; i++) {
            if (_num_mol > 1) Mol_ID = _molecules[i]->Mol_ID();
            _CHAIN *chain = _molecules[i]->GetFirstChain();
            while (chain) {
                 if (chain->chain_type() != ATOMN_TOKEN &&
                     chain->chain_type() != ATOMP_TOKEN) {
                      _RESIDUE *residue = chain->GetFirstResidue();
                      while (residue) {
                           int ret = residue->is_zero_occupancy_residue(Mol_ID, atom_list);
                           if (ret || atom_list.Length() > 0)  {
                                list.Clear();
                                for (int j = 0; j < 5; j++) list.Add(atom_list[j]);
                                 residue_lists.Add(list);
                           }
                           residue = chain->GetNextResidue();
                      }
                 }
                 chain = _molecules[i]->GetNextChain();
            }
       }

       if (residue_lists.Length() == 0) return;

       ndb_to_pdb_get_general_remark(Num_Remark_615, Remark_615, 0, 0, 0, 0);

       char mol_id[4], chain_id[2], insert_code[2], buffer[80];
       for (unsigned int i2 = 0; i2 < residue_lists.Length(); i2++) {
            if (residue_lists[i2][0] != "-1")
                 strcpy(mol_id, residue_lists[i2][0].Text());
            else strcpy(mol_id, "   ");
            if (residue_lists[i2][1] != "")
                 strcpy(chain_id, residue_lists[i2][1].Text());
            else strcpy(chain_id, " ");
            if (residue_lists[i2][4] != "")
                 strcpy(insert_code, residue_lists[i2][4].Text());
            else strcpy(insert_code, " ");
            sprintf(buffer, "%3s %3s %c %5s%c", mol_id, residue_lists[i2][2].Text(),
                      chain_id[0], residue_lists[i2][3].Text(), insert_code[0]);
             One_Card_Info_Format *newCard = _CardInfo[REMARK_TOKEN-1].
                  add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(615, 1);
            newCard->AddElementToField(buffer, 2);
       }
}

int _MAXIT::get_atom_pair(_ATOM &metal_atom, _ATOM &polar_atom, _LINK *link,
              ReVarCifArray<_ATOM>& covalent_atoms)
{
       covalent_atoms.Clear();

       if (_num_mol == 0 || _molecules == NULL) return 0;

       _CHAIN *ch1 = _molecules[link->mol_index]->GetIndexChain(link->FstIndex);
       _CHAIN *ch2 = _molecules[link->mol_index]->GetIndexChain(link->SndIndex);
       if (!ch1 || !ch2) return 0;
       _FIELD *Init = ch1->SeqResIndex(link->FstRes);
       _FIELD *End  = ch2->SeqResIndex(link->SndRes);
       if (!Init || !End) return 0;
       _RESIDUE *beg = Init->Residue;
       _RESIDUE *end = End->Residue;
       if (!beg || !end) return 0;;

       int found = 0, num_a, num_b;
       _ATOM **atom_a = beg->find_atom(link->FstAtom, num_a);
       _ATOM **atom_b = end->find_atom(link->SndAtom, num_b);
       if (atom_a && atom_b) {
            _ATOM *atom1 = NULL;
            _ATOM *atom2 = NULL;
            for (int i = 0; i < num_a; i++) {
                 if (!strcmp(atom_a[i]->alt_loc(), link->altLoc1) &&
                     !strcmp(atom_a[i]->restype(), link->FstName)) {
                      atom1 = atom_a[i];
                      break;
                 }
            }
            for (int i2 = 0; i2 < num_b; i2++) {
                 if (!strcmp(atom_b[i2]->alt_loc(), link->altLoc2) &&
                     !strcmp(atom_b[i2]->restype(), link->SndName)) {
                      atom2 = atom_b[i2];
                      break;
                 }
            }
            if (atom1 && atom2) {
                 found = 1;
                 COORD coord1 = get_coordinate(atom1, _cell, link->SymOP_1);
                 COORD coord2 = get_coordinate(atom2, _cell, link->SymOP_2);
                 _RESIDUE *res = NULL;
                 char AtomName[10], SymOP[10];
                 if (find_metal_flag(atom1->atom_type()) == 2) {
                      metal_atom = *atom1;
                      metal_atom.set_orig(coord1);
                      polar_atom = *atom2;
                      polar_atom.set_orig(coord2);
                      res = beg;
                      strcpy(AtomName, link->FstAtom);
                      strcpy(SymOP, link->SymOP_1);
                 } else {
                      metal_atom = *atom2;
                      metal_atom.set_orig(coord2);
                      polar_atom = *atom1;
                      polar_atom.set_orig(coord1);
                      res = end;
                      strcpy(AtomName, link->SndAtom);
                      strcpy(SymOP, link->SymOP_2);
                 }
                 ConnectFormat *drug = find_drug(res->ResName());
                 if (drug) {
                      _SIndex *index = drug->find_atom(AtomName);
                      if (index) {
                           for (int i = 0; i < drug->atoms(index->pos())->nconect(); ++i) {
                                _ATOM **atoms = res->find_atom(drug->atoms(index->pos())
                                                  ->connectto(i), num_a);
                                for (int j = 0; j < num_a; j++) {
                                     if (!strcmp(metal_atom.alt_loc(), "") ||
                                         !strcmp(atoms[j]->alt_loc(), "") ||
                                         !strcmp(metal_atom.alt_loc(),
                                                 atoms[j]->alt_loc())) {
                                          coord1 = get_coordinate(atoms[j], _cell, SymOP);
                                          _ATOM atom = *atoms[j];
                                          atom.set_orig(coord1);
                                          covalent_atoms.Add(atom);
                                     }
                                }
                                if (atoms) delete [] atoms;
                           }
                      }
                 }
            }
       }
       if (atom_a) delete [] atom_a;
       if (atom_b) delete [] atom_b;

       return found;
}

void _MAXIT::ndb_to_pdb_get_remark_620(ReVarCifArray<_METAL_COORDINATION> &MetalList)
{
       if (MetalList.Length() == 0) return;

       ndb_to_pdb_get_general_remark(Num_Remark_620, Remark_620, 0, 0, 0, 0);

       char buffer[1000], buffer1[1000], chain_id, insert_code, name[5];
       One_Card_Info_Format *pCard = NULL;
       for (unsigned int i = 0; i < MetalList.Length(); i++) {
            pCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            pCard->AddElementToField(620, 1);
            pCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            pCard->AddElementToField(620, 1);
            pCard->AddElementToField("COORDINATION ANGLES FOR:  M RES CSSEQI METAL", 2);

            if (strcmp(MetalList[i].metal_atom.pdb_chnid(), ""))
                 chain_id = MetalList[i].metal_atom.pdb_chnid()[0];
            else chain_id = ' ';
            if (strcmp(MetalList[i].metal_atom.ins_code(), ""))
                 insert_code = MetalList[i].metal_atom.ins_code()[0];
            else insert_code = ' ';
            get_print_atom(MetalList[i].metal_atom.pdb_resnam(),
                       MetalList[i].metal_atom.pdb_atmnam(), name);
            if (MetalList[i].Mol_ID > 0)
                 sprintf(buffer, "                        %3d %3s %c%4d%c %4s",  
                        MetalList[i].Mol_ID, MetalList[i].metal_atom.pdb_resnam(), chain_id,
                        MetalList[i].metal_atom.pdb_resnum(), insert_code, name);
            else sprintf(buffer, "                            %3s %c%4d%c %4s",
                        MetalList[i].metal_atom.pdb_resnam(), chain_id,
                        MetalList[i].metal_atom.pdb_resnum(), insert_code, name);
            pCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            pCard->AddElementToField(620, 1);
            pCard->AddElementToField(buffer, 2);
            pCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            pCard->AddElementToField(620, 1);
            pCard->AddElementToField("N RES CSSEQI ATOM", 2);
            for (unsigned int j = 0; j < MetalList[i].polar_atoms.Length(); j++) {
                 if (j == 9) break; 
                 if (strcmp(MetalList[i].polar_atoms[j].pdb_chnid(), ""))
                      chain_id = MetalList[i].polar_atoms[j].pdb_chnid()[0];
                 else chain_id = ' ';
                 if (strcmp(MetalList[i].polar_atoms[j].ins_code(), ""))
                      insert_code = MetalList[i].polar_atoms[j].ins_code()[0];
                 else insert_code = ' ';
                 get_print_atom(MetalList[i].polar_atoms[j].pdb_resnam(),
                         MetalList[i].polar_atoms[j].pdb_atmnam(), name);
                 sprintf(buffer, "%d %3s %c%4d%c %4s", j + 1,
                          MetalList[i].polar_atoms[j].pdb_resnam(), chain_id,
                          MetalList[i].polar_atoms[j].pdb_resnum(), insert_code, name);
                 strcpy(buffer1, "N                 ");
                 for (unsigned int k = 0; k < j; k++) {
                      double angle = cal_angle(MetalList[i].polar_atoms[k],
                                          MetalList[i].metal_atom,
                                          MetalList[i].polar_atoms[j]);
                      int len = strlen(buffer);
                      sprintf(&buffer[len], "%6.1f", angle);
                      len = strlen(buffer1);
                      sprintf(&buffer1[len], "   %d  ", k + 1);
                 }
                 pCard = _CardInfo[REMARK_TOKEN-1].
                            add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 pCard->AddElementToField(620, 1);
                 pCard->AddElementToField(buffer, 2);
            }
            pCard = _CardInfo[REMARK_TOKEN-1].
                            add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            pCard->AddElementToField(620, 1);
            pCard->AddElementToField(buffer1, 2);
       }
}

void _MAXIT::ndb_to_pdb_get_remark_630()
{
       if (_num_mol == 0 || _molecules == NULL) return;

       // key; het_id
       // vector[0]: type
       // vector[1]: details
       std::map<std::string, std::vector<std::string> > prd_info;
       get_prd_info(_CifObj, _StructureId, prd_info);

       ReVarCifArray<_REMARK_630> het_list;
       het_list.Clear();

       char buffer[100];
       CifString cs, type, source, s_details, o_details;
       ReVarCifArray<CifString> data;
       for (int l = 0; l < _num_mol; l++) {
            _CHAIN *chain = _molecules[l]->GetFirstChain();
            while (chain) {
                 if (chain->ResidueNumbers() == 1) {
                      _RESIDUE *residue = chain->GetFirstResidue();
                      if (residue) {
                           std::map<std::string, std::vector<std::string> >::const_iterator
                               mpos = prd_info.find(residue->ResName());
                           ConnectFormat *drug = find_drug(residue->ResName());
                           if (drug && drug->subcomponent_list() &&
                               strcmp(drug->subcomponent_list(), "") ||
                               mpos != prd_info.end()) {
                                int found = 0;
                                for (unsigned int i = 0; i < het_list.Length(); i++) {
                                     if (het_list[i].data[0] == residue->ResName()) {
                                          found = 1;
                                          if (_num_mol > 1)
                                               sprintf(buffer, "%d", _molecules[l]->Mol_ID());
                                          else strcpy(buffer, "");
                                          cs.Copy(buffer);
                                          data.Clear();
                                          data.Add(cs);
                                          cs.Copy(residue->ResName());
                                          data.Add(cs);
                                          _ATOM *atom = residue->GetFirstAtom();
                                          cs.Copy(atom->pdb_chnid());
                                          data.Add(cs);
                                          sprintf(buffer, "%d", atom->pdb_resnum());
                                          cs.Copy(buffer);
                                          data.Add(cs);
                                          cs.Copy(atom->ins_code());
                                          data.Add(cs);
                                          het_list[i].instances.Add(data);
                                          break;
                                     }
                                }
                                if (!found) {
                                     _REMARK_630 het;
                                     het.data.Clear();
                                     het.subcomp.Clear();
                                     het.instances.Clear();
                                     get_inhibitor_details(_CifObj, _StructureId,
                                       chain->entity_id(), type, source, s_details,
                                       o_details);
                                     if (mpos != prd_info.end()) {
                                          if (type == "") {
                                               type.Copy(mpos->second[0].c_str());
                                               type.ToUpper();
                                          }
                                          if (o_details == "") {
                                               o_details.Copy(mpos->second[1].c_str());
                                               o_details.ToUpper();
                                          }
                                     }
                                     cs.Copy(residue->ResName());
                                     het.data.Add(cs);
                                     het.data.Add(type);
                                     if (drug->chemical_name())
                                          cs.Copy(drug->chemical_name());
                                     het.data.Add(cs);
                                     het.data.Add(source);
                                     het.data.Add(s_details);
                                     het.data.Add(o_details);
                                     char *p = drug->subcomponent_list();
                                     while (p) {
                                          p = get_word(p, buffer);
                                          if (!strcmp(buffer, "")) continue;
                                          cs.Copy(buffer);
                                          het.subcomp.Add(cs);
                                     }
                                     if (_num_mol > 1)
                                          sprintf(buffer, "%d", _molecules[l]->Mol_ID());
                                     else strcpy(buffer, "");
                                     cs.Copy(buffer);
                                     data.Clear();
                                     data.Add(cs);
                                     cs.Copy(residue->ResName());
                                     data.Add(cs);
                                     _ATOM *atom = residue->GetFirstAtom();
                                     cs.Copy(atom->pdb_chnid());
                                     data.Add(cs);
                                     sprintf(buffer, "%d", atom->pdb_resnum());
                                     cs.Copy(buffer);
                                     data.Add(cs);
                                     cs.Copy(atom->ins_code());
                                     data.Add(cs);
                                     het.instances.Add(data);
                                     het_list.Add(het);
                                }
                           }
                      }
                 }
                 chain = _molecules[l]->GetNextChain();
            }
       }

       if (het_list.Length() == 0) return;

       char mol_id[4], chain_id[4], insert_code[4];
       for (unsigned int i = 0; i < het_list.Length(); i++) {
            for (int j = 0; j < NUM_REMARK_630; j++) {
                 cs.Copy(_remark_630_template[j].remark);
                 if (!strcmp(_remark_630_template[j].remark, "  M RES C SSSEQI")) {
                      write_remark(630, cs.Text());
                      for (unsigned k = 0; k < het_list[i].instances.Length(); k++) {
                           if (het_list[i].instances[k][0] != "")
                                strcpy(mol_id, het_list[i].instances[k][0].Text());
                           else strcpy(mol_id, "   ");
                           if (het_list[i].instances[k][2] != "")
                                strcpy(chain_id, het_list[i].instances[k][2].Text());
                           else strcpy(chain_id, " ");
                           if (het_list[i].instances[k][4] != "")
                                strcpy(insert_code, het_list[i].instances[k][4].Text());
                           else strcpy(insert_code, " ");
                           sprintf(buffer, "%3s %3s %c %5s%c", mol_id,
                                        het_list[i].instances[k][1].Text(),
                                        chain_id[0], het_list[i].instances[k][3].Text(),
                                        insert_code[0]);
                           write_remark(630, buffer);
                      }
                 } else if (!strcmp(_remark_630_template[j].remark, "SUBCOMP:")) {
                      if (het_list[i].subcomp.Length() > 0) {
                           cs += "    ";
                           int count = 0;
                           for (unsigned int k = 0; k < het_list[i].subcomp.Length(); k++) {
                                if (count == 10) {
                                     write_remark(630, cs.Text());
                                     count = 0;
                                     cs.Copy("          2 ");
                                }
                                sprintf(buffer, "%3s ", het_list[i].subcomp[k].Text());
                                cs += buffer;
                                count++;
                           }
                      } else cs += " NULL";
                      write_remark(630, cs.Text());
                 } else {
                      if (_remark_630_template[j].index) {
                           if (het_list[i].data[_remark_630_template[j].index] != "")
                                cs += het_list[i].data[_remark_630_template[j].index];
                           else cs += "NULL";
                      }
                      write_remark(630, cs.Text());
                 }
            }
       }
}

void _MAXIT::write_remark(const int remark_no, const char *text)
{
       One_Card_Info_Format *pCard = NULL;
       if (!strcmp(text, "")) {
            pCard = _CardInfo[REMARK_TOKEN-1].
                     add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            pCard->AddElementToField(remark_no, 1);
            pCard->AddElementToField(text, 2);
            return;
       }

       int len = strlen(text)+1;
       char *value = new char[len];
       memset(value, 0, len);
       strcpy(value, text);
       int width = NdbToken[REMARK_TOKEN-1].FieldList[2].FieldWidth - 1;
       char *buffer = new char[width+1];
       memset(buffer, 0, width+1);
       char *p = value;
       while (p) {
            p = get_max_length_word(p, buffer, width);
            if (!strcmp(buffer, "")) continue;
            pCard = _CardInfo[REMARK_TOKEN-1].
                     add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            pCard->AddElementToField(remark_no, 1);
            pCard->AddElementToField(buffer, 2);
       }
       delete [] value;
       delete [] buffer;
}

void _MAXIT::find_secondary_citations(const char* index, const int remark_no)
{
       if (_JrnlCardInfo == NULL) return;

       int vField = 5;
       int jrnl_id = find_jrnl_id("REF");
       One_Card_Info_Format *pJrnlCard = _JrnlCardInfo[jrnl_id].Head();
       while (pJrnlCard) {
            if (strcmp(pJrnlCard->Field(vField+1), "") &&
               !strcmp(pJrnlCard->Field(vField), ""))
                 pJrnlCard->AddElementToField("V.", vField);
            else if (!strcmp(pJrnlCard->Field(vField+1), ""))
                 pJrnlCard->AddElementToField("", vField);
            pJrnlCard = pJrnlCard->Next();
       }

       int astmField = 3;
       int countryField = 5;
       int issnField = 6;
       int refn_id = find_jrnl_id("REFN");
       pJrnlCard = _JrnlCardInfo[refn_id].Head();
       while (pJrnlCard) {
            pJrnlCard->AddElementToField("", astmField);
            pJrnlCard->AddElementToField("", astmField+1);
            pJrnlCard->AddElementToField("", countryField);

            if (strcmp(pJrnlCard->Field(issnField+1), "") &&
               !strcmp(pJrnlCard->Field(issnField), ""))
                 pJrnlCard->AddElementToField("ISSN", issnField);
            else if (!strcmp(pJrnlCard->Field(issnField+1), ""))
                 pJrnlCard->AddElementToField("", issnField);

            if (!strcmp(pJrnlCard->Field(issnField), "ISBN")) {
                 pJrnlCard->AddElementToField("", issnField);
                 pJrnlCard->AddElementToField("", issnField+1);
            }
            pJrnlCard = pJrnlCard->Next();
       }

       std::set<std::string> index_set;
       index_set.clear();
       for (jrnl_id = 0; jrnl_id < MxJrnlTokens; jrnl_id++) {
            pJrnlCard = _JrnlCardInfo[jrnl_id].Head();
            while (pJrnlCard) {
                 index_set.insert(pJrnlCard->Field(1));
                 pJrnlCard = pJrnlCard->Next();
            }
       }
       if (index_set.empty()) return;

       char temp_line[MxLen], FieldFormat[MxStrLen], token[MxLen], buffer[MxLen], *p = NULL;
       One_Card_Info_Format *newCard = NULL;

       std::vector<std::string> index_array;
       index_array.clear();
       for (unsigned int i = 0; i < index_set.size(); ++i) {
            memset(buffer, 0, MxLen);
            sprintf(buffer, "%s%d", index, (int) i + 1);
            if (index_set.find(buffer) != index_set.end()) index_array.push_back(buffer);
       }
       if (index_array.empty()) return;

       std::vector<std::string> original_entry_ids;
       original_entry_ids.clear();
       if (remark_no  == 0) {
            newCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(remark_no, 1);

            if (_CifObj) {
                 ISTable *t = _CifObj->GetTablePtr(_StructureId, "rcsb_database_related");
                 if (!is_empty_table(t)) {
                      CifString cs1, cs2, cs3;
                      int idx1 = t->GetColumnIndex("db_name");
                      int idx2 = t->GetColumnIndex("db_id");
                      int idx3 = t->GetColumnIndex("content_type");
                      int rowNo = t->GetNumRows();
                      for (int i = 0; i < rowNo; ++i) {
                           get_value(cs1, t, idx1, i);
                           get_value(cs2, t, idx2, i);
                           get_value(cs3, t, idx3, i);
                           if (cs1 == "" || cs2 == "" || cs3 == "") continue;
                           cs1.ToUpper();
                           if (cs1 != "PDB") continue; 
                           cs3.ToLower();
                           if (cs3 != "re-refinement") continue;
                           cs2.ToUpper();
                           original_entry_ids.push_back(cs2.Text());
                      }
                 }
            }
       }

       index_set.clear();
       for (unsigned int i = 0; i < index_array.size(); ++i) {
            index_set.insert(index_array[i]);
            newCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(remark_no, 1);
            if (remark_no  == 0)
                 sprintf(temp_line, "ORIGINAL DATA REFERENCE %d", (int) i + 1);
            else sprintf(temp_line, "REFERENCE %d", (int) i + 1);
            newCard->AddElementToField(temp_line, 2);
            if (remark_no  == 0 && !original_entry_ids.empty()) {
                 newCard = _CardInfo[REMARK_TOKEN-1].
                            add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(remark_no, 1);
                 if (i < original_entry_ids.size())
                      sprintf(temp_line, " PDB ID: %s", original_entry_ids[i].c_str());
                 else sprintf(temp_line, " PDB ID: %s", original_entry_ids[0].c_str());
                 newCard->AddElementToField(temp_line, 2);
            }
            for (int displayOrder = 0; displayOrder < MxJrnlTokens; displayOrder++) {
                 for (jrnl_id = 0; jrnl_id < MxJrnlTokens; jrnl_id++) {
                      if (JrnlToken[jrnl_id].DisplayOrder == displayOrder) break;
                 }
                 if (jrnl_id == MxJrnlTokens) continue;

                 pJrnlCard = _JrnlCardInfo[jrnl_id].Head();
                 while (pJrnlCard) {
                      if (index_array[i] == pJrnlCard->Field(1)) break;
                      pJrnlCard = pJrnlCard->Next();
                 }

                 if (pJrnlCard == NULL) continue;

                 int ContNo = 1;
                 int start = 0;
                 int fieldwidth = 0;
                 bool left_adjust = false;
                 while (1) {
                      int pos = 1;
                
                      strcpy(temp_line, " ");
                      for (int field_no = 2; field_no < JrnlToken[jrnl_id].NumField; field_no++) {
                           if (JrnlToken[jrnl_id].FieldList[field_no].NdbOnly) continue;
                           pJrnlCard->UpperCaseField(field_no);
                           if (field_no > 2) {
                                int width = JrnlToken[jrnl_id].FieldList[field_no].FieldStcol
                                          - JrnlToken[jrnl_id].FieldList[field_no-1].FieldStcol
                                          - JrnlToken[jrnl_id].FieldList[field_no-1].FieldWidth;
                                sprintf(FieldFormat, "%%-%ds", width);
                                sprintf(&temp_line[pos], FieldFormat, "");
                                pos += width;
                           }
                           get_field_format(jrnl_id, field_no, 1, NDB_FILE_FORMAT_PDB,
                                            FieldFormat, fieldwidth, left_adjust);
                           if (JrnlToken[jrnl_id].ConField == field_no + 1) {
                                if (ContNo != 1)
                                     // sprintf(&temp_line[pos], FieldFormat, ContNo);
                                     write_format_value(&temp_line[pos], FieldFormat,
                                            ContNo, fieldwidth, left_adjust);
                                else {
                                     sprintf(FieldFormat, "%%%ds",
                                            JrnlToken[jrnl_id].FieldList[field_no].FieldWidth);
                                     sprintf(&temp_line[pos], FieldFormat, "");
                                }                      
                           } else if (JrnlToken[jrnl_id].ContinuedF == field_no + 1)
                                print_jrnl_continuedf_field(jrnl_id, pJrnlCard, pos, start,
                                                     temp_line, FieldFormat);
                           else if (strcmp(pJrnlCard->Field(field_no), "") == 0 ||
                                   ContNo != 1 && field_no != 2) {
                                sprintf(FieldFormat, "%%%ds",
                                       JrnlToken[jrnl_id].FieldList[field_no].FieldWidth);
                                sprintf(&temp_line[pos], FieldFormat, "");
                           } else if (JrnlToken[jrnl_id].FieldList[field_no].FieldType <= 2) {
                                double temp_double = atof(pJrnlCard->Field(field_no));
                                // sprintf(&temp_line[pos], FieldFormat, (int) temp_double);
                                write_format_value(&temp_line[pos], FieldFormat,
                                           (int) temp_double, fieldwidth, left_adjust);
                           } else {
                                // sprintf(&temp_line[pos], FieldFormat, pJrnlCard->Field(field_no));
                                write_format_value(&temp_line[pos], FieldFormat,
                                   pJrnlCard->Field(field_no), fieldwidth, left_adjust);
                           }
                           pos += JrnlToken[jrnl_id].FieldList[field_no].FieldWidth;
                      }

                      newCard = _CardInfo[REMARK_TOKEN-1].
                                 add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                      newCard->AddElementToField(remark_no, 1);
                      newCard->AddElementToField(temp_line, 2);

                      if (JrnlToken[jrnl_id].ContinuedF != 0 && start <
                              (int) strlen(pJrnlCard->Field(JrnlToken[jrnl_id].ContinuedF-1)))
                           ContNo++;
                      else {
                           if (jrnl_id == refn_id) {
                                if (strcmp(pJrnlCard->Field(9), "") &&
                                    strcmp(pJrnlCard->Field(9), "-1")) {
                                     strcpy(token, " PMID");
                                     sprintf(temp_line, "%s   %s", token, pJrnlCard->Field(9));
                                     newCard = _CardInfo[REMARK_TOKEN-1].
                                                add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                                     newCard->AddElementToField(remark_no, 1);
                                     newCard->AddElementToField(temp_line, 2);
                                }
                                if (strcmp(pJrnlCard->Field(10), "")) {
                                     strcpy(token, " DOI ");
                                     p = pJrnlCard->Field(10);
                                     int doi = 0;
                                     while (p) {
                                          doi++;
                                          p = get_word_with_length(p, buffer, 51);
                                          if (doi > 1)
                                               sprintf(temp_line, "%s%2d %s", token, doi, buffer);
                                          else sprintf(temp_line, "%s   %s", token, buffer);
                                          ndb_set_string_to_upper(temp_line);
                                          newCard = _CardInfo[REMARK_TOKEN-1].
                                                     add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                                          newCard->AddElementToField(remark_no, 1);
                                          newCard->AddElementToField(temp_line, 2);
                                     }
                                }
                           }
                           break;
                      }
                 }
            }
       }

       for (int i = 0; i < MxJrnlTokens; i++) {
            pJrnlCard = _JrnlCardInfo[i].Head();
            while (pJrnlCard) {
                 if (index_set.find(pJrnlCard->Field(1)) != index_set.end())
                      pJrnlCard = _JrnlCardInfo[i].delete_a_card(pJrnlCard);
                 else pJrnlCard = pJrnlCard->Next();
            }
       }
}

int _MAXIT::generate_biological_assembly()
{
       if (!_CifObj) return 0;

       bool keep_original_count = false;
       ISTable *Table = _CifObj->GetTablePtr(_StructureId, "ndb_database_status");
       if (!is_empty_table(Table)) {
            CifString cs;
            int index = Table->GetColumnIndex("skip_PDB_REMARK");
            get_value(cs, Table, index, 0);
            cs.ToLower();
            if (strstr(cs.Text(), "assembly")) keep_original_count = true;
       }

       ISTable *t = NULL;
       ReVarCifArray<CifString> id_list;
       double mat[4][4];
       char FieldFormat[82];
       CifString cs1, cs2;
       char *buff = NULL, *p = NULL, buffer[80];
       int len, rowNo, idx1;
       One_Card_Info_Format *newCard = _CardInfo[SPLIT_TOKEN-1].Head();
       if (newCard &&
           is_empty_table(_CifObj->GetTablePtr(_StructureId, "pdbx_struct_assembly"))) {
            ndb_to_pdb_get_general_remark(Num_Remark_350, Remark_350, 0, 0, 0, 0);
            newCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(350, 1);
            newCard->AddElementToField("BIOMOLECULE: 1", 2);
            id_list.Clear();
            t = _CifObj->GetTablePtr(_StructureId, "struct_biol_gen");
            if (!is_empty_table(t)) {
                 rowNo = t->GetNumRows();
                 idx1 = t->GetColumnIndex("ndb_new_pdb_asym_id");
                 for (int i = 0; i < rowNo; i++) {
                      get_value(cs1, t, idx1, i);
                      if (cs1 == "") continue;
                      int found = 0;
                      for (unsigned int j = 0; j < id_list.Length(); j++) {
                           if (cs1 == id_list[j]) {
                                found = 1;
                                break;
                           }
                      }
                      if (!found) id_list.Add(cs1);
                 }
            }
            if (id_list.Length() <= NUM_MER)
                 sprintf(FieldFormat, "QUATERNARY STRUCTURE FOR THIS ENTRY: %s",
                                      _num_mers[id_list.Length()]);
            else sprintf(FieldFormat, "QUATERNARY STRUCTURE FOR THIS ENTRY: %d-MERIC",
                                      (int) id_list.Length());
            newCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(350, 1);
            newCard->AddElementToField(FieldFormat, 2);

            cs1.Clear();
            for (unsigned int i = 0; i < id_list.Length(); i++) {
                 if (cs1 != "") cs1 += ", ";
                 cs1 += id_list[i];
            }
            cs2.Copy("APPLY THE FOLLOWING TO CHAINS: ");
            len = strlen(cs1.Text());
            buff = new char[len + 1];
            memset(buff, 0, len + 1);
            strcpy(buff, cs1.Text());
            p = buff;
            while (p) {
                 p = get_max_length_word(p, buffer, 29);
                 newCard = _CardInfo[REMARK_TOKEN-1].
                            add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(350, 1);
                 cs2 += buffer;
                 newCard->AddElementToField(cs2.Text(), 2);
                 cs2.Copy("                   AND CHAINS: ");
            }
            delete [] buff;

            mat[0][0] = mat[1][1] = mat[2][2] = 1.0;
            mat[0][1] = mat[0][2] = mat[0][3] = 0.0;
            mat[1][0] = mat[1][2] = mat[1][3] = 0.0;
            mat[2][0] = mat[2][1] = mat[2][3] = 0.0;

            for (int k = 0; k < 3; k++) {
                 newCard = _CardInfo[REMARK_TOKEN-1].
                          add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(Remark_SYMMA.Remark_No, 1);
                 for (int j = 0; j < Remark_SYMMA.NumField; j++) {
                      print_space_between_remark_field(newCard, &Remark_SYMMA, j);
                      get_remark_field_format(FieldFormat, &Remark_SYMMA, j);
                      int field_no = Remark_SYMMA.FieldList[j].FieldId-1;
                      if (!strcmp("BIOMT", Remark_SYMMA.FieldList[j].Text))
                           sprintf(buffer, "%s%d",
                               Remark_SYMMA.FieldList[j].Text, k + 1);
                      else if (field_no == 1)
                           sprintf(buffer, FieldFormat, 1.0);
                      else if (j != Remark_SYMMA.NumField - 1)
                           sprintf(buffer, FieldFormat, mat[k][j-2]);
                      else sprintf(buffer, FieldFormat, mat[k][3]);
                      newCard->AttachToField(buffer, 2);
                 }
            }

            return 0;
       }

       if (is_empty_table(_CifObj->GetTablePtr(_StructureId, "pdbx_struct_assembly")) ||
           is_empty_table(_CifObj->GetTablePtr(_StructureId, "pdbx_struct_assembly_gen")) ||
           is_empty_table(_CifObj->GetTablePtr(_StructureId, "pdbx_struct_oper_list"))) {
            return 0;
       }

       ReVarCifArray<_ASSEMBLY> assembly_lists;
       get_assembly_list(_CifObj, _StructureId, assembly_lists, keep_original_count);
       if (assembly_lists.Length() == 0) return 0;

       t = _CifObj->GetTablePtr(_StructureId, "pdbx_struct_oper_list");

       ndb_to_pdb_get_general_remark(Num_Remark_350 - 1, Remark_350, 0, 0, 0, 0);

       for (unsigned int l = 0; l < assembly_lists.Length(); l++) {
            newCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(350, 1);
            newCard->AddElementToField("", 2);
            newCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(350, 1);
            sprintf(buffer, "BIOMOLECULE: %d", l + 1);
            newCard->AddElementToField(buffer, 2);

            if (assembly_lists[l].total_chains <= NUM_MER)
                 strcpy(FieldFormat, _num_mers[assembly_lists[l].total_chains]);
            else sprintf(FieldFormat, "%d-MERIC", assembly_lists[l].total_chains);
            if (assembly_lists[l].type == "author_and_software_defined_assembly" ||
                assembly_lists[l].type == "author_defined_assembly") {
                 newCard = _CardInfo[REMARK_TOKEN-1].
                            add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(350, 1);
                 sprintf(buffer, "AUTHOR DETERMINED BIOLOGICAL UNIT: %s", FieldFormat);
                 newCard->AddElementToField(buffer, 2);
            }
            if (assembly_lists[l].type == "author_and_software_defined_assembly" ||
                assembly_lists[l].type == "software_defined_assembly") {
                 newCard = _CardInfo[REMARK_TOKEN-1].
                            add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(350, 1);
                 sprintf(buffer, "SOFTWARE DETERMINED QUATERNARY STRUCTURE: %s", FieldFormat);
                 newCard->AddElementToField(buffer, 2);
                 if (assembly_lists[l].software == "") assembly_lists[l].software.Copy("PISA");
                 if (assembly_lists[l].software != "") {
                      newCard = _CardInfo[REMARK_TOKEN-1].
                                 add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                      newCard->AddElementToField(350, 1);
                      sprintf(buffer, "SOFTWARE USED: %s", assembly_lists[l].software.Text());
                      newCard->AddElementToField(buffer, 2);
                 }
                 if (assembly_lists[l].absa != "") {
                      newCard = _CardInfo[REMARK_TOKEN-1].
                                 add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                      newCard->AddElementToField(350, 1);
                      sprintf(buffer, "TOTAL BURIED SURFACE AREA: %d ANGSTROM**2",
                                      atoi(assembly_lists[l].absa.Text()));
                      newCard->AddElementToField(buffer, 2);
                 }
                 if (assembly_lists[l].ssa != "") {
                      newCard = _CardInfo[REMARK_TOKEN-1].
                                 add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                      newCard->AddElementToField(350, 1);
                      // sprintf(buffer, "TOTAL SURFACE AREA FOR THE COMPLEX: %d ANGSTROM**2",
                      sprintf(buffer, "SURFACE AREA OF THE COMPLEX: %d ANGSTROM**2",
                                         atoi(assembly_lists[l].ssa.Text()));
                      newCard->AddElementToField(buffer, 2);
                 }
                 if (assembly_lists[l].more != "") {
                      newCard = _CardInfo[REMARK_TOKEN-1].
                                 add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                      newCard->AddElementToField(350, 1);
                      // sprintf(buffer, "GAIN IN SOLVENT FREE ENERGY: %d KCAL/MOL",
                      sprintf(buffer, "CHANGE IN SOLVENT FREE ENERGY: %.1f KCAL/MOL",
                                         atof(assembly_lists[l].more.Text()));
                      newCard->AddElementToField(buffer, 2);
                 }
            }

            int serialNo = 0;
            for (unsigned int m = 0; m < assembly_lists[l].coplists.Length(); m++) {
                 cs2.Copy("APPLY THE FOLLOWING TO CHAINS: ");
/*
                 cs2 += assembly_lists[l].coplists[m].inst_ids;
                 len = strlen(cs2.Text());
*/
                 len = strlen(assembly_lists[l].coplists[m].inst_ids.Text());
                 buff = new char[len + 1];
                 memset(buff, 0, len + 1);
                 // strcpy(buff, cs2.Text());
                 strcpy(buff, assembly_lists[l].coplists[m].inst_ids.Text());
                 p = buff;
                 while (p) {
                      p = get_max_length_word(p, buffer, 29 /*60 */ );
                      newCard = _CardInfo[REMARK_TOKEN-1].
                                 add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                      newCard->AddElementToField(350, 1);
                      // newCard->AddElementToField(buffer, 2);
                      cs2 += buffer;
                      newCard->AddElementToField(cs2.Text(), 2);
                      cs2.Copy("                   AND CHAINS: ");
                 }
                 delete [] buff;

                 for (unsigned int i = 0; i < assembly_lists[l].coplists[m].oplist.Length(); i++) {
                      get_multi_matrix_list(assembly_lists[l].coplists[m].oplist[i].Text(), id_list);
                      if (get_final_matrix(mat, id_list, t)) {
                           serialNo++;
                           for (int k = 0; k < 3; k++) {
                                newCard = _CardInfo[REMARK_TOKEN-1].
                                         add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                                newCard->AddElementToField(Remark_SYMMA.Remark_No, 1);
                                for (int j = 0; j < Remark_SYMMA.NumField; j++) {
                                     print_space_between_remark_field(newCard, &Remark_SYMMA, j);
                                     get_remark_field_format(FieldFormat, &Remark_SYMMA, j);
                                     int field_no = Remark_SYMMA.FieldList[j].FieldId-1;
                                     if (!strcmp("BIOMT", Remark_SYMMA.FieldList[j].Text))
                                          sprintf(buffer, "%s%d",
                                              Remark_SYMMA.FieldList[j].Text, k + 1);
                                     else if (field_no == 1)
                                          sprintf(buffer, FieldFormat, (double) serialNo);
                                     else if (j != Remark_SYMMA.NumField - 1)
                                          sprintf(buffer, FieldFormat, mat[k][j-2]);
                                     else sprintf(buffer, FieldFormat, mat[k][3]);
                                     newCard->AttachToField(buffer, 2);
                                }
                           }
                      }
                 }
            }
       }

       return 1;
}

void _MAXIT::recover_assembly_info()
{
       if (__assembly_array.Length() == 0) {
            CifString cs, cs1;
            int rowNo, idx1, idx2, idx3;

            ReVarCifArray<_ASYM_CHAINID_CORR> asym_chainid_corrs;
            asym_chainid_corrs.Clear();
            ISTable *t = _CifObj->GetTablePtr(_StructureId, "struct_biol_gen");
            if (!is_empty_table(t)) {
                 rowNo = t->GetNumRows();
                 idx1 = t->GetColumnIndex("asym_id");
                 idx2 = t->GetColumnIndex("ndb_new_pdb_asym_id");
                 for (int i = 0; i < rowNo; i++) {
                      get_value(cs, t, idx1, i);
                      get_value(cs1, t, idx2, i);
                      if (cs != "" && cs1 != "") {
                           _ASYM_CHAINID_CORR coor;
                           coor.asym_id.Copy(cs);
                           coor.pdb_chnid.Copy(cs1);
                           asym_chainid_corrs.Add(coor);
                      }
                 }
            }

            ReVarCifArray<CifString> colData;
            t = _CifObj->GetTablePtr(_StructureId, "pdbx_struct_assembly_gen");
            if (!is_empty_table(t)) {
                 idx1 = t->GetColumnIndex("asym_id_list");
                 idx2 = t->GetColumnIndex("auth_asym_id_list");
                 if (idx2 >= 0) {
                      rowNo = t->GetNumRows();
                      if (idx1 < 0) {
                           colData.Clear();
                           cs.Copy("?");                
                           for (int i = 0; i < rowNo; ++i) colData.Add(cs);
                           t->AddColumn("asym_id_list");
                           idx1 = t->GetColumnIndex("asym_id_list");
                           t->FillColumn(colData, idx1);
                      }

                      for (int i = 0; i < rowNo; ++i) {
                           get_value(cs, t, idx2, i);
                           if (cs == "") continue;
                           get_wordlist_from_string_separated_by_delimits(colData,
                                      cs.Text(), ",");
                           cs.Clear();
                           for (unsigned int l = 0; l < asym_chainid_corrs.Length(); ++l) {
                                for (unsigned int j = 0; j < colData.Length(); ++j) {
                                     if (asym_chainid_corrs[l].pdb_chnid == colData[j]) {
                                          if (cs != "") cs += ",";
                                          cs += asym_chainid_corrs[l].asym_id;
                                     }
                                }
                           }
                           if (cs != "") t->UpdateCell(cs, idx1, i);
                      }

                      t->DeleteColumn(idx2);
                      for (int j = 0; j < t->GetNumColumns(); j++) {
                           if (t->ColumnLength(j) == 0) {
                                t->RemoveColumn(j);
                                j--;
                           }
                      }
                      _CifObj->WriteTable(t, _StructureId, "pdbx_struct_assembly_gen");
                 }
            }
 
            t = _CifObj->GetTablePtr(_StructureId, "pdbx_struct_assembly");
            if (!is_empty_table(t)) {
                 rowNo = t->GetNumRows();
                 idx1 = t->GetColumnIndex("details");
                 bool found = false;
                 for (int i = 0; i < rowNo; ++i) {
                      get_value(cs, t, idx1, i);
                      cs.ToLower();
                      if (cs == "representative helical assembly" ||
                          cs == "complete icosahedral assembly" ||
                          cs == "complete point assembly") {
                           found = true;
                           break;
                      }
                 }
                 if (!found) return;
                 idx2 = t->GetColumnIndex("oligomeric_details");
                 idx3 = t->GetColumnIndex("oligomeric_count");

                 colData.Clear();
                 cs.Copy("?");                
                 for (int i = 0; i < rowNo; ++i) colData.Add(cs);
                 if (idx2 < 0) {
                      t->AddColumn("oligomeric_details");
                      idx2 = t->GetColumnIndex("oligomeric_details");
                      t->FillColumn(colData, idx2);
                 }

                 if (idx3 >= 0) {
                      for (int j = 0; j < rowNo; ++j) {
                           get_value(cs, t, idx1, j);
                           cs.ToLower();
                           if (cs == "representative helical assembly") {
                                cs = "helical";
                                t->UpdateCell(cs, idx2, j);
                           } else {
                                get_value(cs, t, idx3, j);
                                if (cs != "") {
                                     cs += "-MERIC";
                                     t->UpdateCell(cs, idx2, j);
                                }
                           }
                      }
                      _CifObj->WriteTable(t, _StructureId, "pdbx_struct_assembly");
                      return;
                 }

                 t->AddColumn("oligomeric_count");
                 idx3 = t->GetColumnIndex("oligomeric_count");
                 t->FillColumn(colData, idx3);
                 ReVarCifArray<_ASSEMBLY> assembly_lists;
                 get_assembly_list(_CifObj, _StructureId, assembly_lists, false);
                 for (unsigned int i = 0; i < assembly_lists.Length(); ++i) {
                      for (int j = 0; j < rowNo; ++j) {
                           get_value(cs, t, idx1, j);
                           cs.ToLower();
                           if (cs == assembly_lists[i].type) {
                                if (cs == "representative helical assembly") {
                                     cs = "helical";
                                     t->UpdateCell(cs, idx2, j);
                                     cs.Clear();
                                     cs += assembly_lists[i].total_chains;
                                     t->UpdateCell(cs, idx3, j);
                                } else {
                                     // get_value(cs, t, idx1, j);
                                     // t->UpdateCell(cs, idx2, j);
                                     cs.Clear();
                                     cs += assembly_lists[i].total_chains;
                                     t->UpdateCell(cs, idx3, j);
                                     cs += "-MERIC";
                                     t->UpdateCell(cs, idx2, j);
                                }
                                break;
                           }
                      }
                 }
                 _CifObj->WriteTable(t, _StructureId, "pdbx_struct_assembly");
            }
            add_nmr_assembly_info();
            return;
       }

       bool keep_original_count = false;
       ISTable *Table = _CifObj->GetTablePtr(_StructureId, "ndb_database_status");
       if (!is_empty_table(Table)) {
            CifString cs;
            int index = Table->GetColumnIndex("skip_PDB_REMARK");
            get_value(cs, Table, index, 0);
            cs.ToLower();
            if (strstr(cs.Text(), "assembly")) keep_original_count = true;
       }

       ReVarCifArray<_PDB_ASYM_CORR> chainids;
       get_pdb_asym_correspondence(_CifObj, _StructureId, chainids);

       Table = pdbx_struct_assembly_Table();
       ISTable *t1 = pdbx_struct_assembly_gen_Table();

       ISTable *t2 = new ISTable("pdbx_struct_assembly_prop");
       t2->AddColumn("biol_id");
       t2->AddColumn("type");
       t2->AddColumn("value");
       t2->AddColumn("details");

       ISTable *t3 = pdbx_struct_oper_list_Table();

       CifString cifstring, assembly_id, sym, oper_id;
       int row, len, k;
       char *value = NULL;
       ReVarCifArray<_SYM_CHAIN_CORR> sym_chain_corrs;
       BTreeObj<_IIndex> Sort_Index;
       _IIndex idx;
       Sort_Index.Clear();
       for (unsigned int i2 = 0; i2 < __assembly_array.Length(); i2++) {
            idx.set_value(atoi(__assembly_array[i2].assembly_id.Text()));
            idx.set_pos(i2);
            Sort_Index.Add(idx);
       }

       ReVarCifArray<_ASSEMBLY_ARRAY> assembly_array_sort;
       assembly_array_sort.Clear();
       Sort_Index.GoFirst();
       idx = Sort_Index.GetCurrData();
       assembly_array_sort.Add(__assembly_array[idx.pos()]);
       while (Sort_Index.GoToNext() >= 0) {
            idx = Sort_Index.GetCurrData();
            assembly_array_sort.Add(__assembly_array[idx.pos()]);
       }

       char buffer[100], *p = NULL;
       for (unsigned int i = 0; i < assembly_array_sort.Length(); i++) {
            int rrow = Table->GetNumRows();
            Table->AddRow();
            sprintf(buffer, "%d", i + 1);
            assembly_id.Copy(buffer);
            Table->UpdateCell(assembly_id, 0, rrow);
            Table->UpdateCell(assembly_array_sort[i].method, 1, rrow);
            Table->UpdateCell(assembly_array_sort[i].software, 2, rrow);
            if (assembly_array_sort[i].absa != "") {
                 row = t2->GetNumRows();
                 t2->AddRow();
                 t2->UpdateCell(assembly_id, 0, row);
                 cifstring.Copy("ABSA (A^2)");
                 t2->UpdateCell(cifstring, 1, row);
                 t2->UpdateCell(assembly_array_sort[i].absa, 2, row);
            }
            if (assembly_array_sort[i].ssa != "") {
                 row = t2->GetNumRows();
                 t2->AddRow();
                 t2->UpdateCell(assembly_id, 0, row);
                 cifstring.Copy("SSA (A^2)");
                 t2->UpdateCell(cifstring, 1, row);
                 t2->UpdateCell(assembly_array_sort[i].ssa, 2, row);
            }
            if (assembly_array_sort[i].more != "") {
                 row = t2->GetNumRows();
                 t2->AddRow();
                 t2->UpdateCell(assembly_id, 0, row);
                 cifstring.Copy("MORE");
                 t2->UpdateCell(cifstring, 1, row);
                 t2->UpdateCell(assembly_array_sort[i].more, 2, row);
            }

            sym_chain_corrs.Clear();
            int count = 0;
            for (unsigned int j = 0; j < assembly_array_sort[i].assembly_chains.Length(); j++) {
                 len = strlen(assembly_array_sort[i].assembly_chains[j].sym_operators.Text())+1;
                 value = new char[len];
                 memset(value, 0, len);
                 strcpy(value, assembly_array_sort[i].assembly_chains[j].sym_operators.Text()); 
                 for (k = 0; k < len - 1; k++) {
                      if (value[k] == ',') value[k] = ' ';
                 }
                 ndb_clean_string(value);
                 p = value;
                 while (p) {
                      p = get_word(p, buffer);
                      if (!strcmp(buffer, "")) continue;
                      count++;
                      row = t3->GetNumRows();
                      for (k = 0; k < row; k++) {
                           t3->GetCell(sym, 2, k);
                           if (sym == buffer) {
                                t3->GetCell(oper_id, 0, k);
                                break;
                           }
                      }
                      if (k == row) {
                           t3->AddRow();
                           sym.Copy(buffer);
                           t3->UpdateCell(sym, 2, row);
                           sprintf(buffer, "%d", row + 1);
                           oper_id.Copy(buffer);
                           t3->UpdateCell(oper_id, 0, row);
                           if (sym == "1_555")
                                cifstring.Copy("identity operation");
                           else cifstring.Copy("crystal symmetry operation");
                           t3->UpdateCell(cifstring, 1, row);
                      }
                      for (k = 0; k < (int) sym_chain_corrs.Length(); k++) {
                           if (sym_chain_corrs[k].sym == sym) {
                                sym_chain_corrs[k].catchainids += assembly_array_sort[i].assembly_chains[j].PDB_chain_ID; 
                                sym_chain_corrs[k].chainids.Add(assembly_array_sort[i].assembly_chains[j].PDB_chain_ID);
                                break;
                           }
                      }
                      if (k == (int) sym_chain_corrs.Length()) {
                           _SYM_CHAIN_CORR sym_corr;
                           sym_corr.sym.Copy(sym);
                           sym_corr.oper_id.Copy(oper_id);
                           sym_corr.catchainids.Copy(assembly_array_sort[i].assembly_chains[j].PDB_chain_ID);
                           sym_corr.chainids.Clear();
                           sym_corr.chainids.Add(assembly_array_sort[i].assembly_chains[j].PDB_chain_ID);
                           sym_chain_corrs.Add(sym_corr);
                      }
                 }
                 delete [] value;
            }
            if (sym_chain_corrs.Length() > 0) {
                 Sort_Index.Clear();
                 for (unsigned int j = 0; j < sym_chain_corrs.Length(); j++) {
                      idx.set_value(sym_chain_corrs[j].chainids.Length());
                      idx.set_pos(j);
                      Sort_Index.Add(idx);
                 }
                 if (Sort_Index.Root() >= 0) {
                      Sort_Index.GoFirst();
                      int *index = new int[sym_chain_corrs.Length()];
                      idx = Sort_Index.GetCurrData();
                      k = 0;
                      index[k++] = idx.pos();
                      while (Sort_Index.GoToNext() >= 0) {
                           idx = Sort_Index.GetCurrData();
                           index[k++] = idx.pos();
                      }
                      for (int f = 0; f < k; f++) {
                           cifstring.Copy(sym_chain_corrs[index[f]].oper_id);
                           int last = f;
                           for (int s = f + 1; s < k; s++) {
                                if (sym_chain_corrs[index[s]].catchainids ==
                                    sym_chain_corrs[index[f]].catchainids) {
                                     cifstring += ",";
                                     cifstring += sym_chain_corrs[index[s]].oper_id;
                                     last = s;
                                }
                           }
                           row = t1->GetNumRows();
                           t1->AddRow();
                           t1->UpdateCell(assembly_id, 0, row);
                           t1->UpdateCell(cifstring, 1, row);
                           cifstring.Clear();
                           for (unsigned int j = 0; j < sym_chain_corrs[index[f]].chainids.Length(); j++) {
                                for (unsigned int l = 0; l < chainids.Length(); l++) {
                                     if (sym_chain_corrs[index[f]].chainids[j] ==
                                         chainids[l].PDB_chain_ID) {
                                          for (unsigned int m = 0; m < chainids[l].Asym_IDs.Length(); m++) {
                                               if (cifstring != "") cifstring += ",";
                                               cifstring += chainids[l].Asym_IDs[m]; 
                                          }
                                          break;
                                     }
                                }
                           }
                           t1->UpdateCell(cifstring, 2, row);
                           f = last;
                      }
                      delete [] index;
                 }
            }

            if (keep_original_count && assembly_array_sort[i].oligomeric_count)
                 count = assembly_array_sort[i].oligomeric_count;

            if (count <= NUM_MER) cifstring.Copy(_num_mers[count]);
            else {
                 sprintf(buffer, "%d-MERIC", count);
                 cifstring.Copy(buffer);
            }
            cifstring.ToLower();
            Table->UpdateCell(cifstring, 3, rrow);
            sprintf(buffer, "%d", count);
            cifstring.Copy(buffer);
            Table->UpdateCell(cifstring, 4, rrow);
       }

       int rowNo = t3->GetNumRows(); 
       int j, op, lx, ly, lz;
       double matrix[3][3], vector[3];
       for (int i3 = 0; i3 < rowNo; i3++) {
            t3->GetCell(cifstring, 2, i3);
            strcpy(buffer, cifstring.Text());
            op = atoi(buffer) - 1;
            len = strlen(buffer);
            for (j = 0; j < len; j++) {
                 if (buffer[j] == '_') break;
            }
            lx = buffer[j + 1] - '5';
            ly = buffer[j + 2] - '5';
            lz = buffer[j + 3] - '5';
            ndb_get_symmetry_matrix(_cell, op, lx, ly, lz, matrix, vector);
            memset(buffer, 0, 100);
            ndb_get_symmetry_operation_name(_cell, op, lx, ly, lz, buffer);
            cifstring.Copy(buffer);
            t3->UpdateCell(cifstring,  3, i3);
            sprintf(buffer, "%.10f", matrix[0][0]);
            cifstring.Copy(buffer);
            t3->UpdateCell(cifstring,  4, i3);
            sprintf(buffer, "%.10f", matrix[0][1]);
            cifstring.Copy(buffer);
            t3->UpdateCell(cifstring,  5, i3);
            sprintf(buffer, "%.10f", matrix[0][2]);
            cifstring.Copy(buffer);
            t3->UpdateCell(cifstring,  6, i3);
            sprintf(buffer, "%.10f", vector[0]);
            cifstring.Copy(buffer);
            t3->UpdateCell(cifstring,  7, i3);
            sprintf(buffer, "%.10f", matrix[1][0]);
            cifstring.Copy(buffer);
            t3->UpdateCell(cifstring,  8, i3);
            sprintf(buffer, "%.10f", matrix[1][1]);
            cifstring.Copy(buffer);
            t3->UpdateCell(cifstring,  9, i3);
            sprintf(buffer, "%.10f", matrix[1][2]);
            cifstring.Copy(buffer);
            t3->UpdateCell(cifstring, 10, i3);
            sprintf(buffer, "%.10f", vector[1]);
            cifstring.Copy(buffer);
            t3->UpdateCell(cifstring, 11, i3);
            sprintf(buffer, "%.10f", matrix[2][0]);
            cifstring.Copy(buffer);
            t3->UpdateCell(cifstring, 12, i3);
            sprintf(buffer, "%.10f", matrix[2][1]);
            cifstring.Copy(buffer);
            t3->UpdateCell(cifstring, 13, i3);
            sprintf(buffer, "%.10f", matrix[2][2]);
            cifstring.Copy(buffer);
            t3->UpdateCell(cifstring, 14, i3);
            sprintf(buffer, "%.10f", vector[2]);
            cifstring.Copy(buffer);
            t3->UpdateCell(cifstring, 15, i3);
       }

       _CifObj->WriteTable(Table, _StructureId, "pdbx_struct_assembly");
       _CifObj->WriteTable(t1, _StructureId, "pdbx_struct_assembly_gen");
       _CifObj->WriteTable(t2, _StructureId, "pdbx_struct_assembly_prop");
       _CifObj->WriteTable(t3, _StructureId, "pdbx_struct_oper_list");

       rowNo = t2->GetNumRows();
       ReVarCifArray<_ASSEMBLY> assembly_lists;
       get_assembly_list(_CifObj, _StructureId, assembly_lists, keep_original_count);
       for (unsigned int i = 0; i < assembly_lists.Length(); i++) {
            if (assembly_lists[i].total_chains == 1) {
                 for (k = 0; k < rowNo; k++) {
                      get_value(cifstring, t2, 0, k);
                      if (cifstring == assembly_lists[i].assembly_id) {
                           t2->DeleteRow(k);
                      }
                 }
            }
       }
       t2->CompressTable();
       _CifObj->WriteTable(t2, _StructureId, "pdbx_struct_assembly_prop");
}

void _MAXIT::add_nmr_assembly_info()
{
       if (!(_experiment_type & EXPERIMENT_TYPE_NMR) &&
           !(_experiment_type & EXPERIMENT_TYPE_NMR_SOLID)) return;

       ISTable *t = _CifObj->GetTablePtr(_StructureId, "pdbx_struct_assembly");
       if (!is_empty_table(t)) return;

       t = _CifObj->GetTablePtr(_StructureId, "struct_biol_gen");
       if (is_empty_table(t)) return;

       TblIndexObj chainIndex;
       chainIndex.Clear();
       int num = 0;
       int rowNo = t->GetNumRows();
       int idx1 = t->GetColumnIndex("asym_id");
       int idx2 = t->GetColumnIndex("ndb_new_pdb_asym_id");
       CifString asym_ids, cs, cs1;
       asym_ids.Clear();

       for (int i = 0; i < rowNo; i++) {
            get_value(cs, t, idx1, i);
            get_value(cs1, t, idx2, i);
            if (cs != "" && cs1 != "") {
                 if (asym_ids != "") asym_ids += ",";
                 asym_ids += cs;
                 if (chainIndex.Seek(cs1) < 0) {
                      chainIndex.Add(cs1);
                      num++;
                 }
            }
       }

       t = pdbx_struct_assembly_Table();
       t->AddRow();
       cs.Copy("1");
       t->UpdateCell(cs, 0, 0);
       cs.Copy("author_defined_assembly");
       t->UpdateCell(cs, 1, 0);

       char buffer[100];
       if (num <= NUM_MER) cs.Copy(_num_mers[num]);
       else {
            sprintf(buffer, "%d-MERIC", num);
            cs.Copy(buffer);
       }
       cs.ToLower();
       t->UpdateCell(cs, 3, 0);
       sprintf(buffer, "%d", num);
       cs.Copy(buffer);
       t->UpdateCell(cs, 4, 0);

       _CifObj->WriteTable(t, _StructureId, "pdbx_struct_assembly");

       t = pdbx_struct_assembly_gen_Table();
       t->AddRow();
       cs.Copy("1");
       t->UpdateCell(cs, 0, 0);
       t->UpdateCell(cs, 1, 0);
       t->UpdateCell(asym_ids, 2, 0);

       _CifObj->WriteTable(t, _StructureId, "pdbx_struct_assembly_gen");

       t = pdbx_struct_oper_list_Table();
       t->AddRow();
       t->UpdateCell(cs, 0, 0);
       cs.Copy("identity operation");
       t->UpdateCell(cs, 1, 0);
       cs.Copy("1_555");
       t->UpdateCell(cs, 2, 0);
       cs.Copy("x,y,z");
       t->UpdateCell(cs, 3, 0);
       cs.Copy("1.0000000000");
       t->UpdateCell(cs, 4, 0);
       t->UpdateCell(cs, 9, 0);
       t->UpdateCell(cs, 14, 0);
       cs.Copy("0.0000000000");
       t->UpdateCell(cs, 5, 0);
       t->UpdateCell(cs, 6, 0);
       t->UpdateCell(cs, 7, 0);
       t->UpdateCell(cs, 8, 0);
       t->UpdateCell(cs, 10, 0);
       t->UpdateCell(cs, 11, 0);
       t->UpdateCell(cs, 12, 0);
       t->UpdateCell(cs, 13, 0);
       t->UpdateCell(cs, 15, 0);

       _CifObj->WriteTable(t, _StructureId, "pdbx_struct_oper_list");
}

int _MAXIT::find_water_tls_group(const int find_only)
{
       if (_num_mol == 0 || _molecules == NULL) return 0;

       if (!_CifObj) return 0;

       ISTable *Table = _CifObj->GetTablePtr(_StructureId, "ccp4_refine_tls_group");
       if (is_empty_table(Table)) return 0;

       int num_water_chain = 0;
       CifString chain_id;
       chain_id.Clear();
       int max_num = -100000;
       int min_num = 100000;
       _CHAIN *chain = _molecules[0]->GetFirstChain();
       while (chain) {
            if (chain->chain_type() != HETAS_TOKEN) {
                 chain = _molecules[0]->GetNextChain();
                 continue;
            }
            _RESIDUE *residue = chain->GetFirstResidue();
            if (strcmp(residue->ResName(), "HOH") &&
                strcmp(residue->ResName(), "DOD")) {
                 chain = _molecules[0]->GetNextChain();
                 continue;
            }

            max_num = -100000;
            min_num = 100000;
            num_water_chain++;
            chain_id.Copy(chain->PDB_ChainID());
            while (residue) {
                 if (residue->pdb_res_no() > max_num)
                      max_num = residue->pdb_res_no();
                 if (residue->pdb_res_no() < min_num)
                      min_num = residue->pdb_res_no();
                 residue = chain->GetNextResidue();
            }
            chain = _molecules[0]->GetNextChain();
       }

       if (num_water_chain == 0) return 0;

       int rowNo = Table->GetNumRows();
       int beg_auth_asym_id = Table->GetColumnIndex("beg_auth_asym_id");
       int end_auth_asym_id = Table->GetColumnIndex("end_auth_asym_id");
       int beg_auth_seq_id = Table->GetColumnIndex("beg_auth_seq_id");
       int end_auth_seq_id = Table->GetColumnIndex("end_auth_seq_id");
       CifString cs1, cs2, cs3, cs4;

       int tls_group = 0;
       for (int i = 0; i < rowNo; i++) {
            get_value(cs1, Table, beg_auth_asym_id, i);
            get_value(cs2, Table, end_auth_asym_id, i);
            get_value(cs3, Table, beg_auth_seq_id, i);
            get_value(cs4, Table, end_auth_seq_id, i);
            if (cs1 == "" || cs2 == "" || cs3 == "" || cs4 == "") continue;
            if (cs1 != cs2) continue;

            if (num_water_chain == 1 && chain_id == cs1 &&
                atoi(cs3.Text()) <= min_num &&
                atoi(cs4.Text()) >= max_num) return 0;

            int first = 1;
            for (int j = 0; j < _num_mol; j++) {
                 _CHAIN *chain = _molecules[j]->GetFirstChain();
                 while (chain) {
                      if (chain->chain_type() != HETAS_TOKEN ||
                          cs1 != chain->PDB_ChainID()) {
                           chain = _molecules[j]->GetNextChain();
                           continue;
                      }
                      _RESIDUE *residue = chain->GetFirstResidue();
                      if (strcmp(residue->ResName(), "HOH") &&
                          strcmp(residue->ResName(), "DOD")) {
                           chain = _molecules[j]->GetNextChain();
                           continue;
                      }
                      while (residue) {
                           if (residue->pdb_res_no() >= atoi(cs3.Text()) &&
                               residue->pdb_res_no() <= atoi(cs4.Text())) {
                                if (find_only) return 1;
                                if (first) {
                                     tls_group++;
                                     first = 0;
                                }
                                residue->set_tls_group(tls_group);
                           }
                           residue = chain->GetNextResidue();
                      }
                      chain = _molecules[j]->GetNextChain();
                 }
            }
       }

       return 0;
}

void _MAXIT::version_update_read()
{
       update_base_pair_step_read();
       update_tls_ncs_read();
       update_rcsb_molecule_read();
       find_water_tls_group(0);

       if (!__need_file_version_update) return;

       if (_num_mol == 0 || _molecules == NULL) return;

       if (!_CifObj) return;

       ISTable *Table = _CifObj->GetTablePtr(_StructureId, "struct_site_gen");
       if (!is_empty_table(Table)) cifndb_process_struct_site_gen(Table);
       Table = _CifObj->GetTablePtr(_StructureId, "rcsb_struct_link");
       if (!is_empty_table(Table)) cifndb_process_rcsb_link(Table);
       Table = _CifObj->GetTablePtr(_StructureId, "struct_conn");
       if (!is_empty_table(Table)) cifndb_process_struct_conn(Table);

       for (int i2 = 0; i2 < _n_site; i2++) {
            if (_site[i2]->numRes == 0) continue;
            _site[i2]->residue = new _RESIDUE*[_site[i2]->numRes];
            for (int j = 0; j < _site[i2]->numRes; j++) {
                 _site[i2]->residue[j] = NULL;
                 _CHAIN *chain = _molecules[0]->GetIndexChain(_site[i2]->initIndex[j]);
                 if (!chain) continue;
                 _FIELD *Init = chain->SeqResIndex(_site[i2]->initRes[j]);
                 if (!Init) continue;
                 _site[i2]->residue[j] = Init->Residue;
            }
       }

       for (int i = 0; i < _n_link; i++) {
            _CHAIN *fst = _molecules[0]->GetIndexChain(_link[i]->FstIndex);
            _CHAIN *snd = _molecules[0]->GetIndexChain(_link[i]->SndIndex);
            if (!fst || !snd) continue;
            _FIELD *Fst = fst->SeqResIndex(_link[i]->FstRes);
            _FIELD *Snd = snd->SeqResIndex(_link[i]->SndRes);
            if (!Fst || !Snd) continue;
            _link[i]->fresidue = Fst->Residue;
            _link[i]->sresidue = Snd->Residue;
            if (!_link[i]->fresidue || !_link[i]->sresidue) continue;

            _ATOM *atom = _link[i]->fresidue->find_atom(_link[i]->FstAtom);
            if (atom) strcpy(_link[i]->FstAtom, atom->pub_atmnam()); 
            atom = _link[i]->sresidue->find_atom(_link[i]->SndAtom);
            if (atom) strcpy(_link[i]->SndAtom, atom->pub_atmnam()); 
       }
}

void _MAXIT::version_update_write()
{
       if (!__need_file_version_update) return;

       if (_num_mol == 0 || _molecules == NULL) return;

       _ATOM *atom1 = NULL, *atom2 = NULL;
       int FstIndex = -1, FstRes = -1, SndIndex = -1, SndRes = -1;

       for (int i2 = 0; i2 < _n_site; i2++) {
            for (int j = 0; j < _site[i2]->numRes; j++) {
                 if (_site[i2]->residue[j] == NULL) {
                      _site[i2]->initIndex[j] = -1;
                      _site[i2]->initRes[j] = -1;
                      continue;
                 }
                 atom1 = _site[i2]->residue[j]->GetFirstAtom();
                 if (_molecules[0]->find_pdb_residue(atom1->pdb_chnid(),
                         atom1->pdb_resnam(), atom1->pdb_resnum(),
                         atom1->ins_code(), FstIndex, FstRes)) {
                      _site[i2]->initIndex[j] = FstIndex;
                      _site[i2]->initRes[j] = FstRes;
                 }
            }
       }

       for (int i = 0; i < _n_link; i++) {
            _link[i]->FstIndex = -1;
            _link[i]->FstRes = -1;
            _link[i]->SndIndex = -1;
            _link[i]->SndRes = -1;
            if (!_link[i]->fresidue || !_link[i]->sresidue) continue;

            atom1 = _link[i]->fresidue->GetFirstAtom();
            while (atom1) {
                 if (!strcmp(_link[i]->FstAtom, atom1->pub_atmnam())) break;
                 atom1 = _link[i]->fresidue->GetNextAtom();
            }
            atom2 = _link[i]->sresidue->GetFirstAtom();
            while (atom2) {
                 if (!strcmp(_link[i]->SndAtom, atom2->pub_atmnam())) break;
                 atom2 = _link[i]->sresidue->GetNextAtom();
            }
            if (!atom1 || !atom2) continue;

            if (_molecules[0]->find_pdb_residue(atom1->pdb_chnid(),
                    atom1->pdb_resnam(), atom1->pdb_resnum(),
                    atom1->ins_code(), FstIndex, FstRes) &&
                _molecules[0]->find_pdb_residue(atom2->pdb_chnid(),
                    atom2->pdb_resnam(), atom2->pdb_resnum(),
                    atom2->ins_code(), SndIndex, SndRes)) {
                 _link[i]->FstIndex = FstIndex;
                 _link[i]->FstRes = FstRes;
                 _link[i]->SndIndex = SndIndex;
                 _link[i]->SndRes = SndRes;
                 strcpy(_link[i]->FstAtom, atom1->pdb_atmnam());
                 strcpy(_link[i]->SndAtom, atom2->pdb_atmnam());
                 strcpy(_link[i]->FstName, atom1->pdb_resnam());
                 strcpy(_link[i]->SndName, atom2->pdb_resnam());
            }
       }
}

void _MAXIT::update_tls_ncs_read()
{
       update_remediation_atom_site_mapping_read(_CifObj, _StructureId);

       if (_num_mol == 0 || _molecules == NULL) return;

       if (!_CifObj) return;

       ISTable *Table = _CifObj->GetTablePtr(_StructureId, "ccp4_refine_tls_group");
       if (!is_empty_table(Table)) {
            get_coor_index(Table, _molecules[0]);
            _CifObj->WriteTable(Table, _StructureId, "ccp4_refine_tls_group");
       }

       Table = _CifObj->GetTablePtr(_StructureId, "struct_ncs_dom_lim");
       if (!is_empty_table(Table)) {
            get_coor_index(Table, _molecules[0]);
            _CifObj->WriteTable(Table, _StructureId, "struct_ncs_dom_lim");
       }
}

void _MAXIT::update_rcsb_molecule_read()
{
       if (_num_mol == 0 || _molecules == NULL) return;
       if (!_CifObj) return;

       ISTable *t = _CifObj->GetTablePtr(_StructureId, "rcsb_molecule");
       if (is_empty_table(t)) return;

       ReVarCifArray<CifString> data;
       data.Clear();
       CifString cs, cs1, cs2;
       cs.Copy(".");
       int rowNo = t->GetNumRows();
       for (int i = 0; i < rowNo; i++) data.Add(cs);
       t->AddColumn("index");
       int index = t->GetColumnIndex("index");
       t->FillColumn(data, index);

       char buffer[20];

       int auth_asym_id = t->GetColumnIndex("auth_asym_id");
       int auth_seq_id  = t->GetColumnIndex("auth_seq_id");
       for (int i = 0; i < rowNo; i++) {
            get_value(cs1, t, auth_asym_id, i);
            get_value(cs2, t, auth_seq_id, i);
            bool is_polymer_chain = false;
            if (cs2 == "") is_polymer_chain = true;

            _CHAIN *chain = _molecules[0]->GetFirstChain();
            while (chain) {
                 if (cs1 == chain->PDB_ChainID()) {
                      _RESIDUE *residue = chain->GetFirstResidue();
                      if (is_polymer_chain) {
                           sprintf(buffer, "%d", residue->fingerprint());
                           cs.Copy(buffer);
                           t->UpdateCell(cs, index, i);
                           break;
                      }
                      bool found = false;
                      while (residue) {
                           if (residue->pdb_res_no() == atoi(cs2.Text())) {
                                sprintf(buffer, "%d", residue->fingerprint());
                                cs.Copy(buffer);
                                t->UpdateCell(cs, index, i);
                                found = true;
                                break;
                           }
                           residue = chain->GetNextResidue();
                      }
                      if (found) break;
                 }
                 chain = _molecules[0]->GetNextChain();
            }
       }
       _CifObj->WriteTable(t, _StructureId, "rcsb_molecule");
}

void _MAXIT::update_base_pair_step_read()
{
       if (_num_mol == 0 || _molecules == NULL) return;

       if (!_CifObj) return;

       for (int i = 0; i < NUM_UPDATE_CATEGORY; ++i) {
            ISTable *Table = _CifObj->GetTablePtr(_StructureId,
                                _update_category[i].category);
            if (is_empty_table(Table)) return;
            for (int j = 0; j < _update_category[i].NumRecord; ++j) {
                 get_coor_index(Table, _molecules[0], _update_category[i].items[j]);
            }
            _CifObj->WriteTable(Table, _StructureId, _update_category[i].category);
       }
}

void _MAXIT::update_tls_ncs_write()
{
       update_remediation_atom_site_mapping_write(_CifObj, _StructureId);

       if (_num_mol == 0 || _molecules == NULL) return;

       if (!_CifObj) return;

       ISTable *Table = _CifObj->GetTablePtr(_StructureId, "ccp4_refine_tls_group");
       if (!is_empty_table(Table)) {
             reverse_coor_index(Table, _molecules[0]);
             _CifObj->WriteTable(Table, _StructureId, "ccp4_refine_tls_group");
       }

       Table = _CifObj->GetTablePtr(_StructureId, "struct_ncs_dom_lim");
       if (!is_empty_table(Table)) {
            reverse_coor_index(Table, _molecules[0]);
            _CifObj->WriteTable(Table, _StructureId, "struct_ncs_dom_lim");
       }

       update_rcsb_molecule_write();

       update_base_pair_step_write();
}

void _MAXIT::update_rcsb_molecule_write()
{
       ISTable *t = _CifObj->GetTablePtr(_StructureId, "rcsb_molecule");
       if (is_empty_table(t)) return;

       int rowNo = t->GetNumRows();
       int index = t->GetColumnIndex("index");
       int auth_asym_id = t->GetColumnIndex("auth_asym_id");
       int auth_seq_id  = t->GetColumnIndex("auth_seq_id");
       if (index < 0) return;

       CifString cs;
       char buffer[20];

       for (int i = 0; i < rowNo; i++) {
            get_value(cs, t, index, i);
            if (cs == "") continue;
            _RESIDUE *residue = get_residue_from_fingerprint(cs, _molecules[0]);
            if (!residue) continue;
            _ATOM *atom = residue->GetFirstAtom();

            cs.Copy(atom->pdb_chnid());
            t->UpdateCell(cs, auth_asym_id, i);

            get_value(cs, t, auth_seq_id, i);
            if (cs != "") {
                 sprintf(buffer, "%d", atom->pdb_resnum());
                 cs.Copy(buffer);
                 t->UpdateCell(cs, auth_seq_id, i);
            }
       }

       t->DeleteColumn(index);
       for (int j = 0; j < t->GetNumColumns(); j++) {
            if (t->ColumnLength(j) == 0) {
                 t->RemoveColumn(j);
                 j--;
            }
       }
       _CifObj->WriteTable(t, _StructureId, "rcsb_molecule");
}

void _MAXIT::update_base_pair_step_write()
{
       if (_num_mol == 0 || _molecules == NULL) return;

       if (!_CifObj) return;

       ISTable *Table = NULL;
       for (int i = 0; i < NUM_UPDATE_CATEGORY; ++i) {
            Table = _CifObj->GetTablePtr(_StructureId,
                                _update_category[i].category);
            if (is_empty_table(Table)) return;
            for (int j = 0; j < _update_category[i].NumRecord; ++j) {
                 reverse_coor_index(Table, _molecules[0], _update_category[i].items[j]);
            }
            _CifObj->WriteTable(Table, _StructureId, _update_category[i].category);
       }

       Table = _CifObj->GetTablePtr(_StructureId, "ndb_struct_na_base_pair");
       if (!is_empty_table(Table)) {
            int rowNo = Table->GetNumRows();
            int i_auth_asym_id = Table->GetColumnIndex("i_auth_asym_id");
            int i_label_comp_id = Table->GetColumnIndex("i_label_comp_id");
            int i_auth_seq_id = Table->GetColumnIndex("i_auth_seq_id");
            int i_PDB_ins_code = Table->GetColumnIndex("i_PDB_ins_code");
            int j_auth_asym_id = Table->GetColumnIndex("j_auth_asym_id");
            int j_label_comp_id = Table->GetColumnIndex("j_label_comp_id");
            int j_auth_seq_id = Table->GetColumnIndex("j_auth_seq_id");
            int j_PDB_ins_code = Table->GetColumnIndex("j_PDB_ins_code");
            int pair_name = Table->GetColumnIndex("pair_name");
            CifString cs1, cs2, cs3, cs4, cs5, cs6, cs7, cs8;
            for (int i = 0; i < rowNo; ++i) {
                 get_value(cs1, Table, i_auth_asym_id, i);
                 get_value(cs2, Table, i_label_comp_id, i);
                 get_value(cs3, Table, i_auth_seq_id, i);
                 get_value(cs4, Table, i_PDB_ins_code, i);
                 get_value(cs5, Table, j_label_comp_id, i);
                 get_value(cs6, Table, j_auth_seq_id, i);
                 get_value(cs7, Table, j_PDB_ins_code, i);
                 get_value(cs8, Table, j_auth_asym_id, i);
                 cs1 += "_"; cs1 += cs2; cs1 += cs3; cs1 += cs4; cs1 += ":";
                 cs1 += cs5; cs1 += cs6; cs1 += cs7; cs1 += "_"; cs1 += cs8;
                 Table->UpdateCell(cs1, pair_name, i);
            }
            _CifObj->WriteTable(Table, _StructureId, "ndb_struct_na_base_pair");
       }

       Table = _CifObj->GetTablePtr(_StructureId, "ndb_struct_na_base_pair_step");
       if (!is_empty_table(Table)) {
            int rowNo = Table->GetNumRows();
            int i_auth_asym_id_1 = Table->GetColumnIndex("i_auth_asym_id_1");
            int i_label_comp_id_1 = Table->GetColumnIndex("i_label_comp_id_1");
            int i_auth_seq_id_1 = Table->GetColumnIndex("i_auth_seq_id_1");
            int i_PDB_ins_code_1 = Table->GetColumnIndex("i_PDB_ins_code_1");
            int j_auth_asym_id_1 = Table->GetColumnIndex("j_auth_asym_id_1");
            int j_label_comp_id_1 = Table->GetColumnIndex("j_label_comp_id_1");
            int j_auth_seq_id_1 = Table->GetColumnIndex("j_auth_seq_id_1");
            int j_PDB_ins_code_1 = Table->GetColumnIndex("j_PDB_ins_code_1");
            int i_auth_asym_id_2 = Table->GetColumnIndex("i_auth_asym_id_2");
            int i_label_comp_id_2 = Table->GetColumnIndex("i_label_comp_id_2");
            int i_auth_seq_id_2 = Table->GetColumnIndex("i_auth_seq_id_2");
            int i_PDB_ins_code_2 = Table->GetColumnIndex("i_PDB_ins_code_2");
            int j_auth_asym_id_2 = Table->GetColumnIndex("j_auth_asym_id_2");
            int j_label_comp_id_2 = Table->GetColumnIndex("j_label_comp_id_2");
            int j_auth_seq_id_2 = Table->GetColumnIndex("j_auth_seq_id_2");
            int j_PDB_ins_code_2 = Table->GetColumnIndex("j_PDB_ins_code_2");
            int step_name = Table->GetColumnIndex("step_name");
            CifString cs1, cs2, cs3, cs4, cs5, cs6, cs7, cs8, cs9, cs10, cs11, cs12;
            CifString cs13, cs14, cs15, cs16;
            for (int i = 0; i < rowNo; ++i) {
                 get_value(cs1, Table, i_auth_asym_id_1, i);
                 get_value(cs2, Table, i_auth_asym_id_2, i);
                 get_value(cs3, Table, i_label_comp_id_1, i);
                 get_value(cs4, Table, i_auth_seq_id_1, i);
                 get_value(cs5, Table, i_PDB_ins_code_1, i);
                 get_value(cs6, Table, i_label_comp_id_2, i);
                 get_value(cs7, Table, i_auth_seq_id_2, i);
                 get_value(cs8, Table, i_PDB_ins_code_2, i);
                 get_value(cs9, Table, j_label_comp_id_2, i);
                 get_value(cs10, Table, j_auth_seq_id_2, i);
                 get_value(cs11, Table, j_PDB_ins_code_2, i);
                 get_value(cs12, Table, j_label_comp_id_1, i);
                 get_value(cs13, Table, j_auth_seq_id_1, i);
                 get_value(cs14, Table, j_PDB_ins_code_1, i);
                 get_value(cs15, Table, j_auth_asym_id_2, i);
                 get_value(cs16, Table, j_auth_asym_id_1, i);
                 cs1 += cs2; cs1 += "_"; cs1 += cs3; cs1 += cs4; cs1 += cs5;
                 cs1 += cs6; cs1 += cs7; cs1 += cs8; cs1 += ":"; cs1 += cs9;
                 cs1 += cs10; cs1 += cs11; cs1 += cs12; cs1 += cs13;
                 cs1 += cs14; cs1 += "_"; cs1 += cs15; cs1 += cs16;
                 Table->UpdateCell(cs1, step_name, i);
            }
            _CifObj->WriteTable(Table, _StructureId, "ndb_struct_na_base_pair_step");
       }
}

static void get_coor_index(ISTable *t, _MOLECULE *molecule)
{
       int beg_auth_asym_id = t->GetColumnIndex("beg_auth_asym_id");
       int beg_auth_seq_id = t->GetColumnIndex("beg_auth_seq_id");
       int end_auth_asym_id = t->GetColumnIndex("end_auth_asym_id");
       int end_auth_seq_id = t->GetColumnIndex("end_auth_seq_id");
       if (beg_auth_asym_id < 0 || beg_auth_seq_id < 0 ||
           end_auth_asym_id < 0 || end_auth_seq_id < 0) return;

       ReVarCifArray<CifString> data;
       data.Clear();
       CifString cs1;
       cs1.Copy(".");
       int rowNo = t->GetNumRows();
       for (int i = 0; i < rowNo; i++) data.Add(cs1);
       t->AddColumn("beg_index");
       t->AddColumn("end_index");
       int beg_index = t->GetColumnIndex("beg_index");
       int end_index = t->GetColumnIndex("end_index");
       t->FillColumn(data, beg_index);
       t->FillColumn(data, end_index);

       CifString cs2, cs3, cs4;
       for (int i = 0; i < rowNo; i++) {
            get_value(cs1, t, beg_auth_asym_id, i);
            get_value(cs2, t, beg_auth_seq_id, i);
            get_value(cs3, t, end_auth_asym_id, i);
            get_value(cs4, t, end_auth_seq_id, i);
            if (cs1 == cs3 && cs2 == "-1" && cs4 == "-1") {
                 get_first_and_last_residue_number(cs1, cs2, cs4, molecule);
                 t->UpdateCell(cs2, beg_auth_seq_id, i);
                 t->UpdateCell(cs4, end_auth_seq_id, i);
            }

            if (cs1 != "" && cs2 != "") {
                 get_coor_index1(cs1, cs2, molecule, 0);
                 t->UpdateCell(cs1, beg_index, i);
            }
            if (cs3 != "" && cs4 != "") {
                 get_coor_index1(cs3, cs4, molecule, 1);
                 t->UpdateCell(cs3, end_index, i);
            }
       }
}

static void get_coor_index(ISTable *t, _MOLECULE *molecule, const _UPDATE_ITEMS& items)
{
       int auth_asym_id = t->GetColumnIndex(items.auth_asym_id);
       int auth_comp_id = t->GetColumnIndex(items.auth_comp_id);
       int auth_seq_id = t->GetColumnIndex(items.auth_seq_id);
       int auth_ins_code = t->GetColumnIndex(items.auth_ins_code);
       if (auth_asym_id < 0 || auth_comp_id < 0 ||
           auth_seq_id < 0 || auth_ins_code < 0) return;

       ReVarCifArray<CifString> data;
       data.Clear();
       CifString cs1;
       cs1.Copy(".");
       int rowNo = t->GetNumRows();
       for (int i = 0; i < rowNo; ++i) data.Add(cs1);
       t->AddColumn(items.temp_key);
       int temp_index = t->GetColumnIndex(items.temp_key);
       t->FillColumn(data, temp_index);

       CifString cs2, cs3, cs4;
       char buffer[20];
       for (int i = 0; i < rowNo; ++i) {
            get_value(cs1, t, auth_asym_id, i);
            get_value(cs2, t, auth_comp_id, i);
            get_value(cs3, t, auth_seq_id, i);
            get_value(cs4, t, auth_ins_code, i);
            int fingerprint = get_fingerprint(cs1, cs2, cs3, cs4, molecule);
            if (fingerprint >= 0) {
                 sprintf(buffer, "%d", fingerprint);
                 cs1.Copy(buffer);
                 t->UpdateCell(cs1, temp_index, i);
            }
       }
}

static void get_first_and_last_residue_number(CifString &pdb_chnid, CifString &first_num,
              CifString &last_num, _MOLECULE *molecule)
{
       int first = 100000, last = -100000;

       _CHAIN *chain = molecule->GetFirstChain();
       while (chain) {
            if (pdb_chnid == chain->PDB_ChainID()) {
                 _RESIDUE *residue = chain->GetFirstResidue();
                 while (residue) {
                      if (residue->pdb_res_no() < first) first = residue->pdb_res_no();
                      if (residue->pdb_res_no() > last) last = residue->pdb_res_no();
                      residue = chain->GetNextResidue();
                 }
            }
            chain = molecule->GetNextChain();
       }

       if (first != 100000 && last != (-100000)) {
            char buffer[20];
            sprintf(buffer, "%d", first);
            first_num.Copy(buffer);
            sprintf(buffer, "%d", last);
            last_num.Copy(buffer);
       }
}

static void get_coor_index1(CifString &pdb_chnid, const CifString &pdb_resnum,
               _MOLECULE *molecule, const int flag)
{
       int number = 100000;
       if (flag) number = -100000;
       int index = -1;

       int range = atoi(pdb_resnum.Text());
       _CHAIN *chain = molecule->GetFirstChain();
       while (chain) {
            if (pdb_chnid == chain->PDB_ChainID()) {
                 _RESIDUE *residue = chain->GetFirstResidue();
                 while (residue) {
                      if (flag) {
                           if (residue->pdb_res_no() <= range) {
                                if (residue->pdb_res_no() > number) {
                                     number = residue->pdb_res_no();
                                     index = residue->fingerprint();
                                }
                           }
                      } else {
                           if (residue->pdb_res_no() >= range) {
                                if (residue->pdb_res_no() < number) {
                                     number = residue->pdb_res_no();
                                     index = residue->fingerprint();
                                }
                           }
                      }
                      residue = chain->GetNextResidue();
                 }
            }
            chain = molecule->GetNextChain();
       }
       pdb_chnid.Copy(".");
       if (index != (-1)) {
            char buffer[20];
            sprintf(buffer, "%d", index);
            pdb_chnid.Copy(buffer);
       }
}

static int get_fingerprint(const CifString &pdb_chnid, const CifString &pdb_resnam,
      const CifString &pdb_resnum, const CifString &ins_code, _MOLECULE *molecule)
{
       _CHAIN *chain = molecule->GetFirstChain();
       while (chain) {
            if (pdb_chnid == chain->PDB_ChainID()) {
                 _RESIDUE *residue = chain->GetFirstResidue();
                 while (residue) {
                      _ATOM *atom = residue->GetFirstAtom();
                      if (atom->pdb_resnum() == atoi(pdb_resnum.Text()) &&
                          pdb_resnam == atom->pdb_resnam() &&
                          ins_code == atom->ins_code()) {
                           return residue->fingerprint();
                      }
                      residue = chain->GetNextResidue();
                 }
            }
            chain = molecule->GetNextChain();
       }
       return (-1);
}

static void reverse_coor_index(ISTable *t, _MOLECULE *molecule)
{
       int beg_index = t->GetColumnIndex("beg_index");
       int end_index = t->GetColumnIndex("end_index");
       if (beg_index < 0 || end_index < 0) {
            if (beg_index >= 0) t->DeleteColumn(beg_index);
            if (end_index >= 0) t->DeleteColumn(end_index);
            for (int j = 0; j < t->GetNumColumns(); j++) {
                 if (t->ColumnLength(j) == 0) {
                      t->RemoveColumn(j);
                      j--;
                 }
            }
            return;
       }
    
       int beg_auth_asym_id = t->GetColumnIndex("beg_auth_asym_id");
       int beg_auth_seq_id = t->GetColumnIndex("beg_auth_seq_id");
       int end_auth_asym_id = t->GetColumnIndex("end_auth_asym_id");
       int end_auth_seq_id = t->GetColumnIndex("end_auth_seq_id");

       int rowNo = t->GetNumRows();
       CifString cs1, cs2, cs3, cs4, cs5, x, y, z;

       for (int i = 0; i < rowNo; i++) {
            get_value(cs1, t, beg_auth_asym_id, i);
            get_value(cs2, t, beg_auth_seq_id, i);
            get_value(cs3, t, beg_index, i);
            reverse_coor_index1(cs1, cs2, cs3, molecule, 0);
            t->UpdateCell(cs1, beg_auth_asym_id, i);
            t->UpdateCell(cs2, beg_auth_seq_id, i);

            get_value(cs1, t, end_auth_asym_id, i);
            get_value(cs2, t, end_auth_seq_id, i);
            get_value(cs3, t, end_index, i);
            reverse_coor_index1(cs1, cs2, cs3, molecule, 1);
            t->UpdateCell(cs1, end_auth_asym_id, i);
            t->UpdateCell(cs2, end_auth_seq_id, i);
       }

       t->DeleteColumn(beg_index);
       t->DeleteColumn(end_index);
       for (int j = 0; j < t->GetNumColumns(); j++) {
            if (t->ColumnLength(j) == 0) {
                 t->RemoveColumn(j);
                 j--;
            }
       }
}

static void reverse_coor_index(ISTable *t, _MOLECULE *molecule, const _UPDATE_ITEMS& items)
{
       int temp_index = t->GetColumnIndex(items.temp_key);
       if (temp_index < 0) return;
    
       int label_asym_id = t->GetColumnIndex(items.label_asym_id);
       int label_comp_id = t->GetColumnIndex(items.label_comp_id);
       int label_seq_id = t->GetColumnIndex(items.label_seq_id);
       int auth_asym_id = t->GetColumnIndex(items.auth_asym_id);
       int auth_comp_id = t->GetColumnIndex(items.auth_comp_id);
       int auth_seq_id = t->GetColumnIndex(items.auth_seq_id);
       int auth_ins_code = t->GetColumnIndex(items.auth_ins_code);

       int rowNo = t->GetNumRows();
       CifString cs;
       char buffer[20];

       for (int i = 0; i < rowNo; i++) {
            get_value(cs, t, temp_index, i);
            if (cs == "") continue;
            _RESIDUE *residue = get_residue_from_fingerprint(cs, molecule);
            if (!residue) continue;
            _ATOM *atom = residue->GetFirstAtom();
            if (label_asym_id >= 0) {
                 cs.Copy(atom->chnid());
                 t->UpdateCell(cs, label_asym_id, i);
            }
            if (label_comp_id >= 0) {
                 cs.Copy(atom->restype());
                 t->UpdateCell(cs, label_comp_id, i);
            }
            if (label_seq_id >= 0) {
                 sprintf(buffer, "%d", atom->resnum());
                 cs.Copy(buffer);
                 t->UpdateCell(cs, label_seq_id, i);
            }
            if (auth_asym_id >= 0) {
                 cs.Copy(atom->pdb_chnid());
                 t->UpdateCell(cs, auth_asym_id, i);
            }
            if (auth_comp_id >= 0) {
                 cs.Copy(atom->pdb_resnam());
                 t->UpdateCell(cs, auth_comp_id, i);
            }
            if (auth_seq_id >= 0) {
                 sprintf(buffer, "%d", atom->pdb_resnum());
                 cs.Copy(buffer);
                 t->UpdateCell(cs, auth_seq_id, i);
            }
            if (auth_ins_code >= 0) {
                 cs.Copy(atom->ins_code());
                 t->UpdateCell(cs, auth_ins_code, i);
            }
       }

       t->DeleteColumn(temp_index);
       for (int j = 0; j < t->GetNumColumns(); j++) {
            if (t->ColumnLength(j) == 0) {
                 t->RemoveColumn(j);
                 j--;
            }
       }
}

static void reverse_coor_index1(CifString &pdb_chnid, CifString &pdb_num,
              const CifString &Idx, _MOLECULE *molecule, const int flag)
{
       int fingerprint = atoi(Idx.Text());
       int range = atoi(pdb_num.Text());
       char buffer[20];
       _CHAIN *chain = molecule->GetFirstChain();
       while (chain) {
            _RESIDUE *residue = chain->GetFirstResidue();
            while (residue) {
                 if (residue->fingerprint() == fingerprint) {
                      pdb_chnid.Copy(chain->PDB_ChainID());
                      if (flag) {
                           if (residue->pdb_res_no() > range) {
                                sprintf(buffer, "%d", residue->pdb_res_no());
                                pdb_num.Copy(buffer);
                           }
                      } else {
                           if (residue->pdb_res_no() < range) {
                                sprintf(buffer, "%d", residue->pdb_res_no());
                                pdb_num.Copy(buffer);
                           }
                      }
                      return;
                 }
                 residue = chain->GetNextResidue();
            }
            chain = molecule->GetNextChain();
       }
}

static _RESIDUE *get_residue_from_fingerprint(const CifString &Idx, _MOLECULE *molecule)
{
       int fingerprint = atoi(Idx.Text());
       _CHAIN *chain = molecule->GetFirstChain();
       while (chain) {
            _RESIDUE *residue = chain->GetFirstResidue();
            while (residue) {
                 if (residue->fingerprint() == fingerprint) return residue;
                 residue = chain->GetNextResidue();
            }
            chain = molecule->GetNextChain();
       }

       return NULL;
}

static void get_assembly_list(CifFileObj *fobj, const char *block_id,
              ReVarCifArray<_ASSEMBLY> &assembly_lists, const bool keep_original_count)
{
       ISTable *t = fobj->GetTablePtr(block_id, "pdbx_struct_assembly");

       CifString cs1, cs2, cs3, cs4;
       _ASSEMBLY assembly;
       assembly_lists.Clear();
       int rowNo = t->GetNumRows();
       int idx1 = t->GetColumnIndex("id");
       int idx2 = t->GetColumnIndex("details");
       int idx3 = t->GetColumnIndex("method_details");
       int idx4 = t->GetColumnIndex("oligomeric_count"); 
       for (int i2 = 0; i2 < rowNo; i2++) {
            get_value(cs2, t, idx2, i2);
            cs2.ToLower();
            if (cs2 == "representative helical assembly" ||
                cs2 == "complete icosahedral assembly" ||
                cs2 == "complete point assembly" ||
                cs2 == "author_and_software_defined_assembly" ||
                cs2 == "author_defined_assembly" ||
                cs2 == "software_defined_assembly") {
                 assembly.type.Copy(cs2);
                 get_value(cs1, t, idx1, i2);
                 assembly.assembly_id.Copy(cs1);
                 get_value(cs3, t, idx3, i2);
                 assembly.software.Copy(cs3);
                 get_value(cs4, t, idx4, i2);
                 assembly.total_chains = atoi(cs4.Text());
                 assembly.absa.Clear();  
                 assembly.ssa.Clear();
                 assembly.more.Clear();
                 assembly.coplists.Clear();
                 assembly_lists.Add(assembly);
            }
       }
       if (assembly_lists.Length() == 0) return;

       t = fobj->GetTablePtr(block_id, "struct_asym");

       ReVarCifArray<_ASYM_CHAINID_CORR> asym_chainid_corrs;
       asym_chainid_corrs.Clear();
       if (!is_empty_table(t)) {
            rowNo = t->GetNumRows();
            idx1 = t->GetColumnIndex("id");
            idx2 = t->GetColumnIndex("ndb_pdb_id");
            idx3 = t->GetColumnIndex("ndb_type");
            for (int i = 0; i < rowNo; i++) {
                 get_value(cs1, t, idx1, i);
                 get_value(cs2, t, idx2, i);
                 get_value(cs3, t, idx3, i);
                 if (cs1 != "" && cs2 != "") {
                      _ASYM_CHAINID_CORR coor;
                      coor.asym_id.Copy(cs1);
                      coor.pdb_chnid.Copy(cs2);
                      if (cs3 == "ATOMP" || cs3 == "ATOMN")
                           coor.is_polymer = 1;
                      else coor.is_polymer = 0;
                      asym_chainid_corrs.Add(coor);
                 }
            }
       }

       t = fobj->GetTablePtr(block_id, "pdbx_struct_assembly_gen");
       rowNo = t->GetNumRows();
       idx1 = t->GetColumnIndex("assembly_id");
       idx2 = t->GetColumnIndex("asym_id_list");
       idx3 = t->GetColumnIndex("oper_expression");
       _CHAIN_OPLIST coplist;
       TblIndexObj ChainIDIndex, PolymerChainIDIndex;
       int len;
       char *buff = NULL, *p = NULL, buffer[80];
       for (unsigned int i = 0; i < assembly_lists.Length(); i++) {
            int count = 0;
            for (int j = 0; j < rowNo; j++) {
                 get_value(cs1, t, idx1, j);
                 get_value(cs2, t, idx2, j);
                 get_value(cs3, t, idx3, j);
                 if (cs1 == assembly_lists[i].assembly_id) {
                      len = strlen(cs2.Text());
                      buff = new char[len + 1];
                      memset(buff, 0, len + 1);
                      strcpy(buff, cs2.Text());
                      for (int k = 0; k < len; k++) {
                           if (buff[k] == ',' || buff[k] == ';') buff[k] = ' ';
                      }
                      ndb_clean_string(buff);
                      p = buff;
                      coplist.inst_ids.Clear();
                      len = 0;
                      ChainIDIndex.Clear();
                      PolymerChainIDIndex.Clear();
                      while (p) {
                           p = get_word(p, buffer);
                           if (!strcmp(buffer, "")) continue;
                           cs2.Copy(buffer);
                           int is_polymer = 0;
                           for (unsigned int k = 0; k < asym_chainid_corrs.Length(); k++) {
                                if (asym_chainid_corrs[k].asym_id == buffer) {
                                     cs2.Copy(asym_chainid_corrs[k].pdb_chnid);
                                     if (asym_chainid_corrs[k].is_polymer) is_polymer = 1;
                                     break;
                                }
                           }
                           if (is_polymer) {
                                if (PolymerChainIDIndex.Seek(cs2) < 0) {
                                     PolymerChainIDIndex.Add(cs2);
                                     len++;
                                }
                           }
                           if (ChainIDIndex.Seek(cs2) < 0) {
                                ChainIDIndex.Add(cs2);
                                if (coplist.inst_ids != "") coplist.inst_ids += ", ";
                                coplist.inst_ids += cs2;
                           }
                      }
                      delete [] buff;
    
                      coplist.oplist.Clear();
                      parseString(cs3, coplist.oplist);
                      if (coplist.oplist.Length() > 0) {
                           assembly_lists[i].coplists.Add(coplist);
                      }
                      count += len * coplist.oplist.Length();
                 }
            }
            if (!keep_original_count || !assembly_lists[i].total_chains)
                 assembly_lists[i].total_chains = count;
       }

       t = fobj->GetTablePtr(block_id, "pdbx_struct_assembly_prop");
       if (!is_empty_table(t)) {
            rowNo = t->GetNumRows();
            idx1 = t->GetColumnIndex("biol_id");
            idx2 = t->GetColumnIndex("type");
            idx3 = t->GetColumnIndex("value");
            for (unsigned int i = 0; i < assembly_lists.Length(); i++) {
                 for (int j = 0; j < rowNo; j++) {
                      get_value(cs1, t, idx1, j);
                      get_value(cs2, t, idx2, j);
                      get_value(cs3, t, idx3, j);
                      if (cs1 == assembly_lists[i].assembly_id) {
                           if (cs2 == "ABSA (A^2)")
                                assembly_lists[i].absa.Copy(cs3);
                           else if (cs2 == "SSA (A^2)")
                                assembly_lists[i].ssa.Copy(cs3);
                           else if (cs2 == "MORE")
                                assembly_lists[i].more.Copy(cs3);
                      }
                 }
            }
       }
}

static void get_inhibitor_details(CifFileObj *fobj, const char *blockId,
                    const int entity_id, CifString &type, CifString &source,
                    CifString &s_details, CifString &o_details)
{
       type.Clear();
       source.Clear();
       s_details.Clear();
       o_details.Clear();

       if (fobj == NULL) return;

       const char *_source_cat[2][3] = {
             { "entity_src_gen", "ndb_gene_src_scientific_name",
                                 "pdbx_gene_src_ncbi_taxonomy_id" },
             { "entity_src_nat", "ndb_organism_scientific", "pdbx_ncbi_taxonomy_id" }
       };

       CifString cs;
       bool found = false;
       for (int i = 0; i < 2; ++i) {
            ISTable *t = fobj->GetTablePtr(blockId, _source_cat[i][0]);
            if (is_empty_table(t)) continue;
            int idx1 = t->GetColumnIndex("entity_id");
            int idx2 = t->GetColumnIndex(_source_cat[i][1]);
            int idx3 = t->GetColumnIndex(_source_cat[i][2]);
            int rowNo = t->GetNumRows();
            for (int j = 0; j < rowNo; ++j) {
                 get_value(cs, t, idx1, j);
                 if (atoi(cs.Text()) == entity_id) {
                      get_value(source, t, idx2, j);
                      source.ToUpper();
                      get_value(s_details, t, idx3, j);
                      s_details.ToUpper();
                      found = true;
                      break;
                 }
            }
            if (found) break;
       }
}

static COORD get_coordinate(_ATOM *atom, const NDBCELLCONST *cell, const char *SymOP)
{
       COORD coord = atom->orig();
       if (strcmp(SymOP, "") && strcmp(SymOP, "1_555")) {
            int op = atoi(SymOP) - 1;
            int t[3], j;
            int len = strlen(SymOP);
            t[0] = t[1] = t[2] = 0;
            for (j = 0; j < len; j++) {
                 if (SymOP[j] == '_') break;
            }
            if (j + 3 < len) {
                 for (int k = 0; k < 3; k++) {
                      t[k] = SymOP[j + k + 1] - '5';
                 }
            }
            symmetry_operation(coord, cell, op, t[0], t[1], t[2]);
       }
       return coord;
}

static ISTable *pdbx_struct_assembly_Table()
{
       ISTable *t = new ISTable("pdbx_struct_assembly");
       t->AddColumn("id");
       t->AddColumn("details");
       t->AddColumn("method_details");
       t->AddColumn("oligomeric_details");
       t->AddColumn("oligomeric_count");
       return t;
}

static ISTable *pdbx_struct_assembly_gen_Table()
{
       ISTable *t = new ISTable("pdbx_struct_assembly_gen");
       t->AddColumn("assembly_id");
       t->AddColumn("oper_expression");
       t->AddColumn("asym_id_list");
       return t;
}

static ISTable *pdbx_struct_oper_list_Table()
{
       ISTable *t3 = new ISTable("pdbx_struct_oper_list");
       t3->AddColumn("id");
       t3->AddColumn("type");
       t3->AddColumn("name");
       t3->AddColumn("symmetry_operation");
       t3->AddColumn("matrix[1][1]");
       t3->AddColumn("matrix[1][2]");
       t3->AddColumn("matrix[1][3]");
       t3->AddColumn("vector[1]");
       t3->AddColumn("matrix[2][1]");
       t3->AddColumn("matrix[2][2]");
       t3->AddColumn("matrix[2][3]");
       t3->AddColumn("vector[2]");
       t3->AddColumn("matrix[3][1]");
       t3->AddColumn("matrix[3][2]");
       t3->AddColumn("matrix[3][3]");
       t3->AddColumn("vector[3]");
       return t3;
}

static void update_remediation_atom_site_mapping_read(CifFileObj *fobj,
                       const char *block_id)
{
       if (!fobj) return;

       ISTable *mTable = fobj->GetTablePtr(block_id, "pdbx_remediation_atom_site_mapping");
       if (is_empty_table(mTable)) return;

       ISTable *aTable = fobj->GetTablePtr(block_id, "atom_site");
       if (is_empty_table(aTable)) return;

       const char *_mitems[6] = { "auth_atom_id", "label_alt_id", "auth_comp_id",
                                  "auth_asym_id", "auth_seq_id", "PDB_ins_code" };
       const char *_aitems[6] = { "auth_atom_id", "label_alt_id", "auth_comp_id",
                                  "auth_asym_id", "auth_seq_id", "ndb_ins_code" };
       const char *_xitems[3] = { "Cartn_x", "Cartn_y", "Cartn_z" };
       int mindex[6], aindex[6], xindex[3];
       for (int i = 0; i < 6; ++i) {
            mindex[i] = mTable->GetColumnIndex(_mitems[i]);
            aindex[i] = aTable->GetColumnIndex(_aitems[i]);
       }
       for (int i = 0; i < 3; ++i) xindex[i] = aTable->GetColumnIndex(_xitems[i]);

       std::map<std::string, std::string> name_coor_mapping;
       name_coor_mapping.clear();

       std::string cs1, cs2;
       CifString cs, model;
       int model_index = aTable->GetColumnIndex("ndb_model");
       get_value(model, aTable, model_index, 0);
       
       int rowNo = aTable->GetNumRows();
       for (int i = 0; i < rowNo; ++i) {
            get_value(cs, aTable, model_index, i);
            if (cs != model) break;
            cs1.clear();
            for (int j = 0; j < 6; ++j) {
                 get_value(cs, aTable, aindex[j], i);
                 if (j) cs1 += "_";
                 cs1 += cs.Text();
            }
            cs2.clear();
            for (int j = 0; j < 3; ++j) {
                 get_value(cs, aTable, xindex[j], i);
                 if (j) cs2 += "_";
                 cs2 += cs.Text();
            }
            name_coor_mapping.insert(make_pair(cs1, cs2));
       }

       ReVarCifArray<CifString> data;
       data.Clear();

       rowNo = mTable->GetNumRows();
       for (int i = 0; i < rowNo; ++i) {
            cs1.clear();
            for (int j = 0; j < 6; ++j) {
                 get_value(cs, mTable, mindex[j], i);
                 if (j) cs1 += "_";
                 cs1 += cs.Text();
            }
            std::map<std::string, std::string>::iterator pos
                     = name_coor_mapping.find(cs1);
            if (pos != name_coor_mapping.end()) {
                 cs.Copy(pos->second.c_str());
                 data.Add(cs);
            }
       }

       if (rowNo == (int) data.Length()) {
            mTable->AddColumn("tmp_index");
            int index = mTable->GetColumnIndex("tmp_index");
            mTable->FillColumn(data, index);
            fobj->WriteTable(mTable, block_id, "pdbx_remediation_atom_site_mapping");
       }
}

static void update_remediation_atom_site_mapping_write(CifFileObj *fobj,
                       const char *block_id)
{
       if (!fobj) return;

       ISTable *mTable = fobj->GetTablePtr(block_id, "pdbx_remediation_atom_site_mapping");
       if (is_empty_table(mTable)) return;

       int index = mTable->GetColumnIndex("tmp_index");
       if (index < 0) return;

       ISTable *aTable = fobj->GetTablePtr(block_id, "atom_site");
       if (is_empty_table(aTable)) {
            mTable->DeleteColumn(index);
            for (int j = 0; j < mTable->GetNumColumns(); j++) {
                 if (mTable->ColumnLength(j) == 0) {
                      mTable->RemoveColumn(j);
                      j--;
                 }
            }
            fobj->WriteTable(mTable, block_id, "pdbx_remediation_atom_site_mapping");
            return;
       }

       const char *_mitems[11] = { "label_atom_id", "label_comp_id", "label_asym_id",
                   "label_seq_id", "label_alt_id", "auth_atom_id", "auth_comp_id",
                   "auth_asym_id", "auth_seq_id", "PDB_ins_code", "occupancy" };

       const char *_aitems[11] = { "label_atom_id", "label_comp_id", "label_asym_id",
                   "label_seq_id", "label_alt_id", "auth_atom_id", "auth_comp_id",
                   "auth_asym_id", "auth_seq_id", "ndb_ins_code", "occupancy" };

       const char *_xitems[3] = { "Cartn_x", "Cartn_y", "Cartn_z" };
       int mindex[11], aindex[11], xindex[3];
       for (int i = 0; i < 11; ++i) {
            mindex[i] = mTable->GetColumnIndex(_mitems[i]);
            aindex[i] = aTable->GetColumnIndex(_aitems[i]);
       }
       for (int i = 0; i < 3; ++i) xindex[i] = aTable->GetColumnIndex(_xitems[i]);

       std::map<std::string, std::vector<std::string> > name_coor_mapping;
       name_coor_mapping.clear();
       std::vector<std::string> data;

       std::string cs2;
       CifString cs, model;
       int model_index = aTable->GetColumnIndex("ndb_model");
       get_value(model, aTable, model_index, 0);
       
       int rowNo = aTable->GetNumRows();
       for (int i = 0; i < rowNo; ++i) {
            get_value(cs, aTable, model_index, i);
            if (cs != model) break;
            data.clear();
            for (int j = 0; j < 11; ++j) {
                 get_value(cs, aTable, aindex[j], i);
                 data.push_back(cs.Text());
            }
            cs2.clear();
            for (int j = 0; j < 3; ++j) {
                 get_value(cs, aTable, xindex[j], i);
                 if (j) cs2 += "_";
                 cs2 += cs.Text();
            }
            name_coor_mapping.insert(make_pair(cs2, data));
       }

       rowNo = mTable->GetNumRows();
       bool found = true;
       for (int i = 0; i < rowNo; ++i) {
            get_value(cs, mTable, index, i);
            if (name_coor_mapping.find(cs.Text()) == name_coor_mapping.end()) {
                 found = false;
                 break;
            }
       }

       if (found) {
            for (int i = 0; i < rowNo; ++i) {
                 get_value(cs, mTable, index, i);
                 std::map<std::string, std::vector<std::string> >::iterator pos =
                             name_coor_mapping.find(cs.Text());
                 if (pos == name_coor_mapping.end()) continue;
                 for (int j = 0; j < 11; ++j) {
                      if (!pos->second[j].empty())
                           cs.Copy(pos->second[j].c_str());
                      else cs.Copy(".");
                      mTable->UpdateCell(cs, mindex[j], i);
                 }
            }
       }

       mTable->DeleteColumn(index);
       for (int j = 0; j < mTable->GetNumColumns(); j++) {
            if (mTable->ColumnLength(j) == 0) {
                 mTable->RemoveColumn(j);
                 j--;
            }
       }

       fobj->WriteTable(mTable, block_id, "pdbx_remediation_atom_site_mapping");
}

static void get_prd_info(CifFileObj *fobj, const char *blockId, std::map<std::string,
                         std::vector<std::string> >& prd_info)
{
       prd_info.clear();
       if (!fobj) return;

       ISTable *t1 = fobj->GetTablePtr(blockId, "rcsb_molecule");
       ISTable *t2 = fobj->GetTablePtr(blockId, "pdbx_molecule_features");
       if (is_empty_table(t1) || is_empty_table(t2)) return;

       ISTable *t3 = fobj->GetTablePtr(blockId, "ndb_nonpoly_scheme");
       if (is_empty_table(t3)) return;

       std::map<std::string, std::string> nonpoly_mapping;
       nonpoly_mapping.clear();
       int rowNo = t3->GetNumRows();
       CifString cs, cs1, cs2, cs3;
       int idx1 = t3->GetColumnIndex("pdb_id");
       int idx2 = t3->GetColumnIndex("pdb_num");
       int idx3 = t3->GetColumnIndex("pdb_mon_id");
       for (int i = 0; i < rowNo; ++i) {
            get_value(cs, t3, idx1, i);
            get_value(cs1, t3, idx2, i);
            get_value(cs2, t3, idx3, i);
            std::string str = cs.Text(); str += "_"; str += cs1.Text();
            nonpoly_mapping.insert(std::make_pair(str, cs2.Text()));
       }

       std::set<std::string> polymer_prds;
       polymer_prds.clear();
       rowNo = t1->GetNumRows();
       idx1 = t1->GetColumnIndex("prd_id");
       idx2 = t1->GetColumnIndex("auth_asym_id");
       idx3 = t1->GetColumnIndex("auth_seq_id");
       for (int i = 0; i < rowNo; ++i) {
            get_value(cs, t1, idx1, i);
            get_value(cs1, t1, idx3, i);
            if (cs1 != "") continue;
            polymer_prds.insert(cs.Text());
       }

       std::map<std::string, std::string> het_prd_id_mapping;
       het_prd_id_mapping.clear();
       for (int i = 0; i < rowNo; ++i) {
            get_value(cs, t1, idx1, i);
            if (polymer_prds.find(cs.Text()) != polymer_prds.end()) continue;
            get_value(cs1, t1, idx2, i);
            get_value(cs2, t1, idx3, i);
            std::string str = cs1.Text(); str += "_"; str += cs2.Text();
            std::map<std::string, std::string>::const_iterator
                mpos = nonpoly_mapping.find(str);
            if (mpos == nonpoly_mapping.end()) continue;
            het_prd_id_mapping.insert(std::make_pair(mpos->second, cs.Text()));
       }
       if (het_prd_id_mapping.empty()) return;

       rowNo = t2->GetNumRows();
       idx1 = t2->GetColumnIndex("prd_id");
       idx2 = t2->GetColumnIndex("type");
       idx3 = t2->GetColumnIndex("details");
       int idx4 = t2->GetColumnIndex("class");
       std::map<std::string, std::vector<std::string> > prd_info_mapping;
       prd_info_mapping.clear();
       std::vector<std::string> data;
       for (int i = 0; i < rowNo; ++i) {
            get_value(cs, t2, idx1, i);
            get_value(cs1, t2, idx2, i);
            get_value(cs2, t2, idx3, i);
            get_value(cs3, t2, idx4, i);
            cs1 += " ";
            cs1 += cs3;
            data.clear();
            data.push_back(cs1.Text());
            data.push_back(cs2.Text());
            prd_info_mapping.insert(std::make_pair(cs.Text(), data));
       }

       for (std::map<std::string, std::string>::const_iterator
            pos = het_prd_id_mapping.begin(); pos != het_prd_id_mapping.end(); ++pos) {
            std::map<std::string, std::vector<std::string> >::const_iterator
                mpos = prd_info_mapping.find(pos->second);
            if (mpos == prd_info_mapping.end()) continue;
            prd_info.insert(std::make_pair(pos->first, mpos->second));
       }
}
