/*
FILE:     cif_to_ndb.C
*/
/*
VERSION:  8.125
*/
/*
DATE:     3/1/2013
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2013 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/stat.h>

#include "ndb_misclib.h"
#include "util.h"
#include "_maxit.h"
#include "align_util.h"
#include "erf_category.h"
#include "erf_util.h"
#include "sg_center.h"

#define N_SWITCH_CATEGORY   3
// #define N_R_MERGE           4
#define N_ENTITY           49
#define N_ENTITY_CATEGORY   4

typedef struct {
        const char *category;
        const char *high_item;
        const char *low_item;
} _Switch_Category;
/*
typedef struct {
        const char *catetory;
        const char *item;
} _R_Merge;
*/
static _Switch_Category _switch_category[N_SWITCH_CATEGORY] =
{
       { "refine",       "ls_d_res_high",     "ls_d_res_low" },
       { "reflns",       "d_resolution_high", "d_resolution_low" },
       { "reflns_shell", "d_res_high",        "d_res_low" }
};
/*
static _R_Merge _r_merge[N_R_MERGE] =
{
        { "reflns",          "ndb_Rmerge_I_obs" },
        { "reflns",          "ndb_Rsym_value" },
        { "reflns_shell",    "Rmerge_I_obs" },
        { "reflns_shell",    "ndb_Rsym_value" }
};
*/
static const char entity_token[N_ENTITY][7] = {
       "ENTAIL", "ENTKEY", "ENTMUT", "FRAGMT", "ISOTYP", "ECNUM",  "BINEMD", "BIOUNI",
       "ENTNAM", "ENTSYS", "PROTEI", "SRCNNM", "SRCNSP", "SRCNTS", "SRCNSE", "SRCNDS",
       "SRCNFR", "SRCNSC", "SRCNCC", "SRCNOR", "SRCNPL", "SRCGFR", "SRCGNM", "SRCNN1",
       "SRCGST", "SRCGSP", "SRCGCC", "SRCGTS", "SRCGTI", "SRCGDS", "SRCSYN", "SRCGE1",
       "SRCGE2", "SRCGE3", "SRCGE4", "SRCGE5", "SRCGHN", "SRCGHS", "SRCGH1", "SRCGHT",
       "SRCGHV", "SRCGHC", "SRCGVE", "SRCGHD", "SRCGPL", "TARGET", "ENTYPE", "ENTITY",
       "ENTDES"
};

static const char *entity_category[N_ENTITY_CATEGORY] = {
       "ndb_entity_poly_na_nonstandard",
       "ndb_entity_poly_na_type",
       "rcsb_entity_name",
       "struct_ref"
};

#define NUM_STOP_WORD  65

const char *_stop_words[NUM_STOP_WORD] = {
       "3D CRYSTAL STRUCTURE",
       "3D-STRUCTURE",
       "ACT_SITE",
       "BINDING",
       "CA_BIND",
       "CARBOHYD",
       "CHAIN",
       "COMPLETE PROTEOME",
       "CONFLICT",
       "CRYO-ELECTRON MICROSCOPY",
       "CRYSTALLOGRAPHY",
       "CRYSTAL STRUCTURE",
       "DIRECT PROTEIN SEQUENCING",
       "DISULFID",
       "DNA_BIND",
       "DOMAIN",
       "ELECTRON CRYSTALLOGRAPHY",
       "ELECTRON DIFFRACTION",
       "ELECTRON MICROSCOPY",
       "ELECTRON TOMOGRAPHY",
       "FIBER DIFFRACTION",
       "FIBRE DIFFRACTION",
       "HELIX TURN",
       "INIT_MET",
       "MECHANISM",
       "METAL",
       "MOD_RES",
       "MUTAGEN",
       "NATIVE CRYSTAL",
       "NEUTRON DIFFRACTION",
       "NMR",
       "NMR STRUCTURE",
       "NON_CONS",
       "NON_TER",
       "NP_BIND",
       "NUCLEAR MAGNETIC RESONANCE",
       "PEPTIDE",
       "PROPEP",
       "REPEAT",
       "REPEAT",
       "SE_CYS",
       "SEQUENCING",
       "SIGNAL",
       "SIGNAL",
       "SITE",
       "SOLID-STATE NMR",
       "SOLUTION NMR",
       "SOLUTION SCATTERING",
       "STRAND",
       "STRUCTURE",
       "THEORETICAL MODEL",
       "THIOETH",
       "THIOLEST",
       "TRANSIT",
       "TRANSMEM",
       "ULTRA-HIGH RESOLUTION",
       "UNSURE",
       "VARIANT",
       "VARSPLIC",
       "X-RAY",
       "X-RAY CRYSTALLOGRAPHY",
       "X-RAY DIFFRACTION",
       "X-RAY POWDER DIFFRACTION",
       "X-RAY STRUCTURE",
       "ZN_FING"
};

static int get_number_strands(ISTable *Table, const char *sheet_id);
static int get_sense(ISTable *Table, const char *sheet_id, const char *id_1, const char *id_2);
static void get_hbond(_MOLECULE *mol, ISTable *Table, const int chain_index,
                 const char *sheet_id, const char *range_id, CifString &curAtom, int &curIndex,
                 int &curRes, CifString &prevAtom, int &prevIndex, int &prevRes);
static void cif_to_ndb_find_serial_no(ISTable *Table, const int id, const int irow,
         const int tokenid, int *idfield, char idvalue[10][82], char SerialNo[10][82]);
static void process_sg_entry(CifFileObj *fobj, const char *blockId);
static void create_dbref_and_seqadv(_CHAIN *chain, const char *pdb_code, CifString &ref_id,
             CifString &db_name, CifString &accession, CifString &start_number,
             ReVarCifArray<CifString> &DB_Seq, const int pdb_range_begin, const int pdb_range_end,
             ISTable *Dbref, ISTable *Seqadv);
static int is_modified_residue(const char *pdb_residue, const char *seq_residue,
              ReVarCifArray<CifString> &parent_list);
static void remove_redundant_keywords(CifFileObj *fobj, const char* blockId);

void _MAXIT::cif_to_ndb_general(const int transfer, const int flag, const int special,
       const int recover_asym, const int exchange, const int together,
       const int short_version, const int symmetry_flag, const int adit_view,
       const int convert_null_value, const int force_renumbering, const int ebi_remark_flag)
{
       int fd, i, k, is_same, card_id, rowNo, entity_id, target_index, is_refmac5 = 0;
       int auth_asym_id, label_asym_id;
       char buffer[5000], buffer1[100];
       _CHAIN *chain = NULL;
       CifString cifstring, cifstring1, target, project, full_name, initials;
       ISTable *Table = NULL;
       One_Card_Info_Format *pCard = NULL;
       ReVarCifArray<CifString> colData, colData1;

       Table = _CifObj->GetTablePtr(_StructureId, "ndb_database_status");
       if (!is_empty_table(Table)) {
            int index = Table->GetColumnIndex("skip_PDB_REMARK");
            if (index >= 0) {
                 CifString cs;
                 Table->GetCell(cs, index, 0);
                 cs.ToLower();
                 if (strstr(cs.Text(), "modres")) _automatic_modres_flag = 0;
            }
       }

       ndb_processing_requests();
       ndb_processing_em_software();
       ndb_processing_struct_category();
       ndb_processing_struct_sheet_hbond();
       ndb_processing_ndb_database_PDB_obs_spr_category();
       ndb_processing_struct_ref_seq_and_diff();

       Table = _CifObj->GetTablePtr(_StructureId, "computing");
       if (!is_empty_table(Table)) {
            target_index = Table->GetColumnIndex("structure_refinement");
            Table->GetCell(cifstring, target_index, 0);
            strcpy(buffer, cifstring.Text());
            ndb_set_string_to_upper(buffer);
            if (strstr(buffer, "REFMAC 5")) is_refmac5 = 1;
       }

       if (transfer != 2) {
            seq_alignment(1, 0);
            automatic_renumber_pdb_numbers(force_renumbering);
            if (ebi_remark_flag & 2) capture_remark_465();
       }

       if (_file_status == FILE_STATUS_MMCIF_PLUS_MOL) {
            char filename[100], line[200];
            char templateFileName[9] = "./XXXXXX";
            strcpy(filename, templateFileName);
            fd = mkstemp(filename);
            close (fd);
            FILE *tlog = fopen(filename, "w");
            ndb_check_file(tlog, NUM_TOKEN, Check_Format, NUM_CHECK_CATEGORY,
                     check_category, 1, 0, "", NULL);
            input_structural_features(NDB_FILE_FORMAT_PDB);
            fclose (tlog);
            tlog = fopen(filename, "r");
            CifString message("");
            while (!feof(tlog)) {
                 fgets(line, 199, tlog);
                 if (feof(tlog)) break;
                 message += line;
            }
            fclose (tlog);
            sprintf(line, "rm -f %s", filename);
            system(line);
            if (message != "") ndb_log_message(NDB_MSG_WARN, "%s\n", message.Text());
       }
       // if (flag) processing_r_factor(100.0);

       process_sg_entry(_CifObj, _StructureId);

       if (is_refmac5 && _num_mol) {
            Table = _CifObj->GetTablePtr(_StructureId, "refine_ls_restr_ncs");
            if (!is_empty_table(Table)) {
                 rowNo = Table->GetNumRows();
                 auth_asym_id = Table->GetColumnIndex("pdbx_auth_asym_id");
                 if (auth_asym_id >= 0) {
                      colData.Clear();
                      for (i = 0; i < rowNo; i++) {
                           get_value(cifstring, Table, auth_asym_id, i);
                           if (cifstring != "") {
                                chain = _molecules[_rep_mol]->GetFirstChain();
                                while (chain) {
                                     if (cifstring == chain->PDB_ChainID()) {
                                          sprintf(buffer1, "%d", chain->index());
                                          cifstring.Copy(buffer1);
                                          colData.Add(cifstring);
                                          break;
                                     }
                                     chain = _molecules[_rep_mol]->GetNextChain();
                                }
                           }
                      }
                      if ((int) colData.Length() == rowNo) {
                           Table->AddColumn("c_index");
                           auth_asym_id = Table->GetColumnIndex("c_index");
                           Table->FillColumn(colData, auth_asym_id);
                           _CifObj->WriteTable(Table, _StructureId, "refine_ls_restr_ncs");
                      }
                 }
            }
       }

       ndb_processing_categories_mapping(transfer);

       if (transfer != 2) {
            find_missing_residues();
            if (!recover_asym) Get_Stbiol(exchange);
            Reorder_Stbiol();
            Input_Entity();
            get_seqifo(0, together);
            match_entity_with_molecules();
            match_stbiol_with_molecules();
            if (force_renumbering > 0) {
                 automatic_renumber_pdb_numbers(force_renumbering);
                 _CifObj->DeleteTable(_StructureId, "struct_site");
                 _CifObj->DeleteTable(_StructureId, "struct_site_gen");
                 Clear_Site();
            }
            int is_reverse_entity = 0;
            Table = _CifObj->GetTablePtr(_StructureId, "entity_poly_seq");
            if (!is_empty_table(Table)) {
                 auth_asym_id = Table->GetColumnIndex("entity_id");
                 get_value(cifstring, Table, auth_asym_id, 0);
                 if (cifstring != "" && cifstring != "1") is_reverse_entity = 1;
            }
            if (_file_status == FILE_STATUS_MMCIF_PLUS_MOL || is_reverse_entity) {
                 is_same = 1;
                 for (i = 0; i < _n_entity; i++) {
                      if (_entity[i].Entity_ID != _entity[i].Old_Entity_ID) {
                           is_same = 0; break;
                      }
                 }
                 if (!is_same) {
                      for (i = 0; i < N_ENTITY - 2; i++) {
                           card_id = find_card_id(entity_token[i]);
                           if (card_id == -1) continue;
                           pCard = _CardInfo[card_id].Head();
                           int field_no = 1;
                           if (!strcmp(entity_token[i], "BINEMD")) field_no = 3;
                           while (pCard) {
                                for (k = 0; k < _n_entity; k++) {
                                     if (atoi(pCard->Field(field_no)) == _entity[k].Old_Entity_ID) {
                                          pCard->AddElementToField(_entity[k].Entity_ID, field_no);
                                          break;
                                     }
                                }
                                pCard = pCard->Next();
                           }
                           _CardInfo[card_id].reorder_cards(1, INTEGER_COMPARE, ASCENT);
                      }

                      for (int l = 0; l < N_ENTITY_CATEGORY; l++) {
                           Table = _CifObj->GetTablePtr(_StructureId, entity_category[l]);
                           if (is_empty_table(Table)) continue;
                           rowNo = Table->GetNumRows();
                           entity_id = Table->GetColumnIndex("entity_id");
                           if (entity_id >= 0) {
                                for (i = 0; i < rowNo; i++) {
                                     get_value(cifstring, Table, entity_id, i);
                                     if (cifstring != "") {
                                          for (k = 0; _n_entity; k++) {
                                               if (atoi(cifstring.Text()) == _entity[k].Old_Entity_ID) {
                                                    sprintf(buffer, "%d", _entity[k].Entity_ID);
                                                    cifstring.Copy(buffer);
                                                    Table->UpdateCell(cifstring, entity_id, i);
                                                    break;
                                               }
                                          }
                                     }
                                }
                           }
                           _CifObj->WriteTable(Table, _StructureId, entity_category[l]);
                      }
                 }
            }
            Table = _CifObj->GetTablePtr(_StructureId, "rcsb_struct_link");
            if (!is_empty_table(Table) && !__need_file_version_update) {
                 cifndb_process_rcsb_link(Table);
            }

            get_entity_card();
            get_compnd_card();
       }

       if (is_refmac5 && _num_mol) {
            Table = _CifObj->GetTablePtr(_StructureId, "refine_ls_restr_ncs");
            if (!is_empty_table(Table)) {
                 rowNo = Table->GetNumRows();
                 auth_asym_id = Table->GetColumnIndex("pdbx_auth_asym_id");
                 label_asym_id = Table->GetColumnIndex("c_index");
                 if (auth_asym_id >= 0 && label_asym_id >= 0) {
                      for (i = 0; i < rowNo; i++) {
                           get_value(cifstring, Table, label_asym_id, i);
                           chain = _molecules[_rep_mol]->GetIndexChain(atoi(cifstring.Text()));
                           if (chain) {
                                cifstring.Copy(chain->PDB_ChainID());
                                Table->UpdateCell(cifstring, auth_asym_id, i);
                           }
                      }
                      Table->DeleteColumn(label_asym_id);
                      Table->RemoveColumn(label_asym_id);
                      _CifObj->WriteTable(Table, _StructureId, "refine_ls_restr_ncs");
                      if (!transfer) {
                           card_id = find_card_id("TMLPTL");
                           pCard = _CardInfo[card_id].Head();
                           while (pCard) pCard = _CardInfo[card_id].delete_a_card(pCard);
                           cifndb_process_refmac_refine_ls_restr_ncs(Table);
                      }
                 }
            }
       }

       check_terminal_atoms(cifstring);
       if (strcmp(cifstring.Text(), "")) ndb_log_message(NDB_MSG_WARN, "%s\n", cifstring.Text());

       post_processing_from_cif_to_ndb();
}

void _MAXIT::entity_to_sequence()
{
       if (_num_mol == 0 || _molecules == NULL) return;

       Output_Entity();

       for (int i = 0; i < N_ENTITY; i++) {
            int card_id = find_card_id(entity_token[i]);
            if (card_id == -1) continue;
            One_Card_Info_Format * pCard = _CardInfo[card_id].Head();
            while (pCard) {
                 int field_no = 1;
                 if (!strcmp(entity_token[i], "BINEMD")) field_no = 3;
                 _CHAIN *chain = _molecules[_rep_mol]->GetFirstChain();
                 while (chain) { 
                      if (atoi(pCard->Field(field_no)) == chain->entity_id()) {
                           CifString sequence;
                           sequence.Clear();
                           for (int j = 0; j < chain->SeqLen(); j++) {
                                if (sequence != "") sequence += " ";
                                sequence += chain->SeqRes(j)->Field[0];
                           }
                           pCard->AddElementToField(sequence.Text(), field_no);
                           break;
                      }
                      chain = _molecules[_rep_mol]->GetNextChain();
                 }
                 pCard = pCard->Next();
            }
       }

       if (!_CifObj) return;

       CifString cs;
       for (int l = 0; l < N_ENTITY_CATEGORY; l++) {
            ISTable *Table = _CifObj->GetTablePtr(_StructureId, entity_category[l]);
            if (is_empty_table(Table)) continue;
            int entity_id = Table->GetColumnIndex("entity_id");
            if (entity_id < 0) continue;
            int rowNo = Table->GetNumRows();
            for (int i = 0; i < rowNo; i++) {
                 get_value(cs, Table, entity_id, i);
                 if (cs == "") continue;
                 _CHAIN *chain = _molecules[_rep_mol]->GetFirstChain();
                 while (chain) { 
                      if (atoi(cs.Text()) == chain->entity_id()) {
                           CifString sequence;
                           sequence.Clear();
                           for (int j = 0; j < chain->SeqLen(); j++) {
                                if (sequence != "") sequence += " ";
                                sequence += chain->SeqRes(j)->Field[0];
                           }
                           Table->UpdateCell(sequence, entity_id, i);
                           break;
                      }
                      chain = _molecules[_rep_mol]->GetNextChain();
                 }
            }
            _CifObj->WriteTable(Table, _StructureId, entity_category[l]);
       }
}

void _MAXIT::sequence_to_entity()
{
       if (_num_mol == 0 || _molecules == NULL) return;

       for (int i = 0; i < N_ENTITY; i++) {
            int card_id = find_card_id(entity_token[i]);
            if (card_id == -1) continue;
            One_Card_Info_Format * pCard = _CardInfo[card_id].Head();
            int field_no = 1;
            if (!strcmp(entity_token[i], "BINEMD")) field_no = 3;
            while (pCard) {
                 _CHAIN *chain = _molecules[_rep_mol]->GetFirstChain();
                 while (chain) { 
                      CifString sequence;
                      sequence.Clear();
                      for (int j = 0; j < chain->SeqLen(); j++) {
                           if (!strcmp(chain->SeqRes(j)->Field[0], "?")) continue;
                           if (sequence != "") sequence += " ";
                           sequence += chain->SeqRes(j)->Field[0];
                      }
                      if (sequence == pCard->Field(field_no)) {
                           pCard->AddElementToField(chain->entity_id(), field_no);
                           break;
                      }
                      chain = _molecules[_rep_mol]->GetNextChain();
                 }
                 pCard = pCard->Next();
            }
            _CardInfo[card_id].reorder_cards(1, INTEGER_COMPARE, ASCENT);
       }
       Input_Entity();

       if (!_CifObj) return;

       CifString cs;
       char buffer[10];
       for (int l = 0; l < N_ENTITY_CATEGORY; l++) {
            ISTable *Table = _CifObj->GetTablePtr(_StructureId, entity_category[l]);
            if (is_empty_table(Table)) continue;
            int entity_id = Table->GetColumnIndex("entity_id");
            if (entity_id < 0) continue;
            int rowNo = Table->GetNumRows();
            for (int i = 0; i < rowNo; i++) {
                 get_value(cs, Table, entity_id, i);
                 if (cs == "") continue;
                 _CHAIN *chain = _molecules[_rep_mol]->GetFirstChain();
                 while (chain) { 
                      CifString sequence;
                      sequence.Clear();
                      for (int j = 0; j < chain->SeqLen(); j++) {
                           if (sequence != "") sequence += " ";
                           sequence += chain->SeqRes(j)->Field[0];
                      }
                      if (sequence == cs) {
                           sprintf(buffer, "%d", chain->entity_id());
                           cs.Copy(buffer);
                           Table->UpdateCell(cs, entity_id, i);
                           break;
                      }
                      chain = _molecules[_rep_mol]->GetNextChain();
                 }
            }
            _CifObj->WriteTable(Table, _StructureId, entity_category[l]);
       }
}

void _MAXIT::cif_to_ndb_exchange(const int exchange, const int adit_view)
{
       read_mmcif_file(exchange);
       pre_processing_mmcif_file(0, 0, 0, 0, adit_view, 0, 0, 0);
       version_update_read();
       seq_alignment(0, 0);
       ndb_processing_em_software();
       ndb_processing_struct_category();
       ndb_processing_struct_sheet_hbond();
       ndb_processing_ndb_database_PDB_obs_spr_category();
       ndb_processing_struct_ref_seq_and_diff();
       // processing_r_factor(100.0);
       ndb_processing_categories_mapping(0);
       Get_Stbiol(exchange);
}

void _MAXIT::ndb_processing_categories_mapping(const int transfer)
{
       int i, index, struct_ref_id = -1, is_refmac5 = 0, is_phenix = 0;
       char buffer[100];
       CifString cifstring;
       ISTable *Struct_Ref = NULL, *Table = NULL;

       Table = _CifObj->GetTablePtr(_StructureId, "computing");
       if (!is_empty_table(Table)) {
            index = Table->GetColumnIndex("structure_refinement");
            Table->GetCell(cifstring, index, 0);
            strcpy(buffer, cifstring.Text());
            ndb_set_string_to_upper(buffer);
            if (strstr(buffer, "REFMAC 5")) is_refmac5 = 1;
            if (strstr(buffer, "PHENIX")) is_phenix = 1;
       }

       if (!_automatic_modres_flag) cifndb_process_modres();

       for (i = 0; i < MxCategory; i++) {
            if (transfer && !(transfer & CifCategory[i].transfer)) continue;

            if (!strcmp(CifCategory[i].category, "struct_conf_type") ||
                !strcmp(CifCategory[i].category, "struct_conn_type") ||
                !strcmp(CifCategory[i].category, "struct_sheet_order") ||
                !strcmp(CifCategory[i].category, "struct_sheet_range") ||
                !strcmp(CifCategory[i].category, "struct_sheet_hbond") ||
                !strcmp(CifCategory[i].category, "rcsb_struct_sheet_hbond") ||
                !strcmp(CifCategory[i].category, "atom_site") ||
                !strcmp(CifCategory[i].category, "atom_site_anisotrop") ||
                !strcmp(CifCategory[i].category, "pdbx_atom_site_aniso_tls") ||
                !strcmp(CifCategory[i].category, "struct_asym") ||
                !strcmp(CifCategory[i].category, "entity_poly_seq") ||
                !strcmp(CifCategory[i].category, "ndb_nonstandard_list") ||
                !strcmp(CifCategory[i].category, "ndb_poly_seq_scheme") ||
                !strcmp(CifCategory[i].category, "database_PDB_rev_record") ||
                // !strcmp(CifCategory[i].category, "ccp4_struct_ncs_dom_lim") ||
                !strcmp(CifCategory[i].category, "ccp4_refine_ls_restr_ncs"))
                 continue;

            Table = _CifObj->GetTablePtr(_StructureId, CifCategory[i].category);
            if (is_empty_table(Table)) continue;

            if (!strcmp(CifCategory[i].category, "struct_ref")) {
                 Struct_Ref = Table;
                 struct_ref_id = i;
                 continue;
            } if (!strcmp(CifCategory[i].category, "struct_conf"))
                 cifndb_process_struct_conf(Table);
            else if (!strcmp(CifCategory[i].category, "struct_sheet"))
                 cifndb_process_sheet();
            else if (!strcmp(CifCategory[i].category, "struct_conn")) {
                 if (!__need_file_version_update) cifndb_process_struct_conn(Table);
            } else if (!strcmp(CifCategory[i].category, "refine_ls_restr")) 
                 cifndb_process_refine_ls_restr(Table);
            else if (!strcmp(CifCategory[i].category, "struct_site_gen"))  {
                 if (!__need_file_version_update) cifndb_process_struct_site_gen(Table);
            } else if (!strcmp(CifCategory[i].category, "rcsb_struct_link")) {
                 if (!__need_file_version_update) cifndb_process_rcsb_link(Table);
            } else if (!strcmp(CifCategory[i].category, "database_PDB_rev"))
                 cifndb_process_database_pdb_rev(Table);
            else if (!strcmp(CifCategory[i].category, "rcsb_database_related"))
                 cifndb_process_rcsb_database_related(Table);
            else if (!strcmp(CifCategory[i].category, "database_PDB_remark") ||
                     !strcmp(CifCategory[i].category, "rcsb_database_PDB_remark")) {
                 if (!strcmp(CifCategory[i].category, "database_PDB_remark"))
                      cifndb_process_remark(Table, PREMRK_TOKEN-1, 2);
                 else {
                      cifndb_process_remark(Table, PDBRMK_TOKEN-1, 3);
                 }
            } else if (!strcmp(CifCategory[i].category, "ndb_database_remark") ||
                     !strcmp(CifCategory[i].category, "ndb_pdb_compnd") ||
                     !strcmp(CifCategory[i].category, "ndb_pdb_source")) {
                 cifndb_process_general_category(Table, i, 1);
            } else if (!strcmp(CifCategory[i].category, "refine_ls_restr_ncs")) {
                 if (is_refmac5 || is_phenix)
                      cifndb_process_refmac_refine_ls_restr_ncs(Table);
                 else cifndb_process_category(Table, i, 0);
/*
            } else if (!strcmp(CifCategory[i].category, "struct_ncs_dom_lim")) {
                 if (is_refmac5)
                      cifndb_process_refmac_struct_ncs_dom_lim(Table);
                 else cifndb_process_category(Table, i, 0);
*/
            } else if (CifCategory[i].general)
                 cifndb_process_general_category(Table, i, 0);
            else if (CifCategory[i].xray_or_nmr & 4)
                 cifndb_process_matrix(Table, i);
            else cifndb_process_category(Table, i, 0);
       }
       if (Struct_Ref) cifndb_process_category(Struct_Ref, struct_ref_id, 1);
}

void _MAXIT::cifndb_process_category(ISTable *Table, const int id, const int is_special)
{
       int  i, j, k, irow, icol, tokenid, jrnlid, fieldno;
       int  idfield[10], SerialNoField[10], cols[200];
       char SerialNo[10][82], idValue[10][82];
       CifString string;

       int colNo = Table->GetNumColumns();
       int rowNo = Table->GetNumRows();
       ReVarCifArray<CifString> *ColumnNames = Table->GetColumnNames();

       for (i = 0; i < 10; i++) idfield[i] = -1;
       for (i = 0; i < colNo; i++) {
            for (j = 0; j < CifCategory[id].nkeywords; j++) {
                 if (!strcasecmp((*ColumnNames)[i].Text(),
                        CifCategory[id].keywords[j].name)) break;
            }
            if (j == CifCategory[id].nkeywords) cols[i] = -1;
            else                                cols[i] = j;
            for (k = 0; k < CifCategory[id].nCifKeyField; k++) {
                 if (j == CifCategory[id].CifKeyField[k]) idfield[k] = i;
            }
       }
       if (ColumnNames) delete ColumnNames;

       for (irow = 0; irow < rowNo; irow++) {
            for (icol = 0; icol < colNo; icol++) {
                 j = cols[icol];
                 if (j < 0 || j > CifCategory[id].nkeywords) continue;
                 Table->GetCell(string, icol, irow);
                 if (string == "" || string == "." || string == "?") continue;

                 for (i = 0; i < CifCategory[id].keywords[j].numNdbInfo; i++) {
                      tokenid = CifCategory[id].keywords[j].NdbInfo[i].NdbTokenId-1;
                      jrnlid  = CifCategory[id].keywords[j].NdbInfo[i].JrnlTokenId-1;
                      fieldno = CifCategory[id].keywords[j].NdbInfo[i].NdbFieldNo-1;
                      for (k = 0; k < CifCategory[id].nCifKeyField; k++)
                           SerialNoField[k] = CifCategory[id].keywords[j].
                                       NdbInfo[i].NdbTokenKeyFieldNo[k];

                      if (tokenid < 0 || tokenid >= MxTokens || fieldno <= 0) continue;
                      cif_to_ndb_find_serial_no(Table, id, irow, tokenid, idfield,
                                       idValue, SerialNo);

                      if (CifCategory[id].nCifKeyField) {
                           for (k = 0; k < CifCategory[id].nCifKeyField; k++) {
                                if (strcmp(idValue[k], "") && strcasecmp(idValue[k],
                                    CifCategory[id].keywords[j].NdbInfo[i].
                                       JrnlTokenName_Type)) break;
                           }
                           if (k < CifCategory[id].nCifKeyField) continue;
                      }
        
                      if (tokenid == JRNL_TOKEN - 1) {
                           if (!strcasecmp(JrnlToken[jrnlid].TokenName, "AUTH") &&
                               !strcasecmp(CifCategory[id].category, "citation"))
                                continue;

                           cifndb_process_general(jrnlid, fieldno, 
                              CifCategory[id].nCifKeyField, SerialNoField, 
                                 SerialNo, string.Text(), TRUE);
                      } else if (is_special)
                           cifndb_process_special(tokenid, fieldno,
                                   CifCategory[id].nCifKeyField,
                                   SerialNoField, SerialNo, string.Text());
                      else cifndb_process_general(tokenid, fieldno, 
                                   CifCategory[id].nCifKeyField, 
                                   SerialNoField, SerialNo, string.Text(), FALSE);
                 }
            }
       }
}

void _MAXIT::cifndb_process_general(const int tokenid, const int fieldno, 
              const int NumKey, int *SerialNoField, char SerialNo[10][82], 
              char *value, const int IsJrnl)
{
       One_Card_Info_Format *pCard = NULL;
       int i, vField = 5, logical = FALSE, jrnl_id;
       char buffer[1000];

       if (tokenid == NDBFIL_TOKEN - 1 &&
          !strcasecmp(value, "DEPOSIT-0")) {
            pCard = _CardInfo[tokenid].Head();
            if (pCard && strcmp(pCard->Field(fieldno), "")) return;
       }

       if (IsJrnl) {
            pCard = _JrnlCardInfo[tokenid].Head();
            while (pCard) {
                 if (NumKey) {
                      for (i = 0; i < NumKey; i++) {
                           if (SerialNoField[i] > 0 && strcasecmp(SerialNo[i],
                               pCard->Field(SerialNoField[i]))) break;
                      }
                      if (i == NumKey) break;
                 } else if (JrnlToken[tokenid].SeqField != 0 && !strcasecmp(SerialNo[0],
                   pCard->Field(JrnlToken[tokenid].SeqField-1)))
                      break;
                 else if (!strcmp(pCard->Field(fieldno), "")) break;
                 pCard = pCard->Next();
            }
       } else if (tokenid == AUTHOR_TOKEN-1)
            pCard = _CardInfo[tokenid].Tail();
       else {
            pCard = _CardInfo[tokenid].Head();
            while (pCard) {
                 if (NumKey) {
                      for (i = 0; i < NumKey; i++) {
                           if (SerialNoField[i] > 0 && strcasecmp(SerialNo[i], 
                               pCard->Field(SerialNoField[i]))) break;
                      }
                      if (i == NumKey) break;
                 } else if (NdbToken[tokenid].SeqField != 0 && !strcasecmp(SerialNo[0],
                   pCard->Field(NdbToken[tokenid].SeqField-1)))
                      break;
                 else if ((tokenid == HEADER_TOKEN - 1 ||
                           tokenid == PDBFIL_TOKEN - 1 ||
                           tokenid == NDBFIL_TOKEN - 1) &&
                          !strcasecmp(value, pCard->Field(fieldno))) break;
                 else if (!strcmp(pCard->Field(fieldno), "")) break;
                 pCard = pCard->Next();
            }
       }

       if (IsJrnl && pCard == NULL) {
            pCard = _JrnlCardInfo[tokenid].add_a_new_card(JrnlToken[tokenid], 1);
            if (NumKey) {
                 for (i = 0; i < NumKey; i++) {
                      if (SerialNoField[i] > 0)
                           pCard->AddElementToField(SerialNo[i], SerialNoField[i]);
                 }
            } else if (atoi(SerialNo[0]) != 0 && JrnlToken[tokenid].SeqField != 0)
                 pCard->AddElementToField(SerialNo[0], JrnlToken[tokenid].SeqField-1);
       } else if (!IsJrnl && pCard == NULL) {
            pCard = _CardInfo[tokenid].add_a_new_card(NdbToken[tokenid], 0);
            if (NumKey) {
                 for (i = 0; i < NumKey; i++) {
                      if (SerialNoField[i] > 0) 
                           pCard->AddElementToField(SerialNo[i], SerialNoField[i]);
                 }
            } else if (strcmp(SerialNo[0], "") && NdbToken[tokenid].SeqField != 0)
                 pCard->AddElementToField(SerialNo[0], NdbToken[tokenid].SeqField-1);
       }

       for (i = 0; i < NumKey; i++) {
            if (fieldno == SerialNoField[i]) logical = TRUE;
       }
       if (!IsJrnl && (logical || fieldno == NdbToken[tokenid].SeqField - 1) ||
           IsJrnl && fieldno == JrnlToken[tokenid].SeqField -1) {}
       else if (!IsJrnl && !strcmp(NdbToken[tokenid].TokenName, "AUTHOR")) {
            ndb_name_conversion_first_last(value);
            if (!strcmp(pCard->Field(fieldno), ""))
                 strcpy(buffer, value);
            else sprintf(buffer, "%s, %s", pCard->Field(fieldno), value);
            pCard->AddElementToField(buffer, fieldno);
       } else if (!IsJrnl || IsJrnl && (!strcmp(pCard->Field(fieldno), "") ||
            strcmp(JrnlToken[tokenid].TokenName, "AUTH") &&
            strcmp(JrnlToken[tokenid].TokenName, "EDIT"))) {
            if (IsJrnl) {
                 jrnl_id = find_jrnl_id("REF");
                 if (jrnl_id == tokenid &&  fieldno == vField + 1)
                      pCard->AddElementToField("V.", vField);
                 if (!strcmp(JrnlToken[tokenid].TokenName, "AUTH") ||
                     !strcmp(JrnlToken[tokenid].TokenName, "EDIT"))
                      ndb_name_conversion_first_last(value);
            }
            pCard->AddElementToField(value, fieldno);
       } else if (atoi(SerialNo[0]) == 
           atoi(pCard->Field(JrnlToken[tokenid].SeqField-1))) {
            ndb_name_conversion_first_last(value);
            sprintf(buffer, "%s, %s", pCard->Field(fieldno), value);
            pCard->AddElementToField(buffer, fieldno);
       }
}

void _MAXIT::cifndb_process_special(const int tokenid, const int fieldno,
              const int NumKey, int *SerialNoField, char SerialNo[10][82],
              const char *value)
{
       int i;
       One_Card_Info_Format *pCard = _CardInfo[tokenid].Head();
       while (pCard) {
            for (i = 0; i < NumKey; i++) {
                 if (SerialNoField[i] > 0 && strcmp(SerialNo[i],
                       pCard->Field(SerialNoField[i]))) break;
            }
            if (i == NumKey) pCard->AddElementToField(value, fieldno);
            pCard = pCard->Next();
       }
}

void _MAXIT::cifndb_process_general_category(ISTable *Table, const int id, 
             const int is_remark)
{
       int ndb_tokenid = -1, ndb_fieldno, i, j, tmp_id, row, col;
       char value[10000], buffer[62], *p;
       CifString string;
       One_Card_Info_Format *pCard = NULL, *qCard = NULL;

       int colNo = Table->GetNumColumns();
       int rowNo = Table->GetNumRows();

       ReVarCifArray<CifString> *ColumnNames = Table->GetColumnNames();
       int *cols = new int[colNo];

       for (i = 0; i < colNo; i++) {
            for (j = 0; j < CifCategory[id].nkeywords; j++) {
                 if (!strcasecmp((*ColumnNames)[i].Text(),
                        CifCategory[id].keywords[j].name)) {
                      tmp_id = CifCategory[id].keywords[j].NdbInfo[0].NdbTokenId-1;
                      if (tmp_id != HEADER_TOKEN-1 && tmp_id >= 0)
                              ndb_tokenid = tmp_id;
                      break;
                 }
            }
            if (j == CifCategory[id].nkeywords) cols[i] = -1;
            else                                cols[i] = j;
       }
       if (ColumnNames) delete ColumnNames;
       if (ndb_tokenid == -1) {
            delete cols;
            return;
       }

       for (row = 0; row < rowNo; row++) {
            pCard = _CardInfo[ndb_tokenid].add_a_new_card(NdbToken[ndb_tokenid], 0);
            for (col = 0; col < colNo; col++) {
                 j = cols[col];
                 if (j < 0 || j > CifCategory[id].nkeywords) continue;

                 string.Clear();
                 Table->GetCell(string, col, row);
                 if (string == "?" || string == "." ||
                     string == "" || string == "@") continue;

                 ndb_fieldno = CifCategory[id].keywords[j].
                            NdbInfo[0].NdbFieldNo - 1;
                 if (ndb_fieldno < 0) continue;

                 strcpy(value, string.Text());
                 if (is_remark) {
                      if (strstr(value, "'") && value[0] == '\n')
                           strcpy(value, &value[1]);
                 } else ndb_clean_string(value);
                 if (is_remark && ndb_tokenid == PREMRK_TOKEN-1 &&
                     strlen(value) > 60) {
                      ndb_clean_string(value);
                      p = get_max_length_word(value, buffer, 60);
                      pCard->AddElementToField(buffer, ndb_fieldno);
                      while (p) {
                           p = get_max_length_word(p, buffer, 60);
                           ndb_clean_string(buffer);
                           if (!strcmp(buffer, "")) break;
                           qCard = _CardInfo[ndb_tokenid].
                                add_a_new_card(NdbToken[ndb_tokenid], 0);
                           qCard->AddElementToField(pCard->Field(1), 1);
                           qCard->AddElementToField(buffer, ndb_fieldno);
                      }
                 } else pCard->AddElementToField(value, ndb_fieldno);
            }
       }
       delete [] cols;

       if (ndb_tokenid == REMARK_TOKEN-1 ||
           ndb_tokenid == PCOMPN_TOKEN-1 ||
           ndb_tokenid == PSOURC_TOKEN-1) {
            pCard = _CardInfo[ndb_tokenid].Head();
            while (pCard) {
                 if (!strcmp(pCard->Field(2), "") ||
                     !strcmp(pCard->Field(2), "@") ||
                     !strcmp(pCard->Field(2), "?"))
                      pCard = _CardInfo[ndb_tokenid].delete_a_card(pCard);
                 else pCard = pCard->Next();
            }
       }
}

void _MAXIT::cifndb_process_refmac_refine_ls_restr_ncs(ISTable *Table)
{
       int i, j, found_match = 0, n_keywords = 8, colNo, rowNo, *fieldno = NULL, card_id;
       CifString cifstring;
       One_Card_Info_Format *pCard = NULL; 
       ReVarCifArray<CifString> *ColumnNames = NULL;
       static const char *keywords[8] = { "", "dom_id", "pdbx_auth_asym_id", "pdbx_number",
             "rms_dev_position", "weight_position", "pdbx_type", "pdbx_ens_id" };

       colNo = Table->GetNumColumns();
       rowNo = Table->GetNumRows();
       ColumnNames = Table->GetColumnNames();

       fieldno = new int[colNo];
       for (i = 0; i < colNo; i++) {
            fieldno[i] = 0;
            for (j = 1; j < n_keywords; j++) {
                 if (!strcasecmp((*ColumnNames)[i].Text(), keywords[j])) {
                      fieldno[i] = j;
                      found_match = 1;
                      break;
                 }
            }
       }
       if (ColumnNames) delete ColumnNames;
       if (!found_match) {
            delete [] fieldno;
            return;
       }

       card_id = find_card_id("TMLPTL");
       for (i = 0; i < rowNo; i++) {
            pCard = _CardInfo[card_id].add_a_new_card(NdbToken[card_id], 0);
            for (j = 0; j < colNo; j++) {
                 if (!fieldno[j]) continue;
                 get_value(cifstring, Table, j, i);
                 if (cifstring != "") pCard->AddElementToField(cifstring.Text(), fieldno[j]);
            }
       }
       delete [] fieldno;
}
/*
void _MAXIT::cifndb_process_refmac_struct_ncs_dom_lim(ISTable *Table)
{
       int i, j, found_match = 0, n_keywords = 8, colNo, rowNo, *fieldno = NULL, card_id;
       CifString cifstring;
       One_Card_Info_Format *pCard = NULL; 
       ReVarCifArray<CifString> *ColumnNames = NULL;
       static const char *keywords[8] = { "", "dom_id", "beg_auth_asym_id", "beg_auth_seq_id",
             "end_auth_asym_id", "end_auth_seq_id", "pdbx_component_id", "pdbx_refine_code" };

       colNo = Table->GetNumColumns();
       rowNo = Table->GetNumRows();
       ColumnNames = Table->GetColumnNames();

       fieldno = new int[colNo];
       for (i = 0; i < colNo; i++) {
            fieldno[i] = 0;
            for (j = 1; j < n_keywords; j++) {
                 if (!strcasecmp((*ColumnNames)[i].Text(), keywords[j])) {
                      fieldno[i] = j;
                      found_match = 1;
                      break;
                 }
            }
       }
       if (ColumnNames) delete ColumnNames;
       if (!found_match) {
            delete [] fieldno;
            return;
       }

       card_id = find_card_id("CCPNCS");
       for (i = 0; i < rowNo; i++) {
            pCard = _CardInfo[card_id].add_a_new_card(NdbToken[card_id], 0);
            for (j = 0; j < colNo; j++) {
                 if (!fieldno[j]) continue;
                 get_value(cifstring, Table, j, i);
                 if (cifstring != "") pCard->AddElementToField(cifstring.Text(), fieldno[j]);
            }
       }
       delete [] fieldno;
}
*/
void _MAXIT::cifndb_process_remark(ISTable *Table, const int card_id, const int field_no)
{
       CifString id, text, old_id;
       char buffer[80], *value, *p;
       int irow, count, length, i, first = 1;
       One_Card_Info_Format *pCard = NULL;

       int width = NdbToken[REMARK_TOKEN-1].FieldList[2].FieldWidth;

       int rowNo = Table->GetNumRows();
       int id_idx = Table->GetColumnIndex("id");
       int txt_idx = Table->GetColumnIndex("text");
       if (id_idx < 0 || txt_idx < 0) return;

       old_id.Copy("-1");
       for (irow = 0; irow < rowNo; irow++) {
            Table->GetCell(id, id_idx, irow);
            Table->GetCell(text, txt_idx, irow);
            if (id != old_id) {
                 old_id.Copy(id);
                 first = 1;
            } else first = 0;
            count = 0;
            length = strlen(text.Text());
            for (i = 0; i < length; i++) {
                 if (text.Text()[i] != '\n') count++;
                 else {
                      if (count > width) break;
                      count = 0;
                 }
            }
            if (count > width) {
                 value = new char[length + 1];
                 strcpy(value, text.Text());
                 ndb_clean_string(value);
                 p = value;
                 while (p) {
                      p = get_max_length_word(p, buffer, width);
                      ndb_clean_string(buffer);
                      if (!strcmp(buffer, "")) break;
                      pCard = _CardInfo[card_id].
                               add_a_new_card(NdbToken[card_id], 0);
                      pCard->AddElementToField(id.Text(), 1);
                      pCard->AddElementToField(buffer, field_no);
                 }
                 delete [] value;
            } else {
                 count = 0;
                 for (i = 0; i < length; i++) {
                      if (text.Text()[i] != '\n') {
                           buffer[count++] = text.Text()[i];
                      } else {
                           buffer[count] = '\0';
                           if (!strcmp(buffer, "?")) strcpy(buffer, "");
                           if (strcmp(buffer, "") || !strcmp(buffer, "") && !first) {
                                pCard = _CardInfo[card_id].
                                        add_a_new_card(NdbToken[card_id], 0);
                                pCard->AddElementToField(id.Text(), 1);
                                pCard->AddElementToField(buffer, field_no);
                                first = 0;
                           }
                           count = 0;
                      }
                 }
                 if (count) {
                      buffer[count] = '\0';
                      if (!strcmp(buffer, "?")) strcpy(buffer, "");
                      if (strcmp(buffer, "") || !strcmp(buffer, "") && !first) {
                           pCard = _CardInfo[card_id].
                                    add_a_new_card(NdbToken[card_id], 0);
                           pCard->AddElementToField(id.Text(), 1);
                           pCard->AddElementToField(buffer, field_no);
                      }
                 }
            }
       }
}

void _MAXIT::cifndb_process_refine_ls_restr(ISTable *Table)
{
       const char *column_name[NDB_CIF_RMS_ITEMS + 2] = { "type", "dev_ideal",
                        "dev_ideal_target", "weight", "number",
                        "pdbx_restraint_function", "pdbx_refine_id" };
       int column_index[NDB_CIF_RMS_ITEMS + 2];

       for (int i = 0; i < NDB_CIF_RMS_ITEMS + 2; i++) {
            column_index[i] = Table->GetColumnIndex(column_name[i]);
       }
       if (column_index[0] < 0) return;

       int rowNo = Table->GetNumRows();

       CifString value, refine_code;
       for (int irow = 0; irow < rowNo; irow++) {
            get_value(value, Table, column_index[0], irow);
            int index = -1;
            for (int i = 0; i < mxRmsRecords; i++) {
                 if (!strcasecmp(value.Text(), rmsRecords[i].type)) {
                      index = i;
                      break;
                 }
            }
            if (index < 0) continue;

            for (int i = 1; i <= NDB_CIF_RMS_ITEMS; i++) {
                 if (column_index[i] >= 0 && 
                     get_value(value, Table, column_index[i], irow)) {
                      int ndb_tokenid = rmsRecords[index].tokens[i - 1];
                      int ndb_fieldno = rmsRecords[index].fields[i - 1];
                      get_value(refine_code, Table, column_index[NDB_CIF_RMS_ITEMS+1], irow);
                      put_into_card(value.Text(), ndb_tokenid, ndb_fieldno,
                             refine_code.Text(), NdbToken[ndb_tokenid].RefineField-1);
                 }
            }
       }
}

void _MAXIT::cifndb_process_matrix(ISTable *Table, const int id)
{
       int colNo = Table->GetNumColumns();
       int rowNo = Table->GetNumRows();

       ReVarCifArray<CifString> *ColumnNames = Table->GetColumnNames();
       int *cols = new int[colNo];

       int ndb_tokenid = -1, i, j, tokenid, fieldno, tmp_id, row;
       CifString value;
       One_Card_Info_Format *pCard1 = NULL, *pCard2 = NULL, *pCard3 = NULL;

       for (i = 0; i < colNo; i++) {
            for (j = 0; j < CifCategory[id].nkeywords; j++) {
                 if (!strcasecmp((*ColumnNames)[i].Text(),
                        CifCategory[id].keywords[j].name)) {
                         tmp_id = CifCategory[id].keywords[j].
                                      NdbInfo[0].NdbTokenId - 1;
                         if (tmp_id != HEADER_TOKEN-1 && tmp_id >= 0)
                              ndb_tokenid = tmp_id;
                      break;
                 }
            }
            if (j == CifCategory[id].nkeywords) cols[i] = -1;
            else                                cols[i] = j;
       }
       if (ndb_tokenid == -1) {
            if (ColumnNames) delete ColumnNames;
            delete cols;
            return;
       }

       for (row = 0; row < rowNo; row++) {
            pCard1 = _CardInfo[ndb_tokenid].add_a_new_card(NdbToken[ndb_tokenid], 0);
            pCard2 = _CardInfo[ndb_tokenid].add_a_new_card(NdbToken[ndb_tokenid], 0);
            pCard3 = _CardInfo[ndb_tokenid].add_a_new_card(NdbToken[ndb_tokenid], 0);
            for (i = 0; i < colNo; i++) {
                 j = cols[i];
                 if (j < 0 || j > CifCategory[id].nkeywords) continue;

                 if (!get_value(value, Table, i, row)) continue;

                 tokenid = CifCategory[id].keywords[j].NdbInfo[0].NdbTokenId-1;
                 fieldno = CifCategory[id].keywords[j].NdbInfo[0].NdbFieldNo-1;
                 if (fieldno < 0 || tokenid != ndb_tokenid) continue;

                 if (strstr((*ColumnNames)[i].Text(), "x[1]") ||
                     strstr((*ColumnNames)[i].Text(), "vector[1]")) {
                      pCard1->AddElementToField(value.Text(), fieldno);
                 } else if (strstr((*ColumnNames)[i].Text(), "x[2]") ||
                            strstr((*ColumnNames)[i].Text(), "vector[2]")) {
                      pCard2->AddElementToField(value.Text(), fieldno);
                 } else if (strstr((*ColumnNames)[i].Text(), "x[3]") ||
                            strstr((*ColumnNames)[i].Text(), "vector[3]")) {
                      pCard3->AddElementToField(value.Text(), fieldno);
                 } else {
                      if (value == "given") value.Copy("1");
                      else if (value == "generate") value.Copy("");
                      pCard1->AddElementToField(value.Text(), fieldno);
                      pCard2->AddElementToField(value.Text(), fieldno);
                      pCard3->AddElementToField(value.Text(), fieldno);
                 }
            }
       }
       if (ColumnNames) delete ColumnNames;
       delete [] cols;
}

void _MAXIT::cifndb_process_struct_site_gen(ISTable *Table)
{
       if (_num_mol == 0 || _molecules == NULL) return;

       int rowNo = Table->GetNumRows();

       int site_id_index = Table->GetColumnIndex("site_id");
       int label_comp_id = Table->GetColumnIndex("label_comp_id");
       int label_asym_id = Table->GetColumnIndex("label_asym_id");
       int label_seq_id  = Table->GetColumnIndex("label_seq_id");
       int auth_comp_id  = Table->GetColumnIndex("auth_comp_id");
       int auth_asym_id  = Table->GetColumnIndex("auth_asym_id");
       int auth_seq_id   = Table->GetColumnIndex("auth_seq_id");
       int ndb_ins_code  = Table->GetColumnIndex("ndb_label_ins_code");
       int symmetry      = Table->GetColumnIndex("symmetry");
 
       CifString site_id, name1, symop;
       int irow, FstIndex, FstRes;
       for (irow = 0; irow < rowNo; irow++) {
            get_value(site_id, Table, site_id_index, irow);
            get_value(symop, Table, symmetry, irow);
            if (get_residue_index(_molecules[0], Table, irow, label_comp_id,
                      label_asym_id, label_seq_id, auth_comp_id, auth_asym_id,
                      auth_seq_id, ndb_ins_code, FstIndex, FstRes, name1)) {
                 Insert_a_Site(site_id.Text(), FstIndex, FstRes, symop.Text());
            }
       }
}

void _MAXIT::cifndb_process_rcsb_link(ISTable *Table)
{
       int rowNo, irow, BegIndex, BegRes, EndIndex, EndRes, beg, end;
       int ptnr1_label_alt_id, ptnr1_label_asym_id, ptnr1_label_atom_id;
       int ptnr1_label_comp_id, ptnr1_label_seq_id, ptnr1_label_ins_code;
       int ptnr1_symmetry, ptnr2_label_alt_id, ptnr2_label_asym_id;
       int ptnr2_label_atom_id, ptnr2_label_comp_id, ptnr2_label_seq_id;
       int ptnr2_label_ins_code, ptnr2_symmetry, details;
       int i, mol_index;
       char buffer[10];
       float d = 0;
       CifString comment, atom1, loc1, name1, atom2, loc2, name2, symm1, symm2;
       CifString type, dist;

       if (_num_mol == 0 || _molecules == NULL) return;

       rowNo = Table->GetNumRows();
       ptnr1_label_alt_id = Table->GetColumnIndex("ptnr1_label_alt_id");
       ptnr1_label_asym_id = Table->GetColumnIndex("ptnr1_label_asym_id");
       ptnr1_label_atom_id = Table->GetColumnIndex("ptnr1_label_atom_id");
       ptnr1_label_comp_id = Table->GetColumnIndex("ptnr1_label_comp_id");
       ptnr1_label_seq_id = Table->GetColumnIndex("ptnr1_label_seq_id");
       ptnr1_label_ins_code = Table->GetColumnIndex("ptnr1_label_ins_code");
       ptnr1_symmetry = Table->GetColumnIndex("ptnr1_symmetry");
       ptnr2_label_alt_id = Table->GetColumnIndex("ptnr2_label_alt_id");
       ptnr2_label_asym_id = Table->GetColumnIndex("ptnr2_label_asym_id");
       ptnr2_label_atom_id = Table->GetColumnIndex("ptnr2_label_atom_id");
       ptnr2_label_comp_id = Table->GetColumnIndex("ptnr2_label_comp_id");
       ptnr2_label_seq_id = Table->GetColumnIndex("ptnr2_label_seq_id");
       ptnr2_label_ins_code = Table->GetColumnIndex("ptnr2_label_ins_code");
       ptnr2_symmetry = Table->GetColumnIndex("ptnr2_symmetry");
       details = Table->GetColumnIndex("details");
       int dist_index = Table->GetColumnIndex("pdbx_dist_value");
       // if (dist_index < 0) dist_index = Table->GetColumnIndex("dist");
       int type_index = Table->GetColumnIndex("type");

       for (irow = 0; irow < rowNo; irow++) {
            get_value(comment, Table, details, irow);
            get_value(loc1, Table, ptnr1_label_alt_id, irow);
            get_value(atom1, Table, ptnr1_label_atom_id, irow);
            get_value(symm1, Table, ptnr1_symmetry, irow);
            get_value(loc2, Table, ptnr2_label_alt_id, irow);
            get_value(atom2, Table, ptnr2_label_atom_id, irow);
            get_value(symm2, Table, ptnr2_symmetry, irow);
            get_value(dist, Table, dist_index, irow);
            get_value(type, Table, type_index, irow);
            if (symm1 == "1_555") symm1.Clear();
            if (symm2 == "1_555") symm2.Clear();
            mol_index = 0;
            beg = 0;
            end = 0;
            for (i = 0; i < _num_mol; i++) {
                 beg = get_residue_index(_molecules[i], Table, irow, -1, -1, -1,
                        ptnr1_label_comp_id, ptnr1_label_asym_id,
                        ptnr1_label_seq_id, ptnr1_label_ins_code,
                        BegIndex, BegRes, name1);
                 end = get_residue_index(_molecules[i], Table, irow, -1, -1, -1,
                        ptnr2_label_comp_id, ptnr2_label_asym_id,
                        ptnr2_label_seq_id, ptnr2_label_ins_code,
                        EndIndex, EndRes, name2);
                 if (beg && end) {
                      _CHAIN *ch1 = _molecules[i]->GetIndexChain(BegIndex);
                      _CHAIN *ch2 = _molecules[i]->GetIndexChain(EndIndex);
                      if (!ch1 || !ch2) continue;
                      _FIELD *field1 = ch1->SeqResIndex(BegRes);
                      _FIELD *field2 = ch2->SeqResIndex(EndRes);
                      if (!field1 || !field2) continue;
                      _RESIDUE *res1 = field1->Residue;
                      _RESIDUE *res2 = field2->Residue;
                      if (!res1 || !res2) continue;

                      int found = 0, num_a = 0, num_b = 0;
                      _ATOM **atom_a = res1->find_atom(atom1.Text(), num_a);
                      _ATOM **atom_b = res2->find_atom(atom2.Text(), num_b);
                      if (atom_a && atom_b) {
                           for (int a = 0; a < num_a; a++) {
                                if (strcmp(atom_a[a]->alt_loc(), loc1.Text()) ||
                                    strcmp(atom_a[a]->restype(), name1.Text())) continue;
                                for (int b = 0; b < num_b; b++) {
                                     if (strcmp(atom_b[b]->alt_loc(), loc2.Text()) ||
                                         strcmp(atom_b[b]->restype(), name2.Text())) continue;
                                     found = 1;
                                     if (type == "") {
                                          type.Copy("covale");
                                          if (find_metal_flag(atom_a[a]->atom_type()) == 2 ||
                                              find_metal_flag(atom_b[b]->atom_type()) == 2)
                                               type.Copy("metalc");
                                     }
                                     // if (dist == "") {
                                          if (symm1 == "" && symm2 == "")
                                               d = cal_distance(atom_a[a], atom_b[b]);
                                          else if (symm1 == "")
                                               d = cal_symmetry_distance(atom_a[a]->orig(),
                                                       atom_b[b]->orig(), symm2.Text(), _cell);
                                          else if (symm2 == "")
                                               d = cal_symmetry_distance(atom_b[b]->orig(),
                                                       atom_a[a]->orig(), symm1.Text(), _cell);
                                          memset(buffer, 0, 10);
                                          sprintf(buffer, "%.3f", d);
                                          dist.Copy(buffer);
                                     // }
                                }
                           }
                      }
                      if (atom_a) delete [] atom_a;
                      if (atom_b) delete [] atom_b;
                      if (found) {
                           mol_index = i;
                           break;
                      }
                 }
            }
            if (!beg || !end) continue;

            Insert_a_Link(mol_index, atom1.Text(), loc1.Text(), name1.Text(), BegIndex,
                          BegRes, atom2.Text(), loc2.Text(), name2.Text(), EndIndex,
                          EndRes, symm1.Text(), symm2.Text(), comment.Text(),
                          type.Text(), dist.Text(), "");
       }
}

void _MAXIT::cifndb_process_struct_conf(ISTable *Table)
{
       int BegIndex = -1, BegRes = -1, EndIndex = -1, EndRes = -1, beg = -1, end = -1;
       int irow, rowNo, id_index, pdb_index, beg_label_comp_id, beg_label_asym_id;
       int beg_label_seq_id, beg_label_ins_code, beg_auth_comp_id, beg_auth_asym_id;
       int beg_auth_seq_id, end_label_comp_id, end_label_asym_id, end_label_seq_id;
       int end_label_ins_code, end_auth_comp_id, end_auth_asym_id, end_auth_seq_id;
       int class_index, details_index;
       CifString pdb_id, type, comment, class_type, name1, name2;
       _CHAIN *chain = NULL;

       if (_num_mol == 0 || _molecules == NULL) return;

       rowNo = Table->GetNumRows();
       id_index = Table->GetColumnIndex("conf_type_id");
       pdb_index = Table->GetColumnIndex("pdb_id");
       beg_label_comp_id = Table->GetColumnIndex("beg_label_comp_id");
       beg_label_asym_id = Table->GetColumnIndex("beg_label_asym_id");
       beg_label_seq_id = Table->GetColumnIndex("beg_label_seq_id");
       beg_label_ins_code = Table->GetColumnIndex("ndb_beg_label_ins_code_pdb");
       beg_auth_comp_id = Table->GetColumnIndex("beg_auth_comp_id");
       beg_auth_asym_id = Table->GetColumnIndex("beg_auth_asym_id");
       beg_auth_seq_id = Table->GetColumnIndex("beg_auth_seq_id");
       end_label_comp_id = Table->GetColumnIndex("end_label_comp_id");
       end_label_asym_id = Table->GetColumnIndex("end_label_asym_id");
       end_label_seq_id = Table->GetColumnIndex("end_label_seq_id");
       end_label_ins_code = Table->GetColumnIndex("ndb_end_label_ins_code_pdb");
       end_auth_comp_id = Table->GetColumnIndex("end_auth_comp_id");
       end_auth_asym_id = Table->GetColumnIndex("end_auth_asym_id");
       end_auth_seq_id = Table->GetColumnIndex("end_auth_seq_id");
       class_index = Table->GetColumnIndex("ndb_helix_class_pdb");
       details_index = Table->GetColumnIndex("details");

       for (irow = 0; irow < rowNo; irow++) {
            beg = get_residue_index(_molecules[0], Table, irow, beg_label_comp_id,
                  beg_label_asym_id, beg_label_seq_id, beg_auth_comp_id,
                  beg_auth_asym_id, beg_auth_seq_id, beg_label_ins_code,
                  BegIndex, BegRes, name1);
            end = get_residue_index(_molecules[0], Table, irow, end_label_comp_id,
                  end_label_asym_id, end_label_seq_id, end_auth_comp_id,
                  end_auth_asym_id, end_auth_seq_id, end_label_ins_code,
                  EndIndex, EndRes, name2);
            if (beg && end && BegIndex == EndIndex) {
                 chain = _molecules[0]->GetIndexChain(BegIndex);
                 if (chain) {
                      get_value(type, Table, id_index, irow);
                      get_value(pdb_id, Table, pdb_index, irow);
                      get_value(comment, Table, details_index, irow);
                      if (!strcasecmp(type.Text(), "HELX_P")) {
                           get_value(class_type, Table, class_index, irow);
                           chain->Insert_a_Helix(pdb_id.Text(), BegRes, EndRes,
                                 atoi(class_type.Text()), comment.Text());
                           _exist_helix = 1;
                      } else if (!strcasecmp(type.Text(), "TURN_P")) {
                           chain->Insert_a_Turn(pdb_id.Text(), BegRes, EndRes, comment.Text());
                      }
                 }
            }
       }
}

void _MAXIT::cifndb_process_sheet()
{
       if (_num_mol == 0 || _molecules == NULL) return;

       ISTable *Table = _CifObj->GetTablePtr(_StructureId, "struct_sheet_range");
       if (is_empty_table(Table)) return;

       ISTable *Table1 = _CifObj->GetTablePtr(_StructureId, "rcsb_struct_sheet_hbond");
       ISTable *Table2 = _CifObj->GetTablePtr(_StructureId, "struct_sheet_order");
       ISTable *Table3 = _CifObj->GetTablePtr(_StructureId, "struct_sheet");

       int rowNo = Table->GetNumRows();
       int sheet_id_index = Table->GetColumnIndex("sheet_id");
       int id_index = Table->GetColumnIndex("id");
       int beg_label_comp_id = Table->GetColumnIndex("beg_label_comp_id");
       int beg_label_asym_id = Table->GetColumnIndex("beg_label_asym_id");
       int beg_label_seq_id = Table->GetColumnIndex("beg_label_seq_id");
       int ndb_beg_label_ins_code_pdb = Table->GetColumnIndex("ndb_beg_label_ins_code_pdb");
       int end_label_comp_id = Table->GetColumnIndex("end_label_comp_id");
       int end_label_asym_id = Table->GetColumnIndex("end_label_asym_id");
       int end_label_seq_id = Table->GetColumnIndex("end_label_seq_id");
       int ndb_end_label_ins_code_pdb = Table->GetColumnIndex("ndb_end_label_ins_code_pdb");
       int beg_auth_comp_id = Table->GetColumnIndex("beg_auth_comp_id");
       int beg_auth_asym_id = Table->GetColumnIndex("beg_auth_asym_id");
       int beg_auth_seq_id = Table->GetColumnIndex("beg_auth_seq_id");
       int end_auth_comp_id = Table->GetColumnIndex("end_auth_comp_id");
       int end_auth_asym_id = Table->GetColumnIndex("end_auth_asym_id");
       int end_auth_seq_id = Table->GetColumnIndex("end_auth_seq_id");

       int row, BegIndex, BegRes, EndIndex, EndRes, beg, end, number_strands;
       int sense = 0, curIndex = -1, curRes = -1, prevIndex = -1, prevRes = -1;
       CifString sheet_id, id, prev_id, curAtom, prevAtom, name1, name2;
       prev_id[0] = '\0';
       prev_id.Clear();
       id.Clear();
       for (row = 0; row < rowNo; row++) {
            prev_id.Copy(id);
            sheet_id.Clear();
            id.Clear();
            get_value(sheet_id, Table, sheet_id_index, row);
            get_value(id, Table, id_index, row);

            beg = get_residue_index(_molecules[0], Table, row, beg_label_comp_id,
                  beg_label_asym_id, beg_label_seq_id, beg_auth_comp_id,
                  beg_auth_asym_id, beg_auth_seq_id, ndb_beg_label_ins_code_pdb,
                  BegIndex, BegRes, name1);
            end = get_residue_index(_molecules[0], Table, row, end_label_comp_id,
                  end_label_asym_id, end_label_seq_id, end_auth_comp_id,
                  end_auth_asym_id, end_auth_seq_id, ndb_end_label_ins_code_pdb,
                  EndIndex, EndRes, name2);
            if (beg && end && BegIndex == EndIndex) {
                 sense = 0; curIndex = -1; curRes = -1; prevIndex = -1; prevRes = -1;
                 curAtom.Clear(); prevAtom.Clear();
                 number_strands = get_number_strands(Table3, sheet_id.Text());
                 if (row) {
                      get_hbond(_molecules[0], Table1, BegIndex, sheet_id.Text(), id.Text(),
                         curAtom, curIndex, curRes, prevAtom, prevIndex, prevRes);
                      sense = get_sense(Table2, sheet_id.Text(), prev_id.Text(), id.Text());
                 }
                 Insert_a_Sheet(sheet_id.Text(), number_strands, BegIndex, BegRes, EndIndex,
                       EndRes, sense, curAtom.Text(), curIndex, curRes,
                       prevAtom.Text(), prevIndex, prevRes);
                 _exist_sheet = 1;
            }
       }
}

#ifdef SUN_OS
#ifdef __cplusplus
extern "C" {
#endif
#endif
static int pdb_rev_compare(const void *a, const void *b)
{
       return (((PDB_REV *) b)->num - ((PDB_REV *) a)->num);
}
#ifdef SUN_OS
#ifdef __cplusplus
}
#endif
#endif

void _MAXIT::cifndb_process_database_pdb_rev(ISTable *Table)
{
       int rowNo = Table->GetNumRows();
       PDB_REV *pdb_rev = (PDB_REV *) calloc(rowNo, sizeof(PDB_REV));
       int n_rev = rowNo;
       int num_index = Table->GetColumnIndex("num");
       int date_index = Table->GetColumnIndex("date");
       int original_index = Table->GetColumnIndex("date_original");
       int replaces_index = Table->GetColumnIndex("replaces");
       int type_index = Table->GetColumnIndex("mod_type");
       int revised_index[4];
       revised_index[0] = Table->GetColumnIndex("rcsb_record_revised_1");
       revised_index[1] = Table->GetColumnIndex("rcsb_record_revised_2");
       revised_index[2] = Table->GetColumnIndex("rcsb_record_revised_3");
       revised_index[3] = Table->GetColumnIndex("rcsb_record_revised_4");

       int irow, i, len, serial_no, pos;
       CifString cifstring, value;
       char buffer[20], *p;
       One_Card_Info_Format *pCard = NULL;
       for (irow = 0; irow < rowNo; irow++) {
            pdb_rev[irow].num = 0;
            pdb_rev[irow].date[0] = '\0';
            pdb_rev[irow].replaces[0] = '\0';
            pdb_rev[irow].mod_type = 0;
            pdb_rev[irow].record = NULL;
            if (num_index >= 0) {
                 get_value(cifstring, Table, num_index, irow);
                 pdb_rev[irow].num = atoi(cifstring.Text());
            }
            if (date_index >= 0) {
                 get_value(cifstring, Table, date_index, irow);
                 strcpy(pdb_rev[irow].date, cifstring.Text());
            }
            if (replaces_index >= 0) {
                 get_value(cifstring, Table, replaces_index, irow);
                 strcpy(pdb_rev[irow].replaces, cifstring.Text());
            }
            if (type_index >= 0) {
                 get_value(cifstring, Table, type_index, irow);
                 pdb_rev[irow].mod_type = atoi(cifstring.Text());
            }
            if (original_index >= 0 && !pdb_rev[irow].mod_type) {
                 get_value(cifstring, Table, original_index, irow);
                 if (cifstring != "") {
                      pCard = _CardInfo[HEADER_TOKEN-1].Head();
                      if (!pCard) pCard = _CardInfo[HEADER_TOKEN-1].
                           add_a_new_card(NdbToken[HEADER_TOKEN-1], 0);
                      pCard->AddElementToField(cifstring.Text(), 2);
                 }
            }
            for (i = 0; i < 4; i++) {
                 if (revised_index[i] < 0) continue;
                 get_value(cifstring, Table, revised_index[i], irow);
                 if (cifstring == "") continue;

                 if (pdb_rev[irow].record == NULL) {
                      len = strlen(cifstring.Text())+1;
                      pdb_rev[irow].record = (char *) calloc(len, sizeof(char));
                      strcpy(pdb_rev[irow].record, cifstring.Text());
                 } else if (!strstr(pdb_rev[irow].record, cifstring.Text())) {
                      len = strlen(cifstring.Text()) + 2 +
                                strlen(pdb_rev[irow].record);
                      pdb_rev[irow].record = (char *) realloc(pdb_rev[irow].record,
                                            len * sizeof(char));
                      strcat(pdb_rev[irow].record, " ");
                      strcat(pdb_rev[irow].record, cifstring.Text());
                 }
            }
       }

       Table = _CifObj->GetTablePtr(_StructureId, "database_PDB_rev_record");
       if (!is_empty_table(Table)) {
            rowNo = Table->GetNumRows();
            num_index = Table->GetColumnIndex("rev_num");
            type_index = Table->GetColumnIndex("type");
            for (irow = 0; irow < rowNo; irow++) {
                 get_value(cifstring, Table, num_index, irow);
                 get_value(value, Table, type_index, irow);
                 if (cifstring == "" || value == "") continue;

                 for (i = 0; i < n_rev; i++)
                      if (pdb_rev[i].num == atoi(cifstring.Text())) break;
                 if (i < n_rev) {
                      if (pdb_rev[i].record == NULL) {
                           len = strlen(value.Text())+1;
                           pdb_rev[i].record = (char *) calloc(len, sizeof(char));
                           strcpy(pdb_rev[i].record, value.Text());
                      } else if (!strstr(pdb_rev[i].record, value.Text())) {
                           len = strlen(value.Text()) + 2 +
                                     strlen(pdb_rev[i].record);
                           pdb_rev[i].record = (char *) realloc(pdb_rev[i].record,
                                                 len * sizeof(char));
                           strcat(pdb_rev[i].record, " ");
                           strcat(pdb_rev[i].record, value.Text());
                      }
                 }
            }
       }
       if (n_rev > 1) qsort((void *) pdb_rev, n_rev, sizeof(PDB_REV), pdb_rev_compare);

       for (i = 0; i < n_rev; i++) {
            if (!pdb_rev[i].num) continue;
            pCard = _CardInfo[REVDAT_TOKEN-1].
                        add_a_new_card(NdbToken[REVDAT_TOKEN-1], 0);
            pCard->AddElementToField(pdb_rev[i].num, 1);
            pCard->AddElementToField(pdb_rev[i].date, 3);
            pCard->AddElementToField(pdb_rev[i].replaces, 4);
            pCard->AddElementToField(pdb_rev[i].mod_type, 5);
            p = pdb_rev[i].record;
            serial_no = 2;
            pos = 6;
            while (p) {
                 if (pos > 9) {
                      pCard = _CardInfo[REVDAT_TOKEN-1].
                              add_a_new_card(NdbToken[REVDAT_TOKEN-1], 0);
                      pCard->AddElementToField(pdb_rev[i].num, 1);
                      pCard->AddElementToField(serial_no, 2);
                      pCard->AddElementToField(pdb_rev[i].mod_type, 5);
                      serial_no++;
                      pos = 6;
                 }
                 p = get_word(p, buffer);
                 pCard->AddElementToField(buffer, pos);
                 pos++;
            }
       }
       for (i = 0; i < n_rev; i++) {
            if (pdb_rev[i].record) free ((void *) pdb_rev[i].record);
       }
       free ((void *) pdb_rev);
}

void _MAXIT::cifndb_process_rcsb_database_related(ISTable *Table)
{
       int rowNo = Table->GetNumRows();
       int db_name_idx = Table->GetColumnIndex("db_name");
       int db_id_idx = Table->GetColumnIndex("db_id");
       int details_idx = Table->GetColumnIndex("details");
       int type_idx = Table->GetColumnIndex("content_type");
       ReVarCifArray<CifString> IDs;
       IDs.Clear();
       CifString cs, cs1, cs2;
       One_Card_Info_Format *pCard = NULL;
       for (int i = 0; i < rowNo; i++) {
            get_value(cs, Table, type_idx, i);
            get_value(cs1, Table, db_id_idx, i);
            cs.ToLower();
            if (cs == "split") {
                 IDs.Add(cs1);
                 continue;
            }
            get_value(cs, Table, db_name_idx, i);
            get_value(cs2, Table, details_idx, i);
            if (cs == "" && cs1 == "" && cs2 == "") continue;

            pCard = _CardInfo[RENTRY_TOKEN-1].
                        add_a_new_card(NdbToken[RENTRY_TOKEN-1], 0);
            pCard->AddElementToField(cs.Text(), 2); 
            pCard->AddElementToField(cs1.Text(), 3); 
            pCard->AddElementToField(cs2.Text(), 4); 
       }

       if (IDs.Length() == 0) return;
       if (SPLIT_TOKEN <= 0) return;

       pCard = _CardInfo[SPLIT_TOKEN-1].add_a_new_card(NdbToken[SPLIT_TOKEN-1], 0);
       int pos = 2;
       int serial_no = 2;
       for (unsigned int i = 0; i < IDs.Length(); i++) {
            if (pos >= pCard->NumField()) {
                 pCard = _CardInfo[SPLIT_TOKEN-1].
                             add_a_new_card(NdbToken[SPLIT_TOKEN-1], 0);
                 pCard->AddElementToField(serial_no, 1);
                 serial_no++;
                 pos = 2;
            }
            pCard->AddElementToField(IDs[i].Text(), pos);
            pos++;
       }
}

void _MAXIT::cifndb_process_struct_conn(ISTable *Table)
{
       if (_num_mol == 0 || _molecules == NULL) return;

       int rowNo = Table->GetNumRows();

       int id_index = Table->GetColumnIndex("conn_type_id");
       // int pdb_index = Table->GetColumnIndex("ndb_PDB_id");
       int name_index = Table->GetColumnIndex("ndb_ptnr1_mod_name");
       int sugar_index = Table->GetColumnIndex("ndb_ptnr1_sugar_name");
       int replaced_index = Table->GetColumnIndex("ndb_ptnr1_replaced_atom");
       // int standard_index = Table->GetColumnIndex("ndb_ptnr1_standard_comp_id");
       int ptnr1_label_comp_id = Table->GetColumnIndex("ptnr1_label_comp_id");
       int ptnr1_label_asym_id = Table->GetColumnIndex("ptnr1_label_asym_id");
       int ptnr1_label_seq_id = Table->GetColumnIndex("ptnr1_label_seq_id");
       int ptnr1_label_atom_id = Table->GetColumnIndex("ptnr1_label_atom_id");
       int ndb_ptnr1_label_alt_loc = Table->GetColumnIndex("ndb_ptnr1_label_alt_loc");
       int ptnr1_label_ins_code = Table->GetColumnIndex("ndb_ptnr1_label_ins_code");
       int ptnr1_symmetry = Table->GetColumnIndex("ptnr1_symmetry");
       int ptnr2_label_comp_id = Table->GetColumnIndex("ptnr2_label_comp_id");
       int ptnr2_label_asym_id = Table->GetColumnIndex("ptnr2_label_asym_id");
       int ptnr2_label_seq_id = Table->GetColumnIndex("ptnr2_label_seq_id");
       int ptnr2_label_atom_id = Table->GetColumnIndex("ptnr2_label_atom_id");
       int ndb_ptnr2_label_alt_loc = Table->GetColumnIndex("ndb_ptnr2_label_alt_loc");
       int ptnr2_label_ins_code = Table->GetColumnIndex("ndb_ptnr2_label_ins_code");
       int ptnr2_symmetry = Table->GetColumnIndex("ptnr2_symmetry");
       int ptnr1_auth_comp_id = Table->GetColumnIndex("ptnr1_auth_comp_id");
       int ptnr1_auth_asym_id = Table->GetColumnIndex("ptnr1_auth_asym_id");
       int ptnr1_auth_seq_id = Table->GetColumnIndex("ptnr1_auth_seq_id");
       int ptnr2_auth_comp_id = Table->GetColumnIndex("ptnr2_auth_comp_id");
       int ptnr2_auth_asym_id = Table->GetColumnIndex("ptnr2_auth_asym_id");
       int ptnr2_auth_seq_id = Table->GetColumnIndex("ptnr2_auth_seq_id");
       int ndb_ptnr3_label_atom_id = Table->GetColumnIndex("ndb_ptnr3_label_atom_id");
       int ndb_ptnr3_label_seq_id = Table->GetColumnIndex("ndb_ptnr3_label_seq_id");
       int ndb_ptnr3_label_asym_id = Table->GetColumnIndex("ndb_ptnr3_label_asym_id");
       int ndb_ptnr3_label_alt_loc = Table->GetColumnIndex("ndb_ptnr3_label_alt_loc");
       int details_index = Table->GetColumnIndex("details");
       int dist_index = Table->GetColumnIndex("pdbx_dist_value");
       // if (dist_index < 0) dist_index = Table->GetColumnIndex("dist");
       int bond_index = Table->GetColumnIndex("pdbx_value_order");
       

       int irow, BegIndex, BegRes, EndIndex, EndRes, beg, end, H_Index, H_Res, i, mol_index;
       CifString type, atom1, loc1, name1, atom2, loc2, name2, symm1, symm2, comment, name;
       CifString Replaced_Atom, atom3, loc3, name3, standard, pdb_id, sugar, dist;
       CifString seq_id1, seq_id2, seq_id3, chain_id1, chain_id2, chain_id3, bondtype;
       char buffer[10];
       float d = 0;
       for (irow = 0; irow < rowNo; irow++) {
            get_value(type, Table, id_index, irow);
            type.ToLower();
            if (type == "") continue;

            get_value(atom1, Table, ptnr1_label_atom_id, irow);
            get_value(loc1, Table, ndb_ptnr1_label_alt_loc, irow);
            get_value(symm1, Table, ptnr1_symmetry, irow);
            get_value(atom2, Table, ptnr2_label_atom_id, irow);
            get_value(loc2, Table, ndb_ptnr2_label_alt_loc, irow);
            get_value(symm2, Table, ptnr2_symmetry, irow);
            get_value(comment, Table, details_index, irow);
            get_value(name, Table, name_index, irow);
            get_value(Replaced_Atom, Table, replaced_index, irow);
            get_value(dist, Table, dist_index, irow);
            get_value(bondtype, Table, bond_index, irow);
            if (symm1 == "1_555") symm1.Clear();
            if (symm2 == "1_555") symm2.Clear();
            mol_index = 0;
            beg = 0;
            end = 0;
            for (i = 0; i < _num_mol; i++) {
                 beg = get_residue_index(_molecules[i], Table, irow, ptnr1_label_comp_id,
                       ptnr1_label_asym_id, ptnr1_label_seq_id, ptnr1_auth_comp_id,
                       ptnr1_auth_asym_id, ptnr1_auth_seq_id, ptnr1_label_ins_code,
                       BegIndex, BegRes, name1);
                 end = get_residue_index(_molecules[i], Table, irow, ptnr2_label_comp_id,
                       ptnr2_label_asym_id, ptnr2_label_seq_id, ptnr2_auth_comp_id,
                       ptnr2_auth_asym_id, ptnr2_auth_seq_id, ptnr2_label_ins_code,
                       EndIndex, EndRes, name2);
                 if (beg && end) {
                      if (atom1 != "") {
                           _CHAIN *ch1 = _molecules[i]->GetIndexChain(BegIndex);
                           if (!ch1) continue;
                           _FIELD *field1 = ch1->SeqResIndex(BegRes);
                           if (!field1) continue;
                           _RESIDUE *res1 = field1->Residue;
                           int found = 0, num_a = 0, num_b = 0;
                           _ATOM **atom_a = res1->find_atom(atom1.Text(), num_a);
                           if (atom_a) {
                                if (type == "link" || type == "covale" ||
                                    type == "metalc" || type == "disulf") {
                                     if (dist == "" && type != "disulf") type.Copy("");
                                     _CHAIN *ch2 = _molecules[i]->GetIndexChain(EndIndex);
                                     if (!ch2) {
                                          delete [] atom_a;
                                          continue;
                                     }
                                     _FIELD *field2 = ch2->SeqResIndex(EndRes);
                                     if (!field2) {
                                          delete [] atom_a;
                                          continue;
                                     }
                                     _RESIDUE *res2 = field2->Residue;
                                     if (!res2) {
                                          delete [] atom_a;
                                          continue;
                                     }
                                     _ATOM **atom_b = res2->find_atom(atom2.Text(), num_b);
                                     if (atom_b) {
                                          for (int a = 0; a < num_a; a++) {
                                               if (strcmp(atom_a[a]->alt_loc(), loc1.Text()) ||
                                                   strcmp(atom_a[a]->restype(), name1.Text())) continue;
                                               for (int b = 0; b < num_b; b++) {
                                                    if (strcmp(atom_b[b]->alt_loc(), loc2.Text()) ||
                                                        strcmp(atom_b[b]->restype(), name2.Text())) continue;
                                                    found = 1;
                                                    if (type == "") {
                                                         type.Copy("covale");
                                                         if (find_metal_flag(atom_a[a]->atom_type()) == 2 ||
                                                             find_metal_flag(atom_b[b]->atom_type()) == 2)
                                                              type.Copy("metalc");
                                                    }
                                                    // if (dist == "") {
                                                         if (symm1 == "" && symm2 == "")
                                                              d = cal_distance(atom_a[a], atom_b[b]);
                                                         else if (symm1 == "")
                                                              d = cal_symmetry_distance(atom_a[a]->orig(),
                                                                      atom_b[b]->orig(), symm2.Text(), _cell);
                                                         else if (symm2 == "")
                                                              d = cal_symmetry_distance(atom_b[b]->orig(),
                                                                      atom_a[a]->orig(), symm1.Text(), _cell); 
                                                         memset(buffer, 0, 10);
                                                         sprintf(buffer, "%.3f", d);
                                                         dist.Copy(buffer);
                                                    // }
                                               }
                                          }
                                          delete [] atom_b;
                                     }
                                } else {
                                     for (int a = 0; a < num_a; a++) {
                                          if (strcmp(atom_a[a]->alt_loc(), loc1.Text()) ||
                                              strcmp(atom_a[a]->restype(), name1.Text())) continue;
                                          found = 1;
                                          break;
                                     }
                                }
                                delete [] atom_a;
                           }
                           if (found) {
                                mol_index = i;
                                break;
                           }
                      } else {
                           mol_index = i;
                           break;
                      }
                 }
            }
            if (!beg) continue;
            if (type == "") continue;

            if (!strcasecmp(type.Text(), "hydrog")) {
                 if (end) {
                      if (!strcasecmp(comment.Text(), "WATSON-CRICK") ||
                          !strcasecmp(comment.Text(), "REVERSED WATSON-CRICK") ||
                          !strcasecmp(comment.Text(), "HOOGSTEEN") ||
                          !strcasecmp(comment.Text(), "REVERSED HOOGSTEEN") ||
                          strstr(comment.Text(), "PAIR") || strstr(comment.Text(), "pair"))
                           Insert_a_Bspair(atom1.Text(), name1.Text(), BegIndex, BegRes,
                                 atom2.Text(), name2.Text(), EndIndex, EndRes, symm2.Text(),
                                 comment.Text());
                      else {
                           seq_id3.Copy("ZZZ"); chain_id3.Copy("ZZZ");
                           get_value(atom3, Table, ndb_ptnr3_label_atom_id, irow);
                           get_value(loc3, Table, ndb_ptnr3_label_alt_loc, irow);
                           get_value(seq_id1, Table, ptnr1_label_seq_id, irow);
                           get_value(chain_id1, Table, ptnr1_label_asym_id, irow);
                           get_value(seq_id2, Table, ptnr2_label_seq_id, irow);
                           get_value(chain_id2, Table, ptnr2_label_asym_id, irow);
                           if (ndb_ptnr3_label_seq_id >= 0)
                                get_value(seq_id3, Table, ndb_ptnr3_label_seq_id, irow);
                           if (ndb_ptnr3_label_asym_id >= 0)
                                get_value(chain_id3, Table, ndb_ptnr3_label_asym_id, irow);
                           H_Index = -1; H_Res = -1;
                           if (seq_id1 == seq_id3 && chain_id1 == chain_id3) {
                                H_Index = BegIndex; H_Res = BegRes;
                                name3.Copy(name1.Text());
                           } else if (seq_id2 == seq_id3 && chain_id2 == chain_id3) {
                                H_Index = EndIndex; H_Res = EndRes;
                                name3.Copy(name2.Text());
                           }
                           Insert_a_Hydbnd(atom1.Text(), loc1.Text(), name1.Text(), BegIndex, BegRes,
                                 atom3.Text(), loc3.Text(), name3.Text(), H_Index, H_Res, atom2.Text(),
                                 loc2.Text(), name2.Text(), EndIndex, EndRes, symm1.Text(), symm2.Text());
                      }
                 }
            } else if (!strcasecmp(type.Text(), "hydbnd")) {
                 if (end) {
                      seq_id3.Copy("ZZZ"); chain_id3.Copy("ZZZ");
                      get_value(atom3, Table, ndb_ptnr3_label_atom_id, irow);
                      get_value(loc3, Table, ndb_ptnr3_label_alt_loc, irow);
                      get_value(seq_id1, Table, ptnr1_label_seq_id, irow);
                      get_value(chain_id1, Table, ptnr1_label_asym_id, irow);
                      get_value(seq_id2, Table, ptnr2_label_seq_id, irow);
                      get_value(chain_id2, Table, ptnr2_label_asym_id, irow);
                      if (ndb_ptnr3_label_seq_id >= 0)
                           get_value(seq_id3, Table, ndb_ptnr3_label_seq_id, irow);
                      if (ndb_ptnr3_label_asym_id >= 0)
                           get_value(chain_id3, Table, ndb_ptnr3_label_asym_id, irow);
                      H_Index = -1; H_Res = -1;
                      if (seq_id1 == seq_id3 && chain_id1 == chain_id3) {
                           H_Index = BegIndex; H_Res = BegRes;
                           name3.Copy(name1.Text());
                      } else if (seq_id2 == seq_id3 && chain_id2 == chain_id3) {
                           H_Index = EndIndex; H_Res = EndRes;
                           name3.Copy(name2.Text());
                      }
                      Insert_a_Hydbnd(atom1.Text(), loc1.Text(), name1.Text(), BegIndex, BegRes,
                             atom3.Text(), loc3.Text(), name3.Text(), H_Index, H_Res, atom2.Text(),
                             loc2.Text(), name2.Text(), EndIndex, EndRes, symm1.Text(), symm2.Text());
                 }
            } else if (!strcasecmp(type.Text(), "link") || !strcasecmp(type.Text(), "covale") ||
                       !strcasecmp(type.Text(), "metalc")) {
                 if (end) Insert_a_Link(mol_index, atom1.Text(), loc1.Text(), name1.Text(), BegIndex, BegRes, atom2.Text(),
                       loc2.Text(), name2.Text(), EndIndex, EndRes, symm1.Text(), symm2.Text(), comment.Text(),
                       type.Text(), dist.Text(), bondtype.Text());
            } else if (!strcasecmp(type.Text(), "modres")) {
/*
                 get_value(pdb_id, Table, pdb_index, irow);
                 get_value(standard, Table, standard_index, irow);
                 Insert_a_Modres(pdb_id.Text(), name1.Text(), BegIndex, BegRes, standard.Text(), comment.Text());
*/
            } else if (!strcasecmp(type.Text(), "sltbrg") || !strcasecmp(type.Text(), "saltbr")) {
                 if (end) Insert_a_Sltbrg(atom1.Text(), loc1.Text(), name1.Text(), BegIndex, BegRes, atom2.Text(),
                        loc2.Text(), name2.Text(), EndIndex, EndRes, symm1.Text(), symm2.Text());
            } else if (!strcasecmp(type.Text(), "disulf")) {
                 if (end) {
                      Insert_a_SSbond(atom1.Text(), loc1.Text(), name1.Text(), BegIndex, BegRes, atom2.Text(),
                           loc2.Text(), name2.Text(), EndIndex, EndRes, symm1.Text(), symm2.Text(), dist.Text());
                      _exist_ssbond = 1;
                 }
            } else if (!strcasecmp(type.Text(), "mismat")) {
                 if (end) Insert_a_Mismat(name1.Text(), BegIndex, BegRes, name2.Text(), EndIndex, EndRes);
            } else if (!strcasecmp(type.Text(), "covale_base")) {
                 if (end) Insert_a_Modifb(name.Text(), atom1.Text(), name1.Text(), BegIndex, BegRes, atom2.Text(),
                                name2.Text(), EndIndex, EndRes, Replaced_Atom.Text());
            } else if (!strcasecmp(type.Text(), "covale_sugar")) {
                 if (end) {
                      get_value(sugar, Table, sugar_index, irow);
                      Insert_a_Modifs(name.Text(), atom1.Text(), name1.Text(), BegIndex, BegRes, atom2.Text(),
                                name2.Text(), EndIndex, EndRes, Replaced_Atom.Text(), sugar.Text());
                 }
            } else if (!strcasecmp(type.Text(), "covale_phosphate")) {
                 if (end) Insert_a_Modifp(name.Text(), atom1.Text(), name1.Text(), BegIndex, BegRes, atom2.Text(),
                                name2.Text(), EndIndex, EndRes, Replaced_Atom.Text());
            }
       }
}

void _MAXIT::cifndb_process_modres()
{
       ISTable *Table = _CifObj->GetTablePtr(_StructureId, "pdbx_struct_mod_residue");
       if (is_empty_table(Table)) return;

       char pdb_code[10];
       memset(pdb_code, 0, 10);
       ISTable *Table1 = _CifObj->GetTablePtr(_StructureId, "database_2");
       if (!is_empty_table(Table1)) get_pdb_id(Table1, pdb_code);

       int rowNo = Table->GetNumRows();
       int standard_index = Table->GetColumnIndex("parent_comp_id");
       int label_comp_id = Table->GetColumnIndex("label_comp_id");
       int label_asym_id = Table->GetColumnIndex("label_asym_id");
       int label_seq_id = Table->GetColumnIndex("label_seq_id");
       int label_ins_code = Table->GetColumnIndex("PDB_ins_code");
       int auth_comp_id = Table->GetColumnIndex("auth_comp_id");
       int auth_asym_id = Table->GetColumnIndex("auth_asym_id");
       int auth_seq_id = Table->GetColumnIndex("auth_seq_id");
       int details_index = Table->GetColumnIndex("details");

       int beg, BegIndex, BegRes;
       CifString name1, standard, details;
       for (int irow = 0; irow < rowNo; irow++) {
            beg = get_residue_index(_molecules[0], Table, irow, label_comp_id, label_asym_id,
                       label_seq_id, auth_comp_id, auth_asym_id, auth_seq_id, label_ins_code,
                       BegIndex, BegRes, name1);
            if (!beg) continue;

            get_value(standard, Table, standard_index, irow);
            get_value(details, Table, details_index, irow);
            Insert_a_Modres(pdb_code, name1.Text(), BegIndex, BegRes, standard.Text(), details.Text());
       }
}

void _MAXIT::put_into_card(const char *value, const int tokenid, const int fieldno,
               const char *refine_code, const int refine_field)
{
       if (tokenid < 0 || tokenid >= MxTokens || fieldno < 0 ||
           fieldno >= NdbToken[tokenid].NumField) return;

       One_Card_Info_Format *pCard = _CardInfo[tokenid].Tail();
       if (refine_field >= 0) {
            while (pCard) {
                 if (!strcmp(refine_code, pCard->Field(refine_field))) break;
                 pCard = pCard->Next();
            }
       }
       if (pCard == NULL || strcmp(pCard->Field(fieldno), ""))
            pCard = _CardInfo[tokenid].add_a_new_card(NdbToken[tokenid], 0);
       pCard->AddElementToField(value, fieldno);
       if (refine_field >= 0) pCard->AddElementToField(refine_code, refine_field);
}

void _MAXIT::ndb_processing_em_software()
{
       int idx;
       CifString string, string1;
       ISTable *Table = NULL;

       Table = _CifObj->GetTablePtr(_StructureId, "em_3d_fitting");
       if (is_empty_table(Table)) return;

       idx = Table->GetColumnIndex("software_name");
       if (idx < 0) return;

       Table->GetCell(string, idx, 0);
       if (string == "" || string == "." || string == "?") return;

       Table = _CifObj->GetTablePtr(_StructureId, "em_2d_projection_selection");
       if (!Table) {
            Table = new ISTable("em_2d_projection_selection");
            Table->AddColumn("entry_id");
            Table->AddColumn("software_name");
            Table->AddColumn("num_particles");
            Table->AddRow();
            _CifObj->WriteTable(Table, _StructureId, "em_2d_projection_selection");
       }
       idx = Table->GetColumnIndex("software_name");
       Table->GetCell(string1, idx, 0);
       if (string1 == "" || string1 == "." || string1 == "?")
            Table->UpdateCell(string, idx, 0);
}

void _MAXIT::ndb_processing_struct_category()
{
       ISTable *Table = _CifObj->GetTablePtr(_StructureId, "struct");
       if (is_empty_table(Table)) return;

       int ndb_descriptor = Table->GetColumnIndex("ndb_descriptor");
       int ndb_title_text = Table->GetColumnIndex("ndb_title_text");
       int title_index    = Table->GetColumnIndex("title");
       
       if (ndb_title_text >= 0 && title_index >= 0 && ndb_descriptor < 0) {
            ISTable *Table1 = new ISTable("struct");
            Table1->AddColumn("entry_id");
            Table1->AddColumn("title");
            Table1->AddColumn("ndb_descriptor");
            Table1->AddColumn("rcsb_model_details");
            Table1->AddColumn("ndb_details");
            int detail_index = Table->GetColumnIndex("pdbx_CASP_flag");
            if (detail_index >= 0) Table1->AddColumn("pdbx_CASP_flag");
            Table1->AddRow();
            CifString string(_StructureId);
            Table1->UpdateCell(string, 0, 0);
            Table->GetCell(string, ndb_title_text, 0);
            Table1->UpdateCell(string, 1, 0);
            Table->GetCell(string, title_index, 0);
            Table1->UpdateCell(string, 2, 0);
            detail_index = Table->GetColumnIndex("rcsb_model_details");
            if (detail_index >= 0) {
                 Table->GetCell(string, detail_index, 0);
                 Table1->UpdateCell(string, 3, 0);
            }
            detail_index = Table->GetColumnIndex("ndb_details");
            if (detail_index >= 0) {
                 Table->GetCell(string, detail_index, 0);
                 Table1->UpdateCell(string, 4, 0);
            }
            detail_index = Table->GetColumnIndex("pdbx_CASP_flag");
            if (detail_index >= 0) {
                 Table->GetCell(string, detail_index, 0);
                 Table1->UpdateCell(string, 5, 0);
            }
            _CifObj->DeleteTable(_StructureId, "struct");
            _CifObj->WriteTable(Table1, _StructureId, "struct");
       }
}

void _MAXIT::ndb_processing_struct_sheet_hbond()
{
       ISTable *Table = _CifObj->GetTablePtr(_StructureId, "struct_sheet_hbond");
       if (is_empty_table(Table)) return;

       ISTable *Table1 = new ISTable("rcsb_struct_sheet_hbond");

       Table1->AddColumn("sheet_id");
       Table1->AddColumn("range_id_2");
       Table1->AddColumn("range_1_label_atom_id");
       Table1->AddColumn("range_1_label_comp_id");
       Table1->AddColumn("range_1_label_asym_id");
       Table1->AddColumn("range_1_label_seq_id");
       Table1->AddColumn("range_1_PDB_ins_code");
       Table1->AddColumn("range_1_auth_comp_id");
       Table1->AddColumn("range_1_auth_asym_id");
       Table1->AddColumn("range_1_auth_seq_id");
       Table1->AddColumn("range_2_label_atom_id");
       Table1->AddColumn("range_2_label_comp_id");
       Table1->AddColumn("range_2_label_asym_id");
       Table1->AddColumn("range_2_label_seq_id");
       Table1->AddColumn("range_2_PDB_ins_code");
       Table1->AddColumn("range_2_auth_comp_id");
       Table1->AddColumn("range_2_auth_asym_id");
       Table1->AddColumn("range_2_auth_seq_id");

       int rowNo = Table->GetNumRows();
       int sheet_id = Table->GetColumnIndex("sheet_id");
       int range_id_1 = Table->GetColumnIndex("range_id_1");
       int beg_label_atom_id = Table->GetColumnIndex("range_1_beg_label_atom_id");
       int beg_label_comp_id = Table->GetColumnIndex("ndb_range_1_beg_label_comp_id");
       int beg_label_asym_id = Table->GetColumnIndex("ndb_range_1_beg_label_asym_id");
       int beg_label_seq_id = Table->GetColumnIndex("range_1_beg_label_seq_id");
       int beg_label_ins_code = Table->GetColumnIndex("ndb_range_1_beg_label_ins_code");
       int beg_auth_comp_id = Table->GetColumnIndex("ndb_range_1_beg_auth_comp_id");
       int beg_auth_asym_id = Table->GetColumnIndex("ndb_range_1_beg_auth_asym_id");
       int beg_auth_seq_id = Table->GetColumnIndex("range_1_beg_auth_seq_id");
       int end_label_atom_id = Table->GetColumnIndex("range_1_end_label_atom_id");
       int end_label_comp_id = Table->GetColumnIndex("ndb_range_1_end_label_comp_id");
       int end_label_asym_id = Table->GetColumnIndex("ndb_range_1_end_label_asym_id");
       int end_label_seq_id = Table->GetColumnIndex("range_1_end_label_seq_id");
       int end_label_ins_code = Table->GetColumnIndex("ndb_range_1_end_label_ins_code");
       int end_auth_comp_id = Table->GetColumnIndex("ndb_range_1_end_auth_comp_id");
       int end_auth_asym_id = Table->GetColumnIndex("ndb_range_1_end_auth_asym_id");
       int end_auth_seq_id = Table->GetColumnIndex("range_1_end_auth_seq_id");
       CifString string;
       for (int i = 0; i < rowNo; i++) {
            Table1->AddRow();
            if (sheet_id >= 0) {
                 Table->GetCell(string, sheet_id, i);
                 Table1->UpdateCell(string, 0, i);
            }
            if (range_id_1 >= 0) {
                 Table->GetCell(string, range_id_1, i);
                 Table1->UpdateCell(string, 1, i);
            }
            if (end_label_atom_id >= 0) {
                 Table->GetCell(string, end_label_atom_id, i);
                 Table1->UpdateCell(string, 2, i);
            }
            if (end_label_comp_id >= 0) {
                 Table->GetCell(string, end_label_comp_id, i);
                 Table1->UpdateCell(string, 3, i);
            }
            if (end_label_asym_id >= 0) {
                 Table->GetCell(string, end_label_asym_id, i);
                 Table1->UpdateCell(string, 4, i);
            }
            if (end_label_seq_id >= 0) {
                 Table->GetCell(string, end_label_seq_id, i);
                 Table1->UpdateCell(string, 5, i);
            }
            if (end_label_ins_code >= 0) {
                 Table->GetCell(string, end_label_ins_code, i);
                 Table1->UpdateCell(string, 6, i);
            }
            if (end_auth_comp_id >= 0) {
                 Table->GetCell(string, end_auth_comp_id, i);
                 Table1->UpdateCell(string, 7, i);
            }
            if (end_auth_asym_id >= 0) {
                 Table->GetCell(string, end_auth_asym_id, i);
                 Table1->UpdateCell(string, 8, i);
            }
            if (end_auth_seq_id >= 0) {
                 Table->GetCell(string, end_auth_seq_id, i);
                 Table1->UpdateCell(string, 9, i);
            }
            if (beg_label_atom_id >= 0) {
                 Table->GetCell(string, beg_label_atom_id, i);
                 Table1->UpdateCell(string, 10, i);
            }
            if (beg_label_comp_id >= 0) {
                 Table->GetCell(string, beg_label_comp_id, i);
                 Table1->UpdateCell(string, 11, i);
            }
            if (beg_label_asym_id >= 0) {
                 Table->GetCell(string, beg_label_asym_id, i);
                 Table1->UpdateCell(string, 12, i);
            }
            if (beg_label_seq_id >= 0) {
                 Table->GetCell(string, beg_label_seq_id, i);
                 Table1->UpdateCell(string, 13, i);
            }
            if (beg_label_ins_code >= 0) {
                 Table->GetCell(string, beg_label_ins_code, i);
                 Table1->UpdateCell(string, 14, i);
            }
            if (beg_auth_comp_id >= 0) {
                 Table->GetCell(string, beg_auth_comp_id, i);
                 Table1->UpdateCell(string, 15, i);
            }
            if (beg_auth_asym_id >= 0) {
                 Table->GetCell(string, beg_auth_asym_id, i);
                 Table1->UpdateCell(string, 16, i);
            }
            if (beg_auth_seq_id >= 0) {
                 Table->GetCell(string, beg_auth_seq_id, i);
                 Table1->UpdateCell(string, 17, i);
            }
       }
       _CifObj->DeleteTable(_StructureId, "struct_sheet_hbond");
       _CifObj->WriteTable(Table1, _StructureId, "rcsb_struct_sheet_hbond");
}

void _MAXIT::ndb_processing_ndb_database_PDB_obs_spr_category()
{
       ISTable *Table = _CifObj->GetTablePtr(_StructureId, "ndb_database_PDB_obs_spr");
       if (is_empty_table(Table)) return;

       int id, nrow, i;
       CifString value;

       id = Table->GetColumnIndex("id");
       if (id >= 0) {
            nrow = Table->GetNumRows();
            for (i = 0; i < nrow; i++) {
                 value.Clear();
                 Table->GetCell(value, id, i);
                 if (strcmp(value.Text(), "SPRSDE") &&
                     strcmp(value.Text(), "OBSLTE")) {
                      value.Copy("SPRSDE");
                      Table->UpdateCell(value, id, i);
                 }
            }
       }
       _CifObj->WriteTable(Table, _StructureId, "ndb_database_PDB_obs_spr");
}

void _MAXIT::ndb_processing_struct_ref_seq_and_diff()
{
       int i, rowNo, auth_beg_idx, auth_end_idx, beg_idx, end_idx;
       CifString string;
       ISTable *Table = NULL;

       Table = _CifObj->GetTablePtr(_StructureId, "struct_ref_seq");
       if (!is_empty_table(Table)) {
            rowNo = Table->GetNumRows();
            auth_beg_idx = Table->GetColumnIndex("ndb_auth_seq_align_beg");
            auth_end_idx = Table->GetColumnIndex("ndb_auth_seq_align_end");
            beg_idx = Table->GetColumnIndex("seq_align_beg");
            end_idx = Table->GetColumnIndex("seq_align_end");
            if (auth_beg_idx >= 0 && auth_end_idx >= 0 && beg_idx >= 0 &&
                end_idx >= 0) {
                 for (i = 0; i < rowNo; i++) {
                      Table->GetCell(string, auth_beg_idx, i);
                      if (string != "?" && string != "" && string != ".")
                           Table->UpdateCell(string, beg_idx, i);
                      Table->GetCell(string, auth_end_idx, i);
                      if (string != "?" && string != "" && string != ".")
                           Table->UpdateCell(string, end_idx, i);
                 }
            }
       }
       Table = _CifObj->GetTablePtr(_StructureId, "struct_ref_seq_dif");
       if (!is_empty_table(Table)) {
            rowNo = Table->GetNumRows();
            auth_beg_idx = Table->GetColumnIndex("ndb_auth_seq_num");
            beg_idx = Table->GetColumnIndex("seq_num");
            int idx = Table->GetColumnIndex("details");
            if (auth_beg_idx >= 0 && beg_idx >= 0) {
                 for (i = 0; i < rowNo; i++) {
                      Table->GetCell(string, idx, i);
                      string.ToUpper();
                      if (string == "DELETION") continue;
                      Table->GetCell(string, auth_beg_idx, i);
                      if (string != "?" && string != "" && string != ".")
                           Table->UpdateCell(string, beg_idx, i);
                 }
            }
       }
}

void _MAXIT::post_processing_from_cif_to_ndb()
{
       int Auth_Token;
       One_Card_Info_Format *pCard = NULL;

       Auth_Token = find_jrnl_id("AUTH");

       if (_JrnlCardInfo[Auth_Token].Head() && _CardInfo[AUTHOR_TOKEN-1].Head() == NULL && 
           strcmp(_JrnlCardInfo[Auth_Token].Head()->Field(4), "")) {
            pCard = _CardInfo[AUTHOR_TOKEN-1].add_a_new_card(NdbToken[AUTHOR_TOKEN-1], 0);
            pCard->AddElementToField(_JrnlCardInfo[Auth_Token].Head()->Field(4), 2);
       }

       get_refn_card();
       _JrnlCardInfo[Auth_Token].reorder_cards(1, INTEGER_COMPARE, ASCENT);
       Auth_Token = find_jrnl_id("TITL");
       _JrnlCardInfo[Auth_Token].reorder_cards(1, INTEGER_COMPARE, ASCENT);
       Auth_Token = find_jrnl_id("EDIT");
       _JrnlCardInfo[Auth_Token].reorder_cards(1, INTEGER_COMPARE, ASCENT);
       Auth_Token = find_jrnl_id("REF");
       _JrnlCardInfo[Auth_Token].reorder_cards(1, INTEGER_COMPARE, ASCENT);
       Auth_Token = find_jrnl_id("PUBL");
       _JrnlCardInfo[Auth_Token].reorder_cards(1, INTEGER_COMPARE, ASCENT);
       Auth_Token = find_jrnl_id("REFN");
       _JrnlCardInfo[Auth_Token].reorder_cards(1, INTEGER_COMPARE, ASCENT);

       Auth_Token = find_card_id("TITLE");
       pCard = _CardInfo[Auth_Token].Head();
       while (pCard) {
            pCard->CleanStringDash(2);
            pCard = pCard->Next();
       }

       if ((_CardInfo[STATUS_TOKEN-1].Head() == NULL ||
           !strcmp(_CardInfo[STATUS_TOKEN-1].Head()->Field(6), "")) &&
            _CardInfo[HEADER_TOKEN-1].Head() &&
            strcmp(_CardInfo[HEADER_TOKEN-1].Head()->Field(2), "")) {
             pCard = _CardInfo[STATUS_TOKEN-1].Head();
             if (pCard == NULL) pCard = _CardInfo[STATUS_TOKEN-1].
                         add_a_new_card(NdbToken[STATUS_TOKEN-1], 0);
             pCard->AddElementToField(_CardInfo[HEADER_TOKEN-1].Head()->Field(2), 6);
       }

       pCard = _CardInfo[PROCES_TOKEN-1].Head();
       if (pCard == NULL) pCard = _CardInfo[PROCES_TOKEN-1].
                       add_a_new_card(NdbToken[PROCES_TOKEN-1], 0);
       if (!strcmp(pCard->Field(5), "")) pCard->AddElementToField("RCSB", 5);
 
       _CardInfo[RADIAT_TOKEN-1].reorder_cards(1, INTEGER_COMPARE, ASCENT);

       pCard = _CardInfo[REFMET_TOKEN-1].Head();
       if (pCard && (!strncasecmp(pCard->Field(3), "CNS", 3) ||
           !strncasecmp(pCard->Field(3), "X-PLOR", 6))) {
            pCard = _CardInfo[RFACTR_TOKEN-1].Head();
            if (pCard) {
                 if (!strcmp(pCard->Field(10), "") && strcmp(pCard->Field(7), ""))
                      pCard->AddElementToField(pCard->Field(7), 10);
                 else if (strcmp(pCard->Field(10), "") && !strcmp(pCard->Field(7), ""))
                      pCard->AddElementToField(pCard->Field(10), 7);
            }
       }
       delete_null_cards();
}

void _MAXIT::update_encapsulated_cif()
{
       int count = 0;
       One_Card_Info_Format *pCard = NULL;
       if (_file_status == FILE_STATUS_MMCIF_PLUS_MOL) {
            pCard = _CardInfo[COMPND_TOKEN-1].Head();
            while (pCard) {
                 count++;
                 pCard = pCard->Next();
            }
            pCard = _CardInfo[SOURCE_TOKEN-1].Head();
            if (pCard && count > 2)
                 pdb_to_ndb_proc_compnd_and_source(0);
       }
}

void _MAXIT::ndb_processing_requests() 
{
       int i, j, k, ph_idx, mtd_idx, tmp_idx, det_idx, rowNo;
       char buffer[1000];
       ISTable *Table = NULL, *Table1 = NULL;
       CifString pH, method, temp, rcsb_details;

       if (_experiment_type & EXPERIMENT_TYPE_NMR) {
            Table = _CifObj->GetTablePtr(_StructureId, "exptl");
            if (!is_empty_table(Table)) {
                 mtd_idx = Table->GetColumnIndex("method");
                 rowNo = Table->GetNumRows();
                 for (i = 0; i < rowNo; i++) {
                      method.Clear();
                      Table->GetCell(method, mtd_idx, i);
                      method.ToUpper();
                      if (strstr(method.Text(), "NMR") && method != "SOLID-STATE NMR") {
#if 0
                 if (/* !strstr(method.Text(), "REGULARIZED MEAN STRUCTURE") && */
                     !strstr(method.Text(), "SOLID-STATE NMR")) {
#endif
/*
                      if (_num_mol > 1)
                           // sprintf(buffer, "NMR, %d STRUCTURES", _num_mol);
                           strcpy(buffer, "SOLUTION NMR");
                      else if (_num_mol == 1) {
                           strcpy(buffer, "SOLUTION NMR");
                           Table1 = _CifObj->GetTablePtr(_StructureId, "rcsb_nmr_representative");
                           if (!is_empty_table(Table1)) {
                                tmp_idx = Table1->GetColumnIndex("selection_criteria");
                                temp.Clear();
                                Table1->GetCell(temp, tmp_idx, 0);
                                if (!strcasecmp(temp.Text(), "minimized average structure")) {
                                     strcpy(buffer, "NMR, minimized average structure");
                                }
                           }
                      } else strcpy(buffer, "SOLUTION NMR"); */
                           method.Copy("SOLUTION NMR");
                           Table->UpdateCell(method, mtd_idx, i);
                      }
                 }
                 _CifObj->WriteTable(Table, _StructureId, "exptl");
            }
            
            Table1 = _CifObj->GetTablePtr(_StructureId, "rcsb_nmr_representative");
            if (!is_empty_table(Table1)) {
                 tmp_idx = Table1->GetColumnIndex("selection_criteria");
                 temp.Clear();
                 Table1->GetCell(temp, tmp_idx, 0);
                 if (!strcasecmp(temp.Text(), "minimized average structure")) {
                      Table = _CifObj->GetTablePtr(_StructureId, "struct");
                      if (!is_empty_table(Table)) {
                           mtd_idx = Table->GetColumnIndex("pdbx_model_type_details");
                           if (mtd_idx < 0) {
                                Table->AddColumn("pdbx_model_type_details");
                                mtd_idx = Table->GetColumnIndex("pdbx_model_type_details");
                                rowNo = Table->GetNumRows();
                                ReVarCifArray<CifString> colData;
                                colData.Clear();
                                temp.Copy("?");
                                for (i = 0; i < rowNo; i++) colData.Add(temp);
                                Table->FillColumn(colData, mtd_idx);
                           }
                           temp.Copy("minimized average");
                           Table->UpdateCell(temp, mtd_idx, 0);
                           _CifObj->WriteTable(Table, _StructureId, "struct");
                      }
                 }
            }

            Table = _CifObj->GetTablePtr(_StructureId, "rcsb_nmr_software");
            if (!is_empty_table(Table)) {
                 mtd_idx = Table->GetColumnIndex("classification");
                 rowNo = Table->GetNumRows();
                 tmp_idx = -1; det_idx = -1;
                 for (i = 0; i < rowNo; i++) {
                      Table->GetCell(method, mtd_idx, i);
                      if (!strcasecmp(method.Text(), "structure solution")) tmp_idx = i;
                      if (!strcasecmp(method.Text(), "refinement")) det_idx = i;
                 }
                 if (tmp_idx >= 0 && det_idx < 0) {
                      ph_idx = Table->GetNumColumns();
                      Table->AddRow();
                      for (i = 0; i < ph_idx; i++) {
                           if (i == mtd_idx)
                                method.Copy("refinement");
                           else Table->GetCell(method, i, tmp_idx);
                           Table->UpdateCell(method, i, rowNo);
                      }
                      _CifObj->WriteTable(Table, _StructureId, "rcsb_nmr_software");
                 }
            }
       }
/*
       for (i = 0; i < N_R_MERGE; i++) {
            Table = _CifObj->GetTablePtr(_StructureId, _r_merge[i].catetory);
            if (is_empty_table(Table)) continue;
            mtd_idx = Table->GetColumnIndex(_r_merge[i].item);
            if (mtd_idx < 0) continue;
            rowNo = Table->GetNumRows();
            for (j = 0; j < rowNo; j++) {
                 Table->GetCell(method, mtd_idx, j);
                 if (method == "" || method == "?" || method == ".") continue;
                 if (atof(method.Text()) > 1.00) {
                      sprintf(buffer, "%.7f", atof(method.Text()) / 100.0);
                      for (k = strlen(buffer) - 1; k >= 0; k--) {
                           if (buffer[k] == '0') buffer[k] = '\0';
                           else break;
                      }
                      method.Copy(buffer);
                      Table->UpdateCell(method, mtd_idx, j);
                 }
            }
            _CifObj->WriteTable(Table, _StructureId, _r_merge[i].catetory);
       }
*/
       for (i = 0; i < N_SWITCH_CATEGORY; i++) {
            Table = _CifObj->GetTablePtr(_StructureId, _switch_category[i].category);
            if (is_empty_table(Table)) continue;
            mtd_idx = Table->GetColumnIndex(_switch_category[i].high_item);
            tmp_idx = Table->GetColumnIndex(_switch_category[i].low_item);
            if (mtd_idx < 0 || tmp_idx < 0) continue;
            Table->GetCell(method, mtd_idx, 0);
            Table->GetCell(temp, tmp_idx, 0);
            if (method == "" || method == "?" || method == "." ||
                temp == "" || temp == "?" || temp == ".") continue;
            if (atof(method.Text()) > atof(temp.Text())) {
                 Table->UpdateCell(temp, mtd_idx, 0);
                 Table->UpdateCell(method, tmp_idx, 0);
                 _CifObj->WriteTable(Table, _StructureId, _switch_category[i].category);
            }
       }

       remove_redundant_keywords(_CifObj, _StructureId);
#if 0
       Table = _CifObj->GetTablePtr(_StructureId, "struct_keywords");
       if (!is_empty_table(Table)) {
            ph_idx = Table->GetColumnIndex("ndb_keywords");
            mtd_idx = Table->GetColumnIndex("text");
            pH.Clear();
            Table->GetCell(pH, ph_idx, 0);
            if (pH == "" || pH == "?" || pH == ".") pH.Clear();
            method.Clear();
            Table->GetCell(method, mtd_idx, 0);
            if (method == "" || method == "?" || method == ".") method.Clear();
/*
            if (pH != "" && method == "") {
                 Table->UpdateCell(pH, mtd_idx, 0);
                 _CifObj->WriteTable(Table, _StructureId, "struct_keywords");
            }
*/
            if (pH != "") {
                 int len = strlen(pH.Text()) + 1;
                 char *header = new char[len];
                 memset(header, 0, len);
                 strcpy(header, pH.Text());
                 ndb_clean_string(header);
                 ndb_set_string_to_upper(header);
                 len = strlen(method.Text()) + 1;
                 char *keywords = new char[len];
                 memset(keywords, 0, len);
                 strcpy(keywords, method.Text());
                 ndb_clean_string(keywords);
                 ndb_set_string_to_upper(keywords);
                 if (!strstr(keywords, header)) {
                      if (method != "") method += ", ";
                      method += pH;
                      if (strstr(header, "/")) {
                           method += " ";
                           method += "COMPLEX";
                      }
                 }
                 delete [] header;
                 delete [] keywords;
                 len = strlen(method.Text()) + 1;
                 keywords = new char[len];
                 memset(keywords, 0, len);
                 header = new char[len];
                 memset(header, 0, len);
                 strcpy(keywords, method.Text());
                 ndb_clean_string(keywords);
                 method.Clear();
                 TblIndexObj UniqueKeywords;
                 UniqueKeywords.Clear();
                 char *p = keywords;
                 while (p) {
                      p = get_words_separate_by_delimits(p, header, ", ");
                      pH.Copy(header);
                      pH.ToUpper();
                      if (UniqueKeywords.Seek(pH) < 0) {
                           if (method != "") method += ", ";
                           method += header;
                           UniqueKeywords.Add(pH);
                      }
                 }
                 Table->UpdateCell(method, mtd_idx, 0);
                 _CifObj->WriteTable(Table, _StructureId, "struct_keywords");
                 delete [] header;
                 delete [] keywords;
            }
       }
#endif
       Table = _CifObj->GetTablePtr(_StructureId, "exptl_crystal_grow");
       if (!is_empty_table(Table)) {
            rowNo = Table->GetNumRows();
            ph_idx = Table->GetColumnIndex("pH");
            mtd_idx = Table->GetColumnIndex("method");
            tmp_idx = Table->GetColumnIndex("temp");
            det_idx = Table->GetColumnIndex("rcsb_details");
            for (i = 0; i < rowNo; i++) {
                 rcsb_details.Clear();
                 Table->GetCell(rcsb_details, det_idx, i);
                 if (rcsb_details == "" || rcsb_details == "?" || rcsb_details == ".")
                      rcsb_details.Clear();

                 pH.Clear();
                 Table->GetCell(pH, ph_idx, i);
                 if (pH == "" || pH == "?" || pH == ".") pH.Clear();
                 if (pH != "" && !strstr(rcsb_details.Text(), pH.Text())) {
                      method.Copy(rcsb_details);
                      method.ToUpper();
                      if (!strstr(method.Text(), "PH ")) {
                           if (rcsb_details != "") rcsb_details += ", ";
                           rcsb_details += "pH ";
                           rcsb_details += pH;
                      }
                 }
                 method.Clear();
                 Table->GetCell(method, mtd_idx, i);
                 if (method == "" || method == "?" || method == ".") method.Clear();
                 if (method != "" && !strstr(rcsb_details.Text(), method.Text())) {
                      if (rcsb_details != "") rcsb_details += ", ";
                      rcsb_details += method;
                 }
                 temp.Clear();
                 Table->GetCell(temp, tmp_idx, i);
                 if (temp == "" || temp == "?" || temp == ".") temp.Clear();
                 if (temp != "" && !strstr(rcsb_details.Text(), temp.Text())) {
                      if (rcsb_details != "") rcsb_details += ", temperature ";
                      rcsb_details += temp;
                      rcsb_details += "K";
                 }
                 Table->UpdateCell(rcsb_details, det_idx, i);
            }
            _CifObj->WriteTable(Table, _StructureId, "exptl_crystal_grow");
       }

       Table = _CifObj->GetTablePtr(_StructureId, "ndb_xplor_file");
       if (!is_empty_table(Table)) {
            rowNo = Table->GetNumRows();
            mtd_idx = Table->GetColumnIndex("param_file");
            tmp_idx = Table->GetColumnIndex("topol_file");
            for (i = 0; i < rowNo; i++) {
                 Table->GetCell(method, mtd_idx, i);
                 if (method != "" && method != "?" && method != ".") {
                      strcpy(buffer, method.Text());
                      for (j = strlen(buffer) - 1; j >= 0; j--) {
                           if (buffer[j] == '/') break;
                      }
                      method.Copy(&buffer[j + 1]);
                      Table->UpdateCell(method, mtd_idx, i);
                 }
                 Table->GetCell(method, tmp_idx, i);
                 if (method != "" && method != "?" && method != ".") {
                      strcpy(buffer, method.Text());
                      for (j = strlen(buffer) - 1; j >= 0; j--) {
                           if (buffer[j] == '/') break;
                      }
                      method.Copy(&buffer[j + 1]);
                      Table->UpdateCell(method, tmp_idx, i);
                 }
            }
            _CifObj->WriteTable(Table, _StructureId, "ndb_xplor_file");
       }

       Table = _CifObj->GetTablePtr(_StructureId, "database_PDB_remark");
       if (!is_empty_table(Table)) {
            mtd_idx = Table->GetColumnIndex("id");
            if (mtd_idx >= 0) {
                 rowNo = Table->GetNumRows();
                 for (i = 0; i < rowNo; i++) {
                      Table->GetCell(method, mtd_idx, i);
                      if (method != "" && method != "?" && method != ".") {
                           strcpy(buffer, method.Text());
                           k = strlen(buffer);
                           for (j = 0; j < k; j++) {
                                if (!isdigit(buffer[j])) break;
                           }
                           if (j < k) {
                                Table->DeleteRow(i);
                                // Table->RemoveRow(i);
                                // i--;
                           }
                      }
                 }
                 Table->CompressTable();
                 _CifObj->WriteTable(Table, _StructureId, "database_PDB_remark");
            }
       }
}

void _MAXIT::generate_self_reference()
{
       ISTable *Table = _CifObj->GetTablePtr(_StructureId, "entity_poly");
       if (is_empty_table(Table)) return;

       int ref_index = Table->GetColumnIndex("rcsb_build_self_reference");
       if (ref_index < 0) return;

       char pdb_code[10];
       memset(pdb_code, 0, 10);
       ISTable *Table1 = _CifObj->GetTablePtr(_StructureId, "database_2");
       if (!is_empty_table(Table1)) get_pdb_id(Table1, pdb_code);

       int rowNo = Table->GetNumRows();
       int id_index = Table->GetColumnIndex("ndb_chain_id");
       CifString cs;
       ReVarCifArray<CifString> data;
       for (int i = 0; i < rowNo; i++) {
            get_value(cs, Table, ref_index, i);
            if (cs != "Y" && cs != "y") continue;
            get_value(cs, Table, id_index, i);
            if (cs == "") continue;
            get_wordlist_from_string_separated_by_delimits(data, cs.Text(), ",");
            cs.Clear();
            _CHAIN *chain = _molecules[_rep_mol]->GetFirstChain();
            while (chain) {
                 if (data[0] == chain->PDB_ChainID()) {
                     cs += chain->entity_id();
                     break;
                 }
                 chain = _molecules[_rep_mol]->GetNextChain();
            }
            if (cs == "") continue;
            generate_dbref_and_seqadv(cs, pdb_code);
       }
}

void _MAXIT::generate_dbref_and_seqadv(CifString &entity_id, const char *pdb_id)
{
       ISTable *Table = _CifObj->GetTablePtr(_StructureId, "struct_ref");
       if (is_empty_table(Table)) {
            Table = new ISTable("struct_ref");
            Table->AddColumn("id");
            Table->AddColumn("db_name");
            Table->AddColumn("db_code");
            Table->AddColumn("rcsb_db_accession");
            Table->AddColumn("entity_id");
       }
       int rowNo = Table->GetNumRows();
       int i;
       int id_idx = Table->GetColumnIndex("id");
       int name_idx = Table->GetColumnIndex("db_name");
       int entity_idx = Table->GetColumnIndex("entity_id");
       int code_idx = Table->GetColumnIndex("db_code");
       int acce_idx = Table->GetColumnIndex("rcsb_db_accession");
       CifString ref_id, cs1, cs2;
       for (i = 0; i < rowNo; i++) {
            get_value(cs1, Table, name_idx, i);
            get_value(cs2, Table, entity_idx, i);
            if (cs1 == "PDB" && cs2 == entity_id) {
                 cs1.Copy(pdb_id);
                 Table->UpdateCell(cs1, code_idx, i);
                 Table->UpdateCell(cs1, acce_idx, i);
                 get_value(ref_id, Table, id_idx, i);
                 break;
            }
       }
       char buffer[10];
       if (i == rowNo) {
            Table->AddRow();
            sprintf(buffer, "%d", i + 1);
            cs1.Copy(buffer);
            ref_id.Copy(buffer);
            Table->UpdateCell(cs1, id_idx, i);
            cs1.Copy("PDB");
            Table->UpdateCell(cs1, name_idx, i);
            cs1.Copy(pdb_id);
            Table->UpdateCell(cs1, code_idx, i);
            Table->UpdateCell(cs1, acce_idx, i);
            Table->UpdateCell(entity_id, entity_idx, i);
       }
       _CifObj->WriteTable(Table, _StructureId, "struct_ref");

       Table = _CifObj->GetTablePtr(_StructureId, "struct_ref_seq");
       if (is_empty_table(Table)) {
            Table = new ISTable("struct_ref_seq");
            Table->AddColumn("align_id");
            Table->AddColumn("ref_id");
            Table->AddColumn("ndb_pdb_id_code");
            Table->AddColumn("ndb_chain_id");
            Table->AddColumn("seq_align_beg");
            Table->AddColumn("ndb_seq_align_beg_ins_code");
            Table->AddColumn("seq_align_end");
            Table->AddColumn("ndb_seq_align_end_ins_code");
            Table->AddColumn("ndb_db_accession");
            Table->AddColumn("db_align_beg");
            Table->AddColumn("db_align_end");
            Table->AddColumn("ndb_auth_seq_align_beg");
            Table->AddColumn("ndb_auth_seq_align_end");
       }

       rowNo = Table->GetNumRows();
       id_idx = Table->GetColumnIndex("align_id");
       int ref_idx = Table->GetColumnIndex("ref_id");
       int pdb_idx = Table->GetColumnIndex("ndb_pdb_id_code");
       int chain_idx = Table->GetColumnIndex("ndb_chain_id");
       int beg_idx = Table->GetColumnIndex("seq_align_beg");
       int beg_ins_idx = Table->GetColumnIndex("ndb_seq_align_beg_ins_code");
       int end_idx = Table->GetColumnIndex("seq_align_end");
       int end_ins_idx = Table->GetColumnIndex("ndb_seq_align_end_ins_code");
       acce_idx = Table->GetColumnIndex("ndb_db_accession");
       int db_beg_idx = Table->GetColumnIndex("db_align_beg");
       int db_end_idx = Table->GetColumnIndex("db_align_end");
       int ndb_beg_idx = Table->GetColumnIndex("ndb_auth_seq_align_beg");
       int ndb_end_idx = Table->GetColumnIndex("ndb_auth_seq_align_end");

       CifString cs3, beg_number, beg_ins, end_number, end_ins;
       _CHAIN *chain = _molecules[_rep_mol]->GetFirstChain();
       while (chain) {
            if (chain->entity_id() == atoi(entity_id.Text())) {
                 beg_number.Clear();
                 beg_ins.Clear();
                 end_number.Clear();
                 end_ins.Clear();
                 for (i = 0; i < chain->SeqLen(); i++) {
                      if (chain->SeqRes(i)->Deleted) continue;
                      if (beg_number == "") {
                           beg_number.Copy(chain->SeqRes(i)->Field[4]);
                           beg_ins.Copy(chain->SeqRes(i)->InsCode);
                      }
                      end_number.Copy(chain->SeqRes(i)->Field[4]);
                      end_ins.Copy(chain->SeqRes(i)->InsCode);
                 }
                 for (i = 0; i < rowNo; i++) {
                      get_value(cs1, Table, ref_idx, i);
                      get_value(cs2, Table, chain_idx, i);
                      get_value(cs3, Table, acce_idx, i);
                      if (cs1 == ref_id && cs2 == chain->PDB_ChainID() && cs3 == pdb_id) {
                           cs1.Copy(pdb_id);
                           Table->UpdateCell(cs1, pdb_idx, i);
                           Table->UpdateCell(beg_number, beg_idx, i);
                           Table->UpdateCell(beg_number, db_beg_idx, i);
                           Table->UpdateCell(beg_number, ndb_beg_idx, i);
                           Table->UpdateCell(beg_ins, beg_ins_idx, i);
                           Table->UpdateCell(end_number, end_idx, i);
                           Table->UpdateCell(end_number, db_end_idx, i);
                           Table->UpdateCell(end_number, ndb_end_idx, i);
                           Table->UpdateCell(end_ins, end_ins_idx, i);
                           break;
                      }
                 }
                 if (i == rowNo) {
                      Table->AddRow();
                      rowNo++;
                      sprintf(buffer, "%d", i + 1);
                      cs1.Copy(buffer);
                      Table->UpdateCell(cs1, id_idx, i);
                      cs1.Copy(chain->PDB_ChainID());
                      Table->UpdateCell(cs1, chain_idx, i);
                      Table->UpdateCell(ref_id, ref_idx, i);
                      cs1.Copy(pdb_id);
                      Table->UpdateCell(cs1, pdb_idx, i);
                      Table->UpdateCell(cs1, acce_idx, i);
                      Table->UpdateCell(beg_number, beg_idx, i);
                      Table->UpdateCell(beg_number, db_beg_idx, i);
                      Table->UpdateCell(beg_number, ndb_beg_idx, i);
                      Table->UpdateCell(beg_ins, beg_ins_idx, i);
                      Table->UpdateCell(end_number, end_idx, i);
                      Table->UpdateCell(end_number, db_end_idx, i);
                      Table->UpdateCell(end_number, ndb_end_idx, i);
                      Table->UpdateCell(end_ins, end_ins_idx, i);
                 }
            }
            chain = _molecules[_rep_mol]->GetNextChain();
       }
       _CifObj->WriteTable(Table, _StructureId, "struct_ref_seq");
}

void _MAXIT::generate_dbref_and_seqadv()
{
       generate_dbref_and_seqadv_from_seqtool_result();

       ISTable *Table = _CifObj->GetTablePtr(_StructureId, "struct_ref_seq");
       if (!is_empty_table(Table)) return;

       Table = _CifObj->GetTablePtr(_StructureId, "struct_ref");
       if (is_empty_table(Table)) return;

       int rowNo = Table->GetNumRows();
       int id_index = Table->GetColumnIndex("id");
       int name_index = Table->GetColumnIndex("db_name");
       int code_index = Table->GetColumnIndex("db_code");
       int accession_index = Table->GetColumnIndex("rcsb_db_accession");
       int entity_index = Table->GetColumnIndex("entity_id");
       int seq_index = Table->GetColumnIndex("rcsb_seq_one_letter_code");
       int start_index = Table->GetColumnIndex("rcsb_align_begin");
       int pdb_start_index = Table->GetColumnIndex("rcsb_pdb_start");
       int pdb_end_index = Table->GetColumnIndex("rcsb_pdb_end");
       if (id_index < 0 || name_index < 0 || accession_index < 0 || entity_index < 0 ||
           seq_index < 0 || start_index < 0) {
            if (id_index < 0)
                 ndb_log_message(NDB_MSG_ERR, "Warning: Missing _struct_ref.id value.\n");
            if (name_index < 0) 
                 ndb_log_message(NDB_MSG_ERR, "Warning: Missing _struct_ref.db_name value.\n");
            if (code_index < 0) 
                 ndb_log_message(NDB_MSG_ERR, "Warning: Missing _struct_ref.db_code value.\n");
            if (accession_index < 0) 
                 ndb_log_message(NDB_MSG_ERR, "Warning: Missing _struct_ref.rcsb_db_accession value.\n");
            if (entity_index < 0) 
                 ndb_log_message(NDB_MSG_ERR, "Warning: Missing _struct_ref.entity_id value.\n");
            if (seq_index < 0) 
                 ndb_log_message(NDB_MSG_ERR, "Warning: Missing _struct_ref.rcsb_seq_one_letter_code value.\n");
            if (start_index < 0) 
                 ndb_log_message(NDB_MSG_ERR, "Warning: Missing _struct_ref.rcsb_align_begin value.\n");
            return;
       }

       _CHAIN *chain = NULL;
       ReVarCifArray<CifString> ID, NAME, ACCESSION, ENTITY, BEGIN_NUMBER, sequence;
       ReVarCifArray<int> PDB_RANGE_BEGIN, PDB_RANGE_END;
       ReVarCifArray<ReVarCifArray<CifString > > SEQUENCES;
       ID.Clear();
       NAME.Clear();
       ACCESSION.Clear();
       ENTITY.Clear();
       BEGIN_NUMBER.Clear();
       SEQUENCES.Clear();
       PDB_RANGE_BEGIN.Clear();
       PDB_RANGE_END.Clear();
       CifString cifstring, s;
       TblIndexObj SelfRef;
       SelfRef.Clear();
       char buff[5];
       int missing_value = 0, is_protein = -1;
       for (int i = 0; i < rowNo; i++) {
            get_value(cifstring, Table, name_index, i);
            cifstring.ToUpper();
            if (cifstring == "PDB") {
                 get_value(cifstring, Table, entity_index, i);
                 if (cifstring != "") SelfRef.Add(cifstring);
                 continue;
            }

            get_value(cifstring, Table, id_index, i);
            if (cifstring == "") {
                 ndb_log_message(NDB_MSG_ERR, "Warning: Missing _struct_ref.id value.\n");
                 missing_value = 1;
            }
            ID.Add(cifstring);

            get_value(cifstring, Table, name_index, i);
            if (cifstring == "") {
                 ndb_log_message(NDB_MSG_ERR, "Warning: Missing _struct_ref.db_name value.\n");
                 missing_value = 1;
            }
            NAME.Add(cifstring);

            get_value(cifstring, Table, code_index, i);
            if (cifstring == "") {
                 ndb_log_message(NDB_MSG_ERR, "Warning: Missing _struct_ref.db_code value.\n");
                 missing_value = 1;
            }

            get_value(cifstring, Table, accession_index, i);
            if (cifstring == "") {
                 ndb_log_message(NDB_MSG_ERR, "Warning: Missing _struct_ref.rcsb_db_accession value.\n");
                 missing_value = 1;
            }
            ACCESSION.Add(cifstring);

            get_value(cifstring, Table, start_index, i);
            if (cifstring == "") {
                 ndb_log_message(NDB_MSG_ERR, "Warning: Missing _struct_ref.rcsb_align_begin value.\n");
                 missing_value = 1;
            }
            BEGIN_NUMBER.Add(cifstring);

            is_protein = -1;
            get_value(cifstring, Table, pdb_start_index, i);
            if (cifstring != "") is_protein = atoi(cifstring.Text());
            PDB_RANGE_BEGIN.Add(is_protein);

            is_protein = -1;
            get_value(cifstring, Table, pdb_end_index, i);
            if (cifstring != "") is_protein = atoi(cifstring.Text());
            PDB_RANGE_END.Add(is_protein);

            is_protein = -1;
            get_value(cifstring, Table, entity_index, i);
            if (cifstring == "") {
                 ndb_log_message(NDB_MSG_ERR, "Warning: Missing _struct_ref.entity_id value.\n");
                 missing_value = 1;
            } else {
                 chain = _molecules[_rep_mol]->GetFirstChain();
                 while (chain) {
                      if (chain->entity_id() == atoi(cifstring.Text())) {
                           if (chain->chain_type() == ATOMP_TOKEN)
                                is_protein = 1;
                           else if (chain->chain_type() == ATOMN_TOKEN)
                                is_protein = 0;
                           break;
                      }  
                      chain = _molecules[_rep_mol]->GetNextChain();
                 }
                 if (is_protein < 0) {
                      ndb_log_message(NDB_MSG_ERR, "Warning: Can't determine the polymer type for entity_id=%s.\n",
                                       cifstring.Text());
                      missing_value = 1;
                 }
            }
            ENTITY.Add(cifstring);

            get_value(cifstring, Table, seq_index, i);
            if (cifstring == "") {
                 ndb_log_message(NDB_MSG_ERR, "Warning: Missing _struct_ref.rcsb_seq_one_letter_code value.\n");
                 missing_value = 1;
            } else if (is_protein >= 0) {
                 sequence.Clear();
                 for (unsigned int j = 0; j < strlen(cifstring.Text()); j++) {
                      if (cifstring.Text()[j] == ' ' ||
                          cifstring.Text()[j] == '\n' ||
                          cifstring.Text()[j] == '\t') continue;

                      if (cifstring.Text()[j] != '(') {
                           memset(buff, 0, 5);
                           if (is_protein)
                                strcpy(buff, get_threeletter_code(cifstring.Text()[j]));
                           else sprintf(buff, "%c", cifstring.Text()[j]);
                           ndb_set_string_to_upper(buff);
                           s.Copy(buff);
                           sequence.Add(s);
                      } else {
                           j++;
                           s.Clear();
                           while (cifstring.Text()[j] != ')') {
                                s += cifstring.Text()[j];
                                j++;
                           }
                           sequence.Add(s);
                      }
                 }
                 SEQUENCES.Add(sequence);
            }

            if (missing_value) break;
       }
       if (missing_value) return;

       if (SelfRef.Root() >= 0) {
            Table = _CifObj->GetTablePtr(_StructureId, "entity_poly");
            if (!is_empty_table(Table)) {
                 rowNo = Table->GetNumRows();
                 int ref_index = Table->GetColumnIndex("rcsb_build_self_reference");
                 int idx_index = Table->GetColumnIndex("entity_id");
                 if (ref_index < 0) {
                      Table->AddColumn("rcsb_build_self_reference");
                      ref_index = Table->GetColumnIndex("rcsb_build_self_reference");
                      ReVarCifArray<CifString> colData;
                      colData.Clear();
                      cifstring.Copy("?");
                      for (int i = 0; i < rowNo; i++) colData.Add(cifstring);
                      Table->FillColumn(colData, ref_index);
                 }
                 s.Copy("Y");
                 for (int i = 0; i < rowNo; ++i) {
                      get_value(cifstring, Table, idx_index, i);
                      if (SelfRef.Seek(cifstring) >= 0) Table->UpdateCell(s, ref_index, i);
                 }
                 _CifObj->WriteTable(Table, _StructureId, "entity_poly");
            }
       }

       if (_CifObj->IsTablePresent(_StructureId, "struct_ref_seq")) {
            _CifObj->DeleteTable(_StructureId, "struct_ref_seq");
       }
       ISTable *Dbref = new ISTable("struct_ref_seq");
       Dbref->AddColumn("align_id");
       Dbref->AddColumn("ref_id");
       Dbref->AddColumn("ndb_pdb_id_code");
       Dbref->AddColumn("ndb_chain_id");
       Dbref->AddColumn("seq_align_beg");
       Dbref->AddColumn("ndb_seq_align_beg_ins_code");
       Dbref->AddColumn("seq_align_end");
       Dbref->AddColumn("ndb_seq_align_end_ins_code");
       Dbref->AddColumn("ndb_db_accession");
       Dbref->AddColumn("db_align_beg");
       Dbref->AddColumn("db_align_end");
       Dbref->AddColumn("ndb_auth_seq_align_beg");
       Dbref->AddColumn("ndb_auth_seq_align_end");

       if (_CifObj->IsTablePresent(_StructureId, "struct_ref_seq_dif")) {
            _CifObj->DeleteTable(_StructureId, "struct_ref_seq_dif");
       }
       ISTable *Seqadv = new ISTable("struct_ref_seq_dif");
       Seqadv->AddColumn("align_id");
       Seqadv->AddColumn("ndb_pdb_id_code");
       Seqadv->AddColumn("mon_id");
       Seqadv->AddColumn("ndb_pdb_chain_id");
       Seqadv->AddColumn("seq_num");
       Seqadv->AddColumn("ndb_pdb_ins_code");
       Seqadv->AddColumn("ndb_seq_db_name");
       Seqadv->AddColumn("ndb_seq_db_accession_code");
       Seqadv->AddColumn("db_mon_id");
       Seqadv->AddColumn("ndb_seq_db_seq_num");
       Seqadv->AddColumn("details");
       Seqadv->AddColumn("ndb_auth_seq_num");
       Seqadv->AddColumn("rcsb_ordinal");

       char pdb_code[10];
       memset(pdb_code, 0, 10);
       Table = _CifObj->GetTablePtr(_StructureId, "database_2");
       if (!is_empty_table(Table)) get_pdb_id(Table, pdb_code);

       chain = _molecules[_rep_mol]->GetFirstChain();
       while (chain) {
            for (unsigned int i = 0; i < ENTITY.Length(); i++) {
                 if (chain->entity_id() == atoi(ENTITY[i].Text())) {
                       create_dbref_and_seqadv(chain, pdb_code, ID[i], NAME[i], ACCESSION[i],
                              BEGIN_NUMBER[i], SEQUENCES[i], PDB_RANGE_BEGIN[i],
                              PDB_RANGE_END[i], Dbref, Seqadv);
                 }
            }
            chain = _molecules[_rep_mol]->GetNextChain();
       }

       _CifObj->WriteTable(Dbref, _StructureId, "struct_ref_seq");
       _CifObj->WriteTable(Seqadv, _StructureId, "struct_ref_seq_dif");
}

static void create_dbref_and_seqadv(_CHAIN *chain, const char *pdb_code, CifString &ref_id,
             CifString &db_name, CifString &accession, CifString &start_number,
             ReVarCifArray<CifString> &DB_Seq, const int pdb_range_begin, const int pdb_range_end,
             ISTable *Dbref, ISTable *Seqadv)
{
       ALIGN *sa = new ALIGN;
       sa->AlignLength = DB_Seq.Length();
       sa->aln = new ALN[sa->AlignLength];
       for (int i0 = 0; i0 < sa->AlignLength; i0++) sa->aln[i0].pos[0] = i0;

       ReVarCifArray<CifString> seq_b;
       seq_b.Clear();
       CifString cs;

       ALIGN *sb = new ALIGN;
       sb->aln = new ALN[chain->SeqLen()];
       sb->AlignLength = 0;
       for (int i1 = 0; i1 < chain->SeqLen(); i1++) {
            if (chain->SeqRes(i1)->Deleted) continue;
            if (pdb_range_begin > 0 && pdb_range_end > 0 &&
                pdb_range_end > pdb_range_begin) {
                 if ((i1 + 1) < pdb_range_begin ||
                     (i1 + 1) > pdb_range_end) continue;
            }
            sb->aln[sb->AlignLength].pos[0] = i1;
            sb->AlignLength++;
            cs.Copy(chain->SeqRes(i1)->Field[0]);
            cs.ToUpper();
            seq_b.Add(cs);
       }

       ALIGN *ss = alignment_new(sa, sb, 2, DB_Seq, seq_b);

       int align_begin = -1;
       int align_end = -1;
       int begin = -1;
       int end = -1;

       for (int i2 = 0; i2 < ss->AlignLength; i2++) {
            if (ss->aln[i2].pos[0] >= 0 && ss->aln[i2].pos[1] >= 0) {
                 if (align_begin < 0) align_begin = i2;
                 align_end = i2;
            }
            if (ss->aln[i2].pos[1] >= 0) {
                 if (begin < 0) begin = i2;
                 end = i2;
            }
       }

       if (align_begin < 0 || align_end < 0 || begin < 0 || end < 0) {
            delete [] sa->aln; delete sa;
            delete [] sb->aln; delete sb;
            delete [] ss->aln; delete ss;
            return;
       }

       char buff[10];
       int align_id = Dbref->GetNumRows();
       Dbref->AddRow();
       sprintf(buff, "%d", align_id+1);
       cs.Copy(buff);
       Dbref->UpdateCell(cs, 0, align_id);
       Dbref->UpdateCell(ref_id, 1, align_id);
       cs.Copy(pdb_code);
       Dbref->UpdateCell(cs, 2, align_id);
       cs.Copy(chain->PDB_ChainID());
       Dbref->UpdateCell(cs, 3, align_id);
       Dbref->UpdateCell(accession, 8, align_id);

       int i = ss->aln[align_begin].pos[1];
       // cs.Copy(chain->SeqRes(i)->Field[2]);
       // Dbref->UpdateCell(cs, 4, align_id);
       cs.Copy(chain->SeqRes(i)->InsCode);
       Dbref->UpdateCell(cs, 5, align_id);
       cs.Copy(chain->SeqRes(i)->Field[4]);
       Dbref->UpdateCell(cs, 4, align_id);
       Dbref->UpdateCell(cs, 11, align_id);

       i = ss->aln[align_end].pos[1];
       // cs.Copy(chain->SeqRes(i)->Field[2]);
       // Dbref->UpdateCell(cs, 6, align_id);
       cs.Copy(chain->SeqRes(i)->InsCode);
       Dbref->UpdateCell(cs, 7, align_id);
       cs.Copy(chain->SeqRes(i)->Field[4]);
       Dbref->UpdateCell(cs, 6, align_id);
       Dbref->UpdateCell(cs, 12, align_id);

       sprintf(buff, "%d", atoi(start_number.Text()) + ss->aln[align_begin].pos[0]);
       cs.Copy(buff);
       Dbref->UpdateCell(cs, 9, align_id);
       sprintf(buff, "%d", atoi(start_number.Text()) + ss->aln[align_end].pos[0]);
       cs.Copy(buff);
       Dbref->UpdateCell(cs, 10, align_id);

       int *match_index = new int[end + 1];
       ReVarCifArray<CifString> parent_List;
       int row_no = Seqadv->GetNumRows();
       for (int j = begin; j <= end; j++) {
            match_index[j] = 0;
            if (ss->aln[j].pos[0] >= 0 && ss->aln[j].pos[1] >= 0) {
                 i = ss->aln[j].pos[1];
                 if (DB_Seq[ss->aln[j].pos[0]] == chain->SeqRes(i)->Field[0] ||
                     DB_Seq[ss->aln[j].pos[0]] == "A" &&
                     !strcmp(chain->SeqRes(i)->Field[0], "DA") ||
                     DB_Seq[ss->aln[j].pos[0]] == "C" &&
                     !strcmp(chain->SeqRes(i)->Field[0], "DC") ||
                     DB_Seq[ss->aln[j].pos[0]] == "G" &&
                     !strcmp(chain->SeqRes(i)->Field[0], "DG") ||
                     DB_Seq[ss->aln[j].pos[0]] == "I" &&
                     !strcmp(chain->SeqRes(i)->Field[0], "DI") ||
                     DB_Seq[ss->aln[j].pos[0]] == "T" &&
                     !strcmp(chain->SeqRes(i)->Field[0], "DT") ||
                     DB_Seq[ss->aln[j].pos[0]] == "U" &&
                     !strcmp(chain->SeqRes(i)->Field[0], "DU"))
                      match_index[j] = 1;
                 else {
                      if (is_modified_residue(chain->SeqRes(i)->Field[0],
                          DB_Seq[ss->aln[j].pos[0]].Text(), parent_List)) {
                           match_index[j] = 1;
                           if (parent_List.Length() > 0 &&
                              (j - (int) parent_List.Length()) >= begin) { 
                                bool not_match = false;
                                for (unsigned int k = 0; k < parent_List.Length(); ++k) {
                                     if (ss->aln[j-parent_List.Length()+k].pos[0] >= 0 &&
                                         ss->aln[j-parent_List.Length()+k].pos[1] < 0) {
                                          if (DB_Seq[ss->aln[j-parent_List.Length()
                                              +k].pos[0]] != parent_List[k]) {
                                               not_match = true;
                                          }
                                     } else not_match = true;
                                }
                                if (!not_match) {
                                     match_index[j] = -i;
                                     for (unsigned int k = 0; k < parent_List.Length(); ++k) {
                                          match_index[j-parent_List.Length()+k] = -i;
                                     }
                                }
                           }
                      }
                 }
            }
       }

       for (int j2 = begin; j2 <= end; j2++) {
            if (match_index[j2] > 0) continue;
            int exist_seqadv = 0;
            if (match_index[j2] < 0) {
                 i = -match_index[j2];
                 if (ss->aln[j2].pos[0] >= 0 &&
                     DB_Seq[ss->aln[j2].pos[0]] != chain->SeqRes(i)->Field[0]) {
                      exist_seqadv = 1;
                      Seqadv->AddRow();
                      cs.Copy(chain->SeqRes(i)->Field[0]);
                      Seqadv->UpdateCell(cs, 2, row_no);
                      cs.Copy(chain->SeqRes(i)->InsCode);
                      Seqadv->UpdateCell(cs, 5, row_no);
                      Seqadv->UpdateCell(DB_Seq[ss->aln[j2].pos[0]], 8, row_no);
                      sprintf(buff, "%d", atoi(start_number.Text()) + ss->aln[j2].pos[0]);
                      cs.Copy(buff);
                      Seqadv->UpdateCell(cs, 9, row_no);
                      cs.Copy("CHROMOPHORE");
                      Seqadv->UpdateCell(cs, 10, row_no);
                      cs.Copy(chain->SeqRes(i)->Field[4]);
                      Seqadv->UpdateCell(cs, 4, row_no);
                      Seqadv->UpdateCell(cs, 11, row_no);
                 }
            } else if (ss->aln[j2].pos[0] >= 0 && ss->aln[j2].pos[1] >= 0) {
                 i = ss->aln[j2].pos[1];
                 if (DB_Seq[ss->aln[j2].pos[0]] != chain->SeqRes(i)->Field[0]) {
                      exist_seqadv = 1;
                      Seqadv->AddRow();
                      cs.Copy(chain->SeqRes(i)->Field[0]);
                      Seqadv->UpdateCell(cs, 2, row_no);
                      cs.Copy(chain->SeqRes(i)->InsCode);
                      Seqadv->UpdateCell(cs, 5, row_no);
                      Seqadv->UpdateCell(DB_Seq[ss->aln[j2].pos[0]], 8, row_no);
                      sprintf(buff, "%d", atoi(start_number.Text()) + ss->aln[j2].pos[0]);
                      cs.Copy(buff);
                      Seqadv->UpdateCell(cs, 9, row_no);
                      cs.Copy("CONFLICT");
                      Seqadv->UpdateCell(cs, 10, row_no);
                      cs.Copy(chain->SeqRes(i)->Field[4]);
                      Seqadv->UpdateCell(cs, 4, row_no);
                      Seqadv->UpdateCell(cs, 11, row_no);
                 }
            } else if (ss->aln[j2].pos[0] >= 0) {
                 exist_seqadv = 1;
                 Seqadv->AddRow();
                 Seqadv->UpdateCell(DB_Seq[ss->aln[j2].pos[0]], 8, row_no);
                 sprintf(buff, "%d", atoi(start_number.Text()) + ss->aln[j2].pos[0]);
                 cs.Copy(buff);
                 Seqadv->UpdateCell(cs, 9, row_no);
                 cs.Copy("DELETION");
                 Seqadv->UpdateCell(cs, 10, row_no);
            } else if (ss->aln[j2].pos[1] >= 0) {
                 exist_seqadv = 1;
                 i = ss->aln[j2].pos[1];
                 Seqadv->AddRow();
                 cs.Copy(chain->SeqRes(i)->Field[0]);
                 Seqadv->UpdateCell(cs, 2, row_no);
                 cs.Copy(chain->SeqRes(i)->InsCode);
                 Seqadv->UpdateCell(cs, 5, row_no);
                 // cs.Copy("INSERTION");
                 cs.Copy("EXPRESSION TAG");
                 Seqadv->UpdateCell(cs, 10, row_no);
                 cs.Copy(chain->SeqRes(i)->Field[4]);
                 Seqadv->UpdateCell(cs, 4, row_no);
                 Seqadv->UpdateCell(cs, 11, row_no);
            }
            if (exist_seqadv) {
                 sprintf(buff, "%d", align_id+1);
                 cs.Copy(buff);
                 Seqadv->UpdateCell(cs, 0, row_no);
                 cs.Copy(pdb_code);
                 Seqadv->UpdateCell(cs, 1, row_no);
                 cs.Copy(chain->PDB_ChainID());
                 Seqadv->UpdateCell(cs, 3, row_no);
                 Seqadv->UpdateCell(db_name, 6, row_no);
                 Seqadv->UpdateCell(accession, 7, row_no);
                 sprintf(buff, "%d", row_no+1);
                 cs.Copy(buff);
                 Seqadv->UpdateCell(cs, 12, row_no);
                 row_no++;
            }
       }
       delete [] match_index;

       delete [] sa->aln; delete sa;
       delete [] sb->aln; delete sb;
       delete [] ss->aln; delete ss;
}

static int get_number_strands(ISTable *Table, const char *sheet_id)
{
       int  id_index, num_index, rowNo, irow;
       CifString id, number;
       if (is_empty_table(Table)) return 1;

       id_index = Table->GetColumnIndex("id");
       num_index = Table->GetColumnIndex("number_strands");
       if (id_index < 0 || num_index < 0) return 1;
       rowNo = Table->GetNumRows();
       for (irow = 0; irow < rowNo; irow++) {
            get_value(id, Table, id_index, irow);
            get_value(number, Table, num_index, irow);
            if (!strcasecmp(sheet_id, id.Text())) {
                 return (atoi(number.Text()));
            }
       }
       return 1;
}

static int get_sense(ISTable *Table, const char *sheet_id, const char *id_1, const char *id_2)
{
       int id_index, range_id_1, range_id_2, sense_index, rowNo, irow;
       char sense[50];
       CifString id, range_1, range_2, cifstring;

       if (is_empty_table(Table)) return 0;
       id_index = Table->GetColumnIndex("sheet_id");
       range_id_1 = Table->GetColumnIndex("range_id_1");
       range_id_2 = Table->GetColumnIndex("range_id_2");
       sense_index  = Table->GetColumnIndex("sense");
       if (id_index < 0 || range_id_1 < 0 || range_id_2 < 0 || sense_index < 0) return 0;
       rowNo = Table->GetNumRows();
       for (irow = 0; irow < rowNo; irow++) {
            get_value(id, Table, id_index, irow);
            get_value(range_1, Table, range_id_1, irow);
            get_value(range_2, Table, range_id_2, irow);
            get_value(cifstring, Table, sense_index, irow);
            if (!strcasecmp(sheet_id, id.Text()) && !strcasecmp(id_1, range_1.Text()) &&
                !strcasecmp(id_2, range_2.Text())) {
                 strcpy(sense, cifstring.Text());
                 ndb_clean_string(sense);
                 if (sense[0] == '-' || isdigit(sense[0])) return (atoi(sense));
                 else if (!strcasecmp(sense, "anti-parallel")) return (-1);
                 else if (!strcasecmp(sense, "parallel")) return (1);
            }
       }
       return 0;
}

static void get_hbond(_MOLECULE *mol, ISTable *Table, const int chain_index,
            const char *sheet_id, const char *range_id, CifString &curAtom, int &curIndex,
            int &curRes, CifString &prevAtom, int &prevIndex, int &prevRes)
{
       if (is_empty_table(Table)) return;
       int id_index = Table->GetColumnIndex("sheet_id");
       int range_id_2 = Table->GetColumnIndex("range_id_2");
       int beg_label_atom_id = Table->GetColumnIndex("range_2_label_atom_id");
       int beg_label_comp_id = Table->GetColumnIndex("range_2_label_comp_id");
       int beg_label_asym_id = Table->GetColumnIndex("range_2_label_asym_id");
       int beg_label_seq_id = Table->GetColumnIndex("range_2_label_seq_id");
       int beg_label_ins_code = Table->GetColumnIndex("range_2_PDB_ins_code");
       int beg_auth_comp_id = Table->GetColumnIndex("range_2_auth_comp_id");
       int beg_auth_asym_id = Table->GetColumnIndex("range_2_auth_asym_id");
       int beg_auth_seq_id = Table->GetColumnIndex("range_2_auth_seq_id");
       int end_label_atom_id = Table->GetColumnIndex("range_1_label_atom_id");
       int end_label_comp_id = Table->GetColumnIndex("range_1_label_comp_id");
       int end_label_asym_id = Table->GetColumnIndex("range_1_label_asym_id");
       int end_label_seq_id = Table->GetColumnIndex("range_1_label_seq_id");
       int end_label_ins_code = Table->GetColumnIndex("range_1_PDB_ins_code");
       int end_auth_comp_id = Table->GetColumnIndex("range_1_auth_comp_id");
       int end_auth_asym_id = Table->GetColumnIndex("range_1_auth_asym_id");
       int end_auth_seq_id = Table->GetColumnIndex("range_1_auth_seq_id");
       int rowNo = Table->GetNumRows();

       int irow, BegIndex, BegRes, EndIndex, EndRes, beg, end;
       CifString id, range_2, name1, name2;
       for (irow = 0; irow < rowNo; irow++) {
            get_value(id, Table, id_index, irow);
            get_value(range_2, Table, range_id_2, irow);
            if (!strcasecmp(sheet_id, id.Text()) && !strcasecmp(range_id, range_2.Text())) {
                 beg = get_residue_index(mol, Table, irow, beg_label_comp_id,
                       beg_label_asym_id, beg_label_seq_id, beg_auth_comp_id,
                       beg_auth_asym_id, beg_auth_seq_id, beg_label_ins_code,
                       BegIndex, BegRes, name1);
                 end = get_residue_index(mol, Table, irow, end_label_comp_id,
                       end_label_asym_id, end_label_seq_id, end_auth_comp_id,
                       end_auth_asym_id, end_auth_seq_id, end_label_ins_code,
                       EndIndex, EndRes, name2);
                 if (beg && end && (BegIndex == chain_index || EndIndex == chain_index)) {
                      curIndex = BegIndex; curRes = BegRes;
                      prevIndex = EndIndex; prevRes = EndRes;
                      get_value(curAtom, Table, beg_label_atom_id, irow);
                      get_value(prevAtom, Table, end_label_atom_id, irow);
                 }
                 return;
            }
       }
}

static void cif_to_ndb_find_serial_no(ISTable *Table, const int id, const int irow,
         const int tokenid, int *idfield, char idvalue[10][82], char SerialNo[10][82])
{
       int i, j, k;
       unsigned int str_len;
       CifString value;

       for (j = 0; j < 10; j++) {
            SerialNo[j][0] = '\0';
            idvalue[j][0] = '\0';
       }

       if (CifCategory[id].nCifKeyField) {
            for (i = 0; i < CifCategory[id].nCifKeyField; i++) {
                 j =  CifCategory[id].CifKeyField[i];
                 get_value(value, Table, idfield[i], irow);
                 if (tokenid == JRNL_TOKEN-1) {
/*
                      if (value == "primary")
                           strcpy(SerialNo[i], "0");
                      else */ if (value != "")
                           strcpy(SerialNo[i], value.Text());
                      else sprintf(SerialNo[i], "%d", irow);
                 } else if (value != "") {
                      strcpy(SerialNo[i], value.Text());
                      for (k = 0; k < CifCategory[id].keywords[j].numNdbInfo; k++) {
                           if (!strcmp(CifCategory[id].category, "database_2")) {
                                if (!strcmp(value.Text(), CifCategory[id].keywords[j].
                                      NdbInfo[k].JrnlTokenName_Type)) {
                                     strcpy(SerialNo[i], "");
                                     strcpy(idvalue[i], CifCategory[id].keywords[j].
                                           NdbInfo[k].JrnlTokenName_Type);
                                     break;
                                }
                           } else {
                                str_len = strlen(CifCategory[id].keywords[j].NdbInfo[k].
                                        JrnlTokenName_Type);
                                if (str_len != 0 && strlen(value.Text()) >= str_len &&
                                   !strncmp(value.Text(), CifCategory[id].keywords[j].
                                      NdbInfo[k].JrnlTokenName_Type, str_len)) {
                                     strcpy(SerialNo[i], &value.Text()[str_len]);
                                     strcpy(idvalue[i], CifCategory[id].keywords[j].
                                           NdbInfo[k].JrnlTokenName_Type);
                                     break;
                                }
                           }
                      }
                 } else sprintf(SerialNo[i], "%d", irow + 1);
            }
       } else sprintf(SerialNo[0], "%d", irow + 1);
}

static void process_sg_entry(CifFileObj *fobj, const char *blockId)
{
       ISTable *Table = fobj->GetTablePtr(blockId, "pdbx_SG_project");
       if (is_empty_table(Table)) return;

       int project_index = Table->GetColumnIndex("project_name");
       int full_index = Table->GetColumnIndex("full_name_of_center");
       int initial_index = Table->GetColumnIndex("initial_of_center");
       int rowNo1 = Table->GetNumRows();
       int is_SG_entry = 0, is_PSI_entry = 0;
       char buffer[2000], buffer1[2000];
       CifString cifstring, cifstring1;
       int i, k; 
       for (k = 0; k < rowNo1; k++) {
            cifstring.Clear();
            Table->GetCell(cifstring, full_index, k);
            for (i = 0; i < num_sg_centers; i++) {
                 if (!strcasecmp(cifstring.Text(), _sg_info[i]._center_name)) {
                      cifstring.Copy(_sg_info[i]._center_name);
                      Table->UpdateCell(cifstring, full_index, k);
                      cifstring.Copy(_sg_info[i]._center_initial);
                      Table->UpdateCell(cifstring, initial_index, k);
                      // cifstring.Copy(_sg_info[i]._project_name);
                      // Table->UpdateCell(cifstring, project_index, k);
                      is_SG_entry = 1;
                      if (_sg_info[i].psi_flag > is_PSI_entry) is_PSI_entry = _sg_info[i].psi_flag;
                      if (_sg_info[i].psi_flag >= 2) {
                           get_value(cifstring, Table, project_index, k);
                           if (cifstring == "") {
                                cifstring.Copy(_sg_info[i]._project_name);
                                Table->UpdateCell(cifstring, project_index, k);
                           }
                      }
                      break;
                 }
            }

            cifstring.Clear();
            Table->GetCell(cifstring, project_index, k);
            if (cifstring == "?" || cifstring == ".") cifstring.Clear();
            if (cifstring != "") is_SG_entry = 1;
       }
       fobj->WriteTable(Table, blockId, "pdbx_SG_project");

       if (is_SG_entry) {
            ISTable *Table1 = fobj->GetTablePtr(blockId, "ndb_database_status");
            if (!is_empty_table(Table1)) {
                 cifstring.Copy("Y");
                 int name_index = Table1->GetColumnIndex("SG_entry");
                 if (name_index < 0) {
                      Table1->AddColumn("SG_entry");
                      name_index = Table1->GetColumnIndex("SG_entry");
                      ReVarCifArray<CifString> colData;
                      colData.Clear();
                      colData.Add(cifstring);
                      Table1->FillColumn(colData, name_index);
                 }
                 Table1->UpdateCell(cifstring, name_index, 0);
                 fobj->WriteTable(Table1, blockId, "ndb_database_status");
            }

            CifString full_name, initials, project, Cap_initials;
            for (k = 0; k < rowNo1; k++) {
                 Table->GetCell(project, project_index, k);
                 if (project == "?" || project == ".") project.Copy("");
                 Table->GetCell(full_name, full_index, k);
                 if (full_name == "?" || full_name == ".") full_name.Copy("");
                 Table->GetCell(initials, initial_index, k);
                 if (initials == "?" || initials == ".") initials.Copy("");

                 if (full_name != "" && is_PSI_entry) {
                      Table1 = fobj->GetTablePtr(blockId, "audit_author");
                      if (is_empty_table(Table1)) {
                           // fobj->DeleteTable(blockId, "audit_author");
                           Table1 = new ISTable("audit_author");
                           Table1->AddColumn("name");
                           ISTable *Table2 = fobj->GetTablePtr(blockId, "citation_author");
                           if (!is_empty_table(Table2)) {
                                int rowNo = Table2->GetNumRows();
                                int citation_id = Table2->GetColumnIndex("citation_id");
                                int name_index = Table2->GetColumnIndex("name");
                                int row = 0;
                                for (i = 0; i < rowNo; i++) {
                                     Table2->GetCell(cifstring, citation_id, i);
                                     if (cifstring == "primary") {
                                          Table2->GetCell(cifstring, name_index, i);
                                          if (cifstring != "?" && cifstring != "." && cifstring != "") {
                                               Table1->AddRow();
                                               Table1->UpdateCell(cifstring, 0, row);
                                               row++;
                                          }
                                     }
                                }
                           }
                           fobj->WriteTable(Table1, blockId, "audit_author");
                      }
                      int name_index = Table1->GetColumnIndex("name");
                      int rowNo = Table1->GetNumRows();
                      for (i = 0; i < rowNo; i++) {
                           Table1->GetCell(cifstring, name_index, i);
                           strcpy(buffer, cifstring.Text());
                           ndb_set_string_to_upper(buffer);
                           strcpy(buffer1, full_name.Text());
                           ndb_set_string_to_upper(buffer1);
                           Cap_initials.Copy(initials);
                           Cap_initials.ToUpper();
                           if (strstr(buffer, buffer1) &&
                               strstr(buffer, Cap_initials.Text())) {
                                cifstring.Copy(full_name.Text());
                                if (initials != "") {
                                     cifstring += " (";
                                     cifstring += initials.Text();
                                     cifstring += ")";
                                }
                                Table1->UpdateCell(cifstring, name_index, i);
                                break;
                           }
                      }
                      if (i == rowNo) {
                           Table1->AddRow();
                           cifstring.Copy(full_name.Text());
                           if (initials != "") {
                                cifstring += " (";
                                cifstring += initials.Text();
                                cifstring += ")";
                           }
                           Table1->UpdateCell(cifstring, name_index, rowNo);
                      }
                      fobj->WriteTable(Table1, blockId, "audit_author");
                 }
#if 0
                 Table1 = fobj->GetTablePtr(blockId, "struct_keywords");
                 if (is_empty_table(Table1)) {
                      // fobj->DeleteTable(blockId, "struct_keywords");
                      Table1 = new ISTable("struct_keywords");
                      Table1->AddColumn("entry_id");
                      Table1->AddColumn("ndb_keywords");
                      Table1->AddColumn("text");
                      Table1->AddRow();
                      cifstring.Copy(blockId);
                      Table1->UpdateCell(cifstring, 0, 0);
                      fobj->WriteTable(Table1, blockId, "struct_keywords");
                 }
                 int name_index = Table1->GetColumnIndex("text");
                 cifstring.Clear();
                 Table1->GetCell(cifstring, name_index, 0);
                 if (cifstring == "?" || cifstring == ".") cifstring.Clear();
                 strcpy(buffer, cifstring.Text());
                 ndb_set_string_to_upper(buffer);
                 if (!strstr(buffer, "STRUCTURAL GENOMICS")) {
                      if (cifstring != "") cifstring += ", ";
                      cifstring += "Structural Genomics";
                 }
                 if (is_PSI_entry == 2) {
                      if (!strstr(buffer, "PSI")) {
                           if (cifstring != "") cifstring += ", ";
                           cifstring += "PSI-2";
                      }
                      if (!strstr(buffer, "PROTEIN STRUCTURE INITIATIVE")) {
                           if (cifstring != "") cifstring += ", ";
                           cifstring += "Protein Structure Initiative"; 
                      }
/*
                 } else if (is_PSI_entry == 1) {
                      if (!strstr(buffer, "PSI")) {
                           if (cifstring != "") cifstring += ", ";
                           cifstring += "PSI";
                      }
                      if (!strstr(buffer, "PROTEIN STRUCTURE INITIATIVE")) {
                           if (cifstring != "") cifstring += ", ";
                           cifstring += "Protein Structure Initiative"; 
                      }
*/
                 } else if (!is_PSI_entry) {
                      strcpy(buffer1, project.Text());
                      ndb_set_string_to_upper(buffer1);
                      if (strcmp(buffer1, "") && !strstr(buffer, buffer1)) {
                           if (cifstring != "") cifstring += ", ";
                           cifstring += project.Text();
                      }
                 }
                 strcpy(buffer1, full_name.Text());
                 ndb_set_string_to_upper(buffer1);
                 if (strcmp(buffer1, "") && !strstr(buffer, buffer1)) {
                      if (cifstring != "") cifstring += ", ";
                      cifstring += full_name.Text();
                 }
                 strcpy(buffer1, initials.Text());
                 ndb_set_string_to_upper(buffer1);
                 if (strcmp(buffer1, "") && !strstr(buffer, buffer1)) {
                      if (cifstring != "") cifstring += ", ";
                      cifstring += initials.Text();
                 }
                 Table1->UpdateCell(cifstring, name_index, 0);
                 fobj->WriteTable(Table1, blockId, "struct_keywords");
#endif
            }

            Table = fobj->GetTablePtr(blockId, "entity_poly"); 
            if (!is_empty_table(Table)) {
                 int target_index = Table->GetColumnIndex("pdbx_target_identifier");
                 rowNo1 = Table->GetNumRows();
                 for (i = 0; i < rowNo1; i++) {
                      cifstring.Clear();
                      Table->GetCell(cifstring, target_index, i);
                      if (cifstring == "." || cifstring == "?") cifstring.Clear();
                      if (cifstring != "") break;
                 }
                 if (i < rowNo1) {
                      Table1 = fobj->GetTablePtr(blockId, "rcsb_database_related");
                      if (is_empty_table(Table1)) {
                           // fobj->DeleteTable(blockId, "rcsb_database_related");
                           Table1 = new ISTable("rcsb_database_related");
                           Table1->AddColumn("db_name");
                           Table1->AddColumn("db_id");
                           Table1->AddColumn("details");
                           fobj->WriteTable(Table1, blockId, "rcsb_database_related");
                      }
                      int db_name_index = Table1->GetColumnIndex("db_name");
                      int db_id_index = Table1->GetColumnIndex("db_id");
                      int rowNo = Table1->GetNumRows();
                      CifString target;
                      for (k = 0; k < rowNo1; k++) {
                           target.Clear();
                           Table->GetCell(target, target_index, k);
                           if (target == "." || target == "?" || target == "") continue; 

                           strcpy(buffer, target.Text());
                           int len = strlen(buffer);
                           for (i = 0; i < len; i++) {
                                if (buffer[i] == ',' || buffer[i] == ';') buffer[i] = ' ';
                           }
                           ndb_clean_string(buffer);

                           char *p = buffer;
                           while (p) {
                                p = get_word(p, buffer1);
                                if (!strcmp(buffer1, "")) continue;

                                for (i = 0; i < rowNo; i++) {
                                     Table1->GetCell(cifstring, db_name_index, i);
                                     Table1->GetCell(cifstring1, db_id_index, i);
                                     if ((!strcasecmp(cifstring.Text(), "TargetDB") ||
                                          !strcasecmp(cifstring.Text(), "TargetTrack")) &&
                                         !strcasecmp(cifstring1.Text(), buffer1)) break;
                                }
                                if (i == rowNo) {
                                     Table1->AddRow();
                                     // cifstring.Copy("TargetDB");
                                     cifstring.Copy("TargetTrack");
                                     Table1->UpdateCell(cifstring, db_name_index, rowNo);
                                     cifstring.Copy(buffer1);
                                     Table1->UpdateCell(cifstring, db_id_index, rowNo);
                                     fobj->WriteTable(Table1, blockId, "rcsb_database_related");
                                     rowNo++;
                                }
                           }
                      }
                 }
            }
       }
}

static int is_modified_residue(const char *pdb_residue, const char *seq_residue,
              ReVarCifArray<CifString> &parent_list)
{
       parent_list.Clear();
       int answer = 0;
       ConnectFormat *drug = find_drug(pdb_residue);
       if (drug && drug->parent_residue() && strcmp(drug->parent_residue(), "")) {
            int len = strlen(drug->parent_residue());
            char *buff = new char[len + 1];
            memset(buff, 0, len + 1);
            strcpy(buff, drug->parent_residue());
            char *buffer = new char[len + 1];
            memset(buffer, 0, len + 1);
            for (int j = 0; j < len; j++) {
                 if (buff[j] == ',' || buff[j] == ';') buff[j] = ' ';
            }
            ndb_clean_string(buff);
            CifString cs;
            char *p = buff;
            ReVarCifArray<CifString> list;
            list.Clear();
            while (p) {
                 p = get_word(p, buffer);
                 cs.Copy(buffer);
                 list.Add(cs);
            }
            delete [] buff;
            delete [] buffer;
            if (list.Length() > 0) {
                 len = list.Length() - 1;
                 if (list[len] == seq_residue) {
                      answer = 1;
                      for (int j = 0; j < len; j++) parent_list.Add(list[j]);
                 }
            }
       }
       return answer;
}

static void remove_redundant_keywords(CifFileObj *fobj, const char* blockId)
{
       ISTable *t = fobj->GetTablePtr(blockId, "struct_keywords");
       if (is_empty_table(t)) return;

       int index = t->GetColumnIndex("text");
       CifString keyword;
       get_value(keyword, t, index, 0);
       // if (keyword == "") return;

       TblIndexObj Stop_words;
       Stop_words.Clear();

       CifString stop_word_file, cs;
       stop_word_file.Copy(root_dir);
       stop_word_file += "/data/ascii/stop_word_list";

       CifString header /*, HEADER */ ;
       // HEADER.Clear();
       index = t->GetColumnIndex("ndb_keywords");
       get_value(header, t, index, 0);
       if (header != "") {
            bool is_complex = false;
            for (unsigned int i = 0; i < header.Length(); ++i) {
                 if (header[i] == '/') {
                      header[i] = '-';
                      is_complex = true;
                 }
            }
            if (is_complex) header += " complex";
            // HEADER.Copy(header);
            // HEADER.ToUpper();
       }
/*
       index = t->GetColumnIndex("ndb_keywords");
       get_value(cs, t, index, 0);
       if (cs != "") {
            cs.ToUpper();
            Stop_words.Add(cs);
       }

       t = fobj->GetTablePtr(blockId, "entity_name_com");
       if (!is_empty_table(t)) {
            int rowNo = t->GetNumRows();
            index = t->GetColumnIndex("name");
            for (int i = 0; i < rowNo; ++i) {
                 get_value(cs, t, index, i);
                 if (cs == "") continue;
                 cs.ToUpper();
                 if (Stop_words.Seek(cs) < 0) Stop_words.Add(cs);
            }
       }

       t = fobj->GetTablePtr(blockId, "entity");
       if (!is_empty_table(t)) {
            int rowNo = t->GetNumRows();
            index = t->GetColumnIndex("ndb_description");
            int index1 = t->GetColumnIndex("type");
            for (int i = 0; i < rowNo; ++i) {
                 get_value(cs, t, index1, i);
                 cs.ToLower();
                 if (cs != "polymer") continue;
                 get_value(cs, t, index, i);
                 if (cs == "") continue;
                 cs.ToUpper();
                 if (Stop_words.Seek(cs) < 0) Stop_words.Add(cs);
            }
       }

       t = fobj->GetTablePtr(blockId, "entity_src_nat");
       if (!is_empty_table(t)) {
            int rowNo = t->GetNumRows();
            index = t->GetColumnIndex("common_name");
            int index1 = t->GetColumnIndex("ndb_organism_scientific");
            int index2 = t->GetColumnIndex("strain");
            CifString cs1;
            for (int i = 0; i < rowNo; ++i) {
                 get_value(cs, t, index, i);
                 if (cs != "") { 
                      cs.ToUpper();
                      if (Stop_words.Seek(cs) < 0) Stop_words.Add(cs);
                 }
                 get_value(cs, t, index1, i);
                 if (cs != "") { 
                      cs.ToUpper();
                      if (Stop_words.Seek(cs) < 0) Stop_words.Add(cs);
                 }
                 get_value(cs1, t, index2, i);
                 if (cs1 != "") { 
                      cs1.ToUpper();
                      if (Stop_words.Seek(cs1) < 0) Stop_words.Add(cs1);
                 }
                 if (cs != "" && cs1 != "") {
                      cs += " ";
                      cs += cs1;
                      cs.ToUpper();
                      if (Stop_words.Seek(cs) < 0) Stop_words.Add(cs);
                 }
            }
       }

       t = fobj->GetTablePtr(blockId, "entity_src_gen");
       if (!is_empty_table(t)) {
            int rowNo = t->GetNumRows();
            index = t->GetColumnIndex("gene_src_common_name");
            int index1 = t->GetColumnIndex("ndb_gene_src_scientific_name");
            int index2 = t->GetColumnIndex("gene_src_strain");
            int index3 = t->GetColumnIndex("host_org_common_name");
            int index4 = t->GetColumnIndex("ndb_host_org_scientific_name");
            int index5 = t->GetColumnIndex("ndb_host_org_strain");
            CifString cs1;
            for (int i = 0; i < rowNo; ++i) {
                 get_value(cs, t, index, i);
                 if (cs != "") { 
                      cs.ToUpper();
                      if (Stop_words.Seek(cs) < 0) Stop_words.Add(cs);
                 }
                 get_value(cs, t, index1, i);
                 if (cs != "") { 
                      cs.ToUpper();
                      if (Stop_words.Seek(cs) < 0) Stop_words.Add(cs);
                 }
                 get_value(cs1, t, index2, i);
                 if (cs1 != "") { 
                      cs1.ToUpper();
                      if (Stop_words.Seek(cs1) < 0) Stop_words.Add(cs1);
                 }
                 if (cs != "" && cs1 != "") {
                      cs += " ";
                      cs += cs1;
                      cs.ToUpper();
                      if (Stop_words.Seek(cs) < 0) Stop_words.Add(cs);
                 }
                 get_value(cs, t, index3, i);
                 if (cs != "") { 
                      cs.ToUpper();
                      if (Stop_words.Seek(cs) < 0) Stop_words.Add(cs);
                 }
                 get_value(cs, t, index4, i);
                 if (cs != "") { 
                      cs.ToUpper();
                      if (Stop_words.Seek(cs) < 0) Stop_words.Add(cs);
                 }
                 get_value(cs1, t, index5, i);
                 if (cs1 != "") { 
                      cs1.ToUpper();
                      if (Stop_words.Seek(cs1) < 0) Stop_words.Add(cs1);
                 }
                 if (cs != "" && cs1 != "") {
                      cs += " ";
                      cs += cs1;
                      cs.ToUpper();
                      if (Stop_words.Seek(cs) < 0) Stop_words.Add(cs);
                 }
            }
       }
*/
       struct stat statbuf;
       if (stat(stop_word_file.Text(), &statbuf) == 0) {
            FILE *fp = fopen(stop_word_file.Text(), "r");
            char buffer[200];
            while (!feof(fp)) {
                 fgets(buffer, 199, fp);
                 if (feof(fp)) break;
                 ndb_clean_string(buffer);
                 ndb_set_string_to_upper(buffer);
                 cs.Copy(buffer);
                 if (Stop_words.Seek(cs) < 0) Stop_words.Add(cs);
            }
            fclose (fp);
       } else {
            for (int i = 0; i < NUM_STOP_WORD; ++i) {
                 cs.Copy(_stop_words[i]);
                 if (Stop_words.Seek(cs) < 0) Stop_words.Add(cs);
            }
       }

       ReVarCifArray<CifString> data, header_and_keywords;
       header_and_keywords.Clear();
       if (keyword != "") header_and_keywords.Add(keyword);
       if (header != "") header_and_keywords.Add(header);
       if (header_and_keywords.Length() == 0) return;

       keyword.Clear();
       for (unsigned int l = 0; l < header_and_keywords.Length(); ++l) {
            get_wordlist_from_string_separated_by_delimits(data,
                        header_and_keywords[l].Text(), ", ");
            for (unsigned int i = 0; i < data.Length(); ++i) {
                 cs.Copy(data[i]);
                 cs.ToUpper();
                 if (strstr(cs.Text(), "/") && strstr(cs.Text(), "COMPLEX")) {
                      for (unsigned int j = 0; j < cs.Length(); ++j) {
                           if (cs[j] == '/') cs[j] = '-';
                      }
                      for (unsigned int j = 0; j < data[i].Length(); ++j) {
                           if (data[i][j] == '/') data[i][j] = '-';
                      }
                 }
                 if (Stop_words.Seek(cs) >= 0) continue;
                 Stop_words.Add(cs);
                 if (keyword != "") keyword += ", ";
                 keyword += data[i];
            }
       }

       t = fobj->GetTablePtr(blockId, "struct_keywords");
       index = t->GetColumnIndex("text");
       t->UpdateCell(keyword, index, 0);
       fobj->WriteTable(t, blockId, "struct_keywords");
}
