#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "align_util.h"
#include "align_standard.h"
#include "align_standard_local.h"
#include "CifString.h"
#include "ndb_misclib.h"
#include "ReVarCifArray.h"

static void print_alignment(FILE *fp, ALIGN *ss, const char *_seqa, const char *_seqb);
static float score(const int k, const int l);
static int find_index(char aa);
static void get_one_line(char *line, const int max_length, FILE *infile);

int check_target_sequence(FILE *report, const char *entity_sequence, 
               const char *entity_id, const char *target_id)
{
       CifString command;
/*
       command.Copy("/usr/bin/curl \"http://targetdb.sbkb.org/servlet/TargetSearch?id=");
       command += target_id;
       command += "&pdbid=&pfamid=&which_lab=&status=&seqname=Query&which_seq=SG&";
       command += "Aftermonth=&Afterday=&Afteryear=&Beforemonth=&Beforeday=&";
       command += "Beforeyear=&p_name=&org_name=&Sequence=&format=fasta\" > seq.result";
*/
       command.Copy("/usr/bin/curl \"http://sbkb.org/tt/search?format=fasta&fileread=Y&targetid=");
       command += target_id;
       command += "\" > seq.result";
       system(command.Text());

       FILE *fp = fopen("seq.result", "r");
       if (fp == NULL) return (0);

       ReVarCifArray<CifString> sequences;
       sequences.Clear();

       command.Clear();
       int found_seq = 0;
       char line[100];
       while (!feof(fp)) {
            get_one_line(line, 99, fp);
            if (feof(fp)) break;
            if (line[0] == '>') {
                 if (command != "") sequences.Add(command);
                 command.Clear();
                 found_seq = 1;
            } else if (found_seq) {
                 ndb_clean_string(line);
                 if (line[strlen(line)-1] == 13) line[strlen(line)-1] = 0;
                 command += line;
            }
       }
       fclose (fp);
       if (command != "") sequences.Add(command);

       if (sequences.Length() == 0) return 0;

       int length_b = strlen(entity_sequence);
       int *index_s = new int[length_b];
       for (int i = 0; i < length_b; i++) {
            index_s[i] = find_index(entity_sequence[i]);
       }

       int best_similarity = 0;
       int best_i = -1;
       ALIGN *best_ss = NULL;
       for (unsigned int i = 0; i < sequences.Length(); i++) {
            int length_a = strlen(sequences[i].Text());
            int *index_t = new int[length_a];
            for (int j = 0; j < length_a; j++) {
                 index_t[j] = find_index(sequences[i].Text()[j]);
            }

            _indexa = index_t;
            _indexb = index_s;

            ALIGN *sa = new ALIGN;
            sa->AlignLength = length_a;
            sa->aln = new ALN[sa->AlignLength];
            for (int j = 0; j < sa->AlignLength; j++) sa->aln[j].pos[0] = j;
            ALIGN *sb = new ALIGN;
            sb->AlignLength = length_b;
            sb->aln = new ALN[sb->AlignLength];
            for (int j = 0; j < sb->AlignLength; j++) sb->aln[j].pos[0] = j;
            ALIGN *ss = alignment_standard(sa, sb, 10, 1, score);

            int count = 0;
            for (int j = 0; j < ss->AlignLength; j++) {
                 if (ss->aln[j].pos[0] >= 0 && ss->aln[j].pos[1] >= 0) {
                      if (sequences[i].Text()[ss->aln[j].pos[0]] ==
                          entity_sequence[ss->aln[j].pos[1]])
                           count++;
                 }
            }

            delete [] sa->aln; delete sa;
            delete [] sb->aln; delete sb;
            delete [] index_t;

            int similarity = 100 * count / length_b;
            if (similarity > best_similarity) {
                 best_i = i;
                 best_similarity = similarity;
                 if (best_ss) { delete [] best_ss->aln; delete best_ss; }
                 best_ss = ss;
            } else {
                 delete [] ss->aln; delete ss;
                 ss = NULL;
            }
       }

       fprintf(report, "Sequence Identity between entity %s and target %s = %d%%\n\n",
                        entity_id, target_id, best_similarity);
       if (best_similarity < 80 && best_ss && best_i >= 0) {
            print_alignment(report, best_ss, sequences[best_i].Text(), entity_sequence);
       }

       delete [] index_s;
       if (best_ss) { delete [] best_ss->aln; delete best_ss; }

       if (best_similarity == 0) best_similarity = 1;
       return best_similarity;
}

static void print_alignment(FILE *fp, ALIGN *ss, const char *_seqa, const char *_seqb)
{
       int i, l, m, n, num_per_line = 60, x, y, u, v;

       l = ss->AlignLength / num_per_line;
       x = ss->AlignLength % num_per_line;
       m = l;
       if (x == 0) m = l - 1;

       for (i = 0; i <= m; i++) {
            n = num_per_line;
            if (i == l) n = x;

	    fprintf(fp, "Target: ");
            for (y = 0; y < n; y++) {
                 u = ss->aln[i * num_per_line + y].pos[0];
                 if (u < 0)
                      fprintf(fp, "-");
                 else fprintf(fp, "%c", _seqa[u]);
            }
            fprintf(fp, "\n");

	    fprintf(fp, "        ");
            for (y = 0; y < n; y++) {
                 u = ss->aln[i * num_per_line + y].pos[0];
                 v = ss->aln[i * num_per_line + y].pos[1];
                 if (u >= 0 && v >= 0 && _seqa[u] == _seqb[v])
                      fprintf(fp, "|");
                 else fprintf(fp, " ");
            }
            fprintf(fp, "\n");

	    fprintf(fp, "Entity: ");
            for (y = 0; y < n; y++) {
                 v = ss->aln[i * num_per_line + y].pos[1];
                 if (v < 0)
                      fprintf(fp, "-");
                 else fprintf(fp, "%c", _seqb[v]);
            }
            fprintf(fp, "\n\n\n");
       }
}

static float score(const int k, const int l)
{
       int index_k, index_l;

       index_k = _indexa[k];
       index_l = _indexb[l];
       if (index_k >= 0 && index_l >= 0)
            return matrix[index_k][index_l];
       else return (-4);
}

static int find_index(char aa)
{
       int i;

       for (i = 0; i < 21; i++) {
            if (aa == AA[i]) return i;
       }
       return (-1);
}

static void get_one_line(char *line, const int max_length, FILE *infile)
{
       int temp_c, length;

       *line = '\0';
       length = 0;
       do {
            temp_c = fgetc(infile);
            *line = temp_c;
            line++;
            length++;
            if (length == max_length) break;
       } while (temp_c != '\n' && temp_c != EOF);
       *line = '\0';

       if (length == max_length) {
            do {
                 temp_c = fgetc(infile);
            } while (temp_c != '\n' && temp_c != EOF );
       }
}
