/*
FILE:     assign_chain_ids.C
*/
/*
VERSION:  8.125
*/
/*
DATE:     3/1/2013
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2013 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <stdio.h>
#include <stdlib.h>

#include "_index.h"
#include "_validation.h"
#include "gridcell.h"
#include "BTreeObj.h"
#include "util.h"
#include "xtal.h"

#define N_CYCLE  4 // 6

static double lower_bond[N_CYCLE] = { 1.2, 3.5,  5.0, 10.0 /* , 20,  50 */ };
static double upper_bond[N_CYCLE] = { 3.5, 5.0, 10.0, 20.0 /* , 50, 100 */ };
static double lower_bond_water_a[N_CYCLE] = { 0.01, 3.5, 5.0, 10.0 /* , 20, 50 */ };
static double lower_bond_water_s[N_CYCLE] = { 2.2,  3.5, 5.0, 10.0 /* , 20, 50 */ };

class _RIndex
{
   public:
      _RIndex();
      _RESIDUE *residue();
      int  num();
      void set_residue(_RESIDUE *a);
      void set_num(const int a);
      void operator=(const _RIndex &index);
      int  operator<(const _RIndex &index);
      int  operator==(const _RIndex &index);
      void DeleteElement() {};
      void Print() {};
   private:
      _RESIDUE *_residue;
      int  _num;
};

_RIndex::_RIndex()
{
       _residue = NULL;
       _num = -1;
}

_RESIDUE* _RIndex::residue() { return _residue; }
int _RIndex::num()     { return _num; }

void _RIndex::set_residue(_RESIDUE *a) { _residue = a; }
void _RIndex::set_num(const int a)     { _num = a; }

void _RIndex::operator=(const _RIndex &index)
{
       _residue = index._residue;
       _num = index._num;
}

int _RIndex::operator<(const _RIndex &index)
{
       if (_num < index._num)
            return 1;
       else return 0;
}

int _RIndex::operator==(const _RIndex &index)
{
       return (_num == index._num);
}

class _VIndex
{
   public:
      _VIndex();
      char *value();
      int  num();
      void set_value(const char *a);
      void set_num(const int a);
      void operator=(const _VIndex &index);
      int  operator<(const _VIndex &index);
      int  operator==(const _VIndex &index);
      void DeleteElement() {};
      void Print() {};
   private:
      char _value[50];
      int  _num;
};

_VIndex::_VIndex()
{
       memset(_value, 0, 50);
       _num = -1;
}

char* _VIndex::value() { return _value; }
int _VIndex::num()     { return _num; }

void _VIndex::set_value(const char *a) { strcpy(_value, a); }
void _VIndex::set_num(const int a)     { _num = a; }

void _VIndex::operator=(const _VIndex &index)
{
       strcpy(_value, index._value);
       _num = index._num;
}

int _VIndex::operator<(const _VIndex &index)
{
       if (_num < index._num)
            return 1;
       else return 0;
}

int _VIndex::operator==(const _VIndex &index)
{
       return (_num == index._num);
}

typedef struct {
       CifString ChainID;
       int count;
} _CHAINID_COUNT;

void _VALIDATION::assign_chain_id_to_non_polymer(const int assign_flag)
{
       if (assign_flag == FORCE_SKIP) return;
       if (!_num_mol || !_molecules) return;

       int need_assign = 0;
       if (assign_flag < 0) {
            need_assign = found_unassigned_chain(stdout, 3);
            if (need_assign)
                 fprintf(stdout, "Ignore -skip_chain_id_assignment option.\n");
            else return;
       }

       if (!need_assign) {
            if (_assign_chain_id_flag || assign_flag > 0) need_assign = 1;
            if (!need_assign) need_assign = found_unassigned_chain(NULL, 0);
       }
       if (!need_assign) return;

       int water_only = 0;
       if (assign_flag == 1) {
            water_only = 1;
            need_assign = found_unassigned_chain(stdout, NON_POLYMER_CHAIN);
            if (need_assign & NON_POLYMER_CHAIN) {
                 fprintf(stdout, "Ignore -water_chain_id_assignment_only option.\n");
                 water_only = 0;
            }
       }

       _ATOM *atom = NULL;
       _CHAIN *chain = NULL;
       _MOLECULE *mol = _molecules[0];
       _RESIDUE *residue = NULL;

       int num_seq = 0;
       for (int i = 0; i < _Num_Seqs; i++) num_seq++;

       int num_polymer = 0;
       char chain_id[20];
       memset(chain_id, 0, 20);
       chain = mol->GetFirstChain();
       while (chain) {
            if (chain->chain_type() == ATOMP_TOKEN ||
                chain->chain_type() == ATOMN_TOKEN /* ||
                chain->chain_type() == ATOMS_TOKEN &&
                chain->TotalResidueNumbers() > 1 */ ) {
                 num_polymer++;
                 residue = chain->GetFirstResidue();
                 strcpy(chain_id, residue->GetFirstAtom()->pdb_chnid());
            }
            chain = mol->GetNextChain();
       }

       if (num_polymer == 0) return;

       if (num_polymer == 1) {
            if (!strcmp(chain_id, "")) strcpy(chain_id, "A");
            if (num_seq == 1) _Seqs[0]->ChainId = chain_id;
            for (int i = 0; i < _num_mol; i++) {
                 chain = _molecules[i]->GetFirstChain();
                 while (chain) {
                      chain->set_ChainID(chain_id);
                      chain->set_PDB_ChainID(chain_id);
                      residue = chain->GetFirstResidue();
                      while (residue) {
                           atom = residue->GetFirstAtom();
                           while (atom) {
                                atom->set_chnid(chain_id);
                                atom->set_pdb_chnid(chain_id);
                                atom = residue->GetNextAtom();
                           }
                           residue = chain->GetNextResidue();
                      }
                      chain = _molecules[i]->GetNextChain();
                 }
                 _molecules[i]->merge_waters(water_only);
                 _re_assign_pdb_number = 1;
                 _molecules[i]->reorder_chain();
            }
            return;
       }

       int ret = 1, ret1 = -1, prev_atom = -1, curr_atom = -1, changed = 0;
       char chain_id1[20], buffer[20];
       double dist = 100.0;
       TblIndexObj PolymerChainIds;
       CifString cifstring;
       xtal *mycrys = NULL;
       CONTACT **contact = NULL;
       int n_contact = 0;
       BTreeObj<_RIndex> _AllWaters, _tlsWaters;
       _RIndex _rindex;
       BTreeObj<_IIndex> _FindWaters;
       _IIndex _iindex;
       BTreeObj<_VIndex> _AssignWaters;
       _VIndex _vindex;
       ReVarCifArray<int> polymer_chains, nonpolymer_chains, water_chains;
       ReVarCifArray<_CHAINID_COUNT> _chainid_count_array;
       _CHAINID_COUNT _chainid_count;
       for (int i = 0; i < _num_mol; i++) {
            PolymerChainIds.Clear();
            polymer_chains.Clear();
            nonpolymer_chains.Clear();
            water_chains.Clear();
            chain = _molecules[i]->GetFirstChain();
            while (chain) {
                 if (chain->chain_type() == ATOMP_TOKEN ||
                     chain->chain_type() == ATOMN_TOKEN /* ||
                     chain->chain_type() == ATOMS_TOKEN &&
                     chain->TotalResidueNumbers() > 1 */ ) {
                      polymer_chains.Add(chain->index());
                      cifstring.Copy(chain->PDB_ChainID());
                      if (PolymerChainIds.Seek(cifstring) < 0) {
                           PolymerChainIds.Add(cifstring);
                      }
                 } else {
                      residue = chain->GetFirstResidue();
                      if (!strcmp(residue->ResName(), "HOH") ||
                          !strcmp(residue->ResName(), "DOD"))
                           water_chains.Add(chain->index());
                      else nonpolymer_chains.Add(chain->index());
                 }
                 chain = _molecules[i]->GetNextChain();
            }
            if (polymer_chains.Length() == 0) continue;
/*
            if (polymer_chains.Length() == 0) {
                 PolymerChainIds.Clear();
                 polymer_chains.Clear();
                 nonpolymer_chains.Clear();
                 water_chains.Clear();
                 chain = _molecules[i]->GetFirstChain();
                 while (chain) {
                      if (strcmp(chain->PDB_ChainID(), "")) {
                           polymer_chains.Add(chain->index());
                           cifstring.Copy(chain->PDB_ChainID());
                           if (PolymerChainIds.Seek(cifstring) < 0) {
                                PolymerChainIds.Add(cifstring);
                           }
                      } else {
                           residue = chain->GetFirstResidue();
                           if (!strcmp(residue->ResName(), "HOH") ||
                               !strcmp(residue->ResName(), "DOD"))
                                water_chains.Add(chain->index());
                           else nonpolymer_chains.Add(chain->index());
                      }
                      chain = _molecules[i]->GetNextChain();
                 }
            }
*/
            if (nonpolymer_chains.Length() == 0 &&
                water_chains.Length() == 0) continue;

            if (!water_only) {
                 mycrys = new xtal;
                 mycrys->setorth('c');
                 if (mycrys->init(_cell, _molecules[i]) <= 0) {
                      delete mycrys;
                      continue;
                 }
                 mycrys->getSymmAtoms();
                 mycrys->moveAllAtomsToCenter();
                 mycrys->placeAtomsInGrid();

                 for (unsigned int j = 0; j < nonpolymer_chains.Length(); j++) {
                      float minimum_distance = 10000.0;
                      int minimum_index = -1;
                      int maximum_contacts = 0;
                      int maximum_index = -1;
                      for (int m = 0; m < N_CYCLE; m++) {
                           for (unsigned int k = 0; k < polymer_chains.Length(); k++) {
                                memset(chain_id, 0, 20);
                                memset(chain_id1, 0, 20);
                                sprintf(chain_id, "chain %d", nonpolymer_chains[j]);
                                sprintf(chain_id1, "chain %d", polymer_chains[k]);
                                contact = get_contact(*mycrys, n_contact, lower_bond[m],
                                   upper_bond[m], "asym", chain_id, chain_id1, 1, 0);
                                if (contact) {
                                     if (n_contact > maximum_contacts) {
                                          maximum_contacts = n_contact;
                                          maximum_index = k;
                                     }
                                     for (int l = 0; l < n_contact; l++) {
                                          if (contact[l]->dist < minimum_distance) {
                                               minimum_distance = contact[l]->dist;
                                               minimum_index = k;
                                          }
                                          delete contact[l];
                                     }
                                     delete [] contact;
                                }
                           }
                           if (minimum_index >= 0 || maximum_index >= 0) break;
                      }
                      if (minimum_distance < 1.75) {}
                      else if (maximum_index >= 0) {
                           minimum_index = maximum_index;
                      }
                      if (minimum_index >= 0) {
                           chain = _molecules[i]->GetIndexChain(nonpolymer_chains[j]);
                           cifstring.Copy(chain->PDB_ChainID());
                           int flag = 1;
                           if (__need_file_version_update &&
                               PolymerChainIds.Seek(cifstring) >= 0) flag = 0;
                           if (flag) {
                                _CHAIN *chain1 = _molecules[i]->
                                          GetIndexChain(polymer_chains[minimum_index]);
                                chain->set_ChainID(chain1->PDB_ChainID());
                                chain->set_PDB_ChainID(chain1->PDB_ChainID());
                                residue = chain->GetFirstResidue();
                                while (residue) {
                                     atom = residue->GetFirstAtom();
                                     while (atom) {
                                          atom->set_chnid(chain1->PDB_ChainID());
                                          atom->set_pdb_chnid(chain1->PDB_ChainID());
                                          atom = residue->GetNextAtom();
                                     }
                                     residue = chain->GetNextResidue();
                                }
                           }
                      }
                 }
                 delete mycrys;

                 _re_assign_pdb_number = 1;
            }

            if (water_chains.Length() == 0) {
                 if (_re_assign_pdb_number) _molecules[i]->reorder_chain();
                 continue;
            }

            mycrys = new xtal;
            mycrys->setorth('c');
            mycrys->init_water(_cell, _molecules[i]);
            mycrys->getSymmAtoms();
            mycrys->moveAllAtomsToCenter();
            mycrys->placeAtomsInGrid();
 
            chain = _molecules[i]->GetFirstChain();
            _AllWaters.Clear();
            _tlsWaters.Clear();
            int tls_group = 0;
            while (chain) {
                 if (chain->chain_type() != HETAS_TOKEN) {
                      chain = _molecules[i]->GetNextChain();
                      continue;
                 }
                 residue = chain->GetFirstResidue();
                 if (residue && strcmp(residue->ResName(), "HOH") &&
                     strcmp(residue->ResName(), "DOD")) {
                      chain = _molecules[i]->GetNextChain();
                      continue;
                 }
                 while (residue) {
                      if (residue->tls_group()) {
                           if (residue->tls_group() > tls_group) {
                                tls_group = residue->tls_group();
                           }
                           _rindex.set_residue(residue);
                           _rindex.set_num(residue->tls_group());
                           _tlsWaters.Add(_rindex);
                      }
                      atom = residue->GetFirstAtom();
                      while (atom) {
                           if (!strcmp(atom->atom_type(), "O")) {
                                 _rindex.set_residue(residue);
                                 _rindex.set_num(atom->atnum());
                                 _AllWaters.Add(_rindex);
                                 break;
                           }
                           atom = residue->GetNextAtom();
                      }
                      residue = chain->GetNextResidue();
                 }
                 chain = _molecules[i]->GetNextChain();
            }

            _AssignWaters.Clear(); 
            for (int m = 0; m < N_CYCLE; m++) {
                 contact = get_contact(*mycrys, n_contact, lower_bond_water_a[m],
                        upper_bond[m], "asym", "kind hetas", "atnum 1 -1", 1, 1);
                 _FindWaters.Clear();
                 for (int j = 0; j < n_contact; j++) {
                      if ((!strcmp(contact[j]->b_resnam, "HOH") ||
                           !strcmp(contact[j]->b_resnam, "DOD")) &&
                            strcmp(contact[j]->a_resnam, "HOH") &&
                            strcmp(contact[j]->a_resnam, "DOD")) {
                           strcpy(buffer, contact[j]->a_type);
                           strcpy(contact[j]->a_type, contact[j]->b_type);
                           strcpy(contact[j]->b_type, buffer);

                           strcpy(buffer, contact[j]->a_resnam);
                           strcpy(contact[j]->a_resnam, contact[j]->b_resnam);
                           strcpy(contact[j]->b_resnam, buffer);

                           strcpy(buffer, contact[j]->a_chnid);
                           strcpy(contact[j]->a_chnid, contact[j]->b_chnid);
                           strcpy(contact[j]->b_chnid, buffer);

                           int num = contact[j]->a_atomnum;
                           contact[j]->a_atomnum = contact[j]->b_atomnum;
                           contact[j]->b_atomnum = num;
                      }
                      if (strcmp(contact[j]->b_resnam, "HOH") &&
                          strcmp(contact[j]->b_resnam, "DOD")) {
                           _iindex.set_value(contact[j]->a_atomnum);
                           _iindex.set_pos(j);
                           _FindWaters.Add(_iindex);
                      }
                 }
                 if (_FindWaters.Root() >= 0) {
                      _FindWaters.GoFirst();
                      ret = 1;
                      prev_atom = -1;
                      curr_atom = -1;
                      dist = 100.0;
                      memset(chain_id1, 0, 20);
                      while (ret >= 0) {
                           _iindex = _FindWaters.GetCurrData();
                           curr_atom = _iindex.value();
                           if (curr_atom != prev_atom) {
                                if (prev_atom != -1) {
                                     _rindex.set_num(prev_atom);
                                     ret1 = _AllWaters.Seek(_rindex);
                                     if (ret1 >= 0) {
                                          _rindex = _AllWaters.GetCurrData();
                                          residue = _rindex.residue();
                                          atom = residue->GetFirstAtom();
                                          while (atom) {
                                               atom->set_chnid(chain_id1);
                                               atom->set_pdb_chnid(chain_id1);
                                               atom = residue->GetNextAtom();
                                          }
                                          _vindex.set_num(prev_atom);
                                          _vindex.set_value(chain_id1);
                                          if (_AssignWaters.Seek(_vindex) < 0) {
                                               _AssignWaters.Add(_vindex);
                                          }
                                          _AllWaters.Delete();
                                     }
                                }
                                dist = contact[_iindex.pos()]->dist;
                                strcpy(chain_id1, contact[_iindex.pos()]->b_chnid);
                                prev_atom = curr_atom;
                           } else if (contact[_iindex.pos()]->dist < dist) {
                                dist = contact[_iindex.pos()]->dist;
                                strcpy(chain_id1, contact[_iindex.pos()]->b_chnid);
                           }
                           ret = _FindWaters.GoToNext();
                      }
                      if (prev_atom != -1) {
                           _rindex.set_num(prev_atom);
                           ret1 = _AllWaters.Seek(_rindex);
                           if (ret1 >= 0) {
                                _rindex = _AllWaters.GetCurrData();
                                residue = _rindex.residue();
                                atom = residue->GetFirstAtom();
                                while (atom) {
                                     atom->set_chnid(chain_id1);
                                     atom->set_pdb_chnid(chain_id1);
                                     atom = residue->GetNextAtom();
                                }
                                _vindex.set_num(prev_atom);
                                _vindex.set_value(chain_id1);
                                if (_AssignWaters.Seek(_vindex) < 0) {
                                     _AssignWaters.Add(_vindex);
                                }
                                _AllWaters.Delete();
                           }
                      }
                 }
                 if (contact) {
                      for (int j = 0; j < n_contact; j++) {
                           if (contact[j]) delete contact[j];
                      }
                      delete [] contact; contact = NULL;
                 }
                 if (tls_group && _FindWaters.Root() >= 0) {
                      for (int j = 1; j <= tls_group; j++) {
                           _chainid_count_array.Clear();
                           ret = 1;
                           _tlsWaters.GoFirst();
                           while (ret >= 0) {
                                _rindex = _tlsWaters.GetCurrData();
                                if (_rindex.num() != j) {
                                     ret = _tlsWaters.GoToNext();
                                     continue;
                                }
                                residue = _rindex.residue();
                                atom = residue->GetFirstAtom();
                                while (atom) {
                                     _iindex.set_value(atom->atnum());
                                     if (_FindWaters.Seek(_iindex) >= 0) {
                                          int found = 0;
                                          for (unsigned int k = 0; k <
                                                _chainid_count_array.Length(); k++) {
                                               if (_chainid_count_array[k].ChainID ==
                                                    atom->chnid()) {
                                                    _chainid_count_array[k].count++;
                                                    found = 1;
                                                    break;
                                               }
                                          }
                                          if (!found) {
                                               _chainid_count.ChainID.Copy(atom->chnid());
                                               _chainid_count.count = 1;
                                               _chainid_count_array.Add(_chainid_count);
                                          }
                                          break;
                                     }
                                     atom = residue->GetNextAtom();
                                }
                                ret = _tlsWaters.GoToNext();
                           }
                           if (_chainid_count_array.Length() == 0) continue;
                           ret1 = 0;
                           cifstring.Clear();
                           for (unsigned int k = 0; k < _chainid_count_array.Length(); k++) {
                                if (_chainid_count_array[k].count > ret1) {
                                     ret1 = _chainid_count_array[k].count;
                                     cifstring.Copy(_chainid_count_array[k].ChainID);
                                }
                           }
                           ret = 1;
                           _tlsWaters.GoFirst();
                           while (ret >= 0) {
                                _rindex = _tlsWaters.GetCurrData();
                                if (_rindex.num() != j) {
                                     ret = _tlsWaters.GoToNext();
                                     continue;
                                }
                                residue = _rindex.residue();
                                atom = residue->GetFirstAtom();
                                while (atom) {
                                     atom->set_chnid(cifstring.Text());
                                     atom->set_pdb_chnid(cifstring.Text());
                                     if (!strcmp(atom->atom_type(), "O")) {
                                          _rindex.set_num(atom->atnum());
                                          ret1 = _AllWaters.Seek(_rindex);
                                          if (ret1 >= 0) _AllWaters.Delete();
                                     }
                                     atom = residue->GetNextAtom();
                                }
                                ret = _tlsWaters.GoToNext();
                           }
                      }
                 }
                 if (_AllWaters.Root() < 0) break;

                 contact = get_contact(*mycrys, n_contact, lower_bond_water_s[m],
                        upper_bond[m], "asym", "kind hetas", "kind hetas", 0, 1);
                 if (contact) {
                      while (1) {
                           changed = 0;
                           prev_atom = -1;
                           curr_atom = -1;
                           dist = 100.0;
                           memset(chain_id1, 0, 20);
                           for (int j = 0; j < n_contact; j++) {
                                if (prev_atom != contact[j]->a_atomnum) {
                                     if (curr_atom >= 0) {
                                          _rindex.set_num(curr_atom);
                                          ret1 = _AllWaters.Seek(_rindex);
                                          if (ret1 >= 0) {
                                               _rindex = _AllWaters.GetCurrData();
                                               residue = _rindex.residue();
                                               atom = residue->GetFirstAtom();
                                               while (atom) {
                                                    atom->set_chnid(chain_id1);
                                                    atom->set_pdb_chnid(chain_id1);
                                                    atom = residue->GetNextAtom();
                                               }
                                               _vindex.set_num(curr_atom);
                                               _vindex.set_value(chain_id1);
                                               if (_AssignWaters.Seek(_vindex) < 0) {
                                                    _AssignWaters.Add(_vindex);
                                               }
                                               _AllWaters.Delete();
                                               changed = 1;
                                          }
                                     }
                                     curr_atom = -1;
                                     dist = 100.0;
                                     memset(chain_id1, 0, 20);
                                }
                                _vindex.set_num(contact[j]->a_atomnum);
                                ret = _AssignWaters.Seek(_vindex);
                                _vindex.set_num(contact[j]->b_atomnum);
                                ret1 = _AssignWaters.Seek(_vindex);
                                if (ret < 0 && ret1 >= 0) {
                                     if (contact[j]->dist < dist) {
                                          dist = contact[j]->dist;
                                          curr_atom = contact[j]->a_atomnum;
                                          _vindex = _AssignWaters.GetCurrData();
                                          strcpy(chain_id1, _vindex.value());
                                     }
                                }
                                prev_atom = contact[j]->a_atomnum;
                           }
                           if (curr_atom >= 0) {
                                _rindex.set_num(curr_atom);
                                ret1 = _AllWaters.Seek(_rindex);
                                if (ret1 >= 0) {
                                     _rindex = _AllWaters.GetCurrData();
                                     residue = _rindex.residue();
                                     atom = residue->GetFirstAtom();
                                     while (atom) {
                                          atom->set_chnid(chain_id1);
                                          atom->set_pdb_chnid(chain_id1);
                                          atom = residue->GetNextAtom();
                                     }
                                     _vindex.set_num(curr_atom);
                                     _vindex.set_value(chain_id1);
                                     if (_AssignWaters.Seek(_vindex) < 0) {
                                          _AssignWaters.Add(_vindex);
                                     }
                                     _AllWaters.Delete();
                                     changed = 1;
                                }
                           }
                           if (!changed) break;
                           if (_AllWaters.Root() < 0) break;
                      }
                      for (int j = 0; j < n_contact; j++) {
                           if (contact[j]) delete contact[j];
                      }
                      delete [] contact; contact = NULL;
                 }
                 if (_AllWaters.Root() < 0) break;
            }

            delete mycrys;

            _molecules[i]->relocate_waters();
            _molecules[i]->merge_waters(water_only);
            _molecules[i]->reorder_chain();

            _re_assign_pdb_number = 1;
       }
}

void _VALIDATION::merge_split_waters()
{
       for (int i = 0; i < _num_mol; i++) {
            _molecules[i]->merge_split_waters();
       }
}
