/*
FILE:     check.C
*/
/*
VERSION:  8.125
*/
/*
DATE:     3/1/2013
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2013 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <sys/stat.h>

#include <map>
#include <set>
#include <string>

#include "_validation.h"
#include "_seqnum.h"
#include "align_standard.h"
#include "BTreeObj.h"
#include "check_global.h"
#include "datacheck.h"
#include "DICFileObj.h"
#include "erf_util.h"
#include "misc_util.h"
#include "sg_center.h"
#include "util.h"

#define  NS               10
// #define  EC_PATTERN       "[1-9][0-9]*(\\.([1-9][0-9]*)|[-]){3}"
#define  EC_PATTERN "([1-6]((.[1-9][0-9]?)|(.-))((.[1-9][0-9]?)|(.-))((.[1-9][0-9]?[0-9]?)|(.-)))(([ ]*,[ ]*)([1-6]((.[1-9][0-9]?)|(.-))((.[1-9][0-9]?)|(.-))((.[1-9][0-9]?[0-9]?)|(.-))))*"
#define  dictxraybinfile  "mmcif_xray.odb"
#define  dictnmrbinfile   "mmcif_nmr.odb"
#define  taxonomyfile     "taxonomy.txt" 

static TblIndexObj TaxnomyList;

static const char *source_category[3] = { "entity_src_gen", "entity_src_nat", "rcsb_entity_src_syn" };
static const char *range_category[2] = { "struct_ncs_dom_lim", "ccp4_refine_tls_group" };
static const char *range_items[2][2] = { { "beg_auth_asym_id", "beg_auth_seq_id" },
                                         { "end_auth_asym_id", "end_auth_seq_id" } };

static void check_tokens(_MOLECULE *mol, Cards_Info_Format *CardInfo,
                 const int Num_Token, CHECK_TOKEN_FORMAT *Check_Format,
                 const int skip, const char *file_name, FILE *fp);
static void check_categorys(_MOLECULE *mol, ISTable *Table, CHECK_CATEGORY *category,
                 FILE *fp, const char *file_name, int &count);
/*
static void check_categorys1(_MOLECULE *mol, ISTable *Table, CHECK_CATEGORY *category,
                 FILE *fp, const char *file_name, int &count);
*/
static void check_seqnum_redundancy(_MOLECULE *mol, const char *infile_name, FILE *fp, int &count);
static int get_wavelength_list(const char *list);
static int check_taxonomy(const char *name);
static char *get_word1(char *string, char *buffer);
static bool check_header_failed(const CifString& header, DICFileObj *dict);
static void split_head(ReVarCifArray<CifString>& enumValues, const char *delimit);
static void check_source(FILE* fp, CifString& spell, ISTable* Table, const int& row,
                         const int& source_index, const int& tax_index);
static void check_missing_source(FILE* fp, int& error, CifFileObj* fobj, const char*
                                 StructureId, const CifString& id, const char* category,
                                 const char* source_item, const char* tax_item);
static void get_wordlist_from_string_separated_by_delimits(ReVarCifArray<CifString> &wordlist,
                  char *string_buff, const char *delimit);

void get_taxonomylist(const int taxonomy_flag)
{
       char filename[300], line[300];
       CifString cifstring;
       FILE *fp = NULL;
       struct stat stattext;

       TaxnomyList.Clear();
       if (!taxonomy_flag) return;
       sprintf(filename, "%s/data/ascii/%s", root_dir, taxonomyfile);
       if (stat(filename, &stattext) == 0) {
            fp = fopen(filename, "r");
            while (!feof(fp)) {
                 fgets(line, 299, fp);
                 if (feof(fp)) break;
                 line[strlen(line)-1] = '\0';
                 cifstring.Copy(line);
                 TaxnomyList.Add(cifstring);
            }
            fclose (fp);
       }
}

void get_EnumerationValue(const char *itemName, DICFileObj *ref,
                          ReVarCifArray<CifString>& enumValues)
{
       CifString cs;
       cs.Copy(itemName);
       get_EnumerationValue(cs, ref, enumValues);
}

void get_EnumerationValue(const CifString& itemName, DICFileObj *ref,
                          ReVarCifArray<CifString>& enumValues)
{
       enumValues.Clear();

       ISTable *Item = ref->GetTablePtr("item_enumeration");
       int ItemIndex = Item->GetColumnIndex("value");
       ISTable *NdbItem = ref->GetTablePtr("ndb_item_enumeration");
       int NdbItemIndex = NdbItem->GetColumnIndex("value");

       int errCode = -1;
       CifString cell;
       ReVarCifArray<CifString> target, nameList;

       nameList.Clear();
       nameList.Add("name");
       target.Clear();
       target.Add(itemName);
       ReVarPCifArray<int> *OutList = Item->Search(target, nameList, errCode);
       if (errCode >= 0) {
            for (unsigned int l = 0; l < OutList->Length(); l++) {
                 Item->GetCell(cell, ItemIndex, (*OutList)[l]);
                 enumValues.Add(cell);
            }
       }
       if (OutList) delete (OutList);
       if (!enumValues.Length()) {
           OutList = NdbItem->Search(target, nameList, errCode);
            if (errCode >= 0) {
                 for (unsigned int l = 0; l < OutList->Length(); l++) {
                      NdbItem->GetCell(cell, NdbItemIndex, (*OutList)[l]);
                      enumValues.Add(cell);
                 }
            }
            if (OutList) delete (OutList);
       }
}

void _VALIDATION::write_check_results(FILE *fp, const int plain_error_flag, const int short_version)
{
       char buffer[100];
       CifString annotator, all_error_messages;

       annotator.Clear();
       bool has_CAVEAT_record = false;
       if (_CifObj) {
            ISTable *Table = _CifObj->GetTablePtr(_StructureId, "ndb_database_status");
            if (!is_empty_table(Table)) {
                 int index = Table->GetColumnIndex("rcsb_annotator");
                 get_value(annotator, Table, index, 0);
            }
            Table = _CifObj->GetTablePtr(_StructureId, "database_PDB_caveat");
            if (!is_empty_table(Table)) {
                 int index = Table->GetColumnIndex("text");
                 get_value(all_error_messages, Table, index, 0);
                 if (all_error_messages != "") has_CAVEAT_record = true;
            }
       }
       
       all_error_messages.Clear();

       if (short_version) {
            if (_Error_Check[3])  all_error_messages += "wrong terminal atom(s), ";
            if (_Error_Check[4])  all_error_messages += "sequence(s) coordinates mismatch, ";
            if (_Error_Check[5])  all_error_messages += "structural feature(s) coordinates mismatch, ";
            if (_Error_Check[6]) {
                 if (_Error_Check_Message[6] != "")
                      all_error_messages += "chirality error (O1P,O2P), ";
                 else if (!has_CAVEAT_record)
                      all_error_messages += "Chirality error without CAVEAT record, ";
                 else all_error_messages += "chirality error, ";
            }
            if (_Error_Check[11]) all_error_messages += "PDB ID mismatch, ";
            if (_Error_Check[13]) all_error_messages += "PDB Chain ID mismatch, ";
            if (_Error_Check[20]) {
                 all_error_messages += "link record(distance";
                 all_error_messages += _Error_Check_Message[20];
                 all_error_messages += "), ";
            }
            if (_Error_Check[21]) all_error_messages += "extra atom(s), ";
            if (_Error_Check[24]) all_error_messages += "multiple source information, ";
            if (_Error_Check[25]) all_error_messages += "atom type error, ";
            if (_Error_Check[26]) all_error_messages += "occupancy error, ";
            if (_Error_Check[27]) all_error_messages += "same coordinates, ";
            if (_error_message != "") {
                 if (strstr(_error_message.Text(), "ATOM") ||
                     strstr(_error_message.Text(), "HETATM")) {
                      all_error_messages += "alternate conformation, ";
                 }
            }
            if (_unrecognized_residue != "") {
                 all_error_messages += "Unrecognized residue(s): ";
                 all_error_messages += _unrecognized_residue;
                 all_error_messages += ", ";
            }
            if (_Error_Check[36]) all_error_messages += "link record(other), ";
            if (_Error_Check[37]) all_error_messages += "link record(alt. conformation), ";
            if (_Error_Check[59]) all_error_messages += "covalent bonds involved hydrogens, ";
            if (_Error_Check[62]) all_error_messages += "lower case insertion code, ";
            if (_Error_Check[64]) all_error_messages += "missing TLS information, ";
            if (_Error_Check[65]) all_error_messages += "short sequence, ";
            if (_Error_Check[66]) all_error_messages += "obsolete/split residue, ";
            if (_Error_Check[67]) {
                 all_error_messages += "terminal residue (";
                 all_error_messages += _Error_Check_Message[67];
                 all_error_messages += "), ";
            }
            if (_Error_Check[68]) {
                 all_error_messages += "dissociated residue (";
                 all_error_messages += _Error_Check_Message[68];
                 all_error_messages += "), ";
            }
            if (_Error_Check[69]) {
                 all_error_messages += "EC number (";
                 all_error_messages += _Error_Check_Message[69];
                 all_error_messages += "), ";
            }
            if (_Error_Check[70]) all_error_messages += "merge peptide inhibitor, ";
            if (_Error_Check[71]) all_error_messages += "PDB residue numbering is larger than 9999, "; 
            if (_Error_Check[72]) all_error_messages += "site record, ";
            if (_Error_Check[73]) all_error_messages += "water far away, ";
            if (_Error_Check[74]) {
                 all_error_messages += "Atom number ";
                 all_error_messages += _Error_Check[74];
                 all_error_messages += " exceeds 99999, ";
            }
            if (_Error_Check[75]) {
                 if (_Error_Check_Message[75] != "")
                      all_error_messages += "HEADER " + _Error_Check_Message[75] + ", ";
                 else all_error_messages += "missing HEADER, ";
            }
            if (_Error_Check[77]) all_error_messages += "REF_ONLY residue, ";
            // if (_Error_Check[79]) all_error_messages += "remediation VERSN REVDAT (13-JUL-11), ";

            if (all_error_messages != "")
                 fprintf(fp, "%s %s\n", _input_filename, all_error_messages.Text());
       } else {
            if (!plain_error_flag) {
                 fprintf(fp, "%-11s  %-12s ", annotator.Text(), _StructureId);
            }

            if (_Error_Check[0]) {
                 sprintf(buffer, "%d", _Error_Check[0]);
                 all_error_messages += buffer;
                 all_error_messages += " close contact(s), ";
            }
            if (_Error_Check[1]) {
                 sprintf(buffer, "%d", _Error_Check[1]);
                 all_error_messages += buffer;
                 all_error_messages += " bad bond distance(s), ";
            }
            if (_Error_Check[2]) {
                 sprintf(buffer, "%d", _Error_Check[2]);
                 all_error_messages += buffer;
                 all_error_messages += " bad bond angles(s), ";
            }
            if (_Error_Check[3])  all_error_messages += "wrong terminal atom(s), ";
            if (_Error_Check[4])  all_error_messages += "sequence(s) coordinates mismatch, ";
            if (_Error_Check[5])  all_error_messages += "structural feature(s) coordinates mismatch, ";
            if (_Error_Check[6]) {
                 if (_Error_Check_Message[6] != "")
                      all_error_messages += "chirality error (O1P,O2P), ";
                 else if (!has_CAVEAT_record)
                      all_error_messages += "Chirality error without CAVEAT record, ";
                 else all_error_messages += "chirality error, ";
            }
            if (_Error_Check[7] || _Error_Check[52])  all_error_messages += "space group and/or cell dimension error, ";
            if (_Error_Check[8])  all_error_messages += "Z-value error, ";
            if (_Error_Check[9])  all_error_messages += "missing keywords, ";
            if (_Error_Check[10]) all_error_messages += "remark 3 or 200 error, ";
            if (_Error_Check[11]) all_error_messages += "PDB ID mismatch, ";
            if (_Error_Check[12]) all_error_messages += "missing source information, ";
            if (_Error_Check[13]) all_error_messages += "PDB Chain ID mismatch, ";
            if (_Error_Check[14]) all_error_messages += "DBREF error, ";
            if (_Error_Check[15]) all_error_messages += "SEQADV error, ";
            if (_Error_Check[16]) all_error_messages += "missing db_name or db_id in rcsb_database_related category, ";
            if (_Error_Check[17]) all_error_messages += "citation, ";
            if (_Error_Check[18]) all_error_messages += "PDB chain ID and number redundancy, ";
            if (_Error_Check[19]) {
                 if (_Error_Check_Message[19] != "") {
                      all_error_messages += "sequence link (";
                      all_error_messages += _Error_Check_Message[19];
                      all_error_messages += "), ";
                 } else all_error_messages += "sequence link, ";
            }
            if (_Error_Check[20]) {
                 all_error_messages += "link record(distance";
                 all_error_messages += _Error_Check_Message[20];
                 all_error_messages += "), ";
            }
            if (_Error_Check[21]) all_error_messages += "extra atom(s), ";
            if (_Error_Check[22]) all_error_messages += "ndb_database_proc category, ";
            if (_Error_Check[23]) {
                 all_error_messages += "mmCIF dictionary compliance ( ";
                 if (_Error_Check[23] & 1)
                      all_error_messages += "missing mandatory item, ";
                 if (_Error_Check[23] & 2)
                      all_error_messages += "duplicate row, ";
                 if (_Error_Check[23] & 4)
                      all_error_messages += "data type error, ";
                 if (_Error_Check[23] & 8)
                      all_error_messages += "out of range, ";
                 if (_Error_Check[23] & 16)
                      all_error_messages += "not defined in enumeration list, ";
                 all_error_messages += "), ";
            }
            if (_Error_Check[24]) all_error_messages += "multiple source information, ";
            if (_Error_Check[25]) all_error_messages += "atom type error, ";
            if (_Error_Check[26]) all_error_messages += "occupancy error, ";
            if (_Error_Check[27]) all_error_messages += "same coordinates, ";
            if (_Error_Check[28]) all_error_messages += "missing TITLE, ";
            if (_Error_Check[29]) {
                 all_error_messages += "wrong date (";
                 all_error_messages += _date_error;
                 all_error_messages += "), ";
            }
            if (_Error_Check[30]) all_error_messages += "MAD experiment, ";
            if (_Error_Check[31]) all_error_messages += "wrong alternate location indicator, ";
            if (_Error_Check[32]) all_error_messages += "wrong wavelength, ";
            if (_Error_Check[33]) all_error_messages += "wrong insertion code, ";
            if (_Error_Check[34]) {
                 all_error_messages += "large cell dimensions (";
                 all_error_messages += _large_cell;
                 all_error_messages += "), ";
            }
            if (_Error_Check[35]) all_error_messages += "missing PubMed ID, ";
            if (_Error_Check[36]) all_error_messages += "link record(other), ";
            if (_Error_Check[37]) all_error_messages += "link record(alt. conformation), ";
            if (_Error_Check[38]) all_error_messages += "NDB ID problem for nucleic acid structure, ";
            if (_spell_check != "") {
                 all_error_messages += "misspelled name (";
                 all_error_messages += _spell_check;
                 all_error_messages += "), ";
            }
            if (_error_message != "") {
                 if (strstr(_error_message.Text(), "ATOM") ||
                     strstr(_error_message.Text(), "HETATM")) {
                      all_error_messages += "alternate conformation, ";
                 }
                 if (strstr(_error_message.Text(), "Error: Z value is")) {
                      all_error_messages += "Z-value error, ";
                 }
            }
            if (_unrecognized_residue != "") {
                 all_error_messages += "Unrecognized residue(s): ";
                 all_error_messages += _unrecognized_residue;
                 all_error_messages += ", ";
            }
            if (_Error_Check[39] || _Error_Check[40] || (_Error_Check[41] && _Error_Check[41] != 100)) {
                 all_error_messages += "Target ( ";
                 if (_Error_Check[39]) all_error_messages += "missing target ID, ";
                 if (_Error_Check[40]) all_error_messages += "target ID did not exist, ";
                 if (_Error_Check[41] && _Error_Check[41] != 100) {
                      all_error_messages += "Sequence Identity = ";
                      sprintf(buffer, "%d%% ", _Error_Check[41]);
                      all_error_messages += buffer;
                 }
                 all_error_messages += "), ";
            }
            if (_Error_Check[42]) all_error_messages += "scale matrices, ";
            if (_Error_Check[43]) all_error_messages += "residue range for TLS/NCS, ";

            if (_Error_Check[44]) all_error_messages += "all upper/lower case in citation.title, ";
            if (_Error_Check[45]) all_error_messages += "all upper/lower case in struct.title, ";
            if (_Error_Check[46]) all_error_messages += "all upper/lower case in citation_author.name, ";
            if (_Error_Check[47]) all_error_messages += "all upper/lower case in audit_author.name, ";

            if (_Error_Check[59]) all_error_messages += "covalent bonds involved hydrogens, ";
            if (_Error_Check[60]) {
                 all_error_messages += "Matthews coefficient & solvent content (";
                 all_error_messages += _Error_Check_Message[60];
                 all_error_messages += "), ";
            }
            if (_Error_Check[61]) {
                 all_error_messages += "crystallization temperature (";
                 all_error_messages += _Error_Check_Message[61];
                 all_error_messages += "), ";
            }
            if (_Error_Check[62]) all_error_messages += "lower case insertion code, ";
            if (_Error_Check[63]) {
                 all_error_messages += "atom(s) on special position with occupancies 1.0 (";
                 all_error_messages += _Error_Check_Message[63];
                 all_error_messages += "), ";
            }
            if (_Error_Check[64]) all_error_messages += "missing TLS information, ";
            if (_Error_Check[65]) all_error_messages += "short sequence, ";
            if (_Error_Check[66]) all_error_messages += "obsolete/split residue, ";
            if (_Error_Check[67]) {
                 all_error_messages += "terminal residue (";
                 all_error_messages += _Error_Check_Message[67];
                 all_error_messages += "), ";
            }
            if (_Error_Check[68]) {
                 all_error_messages += "dissociated residue (";
                 all_error_messages += _Error_Check_Message[68];
                 all_error_messages += "), ";
            }
            if (_Error_Check[69]) {
                 all_error_messages += "EC number (";
                 all_error_messages += _Error_Check_Message[69];
                 all_error_messages += "), ";
            }
            if (_Error_Check[70]) all_error_messages += "merge peptide inhibitor, ";
            if (_Error_Check[71]) all_error_messages += "PDB residue numbering is larger than 9999, "; 
            if (_Error_Check[72]) all_error_messages += "site record, ";
            if (_Error_Check[73]) all_error_messages += "water far away, ";
            if (_Error_Check[74]) {
                 all_error_messages += "Atom number ";
                 all_error_messages += _Error_Check[74];
                 all_error_messages += " exceeds 99999, ";
            }
            if (_Error_Check[75]) {
                 if (_Error_Check_Message[75] != "")
                      all_error_messages += "HEADER " + _Error_Check_Message[75] + ", ";
                 else all_error_messages += "missing HEADER, ";
            }
            if (_Error_Check[77]) all_error_messages += "REF_ONLY residue, ";
            // if (_Error_Check[79]) all_error_messages += "remediation VERSN REVDAT (13-JUL-11), ";

            if (!plain_error_flag) {
                 if (all_error_messages != "")
                      fprintf(fp, "%s\n", all_error_messages.Text());
                 else fprintf(fp, "OK\n");
            } else if (all_error_messages != "") {
                 char *p = all_error_messages.Text();
                 while (p) {
                      p = get_max_length_word(p, buffer, 80);
                      fprintf(fp, "%s\n", buffer);
                 }
            }
       }
       fflush(fp);
}

void _VALIDATION::ndb_check_file(FILE *fp, const int Num_Token, CHECK_TOKEN_FORMAT *Check_Format,
               const int Num_Category, CHECK_CATEGORY *Check_Category,
               const int skip, const int public_flag, const char *File_ID, FILE *fp1)
{
       int is_error = 0, KEYWDS_TOKEN, TITLE_TOKEN, DTTEMP_TOKEN, i = 0, j = 0, k = 0;
       int l = 0, m = 0, high_index = 0, low_index = 0, rowNo = 0, src_index = 0;
       int entity_index = 0, type_index = 0, e_index = 0, s_index = 0;
       int o_index = 0, rowNo1 = 0, is_idential = 1, index[9];
       int Mol_ID = -1, details, card_id, numRemark, id_idx, pubmed_idx;
       int ndb_pdb_id_code, mon_id, ndb_pdb_chain_id, seq_num, ndb_pdb_ins_code;
       int ndb_seq_db_name, ndb_seq_db_accession_code, db_mon_id, ndb_seq_db_seq_num;
       int ndb_auth_seq_num, ret, ret1, first, len, is_nmr = 0;
       int num, *length = NULL, min_length;
       char SpaceGroup[20], pdb_id[20], buffer[1000], binaryfile[300], *p = NULL;
       char date_original[20], today_date[20];
       FILE *fp2 = NULL;
       float d_alpha, d_beta, d_gamma;
       double value[9], svalue[9], rad, va, aa[3], ang[3], alphr[3], cosa[3], sina[3];
       double matthew = 0.0, solvent = 0.0;
       double **backbone = NULL, rt[12], ave_exyz[3], ave_sxyz[3], v, v0, v1, rms, rms1;
       struct stat stattext;
       BTreeObj<_SSIndex> _AsymIndex, _BiolUnit;
       _ATOM **atom = NULL, *atom1 = NULL;
       _CHAIN *chain = NULL;
       CifString string, string1, string2, string3, str[2], res_high, res_low, buff;
       ISTable *Table = NULL, *Table1 = NULL, *Table2 = NULL, *Table3 = NULL;
       One_Card_Info_Format *pCard = NULL;
       _RESIDUE *residue = NULL;
       _SSIndex idx, idx1, *pidx = NULL;
       TblIndexObj _UniIndex, _BiolIndex;
       ReVarCifArray<CifString> wordlist;

       if (_num_mol == 0) return;
       if (_input_format != NDB_FILE_FORMAT_PDB &&
           _input_format != NDB_FILE_FORMAT_MMCIF) return;

       DICFileObj *dict = NULL;
       if (_experiment_type & EXPERIMENT_TYPE_NMR ||
           _experiment_type & EXPERIMENT_TYPE_NMR_SOLID ||
           _experiment_type & EXPERIMENT_TYPE_MODEL) {
            sprintf(binaryfile, "%s/data/binary/%s", root_dir, dictnmrbinfile);
            is_nmr = 1;
       } else {
            sprintf(binaryfile, "%s/data/binary/%s", root_dir, dictxraybinfile);
            is_nmr = 0;
       }

       if (stat(binaryfile, &stattext) != -1) {
            dict = new DICFileObj(binaryfile, READ_MODE);
       }

       if (!_cell->is_artifical) {
            strcpy(SpaceGroup, _cell->space_group);
            is_error = 0;
            if (!strcmp(SpaceGroup, "P 1 2 1") ||
                !strcmp(SpaceGroup, "P 1 21 1") ||
                !strcmp(SpaceGroup, "C 1 2 1")) {
                 d_alpha = fabs(_cell->alpha - 90.0);
                 d_beta  = fabs(_cell->beta - 90.0);
                 d_gamma = fabs(_cell->gamma - 90.0);
                 if (d_alpha > 0.01 || d_beta < 0.01 || d_gamma > 0.01)
                      is_error = 1;
            } else if (!strcmp(SpaceGroup, "P 1 1 2") ||
                       !strcmp(SpaceGroup, "P 1 1 21") ||
                       !strcmp(SpaceGroup, "A 1 1 2")) {
                 d_alpha = fabs(_cell->alpha - 90.0);
                 d_beta  = fabs(_cell->beta - 90.0);
                 d_gamma = fabs(_cell->gamma - 90.0);
                 if (d_alpha > 0.01 || d_beta > 0.01 || d_gamma < 0.01)
                      is_error = 1;
            } else if (!strcmp(SpaceGroup, "H 3") ||
                       !strcmp(SpaceGroup, "H 3 2")) {
                 d_alpha = fabs(_cell->alpha - 90.0);
                 d_beta  = fabs(_cell->beta - 90.0);
                 d_gamma = fabs(_cell->gamma - 120.0);
                 if (d_alpha > 0.01 || d_beta > 0.01 || d_gamma > 0.01)
                      is_error = 1;
            } else if (!strcmp(SpaceGroup, "R 3") ||
                       !strcmp(SpaceGroup, "R 3 2")) {
                 d_alpha = fabs(_cell->alpha - _cell->beta);
                 d_beta  = fabs(_cell->beta - _cell->gamma);
                 if (d_alpha > 0.01 || d_beta > 0.01) is_error = 1;
            }
            if (is_error) {
                 // fprintf(fp, "%s: Error: Space Group", _input_filename);
                 fprintf(fp, "Error: Space Group '%s', alpha=%.2f beta=%.2f gamma=%.2f\n",
                      SpaceGroup, _cell->alpha, _cell->beta, _cell->gamma);
                 _Error_Check[7]++;
            }
            if ((_cell->a > 500 || _cell->b > 500 || _cell->c > 500) && !public_flag) {
                 fprintf(fp, "%s:  WARNING: Since one (or more) cell dimension is larger than 500 A.\n", _input_filename);
                 fprintf(fp, "               Please check the structure packing with ndbview.\n");
                 fprintf(fp, "               If there is large empty space, ask the author to confirm\n");
                 fprintf(fp, "               the accuracy of cell dimensions in the file.\n");
                 // if (_cell->a > 500) fprintf(fp, "a=%.3f ", _cell->a);
                 // if (_cell->b > 500) fprintf(fp, "b=%.3f ", _cell->b);
                 // if (_cell->c > 500) fprintf(fp, "c=%.3f ", _cell->c);
                 fprintf(fp, "\n");
                 if (_cell->a > 500) {
                      sprintf(buffer, " a=%.3f", _cell->a);
                      _large_cell += buffer;
                 }
                 if (_cell->b > 500) {
                      sprintf(buffer, " b=%.3f", _cell->b);
                      _large_cell += buffer;
                 }
                 if (_cell->c > 500) {
                      sprintf(buffer, " c=%.3f", _cell->c);
                      _large_cell += buffer;
                 }
                 _Error_Check[34]++;
            }
            if (!_cell->z) {
                 fprintf(fp, "%s: Error: Z value\n", _input_filename);
                 _Error_Check[8]++;
            }
       } else if (_Error_Check_Message[52] != "") {
            fprintf(fp, "%s\n", _Error_Check_Message[52].Text());
       }

       if (_error_message != "") fprintf(fp, "%s\n", _error_message.Text());
       if (_Error_Check_Message[70] != "") {
             fprintf(fp, "%s\n", _Error_Check_Message[70].Text());
       }

       if (_input_format == NDB_FILE_FORMAT_PDB ||
           _input_format == NDB_FILE_FORMAT_MMCIF &&
           _file_status == FILE_STATUS_MMCIF_PLUS_MOL) {
            check_tokens(_molecules[_rep_mol], _CardInfo, Num_Token, Check_Format,
                    skip, _input_filename, fp);
       } else {
            for (i = 0; i < Num_Category; i++) {
                 Table = _CifObj->GetTablePtr(_StructureId, Check_Category[i].category);
                 if (is_empty_table(Table)) continue;
                 check_categorys(_molecules[_rep_mol], Table, &(Check_Category[i]),
                             fp, _input_filename, _Error_Check[5]);
            }
/*
            for (i = 0; i < NUM_CHECK_CATEGORY1; i++) {
                 Table = _CifObj->GetTablePtr(_StructureId, check_category1[i].category);
                 if (is_empty_table(Table)) continue;
                 check_categorys1(_molecules[_rep_mol], Table, &(check_category1[i]),
                             fp, _input_filename, _Error_Check[43]);
            }
*/
       }

       for (i = 0; i < _num_mol; i++) {
            check_seqnum_redundancy(_molecules[i], _input_filename, fp, _Error_Check[18]);
            int atom_no = _molecules[i]->count_total_atom_number();
            if (atom_no > 99999) {
                 _Error_Check[74] = atom_no;
                 fprintf(fp, "%s: Total atom number %d exceeds 99999.\n\n",
                             _input_filename, atom_no);
            }
       }

       if (_input_format == NDB_FILE_FORMAT_PDB ||
           _input_format == NDB_FILE_FORMAT_NDB) {
            if (!public_flag) {
                 KEYWDS_TOKEN = find_card_id("KEYWDS");
                 pCard = _CardInfo[KEYWDS_TOKEN].Head();
                 if (pCard == NULL || !strcmp(pCard->Field(2), "")) {
                      _Error_Check[9]++;
                      fprintf(fp, "ERROR: Missing KEYWDS records\n\n");
                 }
                 DTTEMP_TOKEN = find_card_id("DTTEMP");
                 pCard = _CardInfo[DTTEMP_TOKEN].Head();
                 if (pCard && strcmp(pCard->Field(1), "")) {
                      if (atof(pCard->Field(1)) < 0)
                           fprintf(fp, "TEMPERATURE WARNING: %s\n", pCard->Field(1));
                 }
                 TITLE_TOKEN = find_card_id("TITLE");
                 pCard = _CardInfo[TITLE_TOKEN].Head();
                 if (pCard == NULL || !strcmp(pCard->Field(2), "")) {
                      fprintf(fp, "ERROR: Missing TITLE record\n\n");
                      _Error_Check[28]++;
                 }
            }
            if (fp1) {
                 char ResName[100];
                 TblIndexObj het_ids;
                 het_ids.Clear();
                 _CHAIN *chain = _molecules[_rep_mol]->GetFirstChain();
                 while (chain) {
                      if (chain->chain_type() == HETAS_TOKEN &&
                          (!strcmp(chain->SeqRes(0)->Field[0], "HOH") ||
                           !strcmp(chain->SeqRes(0)->Field[0], "DOD"))) {
                           chain = _molecules[_rep_mol]->GetNextChain();
                           continue;
                      }

                      for (i = 0; i < chain->SeqLen(); i++) {
                           if (chain->SeqRes(i)->Deleted) continue;
                           _RESIDUE *residue = chain->SeqRes(i)->Residue;
                           if (!residue) continue;

                           char *q = residue->name_list().Text();
                           while (q) {
                                q = get_word(q, ResName);
                                if (!strcmp(ResName, "")) continue;
                                if (!is_a_standard_residue(ResName) &&
                                     strcmp(ResName, "MSE")) {
                                     string1.Copy(ResName);
                                     if (het_ids.Seek(string1) < 0) het_ids.Add(string1);
                                }
                           }
                      }
                      chain = _molecules[_rep_mol]->GetNextChain();
                 }
                 if (het_ids.Root() >= 0) {
                      sprintf(buffer, "%s.ligand", File_ID);
                      fp2 = fopen(buffer, "w");
                      ret = het_ids.GoFirst();
                      while (ret >= 0) {
                           string1 = het_ids.GetCurrData();
                           fprintf(fp2, "%s\n", string1.Text());
                           ret = het_ids.GoToNext();
                      }
                      fclose (fp2);
                      fp2 = NULL;
                 }
            }
       } else {
            Table  = _CifObj->GetTablePtr(_StructureId, "pdbx_struct_assembly");
            Table1 = _CifObj->GetTablePtr(_StructureId, "pdbx_struct_assembly_gen");
            Table2 = _CifObj->GetTablePtr(_StructureId, "pdbx_struct_oper_list");
            if (is_empty_table(Table) || is_empty_table(Table1) ||
                is_empty_table(Table2)) {
                 fprintf(fp, "WARNING: Missing REMARK 350.\n\n");
            } else {
                 rowNo = Table1->GetNumRows();
                 high_index = Table1->GetColumnIndex("asym_id_list");
                 std::set<std::string> biol_assembly_asym_ids;
                 biol_assembly_asym_ids.clear();
                 for (int i = 0; i < rowNo; ++i) {
                      get_value(string1, Table1, high_index, i);
                      if (string1 == "") continue;
                      get_wordlist_from_string_separated_by_delimits(wordlist,
                                       string1.Text(), ",");
                      for (unsigned int j = 0; j < wordlist.Length(); ++j) {
                           biol_assembly_asym_ids.insert(wordlist[j].Text());
                      }
                 }
                 Table = _CifObj->GetTablePtr(_StructureId, "struct_biol_gen");
                 if (!is_empty_table(Table)) {
                      rowNo = Table->GetNumRows();
                      high_index = Table->GetColumnIndex("asym_id");
                      for (int i = 0; i < rowNo; ++i) {
                           get_value(string1, Table, high_index, i);
                           if (biol_assembly_asym_ids.find(string1.Text()) ==
                               biol_assembly_asym_ids.end()) {
                                fprintf(fp, "WARNING: Asym ID '%s' ", string1.Text());
                                fprintf(fp, "is not defined in biological assembly.\n\n");
                           }
                      }
                 }

                 rowNo = Table2->GetNumRows();
                 string.Clear();
                 high_index = Table2->GetColumnIndex("type");
                 low_index = Table2->GetColumnIndex("id");
                 for (int i = 0; i < rowNo; ++i) {
                      get_value(string1, Table2, high_index, i);
                      string1.ToLower();
                      if (string1 == "identity operation") {
                           get_value(string, Table2, low_index, i);
                           break;
                      }
                 }
                 if (string == "") {
                      fprintf(fp, "WARNING: Missing unit matrix in biological assembly.\n\n");
                 } else {
                      rowNo = Table1->GetNumRows();
                      high_index = Table1->GetColumnIndex("assembly_id");
                      low_index = Table1->GetColumnIndex("oper_expression");
                      std::map<std::string, std::string> mapping;
                      mapping.clear();
                      for (int i = 0; i < rowNo; ++i) {
                           get_value(string1, Table1, high_index, i);
                           get_value(string2, Table1, low_index, i);
                           if (string1 == "" || string2 == "") continue;
                           std::map<std::string, std::string>::iterator
                               mpos = mapping.find(string1.Text());
                           if (mpos != mapping.end()) {
                                mpos->second += ",";
                                mpos->second += string2.Text();
                           } else mapping.insert(std::make_pair(string1.Text(), string2.Text()));
                      }

                      for (std::map<std::string, std::string>::const_iterator
                           mpos = mapping.begin(); mpos != mapping.end(); ++mpos) {
                           string1.Clear();
                           string1.Copy(mpos->second.c_str());
                           get_wordlist_from_string_separated_by_delimits(wordlist,
                                       string1.Text(), ",");
                           bool found_unit = false;
                           for (unsigned int i = 0; i < wordlist.Length(); ++i) {
                                if (wordlist[i] == string) {
                                     found_unit = true;
                                     break;
                                }
                           }
                           if (!found_unit) {
                                fprintf(fp, "WARNING: Missing unit matrix for assembly '%s'.\n\n",
                                            mpos->first.c_str());
                           }
                      }
                 }
            }

            CifString header;
            header.Clear();
            Table = _CifObj->GetTablePtr(_StructureId, "struct_keywords");
            if (!is_empty_table(Table)) {
                 details = Table->GetColumnIndex("ndb_keywords");
                 Table->GetCell(header, details, 0);
                 if (header == "?" || header == ".") header.Clear();
            }
            if (header == "") {
                 fprintf(fp, "ERROR: Header is missing, please provide header ");
                 fprintf(fp, "information in _struct_keywords.ndb_keywords.\n\n");
                 _Error_Check[75]++;
            } else if (dict) {
                 if (check_header_failed(header, dict)) {
                      fprintf(fp, "WARNING: HEADER '%s'\n", header.Text());
                      _Error_Check[75]++; 
                      _Error_Check_Message[75].Copy(header);
                 }
            }

            if (_Error_Check[76]) {
                 fprintf(fp, "WARNING: Both MET and MSE residues are present ");
                 fprintf(fp, "in the sequence. Verify with the authors.\n\n");
            }

            Table = _CifObj->GetTablePtr(_StructureId, "chem_comp");
            if (!is_empty_table(Table)) {
                 rowNo = Table->GetNumRows();
                 details = Table->GetColumnIndex("id");
                 high_index = Table->GetColumnIndex("name");
                 for (i = 0; i < rowNo; i++) {
                      Table->GetCell(string, details, i);
                      if (fp1 && !is_a_standard_residue(string.Text()) &&
                           strcmp(string.Text(), "HOH") &&
                           strcmp(string.Text(), "DOD") &&
                           strcmp(string.Text(), "MSE")) {
                           if (fp2 == NULL) {
                                sprintf(buffer, "%s.ligand", File_ID);
                                fp2 = fopen(buffer, "w");
                           }
                           fprintf(fp2, "%s\n", string.Text());
                      }
                      get_value(string1, Table, high_index, i);
                      if (string1 == "") {
                           fprintf(fp, "WARNING: Missing HETNAM for '%s'.\n\n",
                                       string.Text());
                      }
                 }
                 if (fp1 && fp2) {
                      fclose (fp2);
                      fp2 = NULL;
                 }
            }

            if (fp1) {
                 if (header != "") {
                      sprintf(buffer, "%s.header", File_ID);
                      fp2 = fopen(buffer, "w");
                      fprintf(fp2, "%s\n", header.Text());
                      fclose (fp2);
                      fp2 = NULL;
                 }

                 const char *scheme[2] = { "ndb_poly_seq_scheme", "ndb_nonpoly_scheme" };
                 int pdb_mon_id, auth_mon_id;
                 CifString string4;
                 for (j = 0; j < 2; j++) {
                      Table = _CifObj->GetTablePtr(_StructureId, scheme[j]);
                      if (!is_empty_table(Table)) {
                           rowNo = Table->GetNumRows();
                           ndb_pdb_chain_id = Table->GetColumnIndex("pdb_id");
                           seq_num = Table->GetColumnIndex("pdb_num");
                           ndb_pdb_ins_code = Table->GetColumnIndex("pdb_ins_code");
                           pdb_mon_id = Table->GetColumnIndex("pdb_mon_id");
                           auth_mon_id = Table->GetColumnIndex("auth_mon_id");
                           for (i = 0; i < rowNo; i++) {
                                get_value(string, Table, ndb_pdb_chain_id, i); 
                                get_value(string1, Table, seq_num, i);
                                get_value(string2, Table, ndb_pdb_ins_code, i);
                                get_value(string3, Table, pdb_mon_id, i);
                                get_value(string4, Table, auth_mon_id, i);
                                strcpy(buffer, string4.Text());
                                if (find_obsolete_supercede_residue(buffer)) {
                                     if (string3 == buffer) {
                                          if (fp2 == NULL) {
                                               sprintf(buffer, "%s.ligand1", File_ID);
                                               fp2 = fopen(buffer, "w");
                                               fprintf(fp2, "The following residue(s) was(were) superseded:\n");
                                               fprintf(fp2, "   Chain   Seq   Old Res   New Res\n");
                                          }
                                          if (string != "")
                                               fprintf(fp2, "    %2s  ", string.Text());
                                          else fprintf(fp2, "        ");
                                          fprintf(fp2, "  %4s", string1.Text());
                                          if (string2 != "")
                                               fprintf(fp2, "%c  ", string2.Text()[0]);
                                          else fprintf(fp2, "   ");
                                          fprintf(fp2, "  %3s  ", string4.Text());
                                          fprintf(fp2, "     %3s\n", string3.Text());
                                     }
                                }
                           }
                      }
                 }
                 if (fp2) {
                      fprintf(fp2, "\n");
                      fclose (fp2);
                      fp2 = NULL;
                 }
            }

            Table = _CifObj->GetTablePtr(_StructureId, "struct_ref_seq_dif");
            if (!is_empty_table(Table) && !public_flag) {
                 details = Table->GetColumnIndex("details");
                 if (details >= 0) {
                      rowNo = Table->GetNumRows();
                      ndb_pdb_id_code = Table->GetColumnIndex("ndb_pdb_id_code");
                      mon_id = Table->GetColumnIndex("mon_id");
                      ndb_pdb_chain_id = Table->GetColumnIndex("ndb_pdb_chain_id");
                      seq_num = Table->GetColumnIndex("seq_num");
                      ndb_pdb_ins_code = Table->GetColumnIndex("ndb_pdb_ins_code");
                      ndb_seq_db_name = Table->GetColumnIndex("ndb_seq_db_name");
                      ndb_seq_db_accession_code = Table->GetColumnIndex("ndb_seq_db_accession_code");
                      db_mon_id = Table->GetColumnIndex("db_mon_id");
                      ndb_seq_db_seq_num = Table->GetColumnIndex("ndb_seq_db_seq_num");
                      ndb_auth_seq_num = Table->GetColumnIndex("ndb_auth_seq_num");

                      card_id = find_card_id("SEQADV");
                      int has_modified_residue = 0;
                      for (i = 0; i < rowNo; i++) {
                           Table->GetCell(string, details, i);
                           if (!strcasecmp(string.Text(), "MODIFIED RESIDUE")) has_modified_residue = 1;
                           if (strcasecmp(string.Text(), "CONFLICT")) continue;

                           pCard = _CardInfo[card_id].add_a_new_card(NdbToken[card_id], 0);
                           pCard->AddElementToField(string.Text(), 10);

                           string.Clear();
                           Table->GetCell(string, ndb_pdb_id_code, i);
                           if (string != "" && string != "." && string != "?")
                                pCard->AddElementToField(string.Text(), 1);

                           string.Clear();
                           Table->GetCell(string, mon_id, i);
                           if (string != "" && string != "." && string != "?")
                                pCard->AddElementToField(string.Text(), 2);

                           string.Clear();
                           Table->GetCell(string, ndb_pdb_chain_id, i);
                           if (string != "" && string != "." && string != "?")
                                pCard->AddElementToField(string.Text(), 3);

                           string.Clear();
                           Table->GetCell(string, seq_num, i);
                           if (string != "" && string != "." && string != "?")
                                pCard->AddElementToField(string.Text(), 4);

                           string.Clear();
                           Table->GetCell(string, ndb_auth_seq_num, i);
                           if (string != "" && string != "." && string != "?")
                                pCard->AddElementToField(string.Text(), 4);

                           string.Clear();
                           Table->GetCell(string, ndb_pdb_ins_code, i);
                           if (string != "" && string != "." && string != "?")
                                pCard->AddElementToField(string.Text(), 5);

                           string.Clear();
                           Table->GetCell(string, ndb_seq_db_name, i);
                           if (string != "" && string != "." && string != "?")
                                pCard->AddElementToField(string.Text(), 6);

                           string.Clear();
                           Table->GetCell(string, ndb_seq_db_accession_code, i);
                           if (string != "" && string != "." && string != "?")
                                pCard->AddElementToField(string.Text(), 7);

                           string.Clear();
                           Table->GetCell(string, db_mon_id, i);
                           if (string != "" && string != "." && string != "?")
                                pCard->AddElementToField(string.Text(), 8);

                           string.Clear();
                           Table->GetCell(string, ndb_seq_db_seq_num, i);
                           if (string != "" && string != "." && string != "?")
                                pCard->AddElementToField(string.Text(), 9);
                      }
                      if (has_modified_residue) {
                           fprintf(fp, "WARNING: \"modified residue\" exists in _struct_ref_seq_dif.details, ");
                           fprintf(fp, "remove it if it is not chromophore.\n\n");
                      }
                      if (_CardInfo[card_id].Head()) {
                           fprintf(fp, "Please note that the following residues given in your SEQRES conflict\n");
                           fprintf(fp, "with the sequence given in the  database. Do these conflicts represent\n");
                           fprintf(fp, "mutations in the deposited sequence?  Please advise.\n\n");
                           print_general_cards(fp, card_id, NDB_FILE_FORMAT_PDB, numRemark, 0);
                           fprintf(fp, "\n");
                      }
                      if (fp1) {
                           pCard = _CardInfo[card_id].Head();
                           while (pCard) pCard = _CardInfo[card_id].delete_a_card(pCard);

                           for (i = 0; i < rowNo; i++) {
                                Table->GetCell(string, details, i);
                                // if (strcasecmp(string.Text(), "CLONING ARTIFACT")) continue;
                                // if (strcasecmp(string.Text(), "EXPRESSION TAG")) continue;

                                pCard = _CardInfo[card_id].add_a_new_card(NdbToken[card_id], 0);
                                pCard->AddElementToField(string.Text(), 10);

                                string.Clear();
                                Table->GetCell(string, ndb_pdb_id_code, i);
                                if (string != "" && string != "." && string != "?")
                                     pCard->AddElementToField(string.Text(), 1);

                                string.Clear();
                                Table->GetCell(string, mon_id, i);
                                if (string != "" && string != "." && string != "?")
                                     pCard->AddElementToField(string.Text(), 2);

                                string.Clear();
                                Table->GetCell(string, ndb_pdb_chain_id, i);
                                if (string != "" && string != "." && string != "?")
                                     pCard->AddElementToField(string.Text(), 3);

                                string.Clear();
                                Table->GetCell(string, seq_num, i);
                                if (string != "" && string != "." && string != "?")
                                     pCard->AddElementToField(string.Text(), 4);

                                string.Clear();
                                Table->GetCell(string, ndb_auth_seq_num, i);
                                if (string != "" && string != "." && string != "?")
                                     pCard->AddElementToField(string.Text(), 4);

                                string.Clear();
                                Table->GetCell(string, ndb_pdb_ins_code, i);
                                if (string != "" && string != "." && string != "?")
                                     pCard->AddElementToField(string.Text(), 5);

                                string.Clear();
                                Table->GetCell(string, ndb_seq_db_name, i);
                                if (string != "" && string != "." && string != "?")
                                     pCard->AddElementToField(string.Text(), 6);

                                string.Clear();
                                Table->GetCell(string, ndb_seq_db_accession_code, i);
                                if (string != "" && string != "." && string != "?")
                                     pCard->AddElementToField(string.Text(), 7);

                                string.Clear();
                                Table->GetCell(string, db_mon_id, i);
                                if (string != "" && string != "." && string != "?")
                                     pCard->AddElementToField(string.Text(), 8);

                                string.Clear();
                                Table->GetCell(string, ndb_seq_db_seq_num, i);
                                if (string != "" && string != "." && string != "?")
                                     pCard->AddElementToField(string.Text(), 9);
                           }
                           if (_CardInfo[card_id].Head()) {
                                sprintf(buffer, "%s.clone", File_ID);
                                fp2 = fopen(buffer, "w");
                                print_general_cards(fp2, card_id, NDB_FILE_FORMAT_PDB, numRemark, 0);
                                fprintf(fp2, "\n");
                                fclose (fp2);
                                fp2 = NULL;
                           }
                      }
                 }
            }

            for (i = 0; i < NUM_CHECK_RESOLUTION; i++) {
                 Table1 = _CifObj->GetTablePtr(_StructureId, check_res[i].category1);
                 Table2 = _CifObj->GetTablePtr(_StructureId, check_res[i].category2);
                 if (is_empty_table(Table1) || is_empty_table(Table2)) continue;
                 high_index = Table1->GetColumnIndex(check_res[i].item1);
                 low_index  = Table2->GetColumnIndex(check_res[i].item2);
                 if (high_index < 0 || low_index < 0) continue;
                 Table1->GetCell(string, high_index, 0);
                 Table2->GetCell(string1, low_index, 0);
                 if (is_not_empty(string) && is_not_empty(string1)) {
                      if (atof(string.Text()) > atof(string1.Text())) {
                           if (strcmp(check_res[i].message, "")) {
                                if (check_res[i].contain_value)
                                     sprintf(buffer, check_res[i].message, string.Text(), string1.Text());
                                else strcpy(buffer, check_res[i].message);
                                fprintf(fp, "ERROR: %s\n", buffer);
                           } else if (!public_flag) {
                                fprintf(fp, "ERROR: _%s.%s=%s > _%s.%s=%s\n\n",
                                   check_res[i].category1, check_res[i].item1, string.Text(),
                                   check_res[i].category2, check_res[i].item2, string1.Text());
                           }
                           _Error_Check[10]++;
                      }
                 }
            }

            if (_experiment_type & EXPERIMENT_TYPE_XRAY) {
                 for (i = 0; i < NUM_MISSING_VALUE; i++) {
                      Table = _CifObj->GetTablePtr(_StructureId, _Missing_Values[i].category);
                      is_error = 0;
                      if (is_empty_table(Table)) is_error = 1;
                      else {
                           high_index = Table->GetColumnIndex(_Missing_Values[i].item);
                           if (high_index < 0) is_error = 1;
                           else {
                                Table->GetCell(string, high_index, 0);
                                if (!is_not_empty(string)) is_error = 1;
                           }
                      }
                      if (is_error) {
                           if (strcmp(_Missing_Values[i].message, ""))
                                fprintf(fp, "WARNING: %s\n", _Missing_Values[i].message);
                           else if (!public_flag) 
                                fprintf(fp, "WARNING: No _%s.%s value\n\n",
                                     _Missing_Values[i].category, _Missing_Values[i].item);
                      }
                 }
            }

            if (_experiment_type & EXPERIMENT_TYPE_NMR ||
                _experiment_type & EXPERIMENT_TYPE_NMR_SOLID) {
                 Table = _CifObj->GetTablePtr(_StructureId, "rcsb_nmr_software");
                 is_error = 0;
                 if (is_empty_table(Table)) is_error = 1;
                 else {
                      high_index = Table->GetColumnIndex("name");
                      if (high_index < 0) is_error = 1;
                      else {
                           is_error = 1;
                           low_index  = Table->GetColumnIndex("classification");
                           rowNo = Table->GetNumRows();
                           for (i = 0; i < rowNo; i++) {
                                Table->GetCell(string, low_index, i);
                                if (string == "refinement" || string == "solution") {
                                     Table->GetCell(string, high_index, i);
                                     if (is_not_empty(string)) {
                                          is_error = 0;
                                          break;
                                     }
                                }
                           }
                      }
                 }
                 if (is_error) fprintf(fp, "WARNING: Please provide the name of the refinement program.\n\n");

                 Table = _CifObj->GetTablePtr(_StructureId, "rcsb_nmr_ensemble");
                 if (!is_empty_table(Table)) {
                      high_index = Table->GetColumnIndex("conformers_submitted_total_number");
                      if (high_index >= 0) {
                           Table->GetCell(string, high_index, 0);
                           if (is_not_empty(string)) {
                                if (atoi(string.Text()) != _num_mol) {
                                     fprintf(fp, "WARNING: The number %s in ", string.Text());
                                     fprintf(fp, "_rcsb_nmr_ensemble.conformers_submitted_total_number is\n");
                                     fprintf(fp, "         different from the number of deposited conformers %d\n",
                                                    _num_mol);
                                }
                           }
                      }
                 }
            }

            for (i = 0; i < NUM_CATEGORY_COMP; i++) {
                 Table = _CifObj->GetTablePtr(_StructureId, _Check_Category_Comp[i].category1);
                 Table1 = _CifObj->GetTablePtr(_StructureId, _Check_Category_Comp[i].category2);
                 if (is_empty_table(Table) || is_empty_table(Table1)) continue;
                 high_index = Table->GetColumnIndex(_Check_Category_Comp[i].item1);
                 low_index  = Table1->GetColumnIndex(_Check_Category_Comp[i].item2);
                 if (high_index < 0 || low_index < 0) continue;
                 Table->GetCell(string, high_index, 0);
                 Table1->GetCell(string1, low_index, 0);
                 if (is_not_empty(string) && is_not_empty(string1)) {
                      if (_Check_Category_Comp[i].op)
                           d_alpha = atof(string.Text()) / atof(string1.Text());
                      else d_alpha = atof(string.Text()) - atof(string1.Text());
                      if (d_alpha > _Check_Category_Comp[i].value) {
                           if (strcmp(_Check_Category_Comp[i].message, "")) {
                                sprintf(buffer, _Check_Category_Comp[i].message, string.Text(),
                                                 string1.Text());
                                fprintf(fp, "WARNING: %s\n", buffer);
                           } else if (!public_flag) {
                                fprintf(fp, "WARNING: The value of ");
                                if (_Check_Category_Comp[i].op)
                                     fprintf(fp, "_%s.%s / _%s.%s ",
                                        _Check_Category_Comp[i].category1, _Check_Category_Comp[i].item1,
                                        _Check_Category_Comp[i].category2, _Check_Category_Comp[i].item2);
                                else fprintf(fp, "_%s.%s - _%s.%s ",
                                        _Check_Category_Comp[i].category1, _Check_Category_Comp[i].item1,
                                        _Check_Category_Comp[i].category2, _Check_Category_Comp[i].item2);
                                fprintf(fp, "is larger than %.1f\n\n", _Check_Category_Comp[i].value);
                           }
                      }
                 }
            }

            for (i = 0; i < NUM_CATEGORY_COMP3; i++) {
                 Table = _CifObj->GetTablePtr(_StructureId, _Check_Category_Comp3[i].category);
                 if (is_empty_table(Table)) continue;
                 e_index    = Table->GetColumnIndex(_Check_Category_Comp3[i].item1);
                 high_index = Table->GetColumnIndex(_Check_Category_Comp3[i].item2);
                 low_index  = Table->GetColumnIndex(_Check_Category_Comp3[i].item3);
                 if (e_index < 0 || high_index < 0 || low_index < 0) continue;
                 Table->GetCell(string, e_index, 0);
                 Table->GetCell(string1, high_index, 0);
                 Table->GetCell(string2, low_index, 0);
                 if (is_not_empty(string) && is_not_empty(string1) && is_not_empty(string2)) {
                      if (atof(string.Text()) < 0.000001) {
                           if (fabs(atof(string2.Text()) - atof(string1.Text())) > 0.000001) {
                                if (strcmp(_Check_Category_Comp3[i].message3, "")) {
                                     fprintf(fp, "WARNING: ");
                                     fprintf(fp, _Check_Category_Comp3[i].message3, string1.Text(), string2.Text());
                                } else if (!public_flag) {
                                     fprintf(fp, "WARNING: _%s.%s=%s is not same as _%s.%s=%s\n\n",
                                          _Check_Category_Comp3[i].category, _Check_Category_Comp3[i].item3,
                                          string2.Text(), _Check_Category_Comp3[i].category,
                                          _Check_Category_Comp3[i].item2, string1.Text());
                                }
                           }
                      } else {
                           if (atof(string2.Text()) > atof(string1.Text())) {
                                if (strcmp(_Check_Category_Comp3[i].message1, "")) {
                                     fprintf(fp, "WARNING: %s\n", _Check_Category_Comp3[i].message1);
                                } else if (!public_flag) {
                                     fprintf(fp, "WARNING: _%s.%s=%s > _%s.%s=%s\n\n",
                                          _Check_Category_Comp3[i].category, _Check_Category_Comp3[i].item3,
                                          string2.Text(), _Check_Category_Comp3[i].category,
                                          _Check_Category_Comp3[i].item2, string1.Text());
                                }
                           } else if (fabs(atof(string2.Text()) - atof(string1.Text())) < 0.000001) {
                                if (strcmp(_Check_Category_Comp3[i].message2, "")) {
                                     fprintf(fp, "WARNING: %s\n", _Check_Category_Comp3[i].message2);
                                } else if (!public_flag) {
                                     fprintf(fp, "WARNING: _%s.%s=%s = _%s.%s=%s\n\n",
                                          _Check_Category_Comp3[i].category, _Check_Category_Comp3[i].item3,
                                          string2.Text(), _Check_Category_Comp3[i].category,
                                          _Check_Category_Comp3[i].item2, string1.Text());
                                }
                           }
                      }
                 }
            }

            if (!public_flag)
         {
            for (i = 0; i < _n_special; i++) {
                 atom = _molecules[0]->find_ndb_atom(_special[i]->a_atomtyp,
                                 _special[i]->a_chnid, _special[i]->a_resnam,
                                 _special[i]->a_resnum, "", num);
                 if (!atom) continue;
                 if (atom[0]->occ() > 0.99) {
                      _Error_Check[63]++;
                      fprintf(fp, "WARNING: Atom (%s %s %d %s) with occupancy 1.0 on special position\n\n",
                                atom[0]->atmtype(), atom[0]->pdb_resnam(),
                                atom[0]->pdb_resnum(), atom[0]->pdb_chnid());                                       
                      if (_Error_Check_Message[63] != "") _Error_Check_Message[63] += ",";
                      if (strcmp(atom[0]->pdb_chnid(), ""))
                           sprintf(buffer, "(%s %s %d %s)", atom[0]->atmtype(), atom[0]->pdb_resnam(),
                                           atom[0]->pdb_resnum(), atom[0]->pdb_chnid());
                      else sprintf(buffer, "(%s %s %d  )", atom[0]->atmtype(), atom[0]->pdb_resnam(),
                                           atom[0]->pdb_resnum());
                      _Error_Check_Message[63] += buffer;
                 }
                 delete [] atom;
            }

            Table = _CifObj->GetTablePtr(_StructureId, "exptl_crystal");
            if (!is_empty_table(Table) && _experiment_type & EXPERIMENT_TYPE_XRAY) {
                 if (cal_matthew_and_solvent(matthew, solvent)) {
                      high_index = Table->GetColumnIndex("density_Matthews");
                      low_index = Table->GetColumnIndex("density_percent_sol");
                      get_value(string1, Table, high_index, 0);
                      if (string1 != "") {
                           rad = fabs(matthew - atof(string1.Text())) / matthew;
                           if (rad > 0.1) {
                                _Error_Check[60]++;
                                sprintf(buffer, "Matthews coefficient %.2f vs. %s", matthew, string1.Text());
                                _Error_Check_Message[60] += buffer;
                                fprintf(fp, "WARNING: The calculated Matthews coefficient is %.2f, but it was reported as %s in your deposition. Do you approve changing the Matthews coefficient to the calculated value?\n\n",
                                             matthew, string1.Text());
                           }
                           if (atof(string1.Text()) < 1.8 || atof(string1.Text()) > 5) {
                                _Error_Check[60]++;
                                if (_Error_Check_Message[60] != "") _Error_Check_Message[60] += ", ";
                                _Error_Check_Message[60] += "Matthews coefficient=";
                                _Error_Check_Message[60] += string1;
                                fprintf(fp, "WARNING: The Matthews coefficient should be in 1.8 - 5. But the calculated value is %s\n\n", string1.Text());
                           }
                      } else {
                           fprintf(fp, "WARNING: Missing Matthews coefficient value. The calculated value is %.2f.", matthew);
                           if (matthew < 1.8 || matthew > 5) fprintf(fp, " The Matthews coefficient should be in 1.8 - 5.");
                           fprintf(fp, "\n\n");
                      }

                      get_value(string1, Table, low_index, 0);
                      if (string1 != "") {
                           rad = fabs(solvent - atof(string1.Text()) / 100.0) / solvent;
                           if (rad > 0.1) {
                                _Error_Check[60]++;
                                sprintf(buffer, "solvent content %.2f vs. %s", solvent * 100, string1.Text());
                                if (_Error_Check_Message[60] != "") _Error_Check_Message[60] += ", ";
                                _Error_Check_Message[60] += buffer;
                                fprintf(fp, "WARNING: The calculated solvent content is %.2f, but it was reported as %s in your deposition. Do you approve changing the solvent content to the calculated value?\n\n",
                                            solvent * 100, string1.Text());
                           }
                           if (atof(string1.Text()) < 31.6 || atof(string1.Text()) > 75.4) {
                                _Error_Check[60]++;
                                if (_Error_Check_Message[60] != "") _Error_Check_Message[60] += ", ";
                                _Error_Check_Message[60] += "solvent content=";
                                _Error_Check_Message[60] += string1;
                                fprintf(fp, "WARNING: The solvent content should be in 31.6%% - 75.4%%. But the calculated value is %s\n\n", string1.Text());
                           }
                      } else {
                           fprintf(fp, "WARNING: Missing solvent content value. The calculated value is %.2f.", solvent * 100);
                           if (solvent < 31.6 || solvent > 75.4) fprintf(fp, " The solvent content should be in 31.6%% - 75.4%%.");
                           fprintf(fp, "\n\n");
                      }
                 }
            }

            Table = _CifObj->GetTablePtr(_StructureId, "exptl_crystal_grow");
            if (!is_empty_table(Table)) {
                 rowNo = Table->GetNumRows();
                 high_index = Table->GetColumnIndex("temp");
                 if (high_index >= 0) {
                      for (i = 0; i < rowNo; i++) {
                           Table->GetCell(string, high_index, i);
                           if (string != "" && string != "?" && string != ".") {
                                if (atof(string.Text()) < 273) {
                                     _Error_Check[61]++;
                                     if (_Error_Check_Message[61] != "") _Error_Check_Message[61] += ", ";
                                     _Error_Check_Message[61] += string;
                                     fprintf(fp, "WARNING: Crystallization temperature is %s\n\n", string.Text());
                                }
                           }
                      }
                 }
            }

            std::set<std::string> polymer_entity_ids, pdb_chain_ids;
            polymer_entity_ids.clear();
            pdb_chain_ids.clear();

            int has_nucleic_acid = 0;
            int has_targetdb = 0;
            int has_protein = 0;
            Table = _CifObj->GetTablePtr(_StructureId, "entity_poly");
            if (!is_empty_table(Table)) {
                 rowNo = Table->GetNumRows();
                 high_index = Table->GetColumnIndex("type");
                 low_index = Table->GetColumnIndex("pdbx_target_identifier");
                 entity_index = Table->GetColumnIndex("entity_id");
                 s_index = Table->GetColumnIndex("ndb_seq_one_letter_code");
                 o_index = Table->GetColumnIndex("ndb_chain_id");
                 int has_other = 0;
                 char *sequence = NULL, *code = NULL, buffer1[200], buffer2[100];
                 for (i = 0; i < rowNo; i++) {
                      get_value(string, Table, o_index, i);
                      if (string != "") {
                           get_wordlist_from_string_separated_by_delimits(wordlist,
                                       string.Text(), ",");
                           for (unsigned int t = 0; t < wordlist.Length(); ++t)
                                pdb_chain_ids.insert(wordlist[t].Text());
                      }
                      get_value(string, Table, high_index, i);
                      string.ToLower();
                      if (string == "polydeoxyribonucleotide" ||
                          string == "polyribonucleotide" ||
                          string == "polydeoxyribonucleotide/polyribonucleotide hybrid")
                           has_nucleic_acid = 1;

                      if (string == "polypeptide(l)") has_protein = 1;
                      if (string == "other") has_other = 1;

                      get_value(string, Table, low_index, i);
                      get_value(string1, Table, s_index, i);    
                      get_value(string2, Table, entity_index, i);
                      polymer_entity_ids.insert(string2.Text());

                      if (string == "") continue;

                      has_targetdb = 1;
                      if (string1 == "") continue;

                      string1.ToUpper();
                      ret = strlen(string1.Text());
                      sequence = new char[ret + 1];
                      memset(sequence, 0, ret + 1);
                      k = 0;
                      for (j = 0; j < ret; j++) {
                           if (string1.Text()[j] == ' ' ||
                               string1.Text()[j] == '\n') continue;
                           if (string1.Text()[j] != '(') {
                                sequence[k] = string1.Text()[j];
                           } else {
                                j++;
                                l = 0;
                                while (string1.Text()[j] != ')') {
                                     buffer[l] = string1.Text()[j];
                                     l++; j++;
                                }
                                buffer[l] = '\0';
                                ConnectFormat *drug = find_drug(buffer);
                                if (drug && drug->parent_residue() &&
                                    strcmp(drug->parent_residue(), "")) {
                                     code = get_oneletter_code(drug->parent_residue());
                                     if (strlen(code) == 1)
                                          sequence[k] = code[0];
                                     else sequence[k] = 'X';
                                } else sequence[k] = 'X';
                           }
                           k++;
                      }

                      strcpy(buffer1,  string.Text());

                      int low_similarity = 0;
                      p = buffer1;
                      while (p) {
                           p = get_word1(p, buffer2);
                           ndb_clean_string(buffer2);
                           if (!strcmp(buffer2, "")) continue;
                           for (j = 0; j < (int) strlen(buffer2); j++) {
                                if (buffer2[j] == ' ') buffer2[j] = '_';
                           }
                           ret = check_target_sequence(fp, sequence, string2.Text(),
                                    buffer2);
                           if (ret > low_similarity) low_similarity = ret;
                      }
                      if (!low_similarity) {
                           _Error_Check[40]++;
                           fprintf(fp, "WARNING: Target ID %s did not exist in target database\n", string.Text());
                      } else if (low_similarity < 100) _Error_Check[41] = low_similarity;
                      delete [] sequence;
                 }
                 if (has_targetdb) {
                      Table = _CifObj->GetTablePtr(_StructureId, "pdbx_SG_project");
                      if (is_empty_table(Table)) {
                           fprintf(fp, "WARNING: The Target ID is filled in, but the ");
                           fprintf(fp, "_pdbx_SG_project.full_name_of_center is not filled in. ");
                           fprintf(fp, "Please fill in the center name.\n");
                      } else {
                           rowNo = Table->GetNumRows();
                           low_index = Table->GetColumnIndex("full_name_of_center");
                           for (i = 0; i < rowNo; i++) {
                                Table->GetCell(string, low_index, i);
                                if (string != "" && string != "?" && string != ".") break;
                           }
                           if (i == rowNo) {
                                fprintf(fp, "WARNING: The Target ID is filled in, but the ");
                                fprintf(fp, "_pdbx_SG_project.full_name_of_center is not filled in. ");
                                fprintf(fp, "Please fill in the center name.\n");
                           }
                      }
                 }
                 if (has_other) fprintf(fp, "WARNING: _entity_poly.type is OTHER, please correct it.\n\n");
            }

            int need_na_checking1 = 0;
            int need_na_checking2 = 0;
            int need_na_checking3 = 0;
            if (_experiment_type & EXPERIMENT_TYPE_NMR ||
                _experiment_type & EXPERIMENT_TYPE_NMR_SOLID ||
                _experiment_type & EXPERIMENT_TYPE_NEUTRON ||
                _experiment_type & EXPERIMENT_TYPE_XRAY) {
                 if (has_protein && has_nucleic_acid) need_na_checking1 = 1;
                 if (has_nucleic_acid && (_experiment_type & EXPERIMENT_TYPE_NEUTRON ||
                     _experiment_type & EXPERIMENT_TYPE_XRAY)) need_na_checking2 = 1;
                 if (has_nucleic_acid) need_na_checking3 = 1;
            }

            // if (has_nucleic_acid && _experiment_type & EXPERIMENT_TYPE_NMR) {
            if (has_nucleic_acid) {
                 Table = _CifObj->GetTablePtr(_StructureId, "database_2");
                 if (!is_empty_table(Table)) {
                      rowNo = Table->GetNumRows();
                      high_index = Table->GetColumnIndex("database_id");
                      low_index  = Table->GetColumnIndex("database_code");
                      string1.Copy("");
                      string2.Copy("");
                      for (i = 0; i < rowNo; i++) {
                           Table->GetCell(string, high_index, i);
                           if (string == "NDB") {
                                Table->GetCell(string1, low_index, i);
                           } else if (string == "PDB") {
                                Table->GetCell(string2, low_index, i);
                           }
                      }
                      if (_experiment_type & EXPERIMENT_TYPE_NMR ||
                          _experiment_type & EXPERIMENT_TYPE_NMR_SOLID) {
                           if (string1 != string2) {
                                fprintf(fp, "WARNING: This NMR structure needs to have its NDB ID changed to match the PDB ID.\n\n");
                                _Error_Check[38]++; 
                           }
                      } else {
                           if (string1 == _StructureId && !(_experiment_type & EXPERIMENT_TYPE_MODEL) &&
                               !(_experiment_type & EXPERIMENT_TYPE_CRYO_EM)) {
                                fprintf(fp, "WARNING: No NDB ID assigned for the entry.\n\n");
                                _Error_Check[38]++;
                           }
                      }
                 }
            }

            Table = _CifObj->GetTablePtr(_StructureId, "entity_src_gen");
            if (!is_empty_table(Table)) {
                 high_index = Table->GetColumnIndex("ndb_gene_src_scientific_name");
                 int tax_idx1 = Table->GetColumnIndex("pdbx_gene_src_ncbi_taxonomy_id");
                 low_index  = Table->GetColumnIndex("ndb_host_org_scientific_name");
                 int tax_idx2 = Table->GetColumnIndex("pdbx_host_org_ncbi_taxonomy_id");
                 rowNo = Table->GetNumRows();
                 for (i = 0; i < rowNo; i++) {
                      check_source(fp, _spell_check, Table, i, high_index, tax_idx1);
                      check_source(fp, _spell_check, Table, i, low_index, tax_idx2);
                 }
            }

            Table = _CifObj->GetTablePtr(_StructureId, "entity_src_nat");
            if (!is_empty_table(Table)) {
                 high_index = Table->GetColumnIndex("ndb_organism_scientific");
                 low_index = Table->GetColumnIndex("pdbx_ncbi_taxonomy_id");
                 rowNo = Table->GetNumRows();
                 for (i = 0; i < rowNo; i++) {
                      check_source(fp, _spell_check, Table, i, high_index, low_index);
                 }
            }

            Table = _CifObj->GetTablePtr(_StructureId, "rcsb_entity_src_syn");
            if (!is_empty_table(Table)) {
                 high_index = Table->GetColumnIndex("organism_scientific");
                 low_index = Table->GetColumnIndex("ncbi_taxonomy_id");
                 rowNo = Table->GetNumRows();
                 for (i = 0; i < rowNo; i++) {
                      check_source(fp, _spell_check, Table, i, high_index, low_index);
                 }
            }

            Table = _CifObj->GetTablePtr(_StructureId, "exptl");
            if (!is_empty_table(Table)) {
                 high_index = Table->GetColumnIndex("method");
                 string.Clear();
                 Table->GetCell(string, high_index, 0);
                 if (string == "" || string == "?" || string == ".") {
                      fprintf(fp, "WARNING: missing experimental method.\n\n");
                 }
                 if (!strcasecmp(string.Text(), "THEORETICAL MODELING") ||
                     !strcasecmp(string.Text(), "THEORETICAL MODELLING")) {
                      fprintf(fp, "WARNING: _exptl.method is '%s'. It should be 'THEORETICAL MODEL'.\n\n",
                                   string.Text());
                 }
            } else fprintf(fp, "WARNING: missing experimental method.\n\n");

            int suppressed_remark_350 = 0;
            Table = _CifObj->GetTablePtr(_StructureId, "ndb_database_status");
            if (!is_empty_table(Table)) {
                 high_index = Table->GetColumnIndex("skip_PDB_REMARK");
                 string.Clear();
                 Table->GetCell(string, high_index, 0);
                 if (strstr(string.Text(), "350")) suppressed_remark_350 = 1;
            }

         if (__maxit_version == "3.1") {
/*
            Table = _CifObj->GetTablePtr(_StructureId, "pdbx_struct_assembly");
            if (!is_empty_table(Table)) {
                 if (fp1) fprintf(fp1, "2. Please confirm all biomolecules described in remark 350.\n\n");
                 else fprintf(fp, "Please confirm all biomolecules described in remark 350.\n\n");
            }
*/
         } else {
            Table = _CifObj->GetTablePtr(_StructureId, "struct_biol_gen");
            Table1 = _CifObj->GetTablePtr(_StructureId, "struct_asym");
            if (!is_empty_table(Table) && !is_empty_table(Table1) &&
                _experiment_type & EXPERIMENT_TYPE_XRAY) {
                 rowNo = Table1->GetNumRows();
                 high_index = Table1->GetColumnIndex("id");
                 low_index = Table1->GetColumnIndex("ndb_type");
                 _UniIndex.Clear();
                 for (i = 0; i < rowNo; i++) {
                      Table1->GetCell(string, low_index, i);
                      if (string == "ATOMP" || string == "ATOMN" || string == "NA" || string == "PR") {
                           Table1->GetCell(string, high_index, i);
                           _UniIndex.Add(string);
                      }
                 }

                 rowNo = Table->GetNumRows();
                 high_index = Table->GetColumnIndex("symmetry");
                 low_index = Table->GetColumnIndex("biol_id");
                 e_index = Table->GetColumnIndex("asym_id");
                 s_index = Table->GetColumnIndex("ndb_new_pdb_asym_id");
                 if (high_index >= 0 && low_index >= 0 && e_index >= 0 && s_index >= 0) {
                      _BiolIndex.Clear();
                      type_index = 0;
                      _AsymIndex.Clear();
                      for (i = 0; i < rowNo; i++) {
                           Table->GetCell(string, e_index, i);
                           if (_UniIndex.Seek(string) < 0) continue;
                           Table->GetCell(string1, low_index, i);
                           if (_BiolIndex.Seek(string1) < 0) _BiolIndex.Add(string1);
                           Table->GetCell(string2, high_index, i);
                           if (string2 != "?" && string2 != "." && string2 != "" &&
                               string2 != "1_555") type_index = 1;
                           if (string2 == "1_555") {
                                Table->GetCell(string1, s_index, i);
                                idx.set_svalue(string.Text());
                                idx.set_ivalue(string1.Text());
                                _AsymIndex.Add(idx);
                           }
                      }

                      if (_BiolIndex.Root() >= 0) {
                           ret = _BiolIndex.GoFirst();
                           if (_BiolIndex.NumElem() == 1 && !type_index) {
                                if (suppressed_remark_350) {
                                     if (fp1) {
                                          fprintf(fp1, "2. The remark 350 in the PDB file was suppressed since information in the deposition\n");
                                          fprintf(fp1, "indicated that the biological unit was unknown. Please review remark 300 which has\n");
                                          fprintf(fp1, "been modified to explain that.\n\n"); 
                                     } else {
                                          fprintf(fp, "The remark 350 in the PDB file was suppressed since information in the deposition\n");
                                          fprintf(fp, "indicated that the biological unit was unknown. Please review remark 300 which has\n");
                                          fprintf(fp, "been modified to explain that.\n\n"); 
                                     }
                                } else {
                                     if (fp1)
                                          fprintf(fp1, "2. Please confirm, the biological unit is identical to the asymmetric unit.\n\n"); 
                                     else fprintf(fp, "Please confirm, the biological unit is identical to the asymmetric unit.\n\n");
                                }
                           } else {
                                if (!type_index) {
                                     if (fp1)
                                          fprintf(fp1, "2. Please confirm there are %d biological units in the asymmetric unit.\n",
                                                        _BiolIndex.NumElem());
                                     else fprintf(fp, "Please confirm there are %d biological units in the asymmetric unit.\n",
                                                        _BiolIndex.NumElem());
                                } else if (_BiolIndex.NumElem() > 1) {
                                     if (fp1)
                                          fprintf(fp1, "2. Please confirm there are %d biological units.\n", _BiolIndex.NumElem());
                                     else fprintf(fp, "Please confirm there are %d biological units.\n", _BiolIndex.NumElem());
                                }
                                while (ret >= 0) {
                                     _BiolUnit.Clear();
                                     for (i = 0; i < rowNo; i++) {
                                          Table->GetCell(string, e_index, i);
                                          if (_UniIndex.Seek(string) < 0) continue;
                                          Table->GetCell(string1, low_index, i);
                                          if (string1 != _BiolIndex.GetCurrData()) continue;
                                          Table->GetCell(string1, high_index, i);
                                          idx.set_svalue(string.Text());
                                          _AsymIndex.Seek(idx);
                                          idx = _AsymIndex.GetCurrData();
                                          idx1.set_svalue(idx.ivalue());
                                          if (string1 == "1_555") {
                                               idx1.set_ivalue("");
                                               _BiolUnit.Add(idx1);
                                          }
                                          if (string1 != "?" && string1 != "." && string1 != "" &&
                                              string1 != "1_555") {
                                               _BiolUnit.Seek(idx1);
                                               pidx = _BiolUnit.GetCurrDataPtr();
                                               rcsb_get_symmetry_operation_name(_cell, string1.Text(), buffer);
                                               string2.Copy(pidx->ivalue());
                                               if (string2 != "") string2 += ",";
                                               if ((strlen(string2.Text()) % 80) > 70) string2 += "\n";
                                               string2 += buffer;
                                               string2 += " (";
                                               string2 += string1;
                                               string2 += ")";
                                               pidx->set_ivalue(string2.Text());
                                          }
                                     }
                                     if (_BiolUnit.Root() >= 0) {
                                          if (!type_index) {
                                               if (fp1)
                                                    fprintf(fp1, "      The biological unit %s consists of chain(s) ",
                                                                    _BiolIndex.GetCurrData().Text());
                                               else fprintf(fp, "      The biological unit %s consists of chain(s) ",
                                                                    _BiolIndex.GetCurrData().Text());
                                               len = strlen(_BiolIndex.GetCurrData().Text()) + 48;
                                          } else if (_BiolIndex.NumElem() > 1) {
                                               if (fp1)
                                                    fprintf(fp1, "      The biological unit %s is generated by ",
                                                                    _BiolIndex.GetCurrData().Text());
                                               else fprintf(fp, "      The biological unit %s is generated by ",
                                                                    _BiolIndex.GetCurrData().Text());
                                               len = strlen(_BiolIndex.GetCurrData().Text()) + 43;
                                          } else {
                                               if (fp1)
                                                    fprintf(fp1, "2. Please confirm that the biological unit is generated by ");
                                               else fprintf(fp, "Please confirm that the biological unit is generated by ");
                                               len = 56;
                                          }
                                          _BiolUnit.GoFirst();
                                          ret1 = 1;
                                          first = 1;
                                          while (ret1 >= 0) {
                                               idx = _BiolUnit.GetCurrData();
                                               if (!first) {
                                                    if (fp1) fprintf(fp1, ", ");
                                                    else fprintf(fp, ", ");
                                                    len += 2;
                                                    if (type_index) {
                                                         if (fp1) fprintf(fp1, "\n       ");
                                                         else fprintf(fp, "\n       ");
                                                         len = 7;
                                                    }
                                                    if (len > 75) {
                                                         if (fp1) fprintf(fp1, "\n");
                                                         else fprintf(fp, "\n");
                                                         len = 0;
                                                    }
                                               }
                                               if (!type_index) {
                                                    if (fp1) fprintf(fp1, "%s", idx.svalue());
                                                    else fprintf(fp, "%s", idx.svalue());
                                                    len += strlen(idx.svalue());
                                               } else {
                                                    if (fp1) fprintf(fp1, "chain %s", idx.svalue());
                                                    else fprintf(fp, "chain %s", idx.svalue());
                                                    len += strlen(idx.svalue()) + 6;
                                               }
                                               if (strcmp(idx.ivalue(), "")) {
                                                    if (fp1) fprintf(fp1, " (x,y,z (1_555),");
                                                    else fprintf(fp, " (x,y,z (1_555),");
                                                    len += 15;
                                                    if (len > 75) {
                                                         if (fp1) fprintf(fp1, "\n");
                                                         else fprintf(fp, "\n");
                                                         len = 0;
                                                    }
                                                    if (fp1) fprintf(fp1, "%s)", idx.ivalue());
                                                    else fprintf(fp, "%s)", idx.ivalue());
                                                    len += strlen(idx.ivalue())+1;
                                               }
                                               first = 0;
                                               ret1 = _BiolUnit.GoToNext();
                                          }
                                          if (fp1) fprintf(fp1, ".\n");
                                          else fprintf(fp, ".\n");
                                     }
                                     ret = _BiolIndex.GoToNext();
                                }
                                if (fp1) fprintf(fp1, "\n");
                                else fprintf(fp, "\n");
                           }
                      }
                 }
            }
         }

            _UniIndex.Clear();
            for (i = 0; i < 3; i++) {
                 Table = _CifObj->GetTablePtr(_StructureId, source_category[i]);
                 rowNo = 0;
                 if (!is_empty_table(Table)) {
                      rowNo = Table->GetNumRows();
                      entity_index = Table->GetColumnIndex("entity_id");
                      for (j = 0; j < rowNo; j++) {
                           string.Clear();
                           Table->GetCell(string, entity_index, j);
                           if (string != "" && string != "?" && string != ".") {
                                if (_UniIndex.Seek(string) < 0) _UniIndex.Add(string);
                                else {
                                     fprintf(fp, "ERROR: entity '%s' had multiple source information\n\n",
                                                     string.Text());
                                     _Error_Check[24]++;
                                }
                           }
                      }
                 }
            }

            for (i = 0; i < NUM_SAME_VALUE; i++) {
                 Table1 = _CifObj->GetTablePtr(_StructureId, check_same[i].category1);
                 Table2 = _CifObj->GetTablePtr(_StructureId, check_same[i].category2);
                 if (is_empty_table(Table1) || is_empty_table(Table2)) continue;
                 high_index = Table1->GetColumnIndex(check_same[i].item1);
                 low_index  = Table2->GetColumnIndex(check_same[i].item2);
                 if (high_index < 0 || low_index < 0) continue;
                 Table1->GetCell(string, high_index, 0);
                 Table2->GetCell(string1, low_index, 0);
                 if (is_not_empty(string) && is_not_empty(string1) &&
                     fabs(atof(string.Text()) - atof(string1.Text())) > 0.000001) {
                      fprintf(fp, "WARNING: _%s.%s=%s is not same as _%s.%s=%s\n\n",
                            check_same[i].category1, check_same[i].item1, string.Text(),
                            check_same[i].category2, check_same[i].item2, string1.Text());
                 }
            }

            for (i = 0; i < NUM_ZERO_VALUE; i++) {
                 Table = _CifObj->GetTablePtr(_StructureId, _Zero_Values[i].category);
                 if (is_empty_table(Table)) continue;
                 high_index = Table->GetColumnIndex(_Zero_Values[i].item);
                 if (high_index < 0) continue;
                 Table->GetCell(string, high_index, 0);
                 if (is_not_empty(string) && atof(string.Text()) < 0.000001) {
                      fprintf(fp, "WARNING: _%s.%s is %s\n\n", _Zero_Values[i].category,
                               _Zero_Values[i].item, string.Text());
                 }
            }

            int first = 1;
            for (i = 0; i < NUM_UPPER_CASE; i++) {
                 Table = _CifObj->GetTablePtr(_StructureId, _Upper_Case[i].category);
                 if (is_empty_table(Table)) continue;
                 high_index = Table->GetColumnIndex(_Upper_Case[i].item);
                 if (high_index < 0) continue;
                 rowNo = Table->GetNumRows();
                 for (j = 0; j < rowNo; j++) {
                      get_value(string, Table, high_index, j); 
                      if (string == "") continue;
                      len = strlen(string.Text()); 
                      for (k = 0; k < len; k++) {
                           if (isalpha(string.Text()[k]) && islower(string.Text()[k])) break;
                      }
                      if (k == len) {
                           string1.Clear();
                           string1 += "The value '";
                           string1 += string;
                           string1 += "' in _";
                           string1 += _Upper_Case[i].category;
                           string1 += ".";
                           string1 += _Upper_Case[i].item;
                           string1 += " is all in upper case.";
                           first = 1;
                           p = string1.Text();
                           while (p) {
                                p = get_max_length_word(p, buffer, 70);
                                if (!strcmp(buffer, "")) continue;
                                if (first) fprintf(fp, "WARNING: %s\n", buffer);
                                else       fprintf(fp, "         %s\n", buffer);
                                first = 0;
                           }
                           fprintf(fp, "\n");
/*
                           fprintf(fp, "WARNING: The value '%s' in _%s.%s is all in upper case\n\n",
                               string.Text(), _Upper_Case[i].category, _Upper_Case[i].item);
*/
                           _Error_Check[44 + i]++;
                      }
                      for (k = 0; k < len; k++) {
                           if (isalpha(string.Text()[k]) && isupper(string.Text()[k])) break;
                      }
                      if (k == len) {
                           string1.Clear();
                           string1 += "The value '";
                           string1 += string;
                           string1 += "' in _";
                           string1 += _Upper_Case[i].category;
                           string1 += ".";
                           string1 += _Upper_Case[i].item;
                           string1 += " is all in lower case.";
                           first = 1;
                           p = string1.Text();
                           while (p) {
                                p = get_max_length_word(p, buffer, 70);
                                if (!strcmp(buffer, "")) continue;
                                if (first) fprintf(fp, "WARNING: %s\n", buffer);
                                else       fprintf(fp, "         %s\n", buffer);
                                first = 0;
                           }
                           fprintf(fp, "\n");
/*
                           fprintf(fp, "WARNING: The value '%s' in _%s.%s is all in lower case\n\n",
                               string.Text(), _Upper_Case[i].category, _Upper_Case[i].item);
*/
                           _Error_Check[44 + i]++;
                      }
                 }
            }

            for (i = 0; i < NUM_EXTRA_SPACE_CASE; i++) {
                 Table = _CifObj->GetTablePtr(_StructureId, _Extra_Space_Case[i].category);
                 if (is_empty_table(Table)) continue;
                 high_index = Table->GetColumnIndex(_Extra_Space_Case[i].item);
                 if (high_index < 0) continue;
                 rowNo = Table->GetNumRows();
                 for (j = 0; j < rowNo; j++) {
                      get_value(string, Table, high_index, j); 
                      if (string == "") continue;
                      if (check_extra_space_in_name(string.Text())) {
                           fprintf(fp, "WARNING: Extra space in the value '%s' of _%s.%s\n",
                                    string.Text(), _Extra_Space_Case[i].category,
                                      _Extra_Space_Case[i].item); 
                      }
                 }
            }

            for (i = 0; i < NUM_CATEGORY_VALUE; i++) {
                 Table = _CifObj->GetTablePtr(_StructureId, _Check_Category_Values[i].category);
                 if (is_empty_table(Table)) continue;
                 high_index = Table->GetColumnIndex(_Check_Category_Values[i].item);
                 if (high_index < 0) continue;
                 Table->GetCell(string, high_index, 0);
                 if (is_not_empty(string)) {
                      if (!_Check_Category_Values[i].op &&
                          atof(string.Text()) < _Check_Category_Values[i].value ||
                          _Check_Category_Values[i].op &&
                          atof(string.Text()) > _Check_Category_Values[i].value) {
                           fprintf(fp, "WARNING: The value of _%s.%s '%s' is ",
                                 _Check_Category_Values[i].category,
                                 _Check_Category_Values[i].item, string.Text());
                           if (_Check_Category_Values[i].op)
                                fprintf(fp, "larger than %.1f\n\n", _Check_Category_Values[i].value);
                           else fprintf(fp, "less than %.1f\n\n", _Check_Category_Values[i].value);
                      }
                 }
            }

            for (i = 0; i < NUM_CATEGORY_VALUE1; i++) {
                 Table = _CifObj->GetTablePtr(_StructureId, _Check_Category_Values_1[i].category);
                 if (is_empty_table(Table)) continue;
                 high_index = Table->GetColumnIndex(_Check_Category_Values_1[i].item);
                 if (high_index < 0) continue;
                 Table->GetCell(string, high_index, 0);
                 if (is_not_empty(string)) {
                      if (atof(string.Text()) > _Check_Category_Values_1[i].value) {
                           _Error_Check[10]++;
                           fprintf(fp, "WARNING: The value '%s' in _%s.%s is greater than %.1f, ",
                                 string.Text(), _Check_Category_Values_1[i].category,
                                 _Check_Category_Values_1[i].item,
                                 _Check_Category_Values_1[i].value);
                           if (_Check_Category_Values_1[i].op)
                                fprintf(fp, "value in percentage is expected.\n");
                           else fprintf(fp, "value in ratio is expected.\n");
                      }
                      if (_Check_Category_Values_1[i].op && atof(string.Text()) < 1.0) {
                           _Error_Check[10]++;
                           fprintf(fp, "WARNING: The value '%s' in _%s.%s is less than 1.0, ",
                                 string.Text(), _Check_Category_Values_1[i].category,
                                 _Check_Category_Values_1[i].item);
                           fprintf(fp, "value in percentage is expected.\n");
                                
                      }
                 }
            }

            for (i = 0; i < NUM_UNIQUE_INDEX; i++) {
                 Table = _CifObj->GetTablePtr(_StructureId, _Unique_Index[i].category);
                 if (is_empty_table(Table)) continue;
                 k = 0;
                 for (j = 0; j < _Unique_Index[i].num_items; j++) {
                      index[j] = Table->GetColumnIndex(_Unique_Index[i].items[j]);
                      if (index[j] >= 0) k = 1;
                 }
                 if (!k) continue;
                 _UniIndex.Clear();
                 rowNo = Table->GetNumRows();
                 for (j = 0; j < rowNo; j++) {
                      string.Copy("");
                      for (k = 0; k < _Unique_Index[i].num_items; k++) {
                           str[k].Copy("");
                           Table->GetCell(str[k], index[k], j);
                           string += "_";
                           string += str[k];
                      }
                      if (_UniIndex.Seek(string) >= 0) {
                           if (_Unique_Index[i].error_flag)
                                fprintf(fp, "ERROR: ");
                           else fprintf(fp, "WARNING: ");
                           fprintf(fp, "repeat value(s)");
                           for (k = 0; k < _Unique_Index[i].num_items; k++) {
                                fprintf(fp, " _%s.%s=%s", _Unique_Index[i].category,
                                        _Unique_Index[i].items[k], str[k].Text());
                           }
                           fprintf(fp, "\n\n");
                      } else _UniIndex.Add(string);
                 }
            }

            Table = _CifObj->GetTablePtr(_StructureId, "diffrn_source");
            if (!is_empty_table(Table)) {
                 high_index = Table->GetColumnIndex("rcsb_wavelength");
                 low_index = Table->GetColumnIndex("rcsb_wavelength_list");
                 rowNo = Table->GetNumRows();
                 for (i = 0; i < rowNo; i++) {
                      get_value(string, Table, high_index, i);
                      if (string != "") {
                           len = strlen(string.Text());
                           for (j = 0; j < len; j++) {
                                if (string.Text()[j] != ' ' && string.Text()[j] != '.' &&
                                    string.Text()[j] != ',' && !isdigit(string.Text()[j])) {
                                     fprintf(fp, "WARNING: wrong value '%s' for _diffrn_source.rcsb_wavelength.\n",
                                                string.Text());
                                     _Error_Check[32]++;
                                     break;
                                }
                           }
                      }
                      get_value(string, Table, low_index, i);
                      if (string != "") {
                           len = strlen(string.Text());
                           for (j = 0; j < len; j++) {
                                if (string.Text()[j] != ' ' && string.Text()[j] != '.' &&
                                    string.Text()[j] != ',' && !isdigit(string.Text()[j])) {
                                     fprintf(fp, "WARNING: wrong value '%s' for _diffrn_source.rcsb_wavelength_list.\n",
                                                string.Text());
                                     _Error_Check[32]++;
                                     break;
                                }
                           }
                      }
                 }
            }

            Table = _CifObj->GetTablePtr(_StructureId, "refine");
            if (!is_empty_table(Table)) {
                 high_index = Table->GetColumnIndex("ndb_method_to_determine_struct");
                 get_value(string, Table, high_index, 0);
                 if (!strcasecmp(string.Text(), "MAD")) {
                      Table = _CifObj->GetTablePtr(_StructureId, "diffrn_radiation");
                      if (!is_empty_table(Table) && ((rowNo = Table->GetNumRows()) == 1)) {
                           high_index = Table->GetColumnIndex("rcsb_diffrn_protocol");
                           get_value(string1, Table, high_index, 0);
                           if (!strcasecmp(string1.Text(), "single wavelength")) {
                                _Error_Check[30]++;
                                fprintf(fp, "WARNING: _refine.ndb_method_to_determine_struct is MAD");
                                fprintf(fp, " but _diffrn_radiation.rcsb_diffrn_protocol is ");
                                fprintf(fp, "\"single wavelength\"\n");
                           }
                      }
                      Table = _CifObj->GetTablePtr(_StructureId, "diffrn_source");
                      if (!is_empty_table(Table) && ((rowNo = Table->GetNumRows()) == 1)) {
                           high_index = Table->GetColumnIndex("rcsb_wavelength_list");
                           get_value(string1, Table, high_index, 0);
                           high_index = get_wavelength_list(string1.Text());
                           if (high_index == 0) {
                                _Error_Check[30]++;
                                fprintf(fp, "WARNING: _refine.ndb_method_to_determine_struct is MAD");
                                fprintf(fp, " but there is no value in _diffrn_source.rcsb_wavelength_list\n");
                           } else if (high_index == 1) {
                                _Error_Check[30]++;
                                fprintf(fp, "WARNING: _refine.ndb_method_to_determine_struct is MAD");
                                fprintf(fp, " but there is only %s in _diffrn_source.rcsb_wavelength_list\n", string1.Text());
                                fprintf(fp, "         Please provide all wavelengths used for the MAD technique.\n");
                           }
                      }
                 }
            }

            Table = _CifObj->GetTablePtr(_StructureId, "struct_keywords");
            if (is_empty_table(Table)) {
                 _Error_Check[9]++;
                 fprintf(fp, "ERROR: Missing KEYWDS records\n\n");
            } else {
                 high_index = Table->GetColumnIndex("text");
                 if (high_index < 0) {
                      fprintf(fp, "ERROR: Missing KEYWDS records\n\n");
                      _Error_Check[9]++;
                 } else {
                      get_value(string, Table, high_index, 0);
                      if (string == "") {
                           _Error_Check[9]++;
                           fprintf(fp, "ERROR: Missing KEYWDS records\n\n");
                      }
                 }
            }
            Table = _CifObj->GetTablePtr(_StructureId, "struct");
            if (is_empty_table(Table)) {
                 _Error_Check[28]++;
                 fprintf(fp, "ERROR: Missing TITLE record\n\n");
                 fprintf(fp, "ERROR: Missing _struct.ndb_descriptor record\n\n");
            } else {
                 high_index = Table->GetColumnIndex("title");
                 if (high_index < 0) {
                      fprintf(fp, "ERROR: Missing TITLE record\n\n");
                      _Error_Check[28]++;
                 } else {
                      Table->GetCell(string, high_index, 0);
                      get_value(string, Table, high_index, 0);
                      if (string == "") {
                           _Error_Check[28]++;
                           fprintf(fp, "ERROR: Missing TITLE record\n\n");
                      } else {
                           get_wordlist_from_string_separated_by_delimits(wordlist,
                                       string.Text(), "/");
                           bool found_4_character = false;
                           for (unsigned int i = 0; i < wordlist.Length(); ++i) {
                                if (wordlist[i].Length() > 3) {
                                     found_4_character = true;
                                     break;
                                }
                           }
                           if (!found_4_character) {
                                 fprintf(fp, "WARNING: TITLE is '%s'.\n\n", string.Text());
                           }
                      }
                 }

                 high_index = Table->GetColumnIndex("ndb_descriptor");
                 if (high_index < 0) {
                      fprintf(fp, "ERROR: Missing _struct.ndb_descriptor record\n\n");
                 } else {
                      Table->GetCell(string, high_index, 0);
                      get_value(string, Table, high_index, 0);
                      if (string == "") {
                           fprintf(fp, "ERROR: Missing _struct.ndb_descriptor record\n\n");
                      } else {
                           bool found_character = false;
                           for (unsigned int i = 0; i < string.Length(); ++i) {
                                if (isalpha(string[i])) {
                                     found_character = true;
                                     break;
                                }
                           }
                           if (!found_character) {
                                 fprintf(fp, "WARNING: _struct.ndb_descriptor has value '%s'.\n\n",
                                             string.Text());
                           }
                      }
                 }
            }
            Table = _CifObj->GetTablePtr(_StructureId, "ndb_database_proc");
            if (is_empty_table(Table)) {
                 _Error_Check[22]++;
                 fprintf(fp, "WARNING: No ndb_database_proc category\n\n");
            } else {
                 high_index = Table->GetColumnIndex("cycle_id");
                 low_index = Table->GetColumnIndex("date_end_cycle");
                 if (high_index < 0) {
                      _Error_Check[22]++;
                      fprintf(fp, "WARNING: Missing cycle_id in ndb_database_proc category\n\n");
                 } else {
                      rowNo = Table->GetNumRows();
                      for (i = 0; i < rowNo; i++) {
                           get_value(res_high, Table, high_index, i);
                           get_value(res_low, Table, low_index, i);
                           if (res_high == "1") break;
                      }
                      if (i == rowNo) {
                           _Error_Check[22]++;
                           fprintf(fp, "WARNING: Missing cycle_id '1' in ndb_database_proc category\n\n");
                      } else if (res_low == "") {
                           _Error_Check[22]++;
                           fprintf(fp, "WARNING: Missing date in ndb_database_proc.date_end_cycle for cycle_id '1'\n\n");
                      }
                      get_date(today_date);
                      high_index = Table->GetColumnIndex("date_begin_cycle");
                      for (i = 0; i < rowNo; i++) {
                           get_value(res_high, Table, high_index, i);
                           get_value(res_low, Table, low_index, i);
                           if (strcmp(res_high.Text(), res_low.Text()) > 0) {
                                fprintf(fp, "ERROR: _ndb_database_proc.date_end_cycle '%s' is before _ndb_database_proc.date_begin_cycle '%s'.\n",
                                            res_low.Text(), res_high.Text());
                                sprintf(buffer, " end date %s before begin date %s,",  res_low.Text(), res_high.Text());
                                _date_error += buffer;
                                _Error_Check[29]++;
                           }
                           if (strcmp(res_high.Text(), today_date) > 0) {
                                fprintf(fp, "ERROR: _ndb_database_proc.date_begin_cycle '%s' is after today '%s'\n",
                                            res_high.Text(), today_date);
                                sprintf(buffer, " future date %s,", res_high.Text());
                                _date_error += buffer;
                                _Error_Check[29]++;
                           }
                           if (strcmp(res_low.Text(), today_date) > 0) {
                                fprintf(fp, "ERROR: _ndb_database_proc.date_end_cycle '%s' is after today '%s'\n",
                                            res_low.Text(), today_date);
                                sprintf(buffer, " future date %s,", res_low.Text());
                                _date_error += buffer;
                                _Error_Check[29]++;
                           }
                      }
                 }
            }

            pdb_id[0] = '\0';
            Table = _CifObj->GetTablePtr(_StructureId, "database_2");
            if (!is_empty_table(Table)) get_pdb_id(Table, pdb_id);

            Table = _CifObj->GetTablePtr(_StructureId, "database");
            if (!is_empty_table(Table)) {
                 high_index = Table->GetColumnIndex("ndb_code_PDB");
                 if (high_index >= 0) {
                      get_value(res_low, Table, high_index, 0);
                      if (strcasecmp(res_low.Text(), pdb_id)) {
                           fprintf(fp, "WARNING: PDB ID '%s' in database category is different", res_low.Text());
                           fprintf(fp, " '%s' from database_2 category\n\n", pdb_id);
                           _Error_Check[11]++;
                      }
                 }
            }

            memset(date_original, 0, 20);
            Table = _CifObj->GetTablePtr(_StructureId, "database_PDB_rev");
            if (!is_empty_table(Table)) {
                 high_index = Table->GetColumnIndex("date_original");
                 get_value(res_high, Table, high_index, 0);
                 strcpy(date_original, res_high.Text());
                 rowNo = Table->GetNumRows();
                 high_index = Table->GetColumnIndex("num");
                 low_index = Table->GetColumnIndex("replaces");
                 // int idx_index = Table->GetColumnIndex("date");
                 if (high_index >= 0 && low_index >= 0) {
                      for (i = 0; i < rowNo; i++) {
/*
                           get_value(res_high, Table, idx_index, i);
                           if (res_high == "2011-07-13") {
                                _Error_Check[79]++;
                                fprintf(fp, "WARNING: Entry has remediation VERSN REVDAT (13-JUL-11)\n\n");
                           }
*/
                           get_value(res_high, Table, high_index, i);
                           get_value(res_low, Table, low_index, i);
                           if (atoi(res_high.Text()) > 0 && strcasecmp(res_low.Text(), pdb_id)) {
                                fprintf(fp, "WARNING: PDB ID '%s' in database_PDB_rev category is different", res_low.Text());
                                fprintf(fp, " '%s' from database_2 category\n\n", pdb_id);
                                _Error_Check[11]++;
                           }
                      }
                 }
            }

            if (strcmp(date_original, "")) {
                 Table = _CifObj->GetTablePtr(_StructureId, "diffrn_detector");
                 if (!is_empty_table(Table)) {
                      high_index = Table->GetColumnIndex("ndb_collection_date");
                      get_value(res_high, Table, high_index, 0);
                      if (res_high != "" && strcmp(res_high.Text(), date_original) > 0) {
                           fprintf(fp, "WARNING: _database_PDB_rev.date_original '%s' is before _diffrn_detector.ndb_collection_date '%s'.\n",
                                   date_original, res_high.Text());
                           sprintf(buffer, " deposit date %s before data collection date %s,", date_original, res_high.Text());
                           _date_error += buffer;
                           _Error_Check[29]++;
                      }
                 }
            }

            Table = _CifObj->GetTablePtr(_StructureId, "atom_sites");
            if (!is_empty_table(Table) && !_cell->is_artifical) {
                 index[0] = Table->GetColumnIndex("fract_transf_matrix[1][1]");
                 index[1] = Table->GetColumnIndex("fract_transf_matrix[2][2]");
                 index[2] = Table->GetColumnIndex("fract_transf_matrix[3][3]");
                 index[3] = Table->GetColumnIndex("fract_transf_matrix[1][2]");
                 index[4] = Table->GetColumnIndex("fract_transf_matrix[1][3]");
                 index[5] = Table->GetColumnIndex("fract_transf_matrix[2][1]");
                 index[6] = Table->GetColumnIndex("fract_transf_matrix[2][3]");
                 index[7] = Table->GetColumnIndex("fract_transf_matrix[3][1]");
                 index[8] = Table->GetColumnIndex("fract_transf_matrix[3][2]");
                 e_index = 0;
                 for (i = 0; i < 9; i++) {
                      get_value(res_high, Table, index[i], 0);
                      if (res_high != "") {
                           e_index++;
                           value[i] = atof(res_high.Text());
                      }
                 }
                 if (e_index) {
                      if (e_index < 9)
                           fprintf(fp, "ERROR: An incomplete matrix is present in atom_sites category\n\n");
                      else {
                           is_idential = 1;
                           for (i = 0; i < 3; i++) {
                                if (fabs(value[i] - 1.0) > 0.00001) {
                                     is_idential = 0;
                                     break;
                                }
                           }
                           for (i = 0; i < 3; i++) {
                                if (fabs(value[i]) > 0.00001) {
                                     e_index = 0;
                                     break;
                                }
                           }
                           for (i = 3; i < 9; i++) {
                                if (fabs(value[i]) > 0.00001) {
                                     is_idential = 0;
                                     e_index = 0;
                                     break;
                                }
                           }
                           if (is_idential)
                                fprintf(fp, "ERROR: An identity matrix is present in atom_sites category\n\n");
                           if (e_index)
                                fprintf(fp, "ERROR: All values are zero in atom_sites category\n\n");
                           if (!is_idential && !e_index) {
                                aa[0] = _cell->a;
                                aa[1] = _cell->b;
                                aa[2] = _cell->c;
                                ang[0] = _cell->alpha;
                                ang[1] = _cell->beta;
                                ang[2] = _cell->gamma;
                                rad = acos(-1.0)/180.0;
                                for (i = 0; i < 3; i++) {
                                     alphr[i] = ang[i] * rad;
                                     cosa[i]  = cos(alphr[i]);
                                     sina[i]  = sin(alphr[i]);
                                     if (fabs(ang[i] - 90.0) < 0.001) {
                                          cosa[i] = 0.0;
                                          sina[i] = 1.0;
                                     }
                                }
                                va = sqrt(1.00 + 2.0 * cosa[0] * cosa[1] * cosa[2]
                                   - cosa[0] * cosa[0] - cosa[1] * cosa[1]
                                   - cosa[2] * cosa[2]);

                                svalue[0] = 1.0 / aa[0];
                                svalue[1] = 1.0 / (aa[1] * sina[2]);
                                svalue[2] = sina[2] / (aa[2] * va);
                                svalue[3] = -cosa[2] / (aa[0] * sina[2]);
                                svalue[4] = (cosa[2] * cosa[0]-cosa[1]) / (aa[0] * va * sina[2]);
                                svalue[5] = 0.0;
                                svalue[6] = (cosa[1] * cosa[2]-cosa[0]) / (aa[1] * va * sina[2]);
                                svalue[7] = 0.0;
                                svalue[8] = 0.0;
                                for (i = 0; i < 9; i++) {
                                     if (fabs(svalue[i] - value[i]) > 0.00005) break;
                                }
                                if (i < 9) {
                                     fprintf(fp, "WARNING: Scale matrices is different from calculated matrices.\n\n");
                                     _Error_Check[42]++; 
                                }
                           }
                      }
                 }
            }

            Table = _CifObj->GetTablePtr(_StructureId, "entity");
            if (!is_empty_table(Table)) {
                 rowNo = Table->GetNumRows();
                 high_index = Table->GetColumnIndex("id");
                 low_index = Table->GetColumnIndex("ndb_description");
                 src_index = Table->GetColumnIndex("src_method");

                 for (i = 0; i < rowNo; i++) {
                      get_value(res_high, Table, high_index, i);
                      if (polymer_entity_ids.find(res_high.Text()) ==
                          polymer_entity_ids.end()) continue;

                      get_value(res_low, Table, low_index, i);
                      if (res_low == "") {
                           fprintf(fp, "ERROR: Missing polymer description for entity ");
                           fprintf(fp, "'%s' in _entity.ndb_description.\n\n", res_high.Text());
                      }
                      get_value(buff, Table, src_index, i);
                      buff.ToLower();
                      if (buff == "man") {
                           check_missing_source(fp, _Error_Check[12], _CifObj,
                                        _StructureId, res_high, "entity_src_gen",
                                        "ndb_gene_src_scientific_name",
                                        "pdbx_gene_src_ncbi_taxonomy_id");
                           check_missing_source(fp, _Error_Check[12], _CifObj, 
                                         _StructureId, res_high, "entity_src_gen",
                                         "ndb_host_org_scientific_name",
                                         "pdbx_host_org_ncbi_taxonomy_id");
                      } else if (buff == "nat")
                           check_missing_source(fp, _Error_Check[12], _CifObj, 
                                         _StructureId, res_high, "entity_src_nat",
                                         "ndb_organism_scientific",
                                         "pdbx_ncbi_taxonomy_id");
                      else if (buff == "syn")
                           check_missing_source(fp, _Error_Check[12], _CifObj,  
                                         _StructureId, res_high, "rcsb_entity_src_syn",
                                         "organism_scientific", "ncbi_taxonomy_id");
                      else {
                           get_value(buff, Table, src_index, i);
                           fprintf(fp, "ERROR: Incorrect value '%s' in _entity.src_method for entity '%s'.\n",
                                       res_high.Text(), buff.Text());
                      }
                 }
            }

            Table = _CifObj->GetTablePtr(_StructureId, "struct_biol_gen");
            if (!is_empty_table(Table) && _num_mol) {
                 rowNo = Table->GetNumRows();
                 high_index = Table->GetColumnIndex("asym_id");
                 low_index = Table->GetColumnIndex("ndb_new_pdb_asym_id");
                 if (high_index >= 0 && low_index >= 0) {
                      for (i = 0; i < rowNo; i++) {
                           get_value(res_high, Table, high_index, i);
                           get_value(res_low, Table, low_index, i);
                           chain = _molecules[_rep_mol]->GetFirstChain();
                           while (chain) {
                                if (res_high == chain->ChainID() && /*
                                   (chain->chain_type() == ATOMP_TOKEN ||
                                    chain->chain_type() == ATOMN_TOKEN) && */
                                    res_low == "") {
                                     _Error_Check[13]++;
                                     // fprintf(fp, "WARNING: NO PDB Chain ID for macromolecule(s)\n\n");
                                     fprintf(fp, "ERROR: NO PDB Chain ID assigned for asym ID %s.\n\n", chain->ChainID());
                                }
                                chain = _molecules[_rep_mol]->GetNextChain();
                           }
                      }
                 }
            }

            Table = _CifObj->GetTablePtr(_StructureId, "struct_ref");
            if (!is_empty_table(Table)) {
                 rowNo = Table->GetNumRows();
                 high_index = Table->GetColumnIndex("db_name");
                 if (high_index >= 0) {
                      for (i = 0; i < rowNo; i++) {
                           get_value(res_high, Table, high_index, i);
                           if (strcasecmp(res_high.Text(), "PIR") &&
                               strcasecmp(res_high.Text(), "SWS") &&
                               strcasecmp(res_high.Text(), "UNP") &&
                               strcasecmp(res_high.Text(), "NOR") &&
                               strcasecmp(res_high.Text(), "GB") &&
                               strcasecmp(res_high.Text(), "PRF") &&
                               strcasecmp(res_high.Text(), "PDB") &&
                               strcasecmp(res_high.Text(), "EMBL")) {
                                fprintf(fp, "WARNING: db_name=%s in struct_ref category\n\n", res_high.Text());
                                _Error_Check[14]++;
                           }
                      }
                 }
            }

            rowNo1 = 0;
            Table1 = _CifObj->GetTablePtr(_StructureId, "struct_biol_gen");
            if (!is_empty_table(Table1)) {
                 rowNo1 = Table1->GetNumRows();
                 src_index = Table1->GetColumnIndex("ndb_new_pdb_asym_id");
            } else Table1 = NULL;

            Table = _CifObj->GetTablePtr(_StructureId, "struct_ref_seq");
            if (!is_empty_table(Table)) {
                 rowNo = Table->GetNumRows();
                 high_index = Table->GetColumnIndex("ndb_chain_id");
                 low_index = Table->GetColumnIndex("ndb_pdb_id_code");
                 e_index = Table->GetColumnIndex("ndb_auth_seq_align_beg");
                 o_index = Table->GetColumnIndex("ndb_auth_seq_align_end");
                 if (high_index < 0) fprintf(fp, "WARNING: No PDB Chain ID(s) in struct_ref_seq category\n\n");
                 if (low_index < 0) fprintf(fp, "WARNING: No PDB ID in struct_ref_seq category\n\n");

                 for (i = 0; i < rowNo; i++) {
                      get_value(res_high, Table, high_index, i);
                      if (res_high != "") {
                           if (pdb_chain_ids.find(res_high.Text()) != pdb_chain_ids.end())
                                pdb_chain_ids.erase(res_high.Text());
                      }
                      if (high_index >= 0 && low_index >= 0) {
                           get_value(res_high, Table, high_index, i);
                           get_value(res_low, Table, low_index, i);
                           if (res_high == "") {
                                _Error_Check[14]++;
                                _Error_Check[13]++;
                                fprintf(fp, "WARNING: No PDB Chain ID(s) in struct_ref_seq category\n\n");
                           } else if (Table1 && rowNo1) {
                                for (j = 0; j < rowNo1; j++) {
                                     Table1->GetCell(string, src_index, j);
                                     if (string == res_high) break;
                                }
                                if (j == rowNo1) {
                                     _Error_Check[14]++;
                                     _Error_Check[13]++;
                                     fprintf(fp, "ERROR: PDB Chain ID(s) mismatch in struct_ref_seq category\n\n");
                                }
                           }
                           if (strcasecmp(res_low.Text(), pdb_id)) {
                                _Error_Check[14]++;
                                _Error_Check[11]++;
                                fprintf(fp, "WARNING: PDB ID '%s' in struct_ref_seq category is different", res_low.Text());
                                fprintf(fp, " '%s' from database_2 category\n\n", pdb_id);
                           }
                      }
                      get_value(res_high, Table, e_index, i);
                      if (res_high != "") {
                           for (j = 0; j < (int) strlen(res_high.Text()); j++) {
                                if (isalpha(res_high.Text()[j])) {
                                     fprintf(fp, "WARNING: wrong value '%s' in _struct_ref_seq.ndb_auth_seq_align_beg\n\n",
                                                     res_high.Text());
                                     _Error_Check[14]++;
                                     break;
                                }
                           }
                      }
                      get_value(res_high, Table, o_index, i);
                      if (res_high != "") {
                           for (j = 0; j < (int) strlen(res_high.Text()); j++) {
                                if (isalpha(res_high.Text()[j])) {
                                     fprintf(fp, "WARNING: wrong value '%s' in _struct_ref_seq.ndb_auth_seq_align_end\n\n",
                                                     res_high.Text());
                                     _Error_Check[14]++;
                                     break;
                                }
                           }
                      }
                 }
            }

            for (std::set<std::string>::const_iterator
                 spos = pdb_chain_ids.begin(); spos != pdb_chain_ids.end(); ++spos) {
                 fprintf(fp, "ERROR: Missing DBREF for polymer chain '%s'.\n\n", spos->c_str());
                 _Error_Check[14]++;
            }

            Table = _CifObj->GetTablePtr(_StructureId, "struct_ref_seq_dif");
            if (!is_empty_table(Table)) {
                 rowNo = Table->GetNumRows();
                 high_index = Table->GetColumnIndex("ndb_pdb_chain_id");
                 low_index = Table->GetColumnIndex("ndb_pdb_id_code");
                 o_index = Table->GetColumnIndex("ndb_auth_seq_num"); 
                 if (high_index < 0) fprintf(fp, "WARNING: No PDB Chain ID(s) in struct_ref_seq_dif category\n\n");
                 if (low_index < 0) fprintf(fp, "WARNING: No PDB ID in struct_ref_seq_dif category\n\n");

                 for (i = 0; i < rowNo; i++) {
                      if (high_index >= 0 && low_index >= 0) {
                           get_value(res_high, Table, high_index, i);
                           get_value(res_low, Table, low_index, i);
                           if (res_high == "") {
                                _Error_Check[15]++;
                                _Error_Check[13]++;
                                fprintf(fp, "WARNING: No PDB Chain ID(s) in struct_ref_seq_dif category\n\n");
                           } else if (Table1 && rowNo1) {
                                for (j = 0; j < rowNo1; j++) {
                                     Table1->GetCell(string, src_index, j);
                                     if (string == res_high) break;
                                }
                                if (j == rowNo1) {
                                     _Error_Check[15]++;
                                     _Error_Check[13]++;
                                     fprintf(fp, "ERROR: PDB Chain ID(s) mismatch in struct_ref_seq_dif category\n\n");
                                }
                           } 
                           if (strcasecmp(res_low.Text(), pdb_id)) {
                                _Error_Check[15]++;
                                _Error_Check[11]++;
                                fprintf(fp, "WARNING: PDB ID '%s' in struct_ref_seq_dif category is different", res_low.Text());
                                fprintf(fp, " '%s' from database_2 category\n\n", pdb_id);
                           }
                      }
                      get_value(res_high, Table, o_index, i);
                      if (res_high != "") {
                           for (j = 0; j < (int) strlen(res_high.Text()); j++) {
                                if (isalpha(res_high.Text()[j])) {
                                     fprintf(fp, "WARNING: wrong value '%s' in _struct_ref_seq_dif.ndb_auth_seq_num\n\n",
                                                     res_high.Text());
                                     _Error_Check[15]++;
                                     break;
                                }
                           }
                      }
                 }
            }

            Table = _CifObj->GetTablePtr(_StructureId, "rcsb_database_related");
            if (!is_empty_table(Table)) {
                 rowNo = Table->GetNumRows();
                 high_index = Table->GetColumnIndex("db_name");
                 low_index = Table->GetColumnIndex("db_id");
                 for (i = 0; i < rowNo; i++) {
                      get_value(res_high, Table, high_index, i);
                      get_value(res_low, Table, low_index, i);
                      if (res_low != "" && res_high == "") {
                           fprintf(fp, "WARNING: No DB_NAME for related entry %s in rcsb_database_related category\n\n",
                                          res_low.Text());
                           _Error_Check[16]++;
                      }
                      if (res_low == "" && res_high != "") {
                           fprintf(fp, "WARNING: Not a valid ID in rcsb_database_related category\n\n");
                           _Error_Check[16]++;
                      }
                 }
            }

            Table = _CifObj->GetTablePtr(_StructureId, "rcsb_missing_residue_list");
            if (!is_empty_table(Table)) {
                 rowNo = Table->GetNumRows();
                 high_index = Table->GetColumnIndex("pdb_chain_id");
                 if (high_index < 0) {
                      fprintf(fp, "WARNING: No PDB Chain ID(s) in rcsb_missing_residue_list category\n\n");
                      _Error_Check[13]++;
                 } else {
                      for (i = 0; i < rowNo; i++) {
                           get_value(res_high, Table, high_index, i);
                           if (res_high == "") {
                                fprintf(fp, "WARNING: No PDB Chain ID(s) in rcsb_missing_residue_list category\n\n");
                                _Error_Check[13]++;
                           }
                      }
                 }
            }

            _UniIndex.Clear();
            Table = _CifObj->GetTablePtr(_StructureId, "citation");
            if (!is_empty_table(Table)) {
                 rowNo = Table->GetNumRows();
                 id_idx = Table->GetColumnIndex("id");
                 pubmed_idx = Table->GetColumnIndex("rcsb_database_id_PubMed");
                 high_index = Table->GetColumnIndex("journal_abbrev");
                 low_index = Table->GetColumnIndex("journal_id_ISSN");
                 e_index = Table->GetColumnIndex("journal_volume");
                 s_index = Table->GetColumnIndex("page_first");
                 o_index = Table->GetColumnIndex("page_last");
                 int t_index = Table->GetColumnIndex("title");
                 for (i = 0; i < rowNo; i++) {
                      // string.Clear();
                      // Table->GetCell(string, id_idx, i);
                      get_value(string, Table, id_idx, i);
                      get_value(string1, Table, t_index, i);
                      if (string != "" && string1 != "") {
                           if (_UniIndex.Seek(string) < 0) _UniIndex.Add(string);
                      }
                      if (!strcasecmp(string.Text(), "primary")) {
                           string.Clear();
                           Table->GetCell(string, high_index, i);
                           if (string != "." && string != "" && string != "?" &&
                               strcasecmp(string.Text(), "To be Published") &&
                               strcasecmp(string.Text(), "Unpublished")) {
                                string.Clear();
                                Table->GetCell(string, pubmed_idx, i);
                                if (string == "" || string == "." || string == "?") {
                                     fprintf(fp, "WARNING: No PubMed ID for primary citation\n\n");
                                     _Error_Check[35]++;
                                }
                           }
                      }
                      string.Clear();
                      Table->GetCell(string, high_index, i);
                      if (string != "." && string != "" && string != "?" &&
                          strcasecmp(string.Text(), "To be Published")) {
                           get_value(res_low, Table, low_index, i);
                           if (res_low == "") {
                                fprintf(fp, "WARNING: No ISSN for Journal '%s'\n\n", string.Text());
                                _Error_Check[17]++;
                           }
                      }
                      if (!strcasecmp(string.Text(), "To be Published")) {
                           get_value(res_low, Table, e_index, i);
                           get_value(string1, Table, s_index, i);
                           get_value(string2, Table, o_index, i);
                           if (res_low != "" || string1 != "" || string2 != "") {
                                fprintf(fp, "WARNING: Journal name is \"to be published\" ");
                                if (res_low != "" && (string1 != "" || string2 != ""))
                                     fprintf(fp, "but volume and page numbers exist.\n\n");
                                else if (res_low != "")
                                     fprintf(fp, "but volume number exists.\n\n");
                                else if (string1 != "" || string2 != "")
                                     fprintf(fp, "but page numbers exist.\n\n");
                           }
                      }
                 }
            }

            _BiolIndex.Clear();
            Table = _CifObj->GetTablePtr(_StructureId, "citation_author");
            if (!is_empty_table(Table)) {
                 rowNo = Table->GetNumRows();
                 id_idx = Table->GetColumnIndex("citation_id");
                 for (i = 0; i < rowNo; i++) {
                      get_value(string, Table, id_idx, i);
                      if (string != "") {
                           if (_BiolIndex.Seek(string) < 0) _BiolIndex.Add(string);
                      }
                 }
            }
            Table = _CifObj->GetTablePtr(_StructureId, "citation_editor");
            if (!is_empty_table(Table)) {
                 rowNo = Table->GetNumRows();
                 id_idx = Table->GetColumnIndex("citation_id");
                 for (i = 0; i < rowNo; i++) {
                      get_value(string, Table, id_idx, i);
                      if (string != "") {
                           if (_BiolIndex.Seek(string) < 0) _BiolIndex.Add(string);
                      }
                 }
            }

            if (_BiolIndex.Root() >= 0 || _UniIndex.Root() >= 0) {
                 if (_BiolIndex.Root() >= 0) {
                      ret = _BiolIndex.GoFirst();
                      while (ret >= 0) {
                           string = _BiolIndex.GetCurrData();
                           if (_UniIndex.Seek(string) < 0) {
                                fprintf(fp, "WARNING: citation id %s in citation_author/citation_editor ", string.Text());
                                fprintf(fp, "categories did not exist in citation category.\n");
                                _Error_Check[17]++;
                           }
                           ret = _BiolIndex.GoToNext();
                      }
                 }
                 if (_UniIndex.Root() >= 0) {
                      ret = _UniIndex.GoFirst();
                      while (ret >= 0) {
                           string = _UniIndex.GetCurrData();
                           if (_BiolIndex.Seek(string) < 0) {
                                fprintf(fp, "WARNING: Please fill in citation authors since the primary citation is provided.\n");
                                _Error_Check[17]++;
                           }
                           ret = _UniIndex.GoToNext();
                      }
                 }
            }
          
            Table = _CifObj->GetTablePtr(_StructureId, "pdbx_SG_project");
            if (!is_empty_table(Table)) {
                 high_index = Table->GetColumnIndex("initial_of_center");
                 low_index = Table->GetColumnIndex("full_name_of_center");
                 e_index = Table->GetColumnIndex("project_name");
                 rowNo = Table->GetNumRows();
                 int n_center = 0;
                 for (i = 0; i < rowNo; i++) {
                      Table->GetCell(string, high_index, i);
                      Table->GetCell(string1, low_index, i);
                      Table->GetCell(string2, e_index, i);
                      if (!strcasecmp(string2.Text(), "Protein Structure Initiative") ||
                          !strcasecmp(string2.Text(), "PSI, Protein Structure Initiative")) {
                           if (string1 == "" || string1 == "?" || string1 == ".") {
                                fprintf(fp, "WARNING: _pdbx_SG_project.project_name is PSI ");
                                fprintf(fp, "but _pdbx_SG_project.full_name_of_center is empty. ");
                                fprintf(fp, "Please fill in the center name.\n");
                           } else {
                                for (j = 0; j < num_sg_centers; j++) {
                                     if (!strcasecmp(string1.Text(), _sg_info[j]._center_name) &&
                                         _sg_info[j].psi_flag) break;
                                }
                                if (j == num_sg_centers) {
                                     fprintf(fp, "WARNING: _pdbx_SG_project.project_name is PSI ");
                                     fprintf(fp, "but _pdbx_SG_project.full_name_of_center is filled in as\n'%s'", string1.Text());
                                     fprintf(fp, " which is not a PSI center.\n");
                                }
                           }
                      }

                      for (j = 0; j < num_sg_centers; j++) {
                           if (!strcasecmp(string.Text(), _sg_info[j]._center_initial)) break;
                      }
                      if (j < num_sg_centers) n_center++;
                 }
                 if (n_center && !has_targetdb) {
                      _Error_Check[39]++;
                      fprintf(fp, "WARNING: _pdbx_SG_project.full_name_of_center is filled in ");
                      fprintf(fp, "but _entity_poly.pdbx_target_identifier is empty. ");
                      fprintf(fp, "Please fill in the target ID.\n");
                 }
            }

            Table = _CifObj->GetTablePtr(_StructureId, "ndb_database_status");
            if (!is_empty_table(Table)) {
                 high_index = Table->GetColumnIndex("author_release_status_code");
                 low_index = Table->GetColumnIndex("date_hold_coordinates");
                 Table->GetCell(string, high_index, 0);
                 Table->GetCell(string1, low_index, 0);
                 if ((string == "HOLD" || string == "hold") &&
                     (string1 == "?" || string1 == "." || string1 == "")) {
                      fprintf(fp, "WARNING: No _ndb_database_status.date_hold_coordinates value ");
                      fprintf(fp, "for _ndb_database_status.author_release_status_code=%s\n\n", string.Text());
                 }
                 high_index = Table->GetColumnIndex("dep_release_code_struct_fact");
                 low_index = Table->GetColumnIndex("date_hold_struct_fact");
                 Table->GetCell(string, high_index, 0);
                 Table->GetCell(string1, low_index, 0);
                 if ((string == "HOLD FOR 6 MONTHS" || string == "HOLD FOR 1 YEAR") &&
                     (string1 == "?" || string1 == "." || string1 == "")) {
                      fprintf(fp, "WARNING: No _ndb_database_status.date_hold_struct_fact value ");
                      fprintf(fp, "for _ndb_database_status.dep_release_code_struct_fact=%s\n\n", string.Text());
                 }
            }

            Table = _CifObj->GetTablePtr(_StructureId, "ccp4_refine_tls_group");
            Table1 = _CifObj->GetTablePtr(_StructureId, "ccp4_refine_tls");
            if ((!is_empty_table(Table) && is_empty_table(Table1)) ||
                (is_empty_table(Table) && !is_empty_table(Table1))) {
                 _Error_Check[64]++;
                 if (is_empty_table(Table))
                      fprintf(fp, "ERROR: Missing TLS information in ccp4_refine_tls_group category.\n\n");
                 else if (is_empty_table(Table1))
                      fprintf(fp, "ERROR: Missing TLS information in ccp4_refine_tls category.\n\n");
            }

            for (i = 0; i < 2; i++) {
                 Table = _CifObj->GetTablePtr(_StructureId, range_category[i]);
                 if (is_empty_table(Table)) continue;
                 rowNo = Table->GetNumRows();
                 bool missing_range = false;
                 for (j = 0; j < 2; j++) {
                      high_index = Table->GetColumnIndex(range_items[j][0]);
                      low_index = Table->GetColumnIndex(range_items[j][1]);
                      if (high_index < 0 || low_index < 0) continue;
                      for (k = 0; k < rowNo; k++) {
                           get_value(string1, Table, high_index, k);
                           get_value(string2, Table, low_index, k);
                           if (string1 != "" && string2 == "") {
                                missing_range = true;
                                break;
                           }
                      }
                      if (missing_range) break;
                 }
                 if (missing_range) {
                      fprintf(fp, "WARNING: missing range in %s\n\n", range_category[i]);
                      _Error_Check[43]++;
                 }
            }

            Table = _CifObj->GetTablePtr(_StructureId, "entity_keywords");
            if (!is_empty_table(Table)) {
                 regex_t preg;
                 regmatch_t pmatch[NS];
                 rowNo = Table->GetNumRows();
                 regcomp(&preg, EC_PATTERN, REG_EXTENDED);
                 int e_index = Table->GetColumnIndex("ndb_ec");
                 for (int i = 0; i < rowNo; i++) {
                      get_value(string, Table, e_index, i);
                      if (string != "") {
                           ret = regexec(&preg, string.Text(), NS, pmatch, 0);
                           if (ret != 0) {
                                _Error_Check[69]++;
                                if (_Error_Check_Message[69] != "") {
                                     _Error_Check_Message[69] += ", ";
                                }
                                _Error_Check_Message[69] += string;
                           }
                      }
                 }
                 regfree(&preg);
                 if (_Error_Check[69]) fprintf(fp, "ERROR: Wrong EC number %s.\n\n",
                                _Error_Check_Message[69].Text());
            }
/*
            if (dict) {
                 _Error_Check[23] = DictionaryDataChecking(_CifObj, dict, fp, is_nmr,
                             _Error_Check[29], _date_error);
            }
*/

            if (need_na_checking1) {
                 Table = _CifObj->GetTablePtr(_StructureId, "ndb_entity_func_bind_mode");
                 if (is_empty_table(Table)) {
                      fprintf(fp, "WARNING: _ndb_entity_func_bind_mode is not filled in for protein/NA complex.\n\n");
                 }

                 Table  = _CifObj->GetTablePtr(_StructureId, "ndb_entity_func_enzyme");
                 Table1 = _CifObj->GetTablePtr(_StructureId, "ndb_entity_func_regulatory");
                 Table2 = _CifObj->GetTablePtr(_StructureId, "ndb_entity_func_structural");
                 Table3 = _CifObj->GetTablePtr(_StructureId, "ndb_entity_func_other");
                 if (is_empty_table(Table) && is_empty_table(Table1) &&
                     is_empty_table(Table2) && is_empty_table(Table3)) {
                      fprintf(fp, "WARNING: _ndb_entity_func is not filled in for protein/NA complex.\n\n");
                 }
            }
/*
            if (need_na_checking2) {
                 Table = _CifObj->GetTablePtr(_StructureId, "exptl_crystal_grow_comp");
                 if (is_empty_table(Table)) {
                      fprintf(fp, "WARNING: _exptl_crystal_grow_comp is not filled in for NA structure.\n\n");
                 }
            }
*/

            if (need_na_checking3) {
                 Table = _CifObj->GetTablePtr(_StructureId, "ndb_na_struct_keywds");
                 if (is_empty_table(Table)) {
                      fprintf(fp, "WARNING: _ndb_na_struct_keywds is not filled in for NA structure.\n\n");
                 } else {
                      high_index = Table->GetColumnIndex("conformation_type");
                      low_index = Table->GetColumnIndex("strand_description");
                      Table->GetCell(string, high_index, 0);
                      Table->GetCell(string1, low_index, 0);
                      if ((string == "" || string == "?" || string == ".") &&
                          (string1 == "" || string1 == "?" || string1 == ".")) {
                           fprintf(fp, "WARNING: _ndb_na_struct_keywds is not filled in for NA structure.\n\n");
                      }
                      if (string != "" && string != "?" && string != ".") {
                           fprintf(fp, "Please compare results in _ndb_na_struct_keywds.conformation_type ");
                           fprintf(fp, "and _ndb_struct_conf_na.feature.\n\n");
                      }
                 }

                 Table  = _CifObj->GetTablePtr(_StructureId, "ndb_drug_info");
                 Table1 = _CifObj->GetTablePtr(_StructureId, "ndb_inhibitor_info");
                 Table2 = _CifObj->GetTablePtr(_StructureId, "struct_site_keywords");
                 if (!is_empty_table(Table) || !is_empty_table(Table1)) {
                      if (is_empty_table(Table2)) {
                           fprintf(fp, "WARNING: please check to see if drug binding site needs to ");
                           fprintf(fp, "fill in _struct_site_keywords.text for NA.\n\n");
                      }
                 }
            }

            Table = _CifObj->GetTablePtr(_StructureId, "ndb_poly_seq_scheme");
            int ret, idx1 = -1, idx2 = -1, idx3 = -1, idx4 = -1;
            CifString cs, cs1, cs2, cs3, cs4;
            TblIndexObj residues1, residues2;
            residues1.Clear();
            residues2.Clear();
            if (!is_empty_table(Table)) {
                 rowNo = Table->GetNumRows();
                 idx1  = Table->GetColumnIndex("pdb_id");
                 idx2  = Table->GetColumnIndex("pdb_num");
                 idx3  = Table->GetColumnIndex("pdb_ins_code");
                 idx4  = Table->GetColumnIndex("mon_id");
                 for (i = 0; i < rowNo; i++) {
                      get_value(cs1, Table, idx1, i);
                      get_value(cs2, Table, idx2, i);
                      if (atoi(cs2.Text()) > 9999) _Error_Check[71]++;
                      get_value(cs3, Table, idx3, i);
                      get_value(cs4, Table, idx4, i);
                      if (cs3 != "" && !strstr(cs2.Text(), cs3.Text())) {
                           cs2 += cs3;
                      }
                      cs.Copy(cs1);
                      cs += "_";
                      cs += cs2;
                      cs += "_";
                      cs += cs4;
                      if (residues1.Seek(cs) >= 0) {
/*
                           fprintf(fp, "WARNING: residue '%s %s %s' repeat in ndb_poly_seq_scheme category.\n\n",
                                      cs1.Text(), cs2.Text(), cs4.Text());
*/
                      } else residues1.Add(cs);
                      cs.Copy(cs1);
                      cs += "_";
                      cs += cs2;
                      ret = residues2.Seek(cs);
/*
                      if (ret >= 0) {
                           fprintf(fp, "WARNING: The following residues have same chain ID '%s' and residue number '%s' : %s",
                                    cs1.Text(), cs2.Text(), cs4.Text());
                           while (ret >= 0) {
                                get_value(cs4, Table, idx4, ret);
                                fprintf(fp, " %s", cs4.Text());
                                ret = residues2.GoToNext();
                                if (ret >= 0) {
                                     if (residues2.GetCurrData() != cs) break;
                                }
                           }
                           fprintf(fp, ".\n\n");
                      }
*/
                      residues2.Add(cs);
                 }
            }

            Table1 = _CifObj->GetTablePtr(_StructureId, "ndb_nonpoly_scheme");
            int idx11, idx12, idx13, idx14;
            TblIndexObj residues3, residues4;
            residues3.Clear();
            residues4.Clear();
            if (!is_empty_table(Table1)) {
                 rowNo1 = Table1->GetNumRows();
                 idx11  = Table1->GetColumnIndex("pdb_id");
                 idx12  = Table1->GetColumnIndex("pdb_num");
                 idx13  = Table1->GetColumnIndex("pdb_ins_code");
                 idx14  = Table1->GetColumnIndex("mon_id");
                 for (int i = 0; i < rowNo1; i++) {
                      get_value(cs1, Table1, idx11, i);
                      get_value(cs2, Table1, idx12, i);
                      if (atoi(cs2.Text()) > 9999) _Error_Check[71]++;
                      get_value(cs3, Table1, idx13, i);
                      get_value(cs4, Table1, idx14, i);
                      if (cs3 != "" && !strstr(cs2.Text(), cs3.Text())) {
                           cs2 += cs3;
                      }
                      cs.Copy(cs1);
                      cs += "_";
                      cs += cs2;
                      cs += "_";
                      cs += cs4;
                      if (residues3.Seek(cs) >= 0) {
/*
                           fprintf(fp, "WARNING: residue '%s %s %s' repeat in ndb_nonpoly_scheme category.\n\n",
                                      cs1.Text(), cs2.Text(), cs4.Text());
*/
                      } else residues3.Add(cs);
                      cs.Copy(cs1);
                      cs += "_";
                      cs += cs2;
                      ret = residues2.Seek(cs);
/*
                      if (ret >= 0) {
                           fprintf(fp, "WARNING: The following residues have same chain ID '%s' and residue number '%s' : %s",
                                    cs1.Text(), cs2.Text(), cs4.Text());
                           while (ret >= 0) {
                                get_value(cs4, Table, idx4, ret);
                                fprintf(fp, " %s", cs4.Text());
                                ret = residues2.GoToNext();
                                if (ret >= 0) {
                                     if (residues2.GetCurrData() != cs) break;
                                }
                           }
                           fprintf(fp, ".\n\n");
                      }
                      ret = residues4.Seek(cs);
                      if (ret >= 0) {
                           fprintf(fp, "WARNING: The following residues have same chain ID '%s' and residue number '%s' : %s",
                                    cs1.Text(), cs2.Text(), cs4.Text());
                           while (ret >= 0) {
                                get_value(cs4, Table1, idx14, ret);
                                fprintf(fp, " %s", cs4.Text());
                                ret = residues4.GoToNext();
                                if (ret >= 0) {
                                     if (residues4.GetCurrData() != cs) break;
                                }
                           }
                           fprintf(fp, ".\n\n");
                      }
*/
                      residues4.Add(cs);
                 }
            }

            Table = _CifObj->GetTablePtr(_StructureId, "struct_ref_seq_dif");
            if (!is_empty_table(Table)) {
                 rowNo = Table->GetNumRows();
                 idx1  = Table->GetColumnIndex("ndb_pdb_chain_id");
                 idx2  = Table->GetColumnIndex("ndb_auth_seq_num");
                 idx3  = Table->GetColumnIndex("ndb_pdb_ins_code");
                 idx4  = Table->GetColumnIndex("mon_id");
                 for (i = 0; i < rowNo; i++) {
                      get_value(cs1, Table, idx1, i);
                      get_value(cs2, Table, idx2, i);
                      get_value(cs3, Table, idx3, i);
                      get_value(cs4, Table, idx4, i);
                      if (cs2 == "" && cs3 == "" && cs4 == "") continue;
                      cs.Copy(cs1);
                      cs += "_";
                      cs += cs2;
                      cs += cs3;
                      cs += "_";
                      cs += cs4;
                      if (residues1.Seek(cs) < 0) {
                           _Error_Check[5]++;
                           fprintf(fp, "ERROR: residue '%s %s%s %s' in struct_ref_seq_dif can not be found.\n\n",
                                       cs1.Text(), cs2.Text(), cs3.Text(), cs4.Text());
                      }
                 }
            }

            Table = _CifObj->GetTablePtr(_StructureId, "struct_site_gen");
            if (!is_empty_table(Table)) {
                 rowNo = Table->GetNumRows();
                 idx1  = Table->GetColumnIndex("auth_asym_id");
                 idx2  = Table->GetColumnIndex("auth_seq_id");
                 idx3  = Table->GetColumnIndex("ndb_label_ins_code");
                 idx4  = Table->GetColumnIndex("auth_comp_id");
                 for (i = 0; i < rowNo; i++) {
                      get_value(cs1, Table, idx1, i);
                      get_value(cs2, Table, idx2, i);
                      get_value(cs3, Table, idx3, i);
                      get_value(cs4, Table, idx4, i);
                      cs.Copy(cs1);
                      cs += "_";
                      cs += cs2;
                      cs += cs3;
                      cs += "_";
                      cs += cs4;
                      if (residues1.Seek(cs) < 0 && residues3.Seek(cs) < 0) {
                           _Error_Check[5]++;
                           fprintf(fp, "ERROR: residue '%s %s%s %s' in struct_site_gen can not be found.\n\n",
                                       cs1.Text(), cs2.Text(), cs3.Text(), cs4.Text());
                      }
                 }
            }

            Table = _CifObj->GetTablePtr(_StructureId, "rcsb_missing_residue_list");
            if (!is_empty_table(Table)) {
                 rowNo = Table->GetNumRows();
                 idx1  = Table->GetColumnIndex("pdb_chain_id");
                 idx2  = Table->GetColumnIndex("pdb_residue_number");
                 idx3  = Table->GetColumnIndex("pdb_insertion_code");
                 idx4  = Table->GetColumnIndex("pdb_residue_name");
                 for (i = 0; i < rowNo; i++) {
                      get_value(cs1, Table, idx1, i);
                      get_value(cs2, Table, idx2, i);
                      get_value(cs3, Table, idx3, i);
                      get_value(cs4, Table, idx4, i);
                      cs.Copy(cs1);
                      cs += "_";
                      cs += cs2;
                      cs += cs3;
                      cs += "_";
                      cs += cs4;
                      if (residues1.Seek(cs) < 0 && residues3.Seek(cs) < 0) {
                           _Error_Check[5]++;
                           fprintf(fp, "ERROR: residue '%s %s%s %s' in rcsb_missing_residue_list can not be found.\n\n",
                                       cs1.Text(), cs2.Text(), cs3.Text(), cs4.Text());
                      }
                 }
            }

            int numCategory = 0;
            char **categoryName = _CifObj->GetTableNames(numCategory);
            for (i = 0; i < numCategory; i++) {
                 Table = _CifObj->GetTablePtr(_StructureId, categoryName[i]);
                 if (is_empty_table(Table) || !strcmp(categoryName[i], "rcsb_version")) {
                      delete [] categoryName[i];
                      continue;
                 }
                 id_idx = Table->GetColumnIndex("entry_id");
                 if (id_idx < 0) {
                      delete [] categoryName[i];
                      continue;
                 }
                 rowNo = Table->GetNumRows();
                 for (j = 0; j < rowNo; j++) {
                      get_value(cs, Table, id_idx, j);
                      if (cs == "") continue;
                      if (cs != _StructureId) {
                           fprintf(fp, "ERROR: entry id '%s' in %d row of '%s' category is not same as '%s'.\n\n",
                                       cs.Text(), j + 1, categoryName[i], _StructureId);
                      }
                 }
            }
            if (categoryName) delete [] categoryName;
         }
       }

       if (dict) delete dict;

       string.Clear();
       string1.Clear();
       string2.Clear();
       string3.Clear();
       buff.Clear();
       res_high.Clear();
       _UniIndex.Clear();
       int atom_count = 0;
       int zero_occupancy_count = 0;
       int high_b_factor_count = 0;
       int low_b_factor_count = 0;
       int negative_b_factor_count = 0;
       int total_count = 0;
       int same_count = 0;
       int maximum_protein_sequence_length = 0;
       int maxinum_nucleic_acid_sequence_length = 0;
       for (i = 0; i < _num_mol; i++) {
            if (_num_mol > 1) Mol_ID = _molecules[i]->Mol_ID();
            chain = _molecules[i]->GetFirstChain();
            while (chain) {
                 if (chain->chain_type() == ATOMP_TOKEN) {
/*
                      if (!strcmp(chain->SeqRes(0)->Field[0], "ACE")) {
                           _Error_Check[67]++;
                           if (!strstr(_Error_Check_Message[67].Text(), "ACE")) {
                                if (_Error_Check_Message[67] != "") {
                                     _Error_Check_Message[67] += ",";
                                }
                                _Error_Check_Message[67] += "ACE";
                           }
                      }
                      if (!strcmp(chain->SeqRes(chain->SeqLen()-1)->Field[0], "NH2")) {
                           _Error_Check[67]++;
                           if (!strstr(_Error_Check_Message[67].Text(), "NH2")) {
                                if (_Error_Check_Message[67] != "") {
                                     _Error_Check_Message[67] += ",";
                                }
                                _Error_Check_Message[67] += "NH2";
                           }
                      }
*/
                      if (chain->SeqLen() > maximum_protein_sequence_length) {
                           maximum_protein_sequence_length = chain->SeqLen();
                      }
                 } else
                 if (chain->chain_type() == ATOMN_TOKEN) {
                      if (chain->SeqLen() > maxinum_nucleic_acid_sequence_length) {
                           maxinum_nucleic_acid_sequence_length = chain->SeqLen();
                      }
                 }
                 chain->check_inscode(string3);
                 residue = chain->GetFirstResidue();
                 while (residue) {
                      total_count++;
                      residue->find_atom_type_mismatch(Mol_ID, string);
                      if (is_split_residue(residue->ResName())) {
                           fprintf(fp, "WARNING: residue %s need to be split.\n", residue->ResName());
                           _Error_Check[66]++;
                      }
                      if (is_obsolete_residue(residue->ResName())) {
                           fprintf(fp, "WARNING: residue %s need to be obsoleted.\n", residue->ResName());
                           _Error_Check[66]++;
                      }
                      if (is_REF_ONLY_residue(residue->ResName())) {
                           _Error_Check[77]++;
                           fprintf(fp, "WARNING: The component %s is marked to be ",
                                       residue->ResName());
                           fprintf(fp, "REF_ONLY. Not to be used in any entry. The ");
                           fprintf(fp, "presentation of this entity is polymeric.\n");
                      }
                      residue->find_wrong_alt_id(Mol_ID, string2);
                      if (!(_experiment_type & EXPERIMENT_TYPE_NMR) &&
                          !(_experiment_type & EXPERIMENT_TYPE_NMR_SOLID)) {
                           if (residue->check_occupancy_and_B_factor(Mol_ID, string1,
                                 atom_count, zero_occupancy_count, high_b_factor_count,
                                 low_b_factor_count, negative_b_factor_count, res_low)) {
                                same_count++;
                                if (_UniIndex.Seek(res_low) < 0) _UniIndex.Add(res_low);
                           }
                      }
                      residue->check_same_coordinates(Mol_ID, buff);
                      residue->check_insertion_code(Mol_ID, res_high);
                      residue->check_connectivity(Mol_ID, _Error_Check_Message[68]);
                      residue = chain->GetNextResidue();
                 }
                 chain = _molecules[i]->GetNextChain();
            }
       }

       if (maximum_protein_sequence_length < 24 &&
            maxinum_nucleic_acid_sequence_length < 3) {
            bool is_bmrb_entry = false;
            if (_CifObj) {
                 Table = _CifObj->GetTablePtr(_StructureId, "database_2");
                 if (!is_empty_table(Table)) {
                      bool has_bmrb_id = false;
                      bool has_other_id = false;
                      rowNo = Table->GetNumRows();
                      CifString cs1, cs2;
                      high_index = Table->GetColumnIndex("database_id");
                      low_index  = Table->GetColumnIndex("database_code");
                      for (i = 0; i < rowNo; i++) {
                           get_value(cs1, Table, high_index, i);
                           get_value(cs2, Table, low_index, i);
                           if (cs1 != "" && cs1 != "BMRB" && cs2 != "")
                                has_other_id = true;
                           if (cs1 == "BMRB" && cs2 != "")
                                has_bmrb_id = true;
                      }
                      if (has_bmrb_id && !has_other_id) is_bmrb_entry = true;
                 }
            }
            if (!is_bmrb_entry) {
                 fprintf(fp, "\nERROR: The sequence length is less than 24 for protein (or 3 for nucleic acid)\n");
                 fprintf(fp, "       which is not acceptable at PDB!\n\n");
                 _Error_Check[65]++;
            }
       }

       if (string != "") {
            fprintf(fp, "\n==> The following atoms have wrong atom type\n\n");
            fprintf(fp, "     RES MOD#C SEQ          ATOMS    TYPE    TYPE FROM COMPONENT.CIF\n\n");
            fprintf(fp, "%s\n\n", string.Text());
            _Error_Check[25]++;
       }
       if (string1 != "") {
            fprintf(fp, "%s\n\n", string1.Text());
            _Error_Check[26]++;
       }
       if (buff != "") {
            fprintf(fp, "%s\n\n", buff.Text());
            _Error_Check[27]++;
       }
       if (string2 != "") {
            fprintf(fp, "%s\n\n", string2.Text());
            _Error_Check[31]++;
       }
       if (string3 != "") {
            fprintf(fp, "%s\n\n", string3.Text());
            _Error_Check[62]++;
       }
       if (res_high != "") {
            fprintf(fp, "%s\n\n", res_high.Text());
            _Error_Check[33]++;
       }
       if (_Error_Check_Message[67] != "") {
            fprintf(fp, "ERROR: terminal residue(s):\n\t%s\n\n",
                      _Error_Check_Message[67].Text());
       }
       if (_Error_Check_Message[68] != "") {
            _Error_Check[68]++;
            fprintf(fp, "ERROR: dissociated residue(s):\n\t%s\n\n",
                      _Error_Check_Message[68].Text());
       }
       if (_Error_Check[71]) fprintf(fp, "\nPDB residue numbering is larger than 9999.\n");

/*
       if (!(_experiment_type & EXPERIMENT_TYPE_NMR) &&
           !(_experiment_type & EXPERIMENT_TYPE_NMR_SOLID)) {
*/
       if (_experiment_type & EXPERIMENT_TYPE_XRAY ||
           _experiment_type & EXPERIMENT_TYPE_NEUTRON ||
           _experiment_type & EXPERIMENT_TYPE_FIBER ||
           _experiment_type & EXPERIMENT_TYPE_ELECTRON) {
            if (atom_count > 0) {
                 if ((float) zero_occupancy_count / (float) atom_count > 0.5) {
                      fprintf(fp, "WARNING: There are too many atoms with 0 occupancy.\n\n");
                 }
                 // if ((float) high_b_factor_count / (float) atom_count > 0.5) {
                 if (high_b_factor_count == atom_count) {
                      fprintf(fp, "WARNING: B-factor values are high.\n\n");
                 }
                 if (low_b_factor_count == atom_count) {
                      fprintf(fp, "WARNING: B-factor values are low.\n\n");
                 }
                 if (negative_b_factor_count) {
                      fprintf(fp, "WARNING: negative B value presents in the coordinates.\n\n");
                 }
            }
            if (_UniIndex.Root() >= 0 && _UniIndex.NumElem() == 1 && total_count == same_count) {
                 fprintf(fp, "WARNING: All B-factors are same.\n\n");
            }
       }

       if ((_experiment_type & EXPERIMENT_TYPE_NMR ||
            _experiment_type & EXPERIMENT_TYPE_NMR_SOLID) && _num_mol > 1) {
            length = new int[_num_mol];
            for (i = 0; i < _num_mol; i++) {
                 length[i] = 0;
                 chain = _molecules[i]->GetFirstChain();
                 while (chain) {
                      if (chain->chain_type() == ATOMP_TOKEN ||
                          chain->chain_type() == ATOMN_TOKEN) {
                           residue = chain->GetFirstResidue();
                           while (residue) {
                                length[i]++;
                                residue = chain->GetNextResidue();
                           }
                      }
                      chain = _molecules[i]->GetNextChain();
                 }
            }
            backbone = new double*[_num_mol];
            for (i = 0; i < _num_mol; i++) {
                 backbone[i] = NULL;
                 if (length[i] > 0) {
                      backbone[i] = new double[3 * length[i]];
                      length[i] = 0;
                      chain = _molecules[i]->GetFirstChain();
                      while (chain) {
                           if (chain->chain_type() == ATOMP_TOKEN ||
                               chain->chain_type() == ATOMN_TOKEN) {
                                residue = chain->GetFirstResidue();
                                while (residue) {
                                     atom1 = NULL;
                                     if (chain->chain_type() == ATOMP_TOKEN)
                                          atom1 = residue->find_atom("CA");
                                     else atom1 = residue->find_atom("P");
                                     if (atom1) {
                                          backbone[i][3 * length[i]] = atom1->orig().x;
                                          backbone[i][3*length[i]+1] = atom1->orig().y;
                                          backbone[i][3*length[i]+2] = atom1->orig().z;
                                          length[i]++;
                                     }
                                     residue = chain->GetNextResidue();
                                }
                           }
                           chain = _molecules[i]->GetNextChain();
                      }
                 }
            }

            fprintf(fp, "\nRMSD checking between models:\n");
            is_error = 0;
            if (length[0] > 0) {
                 for (i = 1; i < _num_mol; i++) {
                      if (!length[i]) continue;
                      min_length = length[0];
                      if (length[i] < min_length) min_length = length[i];
                      if (min_length < 3) continue;
                      superposition_matrix(min_length, backbone[0], backbone[i],  rt, ave_exyz, ave_sxyz);
                      v = 0.0;
                      v1 = 0.0;
                      int orig_number = 0, supr_number = 0;
                      for (k = 0; k < min_length; k++) {
                           float v2 = 0;
                           float v3 = 0;
                           for (l = 0; l < 3; l++) {
                                v0 = 0.0;
                                for (m = 0; m < 3; m++) {
                                     v0 += rt[l * 3 + m] * (backbone[i][3 * k + m] - rt[9 + m]);
                                }
                                v += (backbone[0][3 * k + l] - backbone[i][3 * k + l])
                                   * (backbone[0][3 * k + l] - backbone[i][3 * k + l]);
                                v2 += (backbone[0][3 * k + l] - backbone[i][3 * k + l])
                                    * (backbone[0][3 * k + l] - backbone[i][3 * k + l]);
                                v1 += (backbone[0][3 * k + l] - v0) * (backbone[0][3 * k + l] - v0);
                                v3 += (backbone[0][3 * k + l] - v0) * (backbone[0][3 * k + l] - v0);
                           }
                           rms1 = sqrt(v2);
                           if (rms1 < 3.6) orig_number++;
                           rms1 = sqrt(v3);
                           if (rms1 < 3.6) supr_number++;
                      }
                      rms = sqrt(v / (double) min_length);
                      rms1 = sqrt(v1 / (double) min_length);
                      if (rms < 0.01) {
                           fprintf(fp, "\tRMSD of CA and/or P atom(s) between Model %3d and Model %3d is %7.3f\n", 
                               _molecules[0]->Mol_ID(), _molecules[i]->Mol_ID(), rms);
                           is_error = 1;
                      } else {
                           v = rms1 / rms;
                           int is_superimposed = 0;
                           if (v > 0.95 || v > 0.8 && rms < 4.0) is_superimposed = 1; 
                           if (orig_number >= supr_number && orig_number > min_length * 2 / 5) is_superimposed = 1;
                           if (v > 0.7 && orig_number > min_length * 7 / 10) is_superimposed = 1;
                           if (orig_number > min_length / 2 && rms < 5.0) is_superimposed = 1;
                           if (!is_superimposed) {
/*
fprintf(stdout, "%d - %d : %d %d %d %.3f (%.3f/%.3f)\n", _molecules[0]->Mol_ID(), _molecules[i]->Mol_ID(),
orig_number, supr_number, min_length, v, rms1,rms);
fprintf(stdout, "\n");
*/
                                fprintf(fp, "\tModel %3d and Model %3d are not superimposed.\n",
                                    _molecules[0]->Mol_ID(), _molecules[i]->Mol_ID());
                                is_error = 1;
                           }
                      }
                 }
            }
            for (i = 1; i < _num_mol - 1; i++) {
                 if (!length[i]) continue;
                 for (j = i + 1; j < _num_mol; j++) {
                      if (!length[j]) continue;
                      min_length = length[i];
                      if (length[j] < min_length) min_length = length[j];
                      // superposition_matrix(min_length, backbone[i], backbone[j], rt, ave_exyz, ave_sxyz); 
                      v = 0.0;
                      for (k = 0; k < min_length; k++) {
                           for (l = 0; l < 3; l++) {
/*
                                v0 = 0.0;
                                for (m = 0; m < 3; m++) {
                                     v0 += rt[l * 3 + m] * (backbone[j][3 * k + m] - rt[9 + m]);
                                }
*/
                                v += (backbone[i][3 * k + l] - backbone[j][3 * k + l])
                                   * (backbone[i][3 * k + l] - backbone[j][3 * k + l]);
                                // v1 += (backbone[i][3 * k + l] - v0) * (backbone[i][3 * k + l] - v0);
                           }
                      }
                      rms = sqrt(v / (double) min_length);
                      if (rms < 0.01) {
                           fprintf(fp, "\tRMSD of CA and/or P atom(s) between Model %3d and Model %3d is %7.3f\n", 
                                    _molecules[i]->Mol_ID(), _molecules[j]->Mol_ID(), rms);
                           is_error = 1;
                      }
                 }
            }
            if (!is_error) fprintf(fp, "\tRMSD checking OK.\n");

            for (i = 0; i < _num_mol; i++) {
                 if (backbone[i]) delete [] backbone[i];
            }
            delete [] backbone;
            delete [] length;
       }
}

static void check_tokens(_MOLECULE *mol, Cards_Info_Format *CardInfo,
                const int Num_Token, CHECK_TOKEN_FORMAT *Check_Format,
                const int skip, const char *file_name, FILE *fp)
{
       int i , j, tokenid, num;
       char AtomName[10], ChainId[10], ResidueName[10], ResidueNo[10],
              InsCode[10], buffer[200];
       _ATOM **atom = NULL;
       _RESIDUE *residue = NULL;
       One_Card_Info_Format *pCard = NULL;

       for (i = 0; i < Num_Token; i++) {
            if (skip && (!strcmp(Check_Format[i].TokenName, "HET") ||
                         !strcmp(Check_Format[i].TokenName, "CISPEP")))
                 continue;
            tokenid = Check_Format[i].Token_ID;
            if (tokenid == -1 || tokenid == DBREF_TOKEN - 1) continue;

            pCard = CardInfo[tokenid].Head();
            while (pCard != NULL) {
                 for (j = 0; j < Check_Format[i].NumField; j++) {
                      AtomName[0] = ChainId[0] = ResidueName[0] =
                            ResidueNo[0] = InsCode[0] = '\0';
                      if (Check_Format[i].Fields[j].AtmName) strcpy(AtomName,
                            pCard->Field(Check_Format[i].Fields[j].AtmName));
                      if (Check_Format[i].Fields[j].ChainId) strcpy(ChainId,
                            pCard->Field(Check_Format[i].Fields[j].ChainId));
                      if (Check_Format[i].Fields[j].ResName) strcpy(ResidueName,
                            pCard->Field(Check_Format[i].Fields[j].ResName));
                      if (Check_Format[i].Fields[j].ResNum) strcpy(ResidueNo,
                            pCard->Field(Check_Format[i].Fields[j].ResNum));
                      if (Check_Format[i].Fields[j].InsCode) strcpy(InsCode,
                            pCard->Field(Check_Format[i].Fields[j].InsCode));

                      if (!strcmp(AtomName, "") && !strcmp(ResidueName, "") &&
                          !strcmp(ChainId, "") && !strcmp(ResidueNo, "") &&
                          !strcmp(InsCode, "")) continue;

                      buffer[0] = '\0';
                      if (strcmp(AtomName, "")) {
                           atom = mol->find_pdb_atom(AtomName, ChainId,
                                        ResidueName, atoi(ResidueNo), InsCode, num);
                           if (!atom) {
                                if (!strcmp(ChainId, "")) strcpy(ChainId, " ");
                                if (strcmp(InsCode, ""))
                                     sprintf(buffer, "ATOM: %s RES: %s CHNID: %s SSEQ: %s%s",
                                         AtomName, ResidueName, ChainId, ResidueNo, InsCode);
                                else sprintf(buffer, "ATOM: %s RES: %s CHNID: %s SSEQ: %s",
                                         AtomName, ResidueName, ChainId, ResidueNo);
                           } else delete [] atom;
                      } else {
                           residue = mol->find_pdb_residue(ChainId,
                                        ResidueName, atoi(ResidueNo), InsCode);
                           if (!residue) {
                                if (!strcmp(ChainId, "")) strcpy(ChainId, " ");
                                if (strcmp(InsCode, "") &&
                                    strcmp(ResidueName, ""))
                                     sprintf(buffer, "RES: %s CHNID: %s SSEQ: %s%s",
                                         ResidueName, ChainId, ResidueNo, InsCode);
                                else if (strcmp(InsCode, ""))
                                     sprintf(buffer, "CHNID: %s SSEQ: %s%s", ChainId,
                                         ResidueNo, InsCode);
                                else if (strcmp(ResidueName, ""))
                                     sprintf(buffer, "RES: %s CHNID: %s SSEQ: %s",
                                         ResidueName, ChainId, ResidueNo);
                                else sprintf(buffer, "CHNID: %s SSEQ: %s", ChainId,
                                         ResidueNo);
                           }
                      }
                      if (strcmp(buffer, "")) {
                           fprintf(fp, "%s: Error: Record (%s) in Token '%s'",
                             file_name, buffer, NdbToken[tokenid].TokenName);
                           fprintf(fp, " can not be found in coordinates\n");
                      }
                 }
                 pCard = pCard->Next();
            }
       }
}

static void check_categorys(_MOLECULE *mol, ISTable *Table, CHECK_CATEGORY *category,
                 FILE *fp, const char *file_name, int &count)
{
       int index[4][5], i, j, k, rowNo, is_useful, num;
       char value[5][15], buffer[200];
       CifString cifstring, record;
       _ATOM **atom = NULL;
       _RESIDUE *residue = NULL;

       for (i = 0; i < 4; i++) {
            for (j = 0; j < 5; j++) index[i][j] = -1;
       }
       category->type_index = Table->GetColumnIndex(category->type);
       for (i = 0; i < category->NumRecord; i++) {
            if (strcmp(category->items[i].atom_label, ""))
                 index[i][0] = Table->GetColumnIndex(category->items[i].atom_label);
            if (strcmp(category->items[i].asym_label, ""))
                 index[i][1] = Table->GetColumnIndex(category->items[i].asym_label);
            if (strcmp(category->items[i].comp_label, ""))
                 index[i][2] = Table->GetColumnIndex(category->items[i].comp_label);
            if (strcmp(category->items[i].seq_label, ""))
                 index[i][3] = Table->GetColumnIndex(category->items[i].seq_label);
            if (strcmp(category->items[i].ins_label, ""))
                 index[i][4] = Table->GetColumnIndex(category->items[i].ins_label);
       }

       rowNo = Table->GetNumRows();
       record.Copy(category->type);
       for (i = 0; i < rowNo; i++) {
            if (category->type_index >= 0) {
                 is_useful = 0;
                 get_value(record, Table, category->type_index, i);
                 if (!strcmp(category->category, "struct_conf")) {
                      if (record == "HELIX")
                           is_useful = 1;
                      else if (record == "TURN")
                           is_useful = 1;
                 } else if (!strcmp(category->category, "struct_conn")) {
                      if (record == "hydbnd") {
                           record.Copy("HYDBND");
                           is_useful = 1;
                      } else if (record == "disulf") {
                           record.Copy("SSBOND");
                           is_useful = 1;
                      } else if (record == "link") {
                           record.Copy("LINK");
                           is_useful = 1;
                      } else if (record == "sltbrg") {
                           record.Copy("SLTBRG");
                           is_useful = 1;
                      } else if (record == "modres") {
                           record.Copy("MODRES");
                           is_useful = 1;
                      }
                 }
                 if (!is_useful) continue;
            }
            for (j = 0; j < category->NumRecord; j++) {
                 buffer[0] = '\0';
                 for (k = 0; k < 5; k++) {
                      memset(value[k], '\0', 15);
                      if (index[j][k] >= 0) {
                           get_value(cifstring, Table, index[j][k], i);
                           strcpy(value[k], cifstring.Text());
                      }
                 }
                 if (strcmp(value[0], "") || strcmp(value[1], "") || strcmp(value[2], "") ||
                     strcmp(value[3], "") || strcmp(value[4], "")) {
                      if (strcmp(value[0], "")) {
                           atom = mol->find_pdb_atom(value[0], value[1],
                                     value[2], atoi(value[3]), value[4], num);
                           if (!atom) {
                                if (!strcmp(value[1], "")) strcpy(value[1], " ");
                                if (!strcmp(value[4], "")) strcpy(value[4], " ");
                                sprintf(buffer, "ATOM: %s RES: %s CHNID: %s SSEQ: %s%s",
                                          value[0], value[2], value[1], value[3], value[4]);
                           } else delete [] atom;
                      } else {
                           residue = mol->find_pdb_residue(value[1],
                                     value[2], atoi(value[3]), value[4]);
                           if (!residue) {
                                if (!strcmp(value[1], "")) strcpy(value[1], " ");
                                if (!strcmp(value[4], "")) strcpy(value[4], " ");
                                sprintf(buffer, "RES: %s CHNID: %s SSEQ: %s%s",
                                          value[2], value[1], value[3], value[4]);
                           }
                      }
                 }
                 if (strcmp(buffer, "")) {
                      fprintf(fp, "%s: Error: Record (%s) in Token '%s'", file_name,
                                 buffer, record.Text());
                      fprintf(fp, " can not be found in coordinates\n");
                      count++;
                 }
            }
       }
}
/*
static void check_categorys1(_MOLECULE *mol, ISTable *Table, CHECK_CATEGORY *category,
                 FILE *fp, const char *file_name, int &count)
{
       int index[4][2], i, j, rowNo, Chain_Index, Res_Index;
       CifString cifstring, cifstring1;

       for (i = 0; i < 4; i++) {
            for (j = 0; j < 2; j++) index[i][j] = -1;
       }

       for (i = 0; i < category->NumRecord; i++) {
            if (strcmp(category->items[i].asym_label, ""))
                 index[i][0] = Table->GetColumnIndex(category->items[i].asym_label);
            if (strcmp(category->items[i].seq_label, ""))
                 index[i][1] = Table->GetColumnIndex(category->items[i].seq_label);
       }

       rowNo = Table->GetNumRows();
       for (i = 0; i < rowNo; i++) {
            for (j = 0; j < category->NumRecord; j++) {
                 get_value(cifstring, Table, index[j][0], i);
                 get_value(cifstring1,Table, index[j][1], i);
                 mol->get_ndb_residue(cifstring.Text(), cifstring1.Text(), Chain_Index, Res_Index);
                 if (Chain_Index < 0 && Res_Index < 0) {
                      fprintf(fp, "%s: Error: Residue ranges (%s %s) in category '%s'", file_name,
                                 cifstring.Text(), cifstring1.Text(), category->category);
                      fprintf(fp, " can not be found in coordinates\n");
                      count++;
                 }
            }
       }
}
*/
static void check_seqnum_redundancy(_MOLECULE *mol, const char *infile_name, FILE *fp, int &count)
{
       int j, first, end, ret;
       BTreeObj<_SEQNUM> SeqNum;
       _CHAIN *chain = NULL;
       _SEQNUM seqnum, tmp;
       char buffer[100];

       SeqNum.Clear();
       chain = mol->GetFirstChain();
       while (chain) {
            for (j = 0; j < chain->SeqLen(); j++) {
                 if (chain->SeqRes(j)->Deleted) continue;
                 if (!chain->SeqRes(j)->Residue) continue;

                 sprintf(buffer, "%s%s", chain->SeqRes(j)->Field[4],
                             chain->SeqRes(j)->InsCode);
                 seqnum.set_chnid(chain->PDB_ChainID());
                 seqnum.set_seqnum(buffer);
                 seqnum.set_name(chain->SeqRes(j)->Field[3]);
                 SeqNum.Add(seqnum);
            }
            chain = mol->GetNextChain();
       }

       if (SeqNum.Root() >= 0) {
            SeqNum.GoFirst();
            tmp = SeqNum.GetCurrData();
            first = 1;
            end = 0;
            ret = SeqNum.GoToNext();
            while (ret >= 0) {
                 seqnum = SeqNum.GetCurrData();
                 if (tmp == seqnum) {
                      if (first) {
                           end = 1;
                           fprintf(fp, "%s: Error: Residues ( ", infile_name);
                           fprintf(fp, "%s %s ", tmp.name(), seqnum.name());
                           first = 0;
                      } else fprintf(fp, "%s ", seqnum.name());
                 } else {
                      if (end) {
                           fprintf(fp, ") have same chain_ID '%s' and residue number '%s'\n",
                           tmp.chnid(), tmp.seqnum());
                           count++;
                      }
                      tmp = seqnum;
                      first = 1;
                      end = 0;
                 }
                 ret = SeqNum.GoToNext();
            }
            if (end) {
                 fprintf(fp, ") have same chain_ID '%s' and residue number '%s'\n",
                           tmp.chnid(), tmp.seqnum());
                 count++;
            }
       }
}

static int get_wavelength_list(const char *list)
{
       int i, count = 0, len;
       char buffer[1000], *p = NULL, buff[100];

       if (list == NULL || !strcmp(list, "")) return 0;

       strcpy(buffer, list);
       len = strlen(buffer);
       for (i = 0; i < len; i++) {
            if (buffer[i] == ',' || buffer[i] == ';') buffer[i] = ' ';
       }
       p = buffer;
       while (p) {
            p = get_word(p, buff);
            if (strcmp(buff, "")) count++;
       }
       return count;
}

static int check_taxonomy(const char *name)
{
       char *buffer = NULL;
       CifString cifstring;

       if (TaxnomyList.NumElem() > 0) {
            buffer = new char[strlen(name)+1];
            strcpy(buffer, name);
            ndb_set_string_to_upper(buffer);
            cifstring.Copy(buffer);
            delete [] buffer;
            if (TaxnomyList.Seek(cifstring) >= 0)
                 return 0;
            else return 1;
       }
       return 0;
}

static char *get_word1(char *string, char *buffer)
{
       *buffer = '\0';
       while ((*string == ',' || *string == ';' || *string == ' ') && *string != '\0') string++;
       if (*string == '\0') return NULL;
       while (*string != ',' && *string != ';' && *string != '\0') {
            *buffer = *string;
            buffer++; string++;
       }
       *buffer = '\0';
       if(*string == '\0') return NULL;
       return string;
}

static bool check_header_failed(const CifString& header, DICFileObj *dict)
{
       ReVarCifArray<CifString> enumValues;
       get_EnumerationValue("_struct_keywords.ndb_keywords", dict, enumValues);
       if (enumValues.Length() == 0) return false;

       TblIndexObj Index;
       Index.Clear();
       for (unsigned int l = 0; l < enumValues.Length(); ++l) Index.Add(enumValues[l]);

       CifString cs;
       cs.Copy(header);
       cs.ToUpper();
       enumValues.Clear();
       enumValues.Add(cs);

       ReVarCifArray<CifString> delimits;
       delimits.Clear();
       delimits.Add(",");
       delimits.Add("/");
       delimits.Add("INHIBITOR");
       delimits.Add("ACTIVATOR");
       delimits.Add("RECEPTOR");
       delimits.Add("REGULATOR");
       // delimits.Add("COMPLEXES");
       for (unsigned int i = 0; i < delimits.Length(); ++i) {
            split_head(enumValues, delimits[i].Text());
       }
       for (unsigned int i = 0; i < enumValues.Length(); ++i) {
            if (Index.Seek(enumValues[i]) < 0) return true;
       }

       return false;
}

static void split_head(ReVarCifArray<CifString>& enumValues, const char *delimit)
{
       char buffer[1000];
       memset(buffer, 0, 1000);
       ReVarCifArray<CifString> wordlist;
       wordlist.Clear();

       CifString cs;
       for (unsigned int i = 0; i < enumValues.Length(); ++i) {
            char *p = enumValues[i].Text();
            while (p) {
                 p = get_words_separate_by_delimits(p, buffer, delimit);
                 if (!strcmp(buffer, "")) continue;
                 cs.Copy(buffer);
                 wordlist.Add(cs);
            }
       }
       enumValues = wordlist;
}

static void check_source(FILE* fp, CifString& spell, ISTable* Table, const int& row,
                         const int& source_index, const int& tax_index)
{
       if (source_index < 0) return;

       CifString cs1, cs2;
       get_value(cs1, Table, source_index, row);
       if (cs1 == "") return;

       if (check_taxonomy(cs1.Text())) {
            fprintf(fp, "ERROR: Misspelled name: %s\n\n", cs1.Text());
            if (spell != "") spell += ", ";
            spell += cs1.Text();
       }

       cs2.Copy(cs1);
       cs2.ToUpper();
       if (strstr(cs2.Text(), "STRAIN")) {
            fprintf(fp, "WARNING: Source organism name '%s' contains 'strain'.\n\n", cs1.Text());
       }

       cs2.Clear();
       get_value(cs2, Table, tax_index, row);
       if (cs2 == "") fprintf(fp, "ERROR: taxonomy id is missing for %s.\n\n", cs1.Text());
}

static void check_missing_source(FILE* fp, int& error, CifFileObj* fobj, const char*
                                 StructureId, const CifString& id, const char* category,
                                 const char* source_item, const char* tax_item)
{
       CifString source, tax_id;
       source.Clear();
       tax_id.Clear();

       ISTable *t = fobj->GetTablePtr(StructureId, category);
       if (!is_empty_table(t)) {
            int rowNo = t->GetNumRows();
            int idx1 = t->GetColumnIndex("entity_id");
            int idx2 = t->GetColumnIndex(source_item);
            int idx3 = t->GetColumnIndex(tax_item);
            CifString cs;
            for (int i = 0; i < rowNo; ++i) {
                 get_value(cs, t, idx1, i);
                 if (cs == id) {
                      get_value(source, t, idx2, i);
                      get_value(tax_id, t, idx3, i);
                      break;
                 }
            }
       }

       if (source == "") {
            error++;
            fprintf(fp, "ERROR: Missing _%s.%s for entity '%s'.\n\n",
                        category, source_item, id.Text());
       }

       if (tax_id == "") {
            error++;
            fprintf(fp, "ERROR: Missing _%s.%s for entity '%s'.\n\n",
                        category, tax_item, id.Text());
       }
}

static void get_wordlist_from_string_separated_by_delimits(ReVarCifArray<CifString> &wordlist,
                  char *string_buff, const char *delimit)
{
       wordlist.Clear();
       char *buffer = new char[strlen(string_buff)+1];
       memset(buffer, 0, strlen(string_buff)+1);
       char *p = string_buff;
       CifString cs;
       while (p) {
            p = get_words_separate_by_delimits(p, buffer, delimit);
            if (!strcmp(buffer, "")) continue;
            cs.Copy(buffer);
            wordlist.Add(cs);
       }
       delete [] buffer;
}
