/*
FILE:     contact.C
*/
/*
VERSION:  8.125
*/
/*
DATE:     3/1/2013
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2013 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <stdio.h>
#include <stdlib.h>
#include <map>
#include <set>
#include <string>

#include "_cindex.h"
#include "_index.h"
#include "_validation.h"
#include "gridcell.h"
#include "BTreeObj.h"
#include "erf_util.h"
#include "util.h"
#include "xtal.h"

static void get_special_and_symmetry_contact(xtal& mycrys, const int _n_a_contact,
                 CONTACT**& _a_contact, int& _n_special, int& _alloc_special,
                 CONTACT**& _special, int& _n_s_contact, int& _alloc_s_contact,
                 CONTACT**& _s_contact, _MOLECULE** _molecules);
static CONTACT **get_all_waters(_MOLECULE *mol, int& all_num,
                         std::map<std::string, int>& _AllWaters);
static void find_outlier_waters_with_shells(xtal& mycrys, const double high_cut_off,
                      CONTACT**& all_waters, std::map<std::string, int>& _AllWaters);
static void find_outlier_waters_without_shells(xtal& mycrys, const double high_cut_off,
                      const bool include_symmetry, CONTACT**& all_waters,
                      std::map<std::string, int>& _AllWaters);
static void get_outlier_and_moving_waters(std::map<std::string, int> _AllWaters,
                 CONTACT**& all_waters, int& _n_waters, int& _alloc_waters,
                 CONTACT**& _waters, int& _n_moving_waters, int& _alloc_moving_waters,
                 CONTACT**& _moving_waters);
static void get_all_shell_atoms(std::map<std::string, int>& _AllWaters,
                                std::set<std::string>& _ShellWaters,
                                const int n_contact, CONTACT **contact);
static std::string get_index(const std::string& chnid, const int num);
static void get_connected_atoms(const char *resname, ReVarCifArray<CifString>& connected_atoms);
// static void print_contact(const int num, CONTACT **contact);

void _VALIDATION::find_all_links(const int water_ion_only)
{
       int n_contact = 0, i, j, k, l, FstIndex = -1, FstRes = -1, SndIndex = -1;
       int SndRes = -1, num_first, num_second, bond_type = 0, add_link = 0;
       char alt_loc_a[5], alt_loc_b[5], buffer[10], type[10];
       float d = 0;
       xtal mycrys;
       CONTACT **contact = NULL;
       _ATOM **atom_f = NULL, **atom_s = NULL;
       _RESIDUE *Fst = NULL, *Snd = NULL;
       _CHAIN *chn1 = NULL, *chn2 = NULL;
       ReVarCifArray<CifString> first_atoms, second_atoms;

       if (_num_mol == 0) return;

       int skip_link = 0;
       if (_CifObj) {
            ISTable *Table = _CifObj->GetTablePtr(_StructureId, "ndb_database_status");
            if (!is_empty_table(Table)) {
                 i = Table->GetColumnIndex("skip_PDB_REMARK");
                 if (i >= 0) {
                      CifString cifstring;
                      Table->GetCell(cifstring, i, 0);
                      cifstring.ToUpper();
                      if (strstr(cifstring.Text(), "LINK")) skip_link = 1;
                 }
            }
       }
       if (skip_link) return;

       if (!water_ion_only) {
            chn1 = _molecules[0]->GetFirstChain();
            while (chn1) {
                 if (chn1->chain_type() != ATOMN_TOKEN &&
                     chn1->chain_type() != ATOMP_TOKEN) {
                      chn1 = _molecules[0]->GetNextChain();
                      continue;
                 }

                 Fst = NULL;
                 j = -1;
                 for (i = 0; i < chn1->SeqLen(); i++) {
                      if (chn1->SeqRes(i)->Deleted) continue;
                      Snd = chn1->SeqRes(i)->Residue;
                      if (!Snd) continue;
                      if (Fst && Snd && (!is_a_standard_residue(Fst->ResName()) ||
                          !strcmp(Fst->ResName(), "ASP") ||
                          !is_a_standard_residue(Snd->ResName()))) {
                           get_connected_atoms(Fst->ResName(), first_atoms);
                           get_connected_atoms(Snd->ResName(), second_atoms);
                           for (unsigned int t1 = 0; t1 < first_atoms.Length(); t1++) {
                                if (!strcmp(Fst->ResName(), "ASP") && first_atoms[t1] == "C" &&
                                     is_a_standard_residue(Snd->ResName())) continue;
                                for (unsigned int t2 = 0; t2 < second_atoms.Length(); t2++) {
                                     atom_f = Fst->find_atom(first_atoms[t1].Text(), num_first);
                                     atom_s = Snd->find_atom(second_atoms[t2].Text(), num_second);
                                     if (atom_f && atom_s) {
                                          for (k = 0; k < num_first; k++)
                                          for (l = 0; l < num_second; l++) {
                                               if ((!strcmp(atom_f[k]->alt_loc(), "") ||
                                                    !strcmp(atom_s[l]->alt_loc(), "") ||
                                                    !strcmp(atom_f[k]->alt_loc(), atom_s[l]->alt_loc())) &&
                                                    is_a_bond(atom_f[k], atom_s[l]) &&
                                                   !strcmp(atom_f[k]->restype(), Fst->ResName()) &&
                                                   !strcmp(atom_s[l]->restype(), Snd->ResName())) {  
                                                    d = cal_distance(atom_f[k], atom_s[l]);
                                                    memset(buffer, 0, 10);
                                                    sprintf(buffer, "%.3f", d);
                                                    Insert_a_Link(0, atom_f[k]->atmtype(), atom_f[k]->alt_loc(),
                                                           Fst->ResName(), chn1->index(), chn1->SeqRes(j)->ResIndex,
                                                           atom_s[l]->atmtype(), atom_s[l]->alt_loc(), Snd->ResName(),
                                                           chn1->index(), chn1->SeqRes(i)->ResIndex, "", "", "",
                                                           "covale", buffer, "");
                                               }
                                          }
                                     }
                                     if (atom_f) delete [] atom_f;
                                     if (atom_s) delete [] atom_s;
                                }
                           }
                      }
                      Fst = Snd;
                      j = i;
                 }
                 chn1 = _molecules[0]->GetNextChain();
            }
       }

       if (_cell == NULL) return;

       mycrys.setorth('c');    
       if (mycrys.init(_cell, _molecules[0], 1) <= 0) return;
       mycrys.getSymmAtoms();      
       mycrys.moveAllAtomsToCenter();   
       mycrys.placeAtomsInGrid();

       contact = get_contact(mycrys, n_contact, 0.85, 4.0, 
                          "asym", "atnum 1 -1", "atnum 1 -1", 1, 0); 
       for (i = 0; i < n_contact; i++) {
            if (contact[i]->sym != 1 || contact[i]->transl[0] ||
                contact[i]->transl[1] || contact[i]->transl[2]) continue;

            bond_type = is_a_link(contact[i]->a_atom_type, contact[i]->b_atom_type,
                              contact[i]->dist);
            if (!bond_type) continue;
 
            if (_molecules[0]->find_ndb_residue(contact[i]->a_chnid,
                   contact[i]->a_resnam, contact[i]->a_resnum, "",
                   FstIndex, FstRes) &&
                _molecules[0]->find_ndb_residue(contact[i]->b_chnid,
                   contact[i]->b_resnam, contact[i]->b_resnum, "",
                   SndIndex, SndRes)) {
                 chn1 = _molecules[0]->GetIndexChain(FstIndex);
                 chn2 = _molecules[0]->GetIndexChain(SndIndex);
                 Fst = _molecules[0]->find_ndb_residue(contact[i]->a_chnid,
                           contact[i]->a_resnam, contact[i]->a_resnum, "");
                 Snd = _molecules[0]->find_ndb_residue(contact[i]->b_chnid,
                           contact[i]->b_resnam, contact[i]->b_resnum, "");
                 if (!chn1 || !chn2 || !Fst || !Snd) continue;

                 add_link = 0;
                 if (chn1->chain_type() == ATOMP_TOKEN &&
                     chn2->chain_type() == ATOMS_TOKEN) {
                      if (!strcmp(contact[i]->b_atom_type, "C") &&
                         (!strcmp(Fst->ResName(), "ASN") &&
                          // !strcmp(Snd->ResName(), "NAG") &&
                          is_sugar_residue(Snd->ResName()) &&
                          !strcmp(contact[i]->a_atomtyp, "ND2") ||
                          !strcmp(Fst->ResName(), "SER") &&
                          // !strcmp(Snd->ResName(), "NGA") &&
                          is_sugar_residue(Snd->ResName()) &&
                          !strcmp(contact[i]->a_atomtyp, "OG") ||
                          !strcmp(Fst->ResName(), "THR") &&
                          // !strcmp(Snd->ResName(), "NGA") &&
                          is_sugar_residue(Snd->ResName()) &&
                          !strcmp(contact[i]->a_atomtyp, "OG1")))
                           add_link = 1;
                 } else if (chn1->chain_type() == ATOMS_TOKEN &&
                            chn2->chain_type() == ATOMP_TOKEN) {
                      if (!strcmp(contact[i]->a_atom_type, "C") &&
                         (// !strcmp(Fst->ResName(), "NAG") &&
                          is_sugar_residue(Fst->ResName()) &&
                          !strcmp(Snd->ResName(), "ASN") &&
                          !strcmp(contact[i]->b_atomtyp, "ND2") ||
                          // !strcmp(Fst->ResName(), "NGA") &&
                          is_sugar_residue(Fst->ResName()) &&
                          !strcmp(Snd->ResName(), "SER") &&
                          !strcmp(contact[i]->b_atomtyp, "OG") ||
                          // !strcmp(Fst->ResName(), "NGA") &&
                          is_sugar_residue(Fst->ResName()) &&
                          !strcmp(Snd->ResName(), "THR") &&
                          !strcmp(contact[i]->b_atomtyp, "OG1")))
                           add_link = 1;
                 } else if (chn1->chain_type() == ATOMS_TOKEN &&
                            chn2->chain_type() == ATOMS_TOKEN) {
                      if (!strcmp(contact[i]->a_atom_type, "C") &&
                          !strcmp(contact[i]->b_atom_type, "O") ||
                          !strcmp(contact[i]->a_atom_type, "O") &&
                          !strcmp(contact[i]->b_atom_type, "C"))
                           add_link = 1;
                 } else if (chn1->chain_type() == ATOMP_TOKEN &&
                            chn2->chain_type() == ATOMP_TOKEN &&
                            !strcmp(Fst->ResName(), "CYS") &&
                            !strcmp(Snd->ResName(), "CYS") &&
                            !strcmp(contact[i]->a_atomtyp, "SG") &&
                            !strcmp(contact[i]->b_atomtyp, "SG")) {
                 } else if (bond_type == 1 &&
                           (!strcmp(Fst->ResName(), "HOH") ||
                            !strcmp(Snd->ResName(), "HOH"))) {
                 } else if (chn1->chain_type() == ATOMP_TOKEN &&
                            chn2->chain_type() == ATOMP_TOKEN &&
                            strcmp(chn1->PDB_ChainID(), chn2->PDB_ChainID())) {
                 } else if (bond_type == 2) add_link = 1;

                 if (add_link) {
                      if (chn1 != chn2) {
                           chn1->set_is_linked(1);
                           chn2->set_is_linked(1);
                      }
                      sprintf(alt_loc_a, "%c", contact[i]->a_alt_loc); 
                      sprintf(alt_loc_b, "%c", contact[i]->b_alt_loc); 

                      strcpy(type,  "covale");
                      if (bond_type == 2) strcpy(type, "metalc");
                      sprintf(buffer, "%.3f", contact[i]->dist);
                      if (!water_ion_only || (water_ion_only && bond_type == 2 &&
                         (!strcmp(Fst->ResName(), "HOH") ||
                          !strcmp(Snd->ResName(), "HOH")))) {
                           Insert_a_Link(0, contact[i]->a_atomtyp, alt_loc_a,
                                 Fst->ResName(), FstIndex, FstRes, contact[i]->b_atomtyp,
                                 alt_loc_b, Snd->ResName(), SndIndex, SndRes, "", "", "",
                                 type, buffer, "");
                      }
                 }
            }
       }
       if (contact) {
            for (i = 0; i < n_contact; i++) if (contact[i]) delete contact[i];
            delete [] contact;
            contact = NULL;
       }
}

CONTACT** _VALIDATION::get_cys_contact(int& n_contact)
{
       n_contact = 0;
       if (_cell->is_artifical == 2) return NULL;

       ReVarCifArray<_RESIDUE*> residues;
       residues.Clear();
       _CHAIN *chain = _molecules[0]->GetFirstChain();
       while (chain) {
            if (chain->chain_type() != ATOMP_TOKEN) {
                 chain = _molecules[0]->GetNextChain();
                 continue;
            }

            _RESIDUE *res = chain->GetFirstResidue();
            while (res) {
                 if (!strcmp(res->ResName(), "CY3") ||
                     !strcmp(res->ResName(), "CYS") ||
                     !strcmp(res->ResName(), "DCY") ||
                     !strcmp(res->ResName(), "LE1") ||
                     !strcmp(res->ResName(), "NCY"))
                      residues.Add(res);
                 res = chain->GetNextResidue();
            }
            chain = _molecules[0]->GetNextChain();
       }
       if (residues.Length() == 0) return NULL;

       xtal mycrys;
       mycrys.setorth('c');    
       mycrys.init(_cell, residues);
       mycrys.getSymmAtoms();      
       mycrys.moveAllAtomsToCenter();   
       mycrys.placeAtomsInGrid();

       return (get_contact(mycrys, n_contact, 1.80, 3.0, 
                    "all", "atnum 1 -1", "atnum 1 -1", 1, 0)); 
}

void _VALIDATION::calculate_contact_for_nmr()
{
       int i = 0, j = 0, n_contact = 0;
       xtal *mycrys = NULL;
       CONTACT **contact = NULL, **tmp = NULL;

       if (_cell->is_artifical == 2) return;
       // mycrys.setorth('c');

       for (i = 0; i < _num_mol; i++) {
            mycrys = new xtal;
            mycrys->setorth('c');
            mycrys->init(_cell, _molecules[i], 0);
            mycrys->getSymmAtoms();
            mycrys->moveAllAtomsToCenter();
            mycrys->placeAtomsInGrid();
            contact = get_contact(*mycrys, n_contact, 0.0, 2.2,
                        "asym", "atnum 1 -1", "atnum 1 -1", 1, 1);
            if (n_contact) {
                 if ((_n_a_contact + n_contact) >= _alloc_a_contact) {
                      _alloc_a_contact = _n_a_contact + n_contact;
                      tmp = _a_contact;
                      _a_contact = new CONTACT*[_alloc_a_contact];
                      for (j = 0; j < _n_a_contact; j++)
                           _a_contact[j] = tmp[j];
                      for (j = _n_a_contact; j < _alloc_a_contact; j++)
                           _a_contact[j] = NULL;
                      delete [] tmp; tmp = NULL;
                 }
                 for (j = 0; j < n_contact; j++) {
                      _a_contact[_n_a_contact] = contact[j];
                      _a_contact[_n_a_contact]->Mol_ID = _molecules[i]->Mol_ID();
                      _n_a_contact++;
                 }
            }
            if (contact) delete [] contact;
            contact = NULL;
            // mycrys.Clear();
            delete mycrys;
       }
}

CONTACT **_VALIDATION::calculate_all_contacts(int &n_all_contact, const double low, const double high)
{
       int i = 0, j = 0, n_contact = 0;
       xtal *mycrys = NULL;
       CONTACT **all_contact = NULL, **contact = NULL, **tmp = NULL;

       if (_cell->is_artifical == 2) return NULL;

       n_all_contact = 0;
       for (i = 0; i < _num_mol; i++) {
            mycrys = new xtal;
            mycrys->setorth('c');
            mycrys->init(_cell, _molecules[i], 0);
            mycrys->getSymmAtoms();
            mycrys->moveAllAtomsToCenter();
            mycrys->placeAtomsInGrid();
            contact = get_contact(*mycrys, n_contact, low, high,
                        "all", "atnum 1 -1", "atnum 1 -1", 1, 0);
            if (n_contact) {
                 tmp = all_contact;
                 all_contact = new CONTACT*[n_all_contact + n_contact];
                 for (j = 0; j < n_all_contact; j++) all_contact[j] = tmp[j];
                 for (j = 0; j < n_contact; j++) {
                      all_contact[j + n_all_contact] = contact[j];
                      all_contact[j + n_all_contact]->Mol_ID = _molecules[i]->Mol_ID();
                 }
                 n_all_contact += n_contact;
                 if (tmp) { delete [] tmp; tmp = NULL; }
            }
            if (contact) delete [] contact;
            contact = NULL;
            delete mycrys;
       }
       return all_contact;
}

void _VALIDATION::calculate_all_contacts_fp(FILE *fp, const int exchange_out)
{
       int i = 0, j = 0, k = 0, l = 0, n_contact = 0, first = 1, num_a = 0, num_b = 0;
       char buffer[100];
       xtal *mycrys = NULL;
       _ATOM **atom_a = NULL, **atom_b = NULL;
       CONTACT **contact = NULL;

       if (_cell->is_artifical == 2) return;

       for (l = 0; l < _num_mol; l++) {
            mycrys = new xtal;
            mycrys->setorth('c');
            mycrys->init(_cell, _molecules[l], 0);
            mycrys->getSymmAtoms();
            mycrys->moveAllAtomsToCenter();
            mycrys->placeAtomsInGrid();
            contact = get_contact(*mycrys, n_contact, 0.0, 3.5,
                        "all", "atnum 1 -1", "atnum 1 -1", 1, 0);
            if (n_contact) {
                 if (first) {
                      first = 0;
                      fprintf(fp, "#\nloop_\n");
                      if (exchange_out)
                           fprintf(fp, "_geom_contact.pdbx_PDB_model_num\n");
                      else fprintf(fp, "_geom_contact.ndb_model_num\n");
                      fprintf(fp, "_geom_contact.atom_site_id_1\n");
                      fprintf(fp, "_geom_contact.atom_site_label_atom_id_1\n");
                      fprintf(fp, "_geom_contact.atom_site_label_asym_id_1\n");
                      fprintf(fp, "_geom_contact.atom_site_label_comp_id_1\n");
                      fprintf(fp, "_geom_contact.atom_site_label_seq_id_1\n");
                      fprintf(fp, "_geom_contact.atom_site_label_alt_id_1\n");
                      fprintf(fp, "_geom_contact.site_symmetry_1\n");
                      fprintf(fp, "_geom_contact.atom_site_auth_atom_id_1\n");
                      fprintf(fp, "_geom_contact.atom_site_auth_asym_id_1\n");
                      fprintf(fp, "_geom_contact.atom_site_auth_comp_id_1\n");
                      fprintf(fp, "_geom_contact.atom_site_auth_seq_id_1\n");
                      if (exchange_out)
                           fprintf(fp, "_geom_contact.pdbx_atom_site_PDB_ins_code_1\n");
                      else fprintf(fp, "_geom_contact.ndb_atom_site_PDB_ins_code_1\n");
                      fprintf(fp, "_geom_contact.atom_site_id_2\n");
                      fprintf(fp, "_geom_contact.atom_site_label_atom_id_2\n");
                      fprintf(fp, "_geom_contact.atom_site_label_asym_id_2\n");
                      fprintf(fp, "_geom_contact.atom_site_label_comp_id_2\n");
                      fprintf(fp, "_geom_contact.atom_site_label_seq_id_2\n");
                      fprintf(fp, "_geom_contact.atom_site_label_alt_id_2\n");
                      fprintf(fp, "_geom_contact.site_symmetry_2\n");
                      fprintf(fp, "_geom_contact.atom_site_auth_atom_id_2\n");
                      fprintf(fp, "_geom_contact.atom_site_auth_asym_id_2\n");
                      fprintf(fp, "_geom_contact.atom_site_auth_comp_id_2\n");
                      fprintf(fp, "_geom_contact.atom_site_auth_seq_id_2\n");
                      if (exchange_out)
                           fprintf(fp, "_geom_contact.pdbx_atom_site_PDB_ins_code_2\n");
                      else fprintf(fp, "_geom_contact.ndb_atom_site_PDB_ins_code_2\n");
                      fprintf(fp, "_geom_contact.dist\n");
                 }
                 for (i = 0; i < n_contact; i++) {
                      atom_a = _molecules[l]->find_ndb_atom(contact[i]->a_atomtyp,
                                    contact[i]->a_chnid, contact[i]->a_resnam,
                                    contact[i]->a_resnum, "", num_a);
                      atom_b = _molecules[l]->find_ndb_atom(contact[i]->b_atomtyp,
                                    contact[i]->b_chnid, contact[i]->b_resnam,
                                    contact[i]->b_resnum, "", num_b);
                      if (!atom_a || !atom_b) {
                           if (atom_a) delete [] atom_a;
                           if (atom_b) delete [] atom_b;
                           continue;
                      }
                      for (j = 0; j < num_a; j++) {
                           for (k = 0; k < num_b; k++) {
                                if (contact[i]->a_alt_loc == atom_a[j]->alt_loc()[0] &&
                                    contact[i]->b_alt_loc == atom_b[k]->alt_loc()[0]) {
                                     sprintf(buffer, "%d_%d%d%d", contact[i]->sym,
                                              contact[i]->transl[0] + 5,
                                              contact[i]->transl[1] + 5,
                                              contact[i]->transl[2] + 5);
                                     fprintf(fp, "%d ", contact[i]->Mol_ID);
                                     write_atom_to_file(fp, atom_a[j], "1_555");
                                     fprintf(fp, " ");
                                     write_atom_to_file(fp, atom_b[k], buffer);
                                     fprintf(fp, " %.3f\n", contact[i]->dist);
                                }
                           }
                      }
                      if (atom_a) delete [] atom_a;
                      if (atom_b) delete [] atom_b;
                      if (contact[i]) delete contact[i];
                 }
            }
            if (contact) delete [] contact;
            contact = NULL;
            delete mycrys;
       }
}

void write_atom_to_file(FILE *fp, _ATOM *atom, const char *symmetry)
{
       char chnid[5], resnum[10], alt_loc[5], pdb_chnid[5], pdb_resnum[10], ins_code[5];

       if (strcmp(atom->chnid(), ""))
            strcpy(chnid, atom->chnid());
       else strcpy(chnid, "?");

       if (atom->resnum() == (-99999))
            strcpy(resnum, ".");
       else sprintf(resnum, "%d", atom->resnum());

       if (strcmp(atom->alt_loc(), ""))
            strcpy(alt_loc, atom->alt_loc());
       else strcpy(alt_loc, "?");

       if (strcmp(atom->pdb_chnid(), ""))
            strcpy(pdb_chnid, atom->pdb_chnid());
       else strcpy(pdb_chnid, "?");

       if (atom->pdb_resnum() == (-99999))
            strcpy(pdb_resnum, ".");
       else sprintf(pdb_resnum, "%d", atom->pdb_resnum());

       if (strcmp(atom->ins_code(), ""))
            strcpy(ins_code, atom->ins_code());
       else strcpy(ins_code, "?");

       fprintf(fp, "%d %s %s %s %s %s %s %s %s %s %s %s", atom->atnum(), atom->atmtype(),
                    chnid, atom->restype(), resnum, alt_loc, symmetry, atom->pdb_atmnam(),
                    pdb_chnid, atom->pdb_resnam(), pdb_resnum, ins_code);
}

void _VALIDATION::calculate_all_contact(const double high_cut_off, const int remove_hydrogen,
                         const int include_type,  const bool include_symmetry)
{
       if (_num_mol == 0) return;
       if (_cell == NULL) return;

       Clear_Contact();

       if (_cell->is_artifical) {
            calculate_contact_for_nmr();
            return;
       }

       xtal mycrys;
       mycrys.setorth('c');    
       if (mycrys.init(_cell, _molecules[0], remove_hydrogen) <= 0) return;
       mycrys.getSymmAtoms();
       mycrys.moveAllAtomsToCenter();
       mycrys.placeAtomsInGrid();

       if (include_type & 1) {
            _a_contact = get_contact(mycrys, _n_a_contact,
                   0.0, 2.2, "asym", "atnum 1 -1", "atnum 1 -1", 1, 1);
            get_special_and_symmetry_contact(mycrys, _n_a_contact, _a_contact, _n_special,
                       _alloc_special, _special, _n_s_contact, _alloc_s_contact,
                       _s_contact, _molecules);
       }

       if (!(include_type & 2)) return;

       bool found = false;
       _CHAIN *chain = _molecules[0]->GetFirstChain();
       while (chain) {
            if (chain->chain_type() == HETAS_TOKEN) {
                 _RESIDUE *residue = chain->GetFirstResidue();
                 if (!strcmp(residue->ResName(), "HOH") ||
                     !strcmp(residue->ResName(), "DOD")) {
                      found = true;
                      break;
                 }
            }
            chain = _molecules[0]->GetNextChain();
       }
       if (!found) return;

       int all_num = 0;
       std::map<std::string, int> _AllWaters;
       CONTACT **all_waters = get_all_waters(_molecules[0], all_num, _AllWaters);
       if (all_waters == NULL || all_num == 0 || _AllWaters.empty()) return;

       bool using_kim_method = false;
       if (using_kim_method)
            find_outlier_waters_without_shells(mycrys, high_cut_off, include_symmetry,
                        all_waters, _AllWaters);
       else find_outlier_waters_with_shells(mycrys, high_cut_off, all_waters, _AllWaters);

       get_outlier_and_moving_waters(_AllWaters, all_waters, _n_waters, _alloc_waters,
                  _waters, _n_moving_waters, _alloc_moving_waters, _moving_waters);

       if (all_waters) {
            for (int i = 0; i < all_num; i++)
                 if (all_waters[i]) delete all_waters[i];
            delete [] all_waters;
       }
}

int _VALIDATION::find_a_link(CONTACT *contact)
{
       int i, FstIndex, FstRes, SndIndex, SndRes;
       char SymOP_1[7], SymOP_2[7];

       strcpy(SymOP_1, "");
       sprintf(SymOP_2, "%d_%d%d%d", contact->sym, contact->transl[0] + 5,
                 contact->transl[1] + 5, contact->transl[2] + 5);
       if (!strcmp(SymOP_2, "1_555")) strcpy(SymOP_2, "");
       _MOLECULE *mol = _molecules[0];
       if (mol->find_ndb_residue(contact->a_chnid, contact->a_resnam,
              contact->a_resnum, "", FstIndex, FstRes) &&
           mol->find_ndb_residue(contact->b_chnid, contact->b_resnam,
              contact->b_resnum, "", SndIndex, SndRes)) {
            for (i = 0; i < _n_link; i++) {
                 if (!strcmp(_link[i]->FstAtom, contact->a_atomtyp) &&
                     !strcmp(_link[i]->SymOP_1, SymOP_1) &&
                     _link[i]->FstIndex == FstIndex &&
                     _link[i]->FstRes == FstRes &&
                     !strcmp(_link[i]->SndAtom, contact->b_atomtyp) &&
                     !strcmp(_link[i]->SymOP_2, SymOP_2) &&
                     _link[i]->SndIndex == SndIndex &&
                     _link[i]->SndRes == SndRes ||
                     !strcmp(_link[i]->FstAtom, contact->b_atomtyp) &&
                     !strcmp(_link[i]->SymOP_1, SymOP_2) &&
                     _link[i]->FstIndex == SndIndex &&
                     _link[i]->FstRes == SndRes &&
                     !strcmp(_link[i]->SndAtom, contact->a_atomtyp) &&
                     !strcmp(_link[i]->SymOP_2, SymOP_1) &&
                     _link[i]->SndIndex == FstIndex &&
                     _link[i]->SndRes == FstRes) return 1;
            }
            for (i = 0; i < _n_ssbond; i++) {
                 if (!strcmp(_ssbond[i]->FstAtom, contact->a_atomtyp) &&
                     !strcmp(_ssbond[i]->SymOP_1, SymOP_1) &&
                     _ssbond[i]->FstIndex == FstIndex &&
                     _ssbond[i]->FstRes == FstRes &&
                     !strcmp(_ssbond[i]->SndAtom, contact->b_atomtyp) &&
                     !strcmp(_ssbond[i]->SymOP_2, SymOP_2) &&
                     _ssbond[i]->SndIndex == SndIndex &&
                     _ssbond[i]->SndRes == SndRes ||
                     !strcmp(_ssbond[i]->FstAtom, contact->b_atomtyp) &&
                     !strcmp(_ssbond[i]->SymOP_1, SymOP_2) &&
                     _ssbond[i]->FstIndex == SndIndex &&
                     _ssbond[i]->FstRes == SndRes &&
                     !strcmp(_ssbond[i]->SndAtom, contact->a_atomtyp) &&
                     !strcmp(_ssbond[i]->SymOP_2, SymOP_1) &&
                     _ssbond[i]->SndIndex == FstIndex &&
                     _ssbond[i]->SndRes == FstRes) return 1;
            }
       }
       return 0;
}

int _VALIDATION::is_residue_linked(_RESIDUE *residue, _RESIDUE *next)
{
       int connect = is_residue_linked(residue);
       if (connect == 2) {
            if (is_connect("C", residue, "N", next))
                 return 3;
            else return connect;
       } else if (residue != next)
            return 1;
       else return 0;
}

int _VALIDATION::is_residue_linked(_RESIDUE *residue)
{
       if (!residue) return 0;
       if (!residue->GetFirstAtom()) return 0;
       int FstIndex, FstRes;
       _MOLECULE *mol = _molecules[0];
       if (mol->find_ndb_residue(residue->GetFirstAtom()->chnid(),
               residue->GetFirstAtom()->restype(),
               residue->GetFirstAtom()->resnum(),
               "", FstIndex, FstRes)) {
            for (int i = 0; i < _n_link; i++) {
                 if (_link[i]->FstIndex == FstIndex &&
                     _link[i]->FstRes == FstRes ||
                     _link[i]->SndIndex == FstIndex &&
                     _link[i]->SndRes == FstRes) {
                      if (!strcmp(residue->ResName(), "ASP")) {
                           if (_link[i]->FstIndex == FstIndex &&
                               _link[i]->FstRes == FstRes &&
                               !strcmp(_link[i]->FstAtom, "CG"))
                                return 2;
                           else if (_link[i]->SndIndex == FstIndex &&
                                    _link[i]->SndRes == FstRes &&
                                    !strcmp(_link[i]->SndAtom, "CG"))
                                return 2;
                      } else if (!strcmp(residue->ResName(), "GLU")) {
                           if (_link[i]->FstIndex == FstIndex &&
                               _link[i]->FstRes == FstRes &&
                               !strcmp(_link[i]->FstAtom, "CD"))
                                return 2;
                           else if (_link[i]->SndIndex == FstIndex &&
                                    _link[i]->SndRes == FstRes &&
                                    !strcmp(_link[i]->SndAtom, "CD"))
                                return 2;
                      }
                      return 1;
                 }
            }
       }
       return 0;
}

CONTACT **get_contact(xtal &mycrys, int &ncontact, const double lo, const double hi,
         const char *inter, const char *fromtype, const char *totype,
         const int DeleteMark, const int filter_flag)
{
       int alloc_contact = 0, asym = 0, i = 1, j, add = 0, ret = -1;
       CONTACT **contact = NULL, **tmp = NULL;
       BTreeObj<_CINDEX> _CIndex;
       _CIndex.Clear();
       _CINDEX idx;
       List<atom*>* fromatoms = mycrys.selectAtoms(fromtype);
       List<atom*>* toatoms   = mycrys.selectAtoms(totype);

       ncontact = 0;
       if (!fromatoms || !toatoms) {
            if (fromatoms) {
                 fromatoms->DeleteList();
                 delete fromatoms;
            }
            if (toatoms) {
                 toatoms->DeleteList();
                 delete toatoms;
            }
            return NULL;
       }

       if (strcmp(inter,"asym")==0) { *toatoms=selectAsym(*toatoms); asym = 1; }
       else if (strcmp(inter,"sym")==0) asym=2;

       List<pair> pl = Contacts(*fromatoms, *toatoms, lo, hi);

       if (pl.length()) {
            ListObj<pair> *ptr = pl.front();
            pair obj;
            while (ptr) {
                 obj = ptr->firstval();
                 if (obj.TheFromAtom().alt_loc() != obj.TheToAtom().alt_loc() &&
                     obj.TheFromAtom().alt_loc() != '\0' &&
                     obj.TheToAtom().alt_loc() != '\0') {
                      ptr = ptr->rest();
                      continue;
                 }
                 i = 1;
                 if (asym == 1) {
                      if (!obj.TheToAtom().symnum() && !obj.TheToAtom().translx() &&
                          !obj.TheToAtom().transly() && !obj.TheToAtom().translz())
                           i = 0;
                 } else if (asym == 2) {
                      if (obj.TheToAtom().symnum() || obj.TheToAtom().translx() ||
                          obj.TheToAtom().transly() || obj.TheToAtom().translz())
                           i = 0;
                 } else i = 0;

                 if (i <= 0) {
                      if (filter_flag) {
/*
                           if (obj.pairDist() >= 1.8 && obj.pairDist() <= 2.2 &&
                               !obj.TheToAtom().symnum() && !obj.TheToAtom().translx() &&
                               !obj.TheToAtom().transly() && !obj.TheToAtom().translz() &&
                               !strncmp(obj.TheFromAtom().attyp(), "SG", 2) &&
                               !strncmp(obj.TheToAtom().attyp(), "SG", 2) &&
                              (!strncmp(obj.TheToAtom().resType(), "CYS", 3) ||
                               !strncmp(obj.TheToAtom().resType(), "DCY", 3)) &&
                              (!strncmp(obj.TheFromAtom().resType(), "CYS", 3) ||
                               !strncmp(obj.TheFromAtom().resType(), "DCY", 3)))
                                i = 1;
*/
                           if (!strcmp(obj.TheFromAtom().typ(), "ATOMP") &&
                               !strcmp(obj.TheToAtom().typ(), "ATOMP") &&
                                !strcmp(obj.TheFromAtom().chain(),
                                        obj.TheFromAtom().chain()) &&
                               (obj.TheFromAtom().res() + 1 == obj.TheToAtom().res() ||
                                obj.TheFromAtom().res() ==  obj.TheToAtom().res() + 1) &&
                                obj.pairDist() >= 1.8)
                                i = 1;
                           else if (!strcmp(obj.TheFromAtom().atomtype(), "D") &&
                                    !strcmp(obj.TheToAtom().atomtype(), "D")) {
                                if (obj.pairDist() >= 1.415) i = 1;
                           } else if ((!strcmp(obj.TheFromAtom().atomtype(), "H") ||
                                     !strcmp(obj.TheFromAtom().atomtype(), "D")) &&
                                    (!strcmp(obj.TheToAtom().atomtype(), "H") ||
                                     !strcmp(obj.TheToAtom().atomtype(), "D"))) {
                                if (obj.pairDist() >= 1.35) i = 1;
                           } else if ((!strcmp(obj.TheFromAtom().atomtype(), "H") ||
                                     !strcmp(obj.TheToAtom().atomtype(), "H") ||
                                     !strcmp(obj.TheFromAtom().atomtype(), "D") ||
                                     !strcmp(obj.TheToAtom().atomtype(), "D")) &&
                                     obj.pairDist() >= 1.6)
                                i = 1;
                           else if (obj.TheFromAtom().num() == obj.TheToAtom().num() &&
                                      obj.pairDist() < 0.3)
                                i = -1;
                           else if (obj.TheFromAtom().occupancy() < 1.0 &&
                                    obj.TheToAtom().occupancy() < 1.0) {
/*
                                if (strcmp(inter, "asym")) i = -1;
                                i = -1;
*/
                                if (obj.TheToAtom().symnum() && 
                                    !strcmp(obj.TheFromAtom().resType(), obj.TheToAtom().resType()) &&
                                    !strcmp(obj.TheFromAtom().chain(), obj.TheToAtom().chain()) &&
                                    obj.TheFromAtom().res() == obj.TheToAtom().res()) {
                                     i = -1;
                                }
                           } else if ((!strcmp(obj.TheFromAtom().atomtype(), "BA") ||
                                     !strcmp(obj.TheFromAtom().atomtype(), "MG") ||
                                     !strcmp(obj.TheFromAtom().atomtype(), "NA") ||
                                     !strcmp(obj.TheFromAtom().atomtype(), "NI") ||
                                     !strcmp(obj.TheFromAtom().atomtype(), "CU") ||
                                     !strcmp(obj.TheFromAtom().atomtype(), "CO") ||
                                    (!strcmp(obj.TheFromAtom().atomtype(), "CA") ||
                                     !strcmp(obj.TheFromAtom().atomtype(), "HG")) &&
                                      strncmp(obj.TheFromAtom().typ(), "ATOM", 4) ||
                                     !strcmp(obj.TheFromAtom().atomtype(), "MN") ||
                                     !strcmp(obj.TheFromAtom().atomtype(), "PT") ||
                                     !strcmp(obj.TheFromAtom().atomtype(), "ZN") ||
                                     !strcmp(obj.TheToAtom().atomtype(), "BA") ||
                                     !strcmp(obj.TheToAtom().atomtype(), "MG") ||
                                     !strcmp(obj.TheToAtom().atomtype(), "NA") ||
                                     !strcmp(obj.TheToAtom().atomtype(), "NI") ||
                                     !strcmp(obj.TheToAtom().atomtype(), "CU") ||
                                     !strcmp(obj.TheToAtom().atomtype(), "CO") ||
                                    (!strcmp(obj.TheToAtom().atomtype(), "CA") ||
                                     !strcmp(obj.TheToAtom().atomtype(), "HG")) &&
                                      strncmp(obj.TheToAtom().typ(), "ATOM", 4) ||
                                     !strcmp(obj.TheToAtom().atomtype(), "MN") ||
                                     !strcmp(obj.TheToAtom().atomtype(), "PT") ||
                                     !strcmp(obj.TheToAtom().atomtype(), "ZN")) &&
                                     obj.pairDist() >= 1.7)
                                i = 1;
                           else if (hi <= 2.2 && obj.pairDist() >= 2.195)
                                i = 1;
                      }
                      if (i <= 0) {
                           add = 1;
                           if (DeleteMark) {
                                if (obj.TheFromAtom().num() <
                                    obj.TheToAtom().num()) {
                                     idx.set_f_num(obj.TheFromAtom().num());
                                     idx.set_t_num(obj.TheToAtom().num());
                                } else {
                                     idx.set_f_num(obj.TheToAtom().num());
                                     idx.set_t_num(obj.TheFromAtom().num());
                                }
                                idx.set_pos(ncontact);
                                ret = _CIndex.Seek(idx);
                                if (ret >= 0) {
                                     idx = _CIndex.GetCurrData();
                                     if (!obj.TheToAtom().symnum() &&
                                         !obj.TheToAtom().translx() &&
                                         !obj.TheToAtom().transly() &&
                                         !obj.TheToAtom().translz() &&
                                         (contact[idx.pos()]->sym > 1 ||
                                          contact[idx.pos()]->transl[0] ||
                                          contact[idx.pos()]->transl[1] ||
                                          contact[idx.pos()]->transl[2]) ||
                                          contact[idx.pos()]->dist > obj.pairDist()) {
                                          strcpy(contact[idx.pos()]->a_type,
                                                 obj.TheFromAtom().typ());
                                          strcpy(contact[idx.pos()]->a_atomtyp,
                                                 obj.TheFromAtom().attyp());
                                          strcpy(contact[idx.pos()]->a_atom_type,
                                                 obj.TheFromAtom().atomtype());
                                          contact[idx.pos()]->a_atomnum =
                                                 obj.TheFromAtom().num();
                                          strcpy(contact[idx.pos()]->a_resnam,
                                                 obj.TheFromAtom().resType());
                                          strcpy(contact[idx.pos()]->a_chnid,
                                                 obj.TheFromAtom().chain());
                                          contact[idx.pos()]->a_resnum =
                                                 obj.TheFromAtom().res();
                                          contact[idx.pos()]->a_alt_loc =
                                                 obj.TheFromAtom().alt_loc();
                                          contact[idx.pos()]->a_occupancy =
                                                 obj.TheFromAtom().occupancy();
                    
                                          strcpy(contact[idx.pos()]->b_type,
                                                 obj.TheToAtom().typ());
                                          strcpy(contact[idx.pos()]->b_atomtyp,
                                                 obj.TheToAtom().attyp());
                                          strcpy(contact[idx.pos()]->b_atom_type,
                                                 obj.TheToAtom().atomtype());
                                          contact[idx.pos()]->b_atomnum =
                                                 obj.TheToAtom().num();
                                          strcpy(contact[idx.pos()]->b_resnam,
                                                 obj.TheToAtom().resType());
                                          strcpy(contact[idx.pos()]->b_chnid,
                                                 obj.TheToAtom().chain());
                                          contact[idx.pos()]->b_resnum =
                                                 obj.TheToAtom().res();
                                          contact[idx.pos()]->b_alt_loc =
                                                 obj.TheToAtom().alt_loc();
                                          contact[idx.pos()]->b_occupancy =
                                                 obj.TheToAtom().occupancy();
               
                                          contact[idx.pos()]->sym =
                                                 obj.TheToAtom().symnum() + 1;
                                          contact[idx.pos()]->transl[0] =
                                                 obj.TheToAtom().translx();
                                          contact[idx.pos()]->transl[1] =
                                                 obj.TheToAtom().transly();
                                          contact[idx.pos()]->transl[2] =
                                                 obj.TheToAtom().translz();
                                          contact[idx.pos()]->dist = obj.pairDist();
                                          contact[idx.pos()]->type = i;
                                     }
                                     add = 0;
                                } else _CIndex.Add(idx);
                           }
                           if (add) {
                                if (ncontact >= alloc_contact) {
                                     alloc_contact += 300;
                                     tmp = contact;
                                     contact = new CONTACT*[alloc_contact];
                                     for (j = 0; j < ncontact; j++)
                                          contact[j] = tmp[j];
                                     for (j = ncontact; j < alloc_contact; j++)
                                          contact[j] = NULL;
                                     delete [] tmp;
                                }
                                contact[ncontact] = new CONTACT;
                                strcpy(contact[ncontact]->a_type,
                                       obj.TheFromAtom().typ());
                                strcpy(contact[ncontact]->a_atomtyp,
                                       obj.TheFromAtom().attyp());
                                strcpy(contact[ncontact]->a_atom_type,
                                       obj.TheFromAtom().atomtype());
                                contact[ncontact]->a_atomnum =
                                       obj.TheFromAtom().num();
                                strcpy(contact[ncontact]->a_resnam,
                                       obj.TheFromAtom().resType());
                                strcpy(contact[ncontact]->a_chnid,
                                       obj.TheFromAtom().chain());
                                contact[ncontact]->a_resnum =
                                       obj.TheFromAtom().res();
                                contact[ncontact]->a_alt_loc =
                                       obj.TheFromAtom().alt_loc();
                                contact[ncontact]->a_occupancy =
                                       obj.TheFromAtom().occupancy();
          
                                strcpy(contact[ncontact]->b_type,
                                       obj.TheToAtom().typ());
                                strcpy(contact[ncontact]->b_atomtyp,
                                       obj.TheToAtom().attyp());
                                strcpy(contact[ncontact]->b_atom_type,
                                       obj.TheToAtom().atomtype());
                                contact[ncontact]->b_atomnum =
                                       obj.TheToAtom().num();
                                strcpy(contact[ncontact]->b_resnam,
                                       obj.TheToAtom().resType());
                                strcpy(contact[ncontact]->b_chnid,
                                       obj.TheToAtom().chain());
                                contact[ncontact]->b_resnum =
                                       obj.TheToAtom().res();
                                contact[ncontact]->b_alt_loc =
                                       obj.TheToAtom().alt_loc();
                                contact[ncontact]->b_occupancy =
                                       obj.TheToAtom().occupancy();
     
                                contact[ncontact]->sym =
                                       obj.TheToAtom().symnum() + 1;
                                contact[ncontact]->transl[0] =
                                       obj.TheToAtom().translx();
                                contact[ncontact]->transl[1] =
                                       obj.TheToAtom().transly();
                                contact[ncontact]->transl[2] =
                                       obj.TheToAtom().translz();
     
                                contact[ncontact]->dist = obj.pairDist();
                                contact[ncontact]->type = i;
                                contact[ncontact]->Mol_ID = 1;
                                ncontact++;
                           }
                      }
                 }
                 ptr = ptr->rest();
            }
       }

       DeletePairsAtoms(pl);
       pl.DeleteList();
  
       unmarkAsToAtoms(*toatoms);
       mycrys.crysGrid()->reset();
       fromatoms->DeleteList();
       toatoms->DeleteList();
       delete fromatoms;
       delete toatoms;

       if (contact == NULL) return contact;

       tmp = contact;
       contact = new CONTACT*[ncontact];
       BTreeObj<_DIndex> _Index;
       _DIndex _idx;
       _Index.Clear();
       for (j = 0; j < ncontact; j++) {
            _idx.set_value(tmp[j]->dist);
            _idx.set_pos(j);
            _Index.Add(_idx);
       }
       _Index.GoFirst();
       ret = 1;
       i = 0;
       while (ret >= 0) {
            _idx = _Index.GetCurrData();
            j = _idx.pos();
            contact[i] = tmp[j];
            i++;
            ret = _Index.GoToNext();
       }
       delete [] tmp;

       return contact;
}

static void get_special_and_symmetry_contact(xtal& mycrys, const int _n_a_contact,
                 CONTACT**& _a_contact, int& _n_special, int& _alloc_special,
                 CONTACT**& _special, int& _n_s_contact, int& _alloc_s_contact,
                 CONTACT**& _s_contact, _MOLECULE** _molecules)
{
       _CINDEX idx;
       BTreeObj<_CINDEX> _CIndex;
       _CIndex.Clear();

       for (int i = 0; i < _n_a_contact; i++) {
            _a_contact[i]->Mol_ID = _molecules[0]->Mol_ID();
            if (_a_contact[i]->a_atomnum < _a_contact[i]->b_atomnum) {
                 idx.set_f_num(_a_contact[i]->a_atomnum);
                 idx.set_t_num(_a_contact[i]->b_atomnum);
            } else {
                 idx.set_f_num(_a_contact[i]->b_atomnum);
                 idx.set_t_num(_a_contact[i]->a_atomnum);
            }
            idx.set_pos(i);
            _CIndex.Add(idx);
       }

       int n_sym_contact = 0;
       CONTACT **sym_contact = get_contact(mycrys, n_sym_contact,
                   0.0, 2.2, "sym", "atnum 1 -1", "atnum 1 -1", 1, 1);
       if (sym_contact == NULL || n_sym_contact == 0) return;

       for (int i = 0; i < n_sym_contact; i++) {
            if (sym_contact[i]->type < 0) {
                 if (sym_contact[i]->dist < 0.3 &&
                     sym_contact[i]->a_resnum == sym_contact[i]->b_resnum &&
                    !strcmp(sym_contact[i]->a_atomtyp, sym_contact[i]->b_atomtyp) &&
                    !strcmp(sym_contact[i]->a_resnam, sym_contact[i]->b_resnam) &&
                    !strcmp(sym_contact[i]->a_chnid, sym_contact[i]->b_chnid)) {
                      if (_n_special >= _alloc_special) {
                           _alloc_special += 10;
                           CONTACT **tmp = _special;
                           _special = new CONTACT*[_alloc_special];
                           for (int j = 0; j < _n_special; j++)
                                _special[j] = tmp[j];
                           for (int j = _n_special; j < _alloc_special; j++)
                                _special[j] = NULL;
                           delete [] tmp;
                      }
                      bool found = false;
                      for (int j = 0; j < _n_special; j++) {
                           if (!strcmp(_special[j]->a_resnam, sym_contact[i]->a_resnam) &&
                               !strcmp(_special[j]->a_chnid, sym_contact[i]->a_chnid) &&
                               !strcmp(_special[j]->a_atomtyp, sym_contact[i]->a_atomtyp) &&
                                _special[j]->a_resnum == sym_contact[i]->a_resnum) {
                                found = true;
                                break;
                           }
                      }
                      if (!found) {
                           _special[_n_special] = sym_contact[i];
                           sym_contact[i] = NULL;
                           _n_special++;
                      }
                 }
            } else {
                 if (sym_contact[i]->a_atomnum < sym_contact[i]->b_atomnum) {
                      idx.set_f_num(sym_contact[i]->a_atomnum);
                      idx.set_t_num(sym_contact[i]->b_atomnum);
                 } else {
                      idx.set_f_num(sym_contact[i]->b_atomnum);
                      idx.set_t_num(sym_contact[i]->a_atomnum);
                 }
                 if (_CIndex.Seek(idx) < 0) {
                      if (_n_s_contact >= _alloc_s_contact) {
                           _alloc_s_contact += 10;
                           CONTACT **tmp = _s_contact;
                           _s_contact = new CONTACT*[_alloc_s_contact];
                           for (int j = 0; j < _n_s_contact; j++)
                                _s_contact[j] = tmp[j];
                           for (int j = _n_s_contact; j < _alloc_s_contact; j++)
                                _s_contact[j] = NULL;
                           delete [] tmp;
                      }
                      _s_contact[_n_s_contact] = sym_contact[i];
                      _s_contact[_n_s_contact]->Mol_ID = _molecules[0]->Mol_ID();
                      sym_contact[i] = NULL;
                      _n_s_contact++;
                      _CIndex.Add(idx);
                 }
            }
       }

       for (int i = 0; i < n_sym_contact; i++) {
            if (sym_contact[i]) delete sym_contact[i];
       }
       delete [] sym_contact;
}

static CONTACT **get_all_waters(_MOLECULE *mol, int& all_num,
                         std::map<std::string, int>& _AllWaters) 
{
       all_num = 0;
       _AllWaters.clear();
       int alloc_num = 0;
       CONTACT **all_waters = NULL;

       _CHAIN *chain = mol->GetFirstChain();
       while (chain) {
            if (chain->chain_type() != HETAS_TOKEN) {
                 chain = mol->GetNextChain();
                 continue;
            }
            _RESIDUE *residue = chain->GetFirstResidue();
            while (residue) {
                 if (strcmp(residue->ResName(), "HOH") &&
                     strcmp(residue->ResName(), "DOD")) {
                      residue = chain->GetNextResidue();
                      continue;
                 }
                 _ATOM *atom = residue->GetFirstAtom();
                 while (atom) {
                      if (!strcmp(atom->atom_type(), "O")) {
                           if (all_num >= alloc_num) {
                                alloc_num += 10;
                                CONTACT **tmp = all_waters;
                                all_waters = new CONTACT*[alloc_num];
                                for (int i = 0; i < all_num; i++)
                                     all_waters[i] = tmp[i];
                                for (int i = all_num; i < alloc_num; i++)
                                     all_waters[i] = NULL;
                                delete [] tmp;
                           }
                           all_waters[all_num] = new CONTACT;
                           strcpy(all_waters[all_num]->a_atomtyp, atom->atmtype());
                           strcpy(all_waters[all_num]->a_atom_type, atom->atom_type());
                           all_waters[all_num]->a_atomnum = atom->atnum();
                           strcpy(all_waters[all_num]->a_resnam, atom->restype());
                           all_waters[all_num]->a_resnum = atom->resnum();
                           strcpy(all_waters[all_num]->a_chnid, atom->chnid());
                           strcpy(all_waters[all_num]->b_chnid, atom->pdb_chnid());
                           all_waters[all_num]->b_resnum = atom->pdb_resnum();
                           all_waters[all_num]->type = 1;
                           all_waters[all_num]->sym = 0;
                           all_waters[all_num]->transl[0] = 0;
                           all_waters[all_num]->transl[1] = 0;
                           all_waters[all_num]->transl[2] = 0;

                           std::string index = get_index(atom->chnid(), atom->resnum());
                           _AllWaters.insert(std::make_pair(index, all_num));
                           all_num++;
                      }
                      atom = residue->GetNextAtom();
                 }
                 residue = chain->GetNextResidue();
            }
            chain = mol->GetNextChain();
       }
       return all_waters;
}

static void find_outlier_waters_with_shells(xtal& mycrys, const double high_cut_off,
                      CONTACT**& all_waters, std::map<std::string, int>& _AllWaters)
{
       int n_asym_water_s_3_5 = 0;
       CONTACT **asym_water_s_3_5 = get_contact(mycrys, n_asym_water_s_3_5, 0.45,
                      WATER_WATER_LIMIT, "asym", "kind hetas", "atnum 1 -1", 1, 1);

       int n_sym_water_s_3_5 = 0;
       CONTACT **sym_water_s_3_5 = get_contact(mycrys, n_sym_water_s_3_5, 0.45,
                     WATER_WATER_LIMIT, "sym", "atnum 1 -1", "kind hetas", 0, 1);

       int n_asym_water_water = 0;
       CONTACT **asym_water_water = get_contact(mycrys, n_asym_water_water, 0.45,
                      WATER_WATER_LIMIT, "asym", "kind hetas", "kind hetas", 0, 1);

       int n_asym_water_S_3_5 = 0;
       CONTACT **asym_water_S_3_5 = NULL;
       if (high_cut_off > WATER_WATER_LIMIT) {
            asym_water_S_3_5 = get_contact(mycrys, n_asym_water_S_3_5, WATER_WATER_LIMIT
                      - 0.01, high_cut_off, "asym", "kind hetas", "atnum 1 -1", 1, 1);
       }

       char type[8];
       for (int i = 0; i < n_sym_water_s_3_5; i++) {
            strcpy(type, sym_water_s_3_5[i]->a_type);
            strcpy(sym_water_s_3_5[i]->a_type, sym_water_s_3_5[i]->b_type);
            strcpy(sym_water_s_3_5[i]->b_type, type);

            strcpy(type, sym_water_s_3_5[i]->a_atomtyp);
            strcpy(sym_water_s_3_5[i]->a_atomtyp, sym_water_s_3_5[i]->b_atomtyp);
            strcpy(sym_water_s_3_5[i]->b_atomtyp, type);

            strcpy(type, sym_water_s_3_5[i]->a_atom_type);
            strcpy(sym_water_s_3_5[i]->a_atom_type, sym_water_s_3_5[i]->b_atom_type);
            strcpy(sym_water_s_3_5[i]->b_atom_type, type);

            int num = sym_water_s_3_5[i]->a_atomnum;
            sym_water_s_3_5[i]->a_atomnum = sym_water_s_3_5[i]->b_atomnum;
            sym_water_s_3_5[i]->b_atomnum = num;

            strcpy(type, sym_water_s_3_5[i]->a_resnam);
            strcpy(sym_water_s_3_5[i]->a_resnam, sym_water_s_3_5[i]->b_resnam);
            strcpy(sym_water_s_3_5[i]->b_resnam, type);

            num = sym_water_s_3_5[i]->a_resnum;
            sym_water_s_3_5[i]->a_resnum = sym_water_s_3_5[i]->b_resnum;
            sym_water_s_3_5[i]->b_resnum = num;

            strcpy(type, sym_water_s_3_5[i]->a_chnid);
            strcpy(sym_water_s_3_5[i]->a_chnid, sym_water_s_3_5[i]->b_chnid);
            strcpy(sym_water_s_3_5[i]->b_chnid, type);
       }

       std::set<std::string> _ShellWaters;
       _ShellWaters.clear();

       for (int i = 0; i < n_asym_water_s_3_5; i++) {
            if ((asym_water_s_3_5[i]->b_atom_type[0] == 'N' ||
                 asym_water_s_3_5[i]->b_atom_type[0] == 'O') &&
                 strcmp(asym_water_s_3_5[i]->b_resnam, "HOH") &&
                 strcmp(asym_water_s_3_5[i]->b_resnam, "DOD")) {
                 std::string index = get_index(asym_water_s_3_5[i]->a_chnid,
                                               asym_water_s_3_5[i]->a_resnum);
                 if (_AllWaters.find(index) != _AllWaters.end()) {
                      _AllWaters.erase(index);
                      _ShellWaters.insert(index);
                 }
            }
       }

       if (!_AllWaters.empty()) {
            if (n_asym_water_water) get_all_shell_atoms(_AllWaters, _ShellWaters,
                                         n_asym_water_water, asym_water_water);

            if (!_AllWaters.empty() && n_asym_water_S_3_5 && asym_water_S_3_5 != NULL) {
                 for (int i = 0; i < n_asym_water_S_3_5; i++) {
                      if ((asym_water_S_3_5[i]->b_atom_type[0] == 'N' ||
                           asym_water_S_3_5[i]->b_atom_type[0] == 'O') &&
                           strcmp(asym_water_S_3_5[i]->b_resnam, "HOH") &&
                           strcmp(asym_water_S_3_5[i]->b_resnam, "DOD")) {
                           std::string index = get_index(asym_water_S_3_5[i]->a_chnid,
                                                         asym_water_S_3_5[i]->a_resnum);
                           _AllWaters.erase(index);
                      }
                 }
            }
       }

       if (!_AllWaters.empty()) {
            for (int i = 0; i < n_sym_water_s_3_5; i++) {
                 if (sym_water_s_3_5[i]->b_atom_type[0] == 'N' ||
                     sym_water_s_3_5[i]->b_atom_type[0] == 'O') {
                      std::string index = get_index(sym_water_s_3_5[i]->a_chnid,
                                                    sym_water_s_3_5[i]->a_resnum);
                      std::map<std::string, int>::iterator pos = _AllWaters.find(index);
                      if (pos != _AllWaters.end()) {
                           all_waters[pos->second]->sym = sym_water_s_3_5[i]->sym - 1;
                           all_waters[pos->second]->transl[0] =
                                                        sym_water_s_3_5[i]->transl[0];
                           all_waters[pos->second]->transl[1] =
                                                        sym_water_s_3_5[i]->transl[1];
                           all_waters[pos->second]->transl[2] =
                                                        sym_water_s_3_5[i]->transl[2];
                      }
                 }
            }
       }

       if (asym_water_s_3_5) {
            for (int i = 0; i < n_asym_water_s_3_5; i++) {
                 if (asym_water_s_3_5[i]) delete asym_water_s_3_5[i];
            }
            delete [] asym_water_s_3_5;
       }

       if (sym_water_s_3_5) {
            for (int i = 0; i < n_sym_water_s_3_5; i++) {
                 if (sym_water_s_3_5[i]) delete sym_water_s_3_5[i];
            }
            delete [] sym_water_s_3_5;
       }
       if (asym_water_water) {
            for (int i = 0; i < n_asym_water_water; i++) {
                 if (asym_water_water[i]) delete asym_water_water[i];
            }
            delete [] asym_water_water;
       }
       if (asym_water_S_3_5) {
            for (int i = 0; i < n_asym_water_S_3_5; i++) {
                 if (asym_water_S_3_5[i]) delete asym_water_S_3_5[i];
            }
            delete [] asym_water_S_3_5;
       }
}

static void find_outlier_waters_without_shells(xtal& mycrys, const double high_cut_off,
                      const bool include_symmetry, CONTACT**& all_waters,
                      std::map<std::string, int>& _AllWaters)
{
       int n_asym_water_s_3_5 = 0;
       CONTACT **asym_water_s_3_5 = get_contact(mycrys, n_asym_water_s_3_5, 0.45,
                         high_cut_off, "asym", "kind hetas", "atnum 1 -1", 1, 1);

       int n_sym_water_s_3_5 = 0;
       CONTACT **sym_water_s_3_5 = get_contact(mycrys, n_sym_water_s_3_5, 0.45,
                         high_cut_off, "sym", "atnum 1 -1", "kind hetas", 0, 1);

       char type[8];
       for (int i = 0; i < n_sym_water_s_3_5; i++) {
            strcpy(type, sym_water_s_3_5[i]->a_type);
            strcpy(sym_water_s_3_5[i]->a_type, sym_water_s_3_5[i]->b_type);
            strcpy(sym_water_s_3_5[i]->b_type, type);

            strcpy(type, sym_water_s_3_5[i]->a_atomtyp);
            strcpy(sym_water_s_3_5[i]->a_atomtyp, sym_water_s_3_5[i]->b_atomtyp);
            strcpy(sym_water_s_3_5[i]->b_atomtyp, type);

            strcpy(type, sym_water_s_3_5[i]->a_atom_type);
            strcpy(sym_water_s_3_5[i]->a_atom_type, sym_water_s_3_5[i]->b_atom_type);
            strcpy(sym_water_s_3_5[i]->b_atom_type, type);

            int num = sym_water_s_3_5[i]->a_atomnum;
            sym_water_s_3_5[i]->a_atomnum = sym_water_s_3_5[i]->b_atomnum;
            sym_water_s_3_5[i]->b_atomnum = num;

            strcpy(type, sym_water_s_3_5[i]->a_resnam);
            strcpy(sym_water_s_3_5[i]->a_resnam, sym_water_s_3_5[i]->b_resnam);
            strcpy(sym_water_s_3_5[i]->b_resnam, type);

            num = sym_water_s_3_5[i]->a_resnum;
            sym_water_s_3_5[i]->a_resnum = sym_water_s_3_5[i]->b_resnum;
            sym_water_s_3_5[i]->b_resnum = num;

            strcpy(type, sym_water_s_3_5[i]->a_chnid);
            strcpy(sym_water_s_3_5[i]->a_chnid, sym_water_s_3_5[i]->b_chnid);
            strcpy(sym_water_s_3_5[i]->b_chnid, type);
       }

       for (int i = 0; i < n_asym_water_s_3_5; i++) {
            if ((!strcmp(asym_water_s_3_5[i]->b_atom_type, "N") ||
                 !strcmp(asym_water_s_3_5[i]->b_atom_type, "O")) &&
                 strcmp(asym_water_s_3_5[i]->b_resnam, "HOH") &&
                 strcmp(asym_water_s_3_5[i]->b_resnam, "DOD") ||
                 find_metal_flag(asym_water_s_3_5[i]->b_atom_type) == 2) {
                 std::string index = get_index(asym_water_s_3_5[i]->a_chnid,
                                               asym_water_s_3_5[i]->a_resnum);
                 if (_AllWaters.find(index) != _AllWaters.end()) {
                      _AllWaters.erase(index);
                 }
            }
       }

       if (include_symmetry) {
            for (int i = 0; i < n_sym_water_s_3_5; i++) {
                 if ((!strcmp(sym_water_s_3_5[i]->b_atom_type, "N") ||
                      !strcmp(sym_water_s_3_5[i]->b_atom_type, "O")) &&
                      strcmp(sym_water_s_3_5[i]->b_resnam, "HOH") &&
                      strcmp(sym_water_s_3_5[i]->b_resnam, "DOD")) {
                      std::string index = get_index(sym_water_s_3_5[i]->a_chnid,
                                                    sym_water_s_3_5[i]->a_resnum);
                      if (_AllWaters.find(index) != _AllWaters.end()) {
                           _AllWaters.erase(index);
                      }
                 }
            }
       }

       if (!_AllWaters.empty()) {
            for (int i = 0; i < n_sym_water_s_3_5; i++) {
                 if ((!strcmp(sym_water_s_3_5[i]->b_atom_type, "N") ||
                      !strcmp(sym_water_s_3_5[i]->b_atom_type, "O")) &&
                      strcmp(sym_water_s_3_5[i]->b_resnam, "HOH") &&
                      strcmp(sym_water_s_3_5[i]->b_resnam, "DOD")) {
                      std::string index = get_index(sym_water_s_3_5[i]->a_chnid,
                                                    sym_water_s_3_5[i]->a_resnum);
                      std::map<std::string, int>::iterator pos = _AllWaters.find(index);
                      if (pos != _AllWaters.end()) {
                           all_waters[pos->second]->sym = sym_water_s_3_5[i]->sym - 1;
                           all_waters[pos->second]->transl[0] =
                                            sym_water_s_3_5[i]->transl[0];
                           all_waters[pos->second]->transl[1] =
                                            sym_water_s_3_5[i]->transl[1];
                           all_waters[pos->second]->transl[2] =
                                            sym_water_s_3_5[i]->transl[2];
                      }
                 }
            }
       }

       if (asym_water_s_3_5) {
            for (int i = 0; i < n_asym_water_s_3_5; i++) {
                 if (asym_water_s_3_5[i]) delete asym_water_s_3_5[i];
            }
            delete [] asym_water_s_3_5;
       }

       if (sym_water_s_3_5) {
            for (int i = 0; i < n_sym_water_s_3_5; i++) {
                 if (sym_water_s_3_5[i]) delete sym_water_s_3_5[i];
            }
            delete [] sym_water_s_3_5;
       }
}

static void get_outlier_and_moving_waters(std::map<std::string, int> _AllWaters,
                 CONTACT**& all_waters, int& _n_waters, int& _alloc_waters,
                 CONTACT**& _waters, int& _n_moving_waters, int& _alloc_moving_waters,
                 CONTACT**& _moving_waters)
{
       if (_AllWaters.empty()) return;
            
       std::map<std::string, std::multimap<int, int> > sort_index;
       sort_index.clear();
       std::map<std::string, std::multimap<int, int> >::iterator spos;
       for (std::map<std::string, int>::iterator pos = _AllWaters.begin();
                                                 pos != _AllWaters.end(); ++pos) {
            int ipos = pos->second;
            spos = sort_index.find(all_waters[ipos]->b_chnid);
            if (spos == sort_index.end()) {
                 std::multimap<int, int> tmp_map;
                 tmp_map.clear();
                 tmp_map.insert(std::make_pair(all_waters[ipos]->b_resnum, ipos));
                 sort_index.insert(std::make_pair(all_waters[ipos]->b_chnid, tmp_map));
            } else spos->second.insert(std::make_pair(all_waters[ipos]->b_resnum, ipos));
       }
       for (spos = sort_index.begin(); spos != sort_index.end(); ++spos) {
          for (std::multimap<int, int>::iterator mpos = spos->second.begin();
                                               mpos != spos->second.end(); ++mpos) {
            if (_n_waters >= _alloc_waters) {
                 _alloc_waters += 10;
                 CONTACT **tmp = _waters;
                 _waters = new CONTACT*[_alloc_waters];
                 for (int j = 0; j < _n_waters; j++)
                      _waters[j] = tmp[j];
                 for (int j = _n_waters; j < _alloc_waters; j++)
                      _waters[j] = NULL;
                 delete [] tmp;
            }
            _waters[_n_waters] = all_waters[mpos->second];
            if (all_waters[mpos->second]->sym || all_waters[mpos->second]->transl[0] ||
                all_waters[mpos->second]->transl[1] || all_waters[mpos->second]->transl[2]) {
                 if (_n_moving_waters >= _alloc_moving_waters) {
                      _alloc_moving_waters += 10;
                      CONTACT **tmp = _moving_waters;
                      _moving_waters = new CONTACT*[_alloc_moving_waters];
                      for (int j = 0; j < _n_moving_waters; j++)
                           _moving_waters[j] = tmp[j];
                      for (int j = _n_moving_waters; j < _alloc_moving_waters; j++)
                           _moving_waters[j] = NULL;
                      delete [] tmp;
                 }
                 _moving_waters[_n_moving_waters] = new CONTACT;
                 strcpy(_moving_waters[_n_moving_waters]->a_atomtyp,
                        _waters[_n_waters]->a_atomtyp);
                 strcpy(_moving_waters[_n_moving_waters]->a_atom_type,
                        _waters[_n_waters]->a_atom_type);
                 strcpy(_moving_waters[_n_moving_waters]->a_chnid,
                        _waters[_n_waters]->a_chnid);
                 strcpy(_moving_waters[_n_moving_waters]->a_resnam,
                        _waters[_n_waters]->a_resnam);
                 _moving_waters[_n_moving_waters]->a_resnum =
                        _waters[_n_waters]->a_resnum;
                 _moving_waters[_n_moving_waters]->a_atomnum =
                        _waters[_n_waters]->a_atomnum;
                 _moving_waters[_n_moving_waters]->type =
                        _waters[_n_waters]->type;
                 _moving_waters[_n_moving_waters]->sym =
                        _waters[_n_waters]->sym;
                 _moving_waters[_n_moving_waters]->transl[0] =
                        _waters[_n_waters]->transl[0];
                 _moving_waters[_n_moving_waters]->transl[1] =
                        _waters[_n_waters]->transl[1];
                 _moving_waters[_n_moving_waters]->transl[2] =
                        _waters[_n_waters]->transl[2];
                 _n_moving_waters++;
            }
            all_waters[mpos->second] = NULL;
            _n_waters++;
          }
       }
}

static void get_all_shell_atoms(std::map<std::string, int>& _AllWaters,
                                std::set<std::string>& _ShellWaters,
                                const int n_contact, CONTACT **contact)
{
       if (_AllWaters.empty()) return; 
       int i, changed = 0;
       for (i = 0; i < n_contact; i++) {
            std::string index1 = get_index(contact[i]->a_chnid, contact[i]->a_resnum);
            std::string index2 = get_index(contact[i]->b_chnid, contact[i]->b_resnum);
            std::map<std::string, int>::iterator mpos1 = _AllWaters.find(index1);
            std::set<std::string>::iterator spos1 = _ShellWaters.find(index2);
            std::map<std::string, int>::iterator mpos2 = _AllWaters.find(index2);
            std::set<std::string>::iterator spos2 = _ShellWaters.find(index1);
            if (mpos1 != _AllWaters.end() && spos1 != _ShellWaters.end()) {
                 _AllWaters.erase(index1);
                 _ShellWaters.insert(index1);
                 changed = 1;
            } else if (mpos2 != _AllWaters.end() && spos2 != _ShellWaters.end()) {
                 _AllWaters.erase(index2);
                 _ShellWaters.insert(index2);
                 changed = 1;
            }
       }
       if (!changed) return;
       get_all_shell_atoms(_AllWaters, _ShellWaters, n_contact, contact);
}

static std::string get_index(const std::string& chnid, const int num)
{
       std::string index = chnid + "_";
       char buffer[10];
       sprintf(buffer, "%d", num);
       index += buffer;
       return index;
}

static void get_connected_atoms(const char *resname, ReVarCifArray<CifString>& connected_atoms)
{
       connected_atoms.Clear();

       CifString cs;
       if (!strcmp(resname, "ACE")) {
            cs.Copy("C");
            connected_atoms.Add(cs);
            return;
       } else if (!strcmp(resname, "NH2")) {
            cs.Copy("N");
            connected_atoms.Add(cs);
            return;
       }

       ConnectFormat *drug = find_drug(resname);
       if (!drug) return;

       for (int i = 0; i < drug->natoms(); ++i) {
            if (strcasecmp(drug->atoms(i)->leaving_atom_flag(), "Y")) continue;
            for (int j = 0; j < drug->atoms(i)->nconect(); ++j) {
                 _SIndex *index = drug->find_atom(drug->atoms(i)->connectto(j));
                 if (!index) continue;
                 if (!strcasecmp(drug->atoms(index->pos())->leaving_atom_flag(), "Y"))
                      continue;
                 cs.Copy(drug->atoms(i)->connectto(j));
                 connected_atoms.Add(cs);
            }
       }
}

void print_contact(const int num, CONTACT **contact)
{
       int i;
       if (!num || !contact) return;
       for (i = 0; i < num; i++) {
            printf("%6s %5s %s %5d %3s %5d %s --- %6s %5s %s %5d %3s %5d %s --- %d_%d%d%d %f\n",
                  contact[i]->a_type,
                  contact[i]->a_atomtyp,
                  contact[i]->a_atom_type,
                  contact[i]->a_atomnum,
                  contact[i]->a_resnam,
                  contact[i]->a_resnum,
                  contact[i]->a_chnid,
                  contact[i]->b_type,
                  contact[i]->b_atomtyp,
                  contact[i]->b_atom_type,
                  contact[i]->b_atomnum,
                  contact[i]->b_resnam,
                  contact[i]->b_resnum,
                  contact[i]->b_chnid,
                  contact[i]->sym,
                  contact[i]->transl[0],contact[i]->transl[1],contact[i]->transl[2],
                  contact[i]->dist);
       }
       printf("\n");
}
