/*
FILE:     ndb_to_pdb.C
*/
/*
VERSION:  8.125
*/
/*
DATE:     3/1/2013
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2013 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <stdio.h>
#include <string.h>
#include <string>
#include <math.h>
#include <map>
#include <set>
#include <vector>

#include "_maxit.h"
#include "util.h"
#include "erf_category.h"
#include "erf_util.h"
#include "ndb_misclib.h"

typedef struct {
     int oper_id;
     int num_chain;
     char chain[50][3];
} NCS_COR;

typedef struct {
        char biol_id[4];
} _BIOL_ID;

typedef struct {
       std::string id;
       std::string prd_id;
       std::string name;
       std::string chainids;
       std::vector<std::string> components;
       std::string description;
} _Remark_400_Group;

#define NUM_TML 6

static const char *tml_token[NUM_TML] = { "RFTPOS", "RFMPOS", "RFLPOS", "RFTTHR", "RFMTHR", "RFLTHR" };
static const char *tml_type[NUM_TML] = { "tight positional", "medium positional", "loose positional",
              "tight thermal", "medium thermal", "loose thermal"};

static void print_null_remark_value(One_Card_Info_Format *newCard,
             const int fieldLength, const int fieldJustification);
// static void clean_descriptor(char *string, char *ec);
static void delete_space_between_string(char *string);
static int check_align(const int Remark_No, REMARK_FIELD *FieldList);
#ifdef SUN_OS
#ifdef __cplusplus
extern "C" {
#endif
#endif
static int biol_id_compare(const void *biol1, const void *biol2);
#ifdef SUN_OS
#ifdef __cplusplus
}
#endif
#endif
static void get_remark_array_from_block(ReVarCifArray<CifString>& remark_array,
               const char *block_text, const int width);

static void get_wordlist_from_string_separated_by_delimits_with_empty_line(
              ReVarCifArray<CifString> &wordlist, char *string_buff, const char *delimit);
static bool has_value(ISTable *Table, const int num, _MOLECULE **mols);
static bool is_same_diffrn(char* single_diffrns, char* combine_diffrns);

void _MAXIT::ndb_to_pdb_general(const int remark)
{
       int i, j, k, len, Auth_Token, token_id[6], sym1[4], sym2[4];
       int num_tls = 0, num_cns = 0, num_twin = 0;
       char buff1[1000], buff2[1000], *p = NULL;
       One_Card_Info_Format *pCard = NULL, *qCard = NULL, *rCard = NULL;

       int is_refmac5 = 0;
       int is_phenix = 0;
       int is_buster = 0;
       pCard = _CardInfo[REFMET_TOKEN-1].Head();
       if (pCard) {
            strcpy(buff1, pCard->Field(3));
            ndb_set_string_to_upper(buff1);
            if (strstr(buff1, "REFMAC 5")) is_refmac5 = 1;
            if (strstr(buff1, "PHENIX")) is_phenix = 1;
            if (strstr(buff1, "BUSTER 2")) is_buster = 1;
       }

       pCard = _CardInfo[REMARK_TOKEN-1].Head();
       while (pCard) pCard = _CardInfo[REMARK_TOKEN-1].delete_a_card(pCard);

       token_id[0] = find_card_id("TWIN");
       num_twin = 0;
       pCard = _CardInfo[token_id[0]].Head();
       while (pCard) {
            pCard = pCard->Next();
            num_twin++;
       }
 
       token_id[0] = find_card_id("TLSGRO");
       token_id[1] = find_card_id("TLSRNG");
       num_tls = 0;
       pCard = _CardInfo[token_id[0]].Head();
       while (pCard) {
            i = 0;
            qCard = _CardInfo[token_id[1]].Head();
            while (qCard) {
                 if (!strcmp(qCard->Field(1), pCard->Field(1))) i++;
                 qCard = qCard->Next();
            }
            pCard->AddElementToField(i, 2);
            num_tls++;
            pCard = pCard->Next();
       }

       if (is_refmac5) {
            ReVarCifArray<ReVarCifArray<CifString> > ncs_data;
            ncs_data.Clear();
            pCard = _CardInfo[NCSGRO_TOKEN-1].Head();
            while (pCard) {
                 bool found = false;
                 for (unsigned int t = 0; t < ncs_data.Length(); t++) {
                      if (ncs_data[t][0] == pCard->Field(4)) {
                           found = true;
                           if (!strstr(ncs_data[t][1].Text(), pCard->Field(3))) {
                                ncs_data[t][1] += " ";
                                ncs_data[t][1] += pCard->Field(3);
                           }
                      }
                 }
                 if (!found) {
                      ReVarCifArray<CifString> data;
                      data.Clear();
                      CifString cs;
                      cs.Copy(pCard->Field(4));
                      data.Add(cs);
                      cs.Copy(pCard->Field(3));
                      data.Add(cs);
                      ncs_data.Add(data);
                 }
                 pCard = pCard->Next();
            }

            if (ncs_data.Length() > 0) {
                 pCard = _CardInfo[NCSGRO_TOKEN-1].Head();
                 while (pCard) pCard = _CardInfo[NCSGRO_TOKEN-1].delete_a_card(pCard);
                 for (unsigned int t = 0; t < ncs_data.Length(); t++) {
                      pCard = _CardInfo[NCSGRO_TOKEN-1].add_a_new_card(NdbToken[NCSGRO_TOKEN-1], 0);
                      pCard->AddElementToField(ncs_data[t][1].Text(), 3);
                      pCard->AddElementToField(ncs_data[t][0].Text(), 4);
                 }
            }

            num_cns = 0;
            int ccpncs_id = find_card_id("CCPNCS");
            if (ccpncs_id >= 0) {
                 pCard = _CardInfo[NCSGRO_TOKEN-1].Head();
                 while (pCard) {
                      i = 0;
                      qCard = _CardInfo[ccpncs_id].Head();
                      while (qCard) {
                           if (!strcmp(qCard->Field(1), pCard->Field(4))) {
                                if (atoi(qCard->Field(6)) > i) i = atoi(qCard->Field(6));
                           }
                           qCard = qCard->Next();
                      }
                      pCard->AddElementToField(i, 2);
                      num_cns++;
                      pCard = pCard->Next();
                 }
            }

            Auth_Token = find_card_id("TMLPTL");
            int card_id = find_card_id("NCSLOC");
            if (Auth_Token >= 0 && _CardInfo[Auth_Token].Head()) {
                 if (card_id >= 0) {
                      pCard = _CardInfo[card_id].Head();
                      while (pCard) pCard = _CardInfo[card_id].delete_a_card(pCard);
                 }
                 for (i = 0; i < NUM_TML; i++) {
                      token_id[i] = find_card_id(tml_token[i]);
                      if (token_id[i] < 0) continue;
                      pCard = _CardInfo[token_id[i]].Head();
                      while (pCard) pCard = _CardInfo[token_id[i]].delete_a_card(pCard);
                 }
                 pCard = _CardInfo[Auth_Token].Head();
                 while (pCard) {
                      pCard->AddElementToField(pCard->Field(7), 1);
                      pCard->CleanStringField(6);
                      if (!strcasecmp(pCard->Field(6), "interatomic distance")) {
                           qCard = _CardInfo[card_id].Head();
                           while (qCard) {
                                if (!strcmp(qCard->Field(1), pCard->Field(1)) &&
                                    !strcmp(qCard->Field(5), "")) {
                                     qCard->AddElementToField(pCard->Field(2), 5);
                                     break;
                                }
                                qCard = qCard->Next();
                           }
                           if (!qCard) {
                                qCard = _CardInfo[card_id].add_a_new_card(NdbToken[card_id], 0);
                                qCard->AddElementToField(pCard->Field(1), 1);
                                qCard->AddElementToField(pCard->Field(2), 2);
                                qCard->AddElementToField(pCard->Field(3), 8);
                                qCard->AddElementToField(pCard->Field(4), 9);
                                qCard->AddElementToField(pCard->Field(5), 10);
                           }
                      } else {
                           for (i = 0; i < NUM_TML; i++) {
                                if (!strcasecmp(pCard->Field(6), tml_type[i])) break;
                           }
                           if (i < NUM_TML) {
                                qCard = _CardInfo[token_id[i]].add_a_new_card(NdbToken[token_id[i]], 0);
                                for (j = 1; j < qCard->NumField(); j++) {
                                     if (j == 2 && !strcmp(pCard->Field(j), ""))
                                          qCard->AddElementToField(" ", j);
                                     else qCard->AddElementToField(pCard->Field(j), j);
                                }
                           }
                      }
                      pCard = pCard->Next();
                 }
            }

            if (card_id >= 0) {
                 pCard = _CardInfo[card_id].Head();
                 if (pCard) num_cns = 0;
                 while (pCard) {
                      num_cns++;
                      pCard = pCard->Next();
                 }
            }

            if (ccpncs_id >= 0 && card_id >= 0) {
                 pCard = _CardInfo[ccpncs_id].Head();
                 while (pCard) {
                      qCard = _CardInfo[card_id].Head();
                      while (qCard) {
                           if (!strcmp(qCard->Field(1), pCard->Field(1))) {
                                if (!strcmp(qCard->Field(2), pCard->Field(2))) {
                                     qCard->AddElementToField(pCard->Field(3), 3);
                                     qCard->AddElementToField(pCard->Field(5), 4);
                                } else if (!strcmp(qCard->Field(5), pCard->Field(2))) {
                                     qCard->AddElementToField(pCard->Field(3), 6);
                                     qCard->AddElementToField(pCard->Field(5), 7);
                                } 
                                break;
                           }
                           qCard = qCard->Next();
                      }
                      pCard = pCard->Next();
                 }
            }

            Auth_Token = find_card_id("TLSRNG");
            if (Auth_Token >= 0 && _CardInfo[Auth_Token].Head()) {
                 pCard = _CardInfo[Auth_Token].Head();
                 while (pCard) {
                      if (!strcmp(pCard->Field(2), "")) pCard->AddElementToField(" ", 2);
                      if (!strcmp(pCard->Field(4), "")) pCard->AddElementToField(" ", 4);
                      pCard = pCard->Next();
                 }
            }

            Auth_Token = find_card_id("CCPNCS");
            if (Auth_Token >= 0 && _CardInfo[Auth_Token].Head()) {
                 pCard = _CardInfo[Auth_Token].Head();
                 while (pCard) {
                      if (!strcmp(pCard->Field(2), "")) pCard->AddElementToField(" ", 2);
                      if (!strcmp(pCard->Field(4), "")) pCard->AddElementToField(" ", 4);
                      pCard = pCard->Next();
                 }
            }
       }

       if (is_phenix || is_buster) {
            Auth_Token = find_card_id("TLSGRO");
            token_id[0] = find_card_id("TLSRNG");
            pCard = _CardInfo[token_id[0]].Head();
            while (pCard) {
                 if (strcmp(pCard->Field(2), "") &&
                     strcmp(pCard->Field(3), "") &&
                     strcmp(pCard->Field(4), "") &&
                     strcmp(pCard->Field(5), "") &&
                    !strcmp(pCard->Field(11), "") && is_buster) {
                      sprintf(buff1, " %2s %4s   %2s %4s", pCard->Field(2),
                             pCard->Field(3), pCard->Field(4), pCard->Field(5));
                      pCard->AddElementToField(buff1, 11);
                 }
                 if (strcmp(pCard->Field(11), "")) {
                      qCard = _CardInfo[Auth_Token].Head();
                      while (qCard) {
                           if (!strcmp(pCard->Field(1), qCard->Field(1))) {
                                qCard->AddElementToField(pCard->Field(11), 7);
                                break;
                           }
                           qCard = qCard->Next();
                      }
                      if (!qCard) {
                           qCard = _CardInfo[Auth_Token].
                                     add_a_new_card(NdbToken[Auth_Token], 0);
                           qCard->AddElementToField(pCard->Field(1), 1);
                           qCard->AddElementToField(pCard->Field(11), 7);
                      }
                 }
                 pCard = pCard->Next();
            }

            Auth_Token = find_card_id("CCPNCS");
            pCard = _CardInfo[Auth_Token].Head();
            while (pCard) {
                 if (strcmp(pCard->Field(15), "")) {
                      qCard = _CardInfo[NCSGRO_TOKEN-1].Head();
                      while (qCard) {
                           if (!strcmp(qCard->Field(4), pCard->Field(1)) &&
                               !strcmp(qCard->Field(1), pCard->Field(14))) {
                                qCard->AddElementToField(pCard->Field(15), 3);
                                break;
                           }
                           qCard = qCard->Next();
                      }
                      if (!qCard) {
                           qCard = _CardInfo[NCSGRO_TOKEN-1].
                                     add_a_new_card(NdbToken[NCSGRO_TOKEN-1], 0);
                           qCard->AddElementToField(pCard->Field(14), 1);
                           qCard->AddElementToField(pCard->Field(15), 3);
                           qCard->AddElementToField(pCard->Field(1), 4);
                      }
                 }
                 pCard = pCard->Next();
            }

            memset(buff1, 0 , 1000);
            Auth_Token = find_card_id("TMLPTL");
            token_id[0] = find_card_id("PHENCS");
            pCard = _CardInfo[NCSGRO_TOKEN-1].Head();
            ReVarCifArray<CifString> Ens_IDs;
            Ens_IDs.Clear();
            CifString cs;
            while (pCard) {
                 bool found = false;
                 for (unsigned t = 0; t < Ens_IDs.Length(); t++) {
                      if (Ens_IDs[t] == pCard->Field(4)) {
                           found = true;
                           break;
                      }
                 }
                 if (!found) {
                      cs.Copy(pCard->Field(4));
                      Ens_IDs.Add(cs);
                 }
                 if (!strcmp(pCard->Field(1), "1")) {
                      p = strstr(pCard->Field(3), "REFERENCE SELECTION:");
                      if (p) strcpy(buff1, p + 20);
                      else strcpy(buff1, pCard->Field(3));
                      ndb_clean_string(buff1);
                 } else {
                      p = strstr(pCard->Field(3), "SELECTION :");
                      if (p) strcpy(buff2, p + 11);
                      else strcpy(buff2, pCard->Field(3));
                      ndb_clean_string(buff2);

                      rCard = _CardInfo[token_id[0]].add_a_new_card(NdbToken[token_id[0]], 0);
                      rCard->AddElementToField(pCard->Field(4), 1);
                      rCard->AddElementToField(buff1, 3);
                      rCard->AddElementToField(buff2, 4);
                      qCard = _CardInfo[Auth_Token].Head();
                      while (qCard) {
                           if (!strcmp(pCard->Field(1), qCard->Field(1)) &&
                               !strcmp(pCard->Field(4), qCard->Field(7))) {
                                rCard->AddElementToField(qCard->Field(3), 5);
                                rCard->AddElementToField(qCard->Field(4), 6); 
                                break;
                           }
                           qCard = qCard->Next();
                      }
                 }
                 pCard = pCard->Next();
            }
            num_cns = Ens_IDs.Length();
            pCard = _CardInfo[NCSGRO_TOKEN-1].Head();
            while (pCard) pCard = _CardInfo[NCSGRO_TOKEN-1].delete_a_card(pCard);
            for (unsigned int t = 0; t < Ens_IDs.Length(); t++) {
                 pCard = _CardInfo[NCSGRO_TOKEN-1].add_a_new_card(NdbToken[NCSGRO_TOKEN-1], 0);
                 pCard->AddElementToField(Ens_IDs[t].Text(), 4);
            }
            pCard = _CardInfo[token_id[0]].Head();
            i = 1;
            strcpy(buff1, "xxx");
            while (pCard) {
                 if (strcmp(pCard->Field(1), buff1)) {
                      i = 1;
                      strcpy(buff1, pCard->Field(1));
                 }
                 pCard->AddElementToField(i, 2);
                 pCard = pCard->Next();
                 i++;
            }
       }

       if (num_cns || num_tls || num_twin) {
            token_id[0] = find_card_id("NCSTLS");
            pCard = _CardInfo[token_id[0]].Head();
            if (!pCard) pCard = _CardInfo[token_id[0]].
                           add_a_new_card(NdbToken[token_id[0]], 0);
            if (num_cns) pCard->AddElementToField(num_cns, 1);
            if (num_tls) pCard->AddElementToField(num_tls, 2);
            if (num_twin) pCard->AddElementToField(num_twin, 3);
       }

       Auth_Token = find_card_id("SITELS");
       if (Auth_Token >= 0 && _CardInfo[Auth_Token].Head()) {
            TblIndexObj site_index;
            CifString cifstring;
            site_index.Clear();
            pCard = _CardInfo[SITE_TOKEN-1].Head();
            while (pCard) {
                 cifstring.Copy(pCard->Field(2));
                 if (site_index.Seek(cifstring) < 0) site_index.Add(cifstring);
                 pCard = pCard->Next();
            }
            pCard = _CardInfo[Auth_Token].Head();
            while (pCard) {
                 cifstring.Copy(pCard->Field(2));
                 if (site_index.Seek(cifstring) < 0)
                      pCard = _CardInfo[Auth_Token].delete_a_card(pCard);
                 else pCard = pCard->Next();
            }
       }

       Auth_Token = find_card_id("SCTYPE");
       if (Auth_Token >= 0 && _CardInfo[Auth_Token].Head()) {
            pCard = _CardInfo[Auth_Token].Head();
            while (pCard) {
                 if (!strcasecmp(pCard->Field(2), "X-RAY"))
                      pCard->AddElementToField("X-RAY DIFFRACTION", 2);
                 else if (!strcasecmp(pCard->Field(2), "NEUTRON"))
                      pCard->AddElementToField("NEUTRON DIFFRACTION", 2);
                 else if (!strcasecmp(pCard->Field(2), "ELECTRON"))
                      pCard->AddElementToField("ELECTRON CRYSTALLOGRAPHY", 2);
                 pCard = pCard->Next();
            }
       }

       ndb_to_pdb_processing_authors();
       ndb_to_pdb_processing_shelx_and_tnt();
       ndb_to_pdb_processing_synchrotron();
       ndb_to_pdb_proc_header();
       replace_pdb_compnd_source_card();
       ndb_to_pdb_get_remarks(remark); 

       get_cispep_cards();

       token_id[0] = SSBOND_TOKEN-1; sym1[0] = 10; sym2[0] = 11;
       token_id[1] = LINK_TOKEN-1;   sym1[1] = 13; sym2[1] = 14;
       token_id[2] = HYDBND_TOKEN-1; sym1[2] = 18; sym2[2] = 19;
       token_id[3] = SLTBRG_TOKEN-1; sym1[3] = 13; sym2[3] = 14;
       for (i = 0; i < 4; i++) {
            pCard = _CardInfo[token_id[i]].Head();
            while (pCard) {
                 strcpy(buff1, pCard->Field(sym1[i]));
                 if (strcmp(buff1, "")) {
                      k = 0;
                      len = strlen(buff1);
                      for (j = 0; j < len; j++) {
                           if (buff1[j] == '_') continue;
                           buff2[k] = buff1[j];
                           k++;
                      }
                      buff2[k] = '\0';
                      pCard->AddElementToField(buff2, sym1[i]);
                 }
                 strcpy(buff1, pCard->Field(sym2[i]));
                 if (strcmp(buff1, "")) {
                      k = 0;
                      len = strlen(buff1);
                      for (j = 0; j < len; j++) {
                           if (buff1[j] == '_') continue;
                           buff2[k] = buff1[j];
                           k++;
                      }
                      buff2[k] = '\0';
                      pCard->AddElementToField(buff2, sym2[i]);
                 }
                 pCard = pCard->Next();
            }
       }
       Auth_Token = find_card_id("CAVEAT");
       pCard = _CardInfo[Auth_Token].Head();
       if (pCard && _CardInfo[PDBFIL_TOKEN-1].Head()) {
            while (pCard) {
                 pCard->AddElementToField(_CardInfo[PDBFIL_TOKEN-1].Head()->Field(1), 2);
                 pCard = pCard->Next();
            }
       }
       if (_title_suppression) {
            Auth_Token = find_card_id("TITLE");
            pCard = _CardInfo[Auth_Token].Head();
            while (pCard) pCard = _CardInfo[Auth_Token].delete_a_card(pCard);
       }

       pCard = _CardInfo[DBREF_TOKEN-1].Head();
       while (pCard) {
            if (strlen(pCard->Field(1)) > 4) pCard->AddElementToField("", 1);
            if ((int) strlen(pCard->Field(8)) >
                 NdbToken[DBREF_TOKEN-1].FieldList[8].FieldWidth ||
                (int) strlen(pCard->Field(9)) >
                 NdbToken[DBREF_TOKEN-1].FieldList[9].FieldWidth ||
                (int) strlen(pCard->Field(10)) >
                 NdbToken[DBREF_TOKEN-1].FieldList[10].FieldWidth ||
                (int) strlen(pCard->Field(12)) >
                 NdbToken[DBREF_TOKEN-1].FieldList[12].FieldWidth) {
                 qCard = _CardInfo[DBREF_TOKEN-1].insert_a_card(pCard,
                                NdbToken[DBREF_TOKEN-1], 0);
                 qCard->AddElementToField("DBREF2", 0);
                 qCard->AddElementToField(pCard->Field(1), 1);
                 qCard->AddElementToField(pCard->Field(2), 2);
                 qCard->AddElementToField(pCard->Field(8), 3);
                 qCard->AddElementToField(pCard->Field(10), 4);
                 qCard->AddElementToField(pCard->Field(11), 5);
                 qCard->AddElementToField(pCard->Field(12), 6);
                 qCard->AddElementToField(pCard->Field(13), 7);

                 pCard->AddElementToField("DBREF1", 0);
                 pCard->AddElementToField(pCard->Field(9), 8);
                 pCard = qCard->Next();
            } else pCard = pCard->Next();
       }

       pCard = _CardInfo[EXPDTA_TOKEN-1].Head();
       if (pCard) {
            qCard = pCard->Next();
            while (qCard) {
                 pCard->AttachToField("; ", 2);
                 pCard->AttachToField(qCard->Field(2), 2);
                 qCard = _CardInfo[EXPDTA_TOKEN-1].delete_a_card(qCard);
            }
       }
/*
       token_id[0] = find_card_id("NMRSMB");
       token_id[1] = find_card_id("NUMMDL");
       pCard = _CardInfo[token_id[0]].Head();
       if (pCard && strcmp(pCard->Field(2), "") && strcmp(pCard->Field(2), "1")) {
            qCard = _CardInfo[token_id[1]].add_a_new_card(NdbToken[token_id[1]], 0);
            qCard->AttachToField(pCard->Field(2), 2);
       }
*/
       if (_num_mol > 1) {
            token_id[1] = find_card_id("NUMMDL");
            pCard = _CardInfo[token_id[1]].add_a_new_card(NdbToken[token_id[1]], 0);
            pCard->AddElementToField(_num_mol, 2);
       }

       CifString Ca_Lists, P_Lists;
       Ca_Lists.Copy(""); P_Lists.Copy("");
       if (_num_mol) {
            CifString Ca_list, P_list;
            Ca_list.Copy(""); P_list.Copy("");
            _CHAIN *chain = _molecules[0]->GetFirstChain();
            while (chain) {
                 if (chain->chain_type() != ATOMP_TOKEN &&
                     chain->chain_type() != ATOMN_TOKEN) {
                      chain = _molecules[0]->GetNextChain();
                      continue;
                 }
                 if (chain->ca_or_p_atom_only() == CA_ATOM_ONLY_CHAIN) {
                      if (Ca_list != "") Ca_list += ", ";
                      Ca_list += chain->PDB_ChainID();
                 } else if (chain->ca_or_p_atom_only() == P_ATOM_ONLY_CHAIN) {
                      if (P_list != "") P_list += ", ";
                      P_list += chain->PDB_ChainID();
                 }
                 chain = _molecules[0]->GetNextChain();
            }
            if (Ca_list != "") {
                 Ca_Lists = "CA ATOMS ONLY, CHAIN ";
                 Ca_Lists += Ca_list;
            }
            if (P_list != "") {
                 P_Lists = "P ATOMS ONLY, CHAIN ";
                 P_Lists += P_list;
            }
       }
       if (Ca_Lists != "" || P_Lists != "") {
            token_id[0] = find_card_id("MDLTYP");
            pCard = _CardInfo[token_id[0]].Head();
            if (!pCard) pCard = _CardInfo[token_id[0]].add_a_new_card(NdbToken[token_id[0]], 0);
            if (Ca_Lists != "" && !strstr(pCard->Field(2), Ca_Lists.Text())) {
                 if (strcmp(pCard->Field(2), "")) pCard->AttachToField("; ", 2);
                 pCard->AttachToField(Ca_Lists.Text(), 2);
            }
            if (P_Lists != "" && !strstr(pCard->Field(2), P_Lists.Text())) {
                 if (strcmp(pCard->Field(2), "")) pCard->AttachToField("; ", 2);
                 pCard->AttachToField(P_Lists.Text(), 2);
            }
       }
}

void _MAXIT::ndb_to_pdb_processing_authors()
{
       int Author_Token = find_card_id("AUTHOR");
       int Auth_Token = find_jrnl_id("AUTH");

       char buffer[1000];
       One_Card_Info_Format *pCard = _JrnlCardInfo[Auth_Token].Head();
       while (pCard) {
            strcpy(buffer, pCard->Field(4));
            delete_space_between_string(buffer);
            pCard->AddElementToField(buffer, 4);
            pCard = pCard->Next();
       }

       pCard = _CardInfo[Author_Token].Head();
       while (pCard) {
            strcpy(buffer, pCard->Field(2));
            delete_space_between_string(buffer);
            pCard->AddElementToField(buffer, 2);
            pCard = pCard->Next();
       }

       if (_JrnlCardInfo[Auth_Token].Head() && !_CardInfo[Author_Token].Head() &&
           strcmp(_JrnlCardInfo[Auth_Token].Head()->Field(4), "")) {
            pCard = _CardInfo[Author_Token].add_a_new_card(NdbToken[Author_Token], 0);
            pCard->AddElementToField(_JrnlCardInfo[Auth_Token].Head()->Field(4), 2);
       }
}

void _MAXIT::ndb_to_pdb_processing_shelx_and_tnt()
{
       One_Card_Info_Format *pCard = NULL, *qCard = NULL, *rCard = NULL, *tCard = NULL;
       int RNOCUT_TOKEN = find_card_id("RNOCUT");

       pCard = _CardInfo[REFMET_TOKEN-1].Head();
       if (pCard) {
            if (!strncasecmp(pCard->Field(3), "SHELX", 5)) {
                 qCard = _CardInfo[RFACTR_TOKEN-1].Head();
                 if (qCard) {
                      rCard = _CardInfo[RNOCUT_TOKEN].Head();
                      if (rCard) {
                           if (!strcmp(rCard->Field(1), ""))
                                rCard->AddElementToField(qCard->Field(8), 1);
                           if (!strcmp(rCard->Field(2), ""))
                                rCard->AddElementToField(qCard->Field(7), 2);
                      } else {
                           rCard = _CardInfo[RNOCUT_TOKEN].
                                 add_a_new_card(NdbToken[RNOCUT_TOKEN], 0);
                           rCard->AddElementToField(qCard->Field(8), 1);
                           rCard->AddElementToField(qCard->Field(7), 2);
                      }
                 }

                 tCard = _CardInfo[FREERF_TOKEN-1].Head();
                 if (qCard && tCard) {
                      rCard = _CardInfo[RNOCUT_TOKEN].Head();
                      if (rCard) {
                           if (!strcmp(rCard->Field(3), ""))
                                rCard->AddElementToField(tCard->Field(1), 3);
                           if (!strcmp(rCard->Field(4), ""))
                                rCard->AddElementToField(tCard->Field(3), 4);
                           if (!strcmp(rCard->Field(5), ""))
                                rCard->AddElementToField(tCard->Field(4), 5);
                      } else {
                           rCard = _CardInfo[RNOCUT_TOKEN].
                                 add_a_new_card(NdbToken[RNOCUT_TOKEN], 0);
                           rCard->AddElementToField(tCard->Field(1), 3);
                           rCard->AddElementToField(tCard->Field(3), 4);
                           rCard->AddElementToField(tCard->Field(4), 5);
                      }
                 }
            } else if (!strncasecmp(pCard->Field(3), "TNT", 3) ||
                       !strncasecmp(pCard->Field(3), "NUCLSQ", 6) ||
                       !strncasecmp(pCard->Field(3), "PROLSQ", 6)) {
                 qCard = _CardInfo[RFACTR_TOKEN-1].Head();
                 if (qCard) {
                      if (!strcmp(qCard->Field(4), "") ||
                          fabs(atof(qCard->Field(4))) < 0.1) {
                           rCard = _CardInfo[RNOCUT_TOKEN].Head();
                           if (rCard) {
                                if (!strcmp(rCard->Field(1), ""))
                                     rCard->AddElementToField(qCard->Field(8), 1);
                           } else {
                                rCard = _CardInfo[RNOCUT_TOKEN].
                                      add_a_new_card(NdbToken[RNOCUT_TOKEN], 0);
                                rCard->AddElementToField(qCard->Field(8), 1);
                           }
                      }
                 }
            }
       }
}

void _MAXIT::ndb_to_pdb_processing_synchrotron()
{
       char  site[100], beamline[1000];
       One_Card_Info_Format *pCard = NULL, *qCard = NULL;

       pCard = _CardInfo[DTMETH_TOKEN-1].Head();
       while (pCard) {
            if (!strcasecmp(pCard->Field(4), "SYNCHROTRON")) {
                 strcpy(pCard->Field(1), "Y");
                 qCard = _CardInfo[RADIAT_TOKEN-1].Head();
                 while (qCard) {
                      if (!strcmp(pCard->Field(NdbToken[DTMETH_TOKEN-1].SeqField-1),
                                  qCard->Field(NdbToken[RADIAT_TOKEN-1].SeqField-1)))
                           break;
                      qCard = qCard->Next();
                 }

                 site[0] = '\0'; beamline[0] = '\0';
                 if (qCard) {
                      qCard->UpperCaseField(5);
                      separation_string_base_word(qCard->Field(5), site, 
                                     beamline, "BEAMLINE");
                      ndb_clean_string(site);
                      ndb_clean_string(beamline); 
                 }

                 if (strcmp(site, "") || strcmp(beamline, "")) {
                      qCard = _CardInfo[DTMEAS_TOKEN-1].Head();
                      while (qCard) {
                           if (!strcmp(pCard->Field(NdbToken[DTMETH_TOKEN-1].SeqField-1),
                                       qCard->Field(NdbToken[DTMEAS_TOKEN-1].SeqField-1)))
                                break;
                           qCard = qCard->Next();
                      }

                      if (qCard == NULL) {
                           qCard = _CardInfo[DTMEAS_TOKEN-1].
                                     add_a_new_card(NdbToken[DTMEAS_TOKEN-1], 0);
                           qCard->AddElementToField(
                              pCard->Field(NdbToken[DTMETH_TOKEN-1].SeqField-1),
                                           NdbToken[DTMEAS_TOKEN-1].SeqField-1);
                      }
                      qCard->AddElementToField(site, 2);
                      qCard->AddElementToField(beamline, 3);
                 }
            } else pCard->AddElementToField("N", 1);
            pCard = pCard->Next();
       }

       pCard = _CardInfo[DTMETH_TOKEN-1].Head();
       while (pCard) {
            if (!strcasecmp(pCard->Field(4), "SYNCHROTRON")) {
                 qCard = _CardInfo[DTMEAS_TOKEN-1].Head();
                 while (qCard) {
                      if (!strcmp(qCard->Field(6), pCard->Field(6))) break;
                      qCard = qCard->Next();
                 }
                 if (qCard)
                      pCard->AddElementToField(qCard->Field(2), 4);
                 else pCard->AddElementToField("", 4);

                 qCard = _CardInfo[RADIAT_TOKEN-1].Head();
                 while (qCard) {
                      if (!strcmp(qCard->Field(1), pCard->Field(6))) break;
                      qCard = qCard->Next();
                 } 
                 if (qCard) qCard->AddElementToField("", 5);
            }
            pCard = pCard->Next();
       }

       pCard = _CardInfo[DTWAVE_TOKEN-1].Head();
       if (_CardInfo[WAVLEN_TOKEN-1].Head() == NULL && pCard) {
            beamline[0] = '\0';
            while (pCard) {
                 strcat(beamline, pCard->Field(3));
                 if (pCard->Next()) strcat(beamline, ", ");
                 pCard = pCard->Next();
            }
            if (strcmp(beamline, "")) {
                 qCard = _CardInfo[WAVLEN_TOKEN-1].
                             add_a_new_card(NdbToken[WAVLEN_TOKEN-1], 0);
                 qCard->AddElementToField(beamline, 3);
            }
       }

       pCard = _CardInfo[WAVLEN_TOKEN-1].Head();
       while (pCard) {
            if (!strcmp(pCard->Field(2), "") && strcmp(pCard->Field(3), ""))
                 pCard->AddElementToField(pCard->Field(3), 2);
            pCard = pCard->Next();
       }

       std::set<int> diffrn_ids, ids;
       diffrn_ids.clear();
       for (int i = 0; i < NUM_200_ALIGN; i++) {
            int card_id = find_card_id(_remark_200_align[i].TokenName);
            pCard = _CardInfo[card_id].Head();
            while (pCard) {
                 if (strcmp(pCard->Field(_remark_200_align[i].Field_ID), "")) {
                      diffrn_ids.insert(atoi(pCard->Field(_remark_200_align[i].Field_ID)));
                      if (strcmp(_remark_200_align[i].TokenName, "DTTEMP"))
                          ids.insert(atoi(pCard->Field(_remark_200_align[i].Field_ID)));
                 }
                 pCard = pCard->Next();
            }
       }
       if (diffrn_ids.empty()) return;

       if (!ids.empty()) diffrn_ids = ids;

       std::string field_4, field_5;
       field_4.clear();
       field_5.clear();
       pCard = _CardInfo[DTMEAS_TOKEN-1].Head();
       if (pCard) {
            field_4 = pCard->Field(4);
            field_5 = pCard->Field(5);
       }

       std::map<int, std::string> diffrn_sctype;
       diffrn_sctype.clear();
       int sctype_id = find_card_id("SCTYPE");
       if (sctype_id >= 0) {
            pCard = _CardInfo[sctype_id].Head();
            while (pCard) {
                 if (strcmp(pCard->Field(1), "") && strcmp(pCard->Field(2), "")) {
                      diffrn_sctype.insert(std::make_pair(atoi(pCard->Field(1)),
                                                          pCard->Field(2)));
                 }
                 pCard = _CardInfo[sctype_id].delete_a_card(pCard);
            }
       }

       std::map<std::string, std::set<int> > sctype_diffrn_ids;
       sctype_diffrn_ids.clear();
       std::map<std::string, std::set<int> >::iterator mvpos;
       if (!diffrn_sctype.empty()) {
            for (std::set<int>::iterator spos = diffrn_ids.begin();
                                    spos != diffrn_ids.end(); ++spos) {
                 std::map<int, std::string>::iterator mpos = diffrn_sctype.find(*spos);
                 if (mpos == diffrn_sctype.end()) mpos = diffrn_sctype.begin();
                 mvpos = sctype_diffrn_ids.find(mpos->second);
                 if (mvpos != sctype_diffrn_ids.end()) {
                      mvpos->second.insert(*spos);
                 } else {
                      ids.clear();
                      ids.insert(*spos);
                      sctype_diffrn_ids.insert(std::make_pair(mpos->second, ids));
                 }
            }
       } else {
            strcpy(site, "X-RAY DIFFRACTION");
            pCard = _CardInfo[EXPDTA_TOKEN-1].Head();
            while (pCard) {
                 if (!strcmp(pCard->Field(2), "X-RAY DIFFRACTION") ||
                     !strcmp(pCard->Field(2), "FIBER DIFFRACTION") ||
                     !strcmp(pCard->Field(2), "NEUTRON DIFFRACTION") ||
                     !strcmp(pCard->Field(2), "ELECTRON CRYSTALLOGRAPHY")) {
                      strcpy(site, pCard->Field(2));
                      break;
                 }
                 pCard = pCard->Next();
            }
            sctype_diffrn_ids.insert(std::make_pair(site, diffrn_ids));
       }

       std::map<int, int> diffrn_crystal_ids;
       diffrn_crystal_ids.clear();

       pCard = _CardInfo[DTTEMP_TOKEN-1].Head();
       while (pCard) {
            if (strcmp(pCard->Field(4), "") && strcmp(pCard->Field(5), "")) {
                 diffrn_crystal_ids.insert(std::make_pair(atoi(pCard->Field(4)),
                                                     atoi(pCard->Field(5))));
            }
            pCard = pCard->Next();
       }

       std::map<int, std::string> crystal_id_ph;
       crystal_id_ph.clear();
       pCard = _CardInfo[PHVAL_TOKEN-1].Head();
       while (pCard) {
            if (strcmp(pCard->Field(1), "") && strcmp(pCard->Field(2), "")) {
                 crystal_id_ph.insert(std::make_pair(atoi(pCard->Field(2)),
                                                          pCard->Field(1)));
            }
            pCard = _CardInfo[PHVAL_TOKEN-1].delete_a_card(pCard);
       }

       std::string cs, cs1;

       for (mvpos = sctype_diffrn_ids.begin(); mvpos != sctype_diffrn_ids.end(); ++mvpos) {
            cs.clear();
            ids.clear();
            for (std::set<int>::iterator spos = mvpos->second.begin();
                                    spos != mvpos->second.end(); ++spos) {
                 if (!cs.empty()) cs += ",";
                 sprintf(site, "%d", *spos);
                 cs += site; // *spos;
                 std::map<int, int>::iterator mpos = diffrn_crystal_ids.find(*spos);
                 if (mpos == diffrn_crystal_ids.end()) continue;
                 ids.insert(mpos->second);
            }
            cs1.clear();
            for (std::set<int>::iterator spos = ids.begin(); spos != ids.end(); ++spos) {
                 std::map<int, std::string>::iterator mpos = crystal_id_ph.find(*spos);
                 if (mpos == crystal_id_ph.end()) continue;
                 if (!cs1.empty()) cs1 += "; ";
                 cs1 += mpos->second;
            }
            if (sctype_id >= 0) {
                 pCard = _CardInfo[sctype_id].add_a_new_card(NdbToken[sctype_id], 0);
                 pCard->AddElementToField(cs.c_str(), 1);
                 pCard->AddElementToField(mvpos->first.c_str(), 2);
            }
            pCard = _CardInfo[PHVAL_TOKEN-1].add_a_new_card(NdbToken[PHVAL_TOKEN-1], 0);
            pCard->AddElementToField(cs1.c_str(), 1);
            pCard->AddElementToField(cs.c_str(), 2);
       }

       std::map<int, std::vector<std::string> > records;
       records.clear();
       std::vector<std::string> data;
       for (int i = 0; i < NUM_200_ALIGN; i++) {
            int card_id = find_card_id(_remark_200_align[i].TokenName);
            records.clear();
            pCard = _CardInfo[card_id].Head();
            while (pCard) {
                 data.clear();
                 for (int j = 0; j < pCard->NumField(); ++j) {
                      data.push_back(pCard->Field(j));
                 }
                 records.insert(std::make_pair(atoi(pCard->Field(_remark_200_align[i].
                                             Field_ID)), data));
                 pCard = _CardInfo[card_id].delete_a_card(pCard);
            }
            
            for (mvpos = sctype_diffrn_ids.begin(); mvpos != sctype_diffrn_ids.end();
                                                    ++mvpos) {
                 cs.clear();
                 data.clear();
                 for (int j = 0; j < NdbToken[card_id].NumField; ++j) data.push_back("");
                 for (std::set<int>::iterator spos = mvpos->second.begin();
                                         spos != mvpos->second.end(); ++spos) {
                      if (!cs.empty()) cs += ",";
                      sprintf(site, "%d", *spos);
                      cs += site; // *spos;

                      std::map<int, std::vector<std::string> >::iterator
                                       mvpos1 = records.find(*spos);
                      for (unsigned int k = 0; k < data.size(); ++k) {
                           if (!data[k].empty()) data[k] += "; ";
                           if (mvpos1 == records.end())
                                data[k] += "NULL";
                           else if (!mvpos1->second[k].empty())
                                data[k] += mvpos1->second[k];
                           else data[k] += "NULL";
                      }
                 }

                 pCard = _CardInfo[card_id].add_a_new_card(NdbToken[card_id], 0);
                 pCard->AddElementToField(cs.c_str(), _remark_200_align[i].Field_ID);
                 for (int j = 0; j < _remark_200_align[i].NumField; j++) {
                      pCard->AddElementToField(data[_remark_200_align[i].Field_No[j]].
                                 c_str(), _remark_200_align[i].Field_No[j]);
                 }
            }
       }

       pCard = _CardInfo[DTMEAS_TOKEN-1].Head();
       while (pCard) {
            pCard->AddElementToField(field_4.c_str(), 4);
            pCard->AddElementToField(field_5.c_str(), 5);
            pCard = pCard->Next();
       }
}

void _MAXIT::ndb_to_pdb_proc_header()
{
       One_Card_Info_Format *pCard = NULL, *qCard = NULL;

       pCard = _CardInfo[HEADER_TOKEN-1].Head();
       if (pCard == NULL) return;
       qCard = _CardInfo[PDBFIL_TOKEN-1].Head();
       if (qCard && strcmp(qCard->Field(1), ""))
            pCard->AddElementToField(qCard->Field(1), 3);
       else {
            qCard = _CardInfo[NDBFIL_TOKEN-1].Head();
            if (qCard && strcmp(qCard->Field(1), ""))
                 pCard->AddElementToField(qCard->Field(1), 3);
            else {
                int BMRBID_TOKEN = find_card_id("BMRBID");
                qCard = _CardInfo[BMRBID_TOKEN].Head();
                if (qCard && strcmp(qCard->Field(1), ""))
                     pCard->AddElementToField(qCard->Field(1), 3);
            }
       }
/*
       if (!strcmp(pCard->Field(1), NDB_DNA_TEXT))
            pCard->AddElementToField("DEOXYRIBONUCLEIC ACID", 1);
       else if (!strcmp(pCard->Field(1), NDB_RNA_TEXT))
            pCard->AddElementToField("RIBONUCLEIC ACID", 1);
       else if (!strcmp(pCard->Field(1), NDB_T_RNA_TEXT))
            pCard->AddElementToField("AMINO-ACID TRANSPORT", 1);
*/
}

void _MAXIT::replace_pdb_compnd_source_card()
{
       int i, found, entity_to_serialno[100];
       int SerialNo = 0; 
       int compndSerialNo = 0;
       int sourceSerialNo = 0;
       int nsource_Nat = 0;
       int nsource_Gen = 0;
       int nsource_Syn = 0;
       int isSynthetic = GEN;
       char ec_number[800], buffer[800], buffer1[800];
       One_Card_Info_Format *pCard = NULL, *qCard = NULL, *rCard = NULL;
       One_Card_Info_Format *sCard = NULL, *compndCard = NULL, *sourceCard = NULL;

       for (i = 0; i < 100; i++) entity_to_serialno[i] = 0;

       pCard = _CardInfo[SOURCE_TOKEN-1].Head();
       while (pCard) pCard = _CardInfo[SOURCE_TOKEN-1].delete_a_card(pCard); 
       pCard = _CardInfo[COMPND_TOKEN-1].Head();
       while (pCard) pCard = _CardInfo[COMPND_TOKEN-1].delete_a_card(pCard); 

       pCard = _CardInfo[SEQIFO_TOKEN-1].Head();
       while (pCard) {
            qCard = _CardInfo[STBIOL_TOKEN-1].Head();
            if (qCard) {
                while (qCard) {
                     if (!strcmp(pCard->Field(2), qCard->Field(12))) {
                          if (strcmp(qCard->Field(15), ""))
                               pCard->AddElementToField(qCard->Field(15), 1);
                          else pCard->AddElementToField(qCard->Field(1), 1);
                          break;
                     }
                     qCard = qCard->Next();
                }
            } else {
                for (i = 0; i < _n_stbiol; i++) {
                     if (!_stbiol[i]) continue;
                     if (_stbiol[i]->NDB_Strand_ID == pCard->Field(2)) {
                          pCard->AddElementToField(_stbiol[i]->order, 1);
                          break;
                     }
                }
            }
            pCard = pCard->Next();
       }
       _CardInfo[SEQIFO_TOKEN-1].reorder_cards(1, INTEGER_COMPARE, ASCENT);

       while (PdbSource_Nat[nsource_Nat].pdbname &&
              strcmp(PdbSource_Nat[nsource_Nat].pdbname, "")) {
            if (PdbSource_Nat[nsource_Nat].NdbTokenId <= 0)
                 PdbSource_Nat[nsource_Nat].NdbTokenId =
                    find_card_id(PdbSource_Nat[nsource_Nat].NdbTokenName)+1;
            nsource_Nat++;
       }

       while (PdbSource_Gen[nsource_Gen].pdbname &&
              strcmp(PdbSource_Gen[nsource_Gen].pdbname, "")) {
            if (PdbSource_Gen[nsource_Gen].NdbTokenId <= 0)
                 PdbSource_Gen[nsource_Gen].NdbTokenId =
                    find_card_id(PdbSource_Gen[nsource_Gen].NdbTokenName)+1;
            nsource_Gen++;
       }

       while (PdbSource_Syn[nsource_Syn].pdbname &&
              strcmp(PdbSource_Syn[nsource_Syn].pdbname, "")) {
            if (PdbSource_Syn[nsource_Syn].NdbTokenId <= 0)
                 PdbSource_Syn[nsource_Syn].NdbTokenId =
                    find_card_id(PdbSource_Syn[nsource_Syn].NdbTokenName)+1;
            nsource_Syn++;
       }

       pCard = _CardInfo[SEQIFO_TOKEN-1].Head();
       if (!pCard) {
            TblIndexObj chnIndex;
            chnIndex.Clear();
            CifString cs;
            qCard = _CardInfo[ENTITY_TOKEN-1].Head();
            if (qCard && strcasecmp(qCard->Field(2), "POLYMER")) {
                 qCard->AddElementToField("POLYMER", 2);
            }
            while (qCard) {
                 if (strcasecmp(qCard->Field(2), "POLYMER")) break;
                 rCard = _CardInfo[STASYM_TOKEN-1].Head();
                 while (rCard) {
                      if (!strcmp(qCard->Field(1), rCard->Field(4))) {
                           sCard = _CardInfo[STBIOL_TOKEN-1].Head();
                           while (sCard) {
                                cs.Copy(sCard->Field(13));
                                if (!strcmp(rCard->Field(3), sCard->Field(3)) &&
                                    chnIndex.Seek(cs) < 0) {
                                     chnIndex.Add(cs);
                                     pCard = _CardInfo[SEQIFO_TOKEN-1].
                                       add_a_new_card(NdbToken[SEQIFO_TOKEN-1], 0);
                                     pCard->AddElementToField(sCard->Field(13), 3);
                                     pCard->AddElementToField(qCard->Field(1), 9);
                                }
                                sCard = sCard->Next();
                           }
                      }
                      rCard = rCard->Next();
                 }
                 qCard = qCard->Next();
            }
       }

       pCard = _CardInfo[SEQIFO_TOKEN-1].Head();
       while (pCard) {
            qCard = _CardInfo[ENTITY_TOKEN-1].Head();
            while (qCard && strcmp(qCard->Field(1), pCard->Field(9)))
                 qCard = qCard->Next();
            if (!qCard || strcasecmp(qCard->Field(2), "POLYMER")) {
                 pCard = pCard->Next();
                 continue;
            }

            ec_number[0] = '\0';
            if (!entity_to_serialno[atoi(qCard->Field(1))]) {
                 sprintf(buffer, "MOL_ID: %d", ++SerialNo);
                 add_compnd_card(buffer, ++compndSerialNo);
                 entity_to_serialno[atoi(qCard->Field(1))] = SerialNo;
                
                 found = FALSE;
                 rCard = _CardInfo[ENTDES_TOKEN-1].Head();
                 while (rCard && strcmp(rCard->Field(1), qCard->Field(1)))
                     rCard = rCard->Next();
                 if (rCard && strcmp(rCard->Field(3), "")) {
                      found = TRUE;
                      if (!strncasecmp(rCard->Field(3), "PROTEIN", 7)) {
                           strcpy(buffer1, rCard->Field(3));
                           // clean_descriptor(buffer1, ec_number);
                           sprintf(buffer, "MOLECULE: %s", buffer1);
                      } else sprintf(buffer, "MOLECULE: %s", rCard->Field(3));
                 } 
  
                 if (!found) {
                      rCard = _CardInfo[PROTEI_TOKEN-1].Head();
                      while (rCard && strcmp(rCard->Field(1), qCard->Field(1)))
                          rCard = rCard->Next();
                      if (rCard) {
                           found = TRUE;
                           sprintf(buffer, "MOLECULE: %s", rCard->Field(3));
                      }
                 }
  
                 if (!found) strcpy(buffer, "MOLECULE: ");
                 ndb_set_string_to_upper(buffer);
                 add_compnd_card(buffer, ++compndSerialNo);
  
                 if (strcmp(pCard->Field(3), ""))
                      sprintf(buffer, "CHAIN: %s", pCard->Field(3));
                 else sprintf(buffer, "CHAIN: %s", "NULL");
                 add_compnd_card(buffer, ++compndSerialNo);
 
                 rCard = _CardInfo[FRAGMT_TOKEN-1].Head();
                 while (rCard && strcmp(rCard->Field(1), qCard->Field(1)))
                      rCard = rCard->Next();
                 if (rCard && strcmp(rCard->Field(3), "")) {
                      sprintf(buffer, "FRAGMENT: %s", rCard->Field(3));
                      ndb_set_string_to_upper(buffer);
                      add_compnd_card(buffer, ++compndSerialNo);
                 }

                 rCard = _CardInfo[ENTNAM_TOKEN-1].Head();
                 while (rCard && strcmp(rCard->Field(1), qCard->Field(1)))
                      rCard = rCard->Next();
                 if (rCard && strcmp(rCard->Field(3), "")) {
                      sprintf(buffer, "SYNONYM: %s", rCard->Field(3));
                      ndb_set_string_to_upper(buffer);
                      add_compnd_card(buffer, ++compndSerialNo);
                 }

                 found = FALSE;
                 rCard = _CardInfo[ECNUM_TOKEN-1].Head();
                 while (rCard && strcmp(rCard->Field(1), qCard->Field(1)))
                      rCard = rCard->Next();
                 if (rCard && strcmp(rCard->Field(2), "")) {
                      sprintf(buffer, "EC: %s", rCard->Field(2));
                      add_compnd_card(buffer, ++compndSerialNo);
                      found = TRUE;
                 }
                 if (!found && strcmp(ec_number, "")) {
                      sprintf(buffer, "EC: %s", ec_number);
                      add_compnd_card(buffer, ++compndSerialNo);
                 }

                 if (!strcasecmp(qCard->Field(3), "NAT")) {
                      isSynthetic = NAT;
                 } else {
                      add_compnd_card("ENGINEERED: YES", ++compndSerialNo);
                      isSynthetic = GEN;
                      if (!strcasecmp(qCard->Field(3), "SYN")) isSynthetic = SYN;
                 }
  
                 found = FALSE;
                 rCard = _CardInfo[ENTMUT_TOKEN-1].Head();
                 while (rCard && strcmp(rCard->Field(1), qCard->Field(1)))
                     rCard = rCard->Next();
                 if (rCard && strcmp(rCard->Field(3), "")) {
                      add_compnd_card("MUTATION: YES", ++compndSerialNo);
                      found = TRUE;
                 }
                 rCard = _CardInfo[ENTAIL_TOKEN-1].Head();
                 while (rCard && strcmp(rCard->Field(1), qCard->Field(1)))
                      rCard = rCard->Next();
                 if (rCard && strcmp(rCard->Field(3), "")) {
                      sprintf(buffer, "OTHER_DETAILS: %s", rCard->Field(3));
                      ndb_set_string_to_upper(buffer);
                      add_compnd_card(buffer, ++compndSerialNo);
                 }
  
                 if (isSynthetic == GEN)
                      ndb_to_pdb_get_source_card(SerialNo, atoi(qCard->Field(1)),
                         isSynthetic, &sourceSerialNo, nsource_Gen, PdbSource_Gen);
                 else if (isSynthetic == SYN)
                      ndb_to_pdb_get_source_card(SerialNo, atoi(qCard->Field(1)),
                         isSynthetic, &sourceSerialNo, nsource_Syn, PdbSource_Syn);
                 else if (isSynthetic == NAT)
                      ndb_to_pdb_get_source_card(SerialNo, atoi(qCard->Field(1)),
                         isSynthetic, &sourceSerialNo, nsource_Nat, PdbSource_Nat);
            } else {
                 compndCard = _CardInfo[COMPND_TOKEN-1].Head();
                 while (compndCard) {
                 if (strstr(compndCard->Field(2), "MOL_ID:") &&
                     atoi(&compndCard->Field(2)[8]) ==
                     entity_to_serialno[atoi(pCard->Field(9))] && compndCard->Next() &&
                     compndCard->Next()->Next() && 
                     !strncmp(compndCard->Next()->Next()->Field(2), "CHAIN", 5))
                      break;
                      compndCard = compndCard->Next();
                 }
                 if (compndCard && compndCard->Next()) {
                      compndCard = compndCard->Next()->Next();
                      if (compndCard != NULL) {
                           compndCard->AttachToField(", ", 2);
                           compndCard->AttachToField(pCard->Field(3), 2);
                      }
                 }
            }
            pCard = pCard->Next();
       }

       compndCard = _CardInfo[COMPND_TOKEN-1].Head();
       while (compndCard && compndCard != _CardInfo[COMPND_TOKEN-1].Tail()) {
            strcpy(buffer, compndCard->Field(2));
            for (i = strlen(buffer) - 1; i >= 0; i--) 
                 if (buffer[i] != ' ') break;
            if (buffer[i] != ';') sprintf(&buffer[i+1], ";");
            compndCard->AddElementToField(buffer, 2);
            compndCard = compndCard->Next();
       }
       sourceCard = _CardInfo[SOURCE_TOKEN-1].Head();
       while (sourceCard && sourceCard != _CardInfo[SOURCE_TOKEN-1].Tail()) {
            strcpy(buffer, sourceCard->Field(2));
            for (i = strlen(buffer) - 1; i >= 0; i--) 
                 if (buffer[i] != ' ') break;
            if (buffer[i] != ';') sprintf(&buffer[i+1], ";");
            sourceCard->AddElementToField(buffer, 2);
            sourceCard = sourceCard->Next();
       }
       compndCard = _CardInfo[COMPND_TOKEN-1].Head();
       if (compndCard == NULL) {
            compndCard = _CardInfo[COMPND_TOKEN-1].
                       add_a_new_card(NdbToken[COMPND_TOKEN-1], 0);
            compndCard->AddElementToField("NULL", 2);
       }
       sourceCard = _CardInfo[SOURCE_TOKEN-1].Head();
       if (sourceCard == NULL) {
            sourceCard = _CardInfo[SOURCE_TOKEN-1].
                       add_a_new_card(NdbToken[SOURCE_TOKEN-1], 0);
            sourceCard->AddElementToField("NULL", 2);
       }
}

void _MAXIT::add_compnd_card(const char *value, const int serialNo)
{
       One_Card_Info_Format *compndCard =
          _CardInfo[COMPND_TOKEN-1].add_a_new_card(NdbToken[COMPND_TOKEN-1], 0);
       compndCard->AddElementToField(value, 2);
       compndCard->AddElementToField(serialNo, 3);
}

void _MAXIT::ndb_to_pdb_get_source_card(const int serialno, const int entityId, 
              const int isSynthetic, int *sourceSerialNo, const int nsource,
              PDB_Source_Format *PdbSource)
{
       One_Card_Info_Format *sourceCard =
            _CardInfo[SOURCE_TOKEN-1].add_a_new_card(NdbToken[SOURCE_TOKEN-1], 0); 
       char buffer[800];
       CifString string_buffer;
       sprintf(buffer, "MOL_ID: %d", serialno);
       sourceCard->AddElementToField(buffer, 2);
       sourceCard->AddElementToField(++(*sourceSerialNo), 3);

       if (isSynthetic == SYN) {
            sourceCard = _CardInfo[SOURCE_TOKEN-1].
                   add_a_new_card(NdbToken[SOURCE_TOKEN-1], 0);
            sourceCard->AddElementToField("SYNTHETIC: YES", 2);
            sourceCard->AddElementToField(++(*sourceSerialNo), 3);
       }

       for (int i = 0; i < nsource; i++) {
            if (PdbSource[i].NdbTokenId != 0 && PdbSource[i].NdbTokenId <= MxTokens) {
                 One_Card_Info_Format *pCard = _CardInfo[PdbSource[i].NdbTokenId-1].Head();
                 while (pCard) {
                      if (strcmp(pCard->Field(PdbSource[i].NdbFieldNo-1), "") &&
                           atoi(pCard->Field(1)) == entityId) break;
          	      pCard = pCard->Next();
                 }
                 if (pCard) {
                      sourceCard = _CardInfo[SOURCE_TOKEN-1].
                             add_a_new_card(NdbToken[SOURCE_TOKEN-1], 0);
                      string_buffer.Copy(PdbSource[i].pdbname);
                      string_buffer += ": ";
                      string_buffer += pCard->Field(PdbSource[i].NdbFieldNo-1);
                      sourceCard->AddElementToField(string_buffer.Text(), 2);
                      sourceCard->AddElementToField(++(*sourceSerialNo), 3);
                 }
            }
       }
}

void _MAXIT::add_remark_90()
{
       One_Card_Info_Format *pCard = _CardInfo[REMARK_TOKEN-1].Head();
       while (pCard) {
            One_Card_Info_Format *pRemarkCard = _CardInfo[PREMRK_TOKEN-1].
                   add_a_new_card(NdbToken[PREMRK_TOKEN-1], 0);
            pRemarkCard->AddElementToField(pCard->Field(1), 1);
            pRemarkCard->AddElementToField(pCard->Field(2), 2);
            pCard = _CardInfo[REMARK_TOKEN-1].delete_a_card(pCard);
       }

       for (int remark_no = 2; remark_no <= 999; remark_no++) {
            if (remark_no == 90)
                 ndb_to_pdb_get_general_remark(Num_Remark_90, Remark_90, 0, 0, 0, 0);
            else ndb_to_pdb_get_remark_from_premark(remark_no);
       }
}

void _MAXIT::ndb_to_pdb_get_remarks(const int remark)
{
       int Num_RemarkNo = 0, RemarkNo[20], RMKSKP_TOKEN, i, j, len;
       int is_electron_microscopy = 0;
       int is_powder_diffration = 0;
       int exist_DNA = 0, CMRNST_TOKEN, RCSBID_TOKEN;
       int /* IsInosine = 0, */ is_skip, /* exist, */ remark_no, card_id, SOLEXP_TOKEN;
       char buff[82], buff1[10], *p;
       const char *model_text = "This theoretical model entry was not annotated and not validated by the wwPDB staff and therefore may not conform to the PDB format.";
       _CHAIN *chain = NULL;
       One_Card_Info_Format *pCard = NULL, *qCard = NULL;

       if (_molecules && _num_mol) {
            chain = _molecules[0]->GetFirstChain();
            while (chain) {
                 if (chain->chain_type() == ATOMN_TOKEN) {
/* */
                      exist_DNA = 1;
                      if (chain->na_type() != ATOMN_TYPE_DNA_ONLY) {
                           exist_DNA = 0;
                           break;
                      }
/* */
/*
                      if (chain->na_type() == ATOMN_TYPE_DNA_ONLY) {
                           exist_DNA = 1;
                           break;
                      }
*/
                 }
                 chain = _molecules[0]->GetNextChain();
            }
       }

       SOLEXP_TOKEN = find_card_id("SOLEXP");
       CMRNST_TOKEN = find_card_id("CMRNST");
       RCSBID_TOKEN = find_card_id("RCSBID");

       for (i = 0; i < 20; i++) RemarkNo[i] = 0;
       
       pCard = _CardInfo[PROCES_TOKEN-1].Head();
       if (pCard) {
            if (!strcasecmp(pCard->Field(5), "OSAKA")) pCard->AddElementToField("PDBJ", 5);
            if (!strcasecmp(pCard->Field(5), "PRAGUE")) pCard->AddElementToField("RCSB", 5);
       }

       if (_CardInfo[PROCES_TOKEN-1].Head() &&
          (_CardInfo[PROCES_TOKEN-1].Head()->Field(7)[0] == 'Y' ||
           _CardInfo[PROCES_TOKEN-1].Head()->Field(7)[0] == 'y')) {
            RemarkNo[0] = 500;
            RemarkNo[1] = 525;
            Num_RemarkNo = 2;
       }

       int is_unprocessed_model = 0;
       RMKSKP_TOKEN = find_card_id("RMKSKP");
       if (_CardInfo[RMKSKP_TOKEN].Head() &&
           strcmp(_CardInfo[RMKSKP_TOKEN].Head()->Field(1), "")) {
            strcpy(buff, _CardInfo[RMKSKP_TOKEN].Head()->Field(1));
            if (!strcasecmp(buff, "unprocessed")) {
                 is_unprocessed_model = 1;
                 RemarkNo[0] = 500;
                 Num_RemarkNo = 1;
            } if (buff[0] == 'Y' || buff[0] == 'y') {
                 RemarkNo[0] = 500;
                 RemarkNo[1] = 525;
                 Num_RemarkNo = 2;
            } else {
                 len = strlen(buff);
                 for (i = 0; i < len; i++) {
                      if (buff[i] == ',') buff[i] = ' ';
                 }
                 ndb_clean_string(buff);
                 p = get_word(buff, buff1);
                 if (atoi(buff1) != 350) RemarkNo[Num_RemarkNo++] = atoi(buff1);
                 while (p) {
                      p = get_word(p, buff1);
                      if (atoi(buff1) != 350) RemarkNo[Num_RemarkNo++] = atoi(buff1);
                 }
            }
       }

       calculate_all_contact(WATER_LIMIT, 1, 3, true);
       get_molecule_topology();
/*
       get_ebi_validate_contact();
       get_ebi_validate_bond();
       get_ebi_validate_angle();
       get_ebi_validate_torsion();
       get_ebi_validate_peptide_omega();
       get_validate_main_chain_plane();
       get_validate_calpha_chiral();
       get_validate_planes();
*/
       if (_experiment_type & EXPERIMENT_TYPE_NMR ||
           _experiment_type & EXPERIMENT_TYPE_NMR_SOLID ||
           _experiment_type & EXPERIMENT_TYPE_MODEL)
            ndb_to_pdb_processing_nmr_remarks();
       if (_experiment_type & EXPERIMENT_TYPE_NMR ||
           _experiment_type & EXPERIMENT_TYPE_NMR_SOLID ||
           _experiment_type & EXPERIMENT_TYPE_MODEL) {
            pCard = _CardInfo[REFMET_TOKEN-1].Head();
            while (pCard) {
                 if (!strcmp(pCard->Field(4), "")) {
                      pCard->UpperCaseField(3);
                      char *Author = find_refine_program_author(pCard->Field(3));
                      if (Author) {
                           pCard->AddElementToField(Author, 4);
                           delete [] Author;
                      }
                 }
                 pCard = pCard->Next();
            }
       }
       pCard = _CardInfo[EXPDTA_TOKEN-1].Head();
       if (pCard) {
            if (strstr(pCard->Field(2), "ELECTRON MICROSCOPY") ||
                strstr(pCard->Field(2), "CRYO-EM RECONSTRUCTION"))
                 is_electron_microscopy = 1;
            if (strstr(pCard->Field(2), "POWDER DIFFRACTION"))
                 is_powder_diffration = 1;
       }

       ReVarCifArray<_METAL_COORDINATION> MetalList, tMetalList;
       _METAL_COORDINATION _metal_coord;
       _ATOM metal_atom, polar_atom;
       tMetalList.Clear();
       MetalList.Clear();
       ReVarCifArray<_ATOM> covalent_atoms;
       for (i = 0; i < _n_link; i++) {
            if (_link[i]->type != "metalc") continue;
            if (get_atom_pair(metal_atom, polar_atom, _link[i], covalent_atoms)) {
                 int found = 0;
                 for (unsigned int j = 0; j < tMetalList.Length(); j++) {
                      if (!strcmp(metal_atom.pdb_atmnam(), tMetalList[j].metal_atom.pdb_atmnam()) &&
                          !strcmp(metal_atom.pdb_resnam(), tMetalList[j].metal_atom.pdb_resnam()) &&
                           metal_atom.pdb_resnum() == tMetalList[j].metal_atom.pdb_resnum() &&
                          !strcmp(metal_atom.ins_code(), tMetalList[j].metal_atom.ins_code()) &&
                          !strcmp(metal_atom.pdb_chnid(), tMetalList[j].metal_atom.pdb_chnid())) {
                           int found1 = 0;
                           for (unsigned int k = 0; k < tMetalList[j].polar_atoms.Length(); k++) {
                                if (!strcmp(polar_atom.pdb_atmnam(),
                                        tMetalList[j].polar_atoms[k].pdb_atmnam()) &&
                                    !strcmp(polar_atom.pdb_resnam(),
                                        tMetalList[j].polar_atoms[k].pdb_resnam()) &&
                                     polar_atom.pdb_resnum() ==
                                        tMetalList[j].polar_atoms[k].pdb_resnum() &&
                                    !strcmp(polar_atom.ins_code(),
                                        tMetalList[j].polar_atoms[k].ins_code()) &&
                                    !strcmp(polar_atom.pdb_chnid(),
                                        tMetalList[j].polar_atoms[k].pdb_chnid())) {
                                     found1 = 1;
                                     break;
                                }
                           }
                           if (!found1) tMetalList[j].polar_atoms.Add(polar_atom);
                           for (unsigned int m = 0; m < covalent_atoms.Length(); m++) {
                                found1 = 0;
                                for (unsigned int k = 0; k < tMetalList[j].polar_atoms.Length(); k++) {
                                     if (!strcmp(covalent_atoms[m].pdb_atmnam(),
                                             tMetalList[j].polar_atoms[k].pdb_atmnam()) &&
                                         !strcmp(covalent_atoms[m].pdb_resnam(),
                                             tMetalList[j].polar_atoms[k].pdb_resnam()) &&
                                          covalent_atoms[m].pdb_resnum() ==
                                             tMetalList[j].polar_atoms[k].pdb_resnum() &&
                                         !strcmp(covalent_atoms[m].ins_code(),
                                             tMetalList[j].polar_atoms[k].ins_code()) &&
                                         !strcmp(covalent_atoms[m].pdb_chnid(),
                                             tMetalList[j].polar_atoms[k].pdb_chnid())) {
                                          found1 = 1;
                                          break;
                                     }
                                }
                                if (!found1) tMetalList[j].polar_atoms.Add(covalent_atoms[m]);
                           }
                           found = 1;
                           break;
                      }
                 }
                 if (!found) {
                      _metal_coord.metal_atom = metal_atom;
                      if (_num_mol > 1)
                           _metal_coord.Mol_ID = _molecules[_link[i]->mol_index]->Mol_ID();
                      else _metal_coord.Mol_ID = -1;
                      _metal_coord.polar_atoms.Clear();
                      _metal_coord.polar_atoms.Add(polar_atom);
                      for (unsigned int m = 0; m < covalent_atoms.Length(); m++) {
                           if (!strcmp(covalent_atoms[m].pdb_atmnam(),
                                   polar_atom.pdb_atmnam()) &&
                               !strcmp(covalent_atoms[m].pdb_resnam(),
                                   polar_atom.pdb_resnam()) &&
                                covalent_atoms[m].pdb_resnum() ==
                                   polar_atom.pdb_resnum() &&
                               !strcmp(covalent_atoms[m].ins_code(),
                                   polar_atom.ins_code()) &&
                               !strcmp(covalent_atoms[m].pdb_chnid(),
                                   polar_atom.pdb_chnid())) continue;
                           _metal_coord.polar_atoms.Add(covalent_atoms[m]);
                      }
                      tMetalList.Add(_metal_coord);
                 }
            }
       }
       for (unsigned int i2 = 0; i2 < tMetalList.Length(); i2++) {
            if (tMetalList[i2].polar_atoms.Length() > 1) MetalList.Add(tMetalList[i2]);
       }

       find_user_defined_remark(0); // REMARK 0 for re-refinement entry
       find_secondary_citations(RE_REFINE, 0);

       find_secondary_citations("", 1);
       
       CifString EBI_ID;
       EBI_ID.Clear();
       int is_solid_stat_nmr = 0;
       int is_version_v315 = 0;
       if (_CifObj) {
            CifString cs;
            int rowNo, idx1, idx2;
            ISTable *Table = _CifObj->GetTablePtr(_StructureId, "database_2");
            if (!is_empty_table(Table)) {
                 rowNo = Table->GetNumRows();
                 idx1  = Table->GetColumnIndex("database_id");
                 idx2  = Table->GetColumnIndex("database_code");
                 for (i = 0; i < rowNo; i++) {
                      get_value(cs, Table, idx1, i);
                      cs.ToUpper();
                      if (cs == "EBI" || cs == "PDBE") {
                           get_value(EBI_ID, Table, idx2, i);
                           break;
                      }
                 }
            }
            Table = _CifObj->GetTablePtr(_StructureId, "exptl");
            if (!is_empty_table(Table)) {
                 idx1  = Table->GetColumnIndex("details");
                 get_value(cs, Table, idx1, 0);
                 if (cs != "") { 
                      cs.ToUpper();
                      char *buff2 = new char[strlen(cs.Text())+1];
                      memset(buff2, 0, strlen(cs.Text())+1);
                      strcpy(buff2, cs.Text());
                      for (unsigned int i = 0; i < strlen(buff2); i++) {
                           if (buff2[i] == '-') buff2[i] = ' ';
                      }
                      ndb_clean_string(buff2);
                      if (strstr(buff2, "SOLID STATE NMR")) is_solid_stat_nmr = 1;
                 }
            }

            Table = _CifObj->GetTablePtr(_StructureId, "rcsb_version");
            if (!is_empty_table(Table)) {
                 rowNo = Table->GetNumRows();
                 idx1  = Table->GetColumnIndex("details");
                 for (i = rowNo - 1; i >= 0; --i) {
                      get_value(cs, Table, idx1, i);
                      if (cs == "compliance with PDB format V.3.15") {
                           is_version_v315 = 1;
                           if (_num_mol <= 1) is_version_v315 = 0;
                           break;
                      }
                 }
            }
       }

       for (remark_no = 2; remark_no <= 999; remark_no++) {
            is_skip = 0;
            for (i = 0; i < Num_RemarkNo; i++) {
                 if (remark_no == RemarkNo[i]) {
                      is_skip = 1;
                      break;
                 }
            }
            if (is_skip) continue;

            if (/* */ remark_no != 350 && /* */ find_user_defined_remark(remark_no)) continue;

            if (remark_no == 2) {
                 if ((_experiment_type & EXPERIMENT_TYPE_XRAY) ||
                     (_experiment_type & EXPERIMENT_TYPE_NEUTRON) ||
                     (_experiment_type & EXPERIMENT_TYPE_CRYO_EM) ||
                     (_experiment_type & EXPERIMENT_TYPE_FIBER) ||
                     (_experiment_type & EXPERIMENT_TYPE_ELECTRON)) {
                      pCard = _CardInfo[RFACTR_TOKEN-1].Head();
                      if (!pCard) pCard = _CardInfo[RFACTR_TOKEN-1].
                              add_a_new_card(NdbToken[RFACTR_TOKEN-1], 0);
                      qCard = _CardInfo[CMRNST_TOKEN].Head();
                      if (qCard && strcmp(qCard->Field(3), "") &&
                          !strcmp(pCard->Field(6), "")) {
                           pCard->AddElementToField(qCard->Field(3), 6);
                      }
                      if (strcmp(pCard->Field(6), "")) {
                           char *resolution = new char[strlen(pCard->Field(6))+1];
                           strcpy(resolution, pCard->Field(6));
                           p = strstr(resolution, ".");
                           if (p) {
                                char *q = resolution + strlen(resolution) - 1;
                                while (q > p + 2) {
                                     if (*q == '0') *q = '\0';
                                     else break;
                                     q--;
                                }
                                pCard->AddElementToField(resolution, 6);
                           }
                           delete [] resolution;
                      }
                      ndb_to_pdb_get_general_remark(1, &Remark_2, 0, 0, 0, 0);
                 } else ndb_to_pdb_get_general_remark(1, &Remark_2_NMR, 0, 0, 0, 0);
            } else if (remark_no == 3) {
                 if (_experiment_type & EXPERIMENT_TYPE_XRAY ||
                     _experiment_type & EXPERIMENT_TYPE_NEUTRON ||
                     _experiment_type & EXPERIMENT_TYPE_FIBER ||
                     _experiment_type & EXPERIMENT_TYPE_ELECTRON)
                      ndb_to_pdb_get_group_remarks();
                 else if (_experiment_type & EXPERIMENT_TYPE_NMR ||
                     _experiment_type & EXPERIMENT_TYPE_NMR_SOLID ||
                     _experiment_type & EXPERIMENT_TYPE_MODEL)
                      ndb_to_pdb_get_general_remark(Num_Remark_3_NMR, Remark_3_NMR,
                                       0, 0, 0, 0);
                 else if (_experiment_type & EXPERIMENT_TYPE_CRYO_EM)
                      ndb_to_pdb_get_general_remark(Num_Remark_3_Cryo_Em,
                                     Remark_3_Cryo_Em, 0, 0, 0, 0);
                 else if (_experiment_type & EXPERIMENT_TYPE_SOLN_SCT) {
                      for (i = 0; i < Remark_3_Solution.Remarks_No; i++) {
                           ndb_to_pdb_get_general_remark(Remark_3_Solution.num_remarks[i],
                                 Remark_3_Solution.remarks[i], 0, 0, 0, 0);
                      }
                 }
            } else if (remark_no == 4) {
                 if (is_version_v315)
                      ndb_to_pdb_get_general_remark(Num_Remark_4_NEW, Remark_4_NEW, 0, 0, 0, 0);
                 else ndb_to_pdb_get_general_remark(Num_Remark_4_NEW, Remark_4_NEW_Default, 0, 0, 0, 0);
#if 0
// #ifdef IUPAC
                 if (_CifObj) {
                      ISTable *Table = _CifObj->GetTablePtr(_StructureId, "rcsb_version");
                      if (!is_empty_table(Table)) {
                           int rowNo = Table->GetNumRows();
                           int idx1  = Table->GetColumnIndex("major_version");
                           int idx2  = Table->GetColumnIndex("minor_version");
                           int idx3  = Table->GetColumnIndex("revision_date");
                           CifString cs, date, version1, version2;
                           get_value(date, Table, idx3, rowNo - 1);
                           get_value(version1, Table, idx1, rowNo - 1);
                           get_value(version2, Table, idx2, rowNo - 1);
/*
                           pCard = _CardInfo[REMARK_TOKEN-1].
                                 add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                           pCard->AddElementToField(4, 1);
                           pCard = _CardInfo[REMARK_TOKEN-1].
                                 add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                           pCard->AddElementToField(4, 1);
                           pCard->AddElementToField("THIS FILE IS A TEST VERSION.", 2);
*/
                           pCard = _CardInfo[REMARK_TOKEN-1].
                                 add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                           pCard->AddElementToField(4, 1);
                           pCard = _CardInfo[REMARK_TOKEN-1].
                                 add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                           pCard->AddElementToField(4, 1);
                           pCard->AddElementToField("THIS IS THE REMEDIATED VERSION OF THIS PDB ENTRY.", 2);
                           cs.Copy("REMEDIATED DATA FILE REVISION ");
                           cs += version1;
                           cs += ".";
                           cs += version2;
                           cs += " (";
                           cs += date;
                           cs += ")";
                           pCard = _CardInfo[REMARK_TOKEN-1].
                                 add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                           pCard->AddElementToField(4, 1);
                           pCard->AddElementToField(cs.Text(), 2);
                      }
                 }
#endif
            } else if (remark_no == 5) ndb_to_pdb_get_remark_5();
            else if (remark_no == 100) {
                 int BMRB_TOKEN = find_card_id("BMRBID");
                 if (is_unprocessed_model) {
                      ndb_convert_time_format(NDB_FILE_FORMAT_PDB, _CardInfo[HEADER_TOKEN-1].Head()->Field(2),
                                   buff, /*12 */ 9);
                      _CardInfo[HEADER_TOKEN-1].Head()->AddElementToField(buff, 2);
                      ndb_to_pdb_get_general_remark(Num_Remark_100_model,
                              Remark_100_model, 0, 0, 0, 0);
                      ndb_convert_time_format(NDB_FILE_FORMAT_PDB, _CardInfo[HEADER_TOKEN-1].Head()->Field(2),
                                   buff, /* 10 */ 9);
                      _CardInfo[HEADER_TOKEN-1].Head()->AddElementToField(buff, 2);
/*
                 } else if (_CardInfo[NDBPRO_TOKEN-1].Head() &&
                      strcmp(_CardInfo[NDBPRO_TOKEN-1].Head()->Field(4), "")) {
                      if (!strcmp(_CardInfo[NDBFIL_TOKEN-1].Head()->Field(1),
                                  _CardInfo[PDBFIL_TOKEN-1].Head()->Field(1)) &&
*/
                 } else if (_CardInfo[PROCES_TOKEN-1].Head() &&
                      strcmp(_CardInfo[PROCES_TOKEN-1].Head()->Field(5), "")) {
                      pCard = _CardInfo[RCSBID_TOKEN].Head();
                      if (!pCard) pCard = _CardInfo[RCSBID_TOKEN].add_a_new_card(NdbToken[RCSBID_TOKEN], 0);
                      pCard = _CardInfo[NDBFIL_TOKEN-1].Head();
                      if (!pCard) pCard = _CardInfo[NDBFIL_TOKEN-1].add_a_new_card(NdbToken[NDBFIL_TOKEN-1], 0);
                      if (strcmp(_CardInfo[RCSBID_TOKEN].Head()->Field(1), "") &&
                          strcmp(_CardInfo[RCSBID_TOKEN].Head()->Field(1),
                                 _CardInfo[NDBFIL_TOKEN-1].Head()->Field(1))) {
                           _CardInfo[NDBFIL_TOKEN-1].Head()->AddElementToField(
                                _CardInfo[RCSBID_TOKEN].Head()->Field(1), 1);
                      }
                      if (EBI_ID != "") {
                           _CardInfo[NDBFIL_TOKEN-1].Head()->AddElementToField(EBI_ID.Text(), 1);
                           _CardInfo[PROCES_TOKEN-1].Head()->AddElementToField("PDBE", 5);
                           ndb_to_pdb_get_general_remark(Num_Remark_100_EBI, Remark_100_EBI, 0, 0, 0, 0);
                      } else if (strstr(_CardInfo[NDBFIL_TOKEN-1].Head()->Field(1), "RCSB"))
                           ndb_to_pdb_get_general_remark(Num_Remark_100_RCSB, Remark_100_RCSB, 0, 0, 0, 0);
                      else if (_CardInfo[PDBFIL_TOKEN-1].Head() == NULL && _CardInfo[BMRB_TOKEN].Head() &&
                              strcmp(_CardInfo[BMRB_TOKEN].Head()->Field(1), ""))
                           ndb_to_pdb_get_general_remark(Num_Remark_100_BMRB, Remark_100_BMRB, 0, 0, 0, 0);
                      else ndb_to_pdb_get_general_remark(1, &Remark_100_BNL, 0, 0, 0, 0);
                 } else ndb_to_pdb_get_general_remark(1, &Remark_100_BNL, 0, 0, 0, 0); 
/*
            } else if (remark_no == 101) {
                 exist = ndb_to_pdb_get_remark_from_premark(101);
                 if (!exist) {
                      pCard = _CardInfo[MODIFB_TOKEN-1].Head();
                      if (pCard)
                           ndb_to_pdb_get_general_remark(1, &Remark_101, 0, 0, 0, 0); 
                 }
            } else if (remark_no == 102) {
                 if (exist_DNA) {
                      if (_CardInfo[MISMAT_TOKEN-1].Head()) {
                           ndb_to_pdb_get_general_remark(1, &Remark_102, 0, 0, 0, 0);  
                      } else ndb_to_pdb_get_remark_from_premark(102);
                 }
            } else if (remark_no == 103) {
                 IsInosine = 0;
                 pCard = _CardInfo[MODIFB_TOKEN-1].Head();
                 while (pCard) {
                      if (!res_name_cmp(pCard->Field(5), "I")) {
                           IsInosine = 1;
                           break;
                      }
                      pCard = pCard->Next();
                 }
                 if (exist_DNA || IsInosine)
                      ndb_to_pdb_get_general_remark(Num_Remark_103, Remark_103, 1, 0, 1, 0);
            } else if (remark_no == 104) {
                 if (IsInosine)
                      ndb_to_pdb_get_general_remark(1, &Remark_104, 0, 0, 1, 0);
            } else if (remark_no == 105) {
                 exist = 0;
                 pCard = _CardInfo[SEQIFO_TOKEN-1].Head();
                 while (pCard != NULL) {
                      if (!strcmp(pCard->Field(5), "NA")) {
                           exist = 1; break;
                      }
                      pCard = pCard->Next();
                 }
                 if (exist) {
                      exist = ndb_to_pdb_get_remark_from_premark(105);
                      if (!exist) ndb_to_pdb_get_general_remark(Num_Remark_105,
                                     Remark_105, 0, 0, 0, 0);
                 }
*/
            } else if (remark_no == 200) {
                 if (!is_powder_diffration) {
                      if (_experiment_type & EXPERIMENT_TYPE_XRAY)
                           ndb_to_pdb_get_general_remark(Num_Remark_200, Remark_200, 0, 0, 0, 0,
                                                         "X-RAY DIFFRACTION");
                      else if (_experiment_type & EXPERIMENT_TYPE_FIBER)
                           ndb_to_pdb_get_general_remark(Num_Remark_200, Remark_200, 0, 0, 0, 0,
                                                         "FIBER DIFFRACTION");
                 } 
            } else if (remark_no == 205) {
                 if (_experiment_type & EXPERIMENT_TYPE_FIBER)
                      ndb_to_pdb_get_general_remark(Num_Remark_205, Remark_205, 0, 0, 0, 0);
            } else if (remark_no == 210) {
                 if (_experiment_type & EXPERIMENT_TYPE_NMR ||
                     _experiment_type & EXPERIMENT_TYPE_NMR_SOLID)
                      ndb_to_pdb_get_general_remark(Num_Remark_210, Remark_210, 0, 0, 0, 0);
            } else if (remark_no == 215) {
                 if  (_experiment_type & EXPERIMENT_TYPE_NMR && _cell->is_artifical)
                      ndb_to_pdb_get_general_remark(Num_Remark_215, Remark_215, 0, 0, 0, 0);
            } else if (remark_no == 217) {
                 if  (_experiment_type & EXPERIMENT_TYPE_NMR_SOLID && _cell->is_artifical)
                      ndb_to_pdb_get_general_remark(Num_Remark_217, Remark_217, 0, 0, 0, 0);
            } else if (remark_no == 220) {
                 if (_experiment_type & EXPERIMENT_TYPE_MODEL) {
                      int MODTLS_TOKEN = find_card_id("MODTLS");
                      if (MODTLS_TOKEN >= 0 && is_unprocessed_model) {
                           pCard = _CardInfo[MODTLS_TOKEN].Head();
                           if (!pCard) pCard = _CardInfo[MODTLS_TOKEN].
                                    add_a_new_card(NdbToken[MODTLS_TOKEN], 0);
                           if (strcmp(pCard->Field(2), "")) {
                                if (pCard->Field(2)[strlen(pCard->Field(2))-1] != '.')
                                     pCard->AttachToField(".", 2);
                                pCard->AttachToField(" ", 2);
                           }
                           pCard->AttachToField(model_text, 2);
                      }
                      ndb_to_pdb_get_general_remark(Num_Remark_220, Remark_220, 0, 0, 0, 0);
                 }
            } else if (remark_no == 225) {
                 if (_experiment_type & EXPERIMENT_TYPE_MODEL)
                      ndb_to_pdb_get_general_remark(Num_Remark_225, Remark_225, 0, 0, 0, 0);
            } else if (remark_no == 230) {
                 if (_experiment_type & EXPERIMENT_TYPE_NEUTRON)
                      ndb_to_pdb_get_general_remark(Num_Remark_230, Remark_230, 0, 0, 0, 0,
                                                    "NEUTRON DIFFRACTION");
            } else if (remark_no == 240) {
                 if (_experiment_type & EXPERIMENT_TYPE_ELECTRON)
                      ndb_to_pdb_get_general_remark(Num_Remark_240, Remark_240, 0, 0, 0, 0,
                                                    "ELECTRON CRYSTALLOGRAPHY");
            } else if (remark_no == 245) {
                 if (_experiment_type & EXPERIMENT_TYPE_CRYO_EM)
                      ndb_to_pdb_get_general_remark(Num_Remark_245, Remark_245, 0, 0, 0, 0);
            } else if (remark_no == 247) {
                 if (_experiment_type & EXPERIMENT_TYPE_CRYO_EM)
                      ndb_to_pdb_get_general_remark(Num_Remark_247, Remark_247, 0, 0, 0, 0);
            } else if (remark_no == 250) {
                 if ((_experiment_type & EXPERIMENT_TYPE_BASIC))
                      ndb_to_pdb_get_general_remark(Num_Remark_250, Remark_250, 0, 0, 0, 0);
            } else if (remark_no == 265) {
                 if (_experiment_type & EXPERIMENT_TYPE_SOLN_SCT) {
                      if (SOLEXP_TOKEN >= 0) {
                           pCard = _CardInfo[SOLEXP_TOKEN].Head();
                           while (pCard) {
                                if (!strcasecmp(pCard->Field(2), "THEORETICAL MODELLING") ||
                                    !strcasecmp(pCard->Field(2), "modelling"))
                                     pCard = _CardInfo[SOLEXP_TOKEN].delete_a_card(pCard);
                                else {
                                     if (!strcasecmp(pCard->Field(2), "x-ray"))
                                          pCard->AddElementToField("SMALL ANGLE X-RAY SCATTERING", 2);
                                     else if (!strcmp(pCard->Field(2), "neutron"))
                                          pCard->AddElementToField("SMALL ANGLE NEUTRON SCATTERING", 2);
                                     pCard = pCard->Next();
                                }
                           }
                      }
                      for (i = 0; i < Remark_265.Remarks_No; i++) {
                           if (strcmp(Remark_265.TokenName[i], "")) {
                                card_id = find_card_id(Remark_265.TokenName[i]);
                                if (card_id == -1) continue;
                                pCard = _CardInfo[card_id].Head();
                                if (!pCard) continue;
                           }
                           j = 0;
                           if (Remark_265.repeat[i]) j = 1;
                           ndb_to_pdb_get_general_remark(Remark_265.num_remarks[i],
                                 Remark_265.remarks[i], j, 0, 0, 1);
                      }
                 }
            } else if (remark_no == 280) {
                 if (_experiment_type & EXPERIMENT_TYPE_XRAY ||
                     _experiment_type & EXPERIMENT_TYPE_NEUTRON)
                      ndb_to_pdb_get_remark_280_crystal();
            } else if (remark_no == 290) {
                 if (_experiment_type & EXPERIMENT_TYPE_XRAY ||
                     _experiment_type & EXPERIMENT_TYPE_NEUTRON ||
                     _experiment_type & EXPERIMENT_TYPE_ELECTRON ||
                     is_electron_microscopy)
                      ndb_to_pdb_get_remark_290();
            } else if (remark_no == 300) {
                 check_virus_ncs_remark_300();
/*
                 if (check_virus_ncs_remark_300()) {
                 } else if (is_electron_microscopy) {
                      // if (!ndb_to_pdb_get_remark_from_premark(300)) {
                           if (_experiment_type & EXPERIMENT_TYPE_CRYO_EM) {
                                ndb_to_pdb_get_remark_300();
                                for (i = 0; i < Remark_Cryo_Em_300.Remarks_No; i++) {
                                     j = 0;
                                     if (Remark_Cryo_Em_300.repeat[i]) j = 1;
                                     ndb_to_pdb_get_general_remark(Remark_Cryo_Em_300.num_remarks[i],
                                           Remark_Cryo_Em_300.remarks[i], j, 0, 0, 0);
                                }
                           }
                      // }
                 } else if (_experiment_type & EXPERIMENT_TYPE_XRAY ||
                          _experiment_type & EXPERIMENT_TYPE_NEUTRON ||
                          _experiment_type & EXPERIMENT_TYPE_ELECTRON) {
                      if (!ndb_to_pdb_get_remark_300()) ;
                           // ndb_to_pdb_get_remark_from_premark(300);
                 } else if (_experiment_type & EXPERIMENT_TYPE_CRYO_EM) {
                      ndb_to_pdb_get_remark_300();
                      for (i = 0; i < Remark_Cryo_Em_300.Remarks_No; i++) {
                           j = 0;
                           if (Remark_Cryo_Em_300.repeat[i]) j = 1;
                           ndb_to_pdb_get_general_remark(Remark_Cryo_Em_300.num_remarks[i],
                                 Remark_Cryo_Em_300.remarks[i], j, 0, 0, 0);
                      }
                 }
*/
            } else if (remark_no == 350) {
                 generate_biological_assembly();
#if 0
                 // if (generate_virus_remark_350()) {
                 if (generate_biological_assembly()) {
                 } else /* */ if (is_electron_microscopy) {
/*
                      exist = ndb_to_pdb_get_remark_from_premark(350);
                      if (!exist) */ ndb_to_pdb_get_remark_350();
                 } else if (_experiment_type & EXPERIMENT_TYPE_XRAY ||
                          _experiment_type & EXPERIMENT_TYPE_NEUTRON ||
                          _experiment_type & EXPERIMENT_TYPE_ELECTRON) {
                      if (!ndb_to_pdb_get_remark_350()) ;
                           // ndb_to_pdb_get_remark_from_premark(350);
                 }
#endif
            } else if (remark_no == 375) {
                 if (_experiment_type & EXPERIMENT_TYPE_XRAY ||
                     _experiment_type & EXPERIMENT_TYPE_NEUTRON)
                      ndb_to_pdb_get_remark_375();
            } else if (remark_no == 400 || remark_no == 450 ||
                       remark_no == 600 || remark_no == 999) {
                 ndb_to_pdb_get_special_remark(remark_no);
                 if (remark_no == 400) ndb_to_pdb_get_remark_400_group();
            } else if (remark_no == 465) ndb_to_pdb_get_remark_465();
            else if (remark_no == 470) ndb_to_pdb_get_remark_470();
            else if (remark_no == 475) ndb_to_pdb_get_remark_475_and_480();
            else if (remark_no == 480) continue;
            else if (remark_no == 500) {
                 // if (!(_experiment_type & EXPERIMENT_TYPE_MODEL)) {
                      ndb_to_pdb_get_remark_500(remark);
                      rms_deviation_of_bonds();
                      rms_deviation_of_angles();
                      outside_torsions();
                      non_cis_trans_torsions();
                      side_chain_plane();
                      main_chain_planarity();
                      alpha_carbon_chiral();
                 // }
            } else if (remark_no == 525) ndb_to_pdb_get_remark_525();
            else if (remark_no == 610) ndb_to_pdb_get_remark_610();
            else if (remark_no == 615) ndb_to_pdb_get_remark_615();
            else if (remark_no == 620) ndb_to_pdb_get_remark_620(MetalList);
            else if (remark_no == 630) ndb_to_pdb_get_remark_630();
            else if (remark_no == 800) {
/*
                 exist = ndb_to_pdb_get_remark_from_premark(800);
                 if (!exist)
*/
                      ndb_to_pdb_get_general_remark(Num_Remark_800, Remark_800, 1, 0, 0, 0);
            } else if (remark_no == 900) {
                 if (_CardInfo[RENTRY_TOKEN-1].Head()) {
                      construct_remark_900();
                 } // else ndb_to_pdb_get_remark_from_premark(900);
            } // else ndb_to_pdb_get_remark_from_premark(remark_no);
       }
}

void _MAXIT::ndb_to_pdb_get_remark_5()
{
       if (!_CifObj) return;
       ISTable *Table = _CifObj->GetTablePtr(_StructureId, "ndb_database_PDB_obs_spr");
       if (is_empty_table(Table)) return;

       int idx = Table->GetColumnIndex("details");
       int idx1 = Table->GetColumnIndex("id");
       if (idx < 0 || idx1 < 0) return;

       CifString block_remark, cs;
       block_remark.Clear();
       int rowNo = Table->GetNumRows();
       for (int i = 0; i < rowNo; ++i) {
           get_value(cs, Table, idx1, i);
           cs.ToUpper();
           if (cs != "OBSLTE") continue;
           get_value(cs, Table, idx, i);
           if (cs == "") continue;
           if (block_remark != "") block_remark += " ";
           block_remark += cs;
       }
       if (block_remark == "") return;

       block_remark.ToUpper();
       block_remark.RemoveBlanks();

       ReVarCifArray<CifString> remark_array;
       int width = NdbToken[REMARK_TOKEN-1].FieldList[2].FieldWidth;
       get_remark_array_from_block(remark_array, block_remark.Text(), width);

       if (remark_array.Length() == 0) return;

       for (unsigned int i1 = 0; i1 < remark_array.Length(); i1++) {
            One_Card_Info_Format *newCard = _CardInfo[REMARK_TOKEN-1].
                         add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(5, 1);
            newCard->AddElementToField(remark_array[i1].Text(), 2);
       }
}

int _MAXIT::find_user_defined_remark(const int remark_no)
{
       char buffer[61], *p;
       int found = 0;
       One_Card_Info_Format *pCard = _CardInfo[PDBRMK_TOKEN-1].Head();
       int width = NdbToken[REMARK_TOKEN-1].FieldList[2].FieldWidth;
       while (pCard) {
            if (atoi(pCard->Field(1)) == remark_no) {
                 found = 1;
                 // if (strcmp(pCard->Field(3), "")) {
                      p = pCard->Field(3);
                      while (p) {
                           p = get_max_length_word(p, buffer, width);
                           One_Card_Info_Format *newCard = _CardInfo[REMARK_TOKEN-1].
                                 add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                           newCard->AddElementToField(remark_no, 1);
                           if (remark_no != 350 && remark_no != 400 && remark_no != 465 &&
                               remark_no != 470 && remark_no != 500)
                                ndb_set_string_to_upper(buffer);
                           newCard->AddElementToField(buffer, 2);
                      }
                 // }
            }
            pCard = pCard->Next();
       }
       return found;
}

void _MAXIT::ndb_to_pdb_get_group_remarks()
{
       int card_id, ContNo = 1, one_flag = 1, i;
       int program = 0;
       char buffer[80], buff[80], *p;
       GROUP_REMARKS *group_remark = NULL;
       One_Card_Info_Format *pCard = NULL, *newCard = NULL;

       pCard = _CardInfo[REFMET_TOKEN-1].Head();
       if (pCard) {
            if (strlen(pCard->Field(3)) < 80) {
                 strcpy(buffer, pCard->Field(3));
                 ndb_set_string_to_upper(buffer);
                 if (strstr(buffer, "CNS"))
                      program = 1;
                 else if (strstr(buffer, "NUCLSQ"))
                      program = 2;
                 else if (strstr(buffer, "PROLSQ") ||
                          strstr(buffer, "PROFFT") ||
                          strstr(buffer, "CCP4") ||
                          strstr(buffer, "PROTEIN") ||
                          strstr(buffer, "RESTRAIN") ||
                          strstr(buffer, "GPRLSA"))
                      program = 3;
                 else if (strstr(buffer, "REFMAC 5"))
                      program = 5;
                 else if (strstr(buffer, "REFMAC"))
                      program = 4;
                 else if (/* strstr(buffer, "BUSTER-TNT") ||
                          strstr(buffer, "BUSTER/TNT") */
                          strstr(buffer, "BUSTER"))
                      program = 7;
                 else if (strstr(buffer,"TNT"))
                      program = 6;
                 else if (strstr(buffer,"SHELX"))
                      program = 8;
                 else if (strstr(buffer, "CNX"))
                      program = 9;
                 else if (strstr(buffer, "PHENIX"))
                      program = 10;
            }

            if (pCard->Next()) {
                 one_flag = 0;
                 ContNo = 1;
            } else one_flag = 1;

            newCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(3, 1);
            newCard->AddElementToField("REFINEMENT.", 2);

            while (pCard) {
                 if (strcmp(pCard->Field(3), "") && strlen(pCard->Field(3)) < 63) {
                      pCard->UpperCaseField(3);
                      newCard = _CardInfo[REMARK_TOKEN-1].
                                  add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                      newCard->AddElementToField(3, 1);

                      if (one_flag) sprintf(buffer, "  %s     : %s",
                                   "PROGRAM", pCard->Field(3));
                      else sprintf(buffer, "  %s %-3d : %s", "PROGRAM",
                                    ContNo, pCard->Field(3));
                      ndb_set_string_to_upper(buffer);
                      newCard->AddElementToField(buffer, 2);

                      char *Author = find_refine_program_author(pCard->Field(3));

                      if (Author) {
                           newCard = _CardInfo[REMARK_TOKEN-1].
                                add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                           newCard->AddElementToField(3, 1);
                           p = get_max_length_word(Author, buff, 42);
                           if (!strcmp(buff, "?")) strcpy(buff, "NULL");
                           if (one_flag)
                                sprintf(buffer, "  %s     : %s", "AUTHORS", buff);
                           else sprintf(buffer, "  %s %-3d : %s", "AUTHORS", ContNo, buff);
                           ndb_set_string_to_upper(buffer);
                           newCard->AddElementToField(buffer, 2);
                           while (p) {
                                p = get_max_length_word(p, buff, 42);
                                if (!strcmp(buff, "?")) strcpy(buff, "NULL");
                                newCard = _CardInfo[REMARK_TOKEN-1].
                                          add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                                newCard->AddElementToField(3, 1);
                                sprintf(buffer, "  %s     : %s", "       ", buff);
                                ndb_set_string_to_upper(buffer);
                                newCard->AddElementToField(buffer, 2);
                           }
                           delete [] Author;
                      }
                 }
                 ContNo++;
                 pCard = pCard->Next();
            }
       } else {
            newCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(3, 1);
            newCard->AddElementToField("REFINEMENT.", 2);
            newCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(3, 1);
            newCard->AddElementToField("  PROGRAM     : NULL", 2);
            newCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(3, 1);
            newCard->AddElementToField("  AUTHORS     : NULL", 2);
       }

       if (!program || program == 1) {
            pCard = _CardInfo[RFACTR_TOKEN-1].Head();
            if (pCard && !strcmp(pCard->Field(10), ""))
                 pCard->AddElementToField(pCard->Field(7), 10);
            pCard = _CardInfo[NCSMOD_TOKEN-1].Head();
            if (!pCard) {
                 pCard = _CardInfo[NCSMOD_TOKEN-1].
                          add_a_new_card(NdbToken[NCSMOD_TOKEN-1], 0);
                 pCard->AddElementToField(1, 1);
            }
            pCard = _CardInfo[XFILES_TOKEN-1].Tail();
            if (!pCard) {
                 pCard = _CardInfo[XFILES_TOKEN-1].
                          add_a_new_card(NdbToken[XFILES_TOKEN-1], 0);
                 pCard->AddElementToField(1, 1);
            } else {
                 for (i = 1; i < pCard->NumField(); i++) {
                      if (strcmp(pCard->Field(i), "")) break;
                 }
                 if (i < pCard->NumField()) {
                      i = atoi(pCard->Field(1)) + 1;
                      pCard = _CardInfo[XFILES_TOKEN-1].
                               add_a_new_card(NdbToken[XFILES_TOKEN-1], 0);
                      pCard->AddElementToField(i, 1);
                 }
            }
       }

       CifString cs;
       _joint_methods.Clear();
       ReVarCifArray<CifString> data;
       for (int card_id = 0; card_id < MxTokens; card_id++) {
            if (!NdbToken[card_id].RefineField) continue;
            data.Clear();
            pCard = _CardInfo[card_id].Head();
            while (pCard) {
                 if (strcmp(pCard->Field(NdbToken[card_id].RefineField-1), "")) {
                      i = 0;
                      for (unsigned int j = 0; j < data.Length(); j++) {
                           if (data[j] == pCard->Field(NdbToken[card_id].RefineField-1)) {
                                i = 1;
                                break;
                           }
                      }
                      if (!i) {
                           cs.Copy(pCard->Field(NdbToken[card_id].RefineField-1));
                           data.Add(cs);
                      }
                 }
                 pCard = pCard->Next();
            }
            if (data.Length() > _joint_methods.Length()) {
                 _joint_methods.Clear();
                 for (unsigned int j = 0; j < data.Length(); j++) {
                      _joint_methods.Add(data[j]);
                 }
            }
       }

       group_remark = group_remarks[program];
       if (program == 5) {
            int card_id = find_card_id("NCSLOC");
            if (card_id >= 0 && _CardInfo[card_id].Head())
                 group_remark = &Remark_Refmac_5_0_local;
       }
       if (_joint_methods.Length() > 1) {
            for (unsigned int j = 0; j < _joint_methods.Length(); j++) {
                  _current_method.Copy(_joint_methods[j]);
                 for (i = 0; i < NUM_METHOD; i++) {
                      if (_current_method == _method_name[i]) {
                           newCard = _CardInfo[REMARK_TOKEN-1].
                                      add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                           newCard->AddElementToField(3, 1);
                           cs.Copy(" ");
                           cs += _method_remark[i];
                           newCard = _CardInfo[REMARK_TOKEN-1].
                                      add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                           newCard->AddElementToField(3, 1);
                           newCard->AddElementToField(cs.Text(), 2);
                           break;
                      }
                 }
                 for (i = 0; i < group_remark->Remarks_No; i++) {
                      if (strcmp(group_remark->TokenName[i], "")) {
                           card_id = find_card_id(group_remark->TokenName[i]);
                           if (card_id == -1) continue;
                           pCard = _CardInfo[card_id].Head();
                           while (pCard) {
                                ndb_to_pdb_get_general_remark(group_remark->num_remarks[i],
                                       group_remark->remarks[i], 0, 
                                       atoi(pCard->Field(group_remark->repeat[i]-1)), 0, 0);
                                pCard = pCard->Next();
                           }
                      } else if (group_remark->repeat[i] < 0)
                           ndb_to_pdb_get_general_remark(group_remark->num_remarks[i],
                                      group_remark->remarks[i], 1, 0, 0, 0);
                      else ndb_to_pdb_get_general_remark(group_remark->num_remarks[i],
                                      group_remark->remarks[i], 0, 0, 0, 0);
                 }
            }
       } else {
            _current_method.Clear();
            for (i = 0; i < group_remark->Remarks_No; i++) {
                 if (strcmp(group_remark->TokenName[i], "")) {
                      card_id = find_card_id(group_remark->TokenName[i]);
                      if (card_id == -1) continue;
                      pCard = _CardInfo[card_id].Head();
                      while (pCard) {
                           ndb_to_pdb_get_general_remark(group_remark->num_remarks[i],
                                  group_remark->remarks[i], 0, 
                                  atoi(pCard->Field(group_remark->repeat[i]-1)), 0, 0);
                           pCard = pCard->Next();
                      }
                 } else if (group_remark->repeat[i] < 0)
                      ndb_to_pdb_get_general_remark(group_remark->num_remarks[i],
                                 group_remark->remarks[i], 1, 0, 0, 0);
                 else ndb_to_pdb_get_general_remark(group_remark->num_remarks[i],
                                 group_remark->remarks[i], 0, 0, 0, 0);
            }
       }
}

void _MAXIT::ndb_to_pdb_get_general_remark(const int num_remarks, REMARKS *Remarks,
          const int repeat, const int index, const int IsInosine, const int space,
          const char *sctype)
{
       int i, j, card_id = -1, field_no = 1, found, *single = NULL;
       One_Card_Info_Format *pCard = NULL, *qCard = NULL, *newCard = NULL;

       for (i = 0; i < num_remarks; i++) {
            card_id = -1;
            for (j = 0; j < Remarks[i].NumField; j++) {
                 if (!strcmp(Remarks[i].FieldList[j].TokenName, "") && 
                              Remarks[i].FieldList[j].FieldId == 0 &&
                              Remarks[i].FieldList[j].FieldWidth == 0)
                      Remarks[i].FieldList[j].FieldWidth = 
                          strlen(Remarks[i].FieldList[j].Text);
                 if (strcmp(Remarks[i].FieldList[j].TokenName, "") && card_id == -1)
                      card_id = find_card_id(Remarks[i].FieldList[j].TokenName);
            }
       }

       if (Remarks == Remark_3_REFMAC_TWIN) {
            card_id = find_card_id("TWIN");
            pCard = _CardInfo[card_id].Head();
            if (pCard == NULL) return;
            newCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(Remarks->Remark_No, 1);
            for (i = 0; i < Num_Remark_3_REFMAC_TWIN_OP; i++) {
                 for (j = 0; j < Remark_3_REFMAC_TWIN_OP[i].NumField; j++) {
                      if (!strcmp(Remark_3_REFMAC_TWIN_OP[i].FieldList[j].TokenName, "") &&
                                   Remark_3_REFMAC_TWIN_OP[i].FieldList[j].FieldId == 0 &&
                                   Remark_3_REFMAC_TWIN_OP[i].FieldList[j].FieldWidth == 0)
                           Remark_3_REFMAC_TWIN_OP[i].FieldList[j].FieldWidth =
                               strlen(Remark_3_REFMAC_TWIN_OP[i].FieldList[j].Text);
                 }
            }
            for (i = 0; i < num_remarks; i++) {
                 ndb_to_pdb_get_general_remark(&Remark_3_REFMAC_TWIN[i], sctype);
            }
            while (pCard) {
                 for (i = 0; i < Num_Remark_3_REFMAC_TWIN_OP; i++) {
                      ndb_to_pdb_get_general_remark(&Remark_3_REFMAC_TWIN_OP[i], pCard);
                 }
                 pCard = pCard->Next();
            }
            return;
       }

       if (repeat && card_id != -1) {
            found = 0;
            pCard = _CardInfo[card_id].Head();
            while (pCard) {
                 if (IsInosine && (card_id == BSPAIR_TOKEN-1 &&
                     (!strcmp(pCard->Field(11), "WATSON-CRICK") ||
                      !strcmp(pCard->Field(5), "") ||
                      !strcmp(pCard->Field(9), "")) ||
                     card_id == MODIFB_TOKEN-1 &&
                     res_name_cmp(pCard->Field(5), "I")) ||
                    !IsInosine && card_id == MODIFB_TOKEN-1 &&
                    !res_name_cmp(pCard->Field(5), "I")) {
                      pCard = pCard->Next();
                      continue;
                 }
                 if (_current_method != "" && NdbToken[card_id].RefineField &&
                     _current_method != pCard->Field(NdbToken[card_id].RefineField-1)) {
                      pCard = pCard->Next();
                      continue;
                 }
                 found = 1;
                 break;
            }

            if (Remarks == Remark_3_PHENIX_TLS_GROUP) {}
            else if (found) {
                 newCard = _CardInfo[REMARK_TOKEN-1].
                           add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(Remarks->Remark_No, 1);
            }
            single = new int[num_remarks];
            for (i = 0; i < num_remarks; i++) single[i] = 1;
            pCard = _CardInfo[card_id].Head();
            while (pCard) {
                 if (IsInosine && (card_id == BSPAIR_TOKEN-1 &&
                     (!strcmp(pCard->Field(11), "WATSON-CRICK") ||
                      !strcmp(pCard->Field(5), "") ||
                      !strcmp(pCard->Field(9), "")) ||
                     card_id == MODIFB_TOKEN-1 &&
                     res_name_cmp(pCard->Field(5), "I")) ||
                    !IsInosine && card_id == MODIFB_TOKEN-1 &&
                    !res_name_cmp(pCard->Field(5), "I")) {
                      pCard = pCard->Next();
                      continue;
                 }
                 if (_current_method != "" && NdbToken[card_id].RefineField &&
                     _current_method != pCard->Field(NdbToken[card_id].RefineField-1)) {
                      pCard = pCard->Next();
                      continue;
                 }
                 if (space && pCard != _CardInfo[card_id].Head()) {
                      newCard = _CardInfo[REMARK_TOKEN-1].
                                add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                      newCard->AddElementToField(Remarks->Remark_No, 1);
                 }
                 for (i = 0; i < num_remarks; i++) {
                      if (Remarks[i].NumField == 1) {
                           if (single[i] || Remarks == Remark_265_2) {
                                ndb_to_pdb_get_general_remark(&Remarks[i], pCard);
                                single[i] = 0;
                           } else if (Remarks == Remark_800) {
                                newCard = _CardInfo[REMARK_TOKEN-1].
                                          add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                                newCard->AddElementToField(Remarks->Remark_No, 1);
                           }
                      } else ndb_to_pdb_get_general_remark(&Remarks[i], pCard);
                 }
                 pCard = pCard->Next();
            }
            delete [] single; single = NULL;
       } else {
            if (Remarks != Remark_3_PHENIX_TLS_GROUP) {
                 newCard = _CardInfo[REMARK_TOKEN-1].
                           add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(Remarks->Remark_No, 1);
            }
            for (i = 0; i < num_remarks; i++) {
                 card_id = -1;
                 for (j = 0; j < Remarks[i].NumField; j++) {
                      if (strcmp(Remarks[i].FieldList[j].TokenName, "")) {
                           card_id = find_card_id(Remarks[i].FieldList[j].TokenName);
                           if (!strcmp(Remarks[i].FieldList[j].TokenName, "NCSGRO"))
                                field_no = 4;
                           else field_no = 1;
                      }
                      if (card_id != -1) break;
                 }
                 if ((Remarks[i].Multiple_Treatment || index) && card_id != -1) {
                      pCard = _CardInfo[card_id].Head();
                      while (pCard) {
                           if (IsInosine && (card_id == BSPAIR_TOKEN-1 &&
                               (!strcmp(pCard->Field(11), "WATSON-CRICK") ||
                                !strcmp(pCard->Field(5), "") ||
                                !strcmp(pCard->Field(9), "")) ||
                               card_id == MODIFB_TOKEN-1 &&
                               res_name_cmp(pCard->Field(5), "I")) ||
                              !IsInosine && card_id == MODIFB_TOKEN-1 &&
                              !res_name_cmp(pCard->Field(5), "I")) {
                                pCard = pCard->Next();
                                continue;
                           }
                           if (_current_method != "" && NdbToken[card_id].RefineField &&
                               _current_method !=
                                       pCard->Field(NdbToken[card_id].RefineField-1)) {
                                pCard = pCard->Next();
                                continue;
                           }
                           if (!index || index && atoi(pCard->Field(field_no)) == index) {
                                ndb_to_pdb_get_general_remark(&Remarks[i], pCard);
                           }
                           pCard = pCard->Next();
                      }
                 } else ndb_to_pdb_get_general_remark(&Remarks[i], sctype);
            }
       }

       char buffer[80];
       if (Remarks == Remark_3_REFMAC_NCS_LOCAL) {
            int card_id = find_card_id("NCSLOC");
            pCard = NULL;
            if (card_id >= 0) pCard = _CardInfo[card_id].Head();
            while (pCard) {
                 sprintf(buffer, "%4s %6s %5s %5s %7s %5s %6s %8s %5s %5s",
                                 pCard->Field(1), pCard->Field(2), pCard->Field(3),
                                 pCard->Field(4), pCard->Field(5), pCard->Field(6),
                                 pCard->Field(7), pCard->Field(8), pCard->Field(9),
                                 pCard->Field(10));
                 newCard = _CardInfo[REMARK_TOKEN-1].
                           add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(Remarks->Remark_No, 1);
                 newCard->AddElementToField(buffer, 2);
                 pCard = pCard->Next();
            }
       }

       if (Remarks == Remark_3_Phenix_BIN) {
            _CardInfo[SHELL_TOKEN-1].reorder_cards(1, FLOAT_COMPARE, DESCENT);
            pCard = _CardInfo[SHELL_TOKEN-1].Head();
            int count = 0;
            while (pCard) {
                 if (_current_method != "" && NdbToken[SHELL_TOKEN-1].RefineField &&
                     _current_method !=
                            pCard->Field(NdbToken[SHELL_TOKEN-1].RefineField-1)) {
                      pCard = pCard->Next();
                      continue;
                 }
                 count++;
                 sprintf(buffer, "%5d%8.4f -%8.4f    %4.2f   %6d %5d  %6.4f %6.4f",
                          count, atof(pCard->Field(2)), atof(pCard->Field(1)),
                          atof(pCard->Field(3)) / 100.0,
                          atoi(pCard->Field(4)), atoi(pCard->Field(8)),
                          atof(pCard->Field(5)), atof(pCard->Field(6)));
                 newCard = _CardInfo[REMARK_TOKEN-1].
                           add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(Remarks->Remark_No, 1);
                 newCard->AddElementToField(buffer, 2);
                 pCard = pCard->Next();
            }
       }

       if (Remarks == Remark_3_PHENIX_NCS) {
            for (j = 0; j < Remark_3_PHENIX_NCS_GROUP.NumField; j++) {
                 if (!strcmp(Remark_3_PHENIX_NCS_GROUP.FieldList[j].TokenName, "") &&
                              Remark_3_PHENIX_NCS_GROUP.FieldList[j].FieldId == 0 &&
                              Remark_3_PHENIX_NCS_GROUP.FieldList[j].FieldWidth == 0)
                      Remark_3_PHENIX_NCS_GROUP.FieldList[j].FieldWidth =
                          strlen(Remark_3_PHENIX_NCS_GROUP.FieldList[j].Text);
            }
            for (i = 0; i < Num_Remark_3_PHENIX_NCS_GROUP_OP; i++) {
                 for (j = 0; j < Remark_3_PHENIX_NCS_GROUP_OP[i].NumField; j++) {
                      if (!strcmp(Remark_3_PHENIX_NCS_GROUP_OP[i].FieldList[j].TokenName, "") &&
                                   Remark_3_PHENIX_NCS_GROUP_OP[i].FieldList[j].FieldId == 0 &&
                                   Remark_3_PHENIX_NCS_GROUP_OP[i].FieldList[j].FieldWidth == 0)
                           Remark_3_PHENIX_NCS_GROUP_OP[i].FieldList[j].FieldWidth =
                               strlen(Remark_3_PHENIX_NCS_GROUP_OP[i].FieldList[j].Text);
                 }
            }
            card_id = find_card_id("PHENCS");
            pCard = _CardInfo[NCSGRO_TOKEN-1].Head();
            while (pCard) {
                 ndb_to_pdb_get_general_remark(&Remark_3_PHENIX_NCS_GROUP, pCard);
                 qCard = _CardInfo[card_id].Head();
                 while (qCard) {
                      if (!strcmp(qCard->Field(1), pCard->Field(4))) {
                           for (i = 0; i < Num_Remark_3_PHENIX_NCS_GROUP_OP; i++) {
                                ndb_to_pdb_get_general_remark(&Remark_3_PHENIX_NCS_GROUP_OP[i], qCard);
                           }
                      }
                      qCard = qCard->Next();
                 }
                 pCard = pCard->Next();
            }
       }
}

void _MAXIT::ndb_to_pdb_get_general_remark(REMARKS *Remark, const char *sctype)
{
       int i, card_id, field_no, align, serial_no_field;
       char diffrn_id[100], bigstring[100000], FieldFormat[80], buffer[100000];
       double temp_double;
       One_Card_Info_Format *newCard = NULL, *pCard = NULL;

       memset(diffrn_id, 0, 100);
       card_id = find_card_id("SCTYPE");
       if (card_id >= 0 && strcmp(sctype, "")) {
            pCard = _CardInfo[card_id].Head();
            while (pCard) {
                 if (!strcmp(pCard->Field(2), sctype)) { 
                      strcpy(diffrn_id, pCard->Field(1));
                      break;
                 }
                 pCard = pCard->Next();
            }
       }

       int is_skip_remark = 0;
       int is_null_value = 1;
       for (i = 0; i < Remark->NumField; i++) {
            if (!strcmp(Remark->FieldList[i].Text, "B VALUE TYPE :")) {
                 is_skip_remark = 1;
            }
            if (strcmp(Remark->FieldList[i].TokenName, "") &&
                Remark->FieldList[i].FieldId) {
                 card_id = find_card_id(Remark->FieldList[i].TokenName);
                 field_no = Remark->FieldList[i].FieldId-1;
                 if (card_id != -1) {
                      pCard = _CardInfo[card_id].Head();
                      if (pCard && strcmp(pCard->Field(field_no), "")) {
                           is_null_value = 0;
                      }
                 }
            }
       }
       if (is_skip_remark && is_null_value) return;

       newCard = _CardInfo[REMARK_TOKEN-1].
                    add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
       newCard->AddElementToField(Remark->Remark_No, 1);

       int found_other_remark = 0;
       for (i = 0; i < Remark->NumField; i++) {
            int nmr_remark = 0;
            if (!strcmp(Remark->FieldList[i].Text, "OTHER REFINEMENT REMARKS:")) {
                 found_other_remark = 1;
            }
            if (!strcmp(Remark->FieldList[i].TokenName, "") &&
                !strcmp(Remark->FieldList[i].Text, "")) continue;

            align = check_align(Remark->Remark_No, &(Remark->FieldList[i]));
            get_remark_field_format(FieldFormat, Remark, i);
            memset(bigstring, 0, 100000);

            if (strcmp(Remark->FieldList[i].TokenName, "") &&
                Remark->FieldList[i].FieldId) {
                 if (!strcmp(Remark->FieldList[i].TokenName, "REFREM") &&
                      found_other_remark) nmr_remark = 1;

                 card_id = find_card_id(Remark->FieldList[i].TokenName);
                 field_no = Remark->FieldList[i].FieldId-1;
                 pCard = NULL;
                 serial_no_field = 0;
                 if (card_id != -1) {
                      if (!strcmp(Remark->FieldList[i].TokenName, "DIFPTL") ||
                          !strcmp(Remark->FieldList[i].TokenName, "DTMEAS") ||
                          !strcmp(Remark->FieldList[i].TokenName, "DTMETH") ||
                          !strcmp(Remark->FieldList[i].TokenName, "DTTEMP") ||
                          !strcmp(Remark->FieldList[i].TokenName, "DTWAVE") ||
                          !strcmp(Remark->FieldList[i].TokenName, "PERCOM") ||
                          !strcmp(Remark->FieldList[i].TokenName, "PHVAL") ||
                          !strcmp(Remark->FieldList[i].TokenName, "RADIAT") ||
                          !strcmp(Remark->FieldList[i].TokenName, "RESTOT") ||
                          !strcmp(Remark->FieldList[i].TokenName, "REFLEC") ||
                          !strcmp(Remark->FieldList[i].TokenName, "SHELC") ||
                          !strcmp(Remark->FieldList[i].TokenName, "WAVLEN")) {
                           serial_no_field = NdbToken[card_id].SeqField - 1;
                           if (card_id == find_card_id("DTWAVE")) serial_no_field = 2;
                      }
                      pCard = _CardInfo[card_id].Head();
                      if (_current_method != "" && NdbToken[card_id].RefineField) {
                           while (pCard) {
                                if (_current_method != pCard->Field(NdbToken[card_id].
                                                 RefineField-1)) {
                                     pCard = pCard->Next();
                                     continue;
                                }
                                break;
                           }
                      }
                      if (strcmp(sctype, "")) {
                           while (pCard) {
                                if (!strcmp(Remark->FieldList[i].TokenName, "EXPDTA")) {
                                     if (strcmp(pCard->Field(2), sctype)) {
                                          pCard = pCard->Next();
                                          continue;
                                     }
                                } else if (!strcmp(Remark->FieldList[i].TokenName,
                                                   "CRMETH")) {
                                     if (strcmp(pCard->Field(5), sctype)) {
                                          pCard = pCard->Next();
                                          continue;
                                     }
                                } else if (serial_no_field > 0) {
                                     // if (strcmp(pCard->Field(serial_no_field), diffrn_id)) {
                                     if (!is_same_diffrn(pCard->Field(serial_no_field),
                                                         diffrn_id)) {
                                          pCard = pCard->Next();
                                          continue;
                                     }
                                }
                                break;
                           }
                      }
                 }
                 if (pCard) {
                      if (Remark->Remark_No == 245) {
                          while (pCard) {
                               if (strcmp(pCard->Field(field_no), "")) {
                                     if (!NdbToken[card_id].FieldList[field_no].MixedCase)
                                          pCard->UpperCaseField(field_no);
                                     if (Remark->FieldList[i].FieldType < 3) {
                                          int len = Remark->FieldList[i].FieldWidth;
                                          if (i == Remark->NumField - 1) {
                                               int len1 = 59 - Remark->FieldList[i].FieldStCol;
                                               if (len1 > len) len = len1;
                                          }
                                          if (_remark_decimal_precision == UNFORCE_DECIMAL_PRECISION &&
                                             (int) strlen(pCard->Field(field_no)) <= len) {
                                               if (Remark->FieldList[i].FieldJustification ==
                                                     PDB_REMARK_LEFT_JUSTIFIED)
                                                    sprintf(FieldFormat, "%%-%ds",
                                                        Remark->FieldList[i].FieldWidth);
                                               else sprintf(FieldFormat, "%%%ds",
                                                        Remark->FieldList[i].FieldWidth);
                                               sprintf(buffer, FieldFormat, pCard->Field(field_no));
                                          } else {
                                               temp_double = atof(pCard->Field(field_no));
/*
                                               if (Remark->Special_Treatment == PDB_REMARK_R_VALUE)
                                                    temp_double = temp_double / 100.0;
*/
                                               sprintf(buffer, FieldFormat, temp_double);
                                          }
                                     } else {
                                          sprintf(buffer, FieldFormat, pCard->Field(field_no));
                                     }
                                     ndb_clean_string(buffer);
                                     if (strcmp(bigstring, "")) strcat(bigstring, "; ");
                                     strcat(bigstring, buffer);
                               }
                               pCard = pCard->Next();
                          }
                      } else if (pCard->Next() && card_id == XFILES_TOKEN-1) {
/*
                         (card_id == EXPDTA_TOKEN-1 || card_id == XFILES_TOKEN-1 ||
                          card_id == find_card_id("DIFPTL") ||
                          card_id == DTMEAS_TOKEN-1 || card_id == PHVAL_TOKEN-1 ||
                          card_id == CRMETH_TOKEN-1 || card_id == DTMETH_TOKEN-1 || 
                          card_id == RADIAT_TOKEN-1 || card_id == DTWAVE_TOKEN-1 ||
                          card_id == WAVLEN_TOKEN-1 || card_id == DTTEMP_TOKEN-1)) {
*/
                           serial_no_field = NdbToken[card_id].SeqField - 1;
                           // if (card_id == find_card_id("DTWAVE")) serial_no_field = 2;
                           while (pCard) {
                                if (_current_method != "" &&
                                    NdbToken[card_id].RefineField && _current_method !=
                                    pCard->Field(NdbToken[card_id].RefineField-1)) {
                                     pCard = pCard->Next();
                                     continue;
                                }
                                if (strcmp(pCard->Field(field_no), "")) {
                                     if (!NdbToken[card_id].FieldList[field_no].MixedCase)
                                          pCard->UpperCaseField(field_no);
                                     strcat(bigstring, pCard->Field(field_no));
                                } else if (!strcmp(bigstring, ""))
                                     strcat(bigstring, "NULL");
                                if (pCard->Next() && strcmp(pCard->Next()->Field(field_no), "")) {
                                     if (serial_no_field > 0 &&
                                        !strcmp(pCard->Field(serial_no_field),
                                           pCard->Next()->Field(serial_no_field)))
                                          strcat(bigstring, ", ");
                                     else strcat(bigstring, "; ");
                                }
                                pCard = pCard->Next();
                           }
                      } else if (strcmp(pCard->Field(field_no), "")) {
                           if (!NdbToken[card_id].FieldList[field_no].MixedCase)
                                pCard->UpperCaseField(field_no);
                           if (Remark->FieldList[i].FieldType < 3) {
                                if (strstr(pCard->Field(field_no), "NULL") ||
                                    strstr(pCard->Field(field_no), ";")) {
                                     strcpy(bigstring, pCard->Field(field_no));
                                } else {
                                     int len = Remark->FieldList[i].FieldWidth;
                                     if (i == Remark->NumField - 1) {
                                          int len1 = 59 - Remark->FieldList[i].FieldStCol;
                                          if (len1 > len) len = len1;
                                     }
                                     if (_remark_decimal_precision == UNFORCE_DECIMAL_PRECISION &&
                                        (int) strlen(pCard->Field(field_no)) <= len) {
                                          if (Remark->FieldList[i].FieldJustification ==
                                              PDB_REMARK_LEFT_JUSTIFIED)
                                               sprintf(FieldFormat, "%%-%ds",
                                                  Remark->FieldList[i].FieldWidth);
                                          else sprintf(FieldFormat, "%%%ds",
                                                  Remark->FieldList[i].FieldWidth);
                                          sprintf(bigstring, FieldFormat, pCard->Field(field_no));
                                     } else {
                                          temp_double = atof(pCard->Field(field_no));
/*
                                          if (Remark->Special_Treatment == PDB_REMARK_R_VALUE)
                                               temp_double = temp_double / 100.0;
*/
                                          sprintf(bigstring, FieldFormat, temp_double);
                                     }
                                }
                           } else {
                                if (card_id != HEADER_TOKEN-1 || (field_no+1) != 2)
                                     sprintf(bigstring, FieldFormat, pCard->Field(field_no));
                           }
                      }
                 }
            } else sprintf(bigstring, FieldFormat, Remark->FieldList[i].Text);

            ndb_to_pdb_print_remark_value(bigstring, newCard, Remark, i, align,
                            nmr_remark);
       }
}

void _MAXIT::ndb_to_pdb_get_general_remark(REMARKS *Remark, One_Card_Info_Format *pCard)
{
       int i, card_id, field_no, align;
       char bigstring[10000], FieldFormat[80];
       double temp_double;
       One_Card_Info_Format *newCard = NULL;

       newCard = _CardInfo[REMARK_TOKEN-1].
                    add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
       newCard->AddElementToField(Remark->Remark_No, 1);
       for (i = 0; i < Remark->NumField; i++) {
            get_remark_field_format(FieldFormat, Remark, i);
            memset(bigstring, 0, 10000);

            align = check_align(Remark->Remark_No, &(Remark->FieldList[i]));

            if (strcmp(Remark->FieldList[i].TokenName, "") &&
                Remark->FieldList[i].FieldId) {
                 card_id = find_card_id(Remark->FieldList[i].TokenName);
                 field_no = Remark->FieldList[i].FieldId-1;
                 if (strcmp(pCard->Field(field_no), "")) {
                      if (!NdbToken[card_id].FieldList[field_no].MixedCase)
                           pCard->UpperCaseField(field_no);
                      if (Remark->FieldList[i].FieldType < 3) {
                           if ((int) strlen(pCard->Field(field_no)) <=
                               Remark->FieldList[i].FieldWidth) {
                                if (Remark->FieldList[i].FieldJustification ==
                                      PDB_REMARK_LEFT_JUSTIFIED)
                                     sprintf(FieldFormat, "%%-%ds",
                                         Remark->FieldList[i].FieldWidth);
                                else sprintf(FieldFormat, "%%%ds",
                                         Remark->FieldList[i].FieldWidth);
                                sprintf(bigstring, FieldFormat, pCard->Field(field_no));
                           } else {
                                temp_double = atof(pCard->Field(field_no));
                                sprintf(bigstring, FieldFormat, temp_double);
                           }
                      } else sprintf(bigstring, FieldFormat, pCard->Field(field_no));
                 }
            } else sprintf(bigstring, FieldFormat, Remark->FieldList[i].Text);
            ndb_to_pdb_print_remark_value(bigstring, newCard, Remark, i, align, 0);
       }
}

void _MAXIT::ndb_to_pdb_print_remark_value(char *value, One_Card_Info_Format *newCard,
      REMARKS *Remark, const int field_no, const int align, const int remark_details)
{
       int str_len, FieldLen, length, max_length, k, align_tmp, tmp_len;
       char *p = NULL, buffer[80], *tmp, *buff = NULL;

       print_space_between_remark_field(newCard, Remark, field_no);
       str_len = strlen(newCard->Field(2));
       FieldLen = NdbToken[REMARK_TOKEN-1].FieldList[2].FieldWidth + 1;
       if (strcmp(value, "")) {
            ReVarCifArray<CifString> remark_array, data;
            remark_array.Clear();
            if (remark_details) {
                 get_wordlist_from_string_separated_by_delimits_with_empty_line(
                            remark_array, value, "\n");
            }

            if (remark_array.Length() > 1) {
                 for (unsigned int i = 0; i < remark_array.Length(); ++i) {
                      if (FieldLen > (int) remark_array[i].Length() + 1) {
                           newCard = _CardInfo[REMARK_TOKEN-1].
                               add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                           newCard->AddElementToField(Remark->Remark_No, 1);
                           newCard->AddElementToField(" ", 2);
                           if (remark_array[i] != "")
                                newCard->AttachToField(remark_array[i].Text(), 2);
                      } else {
                           strcpy(value, remark_array[i].Text());
                           get_remark_array_from_block(data, value, FieldLen - 1);
                           for (unsigned int j = 0; j < data.Length(); ++j) {
                                newCard = _CardInfo[REMARK_TOKEN-1].
                                    add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                                newCard->AddElementToField(Remark->Remark_No, 1);
                                newCard->AddElementToField(" ", 2);
                                newCard->AttachToField(data[j].Text(), 2);
                           }
                      }
                 }
            } else if (FieldLen > ((int) strlen(value) + str_len + 1))
                 newCard->AttachToField(value, 2);
            else {
                 max_length = FieldLen - (str_len + 2);
                 p = get_max_length_word(value, buffer, max_length);
                 newCard->AttachToField(buffer, 2);
                 length = strlen(Remark->FieldList[0].Text) +
                                 Remark->FieldList[0].FieldStCol;
                 align_tmp = 0;
                 if (align) {
                      max_length = FieldLen - length - 1;
                      align_tmp = 1;
                      if (p) {
                           tmp = p;
                           buff = new char[strlen(tmp)+1]; 
                           tmp_len = 0;
                           while (tmp) {
                                tmp = get_word(tmp, buff);
                                if ((int) strlen(buff) > tmp_len)
                                     tmp_len = strlen(buff);
                           }
                           delete [] buff;
                           if (tmp_len > (max_length - 1)) {
                                max_length = FieldLen - 3;
                                align_tmp = 0;
                           }
                      }
                 } else max_length = FieldLen - 3;
                 while (p) {
                      p = get_max_length_word(p, buffer, max_length);
                      newCard = _CardInfo[REMARK_TOKEN-1].
                          add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                      newCard->AddElementToField(Remark->Remark_No, 1);
                      if (align_tmp) {
                           for (k = 0; k < length; k++)
                                newCard->AttachToField(" ", 2);
                           newCard->AttachToField(buffer, 2);
                      } else {
                           newCard->AddElementToField(" ", 2);
                           newCard->AttachToField(buffer, 2);
                      }
                 }
            }
       } else print_null_remark_value(newCard, Remark->FieldList[field_no].FieldWidth,
                            Remark->FieldList[field_no].FieldJustification);

       if (Remark->Remark_No > 99 && Remark->Remark_No != 103 &&
           Remark->Remark_No != 105 && Remark->Remark_No != 106 &&
           field_no == Remark->NumField-1 && Remark->Remark_No != 200 && 
           Remark->Remark_No != 205 &&
           Remark->Remark_No != 210 && Remark->Remark_No != 215 &&
           Remark->Remark_No != 217 && Remark->Remark_No != 220 &&
           Remark->Remark_No != 225 && Remark->Remark_No != 230 &&
           Remark->Remark_No != 240 && Remark->Remark_No != 245 &&
           Remark->Remark_No != 247 && Remark->Remark_No != 265 &&
           Remark->Remark_No != 280 && Remark->Remark_No != 300 &&
           Remark->Remark_No != 350 && Remark->Remark_No != 375 &&
           Remark->Remark_No != 465 && Remark->Remark_No != 470 &&
           Remark->Remark_No != 475 && Remark->Remark_No != 480 &&
           Remark->Remark_No != 500 && Remark->Remark_No != 525 &&
           Remark->Remark_No != 610 && Remark->Remark_No != 615 &&
           Remark->Remark_No != 620 && Remark->Remark_No != 630 &&
           Remark->Remark_No != 800 && Remark->Remark_No != 900) {
            strcpy(buffer, newCard->Field(2));
            ndb_strip_trailing_blanks(buffer);
            strcat(buffer, ".");
            newCard->AddElementToField(buffer, 2);
       }
}

int _MAXIT::ndb_to_pdb_get_remark_from_premark(const int remarkNo)
{
       int exist;
       char line[100], *p;
       One_Card_Info_Format *pCard = NULL, *newCard = NULL;

       if (remarkNo == 550) return 0;
       pCard = _CardInfo[PREMRK_TOKEN-1].Head();
       while (pCard) {
            if (atoi(pCard->Field(1)) == remarkNo) break;
            pCard = pCard->Next();
       }
       if (pCard == NULL) return 0;

       if (remarkNo == 5) {
            newCard = pCard;
            exist = 0;
            while (newCard && atoi(newCard->Field(1))  == remarkNo) {
                 if (strstr(newCard->Field(2), "LAYER 1 RELEASE")) {
                      exist = 1;
                      break;
                 }
                 newCard = newCard->Next();
            }
            if (exist) return 0;
       }

       while (pCard) {
            if (atoi(pCard->Field(1)) == remarkNo) {
                 newCard = _CardInfo[REMARK_TOKEN-1].
                         add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(remarkNo, 1);
                 strcpy(line, pCard->Field(2));
                 p = strstr(line, "N/A");
                 if (p) sprintf(p, "%s", "NULL");
/*
                 p = strstr(line, " NONE");
                 if (p) sprintf(p, "%s", " NULL");
                 p = strstr(line, ":NONE");
                 if (p) sprintf(p, "%s", ": NULL");
*/
                 if (remarkNo != 350 && remarkNo != 400 && remarkNo != 465 &&
                     remarkNo != 470 && remarkNo != 500)
                      ndb_set_string_to_upper(line);
                 newCard->AddElementToField(line, 2);
            }
            pCard = pCard->Next();
       }
       return 1;
}

void _MAXIT::ndb_to_pdb_get_remark_280_crystal() 
{
       int card_id, len;
       char buffer[5000];
       One_Card_Info_Format *pCard = NULL, *qCard = NULL;

       card_id = find_card_id("CRDTLS");
       if (_CardInfo[card_id].Head() == NULL && _CardInfo[CRSOLU_TOKEN-1].Head()) {
            pCard = _CardInfo[CRSOLU_TOKEN-1].Head();
            memset(buffer, 0, 5000);
            while (pCard) {
                 if (!strcmp(pCard->Field(3), "1") &&
                      strcasecmp(pCard->Field(4), "WATER")) {
                       len = strlen(buffer);
                       if (len) sprintf(&buffer[len], ", %s", pCard->Field(4));
                       else strcpy(buffer, pCard->Field(4));
                 }
                 pCard = pCard->Next();
            }
            if (strcmp(buffer, "")) {
                 pCard = _CardInfo[card_id].add_a_new_card(NdbToken[card_id], 0);
                 pCard->AddElementToField(buffer, 2);
            }
       }
       pCard = _CardInfo[card_id].Head();
       if (pCard) {
            qCard = pCard->Next();
            while (qCard) {
                 if (strcmp(qCard->Field(2), "")) {
                      if (strcmp(pCard->Field(2), "")) {
                           len = strlen(pCard->Field(2));
                           if (pCard->Field(2)[len - 1] == '.')
                                pCard->AttachToField(" ", 2);
                           else pCard->AttachToField(". ", 2);
                      }
                      pCard->AttachToField(qCard->Field(2), 2);
                 }
                 qCard = qCard->Next();
            }
       }
       ndb_to_pdb_get_general_remark(Num_Remark_280, Remark_280, 0, 0, 0, 0);
}

void _MAXIT::ndb_to_pdb_get_remark_290()
{
       int i, j, k, lineNo, card_id, field_no;
       double result[3][3], tr[3];
       char FieldFormat[80], tmpString[80], buffer[80];
       One_Card_Info_Format *newCard = NULL, *pCard = NULL;
       NDBCELLCONST *cell = _cell;

       if (cell == NULL) return;

       for (j = 0; j < Num_Remark_290; j++) {
            for (i = 0; i < Remark_290[j].NumField; i++) 
                 if (!strcmp(Remark_290[j].FieldList[i].TokenName, "") && 
                             Remark_290[j].FieldList[i].FieldId == 0 &&
                             Remark_290[j].FieldList[i].FieldWidth == 0)
                      Remark_290[j].FieldList[i].FieldWidth = 
                             strlen(Remark_290[j].FieldList[i].Text);

            lineNo = 1;
            if (Remark_290[j].Special_Treatment == PDB_REMARK_NNNMMM_SYMOP)
                 lineNo = cell->nop;
            else if (Remark_290[j].Special_Treatment == PDB_REMARK_XYZT) 
                 lineNo = cell->nop * 3;
      
            for (k = 0; k < lineNo; k++) {
                 newCard = _CardInfo[REMARK_TOKEN-1].
                            add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(Remark_290[j].Remark_No, 1);
                 if (Remark_290[j].Special_Treatment == PDB_REMARK_XYZT &&
                     k % 3 == 0)
                      ndb_get_symmetry_matrix(cell, k/3, 0, 0, 0, result, tr);
                 for (i = 0; i < Remark_290[j].NumField; i++) {
                      print_space_between_remark_field(newCard, &Remark_290[j], i);
                      get_remark_field_format(FieldFormat, &Remark_290[j], i);
                      if (strcmp(Remark_290[j].FieldList[i].TokenName,"") ||
                                 Remark_290[j].FieldList[i].FieldId != 0) {
                           card_id = find_card_id(Remark_290[j].FieldList[i].TokenName);
                           field_no = Remark_290[j].FieldList[i].FieldId-1;

                           if (!strcmp(Remark_290[j].FieldList[i].Text,"NNNMMM")) {
                                sprintf(tmpString, "%d%d%d%d", k+1, 5, 5, 5);
                                sprintf(buffer, FieldFormat, tmpString);
                                newCard->AttachToField(buffer, 2);
                           } else if (!strcmp(Remark_290[j].FieldList[i].Text,
                                         "SYMOP")) {
                                memset(tmpString, 0, MxStrLen);
                                ndb_get_symmetry_operation_name(cell, k, 0, 0, 0,
                                             tmpString);
                                ndb_set_string_to_upper(tmpString);
                                newCard->AttachToField(tmpString, 2);
                           } else if (!strcmp(Remark_290[j].FieldList[i].Text,
                                           "SMTRY")) {
                                sprintf(tmpString, "%s%d",
                                       Remark_290[j].FieldList[i].Text, k % 3 + 1);
                                sprintf(buffer, FieldFormat, tmpString);
                                newCard->AttachToField(buffer, 2);
                           } else if (!strcmp(Remark_290[j].FieldList[i].Text, "Serial_No")) {
                                sprintf(buffer, FieldFormat, (double) (k / 3 + 1));
                                newCard->AttachToField(buffer, 2);
                           } else if (!strcmp(Remark_290[j].FieldList[i].Text, "X")) { 
                                sprintf(buffer, FieldFormat, result[k%3][0]);
                                newCard->AttachToField(buffer, 2);
                           } else if (!strcmp(Remark_290[j].FieldList[i].Text, "Y")) {
                                sprintf(buffer, FieldFormat, result[k%3][1]);
                                newCard->AttachToField(buffer, 2);
                           } else if (!strcmp(Remark_290[j].FieldList[i].Text, "Z")) {
                                sprintf(buffer, FieldFormat, result[k%3][2]);
                                newCard->AttachToField(buffer, 2);
                           } else if (!strcmp(Remark_290[j].FieldList[i].Text, "T")) {
                                sprintf(buffer, FieldFormat, tr[k%3]);
                                newCard->AttachToField(buffer, 2);
                           } else if (card_id == -1 || card_id > MxTokens ||
                              field_no < 0 || field_no >  NdbToken[card_id].NumField)
                                print_null_remark_value(newCard, 
                                   Remark_290[j].FieldList[i].FieldWidth,
                                   Remark_290[j].FieldList[i].FieldJustification);
                           else {
                                pCard = _CardInfo[card_id].Head();
                                if (pCard && (strcmp(pCard->Field(field_no), "") ||
                                    !strcmp(Remark_290[j].FieldList[i].TokenName,
                                       "CRYST1") && strcmp(pCard->Field(7), ""))) {
                                     if (!strcmp(pCard->Field(field_no), "") &&
                                        !strcmp(Remark_290[j].FieldList[i].TokenName,
                                         "CRYST1") && strcmp(pCard->Field(7), ""))
                                          field_no = 7;
                                     sprintf(buffer, FieldFormat, pCard->Field(field_no));
                                     if (!NdbToken[card_id].FieldList[field_no].MixedCase)
                                          ndb_set_string_to_upper(buffer);					      
                                     newCard->AttachToField(buffer, 2);
                                } else print_null_remark_value(newCard, 
                                    Remark_290[j].FieldList[i].FieldWidth,
                                    Remark_290[j].FieldList[i].FieldJustification);
                           }
                      } else {
                           sprintf(buffer, FieldFormat, Remark_290[j].FieldList[i].Text);
                           newCard->AttachToField(buffer, 2);
                      }
                 }
            }
       }
}

int _MAXIT::ndb_to_pdb_get_remark_300()
{
       int num_chn = 0, card_id, n_biol = 0, i = 0;
       _BIOL_ID BIOL[200];
       _CHAIN *chain = NULL;
       One_Card_Info_Format *pCard = NULL, *stBiolCard = NULL;
       NDBCELLCONST *cell = _cell;

       if (_num_mol == 0 || _molecules == NULL) return 0;

       if (cell == NULL) return 0;

       card_id = find_card_id("CHAINS");
       pCard = _CardInfo[card_id].add_a_new_card(NdbToken[card_id], 0);
       chain = _molecules[0]->GetFirstChain();
       while (chain) {
            if (chain->chain_type() == ATOMP_TOKEN ||
                chain->chain_type() == ATOMN_TOKEN) num_chn++;
            chain = _molecules[0]->GetNextChain();
       }
       if (num_chn) pCard->AddElementToField(num_chn, 2);
/*
       if (check_virus_ncs_remark_300()) {
            pCard = _CardInfo[card_id].Head();
            while (pCard) pCard = _CardInfo[card_id].delete_a_card(pCard);
            return 1;
       }
*/
       stBiolCard = _CardInfo[STBIOL_TOKEN-1].Head();
       while (stBiolCard) {
            for (i = 0; i < n_biol; i++) {
                 if (!strcmp(BIOL[i].biol_id, stBiolCard->Field(2))) break;
            }
            if (i == n_biol) {
                 strcpy(BIOL[i].biol_id, stBiolCard->Field(2));
                 n_biol++;
            }
            stBiolCard = stBiolCard->Next();
       }
       if (n_biol > 1) qsort((void *) BIOL, n_biol, sizeof(_BIOL_ID), biol_id_compare);

       for (i = 0; i < n_biol; i++) {
            if (i == 0) pCard->AddElementToField(BIOL[i].biol_id, 3);
            else {
                 pCard->AttachToField(", ", 3);
                 pCard->AttachToField(BIOL[i].biol_id, 3);
            }
       }

       if (_experiment_type & EXPERIMENT_TYPE_XRAY ||
           _experiment_type & EXPERIMENT_TYPE_NEUTRON)
            ndb_to_pdb_get_general_remark(Num_Remark_300, Remark_300, 0, 0, 0, 0);
            // ndb_to_pdb_get_general_remark(Num_Remark_300_new, Remark_300_new, 0, 0, 0, 0);
       else ndb_to_pdb_get_general_remark(Num_Remark_300_NMR, Remark_300_NMR, 0, 0, 0, 0);

       pCard = _CardInfo[card_id].Head();
       while (pCard) pCard = _CardInfo[card_id].delete_a_card(pCard);

       return 1;
}

int _MAXIT::ndb_to_pdb_get_remark_350()
{
       typedef struct {
              char asym_id[3];
              char pdb_id[500];
              int  n_symmetry;
              char symmetry[100][10];
              int  label;
       } BIOL_UNIT;

       BIOL_UNIT biol_unit[5000];
       int n_biol = 0, i, j, k, l, t, len, serialNo, n_unit;
       int op1, op2, opno, lx, ly, lz, field_no;
       char FieldFormat[80], buffer[100], *p;
       _BIOL_ID BIOL[200];
       double result[3][3], tr[3];
       One_Card_Info_Format *stBiolCard = NULL, *pCard = NULL, *newCard = NULL;
       NDBCELLCONST *cell = _cell;

       if (_num_mol == 0 || _molecules == NULL) return 0;

       if (cell == NULL) return 0;

       if (check_virus_ncs_remark_350()) return 1;

       stBiolCard = _CardInfo[STBIOL_TOKEN-1].Head();
       while (stBiolCard) {
            for (i = 0; i < n_biol; i++) {
                 if (!strcmp(BIOL[i].biol_id, stBiolCard->Field(2))) break;
            }
            if (i == n_biol) {
                 strcpy(BIOL[i].biol_id, stBiolCard->Field(2));
                 n_biol++;
            }
            stBiolCard = stBiolCard->Next();
       }
       if (n_biol > 1) qsort((void *) BIOL, n_biol, sizeof(_BIOL_ID), biol_id_compare);

       ndb_to_pdb_get_general_remark(Num_Remark_350, Remark_350, 0, 0, 0, 0);

       serialNo = 0;
       for (l = 0; l < n_biol; l++) {
            n_unit = 0;
            stBiolCard = _CardInfo[STBIOL_TOKEN-1].Head();
            while (stBiolCard) {
                 if (!strcmp(stBiolCard->Field(2), BIOL[l].biol_id)) {
                      if (!strcmp(stBiolCard->Field(4), NDB_BASE_SYMMETRY) &&
                           strcmp(stBiolCard->Field(13), "")) {
                           strcpy(biol_unit[n_unit].asym_id, stBiolCard->Field(3));
                           strcpy(biol_unit[n_unit].pdb_id, stBiolCard->Field(13));
                           strcpy(biol_unit[n_unit].symmetry[0], stBiolCard->Field(4));
                           biol_unit[n_unit].n_symmetry = 1;
                           biol_unit[n_unit].label = 1;
                           pCard = stBiolCard->Next();
                           while (pCard) {
                                if (strcmp(pCard->Field(4), NDB_BASE_SYMMETRY) &&
                                   !strcmp(pCard->Field(3), stBiolCard->Field(3))) {
                                     for (i = 0; i < biol_unit[n_unit].n_symmetry; i++)
                                          if (!strcmp(biol_unit[n_unit].symmetry[i],
                                                    pCard->Field(4))) break;
                                     if (i == biol_unit[n_unit].n_symmetry) {
                                          op1 = atoi(pCard->Field(4));
                                          for (i = biol_unit[n_unit].n_symmetry - 1;
                                                 i >= 0; i--) {
                                               op2 = atoi(biol_unit[n_unit].symmetry[i]);
                                               if (op1 > op2) break;
                                          }
                                          for (j = biol_unit[n_unit].n_symmetry - 1;
                                                   j >= (i + 1); j--) {
                                               strcpy(biol_unit[n_unit].symmetry[j+1],
                                                      biol_unit[n_unit].symmetry[j]);
                                          }
                                          strcpy(biol_unit[n_unit].symmetry[i+1],
                                                 pCard->Field(4));
                                          biol_unit[n_unit].n_symmetry++;
                                     }
                                }
                                pCard = pCard->Next();
                           }
                           n_unit++;
                      }
                 }
                 stBiolCard = stBiolCard->Next();
            }

            for (i = 0; i < n_unit - 1; i++) {
                 if (!biol_unit[i].label) continue;
                 for (j = i + 1; j < n_unit; j++) {
                      if (!biol_unit[j].label) continue;
                      if (biol_unit[i].n_symmetry != biol_unit[j].n_symmetry)
                           continue;
                      for (k = 0; k < biol_unit[i].n_symmetry; k++) {
                           if (strcmp(biol_unit[i].symmetry[k],
                                      biol_unit[j].symmetry[k])) break;
                      }
                      if (k == biol_unit[i].n_symmetry) {
                           biol_unit[j].label = 0;
                           if (!strstr(biol_unit[i].pdb_id, biol_unit[j].pdb_id)) {
                                strcat(biol_unit[i].pdb_id, ", ");
                                strcat(biol_unit[i].pdb_id, biol_unit[j].pdb_id);
                           }
                      }
                 }
            }

            newCard = _CardInfo[REMARK_TOKEN-1].
                             add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(350, 1);
            sprintf(buffer, "BIOMOLECULE: %s", BIOL[l].biol_id);
            newCard->AddElementToField(buffer, 2);

            for (i = 0; i < n_unit; i++) {
                 if (!biol_unit[i].label) continue;
                 p = get_max_length_word(biol_unit[i].pdb_id, FieldFormat, 28);
                 sprintf(buffer, "APPLY THE FOLLOWING TO CHAINS: %s", FieldFormat);
                 newCard = _CardInfo[REMARK_TOKEN-1].
                                  add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(350, 1);
                 newCard->AddElementToField(buffer, 2);
                 while (p) {
                      p = get_max_length_word(p, buffer, 58);
                      newCard = _CardInfo[REMARK_TOKEN-1].
                                     add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                      newCard->AddElementToField(350, 1);
                      newCard->AddElementToField(buffer, 2);
                 }

                 serialNo = 0;
                 for (t = 0; t < biol_unit[i].n_symmetry; t++) {
                      serialNo++;
                      opno = atoi(biol_unit[i].symmetry[t])-1;
                      len = strlen(biol_unit[i].symmetry[t]);
                      for (j = 0; j < len; j++) {
                           if (biol_unit[i].symmetry[t][j] == '_') break;
                      }
                      lx = biol_unit[i].symmetry[t][j+1] - '5';
                      ly = biol_unit[i].symmetry[t][j+2] - '5';
                      lz = biol_unit[i].symmetry[t][j+3] - '5';
                 
                      ndb_get_symmetry_matrix(cell, opno, lx, ly, lz, result, tr);
                      for (k = 0; k < 3; k++) {
                           newCard = _CardInfo[REMARK_TOKEN-1].
                                  add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                           newCard->AddElementToField(Remark_SYMMA.Remark_No, 1);
                           for (j = 0; j < Remark_SYMMA.NumField; j++) {
                                print_space_between_remark_field(newCard, &Remark_SYMMA, j); 
                                get_remark_field_format(FieldFormat, &Remark_SYMMA, j);
                                field_no = Remark_SYMMA.FieldList[j].FieldId-1;
                                if (!strcmp("BIOMT", Remark_SYMMA.FieldList[j].Text))
                                     sprintf(buffer, "%s%d", Remark_SYMMA.FieldList[j].Text, k + 1);
                                else if (field_no == 1)
                                     sprintf(buffer, FieldFormat, (double) serialNo);
                                else if (j != Remark_SYMMA.NumField - 1)
                                     sprintf(buffer, FieldFormat, result[k][j-2]);
                                else sprintf(buffer, FieldFormat, tr[k]);
                                newCard->AttachToField(buffer, 2);
                           }
                      }
                 }
            }
       }

       return 1;
}

void _MAXIT::ndb_to_pdb_get_remark_375()
{
       int i = 0, num = 0, len = 0;
       char buffer[70], chnid[5];
       _ATOM **atom = NULL;
       One_Card_Info_Format *newCard = NULL;

       if (_n_special == 0 || _special == NULL) return;

       ndb_to_pdb_get_general_remark(Num_Remark_375, Remark_375, 0, 0, 0, 0);
       for (i = 0; i < _n_special; i++) {
            atom = _molecules[0]->find_ndb_atom(_special[i]->a_atomtyp,
                      _special[i]->a_chnid, _special[i]->a_resnam,
                      _special[i]->a_resnum, "", num);
            if (!atom) continue;

            newCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(375, 1);

            if (!strcmp(atom[0]->pdb_resnam(), "HOH")) 
                 sprintf(buffer, "%-4s ", "    ");
            else sprintf(buffer, "%-4s ", atom[0]->pdb_atmnam());

            len = strlen(buffer);
            sprintf(&buffer[len], "%3s ", atom[0]->pdb_resnam());
            
            len = strlen(buffer);
            strcpy(chnid, " ");
            if (strcmp(atom[0]->pdb_chnid(), ""))
                 strcpy(chnid, atom[0]->pdb_chnid());
            sprintf(&buffer[len], "%c", chnid[0]);

            len = strlen(buffer);
            sprintf(&buffer[len], "%4d", atom[0]->pdb_resnum());

            len = strlen(buffer);
            strcpy(chnid, " ");
            if (strcmp(atom[0]->ins_code(), ""))
                 strcpy(chnid, atom[0]->ins_code());
            sprintf(&buffer[len], "%c", chnid[0]);

            len = strlen(buffer);
            sprintf(&buffer[len], " %s",
                          "LIES ON A SPECIAL POSITION.");
            newCard->AddElementToField(buffer, 2);
            delete [] atom;
       }
}

void _MAXIT::ndb_to_pdb_get_special_remark(const int remark_no)
{
       if (!_CifObj) return;
       ISTable *Table = _CifObj->GetTablePtr(_StructureId, "pdbx_entry_details");
       if (is_empty_table(Table)) return;

       int remark_num[4] = { 400, 450, 600, 999 };
       const char *remark_char[4] = { "400", "450", "600", "999" };
       const char *keywords[4] = { "COMPOUND", "SOURCE", "HETEROGEN", "SEQUENCE" };
       const char *items[4] = { "compound_details", "source_details", 
                          "nonpolymer_details", "sequence_details" };

       int index = -1;
       for (int i = 0; i < 4; i++) {
            if (remark_no == remark_num[i]) {
                 index = i;
                 break;
            }
       }
       if (index < 0) return;

       int idx = Table->GetColumnIndex(items[index]);
       CifString block_remark;
       block_remark.Clear();
       Table->GetCell(block_remark, idx, 0);
       if (block_remark == "?" || block_remark == ".") block_remark.Clear();
       if (block_remark == "") return;

       ReVarCifArray<CifString> remark_array;
       int width = NdbToken[REMARK_TOKEN-1].FieldList[2].FieldWidth;
       get_remark_array_from_block(remark_array, block_remark.Text(), width);

       if (remark_array.Length() == 0) return;

       One_Card_Info_Format *newCard = _CardInfo[REMARK_TOKEN-1].
                    add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
       newCard->AddElementToField(remark_char[index], 1);
       newCard->AddElementToField(keywords[index], 2);
       for (unsigned int i1 = 0; i1 < remark_array.Length(); i1++) {
            newCard = _CardInfo[REMARK_TOKEN-1].add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(remark_char[index], 1);
            newCard->AddElementToField(remark_array[i1].Text(), 2);
       }
}

void _MAXIT::ndb_to_pdb_get_remark_400_group()
{
       if (!_CifObj) return;
       ISTable *t1 = _CifObj->GetTablePtr(_StructureId, "rcsb_molecule");
       ISTable *t2 = _CifObj->GetTablePtr(_StructureId, "pdbx_molecule_features");
       if (is_empty_table(t1) || is_empty_table(t2)) return;

       std::set<std::string> polymer_prds;
       polymer_prds.clear();
       int rowNo = t1->GetNumRows();
       int idx2 = t1->GetColumnIndex("prd_id");
       int idx5 = t1->GetColumnIndex("auth_seq_id");
       CifString cs, cs1;
       for (int i = 0; i < rowNo; ++i) {
            get_value(cs, t1, idx2, i);
            get_value(cs1, t1, idx5, i);
            if (cs1 != "") continue;
            polymer_prds.insert(cs.Text());
       }
       if (polymer_prds.empty()) return;

       std::vector<_Remark_400_Group> remark_groups;
       remark_groups.clear();
       std::vector<std::string> general_remarks;
       general_remarks.clear();

       _Remark_400_Group group;
       group.id.clear();
       group.prd_id.clear();
       group.name.clear();
       group.chainids.clear();
       group.components.clear();
       group.description.clear();
       char buffer[80];
       rowNo = t2->GetNumRows();
       int idx1 = t2->GetColumnIndex("prd_id");
       idx2 = t2->GetColumnIndex("name");
       int idx3 = t2->GetColumnIndex("details");
       int idx4 = t2->GetColumnIndex("type");
       idx5 = t2->GetColumnIndex("class");
       int count = 0;
       for (int i = 0; i < rowNo; ++i) {
            get_value(cs, t2, idx1, i);
            group.prd_id = cs.Text();
            if (polymer_prds.find(group.prd_id) == polymer_prds.end()) continue;

            count++;
            sprintf(buffer, "%d", count);
            group.id = buffer;
            get_value(cs, t2, idx2, i);
            cs.ToUpper();
            group.name = cs.Text();
            get_value(cs, t2, idx3, i);
            group.description = cs.Text();
            remark_groups.push_back(group);
            std::string str = "THE " + group.name + " IS ";
            get_value(cs, t2, idx4, i);
            cs.ToUpper();
            str += cs.Text(); str += ", A MEMBER OF ";
            get_value(cs, t2, idx5, i);
            cs.ToUpper();
            str += cs.Text(); str += " CLASS.";
            general_remarks.push_back(str);
       }
       if (remark_groups.empty()) return;

       std::map<std::string, int> polymer_chain_order;
       std::map<std::string, std::string> polymer_chain_type;
       polymer_chain_order.clear();
       polymer_chain_type.clear();
       One_Card_Info_Format *pCard = _CardInfo[SEQIFO_TOKEN-1].Head();
       while (pCard) {
            polymer_chain_order.insert(std::make_pair(pCard->Field(3),
                                      (int) polymer_chain_order.size()));
            polymer_chain_type.insert(std::make_pair(pCard->Field(3), pCard->Field(5)));
            pCard = pCard->Next();
       }

       std::map<std::string, std::set<std::string> > prd_pdb_chnid_mapping;
       std::map<std::string, std::map<std::string, std::vector<int> > > prd_component_id_mapping;
       prd_pdb_chnid_mapping.clear();
       prd_component_id_mapping.clear();
       std::set<std::string> data_set;
       std::vector<int> i_vector;
       std::map<std::string, std::vector<int> > t_map;
       rowNo = t1->GetNumRows();
       idx1 = t1->GetColumnIndex("instance_id");
       idx2 = t1->GetColumnIndex("prd_id");
       idx3 = t1->GetColumnIndex("asym_id");
       idx4 = t1->GetColumnIndex("auth_asym_id");
       idx5 = t1->GetColumnIndex("auth_seq_id");
       CifString prd_id;
       for (int i = 0; i < rowNo; ++i) {
            get_value(prd_id, t1, idx2, i);
            if (polymer_prds.find(prd_id.Text()) == polymer_prds.end()) continue;
            get_value(cs, t1, idx4, i);
            std::map<std::string, std::set<std::string> >::iterator
                mpos1 = prd_pdb_chnid_mapping.find(prd_id.Text());
            if (mpos1 == prd_pdb_chnid_mapping.end()) {
                 data_set.clear();
                 data_set.insert(cs.Text());
                 prd_pdb_chnid_mapping.insert(std::make_pair(prd_id.Text(), data_set));
            } else mpos1->second.insert(cs.Text());

            get_value(cs, t1, idx1, i); 
            std::map<std::string, std::map<std::string, std::vector<int> > >::iterator
                mpos2 = prd_component_id_mapping.find(prd_id.Text());
            if (mpos2 != prd_component_id_mapping.end()) {
                 std::map<std::string, std::vector<int> >::iterator
                     mpos3 = mpos2->second.find(cs.Text());
                 if (mpos3 != mpos2->second.end())
                      mpos3->second.push_back(i);
                 else {
                      i_vector.clear();
                      i_vector.push_back(i);
                      mpos2->second.insert(std::make_pair(cs.Text(), i_vector));
                 }
            } else {
                 i_vector.clear();
                 i_vector.push_back(i);
                 t_map.clear();
                 t_map.insert(std::make_pair(cs.Text(), i_vector));
                 prd_component_id_mapping.insert(std::make_pair(prd_id.Text(), t_map));
            }
       }

       t_map.clear();
       for (std::map<std::string, std::map<std::string, std::vector<int> > >::const_iterator
            mpos = prd_component_id_mapping.begin(); mpos != prd_component_id_mapping.end(); ++mpos) {
            for (std::map<std::string, std::vector<int> >::const_iterator
                 mpos1 = mpos->second.begin(); mpos1 != mpos->second.end(); ++mpos1) {
                 std::map<std::string, std::vector<int> >::iterator
                     mpos2 = t_map.find(mpos->first);
                 if (mpos2 == t_map.end())
                      t_map.insert(std::make_pair(mpos->first, mpos1->second));
                 else if (mpos1->second.size() > mpos2->second.size())
                      mpos2->second = mpos1->second;
            }
       }

       std::map<std::string, std::string> nonpoly_mapping;
       nonpoly_mapping.clear();
       ISTable *t3 = _CifObj->GetTablePtr(_StructureId, "ndb_nonpoly_scheme");
       if (!is_empty_table(t3)) {
            rowNo = t3->GetNumRows();
            CifString cs2;
            int index1 = t3->GetColumnIndex("pdb_id");
            int index2 = t3->GetColumnIndex("pdb_num");
            int index3 = t3->GetColumnIndex("pdb_mon_id");
            for (int i = 0; i < rowNo; ++i) {
                 get_value(cs, t3, index1, i);
                 get_value(cs1, t3, index2, i);
                 get_value(cs2, t3, index3, i);
                 std::string str = cs.Text(); str += "_"; str += cs1.Text();
                 nonpoly_mapping.insert(std::make_pair(str, cs2.Text()));
            }
       }

       std::map<std::string, std::vector<std::string> > prd_component_mapping;
       prd_component_mapping.clear();
       std::vector<std::string> data;
       std::map<std::string, int> non_polymer_component_count;
       for (std::map<std::string, std::vector<int> >::const_iterator
            mpos = t_map.begin(); mpos != t_map.end(); ++mpos) {
            data.clear();
            non_polymer_component_count.clear();
            for (unsigned int i = 0; i < mpos->second.size(); ++i) {
                 get_value(cs, t1, idx4, mpos->second[i]);
                 get_value(cs1, t1, idx5, mpos->second[i]);
                 if (cs1 == "") {
                      std::map<std::string, std::string>::const_iterator
                          mpos1 = polymer_chain_type.find(cs.Text());
                      if (mpos1 != polymer_chain_type.end() &&
                          mpos1->second == "PR")
                           data.push_back("PEPTIDE LIKE POLYMER");
                      else data.push_back("POLYMER");
                 } else {
                      std::string str = cs.Text(); str += "_"; str += cs1.Text();
                      std::map<std::string, std::string>::const_iterator
                          mpos1 = nonpoly_mapping.find(str);
                      if (mpos1 == nonpoly_mapping.end()) continue;
/*
                      str = "RESIDUE " + mpos1->second;
                      ConnectFormat *drug = find_drug(mpos1->second.c_str());
                      if (drug && drug->chemical_name() &&
                          strcmp(drug->chemical_name(), "")) {
                           str += ": "; str += drug->chemical_name();
                      }
                      data.push_back(str);
*/
                      std::map<std::string, int>::iterator
                          mpos2 = non_polymer_component_count.find(mpos1->second);
                      if (mpos2 != non_polymer_component_count.end())
                           mpos2->second++;
                      else non_polymer_component_count.insert(std::make_pair(mpos1->second, 1));
                 }
            }
            for (std::map<std::string, int>::const_iterator
                 mpos2 = non_polymer_component_count.begin();
                 mpos2 != non_polymer_component_count.end(); ++mpos2) {
                 std::string str = "RESIDUE " + mpos2->first;
                 if (mpos2->second > 1) {
                      sprintf(buffer, ", %d COPIES", mpos2->second);
                      str += buffer;
                 }
                 data.push_back(str);
            }
            prd_component_mapping.insert(std::make_pair(mpos->first, data));
       }

       std::multimap<int, std::string> order_chains;
       for (unsigned int i = 0; i < remark_groups.size(); ++i) {
            std::map<std::string, std::set<std::string> >::const_iterator
                mpos = prd_pdb_chnid_mapping.find(remark_groups[i].prd_id);
            if (mpos != prd_pdb_chnid_mapping.end()) {
                 order_chains.clear();
                 int j = polymer_chain_order.size();
                 for (std::set<std::string>::const_iterator
                      spos = mpos->second.begin(); spos != mpos->second.end(); ++spos) {
                      std::map<std::string, int>::const_iterator
                          mpos1 = polymer_chain_order.find(*spos);
                      if (mpos1 != polymer_chain_order.end())
                           order_chains.insert(std::make_pair(mpos1->second, *spos));
                      else order_chains.insert(std::make_pair(j++, *spos));
                 }
                 for (std::multimap<int, std::string>::const_iterator
                      mpos2 = order_chains.begin(); mpos2 != order_chains.end(); ++mpos2) {
                      if (!remark_groups[i].chainids.empty())
                           remark_groups[i].chainids += ", ";
                      remark_groups[i].chainids += mpos2->second;
                 }
            }
            std::map<std::string, std::vector<std::string> >::const_iterator
                mpos3 = prd_component_mapping.find(remark_groups[i].prd_id);
            if (mpos3 != prd_component_mapping.end()) {
                 remark_groups[i].components = mpos3->second;
            }
       }

       ReVarCifArray<CifString> remark_array;
       int width = NdbToken[REMARK_TOKEN-1].FieldList[2].FieldWidth;

       bool found_400 = false;
       pCard = _CardInfo[REMARK_TOKEN-1].Tail();
       if (pCard && !strcmp(pCard->Field(1), "400")) found_400 = true;
       if (!found_400) ndb_to_pdb_add_remark(400, "COMPOUND");
 
       for (unsigned int i = 0; i < general_remarks.size(); ++i) {
            ndb_to_pdb_add_remark(400, "");
            get_remark_array_from_block(remark_array, general_remarks[i].c_str(), width);
            for (unsigned int j = 0; j < remark_array.Length(); ++j) {
                 ndb_to_pdb_add_remark(400, remark_array[j].Text());
            }
       }

       for (unsigned int i = 0; i < remark_groups.size(); ++i) {
            ndb_to_pdb_add_remark(400, "");

            memset(buffer, 0, 80);
            sprintf(buffer, " GROUP: %s", remark_groups[i].id.c_str()); 
            ndb_to_pdb_add_remark(400, buffer);

            get_remark_array_from_block(remark_array, remark_groups[i].name.c_str(),
                                        width - 8);
            if (remark_array.Length() == 0) {
                 cs.Copy("NULL");
                 remark_array.Add(cs);
            }
            memset(buffer, 0, 80);
            sprintf(buffer, "  NAME: %s", remark_array[0].Text());
            ndb_to_pdb_add_remark(400, buffer);
            for (unsigned int j = 1; j < remark_array.Length(); ++j) {
                 memset(buffer, 0, 80);
                 sprintf(buffer, "        %s", remark_array[j].Text());
                 ndb_to_pdb_add_remark(400, buffer);
            }

            get_remark_array_from_block(remark_array, remark_groups[i].chainids.c_str(),
                                        width - 9);
            if (remark_array.Length() == 0) {
                 cs.Copy("NULL");
                 remark_array.Add(cs);
            }
            memset(buffer, 0, 80);
            sprintf(buffer, "  CHAIN: %s", remark_array[0].Text());
            ndb_to_pdb_add_remark(400, buffer);
            for (unsigned int j = 1; j < remark_array.Length(); ++j) {
                 memset(buffer, 0, 80);
                 sprintf(buffer, "         %s", remark_array[j].Text());
                 ndb_to_pdb_add_remark(400, buffer);
            }

            for (unsigned int k = 0; k < remark_groups[i].components.size(); ++k) {
                 get_remark_array_from_block(remark_array,
                    remark_groups[i].components[k].c_str(), width - 15);
                 if (remark_array.Length() == 0) {
                      cs.Copy("NULL");
                      remark_array.Add(cs);
                 }
                 memset(buffer, 0, 80);
                 sprintf(buffer, "  COMPONENT_%d: %s", (int) k + 1,
                                 remark_array[0].Text());
                 ndb_to_pdb_add_remark(400, buffer);
                 for (unsigned int j = 1; j < remark_array.Length(); ++j) {
                      memset(buffer, 0, 80);
                      sprintf(buffer, "               %s", remark_array[j].Text());
                      ndb_to_pdb_add_remark(400, buffer);
                 }
            }

            get_remark_array_from_block(remark_array,
               remark_groups[i].description.c_str(), width - 15);
            if (remark_array.Length() == 0) {
                 cs.Copy("NULL");
                 remark_array.Add(cs);
            }
            memset(buffer, 0, 80);
            sprintf(buffer, "  DESCRIPTION: %s", remark_array[0].Text());
            ndb_to_pdb_add_remark(400, buffer);
            for (unsigned int j = 1; j < remark_array.Length(); ++j) {
                 memset(buffer, 0, 80);
                 sprintf(buffer, "               %s", remark_array[j].Text());
                 ndb_to_pdb_add_remark(400, buffer);
            }
       }
}

void _MAXIT::ndb_to_pdb_add_remark(const int remark_no, const char *value)
{
       One_Card_Info_Format *newCard = _CardInfo[REMARK_TOKEN-1].
                    add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
       newCard->AddElementToField(remark_no, 1);
       newCard->AddElementToField(value, 2);
}

void _MAXIT::ndb_to_pdb_get_remark_465()
{
       char mol_id[4], chain_id[2], insert_code[2], buffer[80], prv_mol_id[4];
       CifString cs;
       One_Card_Info_Format *pCard = NULL, *newCard = NULL;

       pCard = _CardInfo[MISRES_TOKEN-1].Head();
       if (pCard == NULL) return;

       int need_new_nmr_template = 0;
       if (_experiment_type & EXPERIMENT_TYPE_NMR ||
           _experiment_type & EXPERIMENT_TYPE_NMR_SOLID) {
            need_new_nmr_template = 1;
            if (_num_mol > 1) {
                 TblIndexObj Index1, Index2;
                 Index1.Clear(); Index2.Clear();
                 strcpy(prv_mol_id, pCard->Field(2));
                 while (pCard) {
                      if (strcmp(prv_mol_id, pCard->Field(2))) break;
                      cs.Copy(pCard->Field(4));
                      cs += "_";
                      if (strcmp(pCard->Field(3), "")) cs += pCard->Field(3);
                      cs += "_";
                      cs += pCard->Field(5);
                      cs += "_";
                      if (strcmp(pCard->Field(6), "")) cs += pCard->Field(6);
                      Index1.Add(cs);
                      pCard = pCard->Next();
                 }

                 pCard = _CardInfo[MISRES_TOKEN-1].Head();
                 while (pCard) {
                      if (strcmp(prv_mol_id, pCard->Field(2))) {
                           strcpy(prv_mol_id, pCard->Field(2));
                           if (Index1.NumElem() != Index2.NumElem()) {
                                need_new_nmr_template = 0;
                                break;
                           }
                           if (Index1.Root() >= 0 && Index2.Root() >= 0) {
                                int ret = Index1.GoFirst();
                                while (ret >= 0) {
                                     cs = Index1.GetCurrData();
                                     if (Index2.Seek(cs) < 0) {
                                          need_new_nmr_template = 0;
                                          break;
                                     }
                                     ret = Index1.GoToNext();
                                }
                                if (need_new_nmr_template == 0) break;

                                ret = Index2.GoFirst();
                                while (ret >= 0) {
                                     cs = Index2.GetCurrData();
                                     if (Index1.Seek(cs) < 0) {
                                          need_new_nmr_template = 0;
                                          break;
                                     }
                                     ret = Index2.GoToNext();
                                }
                                if (need_new_nmr_template == 0) break;
                           }
                           Index2.Clear();
                      }
                      cs.Copy(pCard->Field(4));
                      cs += "_";
                      if (strcmp(pCard->Field(3), "")) cs += pCard->Field(3);
                      cs += "_";
                      cs += pCard->Field(5);
                      cs += "_";
                      if (strcmp(pCard->Field(6), "")) cs += pCard->Field(6);
                      Index2.Add(cs);
                      pCard = pCard->Next();
                 }

                 if (Index1.NumElem() != Index2.NumElem())
                      need_new_nmr_template = 0;
                 else if (Index1.Root() >= 0 && Index2.Root() >= 0) {
                      int ret = Index1.GoFirst();
                      while (ret >= 0) {
                           cs = Index1.GetCurrData();
                           if (Index2.Seek(cs) < 0) {
                                need_new_nmr_template = 0;
                                break;
                           }
                           ret = Index1.GoToNext();
                      }
                      if (need_new_nmr_template == 1) {
                           ret = Index2.GoFirst();
                           while (ret >= 0) {
                                cs = Index2.GetCurrData();
                                if (Index1.Seek(cs) < 0) {
                                     need_new_nmr_template = 0;
                                     break;
                                }
                                ret = Index2.GoToNext();
                           }
                      }
                 }
                 Index1.Clear(); Index2.Clear();
            }
       }

       if (need_new_nmr_template) {
            ndb_to_pdb_get_general_remark(Num_Remark_465_NMR, Remark_465_NMR, 0, 0, 0, 0);
            if (_num_mol > 1) {
                 newCard = _CardInfo[REMARK_TOKEN-1].add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(465, 1);
                 sprintf(buffer, "  MODELS %d-%d", _molecules[0]->Mol_ID(),
                               _molecules[_num_mol - 1]->Mol_ID());
                 newCard->AddElementToField(buffer, 2);
            }
            newCard = _CardInfo[REMARK_TOKEN-1].add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(465, 1);
            newCard->AddElementToField("    RES C SSSEQI", 2);
       } else ndb_to_pdb_get_general_remark(Num_Remark_465, Remark_465, 0, 0, 0, 0);

       pCard = _CardInfo[MISRES_TOKEN-1].Head();
       strcpy(prv_mol_id, pCard->Field(2));
       while (pCard) {
            if (need_new_nmr_template && strcmp(prv_mol_id, pCard->Field(2))) break;
            newCard = _CardInfo[REMARK_TOKEN-1].add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(465, 1);
            if (strcmp(pCard->Field(2), ""))
                 strcpy(mol_id, pCard->Field(2));
            else strcpy(mol_id, "   ");
            if (need_new_nmr_template) strcpy(mol_id, "   ");
            if (strcmp(pCard->Field(3), ""))
                 strcpy(chain_id, pCard->Field(3));
            else strcpy(chain_id, " ");
            if (strcmp(pCard->Field(6), ""))
                 strcpy(insert_code, pCard->Field(6));
            else strcpy(insert_code, " ");
            sprintf(buffer, "%3s %3s %c %5s%c", mol_id, pCard->Field(4),
                      chain_id[0], pCard->Field(5), insert_code[0]);
            newCard->AddElementToField(buffer, 2);
            pCard = pCard->Next();
       }
}

void _MAXIT::ndb_to_pdb_get_remark_470()
{
       int first = TRUE, num = 0;
       char head[20], model[5], buff[10], buffer[10], chnid = ' ', *p = NULL;
       char ins = ' ';
       One_Card_Info_Format *newCard = NULL;

       finding_missing_and_extra_atoms(1);

       int need_new_nmr_template = 0;
       if (_experiment_type & EXPERIMENT_TYPE_NMR ||
           _experiment_type & EXPERIMENT_TYPE_NMR_SOLID) {
            need_new_nmr_template = 1;
            if (_num_mol > 1) {
                 TblIndexObj Index1, Index2;
                 CifString cs, cs1;

                 Index1.Clear();
                 _CHAIN *chain = _molecules[0]->GetFirstChain();
                 while (chain) {
                      if (chain->chain_type() != ATOMN_TOKEN &&
                          chain->chain_type() != ATOMP_TOKEN) {
                           chain = _molecules[0]->GetNextChain();
                           continue;
                      }
                      if (chain->ca_or_p_atom_only()) {
                           chain = _molecules[0]->GetNextChain();
                           continue;
                      }
                      _RESIDUE *residue = chain->GetFirstResidue();
                      while (residue) {
                           if (!strcmp(residue->missing().Text(), "")) {
                                residue = chain->GetNextResidue();
                                continue;
                           }
                           cs.Copy(chain->PDB_ChainID());
                           cs += "_"; cs += residue->ResName();
                           cs += "_"; cs += residue->pdb_res_no();
                           if (strcmp(residue->GetFirstAtom()->ins_code(), "")) {
                                cs += residue->GetFirstAtom()->ins_code();
                           }
                           p = residue->missing().Text();
                           while (p) {
                                p = get_word(p, buff);
                                cs1.Copy(cs);
                                cs1 += "_"; cs1 += buff;
                                Index1.Add(cs1);
                           }
                           residue = chain->GetNextResidue();
                      }
                      chain = _molecules[0]->GetNextChain();
                 }

                 for (int i = 1; i < _num_mol; i++) {
                      Index2.Clear();
                      chain = _molecules[i]->GetFirstChain();
                      while (chain) {
                           if (chain->chain_type() != ATOMN_TOKEN &&
                               chain->chain_type() != ATOMP_TOKEN) {
                                chain = _molecules[i]->GetNextChain();
                                continue;
                           }
                           if (chain->ca_or_p_atom_only()) {
                                chain = _molecules[i]->GetNextChain();
                                continue;
                           }
                           _RESIDUE *residue = chain->GetFirstResidue();
                           while (residue) {
                                if (!strcmp(residue->missing().Text(), "")) {
                                     residue = chain->GetNextResidue();
                                     continue;
                                }
                                cs.Copy(chain->PDB_ChainID());
                                cs += "_"; cs += residue->ResName();
                                cs += "_"; cs += residue->pdb_res_no();
                                if (strcmp(residue->GetFirstAtom()->ins_code(), "")) {
                                     cs += residue->GetFirstAtom()->ins_code();
                                }
                                p = residue->missing().Text();
                                while (p) {
                                     p = get_word(p, buff);
                                     cs1.Copy(cs);
                                     cs1 += "_"; cs1 += buff;
                                     Index2.Add(cs1);
                                }
                                residue = chain->GetNextResidue();
                           }
                           chain = _molecules[i]->GetNextChain();
                      }

                      if (Index1.NumElem() != Index2.NumElem()) {
                           need_new_nmr_template = 0;
                           break;
                      } else if (Index1.Root() >= 0 && Index2.Root() >= 0) {
                           int ret = Index1.GoFirst();
                           while (ret >= 0) {
                                cs = Index1.GetCurrData();
                                if (Index2.Seek(cs) < 0) {
                                     need_new_nmr_template = 0;
                                     break;
                                }
                                ret = Index1.GoToNext();
                           }
                           if (need_new_nmr_template == 0) break;

                           ret = Index2.GoFirst();
                           while (ret >= 0) {
                                cs = Index2.GetCurrData();
                                if (Index1.Seek(cs) < 0) {
                                     need_new_nmr_template = 0;
                                     break;
                                }
                                ret = Index2.GoToNext();
                           }
                           if (need_new_nmr_template == 0) break;
                      }
                 }
                 Index1.Clear(); Index2.Clear();
            }
       }

       for (int i = 0; i < _num_mol; i++) {
            if (need_new_nmr_template && i) break;
            _CHAIN *chain = _molecules[i]->GetFirstChain();
            while (chain) {
                 if (chain->chain_type() != ATOMN_TOKEN &&
                     chain->chain_type() != ATOMP_TOKEN) {
                      chain = _molecules[i]->GetNextChain();
                      continue;
                 }
                 if (chain->ca_or_p_atom_only()) {
                      chain = _molecules[i]->GetNextChain();
                      continue;
                 }

                 _RESIDUE *residue = chain->GetFirstResidue();
                 while (residue) {
                      if (!strcmp(residue->missing().Text(), "")) {
                           residue = chain->GetNextResidue();
                           continue;
                      }
                  
                      if (first) {
                           first = FALSE;
                           if (need_new_nmr_template) {
                                ndb_to_pdb_get_general_remark(Num_Remark_470_NMR,
                                            Remark_470_NMR, 0, 0, 0, 0);
                                if (_num_mol > 1) {
                                     newCard = _CardInfo[REMARK_TOKEN-1].
                                           add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                                     newCard->AddElementToField(470, 1);
                                     sprintf(buffer, "  MODELS %d-%d",
                                                  _molecules[0]->Mol_ID(),
                                                  _molecules[_num_mol - 1]->Mol_ID());
                                     newCard->AddElementToField(buffer, 2);
                                }
                                newCard = _CardInfo[REMARK_TOKEN-1].add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                                newCard->AddElementToField(470, 1);
                                newCard->AddElementToField("    RES CSSEQI  ATOMS", 2);
                           } else ndb_to_pdb_get_general_remark(Num_Remark_470, Remark_470, 0, 0, 0, 0);
                      }
     
                      head[0] = '\0';
                      model[0] = '\0';
                      if (_num_mol > 1) sprintf(model, "%3d", _molecules[i]->Mol_ID());
                      else strcpy(model, "   ");
                      if (need_new_nmr_template) strcpy(model, "   ");
                      ins = residue->GetFirstAtom()->ins_code()[0];
                      if (ins == '\0') ins = ' ';
                      chnid = chain->PDB_ChainID()[0];
                      if (chnid == '\0') chnid = ' ';
     
                      sprintf(head, "%3s %3s %c%4d%c ", model,
                          residue->ResName(), chnid, residue->pdb_res_no(), ins);
     
                      num = 0;
                      newCard = _CardInfo[REMARK_TOKEN-1].
                                   add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                      newCard->AddElementToField(470, 1);
                      newCard->AddElementToField(head, 2);
     
                      p = residue->missing().Text();
                      while (p) {
                           p = get_word(p, buff);
                           ndb_clean_string(buff);
                           if (!strcmp(buff, "")) continue;
                           if (num == 7) {
                                newCard = _CardInfo[REMARK_TOKEN-1].
                                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                                newCard->AddElementToField(470, 1);
                                newCard->AddElementToField(head, 2);
                                num = 0;
                           }
                           if (strlen(buff) == 4)
                                sprintf(buffer, "%4s", buff);
                           else if (strlen(buff) == 3)
                                sprintf(buffer, " %3s", buff);
                           else if (strlen(buff) == 2) {
                                if (!strcmp(residue->ResName(), "MSE") &&
                                    !strcmp(buff, "SE"))
                                     sprintf(buffer, "%-4s", buff);
                                else sprintf(buffer, " %-3s", buff);
                           } else if (strlen(buff) == 1)
                                sprintf(buffer, " %s  ", buff);
                           sprintf(buff, " %4s", buffer);
                           newCard->AttachToField(buff, 2);
                           num++;
                      }
                      residue = chain->GetNextResidue();
                 }
                 chain = _molecules[i]->GetNextChain();
            }
       }
}

void _MAXIT::ndb_to_pdb_get_remark_500(const int remark)
{
       if (!_CifObj) return;
       if (!_molecules) return;
       ISTable *Table = _CifObj->GetTablePtr(_StructureId, "pdbx_validate_close_contact");
       if (has_value(Table, _num_mol, _molecules)) {
            if (_experiment_type & EXPERIMENT_TYPE_MODEL ||
                _experiment_type & EXPERIMENT_TYPE_NMR_SOLID ||
                _experiment_type & EXPERIMENT_TYPE_NMR)
                 ndb_to_pdb_get_general_remark(Num_Remark_500_NMR, Remark_500_NMR, 0, 0, 0, 0);
            else ndb_to_pdb_get_general_remark(Num_Remark_500_A, Remark_500_A, 0, 0, 0, 0);
            output_close_contact(Table, 0);
    
       }
       Table = _CifObj->GetTablePtr(_StructureId, "pdbx_validate_symm_contact");
       if (!remark && has_value(Table, _num_mol, _molecules)) {
            ndb_to_pdb_get_general_remark(Num_Remark_500_S, Remark_500_S, 0, 0, 0, 0);
            output_close_contact(Table, 1);
       }
}

void _MAXIT::ndb_to_pdb_get_remark_525()
{
       if (_n_waters == 0 || _waters == NULL) return;

       if (__maxit_version == "3.1")
            ndb_to_pdb_get_general_remark(Num_Remark_525_NEW, Remark_525_NEW, 0, 0, 0, 0);
       else ndb_to_pdb_get_general_remark(Num_Remark_525, Remark_525, 0, 0, 0, 0);

       int num = 0, len = 0, i;
       double dist = 0;
       _ATOM **atom = NULL;
       char buffer[70], chnid[5], ins_code[5];
       One_Card_Info_Format *newCard = NULL;

       for (i = 0; i < _n_waters; i++) {
            atom = _molecules[0]->find_ndb_atom(_waters[i]->a_atomtyp,
                      _waters[i]->a_chnid, _waters[i]->a_resnam,
                      _waters[i]->a_resnum, "", num);
            if (!atom) continue;

            newCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(525, 1);

            strcpy(chnid, " ");
            strcpy(ins_code, " ");
            if (strcmp(atom[0]->pdb_chnid(), ""))
                 strcpy(chnid, atom[0]->pdb_chnid());
            if (strcmp(atom[0]->ins_code(), ""))
                 strcpy(ins_code, atom[0]->ins_code());
            sprintf(buffer, "   %3s %c%4d%c", atom[0]->pdb_resnam(),
                      chnid[0], atom[0]->pdb_resnum(), ins_code[0]);
            
            dist = cal_min_distance(_molecules[0], atom[0], WATER_LIMIT);
            delete [] atom;
            len = strlen(buffer);
            sprintf(&buffer[len], 
                    "       DISTANCE = %5.2f ANGSTROMS", dist);
            newCard->AddElementToField(buffer, 2);
       }
}

void _MAXIT::ndb_to_pdb_processing_nmr_remarks()
{
       int count = 0, i = 0;
       char buff[1000], program[1000], authors[1000];
       char maker[100], model[100], strengh[100], *p;
       One_Card_Info_Format *pCard = NULL, *qCard = NULL;

       pCard = _CardInfo[REFMET_TOKEN-1].Head();
       while (pCard) pCard = _CardInfo[REFMET_TOKEN-1].delete_a_card(pCard);
       pCard = _CardInfo[NMRECD_TOKEN-1].Head();
       while (pCard) pCard = _CardInfo[NMRECD_TOKEN-1].delete_a_card(pCard);
       pCard = _CardInfo[NMRSFT_TOKEN-1].Head();
       while (pCard) pCard = _CardInfo[NMRSFT_TOKEN-1].delete_a_card(pCard);

       pCard = _CardInfo[SFTWAR_TOKEN-1].Head();
       while (pCard) {
            if (!strcasecmp(pCard->Field(2), "refinement")) count++;
            pCard = pCard->Next();
       }
       program[0] = '\0'; authors[0] = '\0';
       i = 0;
       pCard = _CardInfo[SFTWAR_TOKEN-1].Head();
       while (pCard) {
            if (!strcasecmp(pCard->Field(2), "refinement")) {
                 i++; 
                 if (strcmp(pCard->Field(4), ""))
                      sprintf(buff, "%s %s", pCard->Field(3), pCard->Field(4));
                 else sprintf(buff, "%s", pCard->Field(3));
                 if (strcmp(program, "")) strcat(program, ", ");
                 strcat(program, buff);

                 if (count > 1)
                      sprintf(buff, "%s (%s)", pCard->Field(5), pCard->Field(3));
                 else strcpy(buff, pCard->Field(5));
                 if (strcmp(authors, "")) strcat(authors, ", ");
                 strcat(authors, buff);
            }
            pCard = pCard->Next();
       }
       pCard = _CardInfo[REFMET_TOKEN-1].add_a_new_card(NdbToken[REFMET_TOKEN-1], 0);
       pCard->AddElementToField(program, 3);
       pCard->AddElementToField(authors, 4);

       count = 0;
       pCard = _CardInfo[SFTWAR_TOKEN-1].Head();
       while (pCard) {
            if (strcasecmp(pCard->Field(2), "refinement")) count++;
            pCard = pCard->Next();
       }
       program[0] = '\0'; i = 0;
       pCard = _CardInfo[SFTWAR_TOKEN-1].Head();
       while (pCard) {
            if (strcasecmp(pCard->Field(2), "refinement")) {
                 i++;
                 if (strcmp(pCard->Field(4), ""))
                      sprintf(buff, "%s %s", pCard->Field(3), pCard->Field(4));
                 else sprintf(buff, "%s", pCard->Field(3));
                 if (!strstr(program, buff)) {
                      if (strcmp(program, "")) strcat(program, ", ");
                      strcat(program, buff);
                 }
            }
            pCard = pCard->Next();
       }
       pCard = _CardInfo[NMRSFT_TOKEN-1].add_a_new_card(NdbToken[NMRSFT_TOKEN-1], 0);
       pCard->AddElementToField(program, 2);

       pCard = _CardInfo[NMREXP_TOKEN-1].Head();
       program[0] = '\0';
       while (pCard) {
            if (!strstr(program, pCard->Field(3))) {
                 if (!strcmp(program, "")) strcpy(program, pCard->Field(3));
                 else {
                      strcat(program, "; ");
                      strcat(program, pCard->Field(3));
                 }
            }
            pCard = pCard->Next();
       }
       pCard = _CardInfo[NMRECD_TOKEN-1].add_a_new_card(NdbToken[NMRECD_TOKEN-1], 0);
       pCard->AddElementToField(program, 2);

       pCard = _CardInfo[NMRSPM_TOKEN-1].Head();
       if (pCard) {
            maker[0] = model[0] = strengh[0] = '\0';
            while (pCard) {
                 p = strstr(pCard->Field(4), "MHZ");
                 if (p == NULL) {
                      sprintf(buff, "%s MHZ", pCard->Field(4));
                      pCard->AddElementToField(buff, 4);
                 }
                 if (strcmp(pCard->Field(2), "") && !strstr(model, pCard->Field(2))) {
                      if (strcmp(model, "")) {
                           strcat(model, "; ");
                           strcat(model, pCard->Field(2));
                      } else strcpy(model, pCard->Field(2));
                 }
                 if (strcmp(pCard->Field(3), "") && !strstr(maker, pCard->Field(3))) {
                      if (strcmp(maker, "")) {
                           strcat(maker, "; ");
                           strcat(maker, pCard->Field(3));
                      } else strcpy(maker, pCard->Field(3));
                 }
                 if (strcmp(pCard->Field(4), "") && !strstr(strengh, pCard->Field(4))) {
                      if (strcmp(strengh, "")) {
                           strcat(strengh, "; ");
                           strcat(strengh, pCard->Field(4));
                      } else strcpy(strengh, pCard->Field(4));
                 }
                 pCard = pCard->Next();
            }
     
            pCard = _CardInfo[NMRSPM_TOKEN-1].Head();
            while (pCard) pCard = _CardInfo[NMRSPM_TOKEN-1].delete_a_card(pCard);
            pCard = _CardInfo[NMRSPM_TOKEN-1].add_a_new_card(NdbToken[NMRSPM_TOKEN-1], 0);
            pCard->AddElementToField(model, 2);
            pCard->AddElementToField(maker, 3);
            pCard->AddElementToField(strengh, 4);
       }

       pCard = _CardInfo[NMRSMP_TOKEN-1].Head();
       while (pCard) {
            if (strcmp(pCard->Field(7), "")) {
                 pCard->AttachToField(" ", 4);
                 pCard->AttachToField(pCard->Field(7), 4);
            }
            pCard = pCard->Next();
       }
       pCard = _CardInfo[NMRSMP_TOKEN-1].Head();
       if (pCard) {
            qCard = pCard->Next();
            if (qCard) {
                 for (i = 3; i < 7; i++) {
                      if (!strcmp(pCard->Field(i), ""))
                           pCard->AddElementToField("NULL", i);
                 }
            }
            while (qCard) {
                 for (i = 3; i < 7; i++) {
                      pCard->AttachToField("; ", i);
                      if (strcmp(qCard->Field(i), ""))
                           pCard->AttachToField(qCard->Field(i), i);
                      else pCard->AttachToField("NULL", i);
                 }
                 qCard = qCard->Next();
            }
       }

       int NMRSDT_TOKEN = find_card_id("NMRSDT");
       pCard = _CardInfo[NMRSDT_TOKEN].Head();
       if (pCard) {
            qCard = pCard->Next();
            if (qCard) {
                 if (!strcmp(pCard->Field(2), ""))
                      pCard->AddElementToField("NULL", 2);
            }
            while (qCard) {
                 pCard->AttachToField("; ", 2);
                 if (strcmp(qCard->Field(2), ""))
                      pCard->AttachToField(qCard->Field(2), 2);
                 else pCard->AttachToField("NULL", 2);
                 qCard = qCard->Next();
            }
       }
}

int _MAXIT::check_virus_ncs_remark_300()
{
       int idx1 = -1, idx2 = -1, idx3 = -1, idx4 = -1, idx5 = -1 /* , num_chn = 0 */ ;
       CifString cs1, cs2, cs3, cs4, cs5, remark;
       One_Card_Info_Format *newCard = NULL;

       if (!_CifObj) return 0;
 
       char *author_details = NULL;
       ISTable *t = _CifObj->GetTablePtr(_StructureId, "struct_biol");
       if (!is_empty_table(t)) {
             int idx = t->GetColumnIndex("details");
             t->GetCell(cs1, idx, 0);
             if (cs1 != "" && cs1 != "?" && cs1 != ".") {
                  idx = strlen(cs1.Text());
                  author_details = new char[idx + 10];
                  memset(author_details, 0, idx + 10);
                  strcpy(author_details, "REMARK: ");
                  strcat(author_details, cs1.Text());
                  ndb_set_string_to_upper(author_details);
             }
       }
 
       One_Card_Info_Format *pCard = _CardInfo[SPLIT_TOKEN-1].Head();
       if (pCard) {
            int card_id = find_card_id("CHAINS");
            pCard = _CardInfo[card_id].add_a_new_card(NdbToken[card_id], 0);
            pCard->AddElementToField(1, 2);
            ndb_to_pdb_get_general_remark(Num_Remark_300_NEW3_1, Remark_300_NEW3_1,
                                    0, 0, 0, 0);
            if (author_details) {
                 get_author_defined_remark_300_details(author_details);
                 delete [] author_details;
            }
            return 0;
       }

       t = _CifObj->GetTablePtr(_StructureId, "pdbx_struct_assembly");
       if (is_empty_table(t)) {
            if (author_details) {
                 ndb_to_pdb_get_general_remark(Num_Remark_300_NEW3_1, Remark_300_NEW3_1,
                                       0, 0, 0, 0);
                 get_author_defined_remark_300_details(author_details);
                 delete [] author_details;
            }
            return 0;
       }
/*
       One_Card_Info_Format *pCard = _CardInfo[card_id].add_a_new_card(NdbToken[card_id], 0);
       _CHAIN *chain = _molecules[0]->GetFirstChain();
       while (chain) {
            if (chain->chain_type() == ATOMP_TOKEN ||
                chain->chain_type() == ATOMN_TOKEN) num_chn++;
            chain = _molecules[0]->GetNextChain();
       }
       if (num_chn) pCard->AddElementToField(num_chn, 2);
       pCard->AddElementToField(1, 3);
*/
       cs1.Clear();
       int rowNo = t->GetNumRows();
       idx2 = t->GetColumnIndex("details");
       int biol_unit = 0;
       char buff[10];
       // int has_author_defined_assembly = 0;
       for (int i = 0; i < rowNo; i++) {
            get_value(cs2, t, idx2, i);
            cs2.ToLower();
            if (cs2 == "representative helical assembly" ||
                cs2 == "complete icosahedral assembly" ||
                cs2 == "complete point assembly" ||
                cs2 == "author_and_software_defined_assembly" ||
                cs2 == "author_defined_assembly" ||
                cs2 == "software_defined_assembly") {
                 // if (cs2 != "software_defined_assembly") has_author_defined_assembly = 1; 
                 biol_unit++;
                 sprintf(buff, "%d", biol_unit);
                 if (cs1 != "") cs1 += ", ";
                 cs1 += buff;
            }
       }
       if (cs1 == "") {
            if (author_details) {
                 ndb_to_pdb_get_general_remark(Num_Remark_300_NEW3_1, Remark_300_NEW3_1, 0, 0, 0, 0);
                 get_author_defined_remark_300_details(author_details);
                 delete [] author_details;
            }
            return 0;
       }

       int card_id = find_card_id("CHAINS");
       pCard = _CardInfo[card_id].add_a_new_card(NdbToken[card_id], 0);
       pCard->AddElementToField(cs1.Text(), 2);

       ndb_to_pdb_get_general_remark(Num_Remark_300_NEW3_1, Remark_300_NEW3_1, 0, 0, 0, 0);
       if (author_details) {
            // if (!has_author_defined_assembly)
            get_author_defined_remark_300_details(author_details);
            delete [] author_details;
       }
       
       // ndb_to_pdb_get_general_remark(Num_Remark_300_VIRUS_new, Remark_300_VIRUS_new, 0, 0, 0, 0);
       t = _CifObj->GetTablePtr(_StructureId, "pdbx_point_symmetry");
       if (!is_empty_table(t)) {
            idx1 = t->GetColumnIndex("Schoenflies_symbol");
            idx2 = t->GetColumnIndex("circular_symmetry");
            cs1.Clear();
            t->GetCell(cs1, idx1, 0);
            if (cs1 == "?" || cs1 == ".") cs1.Clear();
            cs2.Clear();
            t->GetCell(cs2, idx2, 0);
            if (cs2 == "?" || cs2 == ".") cs2.Clear();

            if (cs1 == "I" || cs1 == "O" || cs1 == "T" || cs1 == "D" || cs1 == "C") {
                 remark.Copy("THE ASSEMBLY REPRESENTED IN THIS ENTRY HAS REGULAR");
                 newCard = _CardInfo[REMARK_TOKEN-1].
                             add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(300, 1);
                 newCard->AddElementToField(remark.Text(), 2);

                 if (cs1 == "I")
                      remark.Copy("ICOSAHEDRAL POINT SYMMETRY (SCHOENFLIES SYMBOL = ");
                 else if (cs1 == "O")
                      remark.Copy("OCTAHEDRAL POINT SYMMETRY (SCHOENFLIES SYMBOL = ");
                 else if (cs1 == "T")
                      remark.Copy("TETRAHEDRAL POINT SYMMETRY (SCHOENFLIES SYMBOL = ");
                 else if (cs1 == "D")
                      remark.Copy("DIHEDRAL POINT SYMMETRY (SCHOENFLIES SYMBOL = ");
                 else if (cs1 == "C")
                      remark.Copy("CYCLIC POINT SYMMETRY (SCHOENFLIES SYMBOL = ");
                 remark += cs1;
                 if (cs1 == "D" || cs1 == "C") remark += cs2;
                 remark += ").";
                 newCard = _CardInfo[REMARK_TOKEN-1].
                             add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(300, 1);
                 newCard->AddElementToField(remark.Text(), 2);
            }
       }
       t = _CifObj->GetTablePtr(_StructureId, "pdbx_helical_symmetry");
       if (!is_empty_table(t)) {
            idx1 = t->GetColumnIndex("rotation_per_n_subunits");
            idx2 = t->GetColumnIndex("rise_per_n_subunits");
            idx3 = t->GetColumnIndex("n_subunits_divisor");
            idx4 = t->GetColumnIndex("dyad_axis");
            idx5 = t->GetColumnIndex("circular_symmetry");
            cs1.Clear();
            t->GetCell(cs1, idx1, 0);
            if (cs1 == "?" || cs1 == ".") cs1.Clear();
            cs2.Clear();
            t->GetCell(cs2, idx2, 0);
            if (cs2 == "?" || cs2 == ".") cs2.Clear();
            cs3.Clear();
            t->GetCell(cs3, idx3, 0);
            if (cs3 == "?" || cs3 == ".") cs3.Clear();
            cs4.Clear();
            t->GetCell(cs4, idx4, 0);
            if (cs4 == "?" || cs4 == ".") cs4.Clear();
            cs5.Clear();
            t->GetCell(cs5, idx5, 0);
            if (cs5 == "?" || cs5 == ".") cs5.Clear();

            char twist[100], height[100];
            memset(twist, 0, 100);
            memset(height, 0, 100);
            if (atoi(cs3.Text()) > 1) {
                 sprintf(twist, "%.2f/%d", atof(cs1.Text()), atoi(cs3.Text()));
                 sprintf(height, "%.2f/%d", atof(cs2.Text()), atoi(cs3.Text()));
            } else {
                 sprintf(twist, "%.2f", atof(cs1.Text()));
                 sprintf(height, "%.2f", atof(cs2.Text()));
            }

            remark.Copy("THE ASSEMBLY REPRESENTED IN THIS ENTRY HAS REGULAR");
            newCard = _CardInfo[REMARK_TOKEN-1].
                             add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(300, 1);
            newCard->AddElementToField(remark.Text(), 2);

            remark.Copy("HELICAL SYMMETRY WITH THE FOLLOWING PARAMETERS:");
            newCard = _CardInfo[REMARK_TOKEN-1].
                             add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(300, 1);
            newCard->AddElementToField(remark.Text(), 2);

            remark.Copy("ROTATION PER SUBUNIT (TWIST) = ");
            remark += twist;
            remark += " DEGREES";
            newCard = _CardInfo[REMARK_TOKEN-1].
                             add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(300, 1);
            newCard->AddElementToField(remark.Text(), 2);

            remark.Copy("RISE PER SUBUNIT (HEIGHT) = ");
            remark += height;
            remark += " ANGSTROMS";
            newCard = _CardInfo[REMARK_TOKEN-1].
                             add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(300, 1);
            newCard->AddElementToField(remark.Text(), 2);

            if (atoi(cs5.Text()) > 1) {
                 remark.Copy("IN ADDITION, THERE IS ");
                 remark += cs5;
                 remark += "-FOLD CIRCULAR";
                 newCard = _CardInfo[REMARK_TOKEN-1].
                                  add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(300, 1);
                 newCard->AddElementToField(remark.Text(), 2);

                 remark.Copy("SYMMETRY AROUND THE HELIX AXIS");
                 newCard = _CardInfo[REMARK_TOKEN-1].
                                  add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(300, 1);
                 newCard->AddElementToField(remark.Text(), 2);
            }

            cs4.ToLower();
            if (cs4 == "yes") {
                 remark.Copy("IN ADDITION, THERE IS DYAD SYMMETRY");
                 newCard = _CardInfo[REMARK_TOKEN-1].
                                  add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(300, 1);
                 newCard->AddElementToField(remark.Text(), 2);

                 remark.Copy("PERPENDICULAR TO THE HELIX AXIS");
                 newCard = _CardInfo[REMARK_TOKEN-1].
                                  add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(300, 1);
                 newCard->AddElementToField(remark.Text(), 2);
            }
       }

       pCard = _CardInfo[card_id].Head();
       while (pCard) pCard = _CardInfo[card_id].delete_a_card(pCard);

       return 1;
}
/*
int _MAXIT::check_virus_ncs_remark_300()
{
       int n_ncs = 0, card_id, i;
       NCS_COR *ncs = NULL;

       card_id = find_card_id("NCSRMK");
       if (card_id < 0) return 0;

       One_Card_Info_Format *pCard = _CardInfo[card_id].Head();
       if (pCard == NULL) return 0;

       while (pCard) {
            for (i = 0; i < n_ncs; i++) {
                 if (ncs[i].oper_id == atoi(pCard->Field(2))) break;
            }
            if (i == n_ncs) {
                 n_ncs++;
                 ncs = (NCS_COR *) realloc(ncs, n_ncs * sizeof(NCS_COR));
                 ncs[i].oper_id = atoi(pCard->Field(2));
                 strcpy(ncs[i].chain[0], pCard->Field(4));
                 ncs[i].num_chain = 1;
            } else {
                 strcpy(ncs[i].chain[ncs[i].num_chain], pCard->Field(4));
                 ncs[i].num_chain++;
            }
            pCard = pCard->Next();
       }

       if (!n_ncs) return 0;

       ndb_to_pdb_get_general_remark(Num_Remark_300_VIRUS, Remark_300_VIRUS, 0, 0, 0, 0);

       if (ncs) free ((void *) ncs);
       return 1;
}
*/
int _MAXIT::check_virus_ncs_remark_350()
{
       int n_ncs = 0, card_id, i, j, k, n, field_no, str_len;
       char FieldFormat[82], buffer[60];
       NCS_COR *ncs = NULL;
       One_Card_Info_Format *pCard = NULL, *newCard = NULL;

       card_id = find_card_id("NCSRMK");
       if (card_id < 0) return 0;

       pCard = _CardInfo[card_id].Head();
       if (pCard == NULL) return 0;

       while (pCard) {
            for (i = 0; i < n_ncs; i++) {
                 if (ncs[i].oper_id == atoi(pCard->Field(2))) break;
            }
            if (i == n_ncs) {
                 n_ncs++;
                 ncs = (NCS_COR *) realloc(ncs, n_ncs * sizeof(NCS_COR));
                 ncs[i].oper_id = atoi(pCard->Field(2));
                 strcpy(ncs[i].chain[0], pCard->Field(4));
                 ncs[i].num_chain = 1;
            } else {
                 strcpy(ncs[i].chain[ncs[i].num_chain], pCard->Field(4));
                 ncs[i].num_chain++;
            }
            pCard = pCard->Next();
       }

       if (!n_ncs) return 0;

       ndb_to_pdb_get_general_remark(Num_Remark_350, Remark_350, 0, 0, 0, 0);

       for (n = 0; n < n_ncs; n++) {
            if (!n) {
                 for (i = 0; i < 2; i++) {
                      newCard = _CardInfo[REMARK_TOKEN-1].
                                 add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                      newCard->AddElementToField(Remark_SYMMA_CHAIN[i].Remark_No, 1);
                      strcpy(buffer, "");
                      for (j = 0; j < Remark_SYMMA_CHAIN[i].NumField; j++) {
                           get_remark_field_format(FieldFormat,
                                    &Remark_SYMMA_CHAIN[i], j);
                           field_no = Remark_SYMMA_CHAIN[i].FieldList[j].FieldId-1;
                           str_len = strlen(buffer);
                           if (strcmp(Remark_SYMMA_CHAIN[i].FieldList[j].TokenName,
                                        "SYMMA") || field_no == 0) {
                                sprintf(&buffer[str_len], "%s",
                                    Remark_SYMMA_CHAIN[i].FieldList[j].Text);
                           } else {
                                for (k = 0; k < ncs[n].num_chain; k++) {
                                     str_len = strlen(buffer);
                                     if (k) sprintf(&buffer[str_len], ", %s",
                                                  ncs[n].chain[k]);
                                     else sprintf(&buffer[str_len], " %s",
                                                  ncs[n].chain[k]);
                                }
                           }
                      }
                      newCard->AddElementToField(buffer, 2);
                 }
            }
            double result[3][4];
            if (!get_matrix(result, ncs[n].oper_id)) continue;

            for (k = 0; k < 3; k++) {
                 newCard = _CardInfo[REMARK_TOKEN-1].
                          add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(Remark_SYMMA.Remark_No, 1);
                 for (j = 0; j < Remark_SYMMA.NumField; j++) {
                      print_space_between_remark_field(newCard, &Remark_SYMMA, j);
                      get_remark_field_format(FieldFormat, &Remark_SYMMA, j);
                      field_no = Remark_SYMMA.FieldList[j].FieldId-1;
                      if (!strcmp("BIOMT", Remark_SYMMA.FieldList[j].Text))
                           sprintf(buffer, "%s%d",
                               Remark_SYMMA.FieldList[j].Text, k + 1);
                      else if (field_no == 1)
                           sprintf(buffer, FieldFormat, (double) (n+1));
                      else if (j != Remark_SYMMA.NumField - 1)
                           sprintf(buffer, FieldFormat, result[k][j-2]);
                      else sprintf(buffer, FieldFormat, result[k][3]);
                      newCard->AttachToField(buffer, 2);
                 }
            }
       }

       if (ncs) free ((void *) ncs);
       return 1;
}

int _MAXIT::generate_virus_remark_350()
{
       if (!_CifObj) return 0;
 
       ISTable *t = _CifObj->GetTablePtr(_StructureId, "pdbx_struct_legacy_oper_list");
       if (is_empty_table(t)) return 0;

       CifString cs, assembly_id;
       assembly_id.Clear();
       t = _CifObj->GetTablePtr(_StructureId, "pdbx_struct_assembly");
       if (!is_empty_table(t)) {
            int rowNo = t->GetNumRows();
            int index = t->GetColumnIndex("id");
            int index1 = t->GetColumnIndex("details");
            for (int i = 0; i < rowNo; i++) {
                 get_value(cs, t, index1, i);
                 cs.ToLower();
                 if (cs == "representative helical assembly" ||
                     cs == "complete point assembly" ||
                     cs == "complete icosahedral assembly") {
                      get_value(assembly_id, t, index, i);
                      break;
                 }
            }
       }

       CifString entity_inst_id;
       entity_inst_id.Clear();
       t = _CifObj->GetTablePtr(_StructureId, "pdbx_struct_assembly_gen");
       if (!is_empty_table(t)) {
            int rowNo = t->GetNumRows();
            int index = t->GetColumnIndex("assembly_id");
            // int index1 = t->GetColumnIndex("entity_inst_id");
            int index1 = t->GetColumnIndex("asym_id_list");
            for (int i = 0; i < rowNo; i++) {
                 get_value(cs, t, index, i);
                 if (cs == assembly_id) {
                      get_value(entity_inst_id, t, index1, i);
                      break;
                 }
            }
            t = _CifObj->GetTablePtr(_StructureId, "struct_asym");
            if (entity_inst_id != "" && !is_empty_table(t)) {
                 int len = strlen(entity_inst_id.Text()) + 1;
                 char *buff = new char[len];
                 memset(buff, 0, len);
                 strcpy(buff, entity_inst_id.Text());
                 for (unsigned int i = 0; i < strlen(buff); i++) {
                      if (buff[i] == ',' || buff[i] == ';') buff[i] = ' ';
                 }
                 ndb_clean_string(buff);

                 rowNo = t->GetNumRows();
                 index = t->GetColumnIndex("id");
                 index1 = t->GetColumnIndex("ndb_pdb_id");
                 char *p = buff;
                 char buff1[10];
                 entity_inst_id.Clear();
                 while (p) {
                      p = get_word(p, buff1);
                      if (!strcmp(buff1, "")) continue;
                      for (int i = 0; i < rowNo; i++) {
                           get_value(cs, t, index, i);
                           if (cs == buff1) {
                                get_value(cs, t, index1, i);
                                if (cs != "") {
                                     if (entity_inst_id != "") entity_inst_id += ", ";
                                     entity_inst_id += cs;
                                }
                                break;
                           }
                      }
                 }
            }
       }

       ndb_to_pdb_get_general_remark(Num_Remark_350, Remark_350, 0, 0, 0, 0);

       int i, j, k, field_no;
       One_Card_Info_Format *newCard = NULL;
       for (i = 0; i < 2; i++) {
            newCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(Remark_SYMMA_CHAIN[i].Remark_No, 1);
            cs.Clear();
            for (j = 0; j < Remark_SYMMA_CHAIN[i].NumField; j++) {
                 field_no = Remark_SYMMA_CHAIN[i].FieldList[j].FieldId-1;
                 if (strcmp(Remark_SYMMA_CHAIN[i].FieldList[j].TokenName,
                              "SYMMA") || field_no == 0) {
                      cs += Remark_SYMMA_CHAIN[i].FieldList[j].Text;
                 } else {
                      cs += " ";
                      cs += entity_inst_id;
                 }
            }
            newCard->AddElementToField(cs.Text(), 2);
       }

       t = _CifObj->GetTablePtr(_StructureId, "pdbx_struct_legacy_oper_list");
       double result[3][4];
       char FieldFormat[82], buffer[60];
       int rowNo = t->GetNumRows();
       int index11 = t->GetColumnIndex("matrix[1][1]");
       int index12 = t->GetColumnIndex("matrix[1][2]");
       int index13 = t->GetColumnIndex("matrix[1][3]");
       int index1  = t->GetColumnIndex("vector[1]");
       int index21 = t->GetColumnIndex("matrix[2][1]");
       int index22 = t->GetColumnIndex("matrix[2][2]");
       int index23 = t->GetColumnIndex("matrix[2][3]");
       int index2  = t->GetColumnIndex("vector[2]");
       int index31 = t->GetColumnIndex("matrix[3][1]");
       int index32 = t->GetColumnIndex("matrix[3][2]");
       int index33 = t->GetColumnIndex("matrix[3][3]");
       int index3  = t->GetColumnIndex("vector[3]");
       for (i = 0; i < rowNo; i++) {
            get_value(cs, t, index11, i);
            result[0][0] = atof(cs.Text());
            get_value(cs, t, index12, i);
            result[0][1] = atof(cs.Text());
            get_value(cs, t, index13, i);
            result[0][2] = atof(cs.Text());
            get_value(cs, t, index1, i);
            result[0][3] = atof(cs.Text());
            get_value(cs, t, index21, i);
            result[1][0] = atof(cs.Text());
            get_value(cs, t, index22, i);
            result[1][1] = atof(cs.Text());
            get_value(cs, t, index23, i);
            result[1][2] = atof(cs.Text());
            get_value(cs, t, index2, i);
            result[1][3] = atof(cs.Text());
            get_value(cs, t, index31, i);
            result[2][0] = atof(cs.Text());
            get_value(cs, t, index32, i);
            result[2][1] = atof(cs.Text());
            get_value(cs, t, index33, i);
            result[2][2] = atof(cs.Text());
            get_value(cs, t, index3, i);
            result[2][3] = atof(cs.Text());

            for (k = 0; k < 3; k++) {
                 newCard = _CardInfo[REMARK_TOKEN-1].
                          add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(Remark_SYMMA.Remark_No, 1);
                 for (j = 0; j < Remark_SYMMA.NumField; j++) {
                      print_space_between_remark_field(newCard, &Remark_SYMMA, j);
                      get_remark_field_format(FieldFormat, &Remark_SYMMA, j);
                      field_no = Remark_SYMMA.FieldList[j].FieldId-1;
                      if (!strcmp("BIOMT", Remark_SYMMA.FieldList[j].Text))
                           sprintf(buffer, "%s%d",
                               Remark_SYMMA.FieldList[j].Text, k + 1);
                      else if (field_no == 1)
                           sprintf(buffer, FieldFormat, (double) (i+1));
                      else if (j != Remark_SYMMA.NumField - 1)
                           sprintf(buffer, FieldFormat, result[k][j-2]);
                      else sprintf(buffer, FieldFormat, result[k][3]);
                      newCard->AttachToField(buffer, 2);
                 }
            }
       }

       return 1;
}

void _MAXIT::rms_deviation_of_bonds()
{
       int i, len, rowNo, auth_atom_id_1, auth_asym_id_1, auth_comp_id_1, auth_seq_id_1;
       int insert_id_1, auth_atom_id_2, auth_asym_id_2, auth_comp_id_2, auth_seq_id_2;
       int insert_id_2, bond_deviation, a_resnum, b_resnum, model_id, Mol_ID = 0;
       char buffer[80], a_atomtyp[5], b_atomtyp[5];
       CifString cs, a_chnid, a_resnam, a_inscode, cifstring, b_chnid, b_resnam, b_inscode;
       float bond_length;
       ISTable *Table = NULL;
       One_Card_Info_Format *newCard = NULL;
       
       if (!_CifObj) return;
       Table = _CifObj->GetTablePtr(_StructureId, "pdbx_validate_rmsd_bond");
       if (is_empty_table(Table)) return;

       ndb_to_pdb_get_general_remark(Num_Remark_500_BOND, Remark_500_BOND, 0, 0, 0, 0);

       TblIndexObj Index;
       Index.Clear();

       rowNo = Table->GetNumRows();
       model_id = Table->GetColumnIndex("PDB_model_num");
       auth_atom_id_1 = Table->GetColumnIndex("auth_atom_id_1");
       auth_asym_id_1 = Table->GetColumnIndex("auth_asym_id_1");
       auth_comp_id_1 = Table->GetColumnIndex("auth_comp_id_1");
       auth_seq_id_1 = Table->GetColumnIndex("auth_seq_id_1");
       insert_id_1 = Table->GetColumnIndex("PDB_ins_code_1");
       auth_atom_id_2 = Table->GetColumnIndex("auth_atom_id_2");
       auth_asym_id_2 = Table->GetColumnIndex("auth_asym_id_2");
       auth_comp_id_2 = Table->GetColumnIndex("auth_comp_id_2");
       auth_seq_id_2 = Table->GetColumnIndex("auth_seq_id_2");
       insert_id_2 = Table->GetColumnIndex("PDB_ins_code_2");
       bond_deviation = Table->GetColumnIndex("bond_deviation");
       for (i = 0; i < rowNo; i++) {
            if (i >= 50) {
                 newCard = _CardInfo[REMARK_TOKEN-1].
                            add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(500, 1);
                 newCard = _CardInfo[REMARK_TOKEN-1].
                            add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(500, 1);
                 sprintf(buffer, "THIS ENTRY HAS %7d BOND DEVIATIONS.", rowNo);
                 newCard->AddElementToField(buffer, 2);
                 break;
            }
            get_value(a_chnid, Table, auth_asym_id_1, i);
            cs.Copy(a_chnid);
            get_value(a_resnam, Table, auth_comp_id_1, i);
            cs += "_" + a_resnam;
            get_value(cifstring, Table, auth_seq_id_1, i);
            cs += "_" + cifstring;
            a_resnum = atoi(cifstring.Text());
            get_value(a_inscode, Table, insert_id_1, i);
            cs += "_" + a_inscode;
            get_value(cifstring, Table, auth_atom_id_1, i);
            cs += "_" + cifstring;
            print_atom_atomName_field("", 0, a_atomtyp, cifstring.Text(), a_resnam.Text(), 0);
            if (a_chnid == "") a_chnid.Copy(" ");
            if (a_inscode == "") a_inscode.Copy(" ");

            get_value(b_chnid, Table, auth_asym_id_2, i);
            cs += "_" + b_chnid;
            get_value(b_resnam, Table, auth_comp_id_2, i);
            cs += "_" + b_resnam;
            get_value(cifstring, Table, auth_seq_id_2, i);
            cs += "_" + cifstring;
            b_resnum = atoi(cifstring.Text());
            get_value(b_inscode, Table, insert_id_2, i);
            cs += "_" + b_inscode;
            get_value(cifstring, Table, auth_atom_id_2, i);
            cs += "_" + cifstring;
            print_atom_atomName_field("", 0, b_atomtyp, cifstring.Text(), b_resnam.Text(), 0);
            if (b_chnid == "") b_chnid.Copy(" ");
            if (b_inscode == "") b_inscode.Copy(" ");

            get_value(cifstring, Table, bond_deviation, i);
            bond_length = atof(cifstring.Text());
            get_value(cifstring, Table, model_id, i);
            cs += "_" + cifstring;
            Mol_ID = atoi(cifstring.Text());

            if (Index.Seek(cs) >= 0) continue;
            Index.Add(cs);

            if (_num_mol > 1)
                 sprintf(buffer, "%2d %3s %c%4d%c %4s   %3s %c%4d%c %4s", Mol_ID,
                   a_resnam.Text(), a_chnid.Text()[0], a_resnum, a_inscode.Text()[0], a_atomtyp,
                   b_resnam.Text(), b_chnid.Text()[0], b_resnum, b_inscode.Text()[0], b_atomtyp);
            else sprintf(buffer, "   %3s %c%4d%c %4s   %3s %c%4d%c %4s",
                   a_resnam.Text(), a_chnid.Text()[0], a_resnum, a_inscode.Text()[0], a_atomtyp,
                   b_resnam.Text(), b_chnid.Text()[0], b_resnum, b_inscode.Text()[0], b_atomtyp);
            len = strlen(buffer);
            sprintf(&buffer[len], "    %6.3f", bond_length);

            newCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(500, 1);
            newCard->AddElementToField(buffer, 2);
       }
       newCard = _CardInfo[REMARK_TOKEN-1].
                  add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
       newCard->AddElementToField(500, 1);
       newCard = _CardInfo[REMARK_TOKEN-1].
                  add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
       newCard->AddElementToField(500, 1);
       newCard->AddElementToField("REMARK: NULL", 2);
}

void _MAXIT::rms_deviation_of_angles()
{
       int i, len, rowNo, auth_atom_id_1, auth_asym_id_2, auth_comp_id_2, auth_seq_id_2;
       int insert_id_2, auth_atom_id_2, auth_atom_id_3, resnum_2, model_id, Mol_ID = 0;
       int angle_deviation;
       char atomtyp_1[5], atomtyp_2[5], atomtyp_3[5], buffer[80];
       CifString cs, cifstring, chnid_2, resnam_2, inscode_2;
       float angle_value;
       ISTable *Table = NULL;
       One_Card_Info_Format *newCard = NULL;
       
       if (!_CifObj) return;
       Table = _CifObj->GetTablePtr(_StructureId, "pdbx_validate_rmsd_angle");
       if (is_empty_table(Table)) return;

       ndb_to_pdb_get_general_remark(Num_Remark_500_ANGLE, Remark_500_ANGLE, 0, 0, 0, 0);

       TblIndexObj Index;
       Index.Clear();

       rowNo = Table->GetNumRows();
       model_id = Table->GetColumnIndex("PDB_model_num");
       auth_atom_id_1 = Table->GetColumnIndex("auth_atom_id_1");
/*
       auth_asym_id_2 = Table->GetColumnIndex("auth_asym_id_3");
       auth_comp_id_2 = Table->GetColumnIndex("auth_comp_id_3");
       auth_seq_id_2 = Table->GetColumnIndex("auth_seq_id_3");
       insert_id_2 = Table->GetColumnIndex("PDB_ins_code_3");
*/
       auth_asym_id_2 = Table->GetColumnIndex("auth_asym_id_2");
       auth_comp_id_2 = Table->GetColumnIndex("auth_comp_id_2");
       auth_seq_id_2 = Table->GetColumnIndex("auth_seq_id_2");
       insert_id_2 = Table->GetColumnIndex("PDB_ins_code_2");
       auth_atom_id_2 = Table->GetColumnIndex("auth_atom_id_2");
       auth_atom_id_3 = Table->GetColumnIndex("auth_atom_id_3");
       angle_deviation = Table->GetColumnIndex("angle_deviation");
       for (i = 0; i < rowNo; i++) {
            if (i >= 50) {
                 newCard = _CardInfo[REMARK_TOKEN-1].
                            add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(500, 1);
                 newCard = _CardInfo[REMARK_TOKEN-1].
                            add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(500, 1);
                 sprintf(buffer, "THIS ENTRY HAS %7d ANGLE DEVIATIONS.", rowNo);
                 newCard->AddElementToField(buffer, 2);
                 break;
            }
            get_value(chnid_2, Table, auth_asym_id_2, i);
            cs.Copy(chnid_2);
            get_value(resnam_2, Table, auth_comp_id_2, i);
            cs += "_" + resnam_2;
            get_value(cifstring, Table, auth_seq_id_2, i);
            cs += "_" + cifstring;
            resnum_2 = atoi(cifstring.Text());
            get_value(inscode_2, Table, insert_id_2, i);
            cs += "_" + inscode_2;
            get_value(cifstring, Table, auth_atom_id_1, i);
            cs += "_" + cifstring;
            print_atom_atomName_field("", 0, atomtyp_1, cifstring.Text(), resnam_2.Text(), 0);
            get_value(cifstring, Table, auth_atom_id_2, i);
            cs += "_" + cifstring;
            print_atom_atomName_field("", 0, atomtyp_2, cifstring.Text(), resnam_2.Text(), 0);
            get_value(cifstring, Table, auth_atom_id_3, i);
            cs += "_" + cifstring;
            print_atom_atomName_field("", 0, atomtyp_3, cifstring.Text(), resnam_2.Text(), 0);
            get_value(cifstring, Table, angle_deviation, i);
            angle_value = atof(cifstring.Text());
            get_value(cifstring, Table, model_id, i);
            cs += "_" + cifstring;
            Mol_ID = atoi(cifstring.Text());

            if (Index.Seek(cs) >= 0) continue;
            Index.Add(cs);

            if (chnid_2 == "") chnid_2.Copy(" ");
            if (inscode_2 == "") inscode_2.Copy(" ");

            if (_num_mol > 1)
                 sprintf(buffer, "%2d %3s %c%4d%c", Mol_ID, resnam_2.Text(), chnid_2.Text()[0],
                      resnum_2, inscode_2.Text()[0]);
            else sprintf(buffer, "   %3s %c%4d%c", resnam_2.Text(), chnid_2.Text()[0],
                      resnum_2, inscode_2.Text()[0]);
            len = strlen(buffer);
            sprintf(&buffer[len], " %4s - %4s - %4s ANGL. DEV. = %5.1f DEGREES",
                      atomtyp_1, atomtyp_2, atomtyp_3, angle_value); 
            
            newCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(500, 1);
            newCard->AddElementToField(buffer, 2);
       }
       newCard = _CardInfo[REMARK_TOKEN-1].
                  add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
       newCard->AddElementToField(500, 1);
       newCard = _CardInfo[REMARK_TOKEN-1].
                  add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
       newCard->AddElementToField(500, 1);
       newCard->AddElementToField("REMARK: NULL", 2);
}

void _MAXIT::outside_torsions()
{
       int i, rowNo, model_id, auth_comp_id, auth_asym_id, auth_seq_id, insert_id;
       int phi_index, psi_index, Mol_ID, resnum;
       char buffer[80];
       CifString cs, cifstring, resnam, chnid, inscode;
       float phi, psi;
       ISTable *Table = NULL;
       One_Card_Info_Format *newCard = NULL;

       if (!_CifObj) return;
       Table = _CifObj->GetTablePtr(_StructureId, "pdbx_validate_torsion");
       if (is_empty_table(Table)) return;

       ndb_to_pdb_get_general_remark(Num_Remark_500_TORSION, Remark_500_TORSION, 0, 0, 0, 0);

       TblIndexObj Index;
       Index.Clear();

       rowNo = Table->GetNumRows();
       model_id = Table->GetColumnIndex("PDB_model_num");
       auth_comp_id = Table->GetColumnIndex("auth_comp_id");
       auth_asym_id = Table->GetColumnIndex("auth_asym_id");
       auth_seq_id = Table->GetColumnIndex("auth_seq_id");
       insert_id = Table->GetColumnIndex("PDB_ins_code");
       phi_index = Table->GetColumnIndex("phi");
       psi_index = Table->GetColumnIndex("psi");
       for (i = 0; i < rowNo; i++) {
            if (i >= 50) {
                 newCard = _CardInfo[REMARK_TOKEN-1].
                            add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(500, 1);
                 newCard = _CardInfo[REMARK_TOKEN-1].
                            add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(500, 1);
                 sprintf(buffer, "THIS ENTRY HAS %7d RAMACHANDRAN OUTLIERS.", rowNo);
                 newCard->AddElementToField(buffer, 2);
                 break;
            }
            get_value(cifstring, Table, model_id, i);
            cs.Copy(cifstring);
            Mol_ID = atoi(cifstring.Text());
            get_value(resnam, Table, auth_comp_id, i);
            cs += "_" + resnam;
            get_value(chnid, Table, auth_asym_id, i);
            cs += "_" + chnid;
            get_value(cifstring, Table, auth_seq_id, i);
            cs += "_" + cifstring;
            resnum = atoi(cifstring.Text());
            get_value(inscode, Table, insert_id, i);
            cs += "_" + inscode;
            get_value(cifstring, Table, phi_index, i);
            phi = atof(cifstring.Text());
            get_value(cifstring, Table, psi_index, i);
            psi = atof(cifstring.Text());

            if (Index.Seek(cs) >= 0) continue;
            Index.Add(cs);
 
            if (chnid == "") chnid.Copy(" ");
            if (inscode == "") inscode.Copy(" ");

            if (_num_mol > 1)
                 sprintf(buffer, "%2d %3s %c%4d%c    %7.2f   %7.2f", Mol_ID,
                     resnam.Text(), chnid.Text()[0], resnum, inscode.Text()[0], psi, phi);
            else sprintf(buffer, "   %3s %c%4d%c    %7.2f   %7.2f",
                     resnam.Text(), chnid.Text()[0], resnum, inscode.Text()[0], psi, phi);

            newCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(500, 1);
            newCard->AddElementToField(buffer, 2);
       }
       newCard = _CardInfo[REMARK_TOKEN-1].
                  add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
       newCard->AddElementToField(500, 1);
       newCard = _CardInfo[REMARK_TOKEN-1].
                  add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
       newCard->AddElementToField(500, 1);
       newCard->AddElementToField("REMARK: NULL", 2);
}

void _MAXIT::non_cis_trans_torsions()
{
       int i, rowNo, model_id, auth_asym_id_1, auth_comp_id_1, auth_seq_id_1;
       int insert_id_1, auth_asym_id_2, auth_comp_id_2, auth_seq_id_2, insert_id_2;
       int omega_index, a_resnum, b_resnum, Mol_ID;
       char buffer[80];
       CifString cs, a_chnid, a_resnam, a_inscode, b_chnid, b_resnam, b_inscode, cifstring;
       float omega;
       ISTable *Table = NULL;
       One_Card_Info_Format *newCard = NULL;

       if (!_CifObj) return;
       Table = _CifObj->GetTablePtr(_StructureId, "pdbx_validate_peptide_omega");
       if (is_empty_table(Table)) return;

       ndb_to_pdb_get_general_remark(Num_Remark_500_NON_CIS, Remark_500_NON_CIS, 0, 0, 0, 0);

       TblIndexObj Index;
       Index.Clear();

       rowNo = Table->GetNumRows();
       model_id = Table->GetColumnIndex("PDB_model_num");
       auth_asym_id_1 = Table->GetColumnIndex("auth_asym_id_1");
       auth_comp_id_1 = Table->GetColumnIndex("auth_comp_id_1");
       auth_seq_id_1 = Table->GetColumnIndex("auth_seq_id_1");
       insert_id_1 = Table->GetColumnIndex("PDB_ins_code_1");
       auth_asym_id_2 = Table->GetColumnIndex("auth_asym_id_2");
       auth_comp_id_2 = Table->GetColumnIndex("auth_comp_id_2");
       auth_seq_id_2 = Table->GetColumnIndex("auth_seq_id_2");
       insert_id_2 = Table->GetColumnIndex("PDB_ins_code_2");
       omega_index = Table->GetColumnIndex("omega");

       for (i = 0; i < rowNo; i++) {
            if (i >= 50) {
                 newCard = _CardInfo[REMARK_TOKEN-1].
                            add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(500, 1);
                 newCard = _CardInfo[REMARK_TOKEN-1].
                            add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(500, 1);
                 sprintf(buffer, "THIS ENTRY HAS %7d RAMACHANDRAN OUTLIERS.", rowNo);
                 newCard->AddElementToField(buffer, 2);
                 break;
            }
            get_value(a_chnid, Table, auth_asym_id_1, i);
            cs.Copy(a_chnid);
            get_value(a_resnam, Table, auth_comp_id_1, i);
            cs += "_" + a_resnam;
            get_value(cifstring, Table, auth_seq_id_1, i);
            cs += "_" + cifstring;
            a_resnum = atoi(cifstring.Text());
            get_value(a_inscode, Table, insert_id_1, i);
            cs += "_" + a_inscode;
            if (a_chnid == "") a_chnid.Copy(" ");
            if (a_inscode == "") a_inscode.Copy(" ");

            get_value(b_chnid, Table, auth_asym_id_2, i);
            cs += "_" + b_chnid;
            get_value(b_resnam, Table, auth_comp_id_2, i);
            cs += "_" + b_resnam;
            get_value(cifstring, Table, auth_seq_id_2, i);
            cs += "_" + cifstring;
            b_resnum = atoi(cifstring.Text());
            get_value(b_inscode, Table, insert_id_2, i);
            cs += "_" + b_inscode;
            if (b_chnid == "") b_chnid.Copy(" ");
            if (b_inscode == "") b_inscode.Copy(" ");

            get_value(cifstring, Table, omega_index, i);
            omega = atof(cifstring.Text());
            get_value(cifstring, Table, model_id, i);
            cs += "_" + cifstring;
            Mol_ID = atoi(cifstring.Text());

            if (Index.Seek(cs) >= 0) continue;
            Index.Add(cs);

            if (_num_mol > 1)
                 sprintf(buffer, "%3s %c %4d%c    %3s %c %4d%c       %3d      %7.2f",
                  a_resnam.Text(), a_chnid.Text()[0], a_resnum, a_inscode.Text()[0],
                  b_resnam.Text(), b_chnid.Text()[0], b_resnum, b_inscode.Text()[0],
                  Mol_ID, omega);
            else sprintf(buffer, "%3s %c %4d%c    %3s %c %4d%c                %7.2f",
                  a_resnam.Text(), a_chnid.Text()[0], a_resnum, a_inscode.Text()[0],
                  b_resnam.Text(), b_chnid.Text()[0], b_resnum, b_inscode.Text()[0], omega);

            newCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(500, 1);
            newCard->AddElementToField(buffer, 2);
       }
       newCard = _CardInfo[REMARK_TOKEN-1].
                  add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
       newCard->AddElementToField(500, 1);
       newCard = _CardInfo[REMARK_TOKEN-1].
                  add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
       newCard->AddElementToField(500, 1);
       newCard->AddElementToField("REMARK: NULL", 2);
}

void _MAXIT::side_chain_plane()
{
       if (!_CifObj) return;
       ISTable *Table = _CifObj->GetTablePtr(_StructureId, "pdbx_validate_planes");
       if (is_empty_table(Table)) return;

       ndb_to_pdb_get_general_remark(Num_Remark_500_SPLANE, Remark_500_SPLANE, 0, 0, 0, 0);

       TblIndexObj Index;
       Index.Clear();

       int rowNo = Table->GetNumRows();
       int model_id = Table->GetColumnIndex("PDB_model_num");
       int auth_asym_id = Table->GetColumnIndex("auth_asym_id");
       int auth_comp_id = Table->GetColumnIndex("auth_comp_id");
       int auth_seq_id = Table->GetColumnIndex("auth_seq_id");
       int insert_id = Table->GetColumnIndex("PDB_ins_code");
       int rmsd_index = Table->GetColumnIndex("rmsd");
       int type_index = Table->GetColumnIndex("type");

       CifString cs, chnid, resnam, inscode, cifstring;
       char buffer[100];
       One_Card_Info_Format *newCard = NULL;
       for (int i = 0; i < rowNo; i++) {
            if (i >= 50) {
                 newCard = _CardInfo[REMARK_TOKEN-1].
                            add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(500, 1);
                 newCard = _CardInfo[REMARK_TOKEN-1].
                            add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(500, 1);
                 sprintf(buffer, "THIS ENTRY HAS %7d PLANE DEVIATIONS.", rowNo);
                 newCard->AddElementToField(buffer, 2);
                 break;
            }
            get_value(chnid, Table, auth_asym_id, i);
            cs.Copy(chnid);
            get_value(resnam, Table, auth_comp_id, i);
            cs += "_" + resnam;
            get_value(cifstring, Table, auth_seq_id, i);
            cs += "_" + cifstring;
            int resnum = atoi(cifstring.Text());
            get_value(inscode, Table, insert_id, i);
            cs += "_" + inscode;
            if (chnid == "") chnid.Copy(" ");
            if (inscode == "") inscode.Copy(" ");

            get_value(cifstring, Table, rmsd_index, i);
            double rmsd = atof(cifstring.Text());
            get_value(cifstring, Table, model_id, i);
            cs += "_" + cifstring;
            int Mol_ID = atoi(cifstring.Text());

            if (Index.Seek(cs) >= 0) continue;
            Index.Add(cs);

            get_value(cifstring, Table, type_index, i);

            if (_num_mol > 1)
                 sprintf(buffer, "%2d %3s %c%4d%c        %4.2f    %s",
                       Mol_ID, resnam.Text(), chnid.Text()[0], resnum,
                          inscode.Text()[0], rmsd, cifstring.Text());
            else sprintf(buffer, "   %3s %c%4d%c        %4.2f    %s",
                       resnam.Text(), chnid.Text()[0], resnum,
                       inscode.Text()[0], rmsd, cifstring.Text());

            newCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(500, 1);
            newCard->AddElementToField(buffer, 2);
       }
       newCard = _CardInfo[REMARK_TOKEN-1].
                  add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
       newCard->AddElementToField(500, 1);
       newCard = _CardInfo[REMARK_TOKEN-1].
                  add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
       newCard->AddElementToField(500, 1);
       newCard->AddElementToField("REMARK: NULL", 2);
}

void _MAXIT::main_chain_planarity()
{
       if (!_CifObj) return;
       ISTable *Table = _CifObj->GetTablePtr(_StructureId, "pdbx_validate_main_chain_plane");
       if (is_empty_table(Table)) return;

       ndb_to_pdb_get_general_remark(Num_Remark_500_MPLANE, Remark_500_MPLANE, 0, 0, 0, 0);

       TblIndexObj Index;
       Index.Clear();

       int rowNo = Table->GetNumRows();
       int model_id = Table->GetColumnIndex("PDB_model_num");
       int auth_asym_id = Table->GetColumnIndex("auth_asym_id");
       int auth_comp_id = Table->GetColumnIndex("auth_comp_id");
       int auth_seq_id = Table->GetColumnIndex("auth_seq_id");
       int insert_id = Table->GetColumnIndex("PDB_ins_code");
       int pseudo_index = Table->GetColumnIndex("improper_torsion_angle");

       CifString cs, chnid, resnam, inscode, cifstring;
       char buffer[100];
       One_Card_Info_Format *newCard = NULL;
       for (int i = 0; i < rowNo; i++) {
            if (i >= 50) {
                 newCard = _CardInfo[REMARK_TOKEN-1].
                            add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(500, 1);
                 newCard = _CardInfo[REMARK_TOKEN-1].
                            add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(500, 1);
                 sprintf(buffer, "THIS ENTRY HAS %7d MAIN CHAIN PLANARITY DEVIATIONS.", rowNo);
                 newCard->AddElementToField(buffer, 2);
                 break;
            }
            get_value(chnid, Table, auth_asym_id, i);
            cs.Copy(chnid);
            get_value(resnam, Table, auth_comp_id, i);
            cs += "_" + resnam;
            get_value(cifstring, Table, auth_seq_id, i);
            cs += "_" + cifstring;
            int resnum = atoi(cifstring.Text());
            get_value(inscode, Table, insert_id, i);
            cs += "_" + inscode;
            if (chnid == "") chnid.Copy(" ");
            if (inscode == "") inscode.Copy(" ");

            get_value(cifstring, Table, pseudo_index, i);
            double pseudo_torsion = atof(cifstring.Text());
            get_value(cifstring, Table, model_id, i);
            cs += "_" + cifstring;
            int Mol_ID = atoi(cifstring.Text());

            if (Index.Seek(cs) >= 0) continue;
            Index.Add(cs);

            if (_num_mol > 1)
                 sprintf(buffer, "%2d %3s %c%4d%c      %7.2f",
                       Mol_ID, resnam.Text(), chnid.Text()[0], resnum,
                          inscode.Text()[0], pseudo_torsion);
            else sprintf(buffer, "   %3s %c%4d%c      %7.2f",
                       resnam.Text(), chnid.Text()[0], resnum,
                       inscode.Text()[0], pseudo_torsion);

            newCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(500, 1);
            newCard->AddElementToField(buffer, 2);
       }
       newCard = _CardInfo[REMARK_TOKEN-1].
                  add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
       newCard->AddElementToField(500, 1);
       newCard = _CardInfo[REMARK_TOKEN-1].
                  add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
       newCard->AddElementToField(500, 1);
       newCard->AddElementToField("REMARK: NULL", 2);
}

void _MAXIT::alpha_carbon_chiral()
{
       if (!_CifObj) return;
       ISTable *Table = _CifObj->GetTablePtr(_StructureId, "pdbx_validate_chiral");
       if (is_empty_table(Table)) return;

       ndb_to_pdb_get_general_remark(Num_Remark_500_CHIRAL, Remark_500_CHIRAL, 0, 0, 0, 0);

       TblIndexObj Index;
       Index.Clear();

       int rowNo = Table->GetNumRows();
       int model_id = Table->GetColumnIndex("PDB_model_num");
       int auth_asym_id = Table->GetColumnIndex("auth_asym_id");
       int auth_comp_id = Table->GetColumnIndex("auth_comp_id");
       int auth_seq_id = Table->GetColumnIndex("auth_seq_id");
       int insert_id = Table->GetColumnIndex("PDB_ins_code");
       int details_index = Table->GetColumnIndex("details");
       int angle_index = Table->GetColumnIndex("omega");

       CifString cs, chnid, resnam, inscode, details, cifstring;
       CifString expected, found;
       char buffer[100];
       One_Card_Info_Format *newCard = NULL;
       for (int i = 0; i < rowNo; i++) {
            if (i >= 50) {
                 newCard = _CardInfo[REMARK_TOKEN-1].
                            add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(500, 1);
                 newCard = _CardInfo[REMARK_TOKEN-1].
                            add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(500, 1);
                 sprintf(buffer, "THIS ENTRY HAS %7d CHIRALITY DEVIATIONS.", rowNo);
                 newCard->AddElementToField(buffer, 2);
                 break;
            }
            get_value(chnid, Table, auth_asym_id, i);
            cs.Copy(chnid);
            get_value(resnam, Table, auth_comp_id, i);
            cs += "_" + resnam;
            get_value(cifstring, Table, auth_seq_id, i);
            cs += "_" + cifstring;
            int resnum = atoi(cifstring.Text());
            get_value(inscode, Table, insert_id, i);
            cs += "_" + inscode;
            if (chnid == "") chnid.Copy(" ");
            if (inscode == "") inscode.Copy(" ");
            get_value(cifstring, Table, details_index, i);
            expected.Copy(" ");
            found.Copy(" ");
            if (strstr(cifstring.Text(), "Expecting L Found D")) {
                 expected.Copy("L");
                 found.Copy("D");
                 details.Copy(&(cifstring.Text()[20]));
            } else if (strstr(cifstring.Text(), "Expecting D Found L")) {
                 expected.Copy("D");
                 found.Copy("L");
                 details.Copy(&(cifstring.Text()[20]));
            } else if (strstr(cifstring.Text(), "Expecting D Found D")) {
                 expected.Copy("D");
                 found.Copy("D");
                 details.Copy(&(cifstring.Text()[20]));
            } else if (strstr(cifstring.Text(), "Expecting L Found L")) {
                 expected.Copy("L");
                 found.Copy("L");
                 details.Copy(&(cifstring.Text()[20]));
            } else if (strstr(cifstring.Text(), "Expecting R Found S")) {
                 expected.Copy("R");
                 found.Copy("S");
                 details.Copy(&(cifstring.Text()[20]));
            } else if (strstr(cifstring.Text(), "Expecting S Found R")) {
                 expected.Copy("S");
                 found.Copy("R");
                 details.Copy(&(cifstring.Text()[20]));
            } else details.Copy(cifstring);

            get_value(cifstring, Table, angle_index, i);
            double pseudo_torsion = atof(cifstring.Text());
            get_value(cifstring, Table, model_id, i);
            cs += "_" + cifstring;
            int Mol_ID = atoi(cifstring.Text());

            if (Index.Seek(cs) >= 0) continue;
            Index.Add(cs);
/*
            if (_num_mol > 1)
                 sprintf(buffer, "%2d %3s %c%4d%c %17s%7.2f",
                    Mol_ID, resnam.Text(), chnid.Text()[0], resnum,
                    inscode.Text()[0], details.Text(), pseudo_torsion);
            else sprintf(buffer, "   %3s %c%4d%c %17s%7.2f",
                    resnam.Text(), chnid.Text()[0], resnum,
                    inscode.Text()[0], details.Text(), pseudo_torsion);
*/
            if (_num_mol > 1)
                 sprintf(buffer, "%2d %3s %c%4d%c      %5.1f      %c          %c   %s",
                    Mol_ID, resnam.Text(), chnid.Text()[0], resnum,
                    inscode.Text()[0], pseudo_torsion, expected.Text()[0],
                    found.Text()[0], details.Text());
            else sprintf(buffer, "   %3s %c%4d%c      %5.1f      %c          %c   %s",
                    resnam.Text(), chnid.Text()[0], resnum,
                    inscode.Text()[0], pseudo_torsion, expected.Text()[0],
                    found.Text()[0], details.Text());

            newCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(500, 1);
            newCard->AddElementToField(buffer, 2);
       }
       newCard = _CardInfo[REMARK_TOKEN-1].
                  add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
       newCard->AddElementToField(500, 1);
       newCard = _CardInfo[REMARK_TOKEN-1].
                  add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
       newCard->AddElementToField(500, 1);
       newCard->AddElementToField("REMARK: NULL", 2);
}

void _MAXIT::output_close_contact(ISTable *Table, const int sym)
{
       int num = 0;
       char buffer[80], atomName[10];
       CifString a_atomtyp, a_chnid, a_resnam, a_inscode, cifstring;
       CifString b_atomtyp, b_chnid, b_resnam, b_inscode, symmetry, cs;
       float dist;
       _ATOM **atom_a = NULL, **atom_b = NULL;
       One_Card_Info_Format *newCard = NULL;

       TblIndexObj Index;
       Index.Clear();

       int model_index = Table->GetColumnIndex("PDB_model_num");
       int auth_atom_id_1 = Table->GetColumnIndex("auth_atom_id_1");
       int auth_asym_id_1 = Table->GetColumnIndex("auth_asym_id_1");
       int auth_comp_id_1 = Table->GetColumnIndex("auth_comp_id_1");
       int auth_seq_id_1 = Table->GetColumnIndex("auth_seq_id_1");
       int insert_id_1 = Table->GetColumnIndex("PDB_ins_code_1");
       int auth_atom_id_2 = Table->GetColumnIndex("auth_atom_id_2");
       int auth_asym_id_2 = Table->GetColumnIndex("auth_asym_id_2");
       int auth_comp_id_2 = Table->GetColumnIndex("auth_comp_id_2");
       int auth_seq_id_2 = Table->GetColumnIndex("auth_seq_id_2");
       int insert_id_2 = Table->GetColumnIndex("PDB_ins_code_2");
       int symmetry_op_2 = Table->GetColumnIndex("site_symmetry_2");
       int dist_index = Table->GetColumnIndex("dist");
       int rowNo = Table->GetNumRows();
       for (int i = 0; i < rowNo; i++) {
            get_value(cs, Table, model_index, i);
            if (atoi(cs.Text()) != _molecules[0]->Mol_ID()) break;
            if (i >= 50) {
                 newCard = _CardInfo[REMARK_TOKEN-1].
                            add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(500, 1);
                 newCard = _CardInfo[REMARK_TOKEN-1].
                            add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(500, 1);
                 if (sym)
                      sprintf(buffer, "THIS ENTRY HAS %7d SYMMETRY CONTACTS", rowNo);
                 else sprintf(buffer, "THIS ENTRY HAS %7d CLOSE CONTACTS", rowNo);
                 newCard->AddElementToField(buffer, 2);
                 break;
            }

            get_value(a_atomtyp, Table, auth_atom_id_1, i);
            cs.Copy(a_atomtyp);
            get_value(a_chnid, Table, auth_asym_id_1, i);
            cs += "_" + a_chnid;
            get_value(a_resnam, Table, auth_comp_id_1, i);
            cs += "_" + a_resnam;
            get_value(cifstring, Table, auth_seq_id_1, i);
            cs += "_" + cifstring;
            int a_resnum = atoi(cifstring.Text());
            get_value(a_inscode, Table, insert_id_1, i);
            cs += "_" + a_inscode;
            get_value(b_atomtyp, Table, auth_atom_id_2, i);
            cs += "_" + b_atomtyp;
            get_value(b_chnid, Table, auth_asym_id_2, i);
            cs += "_" + b_chnid;
            get_value(b_resnam, Table, auth_comp_id_2, i);
            cs += "_" + b_resnam;
            get_value(cifstring, Table, auth_seq_id_2, i);
            cs += "_" + cifstring;
            int b_resnum = atoi(cifstring.Text());
            get_value(b_inscode, Table, insert_id_2, i);
            cs += "_" + b_inscode;
            if (Index.Seek(cs) >= 0) continue;
            Index.Add(cs);

            get_value(symmetry, Table, symmetry_op_2, i);
            get_value(cifstring, Table, dist_index, i);
            dist = atof(cifstring.Text());

            strcpy(buffer, symmetry.Text());
            symmetry.Clear();
            int len = strlen(buffer); 
            for (int j = 0; j < len; j++) {
                 if (buffer[j] == '_') continue;
                 symmetry += buffer[j];
            }

            atom_a = _molecules[0]->find_pdb_atom(a_atomtyp.Text(), a_chnid.Text(),
                       a_resnam.Text(), a_resnum, a_inscode.Text(), num);
            atom_b = _molecules[0]->find_pdb_atom(b_atomtyp.Text(), b_chnid.Text(),
                       b_resnam.Text(), b_resnum, b_inscode.Text(), num);
            if (!atom_a || !atom_b) {
                 if (atom_a) delete [] atom_a;
                 if (atom_b) delete [] atom_b;
                 continue;
            }

            print_atom_atomName_field(atom_a[0]->atom_type(), 0, atomName,
                atom_a[0]->pdb_atmnam(), atom_a[0]->pdb_resnam(), atom_a[0]->align());
            sprintf(buffer, " %-4s  %3s", atomName, atom_a[0]->pdb_resnam());
            len = strlen(buffer);
            strcpy(atomName, " ");
            if (strcmp(atom_a[0]->pdb_chnid(), ""))
                 strcpy(atomName, atom_a[0]->pdb_chnid());
            sprintf(&buffer[len], " %c  %4d", atomName[0], 
                          atom_a[0]->pdb_resnum());
            len = strlen(buffer);
            strcpy(atomName, " ");
            if (strcmp(atom_a[0]->ins_code(), ""))
                 strcpy(atomName, atom_a[0]->ins_code());
            sprintf(&buffer[len], "%c  ", atomName[0]);

            print_atom_atomName_field(atom_b[0]->atom_type(), 0, atomName,
                atom_b[0]->pdb_atmnam(), atom_b[0]->pdb_resnam(), atom_b[0]->align());
            len = strlen(buffer);
            sprintf(&buffer[len], " %-4s  %3s", atomName,
                             atom_b[0]->pdb_resnam());
            len = strlen(buffer);
            strcpy(atomName, " ");
            if (strcmp(atom_b[0]->pdb_chnid(), ""))
                 strcpy(atomName, atom_b[0]->pdb_chnid());
            sprintf(&buffer[len], " %c  %4d", atomName[0],
                        atom_b[0]->pdb_resnum());
            len = strlen(buffer);
            strcpy(atomName, " ");
            if (strcmp(atom_b[0]->ins_code(), ""))
                 strcpy(atomName, atom_b[0]->ins_code());
            sprintf(&buffer[len], "%c", atomName[0]);

            if (atom_a) delete [] atom_a;
            if (atom_b) delete [] atom_b;

            len = strlen(buffer);
            if (!sym) sprintf(&buffer[len], "            %5.2f", dist);
            else sprintf(&buffer[len], "   %5s    %5.2f", symmetry.Text(), dist);

            newCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(500, 1);
            newCard->AddElementToField(buffer, 2);
       }

       newCard = _CardInfo[REMARK_TOKEN-1].
                  add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
       newCard->AddElementToField(500, 1);
       newCard = _CardInfo[REMARK_TOKEN-1].
                  add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
       newCard->AddElementToField(500, 1);
       newCard->AddElementToField("REMARK: NULL", 2);
}

void _MAXIT::get_cispep_cards()
{
       if (_n_torsion == 0 || _torsion == NULL) return;
       if (_torsionIndex.Root() < 0) return;

       int CISPEP_TOKEN = find_card_id("CISPEP");
       if (CISPEP_TOKEN == -1) return;
       _CardInfo[CISPEP_TOKEN].clear_Cards();
       
       _RESIDUE *residue = NULL;
       One_Card_Info_Format *pCard = NULL;
       _ATOM *atom = NULL;

       CifString index("OMEGA");
       int serial_no = 0;
       int ret = _torsionIndex.Seek(index);
       while (ret >= 0) {
            if (index != _torsionIndex.GetCurrData()) break;
            if (fabs(_torsion[ret]->val) > 30.0) {
                 ret = _torsionIndex.GoToNext();
                 continue;
            }

            residue = _molecules[0]->find_prev_residue(_torsion[ret]->ChainID,
                   _torsion[ret]->ResName, _torsion[ret]->ResNum,
                   _torsion[ret]->ins_code);
            if (!residue) {
                 ret = _torsionIndex.GoToNext();
                 continue;
            }

            serial_no++;
            pCard = _CardInfo[CISPEP_TOKEN].
                   add_a_new_card(NdbToken[CISPEP_TOKEN], 0);
            atom = residue->GetFirstAtom();
            pCard->AddElementToField(serial_no, 1);
            pCard->AddElementToField(residue->ResName(), 2);
            pCard->AddElementToField(_torsion[ret]->ChainID, 3);
            pCard->AddElementToField(atom->pdb_resnum(), 4);
            pCard->AddElementToField(atom->ins_code(), 5);
            pCard->AddElementToField(_torsion[ret]->ResName, 6);
            pCard->AddElementToField(_torsion[ret]->ChainID, 7);
            pCard->AddElementToField(_torsion[ret]->ResNum, 8);
            pCard->AddElementToField(_torsion[ret]->ins_code, 9);
            if (_num_mol > 1)
                 pCard->AddElementToField(_torsion[ret]->Mol_ID, 10);
            else pCard->AddElementToField(0, 10);
            if (fabs(_torsion[ret]->val) < 0.005)
                 pCard->AddElementToField(0.0, 11);
            else pCard->AddElementToField(_torsion[ret]->val, 11);

            ret = _torsionIndex.GoToNext();
       }
}

void _MAXIT::construct_remark_900()
{
       int max_length = 60;
       char buffer[62], *p;
       One_Card_Info_Format *pCard = _CardInfo[RENTRY_TOKEN-1].Head();

       One_Card_Info_Format *newCard = _CardInfo[REMARK_TOKEN-1].
                  add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
       newCard->AddElementToField(900, 1);
       newCard = _CardInfo[REMARK_TOKEN-1].add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
       newCard->AddElementToField(900, 1);
       newCard->AddElementToField("RELATED ENTRIES", 2);
       while (pCard) {
            newCard = _CardInfo[REMARK_TOKEN-1].
                    add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(900, 1);
            sprintf(buffer, "RELATED ID: %s   RELATED DB: %s",
                     pCard->Field(3), pCard->Field(2));
            ndb_set_string_to_upper(buffer);		  
            newCard->AddElementToField(buffer, 2);
            if (strcmp(pCard->Field(4), "")) {
                 pCard->CleanStringField(4);
                 p = get_max_length_word(pCard->Field(4), buffer, max_length);
                 newCard = _CardInfo[REMARK_TOKEN-1].
                         add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(900, 1);
                 ndb_set_string_to_upper(buffer);		  
                 newCard->AddElementToField(buffer, 2);
                 while (p) {
                      p = get_max_length_word(p, buffer, max_length);
                      newCard = _CardInfo[REMARK_TOKEN-1].
                              add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                      newCard->AddElementToField(900, 1);
                      ndb_set_string_to_upper(buffer);		  
                      newCard->AddElementToField(buffer, 2);
                 }
            }
            pCard = pCard->Next();
       }
}

void _MAXIT::check_atom_number_count()
{
       if (_num_mol == 0) return;

       int card_id = find_card_id("ATNUMS");
       One_Card_Info_Format *pCard = _CardInfo[card_id].Head();
       if (pCard == NULL) {
            pCard = _CardInfo[card_id].add_a_new_card(NdbToken[card_id], 0);
            int atom_no = _molecules[0]->count_atom_no(ATOMP_TOKEN);
            pCard->AddElementToField(atom_no, 2);

            atom_no = _molecules[0]->count_atom_no(ATOMN_TOKEN);
            pCard->AddElementToField(atom_no, 3);

            atom_no = 0;
            atom_no += _molecules[0]->count_atom_no(HETAD_TOKEN);
            atom_no += _molecules[0]->count_atom_no(ATOMMN_TOKEN);
            atom_no += _molecules[0]->count_atom_no(ATOMMP_TOKEN);
            atom_no += _molecules[0]->count_atom_no(HETAC_TOKEN);
            atom_no += _molecules[0]->count_atom_no(HETAI_TOKEN);
            atom_no += _molecules[0]->count_atom_no(HETIC_TOKEN);
            atom_no += _molecules[0]->count_atom_no(HETAIN_TOKEN);
            atom_no += _molecules[0]->count_atom_no(ATOMS_TOKEN);
            pCard->AddElementToField(atom_no, 4);

            atom_no = _molecules[0]->count_atom_no(HETAS_TOKEN);
            pCard->AddElementToField(atom_no, 5);
       }
       if (!strcmp(pCard->Field(1), ""))
            pCard->AddElementToField("LAST", 1);
       if (!strcmp(pCard->Field(6), "")) {
            int atom_no = atoi(pCard->Field(2)) + atoi(pCard->Field(3))
                        + atoi(pCard->Field(4)) + atoi(pCard->Field(5));
            pCard->AddElementToField(atom_no, 6);
       }
}

void _MAXIT::get_author_defined_remark_300_details(char *details)
{
       ReVarCifArray<CifString> remark_array;
       int width = NdbToken[REMARK_TOKEN-1].FieldList[2].FieldWidth;
       get_remark_array_from_block(remark_array, details, width);

       if (remark_array.Length() == 0) return;

       One_Card_Info_Format *newCard = NULL;
       for (unsigned int i = 0; i < remark_array.Length(); ++i) {
            newCard = _CardInfo[REMARK_TOKEN-1].add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(300, 1);
            newCard->AddElementToField(remark_array[i].Text(), 2);
       }
}

void separation_string_base_word(char *old_string, char *str1,
               char *str2, const char *word)
{
       str1[0] = '\0';
       str2[0] = '\0';

       if (old_string == NULL) return;

       char *p = strstr(old_string, word);
       if (!p) strcpy(str1, old_string);

       if (!strcmp(old_string, "") || p == NULL) return;

       char *q = old_string;
       while (q < p) {
            *str1 = *q;
            str1++; q++;
       }
       *str1 = '\0';

       p += strlen(word);

       strcpy(str2, p);
}

static void print_null_remark_value(One_Card_Info_Format *newCard,
         const int fieldLength, const int fieldJustification)
{
       char fieldFormat[82], buffer[82];
       if (fieldJustification == PDB_REMARK_LEFT_JUSTIFIED)
            sprintf(fieldFormat, "%%-%ds", fieldLength);
       else sprintf(fieldFormat, "%%%ds", fieldLength);
       sprintf(buffer, fieldFormat, "NULL");
       newCard->AttachToField(buffer, 2);
}

void print_space_between_remark_field(One_Card_Info_Format *newCard,
             REMARKS *Remark, const int field_no)
{
       char FieldFormat[82], buffer[82];
       if (field_no == 0) 
            sprintf(FieldFormat, "%%-%ds",
               Remark->FieldList[field_no].FieldStCol - 1);
       else sprintf(FieldFormat, "%%-%ds",
               Remark->FieldList[field_no].FieldStCol -
              (Remark->FieldList[field_no-1].FieldStCol +
               Remark->FieldList[field_no-1].FieldWidth));
       sprintf(buffer, FieldFormat, "");
       newCard->AttachToField(buffer, 2);
}

void get_remark_field_format(char *FieldFormat, REMARKS *Remark,
             const int field_no)
{
       if (Remark->FieldList[field_no].FieldType >=3) {
            if (Remark->FieldList[field_no].FieldJustification == 
                  PDB_REMARK_LEFT_JUSTIFIED) 
                 sprintf(FieldFormat, "%%-%ds",
                     Remark->FieldList[field_no].FieldWidth);
            else sprintf(FieldFormat, "%%%ds",
                     Remark->FieldList[field_no].FieldWidth);
       } else {
            if (Remark->FieldList[field_no].FieldJustification == 
                  PDB_REMARK_LEFT_JUSTIFIED)
                 sprintf(FieldFormat, "%%-%d.%dlf", 
                     Remark->FieldList[field_no].FieldWidth,
                     Remark->FieldList[field_no].FieldPrec);
            else sprintf(FieldFormat, "%%%d.%dlf", 
                     Remark->FieldList[field_no].FieldWidth,
                     Remark->FieldList[field_no].FieldPrec);
       }
}
/*
static void clean_descriptor(char *string, char *ec)
{
       ec[0] = '\0';
       char *q = ec;
       char *p = strstr(string, "PROTEIN (");
       if (p) strcpy(string, &string[9]);

       for (int i = strlen(string) - 1; i >= 0; i--) {
            if (string[i] == ')') { string[i] = '\0'; break; }
            else if (string[i] != ' ' && string[i] != '\0') break;
       }
       if ((p = strstr(string, "E.C."))) {
            if (*(p-1) == '(' && isdigit(*(p+4))) {
                *(p-1) = ' ';
                while (p) {
                     if (*p == ')') { *p = ' '; break; }
                     if (*p == '\0') break;
                     *q = *p; *p = ' ';
                     q++; p++;
                }
            }
            *q = '\0';
            strcpy(ec, &ec[4]);
       }
       ndb_clean_string(string);
}
*/
static void delete_space_between_string(char *string)
{
       int str_len = strlen(string);
       for (int i = str_len - 1; i > 0; i--) {
            if (string[i] == ' ' &&
                string[i-1] == ',')
                 strcpy(&string[i], &string[i+1]);
       }
}

static int check_align(const int Remark_No, REMARK_FIELD *FieldList)
{
       int align = 1;
       if (!strcmp(FieldList->TokenName, "XFILE1") && FieldList->FieldId == 3 &&
                   Remark_No != 210 ||
           !strcmp(FieldList->TokenName, "DTMEA1") && FieldList->FieldId == 2 ||
           !strcmp(FieldList->TokenName, "XFILE2") && FieldList->FieldId == 2 ||
           !strcmp(FieldList->TokenName, "CRMET1") && FieldList->FieldId == 3 ||
           !strcmp(FieldList->TokenName, "NMREMK") && FieldList->FieldId == 3 ||
           !strcmp(FieldList->TokenName, "REFREM") && FieldList->FieldId == 3 ||
           !strcmp(FieldList->TokenName, "CRDTLS") && FieldList->FieldId == 3 ||
           !strcmp(FieldList->TokenName, "CEMFIT") && FieldList->FieldId == 7 ||
           !strcmp(FieldList->TokenName, "CMRNST") && FieldList->FieldId == 7 ||
           !strcmp(FieldList->TokenName, "CMRNST") && FieldList->FieldId == 8 ||
           !strcmp(FieldList->TokenName, "EMSAMP") && FieldList->FieldId == 9 ||
           !strcmp(FieldList->TokenName, "EMDATA") && FieldList->FieldId == 19 ||
           !strcmp(FieldList->TokenName, "ASSEMB") && FieldList->FieldId == 7 ||
           !strcmp(FieldList->TokenName, "SOLMDL") && (FieldList->FieldId == 6 ||
                   FieldList->FieldId == 9))
            align = 0;
       if (Remark_No == 800) align = 0;
       return align;
}

static int biol_id_compare(const void *biol1, const void *biol2)
{
       return (atoi(((_BIOL_ID *) biol1)->biol_id) -
               atoi(((_BIOL_ID *) biol2)->biol_id));
}

static void get_remark_array_from_block(ReVarCifArray<CifString>& remark_array,
               const char *block_text, const int width)
{
       char *string_buff = NULL;
       remark_array.Clear();
       CifString cs;
       cs.Clear();
       int more_than_60 = 0;
       int len = strlen(block_text);
       for (int i = 0; i < len; i++) {
            if (block_text[i] == '\n') {
                 int len1 = cs.Length() + 1;
                 string_buff = new char[len1];
                 memset(string_buff, 0, len1);
                 strcpy(string_buff, cs.Text());
                 ndb_strip_trailing_blanks(string_buff);
                 if ((int) strlen(string_buff) > width) {
                      more_than_60 = 1;
                      delete [] string_buff;
                      break;
                 }
                 cs.Copy(string_buff);
                 delete [] string_buff;
                 remark_array.Add(cs);
                 cs.Clear();
            } else cs += block_text[i];
       }
       if (cs != "") {
            int len1 = cs.Length() + 1;
            string_buff = new char[len1];
            memset(string_buff, 0, len1);
            strcpy(string_buff, cs.Text());
            ndb_strip_trailing_blanks(string_buff);
            if ((int) strlen(string_buff) > width) more_than_60 = 1;
            cs.Copy(string_buff);
            delete [] string_buff;
            remark_array.Add(cs);
       }

       if (!more_than_60) return;

       remark_array.Clear();

       string_buff = new char[len + 1];
       memset(string_buff, 0, len + 1);
       strcpy(string_buff, block_text);
       ndb_clean_string(string_buff);
       char *p = string_buff;
       char *buffer = new char[len + 1];
       memset(buffer, 0, len + 1);
       int count = 0;
       while (p) {
            p = get_max_length_word(p, buffer, width);
            count++;
            if (!strcmp(buffer, "")) {
                 if (count < 10) continue;
                 cs.Copy(p);
                 remark_array.Add(cs);
                 break;
            }
            cs.Copy(buffer);
            remark_array.Add(cs);
       }
       delete [] string_buff;
       delete [] buffer;
}

char *get_words_separate_by_delimits_with_space(char *string, char *buffer, const char *delimit)
{
       *buffer = '\0';
       // while (*string == ' ' && *string != '\0') string++;
       if (*string == '\0') return NULL;
       char *p = strstr(string, delimit);
       if (p == NULL) {
            strcpy(buffer, string);
            // ndb_clean_string(buffer);
            return NULL;
       }

       char *begin = buffer;
       while (string != p) {
            *buffer = *string;
            buffer++; string++;
       }
       *buffer = '\0';
       // ndb_clean_string(begin);
       int len = strlen(begin);
       for (int i = len - 1; i >= 0; --i) {
            if (begin[i] == ' ') begin[i] = 0;
            else break;
       }
       string = p + strlen(delimit);
       // while (*string == ' ' && *string != '\0') string++;
       if (*string == '\0') return NULL;
       return string;
}

static void get_wordlist_from_string_separated_by_delimits_with_empty_line(
              ReVarCifArray<CifString> &wordlist, char *string_buff, const char *delimit)
{
       wordlist.Clear();
       char *buffer = new char[strlen(string_buff)+1];
       memset(buffer, 0, strlen(string_buff)+1);
       char *p = string_buff;
       CifString cs;
       int first = 1;
       while (p) {
            p = get_words_separate_by_delimits_with_space(p, buffer, delimit);
            if (!strcmp(buffer, "") && first) continue;
            first = 0;
            cs.Copy(buffer);
            wordlist.Add(cs);
       }
       delete [] buffer;
}

static bool has_value(ISTable *Table, const int num, _MOLECULE **mols)
{
       if (is_empty_table(Table)) return false;
       if (num == 1) return true;
       int model_index = Table->GetColumnIndex("PDB_model_num");
       CifString cs;
       get_value(cs, Table, model_index, 0);
       if (atoi(cs.Text()) == mols[0]->Mol_ID()) return true;
       return false;
}

static bool is_same_diffrn(char* single_diffrns, char* combine_diffrns)
{
       if (!strcmp(single_diffrns, combine_diffrns)) return true;

       ReVarCifArray<CifString> list, list1;
       get_wordlist_from_string_separated_by_delimits_with_empty_line(list,
                   combine_diffrns, ",");
       get_wordlist_from_string_separated_by_delimits_with_empty_line(list1,
                   single_diffrns, ",");
       for (unsigned int i = 0; i < list.Length(); ++i) {
            for (unsigned int j = 0; j < list1.Length(); ++j) {
                 if (list[i] == list1[j]) return true;
            }
       }

       return false;
}
